  (* la/art.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(** Define types and functions of abstract reachability tree & abstract reachability path. 
 *)

open LogicUtil;;
open LogicUtil.Logic;;
open Cfa;;
open La_common;;
(* module Log = ConfigUtil.Log;; *)

(* flag: need print and pause() to debug *)
let la_debug = ref false
let pause () = 
    begin 
        print_string ("\npause  -> press a number to continue: \n"); 
        let _ = read_int () in
        ()
    end
  
(* ~~~~~~~~~~~~~~~~~~~~ define type of art_node ~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
type art_stmt =
    | ART_FORM 
    | ART_COND 
    | ART_CALL 
    | ART_ASSUME 
    | ART_RESET
    | ART_ENTRY
    | ART_RETURN
    | ART_ERROR
    | ART_FAIL
    | ART_NORETURN

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ art_node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
type art_node = {
    art_type : art_stmt;
    mutable art_label : string * int;
    mutable art_outgoing_labels : (string * int) list;
    mutable art_outgoing_edges : predicate list;
    art_incoming_label : string * int;
    mutable art_reachable_region : predicate;
    mutable art_old_reachable_regions : predicate list;
    mutable art_support_predicates : predicate list;
    mutable art_weakest_precond : predicate;
    mutable art_strongest_postcond : predicate;
    mutable art_is_cover_node : bool;
    mutable art_covered_by : string * int;
    mutable art_is_pivot_node : bool;
    (*mutable art_dependent_labels : (string * int) list;*)
    mutable art_trust : bool;
    mutable art_memo : (string * string * string * string * string) list;
}

(* ~~~~~~~~~~~~~~~~~~~~~~~ abstract-reachability-tree ~~~~~~~~~~~~~~~~~~~~ *)
type art = {
    (* Hashtbl.t contains all art_nodes in art *)
    art_node_hash : ((string * int), art_node) Hashtbl.t;
    (* Hashtbl contains versions & reachable_region of same-named cfa_nodes *)
    art_version_hash : (string, (int * predicate)) Hashtbl.t;
    art_trust_nodes: (string * int, art_node) Hashtbl.t;
    art_label_hash : (string, int) Hashtbl.t;
    (* List contains all of art_nodes in error_path (if art has error_path) *)
    mutable art_labels_in_error_path : (string * int) list;
    (* Hash stores maximum indexes of cfa node in art *)
    art_max_indexes : (string, int) Hashtbl.t;
    (* label of root of art *)
    art_label_of_root : string * int;
    mutable art_queue : ((string * int) (* art label *)
                         * int (* branch count *)
                         * Cfa.cfa_fundef_t (* cfa that contains this node *)
                         * (int * cfa_node * cfa_fundef_t) list (* call stack*)) list;
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~path~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This struct is used to save path in art                                 *)
type art_path = {
    mutable art_nodes_in_path : art_node list;
    mutable art_edges_in_path : predicate list;
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_art_stmt (a_art_stmt : art_stmt) : string =
begin
    match a_art_stmt with
    | ART_FORM -> "ART_FORM"
    | ART_COND -> "ART_COND"
    | ART_CALL -> "ART_CALL"
    | ART_ASSUME -> "ART_ASSUME"
    | ART_RESET -> "ART_RESET"
    | ART_ENTRY -> "ART_ENTRY"
    | ART_RETURN -> "ART_RETURN"
    | ART_ERROR -> "ART_ERROR"
    | ART_FAIL -> "ART_FAIL"
    | ART_NORETURN -> "ART_NORETURN"
end

(* ~~~~~~~~~~~~~~~~~~ transform cfa_stmt to art_stmt ~~~~~~~~~~~~~~~~~~~~~ *)
let transform_cfa_stmt_to_art_stmt_type (c_node : cfa_node) =
begin
    match c_node.cfa_node_stmt with
    | CFA_FORM f -> ART_FORM
    | CFA_COND c -> ART_COND
    | CFA_CALL (c, _) -> ART_CALL
    | CFA_ASSUME a -> ART_ASSUME
    | CFA_RESET r -> ART_RESET
    | CFA_ENTRY f -> ART_ENTRY
    | CFA_RETURN -> ART_RETURN
    | CFA_ERROR -> ART_ERROR
    | CFA_FAIL -> ART_FAIL
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_new_art_node (a_type : art_stmt) 
                        (label : string * int)
                        (outgoing_labels : (string * int) list)
                        (outgoing_edges : predicate list)
                        (incoming_label : string * int)
                        (reachable_region : predicate)
                        (old_reachable_regions : predicate list)
                        (support_predicates : predicate list)
                        (covered_node : bool)
                        (is_pivot_node : bool)
                        (*dependent_labels : (string * int) list*)
                        (trust : bool)
                        (memo : (string * string * string * string * string) list (*string*))
: art_node =
begin
    let new_art_node : art_node = {
        art_type = a_type;
        art_label = label;
        art_outgoing_labels = outgoing_labels;
        art_outgoing_edges = outgoing_edges;
        art_incoming_label = incoming_label;
        art_reachable_region = reachable_region;
        art_old_reachable_regions = old_reachable_regions;
        art_support_predicates = support_predicates;
        art_weakest_precond = True;
        art_strongest_postcond = True;
        art_is_cover_node = covered_node;
        art_is_pivot_node = is_pivot_node;
	art_covered_by = ("", -1);
        (*art_dependent_labels = dependent_labels;*)
        art_trust = trust;
        art_memo = memo;
    } in
    new_art_node
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_art_label ((name, version) : string * int) : string =
begin
    name ^ "#" ^ (string_of_int version)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let art_label_to_cfa_label ((name, version) : string * int) : string = name

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* input: label-of-art-node1, label-of-art-node2 output: 0 if label1 =     *)
(* label2; 1 otherwise                                                     *)
let compare_art_labels (name1, version1) (name2, version2) : int =
begin
    if (String.compare name1 name2) = 0 then begin
        if version1 = version2 then 0
        else 1
    end else 1
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_anode_in_path (nodes_in_path : art_node list) (alabel : string * int) 
: art_node option = begin
    match nodes_in_path with
    | [] -> None
    | [anode] -> (if (compare_art_labels alabel anode.art_label) = 0 then Some anode
                  else None)
    | head :: tail -> (if (compare_art_labels alabel head.art_label) = 0 then Some head
                       else (find_anode_in_path tail alabel))
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let is_mem_of_art_labels (alabel : string * int)
                         (art_labels : (string * int) list)
: bool = 
begin
    let is_mem : bool ref = ref false in
    List.iter
        (fun l -> 
            if !is_mem = false then begin
                if (compare_art_labels alabel l) = 0 then is_mem := true
            end     
        )
        art_labels;
    (* Return *)
    !is_mem  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in art_dumper.ml to check if an edge is in        *)
(* error_path or not => to color the error_path                            *)
let is_in_error_path (a_art : art) (a_node1: art_node) (a_node2 : art_node)
: bool =
begin
    let flag1 = ref false in
    let flag2 = ref false in
    (* Check if node1 & node2 are in error_path or not *)
    List.iter
        (fun a_label_in_error_path ->
            if !flag1 = false then begin
                if a_label_in_error_path = a_node1.art_label then
                    flag1 := true;
            end;
            if !flag2 = false then begin
                if a_label_in_error_path = a_node2.art_label then
                    flag2 := true;
            end
        )
        a_art.art_labels_in_error_path;
    (* return *)
    (!flag1 & !flag2)
end

(* ~~~~~~~~~~ decide to remove a version in art or not ~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in lazyabstraction.ml when refine. If an art_node *)
(* is completed and doesn't depend on any art_node in current error_path,  *)
(* we do not remove it                                                     *)
let remove_a_version_in_art (an_art : art) 
                            ((art_name, version_to_remove) : string * int) 
                            : unit =
begin
    (* coppy versions & reachable_regions of an art_name in an_art *)
    let versions_list = Hashtbl.find_all (an_art.art_version_hash) (art_name) in
    (* remove values of key "art_name" in an_art.art_version_hash *)
    La_common.remove_elements_in_hashtbl an_art.art_version_hash art_name;
    (* add versions *)
    List.iter
        (fun (version, pred) ->
            if version != version_to_remove then begin
                Hashtbl.add (an_art.art_version_hash) (art_name) (version, pred)
            end else begin
                (* check if current_art_node is completed or not *)
                let is_trust =
                    Hashtbl.mem (an_art.art_trust_nodes) (art_name, version) in
                if is_trust = true then begin
                    (* find current_art_node *)
                    let cr_node =
                        try Hashtbl.find (an_art.art_trust_nodes) (art_name, version)
                        with Not_found -> assert false 
                    in
                    Hashtbl.add (an_art.art_version_hash)
                                (art_name)
                                (-version, pred);
                    Hashtbl.remove (an_art.art_trust_nodes)
                                   (art_name, version);
                    cr_node.art_label <- (art_name, -version);
                    Hashtbl.add (an_art.art_trust_nodes)
                                (art_name, -version)
                                cr_node
                end
            end
        )
        versions_list;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_new_art (length_of_hashtbl : int) (label_of_root : string * int)
: art =
begin
    let new_art : art = {
        art_node_hash = Hashtbl.create (length_of_hashtbl);
        art_version_hash = Hashtbl.create (length_of_hashtbl);
        art_trust_nodes = Hashtbl.create (length_of_hashtbl);
        art_label_hash = Hashtbl.create (length_of_hashtbl);
        art_labels_in_error_path = [];
        art_label_of_root = label_of_root;
        art_max_indexes = Hashtbl.create length_of_hashtbl;
        art_queue = [];
    } in
    (* return *)
    new_art
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let find_max_version_of_cfa_label_in_art (an_art : art)
                                         (cfa_label_of_art_node : string) 
: int =
begin
    (* find all versions of the cfa_label in an_art *)
    let all_versions : (int * predicate) list =
        Hashtbl.find_all (an_art.art_version_hash) (cfa_label_of_art_node)
    in
    let max_version : int ref = ref (0) in
    List.iter
        (fun (a_version, a_reachable_region) ->
            if a_version > !max_version then max_version := a_version
        )
        all_versions;
    (* return *)
    !max_version
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_max_version_of_cfa_label_in_art (an_art : art)
                                         (cfa_label_of_art_node : string) 
: int =
begin
    (* find all indexes of the cfa_label in an_art *)
    let all_indexes: int list =
        Hashtbl.find_all (an_art.art_label_hash) (cfa_label_of_art_node)
    in
    let max_index : int ref = ref (0) in
    List.iter
        (fun a_index->
            if a_index > !max_index then max_index := a_index;
        )
        all_indexes;
    (* return *)
    !max_index
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let copy_art_node (a_node : art_node) : art_node =
begin
    let new_art_node : art_node = {
        art_type = a_node.art_type;
        art_label = a_node.art_label;
        art_outgoing_labels = a_node.art_outgoing_labels;
        art_outgoing_edges = a_node.art_outgoing_edges;
        art_incoming_label = a_node.art_incoming_label;
        art_reachable_region = a_node.art_reachable_region;
        art_old_reachable_regions = a_node.art_old_reachable_regions;
        art_support_predicates = a_node.art_support_predicates;
        art_weakest_precond = a_node.art_weakest_precond;
        art_strongest_postcond = a_node.art_strongest_postcond;
        art_is_cover_node = a_node.art_is_cover_node;
        art_is_pivot_node = a_node.art_is_pivot_node;
        art_covered_by = a_node.art_covered_by;	
        (*art_dependent_labels = a_node.art_dependent_labels;*)
        art_trust = a_node.art_trust;
        art_memo = a_node.art_memo;
    } in
    (* return *)
    new_art_node
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used to create temp_art in lazyabstraction.ml          *)
let copy_art (an_art : art) : art =
begin
    (* create new art to copy *)
    let length : int = Hashtbl.length an_art.art_node_hash in
    let new_art = create_new_art (length) (an_art.art_label_of_root) in
    (* start copy *)
    Hashtbl.iter
        (fun a_label an_art_node ->
            let cp_art_node : art_node = copy_art_node (an_art_node) in
            Hashtbl.add (new_art.art_node_hash) (a_label) (cp_art_node)
        )
        an_art.art_node_hash;
    (* return *)
    new_art
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let copy_art_path (apath : art_path) : art_path = 
begin
    let cp_apath = {
        art_nodes_in_path = apath.art_nodes_in_path;
        art_edges_in_path = apath.art_edges_in_path;
    } in
    cp_apath    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* check if an art_node is covered by another art_node or not              *)
let is_covered (an_art : art) 
               (current_rr : predicate)
               (checking_clabel : string)
: bool =
begin
    (* find versions of all same_labed art_nodes to check *)
    let all_art_versions : (int * predicate) list =
        Hashtbl.find_all (an_art.art_version_hash) (checking_clabel)
    in
    (* flag to return *)
    let flag = ref false in
    (* with each version, check if it covers current node or not *)
    List.iter
        (fun (old_version, old_rr) ->
            if !flag = false then begin
                let fol_implication = Or ((Not current_rr) :: [old_rr]) in
                let valid = Logic.is_valid (fol_implication) in
                if valid = true then flag := true
            end
        )
        all_art_versions;
    (* return *)
    !flag
end

let rec remove_node_and_childs (art : art) (art_node_label : string * int)
: unit = (
    let art_node = try Hashtbl.find art.art_node_hash art_node_label
                   with Not_found -> assert false
    in
    let labels_of_childs = art_node.art_outgoing_labels in
    Hashtbl.remove art.art_node_hash art_node_label;
    art.art_queue <-
        List.filter (fun (label, _, _, _) ->
            (compare_art_labels label art_node_label) <> 0
        ) art.art_queue;
    art.art_labels_in_error_path <-
        List.filter (fun label ->
            (compare_art_labels label art_node_label) <> 0
        ) art.art_labels_in_error_path;
    List.iter (remove_node_and_childs art) labels_of_childs;
)

let remove_childrens_of_node (art : art) (art_node : art_node)
: unit = (
    let labels_of_childs = art_node.art_outgoing_labels in
    art_node.art_outgoing_labels <- [];
    art_node.art_outgoing_edges <- [];
    Hashtbl.replace art.art_node_hash art_node.art_label art_node;
    List.iter (remove_node_and_childs art) labels_of_childs;
    (* check *)
    List.iter (fun (label, _, _, _) ->
        if Hashtbl.mem art.art_node_hash label = false then
        assert false;
        ()
    ) art.art_queue;
)

let rec is_ancestor (art : art)
                    (art_node_label1 : string * int)
                    (art_node_label2 : string * int)
: bool = (
    try
        let art_node1 =
            Hashtbl.find art.art_node_hash art_node_label2
        in
        if (compare_art_labels art_node_label1 art_node1.art_incoming_label) = 0 then
            true
        else
            is_ancestor art art_node_label1 art_node1.art_incoming_label
    with Not_found -> false
)

(** check cover of node checking_label that having rr is current_rr
    @param art : art
    @param current_rr : reachable region of art_node checking_label
    @param checking_label : art_label of node is being reviewed
*)
let covered_by1 (art : art)
               (current_rr : predicate)
               (checking_label : string * int)
: string * int = (
    let checking_clabel, checking_version = checking_label in
    (* Find rrs of all art_node that have cfa_label as well as checking_label *)
    let labels_and_rrs =
        Hashtbl.fold (fun (clabel, version) art_node labels_and_preds ->
            if (String.compare clabel checking_clabel) = 0
                & (version <> checking_version)
                & (is_ancestor art checking_label (clabel, version)) = false then
                labels_and_preds
                @ [((clabel, version), art_node.art_old_reachable_regions)]
            else labels_and_preds
        ) art.art_node_hash [] in
    (* check cover *)
    let rec check_cover (current_rr : predicate)
                        (labels_and_rrs : ((string * int) * (predicate list)) list)
    : string * int = (
        match labels_and_rrs with
        | [] -> ("", -1)
        | (label, rrs) :: remain_labels_and_rrs -> (
            let is_cover =
                match rrs with
                | [] -> false
                | _ -> is_valid (Or [Not current_rr; And rrs])
            in
            if is_cover then (
                label)
            else check_cover current_rr remain_labels_and_rrs
          )
    )
    in
    check_cover current_rr labels_and_rrs;
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Now, replace "is_covered" with "covered_by" in the following            *)
let covered_by (an_art : art) 
               (current_rr : predicate)
               (checking_clabel : string)
: string * int =
begin
    (* find versions of all same_labed art_nodes to check *)
    let all_art_versions : (int * predicate) list =
        Hashtbl.find_all (an_art.art_version_hash) (checking_clabel)
    in
    (* 19.01.2009. reverse order of all_art_versions to find covered_node *)
    (* Because we want to get the minimum-version of art_node covered     *)
    (* this node.                                                         *) 
    let rev_all_art_versions = List.rev all_art_versions in

    (* create label to keep label of art_node which covered current node *)
    let covering_label = ref ("", -1) in
    (* flag to check *)
    let flag = ref false in
    (* with each version, check if it covered current_node or not *)
    List.iter
        (fun (old_version, old_rr) ->
            if ((current_rr <> True) || (old_rr <> True)) then
                if !flag = false then
                begin
                    let fol_implication = Or ((Not current_rr) :: [old_rr]) in
                    let valid = is_valid (fol_implication) in
                    if valid = true then begin
                        flag := true;
                        covering_label := (checking_clabel, old_version)
                    end
                end
        )
        rev_all_art_versions (*all_art_versions*);

    (* return *)
    !covering_label
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_art_labels_in_path (path : art_path) : (string * int) list =
begin
    let list_to_return = ref [] in
    List.iter
        (fun a_art_node ->
            list_to_return := !list_to_return @ [a_art_node.art_label])
        path.art_nodes_in_path;
    !list_to_return
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_new_art_path () =
begin
    let nil_art_path : art_path = {
        art_nodes_in_path = [];
        art_edges_in_path = [];
    } in
    (* return *)
    nil_art_path
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_based_vars_in_path (apath : art_path) : art_path = 
begin
    let new_art_edges_in_path : predicate list ref = ref [] in
    (* Foreach edge in path, create based_vars for it *)
    List.iter
        (fun an_edge ->
            let new_edge : predicate = 
                create_based_vars_in_predicate (*false*) (an_edge) in
            new_art_edges_in_path := !new_art_edges_in_path @ [new_edge]
        )
        apath.art_edges_in_path;
    (* Create new_path to return *)
    let new_path : art_path = {
        art_nodes_in_path = apath.art_nodes_in_path;
        art_edges_in_path = !new_art_edges_in_path;
    } in
    new_path  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* get string of labels in path. This function is used to debug            *)
let get_string_of_labels_in_art_path (a_path : art_path) : string =
begin
    (* create var to return *)
    let string_of_art_labels = ref "" in
    (* with each node in path, we get string of art_label of it *)
    List.iter
        (fun a_node_in_path ->
            let new_s = string_of_art_label (a_node_in_path.art_label) in
            string_of_art_labels := !string_of_art_labels ^ "  &  " ^ new_s;
        )
        a_path.art_nodes_in_path;
    (* return *)
    !string_of_art_labels
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)    
let change_index_of_variables_of_a_call_block 
      (index_hash : (expression, int) Hashtbl.t) 
      (int_hash : (int, (int list) * int) Hashtbl.t) 
      (var_to_index : (expression, int) Hashtbl.t) 
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref) 
      (error_path_formulars : predicate list ref) 
      (vars_in_fun_stack : expression list Stack.t) 
      (gvars : expression list)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (anode : art_node) 
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (edge : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t) 
: unit = 
begin
    (* save information of vars of function *)
    Stack.push (!current_vars_in_fun) (vars_in_fun_stack);
    (* create vars_in_fun_hash of new function call *)
    let vars_in_fun_hash : expression list ref = ref [] in
    current_vars_in_fun := !vars_in_fun_hash;
    (* change index *)
    let a_new_edge =
        change_index_of_variables_in_non_condition_predicate
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (is_flatten) 
          (array_asgn_is_flatten) (elements_of_arrays) 
          (check_pointer) (null_pointer_exceptions) (var_eq_hash) (edge)
          (uniqueness_hash)
    in
    (* update information: *)
    error_path_formulars := !error_path_formulars @ [a_new_edge];
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let rec remove_local_vars (int_hash : (int, (int list) * int) Hashtbl.t) 
      (var_to_index : (expression, int) Hashtbl.t) 
      (index_to_var : (int, expression) Hashtbl.t)  
      (current_vars_in_fun : expression list ref) 
      (gvars : expression list) 
: unit =
begin
    let rec_flag : bool ref = ref false in
    let new_current_vars_in_fun : expression list ref = ref [] in
    (* Remove all not-pointed vars *)
    List.iter
        (fun a_var_expr -> 
            (* debug *)
            (*debug ("remove local vars: " ^ (string_of_expression a_var_expr));*)
            let a_based_var_expr : expression = 
                create_based_vars_in_expression (*false*) (a_var_expr) in
            let is_gvar : bool = is_mem_of_expr_list (a_based_var_expr) (gvars) in
            if is_gvar = false then begin
                (* Find index of var_expr *)
                let index_of_var_expr = try Hashtbl.find (var_to_index) (a_var_expr)
                                        with Not_found -> - 1
                in
                if index_of_var_expr >= 0 then begin
                    let (points_to_expr, expr_points_to) = 
                        try Hashtbl.find (int_hash) (index_of_var_expr)
                        with Not_found -> assert false
                    in
                    (* Check if exists points_to_expr or not *)
                    if points_to_expr = [] then begin
                        (* Remove a_var_expr *)
                        Hashtbl.remove (int_hash) (index_of_var_expr);
                        Hashtbl.remove (var_to_index) (a_var_expr);
                        Hashtbl.remove (index_to_var) (index_of_var_expr);
                        (* Remove information of a_var_expr in where it points to (called c) *)
                        (if expr_points_to >= 0 then begin
                            let ((points_to_c : int list), (c_points_to : int)) = 
                                try Hashtbl.find (int_hash) (expr_points_to)
                                with Not_found -> assert false
                            in  
                            let new_points_to_c : int list = remove_mem (index_of_var_expr) (points_to_c) in
                            Hashtbl.replace (int_hash) (expr_points_to) (new_points_to_c, c_points_to);
                        end); 
                        (* Set rec_flag to call remove_local_vars *)
                        rec_flag := true;
                    (* If exists, save to *)
                    end else begin
                        new_current_vars_in_fun := !new_current_vars_in_fun @ [a_var_expr]
                    end                    
                end else begin
                    new_current_vars_in_fun := !new_current_vars_in_fun @ [a_var_expr]
                end 
            end else new_current_vars_in_fun := !new_current_vars_in_fun @ [a_var_expr]
        )
        !current_vars_in_fun;
    (* Check rec_lag *)
    if !rec_flag = true then begin
        remove_local_vars (int_hash) (var_to_index) (index_to_var)
          (new_current_vars_in_fun) (gvars) 
    end else ()
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_variables_of_a_return_block
      (index_hash : (expression, int) Hashtbl.t) 
      (int_hash : (int, (int list) * int) Hashtbl.t) 
      (var_to_index : (expression, int) Hashtbl.t) 
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref) 
      (error_path_formulars : predicate list ref) 
      (vars_in_fun_stack : expression list Stack.t) 
      (gvars : expression list)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (anode : art_node) 
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (edge : predicate) 
      (uniqueness_hash : (int, expression) Hashtbl.t)
: unit = 
begin
    (* debug: *)
    (*debug ("edge: " ^ (string_of_predicate edge));
    read_int();
    debug ("current_vars_in_fun: " ^ string_of_expression (Sum !current_vars_in_fun));*)
    (* Remove outgoing from current_vars_in_fun: *)
    (match edge with
    | And [Eq (left_expr, right_expr)] | Eq (left_expr, right_expr) -> begin
        (* Change index of right_expr *)
        let new_right_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
            (int_hash) (var_to_index) (index_to_var) (structs_hash)
            (ref []) (right_expr) (uniqueness_hash)
        in
        (* Remove right_expr: *)
        current_vars_in_fun := 
            remove_expr_from_expr_list (new_right_expr) (!current_vars_in_fun);
      end
    | _ -> (*debug (string_of_predicate edge); assert false*) ());
    (* remove local-variables *)
    (*remove_local_vars (int_hash) (var_to_index) (index_to_var)
      (current_vars_in_fun) (gvars);*)
    (* updates current_vars_in_fun *)
    current_vars_in_fun := (try Stack.pop (vars_in_fun_stack)
                           with
                           | Stack.Empty -> assert false);
    (* change index *)
    let a_new_edge =
        change_index_of_variables_in_non_condition_predicate
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (is_flatten) 
          (array_asgn_is_flatten) (elements_of_arrays) 
          (check_pointer) (null_pointer_exceptions) (var_eq_hash) (edge)
          (uniqueness_hash)
    in
    (* update informations: *)
    error_path_formulars := !error_path_formulars @ [a_new_edge]; 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_variables_of_a_block 
      (index_hash : (expression, int) Hashtbl.t) 
      (int_hash : (int, (int list) * int) Hashtbl.t) 
      (var_to_index : (expression, int) Hashtbl.t) 
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref) 
      (error_path_formulars : predicate list ref) 
      (vars_in_fun_stack : expression list Stack.t) 
      (gvars : expression list)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (temp_reset_constraint : predicate ref) 
      (is_reset : bool ref)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (anode : art_node) 
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (edge : predicate) 
      (uniqueness_hash : (int, expression) Hashtbl.t)
: unit = 
begin
    (*try ( *)
        (* check type of anode *)
        match anode.art_type with
        | ART_RESET -> begin
            let a_new_edge : predicate =
                change_index_of_variables_in_reset (index_hash)
                  (int_hash) (var_to_index) (index_to_var) (structs_hash)
                  (current_vars_in_fun) (elements_of_arrays) (is_flatten) 
                  (check_pointer) (null_pointer_exceptions) (edge)
                  (uniqueness_hash)
            in
            (* update information: *)
            error_path_formulars := !error_path_formulars @ [a_new_edge];
            temp_reset_constraint := True;
            is_reset := true;
          end
        | ART_ASSUME -> begin
            let a_new_edge : predicate =
                change_index_of_variables_in_assume (index_hash)
                  (int_hash) (var_to_index) (index_to_var) (structs_hash)
                  (current_vars_in_fun) (elements_of_arrays) (is_flatten) 
                  (check_pointer) (null_pointer_exceptions) (edge)
                  (uniqueness_hash)
            in
            (* Check function pointer in assume *)
            (*let _ = 
                check_pointers_in_predicate (int_hash) (var_to_index)
                  (index_to_var) (index_hash) (a_new_edge) in*)
            (* update information: *)
            if !is_reset = true then begin
                (* Note that edge of ART_ASSUME has type: *)
                (* And[pred_of_assume; pred_of_reset]     *)
                let final_edge : predicate = 
                    And [a_new_edge; !temp_reset_constraint] in
                error_path_formulars := !error_path_formulars @ [final_edge];
                is_reset := false;
                temp_reset_constraint := True;                  
            end else begin
                let final_edge : predicate = And [a_new_edge; True] in
                error_path_formulars := !error_path_formulars @ [final_edge]
            end
          end
        | ART_COND -> begin
            let a_new_edge : predicate =
                change_index_of_variables_in_condition (index_hash) 
                  (int_hash) (var_to_index) (index_to_var) (structs_hash) 
                  (current_vars_in_fun) (elements_of_arrays) (false) 
                  (check_pointer) (null_pointer_exceptions) (edge)
                  (uniqueness_hash)
            in
            (* Check function pointer in condition *)
            (*let _ = 
                check_pointers_in_predicate (int_hash) (var_to_index)
                  (index_to_var) (index_hash) (a_new_edge) in*)
            (* update information: *)
            error_path_formulars := !error_path_formulars @ [a_new_edge]
          end
        | ART_CALL -> 
            change_index_of_variables_of_a_call_block (index_hash) (int_hash)
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
              (error_path_formulars) (vars_in_fun_stack) (gvars)
              (elements_of_arrays) (is_flatten) (array_asgn_is_flatten) 
              (check_pointer) (null_pointer_exceptions) (anode) (var_eq_hash) (edge)
              (uniqueness_hash)
        | ART_RETURN -> 
            change_index_of_variables_of_a_return_block (index_hash) (int_hash)
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
              (error_path_formulars) (vars_in_fun_stack) (gvars)
              (elements_of_arrays) (is_flatten) (array_asgn_is_flatten) 
              (check_pointer) (null_pointer_exceptions) (anode) (var_eq_hash) (edge)
              (uniqueness_hash)
        | ART_ENTRY | ART_FORM -> begin
            let a_new_edge =
                change_index_of_variables_in_non_condition_predicate
                  (index_hash) (int_hash) (var_to_index) (index_to_var) 
                  (structs_hash) (current_vars_in_fun) (is_flatten) 
                  (array_asgn_is_flatten) (elements_of_arrays) 
                  (check_pointer) (null_pointer_exceptions) (var_eq_hash) (edge)
                  (uniqueness_hash)
            in
            error_path_formulars := !error_path_formulars @ [a_new_edge];
          end
        | anything_else -> (
            if edge = True then
                error_path_formulars := !error_path_formulars @ [True]
            else assert false
          )
    (* Catch exception: *)
    (* ) with La_exception.Not_point_to (_, e1, e2, need_addrof_exprs) -> 
        raise (La_exception.Not_point_to (anode.art_label, e1, e2, need_addrof_exprs)) *)
end

let add_check_uniqueness (check_uniqueness : expression list)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
      (error_path_formulars : predicate list)
: (expression * expression * expression list) list = (
    let empty_exprs : expression list = [] in
    List.fold_left (fun added_uniqueness expr ->
        let exprs_checked_uniqueness =
            Hashtbl.find_all var_eq_hash expr in
        let new_added_uniqueness =
            List.fold_left (fun added_uniqueness expr ->
                match expr with
                | Sum (a_expr :: remain_expr_list) -> (
                  try
                    let index =
                        Hashtbl.find var_to_index a_expr
                    in
                    let size =
                        Hashtbl.find uniqueness_hash index in
                    let pred = Lt ((Sum remain_expr_list), size) in
                    let add_uniqueness_to_null_pointer =
                        if (LogicUtil.Logic.is_sat pred) = false then true
                        else (
                            if (LogicUtil.Logic.is_valid pred) then false
                            else (
                                not (LogicUtil.Logic.is_valid
                                        (Or [Not (And error_path_formulars); pred]))) 
                    ) in
                    if add_uniqueness_to_null_pointer then (
                        added_uniqueness
                        @ [(La_common.create_based_vars_in_expression expr,
                            expr,
                            empty_exprs)])
                    else added_uniqueness

                  with Not_found -> added_uniqueness
                )
                | _ -> added_uniqueness;
            ) [] exprs_checked_uniqueness
        in
        added_uniqueness @ new_added_uniqueness;
    ) [] check_uniqueness;
)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_variables_in_path (path : art_path)
     (gvars : expression list)
     (is_flatten : bool)
     (array_asgn_is_flatten : bool)
     (check_pointer : bool)
     (null_pointer_exceptions : (expression * expression * (expression list)) list list ref)
: art_path =
begin
    (* Step1: create some vars *)
    let int_hash : (int, (int list) * int) Hashtbl.t = Hashtbl.create (0) in
    let var_to_index : (expression, int) Hashtbl.t = Hashtbl.create (0) in
    let index_to_var : (int, expression) Hashtbl.t = Hashtbl.create (0) in
    let error_path_formulars = ref [] in
    let index_hash : (expression, int) Hashtbl.t = Hashtbl.create (0) in
    let var_eq_hash : (expression, expression) Hashtbl.t = Hashtbl.create (0) in
    let structs_hash : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let vars_in_fun_stack = Stack.create () in
    let current_vars_in_fun : expression list ref = ref [] in
    let elements_of_arrays : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let uniqueness_hash : (int, (* inde of pointer variable *)
                           expression (* size of array *)) Hashtbl.t =
        Hashtbl.create 0 in
    (* Create temp_reset_constraints *)
    let temp_reset_constraint : predicate ref = ref True in
    let is_reset : bool ref  = ref false in
    (* Step2: foreach block, change index of vars for it: *)
    (try List.iter2 
        (fun anode an_edge -> 
            (* refresh check_uniqueness *)
            let exceptions_of_edge = ref [] in
            change_index_of_variables_of_a_block (index_hash) (int_hash) 
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun) 
              (error_path_formulars) (vars_in_fun_stack) (gvars) 
              (elements_of_arrays) (is_flatten) (array_asgn_is_flatten : bool)
              (temp_reset_constraint) (is_reset) (check_pointer) 
              (exceptions_of_edge) (anode) var_eq_hash (an_edge)
              (uniqueness_hash);
            let added_uniqueness = add_check_uniqueness !Cfa.check_uniqueness
                                                        var_eq_hash
                                                        var_to_index
                                                        uniqueness_hash
                                                        !error_path_formulars
            in
            exceptions_of_edge := !exceptions_of_edge @ added_uniqueness;
            (* Update to exceptions in path *)
            if check_pointer = true then begin 
                null_pointer_exceptions := !null_pointer_exceptions @ [!exceptions_of_edge];
            end;
        )
        path.art_nodes_in_path
        path.art_edges_in_path;
        (* Step3: create new_path & return *)
        let new_path : art_path = {
            art_nodes_in_path = path.art_nodes_in_path;
            art_edges_in_path = !error_path_formulars;
        } in
        (* Return *)
        new_path
    with
    | Invalid_argument _ -> assert false);
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec encode_vars_in_application (app : expression)
                                   (encoding_hash : (expression, expression) Hashtbl.t)
: expression =
begin
    match app with
    | Application ("addrof", _ (*[_]*))
    | Application ("pref", _ (*[_; _]*))
    | Application ("sref", _ (*[_; _; _]*))
    | Application ("srefp", _ (*[_; _; _]*)) -> begin
        let encoded_application = "ec" ^ (string_of_expression app) in
        let _ =
            (try Hashtbl.find (encoding_hash) (Variable encoded_application)
            with
            | Not_found ->
                Hashtbl.add encoding_hash (Variable encoded_application) app;
                app)
        in
        (* return *)
        Variable encoded_application
      end
    | Application (s, expr_list) -> begin
        let encoded_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let an_encoded_expr =
                    encode_vars_in_expression (an_expr) (encoding_hash) in
                encoded_expr_list := !encoded_expr_list @ [an_encoded_expr])
            expr_list;
        (* return *)
        Application (s, !encoded_expr_list)
      end
    | anything_else -> assert false
  end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and encode_vars_in_expression (expr : expression)
                              (encoding_hash : (expression, expression) Hashtbl.t)
: expression =
begin
    (* Simplify expr *)
    let simplified_expr : expression = simplify_expr expr in
    match simplified_expr with
    | Constant f -> Constant f
    | Variable s -> Variable s
    | Application (s, expr_list) ->
        encode_vars_in_application (simplified_expr) (encoding_hash)
    | Sum expr_list -> begin
        let encoded_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let encoded_expr =
                    encode_vars_in_expression (an_expr) (encoding_hash) in
                encoded_expr_list := !encoded_expr_list @ [encoded_expr])
            expr_list;
        (* return *)
        Sum !encoded_expr_list
      end
    | Coeff (f, an_expr) -> begin
        (* debug debug "Coeff"; *)
        let encoded_expr = 
            encode_vars_in_expression (an_expr) (encoding_hash) in
        (* return *)
        Coeff (f, encoded_expr)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let encode_vars_in_expressions (exprs : expression list)
                               (encoding_hash : (expression, expression) Hashtbl.t)
: expression list = 
begin
    let encoded_exprs : expression list ref = ref [] in
    List.iter
        (fun e -> 
            let encoded_e : expression = 
                encode_vars_in_expression e encoding_hash in
            encoded_exprs := !encoded_exprs @ [encoded_e]
        )
        exprs;
    !encoded_exprs;
end     
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec encode_vars_in_predicate (an_edge : predicate)
                                 (encoding_hash : (expression, expression) Hashtbl.t)
: predicate =
begin
    match an_edge with
    | True -> True
    | False -> False
    | And pred_list -> begin
        let encoded_pred_list = ref [] in
        List.iter
            (fun a_pred ->
                let encoded_pred = 
                    encode_vars_in_predicate (a_pred) (encoding_hash) in
                encoded_pred_list := !encoded_pred_list @ [encoded_pred]
            )
            pred_list;
        (* return *)
        And (!encoded_pred_list)
      end
    | Or pred_list -> begin
        let encoded_pred_list = ref [] in
        List.iter
            (fun a_pred ->
                let encoded_pred =
                    encode_vars_in_predicate (a_pred) (encoding_hash) in
                encoded_pred_list := !encoded_pred_list @ [encoded_pred]
            )
            pred_list;
        (* return *)
        Or (!encoded_pred_list)
      end
    | Not pred -> begin
        let encoded_pred = encode_vars_in_predicate (pred) (encoding_hash) in
        (* return *)
        Not encoded_pred
      end
    | Eq (expr1, expr2) -> begin
        (* debug debug "eq"; *)
        let encoded_expr1 = encode_vars_in_expression (expr1) (encoding_hash) in
        let encoded_expr2 = encode_vars_in_expression (expr2) (encoding_hash) in
        (* return *)
        Eq (encoded_expr1, encoded_expr2)
      end
    | Lt (expr1, expr2) -> begin
        let encoded_expr1 = encode_vars_in_expression (expr1) (encoding_hash) in
        let encoded_expr2 = encode_vars_in_expression (expr2) (encoding_hash) in
        (* return *)
        Lt (encoded_expr1, encoded_expr2)
      end
    | Leq (expr1, expr2) -> begin
        let encoded_expr1 = encode_vars_in_expression (expr1) (encoding_hash) in
        let encoded_expr2 = encode_vars_in_expression (expr2) (encoding_hash) in
        (* return *)
        Leq (encoded_expr1, encoded_expr2)
      end
    | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let encode_vars_in_path (art_edges_in_path : predicate list)
: ((predicate list) * ((expression, expression) Hashtbl.t)) =
begin
    (* create some vars: *)
    let encoded_art_edges = ref [] in
    let encoded_hash = Hashtbl.create (List.length art_edges_in_path) in

    (* foreach edge, encode: *)
    List.iter
        (fun an_edge ->
            let new_edge : predicate =
                encode_vars_in_predicate (an_edge) (encoded_hash) in
            encoded_art_edges := !encoded_art_edges @ [new_edge]
        )
        art_edges_in_path;
    
    (* return *)
    (!encoded_art_edges, encoded_hash)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec decode_vars_in_application (app : expression)
                                   (encoded_hash : (expression, expression) Hashtbl.t)
: expression =
begin
    match app with
    | Application ("addrof", _ (*[_]*))
    | Application ("pref", _ (*[_; _]*))
    | Application ("sref", _ (*[_; _; _]*))
    | Application ("srefp", _ (*[_; _; _]*)) -> assert false
    | Application (s, expr_list) -> begin
        let decoded_expr_list : expression list ref = ref [] in
        List.iter
            (fun an_expr -> 
                let decoded_expr : expression = 
                    decode_vars_in_expression (an_expr) (encoded_hash) in
                decoded_expr_list := !decoded_expr_list @ [decoded_expr]
            )
            expr_list;
        Application (s, !decoded_expr_list)  
      end
    (*| Application ("srefp", [pstr; field; version]) -> begin
        let decoded_pstr = decode_vars_in_expression (pstr) (encoded_hash) in
        let decoded_version =
            try Hashtbl.find (encoded_hash) (version)
            with
            | Not_found -> assert false
        in
        Application ("srefp", [decoded_pstr; field; decoded_version])
      end*)
    (*| Application ("array", [memory_expr; index_expr]) -> begin
        let decoded_memory_expr : expression = 
            decode_vars_in_expression (memory_expr) (encoded_hash) in
        let decoded_index_expr : expression = 
            decode_vars_in_expression (index_expr) (encoded_hash) in
        Application ("array", [decoded_memory_expr; decoded_index_expr])
      end*)
    | anything_else -> begin
        (* debug *)
        emerg (string_of_expression app);
        assert false
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and decode_vars_in_expression (expr : expression)
                              (encoded_hash : (expression, expression) Hashtbl.t)
: expression =
begin
    match expr with
    | Constant f -> Constant f
    | Variable s -> begin
        (* attempt to find in encoded_hash *)
        try Hashtbl.find (encoded_hash) (expr)
        with
        | Not_found -> Variable s
      end
    | Application (s, expr_list) ->
        decode_vars_in_application (expr) (encoded_hash)
    | Sum expr_list -> begin
        let decoded_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let a_decoded_expr =
                    decode_vars_in_expression (an_expr) (encoded_hash)
                in
                decoded_expr_list := !decoded_expr_list @ [a_decoded_expr])
            expr_list;
        (* return *)
        Sum !decoded_expr_list
      end
    | Coeff (f, an_expr) -> begin
        let a_decoded_expr = 
            decode_vars_in_expression (an_expr) (encoded_hash) in
        (* return *)
        Coeff (f, a_decoded_expr)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec decode_vars_in_predicate (encoded_pred : predicate)
                                 (encoded_hash : (expression, expression) Hashtbl.t)
: predicate =
begin
    match encoded_pred with
    | True -> True
    | False -> False
    | And pred_list -> begin
        let decoded_pred_list = ref [] in
        List.iter
            (fun an_encoded_pred ->
                let a_decoded_pred =
                    decode_vars_in_predicate (an_encoded_pred) (encoded_hash)
                in
                decoded_pred_list := !decoded_pred_list @ [a_decoded_pred])
            pred_list;
        (* return *)
        And !decoded_pred_list
      end
    | Or pred_list -> begin
        let decoded_pred_list = ref [] in
        List.iter
            (fun an_encoded_pred ->
                let a_decoded_pred = 
                    decode_vars_in_predicate (an_encoded_pred) (encoded_hash)
                in
                decoded_pred_list := !decoded_pred_list @ [a_decoded_pred])
            pred_list;
        (* return *)
        Or !decoded_pred_list
      end
    | Not pred -> begin
        let decoded_pred = decode_vars_in_predicate (pred) (encoded_hash) in
        (* return *)
        Not decoded_pred
      end
    | Eq (expr1, expr2) -> begin
        let decoded_expr1 = decode_vars_in_expression (expr1) (encoded_hash) in
        let decoded_expr2 = decode_vars_in_expression (expr2) (encoded_hash) in
        (* return *)
        Eq (decoded_expr1, decoded_expr2)
      end
    | Lt (expr1, expr2) -> begin
        let decoded_expr1 = decode_vars_in_expression (expr1) (encoded_hash) in
        let decoded_expr2 = decode_vars_in_expression (expr2) (encoded_hash) in
        (* return *)
        Lt (decoded_expr1, decoded_expr2)
      end
    | Leq (expr1, expr2) -> begin
        let decoded_expr1 = decode_vars_in_expression (expr1) (encoded_hash) in
        let decoded_expr2 = decode_vars_in_expression (expr2) (encoded_hash) in
        (* return *)
        Leq (decoded_expr1, decoded_expr2)
      end
    | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let decode_vars_in_path (encoded_edges_in_path : predicate list)
                        (encoded_hash : (expression, expression) Hashtbl.t)
: predicate list =
begin
    let decoded_edges = ref [] in
    List.iter
        (fun an_encoded_edge ->
            let a_decoded_edge : predicate =
                decode_vars_in_predicate (an_encoded_edge) (encoded_hash)
            in
            decoded_edges := !decoded_edges @ [a_decoded_edge]
        )
        encoded_edges_in_path;
    (* return *)
    !decoded_edges
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* If an edge contains mutiassignments, then we print each of it with memo *)
let get_string_of_memo_and_edge (art_memo : (string * string * string * string * string) list)
                                (edge : predicate)
: string = 
begin
    (* debug *)
    (*debug "memo:";
    List.iter (fun m -> debug m) art_memo;
    debug "edge:";
    debug (string_of_predicate edge);*)
    
    let s : string ref = ref "" in
    (* Check if the edge contains some assignments or not *)
    (match edge with
    | And preds -> begin
        try (
            List.iter2
                (fun an_asgn (filename, line, _, _, _) ->
                    s := !s  ^ "\n\t\t" ^ filename ^ "--realline--" ^ line ^
                         " : " ^ (string_of_predicate an_asgn);
                )
                preds
                art_memo
        ) with Invalid_argument _ -> (*ssert false*) begin
            (* art_memo must has an element *)
            (match art_memo with
            | (filename, line, _, _, _) :: _ -> s := !s ^ "\n\t\t" ^ filename ^ "--realline--" ^ line ^
                                            " : " ^ (string_of_predicate edge);
            | _ -> assert false)
        end
      end
    | True | False | Or _ | Not _ | Eq _ | Lt _ | Leq _ | Atom _ -> begin
        (* art_memo must has an element *)
        (match art_memo with
        | (filename, line, _, _, _) :: _ -> s := !s ^ "\n\t\t" ^ filename ^ "--realline--" ^ line
                                        ^ " : " ^ (string_of_predicate edge);
        | _ -> assert false)
      end);
    (* Return *)
    !s 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_string_of_art_path (path : art_path) : string = 
begin
    let string_to_return = ref "" in
    (try
        (List.iter2
            (fun anode an_edge ->
                let s1 = string_of_art_label anode.art_label in 
                let s2 = get_string_of_memo_and_edge (anode.art_memo) (an_edge) in
                string_to_return := !string_to_return ^ s1 ^ s2 ^ "\n";
            )
            path.art_nodes_in_path
            path.art_edges_in_path)
    with
    | Invalid_argument _ -> assert false);
    (* return *)
    !string_to_return
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_string_of_cutting_apath (path : art_path) : string = 
begin
    let string_to_return = ref "" in
    (try
        (List.iter2
            (fun anode an_edge ->
                if an_edge != True then begin
                    let s1 = string_of_art_label anode.art_label in 
                    let s2 = get_string_of_memo_and_edge (anode.art_memo) (an_edge) in
                    let s = s1 ^ "\n" ^ s2 in
                    string_to_return := !string_to_return ^ s ^ "\n";
                end
            )
            path.art_nodes_in_path
            path.art_edges_in_path)
    with
    | Invalid_argument _ -> assert false);
    (* return *)
    !string_to_return
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let get_string_of_cutting_apath_2 (sliced_error_path : art_path)
                                  (prime_path : art_path) 
: string = 
begin
    (* Check length of cutting_path & prime_path *)
    let l11 : int = List.length sliced_error_path.art_nodes_in_path in
    let l21 : int = List.length prime_path.art_nodes_in_path in
    if l11 != l21 then assert false;
    (* Create some vars *)
    let string_to_return = ref "" in
    let index : int ref = ref (-1) in
    (* Foreach edge in cutting_path, check if it True or not. If not, 
    get string of the edge & node to return                           *)
    (try
        (List.iter2
            (fun anode an_edge ->
                (* Increase index *)
                index := !index + 1;
                (* Check an_edge *)
                if (normalize an_edge) != True then begin
                    (* Find prime_edge in prime_path *)
                    let prime_edge : predicate = 
                        (try List.nth prime_path.art_edges_in_path !index
                        with Invalid_argument _ -> assert false)
                    in
                    (* Get string of the node & edge *)
                    let memo_s = anode.art_memo in 
                    let label_s = string_of_art_label (anode.art_label) in
                    let prime_edge_s = string_of_predicate prime_edge in
                    let edge_s = string_of_predicate an_edge in
                    let s = memo_s ^ " --> \n\t" ^ label_s ^ ": " ^ prime_edge_s ^ 
                            " --- (" ^ edge_s ^ ") ---" in
                    string_to_return := !string_to_return ^ s ^ "\n";
                end
            )
            sliced_error_path.art_nodes_in_path
            sliced_error_path.art_edges_in_path)
    with
    | Invalid_argument _ -> begin
        (*let ni : int = List.length (sliced_error_path.art_nodes_in_path) in
        let ei : int = List.length (sliced_error_path.art_edges_in_path) in
        emerg ("ni = " ^ (string_of_int ni));
        emerg ("ei = " ^ (string_of_int ei));
        emerg (get_string_of_art_path (sliced_error_path));*)
        assert false
    end);
    (* return *)
    !string_to_return
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let cut_path (cutting_path : art_path) (alabel : string * int) : art_path = 
begin
    let flag : bool ref = ref true in
    let new_nodes_in_path : art_node list ref = ref [] in
    let new_edges_in_path : predicate list ref = ref [] in
    (* Foreach node & edge in the path, check if it's before alabel-point 
    then add it to new_nodes_in_path & new_edges_in_path                   *)
    (try
        (List.iter2
            (fun anode an_edge ->
                 if !flag = true then begin
                     if (compare_art_labels alabel anode.art_label) = 0 then begin
                         flag := false;
                         new_nodes_in_path := !new_nodes_in_path @ [anode];
                         new_edges_in_path := !new_edges_in_path @ [an_edge];
                     end else begin
                         new_nodes_in_path := !new_nodes_in_path @ [anode];
                         new_edges_in_path := !new_edges_in_path @ [an_edge];
                     end; 
                 end;
            )
            cutting_path.art_nodes_in_path
            cutting_path.art_edges_in_path)
    with
    | Invalid_argument _ -> assert false);
    (* Return *)
    let new_path : art_path = {
        art_nodes_in_path = !new_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec get_strongest_post_eq_condition_with_criteria 
    (pre_eq_condition : predicate)
    (edge : predicate)
    (type_of_edge : art_stmt)
    (var_criteria : expression list)
: predicate list = 
begin
    match type_of_edge with
    | ART_ENTRY | ART_FORM | ART_RETURN -> begin
        let spcs : predicate list = 
            get_strongest_post_eq_condition_for_assignments_edge (pre_eq_condition)
              (edge) in
        choose_eq_predicates_with_criteria (spcs) (var_criteria)
      end
    | ART_COND -> begin
        (* Because we comput spcs to deal with function pointer, we don't     *)
        (* need to consider condition. Conditions will consider when compute  *)
        (* Interpolations                                                     *)
        let spcs : predicate list = [pre_eq_condition(*; edge*)] in
        choose_eq_predicates_with_criteria (spcs) (var_criteria)
      end
    | ART_CALL -> begin
        (* Temporarily consider ART_CALL is like ART_FORM *)
        let spcs : predicate list = 
            get_strongest_post_eq_condition_for_assignments_edge (pre_eq_condition)
              (edge) in
        choose_eq_predicates_with_criteria (spcs) (var_criteria)
      end
    | ART_ASSUME -> 
        get_strongest_post_eq_condition_for_art_assume (edge)
          (pre_eq_condition) (var_criteria)
    | ART_RESET (* Normally edge of ART_RESET is True *) -> begin
        let spcs : predicate list = [pre_eq_condition] in
        choose_eq_predicates_with_criteria (spcs) (var_criteria)    
      end   
    (* Other cases, assert false *)   
    | ART_ERROR | ART_FAIL | ART_NORETURN -> assert false
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec get_strongest_post_eq_conditions_with_criteria 
    (pre_eq_conditions : predicate list)
    (edge : predicate)
    (type_of_edge : art_stmt)
    (var_criteria : expression list)
: predicate list = 
begin
    (* debug *)
    (*debug ("get_strongest_post_eq_conditions_with_criteria");
    debug ("pre_eq_conditions: " ^ (string_of_predicate (And pre_eq_conditions)));
    debug ("edge: " ^ (string_of_predicate edge));
    debug ("var_criteria: " ^ (string_of_expression (Sum var_criteria)));*)
    
    let result : predicate list = 
        (match type_of_edge with
        | ART_ENTRY | ART_FORM | ART_RETURN -> begin
            let spcs : predicate list = 
                get_strongest_post_eq_conditions_for_assignments_edge (pre_eq_conditions)
                  (edge) in
            choose_eq_predicates_with_criteria (spcs) (var_criteria)
          end
        | ART_COND -> begin
            (* Because we comput spcs to deal with function pointer, we don't *)
            (* need to consider condition. Conditions will consider when      *)
            (* compute Interpolations                                         *)
            let spcs : predicate list = pre_eq_conditions(*; edge*) in
            choose_eq_predicates_with_criteria (spcs) (var_criteria)
          end
        | ART_CALL -> begin
            (* Temporarily consider ART_CALL is like ART_FORM *)
            let spcs : predicate list = 
                get_strongest_post_eq_conditions_for_assignments_edge (pre_eq_conditions)
                  (edge) in
            choose_eq_predicates_with_criteria (spcs) (var_criteria)
          end
        | ART_ASSUME -> 
            get_strongest_post_eq_conditions_for_art_assume (edge)
              (pre_eq_conditions) (var_criteria)
        | ART_RESET (* Normally edge of ART_RESET is True *) -> begin
            let spcs : predicate list = pre_eq_conditions in
            choose_eq_predicates_with_criteria (spcs) (var_criteria)    
          end   
        (* Other cases, assert false *)   
        | ART_ERROR | ART_FAIL | ART_NORETURN -> assert false)
    in
    (* debug *)
    (*debug ("spcs: " ^ (string_of_predicate (And result)));*)
    (* Return *)
    result
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_node_and_edge_tuple (apath : art_path)

: (art_node * predicate) list = 
begin
    (* Create var to contains results *)
    let tuples : (art_node * predicate) list ref = ref [] in
    (* Foreach node & its outgoing_edge in the apath, create a tuple *)
    (try
        (List.iter2
            (fun a_node an_edge ->
                let new_tuple = (a_node, an_edge) in
                tuples := !tuples @ [new_tuple]
            )
            apath.art_nodes_in_path
            apath.art_edges_in_path;
    )
    with
    | Invalid_argument _ -> assert false);  
    (* Return *)
    !tuples
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_string_of_path_formulae (path_formulae : predicate list) : string =
begin
    let tab = "" in
    let s : string ref = ref "" in
    (* Foreach edge of the path, get format string of it, then update to s *)
    List.iter
        (fun edge ->
            let new_s : string = get_format_string_of_predicate (tab) (edge) in
            s := !s ^ "\n" ^ new_s;    
        )
        path_formulae;
    (* return *)
    !s
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_all_func_names_in_path (path : art_path) : string =
begin
    let func_names : string ref = ref "" in
    let const_tab : string ref = ref "    " in
    let tab_list : string list ref = ref [""] in
    let tab_index : int ref = ref 0 in
    let before_entry : bool ref = ref false in 
    let current_tab = ref "" in
    (try (
        List.iter2
            (fun n e -> 
                (* Check if this node is art_call or art_return or entry *)
                (match n.art_type with
                | ART_CALL -> begin
                    (*if (normalize e) != True then begin
                        (* concate label of n to func_names tring *)
                        let new_funcname = (string_of_art_label n.art_label) ^
                                           ":  " ^ (string_of_predicate e) in   
                        func_names := !func_names ^ new_funcname ^ "\n"; 
                        ()
                    end*) ()
                  end
                | ART_RETURN -> begin
                    (* concate label of n to func_names tring *)
                    let new_funcname = (string_of_art_label n.art_label) (*^
                                       ":  " ^ (string_of_predicate e)*) in  
                    (* Find tab *)
                    if !before_entry = false then begin
                        tab_index := !tab_index - 1;
                        tab_list := List.rev (List.tl (List.rev (!tab_list)));
                        if !tab_index < 0 then assert false;
                        current_tab := List.nth !tab_list !tab_index;
                    end; 
                    (* Update func_name *)
                    func_names := !func_names ^ !current_tab ^ new_funcname ^ "\n";
                    before_entry := false; 
                    () 
                  end
                | _ -> begin
                    let (cfa_label, _) = n.art_label in
                    let index_of_fname = try String.index cfa_label '%'
                                         with Not_found -> -1
                    in
                    let based_label = try String.sub cfa_label 0 index_of_fname
                                      with Invalid_argument _ -> ""
                    in
                    if (String.compare based_label "#Lentry") = 0 then begin
                        (* Find tab *)
                        if !before_entry = true then begin
                            let new_tab : string = (List.nth !tab_list !tab_index) ^ !const_tab in
                            tab_list := !tab_list @ [new_tab];
                            tab_index := !tab_index + 1;
                            current_tab := new_tab;
                        end;
                        (* concate label of n to func_names tring *)
                        let new_funcname = (string_of_art_label n.art_label) (*^
                                           ":  " ^ (string_of_predicate e)*) in   
                        (* Update func_name *)
                        func_names := !func_names ^ !current_tab ^ new_funcname ^ "\n";
                        before_entry := true;  
                        () 
                    end
                  end)
            )
            path.art_nodes_in_path
            path.art_edges_in_path 
    ) with Invalid_argument _ -> assert false);
    (* Return *)
    !func_names
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let art_path_to_cfa_path (apath : art_path) : string list = 
begin
    let cpath : string list ref = ref [] in
    (* Foreach anode in apath, get cfa_label of it, then insert in cpath *)
    List.iter
        (fun anode -> 
            let (clabel, _) = anode.art_label in
            cpath := !cpath @ [clabel];
        )
        apath.art_nodes_in_path;
    (* Return *)
    !cpath
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_edge (apath : art_path)
             (getting_label : string * int)
: predicate =
begin
    let edge : predicate ref = ref True in
    let found_edge : bool ref = ref false in
    (try (
        List.iter2
            (fun anode an_edge -> 
                if !found_edge = false then begin
                    if (compare_art_labels anode.art_label getting_label) = 0 then begin
                        edge := an_edge;
                        found_edge := true;
                    end                   
                end       
            )
            apath.art_nodes_in_path
            apath.art_edges_in_path;
    )
    with Invalid_argument _ -> assert false);
    (* return *)
    !edge
end


let increase_start_of_ilist (ilst : int list) : int list = 
    try ((List.hd ilst) + 1) :: (List.tl ilst) 
    with Failure "hd" -> assert false
    | Failure "tl" -> assert false

let rec rev_ilist_to_string (ilst : int list) : string = begin
    match ilst with
    | [] -> ""
    | [i] -> string_of_int i
    | head :: tail -> (rev_ilist_to_string tail) ^ "-" ^ (string_of_int head)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let xml_of_a_memo (memo : string * string * string * string * string) 
: string = begin
    let (filename, startline, startposition, endline, endposition) = memo in
    let xmlfile = "<filename>" ^ filename ^ "</filename>" in
    let xmlstartline = "<startline>" ^ startline ^ "</startline>" in
    let xmlstartposition = 
        (if (String.compare startposition "") = 0 then "" 
        else "<startposition>" ^ startposition ^ "</startposition>") 
    in
    let xmlendline = 
        (if (String.compare endline "") = 0 then "" 
        else "<endline>" ^ endline ^ "</endline>") 
    in
    let xmlendposition = 
        (if (String.compare endposition "") = 0 then "" 
        else "<endposition>" ^ endposition ^ "</endposition>") 
    in  
    xmlfile ^ xmlstartline ^ xmlstartposition ^ xmlendline ^ xmlendposition;
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_a_memo (memo : string * string * string * string * string) 
: string = begin
    let (filename, startline, startposition, endline, endposition) = memo in
    let sfile = "filename: " ^ filename in
    let sstartline = "\nstartline: " ^ startline in
    let sstartposition = 
        (if (String.compare startposition "") = 0 then "" 
        else "\nstartposition: " ^ startposition) 
    in
    let sendline = 
        (if (String.compare endline "") = 0 then "" 
        else "\nendline: " ^ endline) 
    in
    let sendposition = 
        (if (String.compare endposition "") = 0 then "" 
        else "\nendposition: " ^ endposition) 
    in  
    sfile ^ sstartline ^ sstartposition ^ sendline ^ sendposition;
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let start_xml_of_edge etype enode e initial_flag ref_counters : string = 
begin
    let xmls = ref "" in
    let initial = (if initial_flag = true then "<initial>true</initial>" else "") in
    (* Get reachable regions *)
    let weakest_precond = 
        "<weakest-pre-condition>" ^ 
        (string_of_predicate (create_based_vars_in_predicate enode.art_weakest_precond)) ^ 
        "</weakest-pre-condition>"
    in
    let reachable_region =
        "<reachable-region>" ^
        (string_of_predicate (create_based_vars_in_predicate enode.art_reachable_region)) ^
        "</reachable-region>"
    in
    let strongest_postcond = ""
(*        "<strongest-post-condition>" ^                                                       *)
(*        (string_of_predicate (create_based_vars_in_predicate enode.art_strongest_postcond)) ^*)
(*        "</strongest-post-condition>"                                                        *)
    in
    (* Get function - name *)
    let func_name = ref "" in
    let e : predicate = 
        (if etype = ART_CALL then begin
            match e with
            | And ((Eq (left, right)) :: tail) -> 
                func_name := (" name=\"" ^ (string_of_expression right) ^ "\"");
                And tail
            | Eq (left, right) ->
                func_name := (" name=\"" ^ (string_of_expression right) ^ "\"");
                True
            | _ -> assert false
        end else e)
    in
    let state = weakest_precond ^ reachable_region ^ strongest_postcond in
    (* Check if the edge contains some assignments or not *)
    (match e with
    | And [] -> 
        (* art_memo must has an element *)
        (match enode.art_memo with
        | amemo :: _ -> begin
            let xmlmemo = xml_of_a_memo amemo in
            ref_counters := increase_start_of_ilist (!ref_counters);
            xmls := "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                    "label=\"" ^ (rev_ilist_to_string !ref_counters) ^ "\"" ^
                    !func_name ^">\n" ^ xmlmemo ^ state ^ initial ^ "<instruction>()</instruction>"
          end
        | _ -> assert false)
    | And [p] ->
        (* art_memo must has an element *)
        (match enode.art_memo with
        | amemo :: _ -> begin
            let xmlmemo = xml_of_a_memo amemo in
            ref_counters := increase_start_of_ilist (!ref_counters);            
            xmls := "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                    "label=\"" ^ (rev_ilist_to_string !ref_counters) ^ "\"" ^
                    !func_name ^ ">\n" ^ xmlmemo ^ state ^ initial ^
                    "<instruction>\"" ^ (string_of_predicate (create_based_vars_in_predicate e)) ^ 
                    "\"</instruction>" 
          end
        | _ -> assert false)
    | And preds -> begin
        if (List.length preds) = (List.length enode.art_memo) then (
            List.iter2
                (fun an_asgn amemo ->
                    let xmlmemo = xml_of_a_memo amemo in
                    ref_counters := increase_start_of_ilist (!ref_counters); 
                    let new_xmls = "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                                   "label=\"" ^ (rev_ilist_to_string !ref_counters)  ^ "\"" ^
                                   !func_name ^ ">\n" ^ xmlmemo ^ initial ^
                                   "<reachable-region>computing...please go to next</reachable-region>" ^ 
                                   "<instruction>\"" ^ (string_of_predicate (create_based_vars_in_predicate an_asgn)) ^ 
                                   "\"</instruction>" ^ "</statement>\n"
                    in
                    xmls := !xmls ^ new_xmls
                )
                (List.rev (List.tl (List.rev preds)))
                (List.rev (List.tl (List.rev enode.art_memo)));
            (***)
            let lastmemo = List.hd (List.rev enode.art_memo) in
            let lastxmlmemo = xml_of_a_memo lastmemo in
            let an_asgn = List.hd (List.rev preds) in
            ref_counters := increase_start_of_ilist (!ref_counters); 
            let new_xmls = "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                           "label=\"" ^ (rev_ilist_to_string !ref_counters)  ^ "\"" ^
                           !func_name ^ ">\n" ^ lastxmlmemo ^ state ^ initial ^
                           "<instruction>\"" ^ (string_of_predicate (create_based_vars_in_predicate an_asgn)) ^ 
                           "\"</instruction>" in
            xmls := !xmls ^ new_xmls; 
        ) else begin
            (* art_memo must has an element *)
            (match enode.art_memo with
            | amemo :: _ -> begin
                let xmlmemo = xml_of_a_memo amemo in
                ref_counters := increase_start_of_ilist (!ref_counters);
                xmls := "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                        "label=\"" ^ (rev_ilist_to_string !ref_counters)  ^ "\"" ^
                        !func_name ^ ">\n" ^ xmlmemo ^ state ^ initial ^ 
                        "<instruction>\"" ^ (string_of_predicate (create_based_vars_in_predicate e)) ^ 
                        "\"</instruction>"
              end
            | _ -> assert false)
        end
      end
    | True | False | Or _ | Not _ | Eq _ | Lt _ | Leq _ | Atom _ -> begin
        (* art_memo must has an element *)
        (match enode.art_memo with
        | amemo :: _ -> begin
            let xmlmemo = xml_of_a_memo amemo in
            ref_counters := increase_start_of_ilist (!ref_counters);
            xmls := "<statement type=\"" ^ (string_of_art_stmt etype) ^ "\" " ^ 
                    "label=\"" ^ (rev_ilist_to_string !ref_counters)  ^ "\"" ^
                    !func_name ^ ">\n" ^ xmlmemo ^ state ^ initial ^
                    "<instruction>\"" ^ (string_of_predicate (create_based_vars_in_predicate e)) ^ 
                    "\"</instruction>"
          end
        | _ -> assert false)
      end);
    (* Return *)
    !xmls
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let xml_of_path (apath_with_rr : art_path) : string = 
begin
    let xmls = ref "<path>\n" in
    let count_start_xml = ref 0 in
    let is_call : bool ref = ref false in
    let initial = ref 2 in
    let initial_flag = ref true in
    let counters = ref [0] in
    let current_memo : (string * string * string * string * string) ref =
        ref ("", "", "", "", "") in 
    (try (
        List.iter2
            (fun n e ->
                let n = (
                    match n.art_memo with
                    | amemo :: _ -> (
                        let (fn, _, _, _, _) = amemo in
                        if (String.length fn) = 0 then
                            {n with art_memo = [!current_memo]}
                        else (
                            current_memo := amemo;
                            n))
                    | _ -> {n with art_memo = [!current_memo]}
                ) in
                if !initial_flag = true then (
                    initial := !initial - 1; 
                    if !initial < 0 then initial_flag := false);

                (match n.art_type with
                | ART_FORM | ART_COND | ART_ASSUME | ART_RESET -> begin
                    if !is_call = true then begin
                        xmls := !xmls ^ "<defined>false</defined></statement>\n";
                        is_call := false;
                    end;
                    xmls := !xmls ^ (start_xml_of_edge n.art_type n e !initial_flag counters) 
                            ^ "</statement>\n"
                  end
                | ART_CALL -> begin
                    if !is_call = true then begin
                        xmls := !xmls ^ "<defined>false</defined></statement>\n";
                        is_call := false;
                    end;
                    is_call := true;
                    xmls := !xmls ^ (start_xml_of_edge n.art_type n e !initial_flag counters);
                  end
                | ART_ENTRY -> begin
                    if !is_call = true then begin
                        xmls := !xmls ^ "<defined>true</defined></statement>\n";
                        is_call := false;
                    end;
                    counters := (-1) :: !counters;
                    xmls := !xmls ^ (start_xml_of_edge n.art_type n e !initial_flag counters);
                    count_start_xml := !count_start_xml + 1;
                  end
                | ART_RETURN -> begin
                    if !is_call = true then begin
                        xmls := !xmls ^ "<defined>false</defined>" ^ "</statement>\n" ;
                        is_call := false;
                    end;
                    xmls := !xmls ^ "</statement>\n";
                    counters := List.tl !counters;
                    count_start_xml := !count_start_xml - 1;
                  end
                | _ -> assert false)
            )
            apath_with_rr.art_nodes_in_path
            apath_with_rr.art_edges_in_path
    ) with Invalid_argument _ -> assert false);
    
    xmls := !xmls ^ (multiply_string_by_n "</statement>\n" !count_start_xml) ^ "\n</path>";        
    (* Return *)
    !xmls
end

(* Export the checking result to a specified output xml file *)
let save_result_as_xml_file (filename : string) 
                            (successful : bool) 
			    (error_path : art_path)
                            (error_location_memo : string * string * string * string * string)
                            (error_message : string)
                            : unit = (
    let outchan : out_channel = 
        try open_out filename 
        with Sys_error _ -> (
             emerg ("can not open " ^ filename);
             assert false)
    in
    output_string outchan ("<!-- ModCheck result -->\n<errorpath>");
    if (successful == true) then 
        output_string outchan ("<result val=\"safe\"/>")
    else 
        begin
            output_string outchan ("<result val=\"unsafe\">"
	                           ^ (xml_of_a_memo error_location_memo) 
				   ^ ("<error-message>" 
	                           ^ error_message
                                   ^ "</error-message>")
                                   ^ "</result>")
        end;
    (* FIXME: temporarily ignore s_f_contract 
    output_string outchan !s_f_contract;
    *)
    (*output_string outchan (Cfa.xmls_of_unchecked_nodes 
                                     (Cfa.find_unchecked_cnode_in_main cfas (m_name, 1)));*)
    output_string outchan (xml_of_path error_path);
    output_string outchan ("</errorpath>");
    try close_out outchan
    with Sys_error _ -> (
         emerg "can not close or flush data to errorpath.xml";
         assert false))

(* get call stack of an art_path*)
let get_call_stack (path : art_path) : (art_node * (string * string)) list =
begin
    (* Extract full function name in label of ART_ENTRY
       or ART_RETURN node*)
    let extract_full_func (label : string) : string =
    begin
        let b_index = (try String.index label '%' + 1
                       with Failure _ -> 0) in
        let e_index = (try String.rindex label '#' - 1
                       with Failure _ -> (String.length label) - 1) in
        let len = e_index - b_index + 1 in
        (try String.sub label b_index len
         with Invalid_argument _ -> assert false)
    end in
    let nodes = path.art_nodes_in_path in
    let call_stack = (
        (* Traverse through every nodes in error path *)
        List.fold_left (
            fun stack node ->
            begin
                match node.art_type with
                (* If the current node is an "Entry" node, push it to stack with
                   the default location *)
                | ART_ENTRY -> (node, ("unidentified", "-1")) :: stack
                (* If the current node is a "Call" node, replace the default
                   location with the new location (file name, line) *)
                | ART_CALL  -> (
                      match stack with
                      | [] -> stack
                      | head :: tail -> (
                            let entry_node = fst head in
                            let memo = node.art_memo in
                            let (file, line) = (
                                match memo with
                                | (f, l, _, _, _) :: [] -> (f, l)
                                | _ -> ("unidentified", "-1")
                            ) in
                            (entry_node, (file, line)) :: tail;
                        )
                  )
                (* If the current node is a "Return" node:
                     + If its full name (include: filename, function name,
                       level) is the same as full name of the top node in stack
                       then remove top element
                     + Otherwise, error path is invalid *)
                | ART_RETURN -> (
                      match stack with
                      | [] -> assert false
                      | head :: tail -> (
                            let entry_node = fst head in
                            let entry_func = (
                                let entry_label = entry_node.art_label in
                                let lb1 = string_of_art_label entry_label in
                                extract_full_func lb1
                            ) in
                            let return_func = (
                                let lb2 = string_of_art_label node.art_label in
                                extract_full_func lb2
                            ) in
                            if (entry_func <> return_func) then
                                assert false
                            else
                                tail
                       )
                  )
                | _ -> stack
            end
        ) [] nodes
    ) in
    call_stack;
end

(* get string of data in call stack (call stack contains function name) *)
let string_of_call_stack (path : art_path)
                         : string =
begin
    let stack = get_call_stack path in
    let callstack_indent = ref "    " in
    let stack_data = (
        List.fold_left (
            fun data node_and_location ->
            begin
                let node = fst node_and_location in
                let (file_name, line) = snd node_and_location in
                let label = fst(node.art_label) in
                let func_id = (
                    let e_index = (try String.rindex label '%' - 1
                                   with Not_found -> assert false) in
                    let b_index = (try String.index label '%' + 1
                                   with Not_found -> assert false) in
                    let len = e_index - b_index + 1 in
                    (try String.sub label b_index len
                     with Invalid_argument _ -> assert false)
                ) in
                let func_name = (
                    let e_index = String.length func_id - 1 in
                    let b_index = (try String.rindex func_id '%' + 1
                                   with Not_found -> 0) in
                    let len = e_index - b_index + 1 in
                    (try String.sub func_id b_index len
                     with Invalid_argument _ -> assert false)
                ) in
                callstack_indent := !callstack_indent ^ "    ";
                data ^ "  " ^ !callstack_indent ^ func_name ^ " <" ^ file_name
                     ^ "; " ^ line ^ ">\n"
            end
        ) "" stack;
    ) in
    stack_data;
end

let string_of_art_node (art_node : art_node) : string = (
    let string_of_art_node_label = string_of_art_label (art_node.art_label) in
    let string_of_incoming = string_of_art_label (art_node.art_incoming_label) in
    let string_of_outgoings =
        String.concat " -- "
                      (List.map2 (fun label edge ->
                          (string_of_art_label label) ^ " : " ^ (string_of_predicate edge)
                      ) art_node.art_outgoing_labels art_node.art_outgoing_edges)
    in
    let tab = "      " in
    string_of_art_node_label ^ "\n"
    ^ tab ^ "in_coming : " ^ string_of_incoming ^ "\n"
    ^ tab ^ "out_goings : " ^ string_of_outgoings
)

let string_of_art (art : art) : string = (
    Hashtbl.fold (fun label art_node str ->
        str ^ (string_of_art_node art_node) ^ "\n"
    ) art.art_node_hash ""
)

(** find edge from current_art_node to child node *)
let find_edge_in_art (current_art_node : art_node)
                     (child_label : string * int)
: predicate = (
    let rec find_edge (labels : (string * int) list)
                      (edges : predicate list)
                      (child_label : string * int)
    : predicate = (
        match labels, edges with
        | a_label :: remain_labels, a_edge :: remain_edges ->
            if (compare_art_labels a_label child_label) = 0 then
                a_edge
            else find_edge remain_labels remain_edges child_label
        | _ -> assert false
    ) in
    (find_edge current_art_node.art_outgoing_labels
               current_art_node.art_outgoing_edges
               child_label)
)