(* loligoedulis/checker.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Form

let xTRACE_toplevel = ref true

(* Headers of Saved CFA File (20 chars). *)

let ofile_header = "LOLIGO-EDULIS-CFA-O\n"
let afile_header = "LOLIGO-EDULIS-CFA-A\n"
let efile_header = "LOLIGO-EDULIS-CFA-E\n"

let string_equal = Misc.string_equal

module String2KeyedMap = Map.Make (
  struct
  type t = (string * string)
  let compare x y = (compare x y)
  end)

(* ================================================================ *)

let info = Cexpr.info
let debug = Cexpr.debug

let save_ofile (f : string) (defs : Cfa.ofile_t) : unit = (
  (debug "saving to file...") ;
  let c : out_channel = (open_out_bin f) in
  (output_string c ofile_header) ;
  (Marshal.to_channel c defs [(*flags*)]) ;
  (flush c) ;
  (close_out c) ;
  (debug "saving to file...done"))

let load_ofile (f : string) : Cfa.ofile_t = (
  (debug "loading from file...") ;
  let c : in_channel = (open_in_bin f) in
  let h = "--------------------" in
  let n = (input c h 0 20) in
  if (n != 20) then
    (failwith "no header") ;
  if (not (string_equal h ofile_header)) then
    (failwith "bad header") ;
  let defs : Cfa.ofile_t = (Marshal.from_channel c) in
  (close_in c) ;
  (debug "loading from file...done") ;
  defs)

let save_ofile_list (ofiles : Cfa.ofile_t list)
: unit = (
    List.iter
        (fun a_ofile ->
            let file_path = a_ofile.Cfa.ofile_file_name in
            let n = String.length file_path in
            let p = (String.sub file_path 0 (n - 2)) in
            save_ofile (p ^ ".o") a_ofile
        ) ofiles  
)

let save_afile (f : string) (defs : Cfa.afile_t) : unit = (
  (debug "saving to file...") ;
  let c : out_channel = (open_out_bin f) in
  (output_string c afile_header) ;
  (Marshal.to_channel c defs [(*flags*)]) ;
  (flush c) ;
  (close_out c) ;
  (debug "saving to file...done"))

let load_afile (f : string) : Cfa.afile_t = (
  (debug "loading from file...") ;
  let c : in_channel = (open_in_bin f) in
  let h = "--------------------" in
  let n = (input c h 0 20) in
  if (n != 20) then
    (failwith "no header") ;
  if (not (string_equal h afile_header)) then
    (failwith "bad header") ;
  let defs : Cfa.afile_t = (Marshal.from_channel c) in
  (close_in c) ;
  (debug "loading from file...done") ;
  defs)

(* ================================================================ *)
type strategy_t =
    | DFS | HRBC
(* Options for the specification reader & the checker *)
type ps_options_t = {
    psopt_entries : Sexpr.function_entry_t Vector.t;
    mutable psopt_iter_max : int;
    mutable psopt_recursive_max : int;
    mutable psopt_timeout : float;
    mutable psopt_slice_path : int;
    mutable psopt_cutlist : Sexpr.function_entry_t list;
    mutable psopt_la_debug : bool;
    mutable psopt_dot_output : bool;
    mutable psopt_xml_location : string;
    mutable psopt_intensive_pointer_checking : bool;
    mutable psopt_debug_mode : bool;
    mutable psopt_bmc : bool;
    mutable psopt_check_in_callees : bool;
    mutable psopt_check_array_index : bool;
    mutable psopt_strategy : strategy_t;
    mutable psopt_no_assert : bool;
    mutable psopt_no_subtyping_check : bool;
    mutable psopt_temp_dir : string;
    mutable psopt_round_coeffs : bool;
    psopt_verbose_components : string Vector.t;
}

(* Parses options of the ps sub-command *)

let rec make_ps_options (opts : string list list) : ps_options_t = (
    let options : ps_options_t = {
        psopt_entries = (Vector.create 4 Sexpr.nil_function_entry);
        psopt_iter_max = 5;			(* default value *)
        psopt_recursive_max = 5;		(* default value *)
        psopt_timeout = 0.;   (* default value *)
        psopt_slice_path = 2;			(* default value *)
        psopt_cutlist = [];			(* default value *)
        psopt_la_debug = false;			(* default value *)
        psopt_dot_output = false;		(* default value *)
        psopt_xml_location = "errorpath.xml";	(* default value *)
        psopt_intensive_pointer_checking = true;       (* default value *)
        psopt_debug_mode = false;               (* default value *)
        psopt_bmc = true;                      (* default value *)
        psopt_check_in_callees = false;          (* default value *)
        psopt_check_array_index = true;         (* default value *)
        psopt_strategy = HRBC;          (* default value *)
        psopt_no_assert = false;                (* default value *)
        psopt_no_subtyping_check = false;       (* default value *)
        psopt_temp_dir = Filename.temp_dir_name; (* default value *)
        psopt_round_coeffs = false;
        psopt_verbose_components = Vector.create 2 "";
    } in
    List.iter (set_ps_options options) opts;
    (* Some post-processing of the make_ps_options *)
    if (Vector.length options.psopt_entries) == 0 then
        (Vector.add options.psopt_entries { Sexpr.fe_filelocation_opt = None;
                                            Sexpr.fe_funcname = "main";
                                            Sexpr.fe_raw_data = "";
                                            Sexpr.fe_location = None });
                                            (* default value *)
    options)

and set_ps_options (options : ps_options_t) (opt : string list) : unit = (begin
    let pmsg (l : string list) : unit = (
        if (!xTRACE_toplevel) then
            (debug ("ignore ps option: '" ^ (String.concat " " l) ^ "'")))
    in
    let translate_yesorno (y_or_n : string) : bool option =
        let y_or_n = ExtString.String.uppercase y_or_n in
        if (ExtString.String.compare y_or_n "Y") == 0 then
            Some true
        else if (ExtString.String.compare y_or_n "N") == 0 then
            Some false
        else
            None
    in
    let badoption (command : string) (opt : string list) : 'a = (
        (failwith ("Bad option to " ^ command ^ " '" ^ (String.concat " " opt) ^ "'"))) 
    in       
    (match opt with
    | [] -> (assert false)
    | ["-c"] -> ()
    | ["-g"] -> ()
    | ["-s"] -> ()
    | ["-o"; _]
    | ["-D"; _] -> ()
    | ("-E" :: _) -> ()
    | ["-I"; _] -> ()
    | ["-L"; _] -> ()
    | ["-U"; _] -> ()
    | ["-nostdinc"]
    | ["-O"]
    | ["-Os"]
    | ["-include"; _]
    | ["-isystem"; _]
    | ["-imacros"; _]
    | ["-MD"; _]
    | ["-pipe"]
    | ("-Wp" :: _) -> ()
    (*| [o] -> ()		(* skipped as these are cc options *) *)
    (** The following are ps options *)
    | ["--iter-max"; a] -> 
        (options.psopt_iter_max <- (int_of_string a))
    | ["--recursive-max"; a] -> 
        (options.psopt_recursive_max <- (int_of_string a))
    | ["--timeout"; a] ->
        (options.psopt_timeout <- (float_of_string a))
    | ["--slice-path"; a] -> 
        (options.psopt_slice_path <- (int_of_string a))
    | ["--cut-list"; filename] -> (
        let cutlist = Sexpr.parse_cutlist_file filename in
        options.psopt_cutlist <- options.psopt_cutlist @ cutlist)
    | ["--main"; a] -> 
        let entry_point = Sexpr.parse_function_entry a () in
        (Vector.add options.psopt_entries entry_point)
    | ["--la-debug"] -> 
        (options.psopt_la_debug <- true)
    | ["--noasrt"] ->
        (options.psopt_no_assert <- true)
    | ["--nostpcheck"] ->
        (options.psopt_no_subtyping_check <- true)
    | ["--output-xml"; a ] -> 
        (options.psopt_xml_location <- a)
    | ["--dot"] -> 
        (options.psopt_dot_output <- true)
    | ["-d"]
    | ["--debug"] ->
        (options.psopt_debug_mode <- true)
    | ["--round"] ->
        (options.psopt_round_coeffs <- true)
    | ["--verbose"; a] -> 
        (Vector.add options.psopt_verbose_components a)
    | ["--intensive-pointer-checking"; y_or_n] -> (
        match (translate_yesorno y_or_n) with
          Some b ->
            options.psopt_intensive_pointer_checking <- b
        | None ->
            badoption "ps" opt)
    | ["--check-in-callees"; y_or_n] -> (
        match (translate_yesorno y_or_n) with
          Some b ->
            options.psopt_check_in_callees <- b
        | None ->
            badoption "ps" opt)
    | ["--check-array-index"; y_or_n] -> (
        match (translate_yesorno y_or_n) with
          Some b ->
            options.psopt_check_array_index <- b
        | None ->
            badoption "ps" opt)
    | ["--strategy"; a] ->
        if (String.compare a "dfs") == 0 then
            options.psopt_strategy <- DFS
        else if (String.compare a "hrbc") == 0 then
            options.psopt_strategy <- HRBC
        else
            badoption "ps" opt
    | ["--bmc"; y_or_n] -> (
        match (translate_yesorno y_or_n) with
          Some b ->
            options.psopt_bmc <- b
        | None ->
            badoption "ps" opt)
    | ["--temp-dir"; dirname] -> (
        let _ = 
            try if not (Sys.is_directory dirname) then (
                    info ("'" ^ dirname ^ "' is not a directory.");
                    (badoption "ps" opt))
                else (
                    (* Check for READ and WRITE permissions *)
                    try Unix.access dirname [Unix.R_OK; Unix.W_OK]
                    with
                      Unix.Unix_error _ -> (
                          info ("Cannot read and/or write to '"
                                ^ dirname ^ "'.");
                          badoption "ps" opt))
            with
              Sys_error _ -> (
                  info ("'" ^ dirname ^ "' does not exist.");
                  (badoption "ps" opt))
        in
        options.psopt_temp_dir <- dirname)

    | _ -> 
        (badoption "ps" opt))
end)

(* ================================================================ *)

(* TOPLEVELS OF TRANSLATER/INTERPRETER *)

(* Parses and translates a source file.
   {flag_c} is true if the flag '-c' is on, {filename} is of a file
   output from preprocessor, while {file_path} is the one
   passed. *)

let parse_source_file (filename : string) 
                      (file_path : string)
                      (loop_bound_opt : int option)
                      (subtyping_check : bool)
                      : Cexpr.ofile_t =
begin
    let cexpr_ofile =
        Smain.parse_source_file filename
                                file_path
                                loop_bound_opt
                                subtyping_check
    in
(*    Hashtbl.clear Sinstrument.f_pointer_field_hashtbl; *)
(*    Hashtbl.clear Ccommon.contract_of_field_hashtbl; *)
    (cexpr_ofile)
end

let create_cfa_ofile (stmt_ofile : Cexpr.ofile_t) : Cfa.ofile_t =
begin
    let translate_ofile : Translate.ofile_t =
        Smain.translate_ofile stmt_ofile
    in
    let ofile = Cfa.create_cfa_ofile translate_ofile in
    (* get contracts of fields *)
    {ofile with Cfa.ofile_contract_of_field =
                    Cfa.translate_contract_of_field stmt_ofile.Cexpr.f_pointer_field_hashtbl
                                                    stmt_ofile.Cexpr.contract_of_field_hashtbl}    
end

(** Insert pre_main_funcs that call to main funcs
    Insert specs of vars
    Insert func contracts in caller side
    Translate processed Cexpr.ofile_t list to Cfa.ofile_t list
    @param stmt_ofiles : Cexpr.ofile_t list
    @param ps_options : ps_option_t
    @return Cfa.ofile_t list * ps_options_t 
*)
let create_cfa_ofiles (stmt_ofiles : Cexpr.ofile_t list)
                      (ps_options : ps_options_t)
                      (flag_c : bool)
: Cfa.ofile_t list * ps_options_t = (
    let processed_stmt_ofiles, new_main_entries =
        Sinstrument.process_ofiles stmt_ofiles
                                   ps_options.psopt_entries
    in
    
    if (flag_c) then
    (begin
        List.iter
            (fun a_ofile ->
                let stmts =
                    List.map
                        (fun (_, st) -> st) !(a_ofile.Cexpr.ofile_defs) in
                let n = (String.length a_ofile.Cexpr.ofile_file_name) in
                (assert (n > 2));
                let p = (String.sub a_ofile.Cexpr.ofile_file_name 0 (n-2)) in
                Sdump.save_inter_result (p ^ ".inter.c") stmts
            ) processed_stmt_ofiles
    end);
    
    let new_ps_options =
        { ps_options with psopt_entries = new_main_entries }
    in
    let cfa_ofiles : (Cfa.ofile_t list) =
        List.map create_cfa_ofile processed_stmt_ofiles in
    if (flag_c) then
        (save_ofile_list cfa_ofiles);
    (cfa_ofiles, new_ps_options)
)

(* Gathers definitions of files and makes an archive.
   Mostly a copy of Interp.package_definitions *)

let package_definitions (afile : Cfa.afile_t option) (ofiles : Cfa.ofile_t list) : Cfa.afile_t = (
  let find_cfa id directory libs =
    try
      let hashname = Hashtbl.find directory id in
      let ofile = Hashtbl.find libs hashname in
      Hashtbl.find ofile.Cfa.ofile_cfas id
    with 
    | Not_found -> raise Not_found
  in
  let abstpreds : (string, form_t list) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).Cfa.afile_global_predicates
    else
      (Hashtbl.create 10) in
  let libs : (string, Cfa.ofile_t) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).Cfa.afile_libs
    else
      (Hashtbl.create 10) in
  let directory : (string, string) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).Cfa.afile_directory
    else
      (Hashtbl.create 10) in
  let rec compare_types =
    fun tn0 file0 tn1 file1 table ->
      if compare file0 file1 <= 0 then
	compare_types1 tn0 file0 tn1 file1 table
      else
	compare_types1 tn1 file1 tn0 file0 table
  and compare_types1 = 
    fun tn0 file0 tn1 file1 table ->
      assert (compare file0 file1 <= 0);
      let omap =
	if String2KeyedMap.mem (file0, file1) table then
	  String2KeyedMap.find (file0, file1) table
	else
	  Cexpr.StringKeyedMap.empty in
      let (compat, nmapopt) =
	Cexpr.check_compatible_type (Some omap) tn0 tn1 in
      let nmap = match nmapopt with None -> assert false | (Some m) -> m in
(*       assert (Cexpr.StringKeyedMap.cardinal nmap  *)
(* 	       >= Cexpr.StringKeyedMap.cardinal omap); *)
      (Cexpr.proper_type_p compat,
       String2KeyedMap.add (file0, file1) nmap table) in
  (* Collect abstraction predicates. *)
  let globalpreds =
    (uniqify_form_list
	(List.flatten
	    (List.map (fun o -> o.Cfa.ofile_global_predicates) ofiles))) in
  (List.iter
      (fun p ->
	let vs : StringSet.t = (Smt.find_free_variables (StringSet.empty) p) in
	(StringSet.iter
	    (fun (v : string) ->
	      (Hashtbl.replace abstpreds v
		  (if (Hashtbl.mem abstpreds v) then
		    ([p] @ (Hashtbl.find abstpreds v))
		    else
		      [p])))
	    vs))
      globalpreds) ;
  if (!Interp.aiTRACE_linking) then
    (debug (";; global.predicates="
		     ^ (form_list_to_string globalpreds))) ;
  (* Put files into archive. *)
  let _ =
  (List.fold_left
      (fun typetable0 (o : Cfa.ofile_t) ->
	let filename = o.Cfa.ofile_file_name in
	if (!Interp.aiTRACE_linking) then
	  (debug (";; LINKING FILE '" ^ filename ^ "'...")) ;
	let hashname = o.Cfa.ofile_hash_name in
        let typetable3 =
	(Hashtbl.fold
	    (fun (id : string) (cfa : Cfa.cfa) typetable1 ->
	      match cfa with
		Cfa.CFA_FUNDEF {
		  Cfa.cfa_f_meta_data = (meta_data : Cfa.cfa_meta_data_t)
		} ->
		  (begin
		    if (!Interp.aiTRACE_linking) then
		      (debug (";; looking at symbol(f)="
				       ^ (meta_data.Cfa.cfa_id))) ;
		    match meta_data.Cfa.cfa_scls with
		      Cexpr.NILSCLS | Cexpr.EXTERN ->
		        (begin
			  try
                            let old_cfa = find_cfa id directory libs in
                            match old_cfa with
                            | Cfa.CFA_FUNDEF _ ->
                              (failwith ("multiply defined: " ^ id))
                            | Cfa.CFA_VARDEF v ->
                              (begin
                                let old_tn =
                                    v.Cfa.cfa_v_meta_data.Cfa.cfa_tn
                                and cur_tn =
				    meta_data.Cfa.cfa_tn
				  and old_hashname = Hashtbl.find directory id in
				  
				  let (compat, typetable2) =
				    compare_types old_tn old_hashname
				      cur_tn hashname typetable1 in
				  if not compat then
				  failwith ("Multiple definitions with inconsistent types of " ^ id);
                                  (Hashtbl.replace directory id hashname);
				  typetable2
                               end)
                            | Cfa.CFA_OTHERDEF ->
                              (assert false)
                          with
                          | Not_found ->
			      ((Hashtbl.replace directory id hashname);
			       typetable1)
                        end)
		    | _ ->
			typetable1
		  end)
	      | Cfa.CFA_VARDEF {
		  Cfa.cfa_v_meta_data = (meta_data : Cfa.cfa_meta_data_t)
		} ->
		  (* MEMO: Empty name is for structure definition. *)
		  if (not (string_equal id "")) then
		    (begin
		      if (!Interp.aiTRACE_linking) then
			(debug (";; looking at symbol(v)="
					 ^ (meta_data.Cfa.cfa_id))) ;
		      match meta_data.Cfa.cfa_scls with
                      | Cexpr.NILSCLS
                      | Cexpr.EXTERN ->
                        (begin
                          try
                            let old_cfa = find_cfa id directory libs in
                            let old_meta_data =
                                match old_cfa with
                                | Cfa.CFA_FUNDEF f ->
                                    f.Cfa.cfa_f_meta_data
                                | Cfa.CFA_VARDEF v ->
                                    v.Cfa.cfa_v_meta_data
                                | Cfa.CFA_OTHERDEF ->
                                    assert false
                            in
                            let old_tn = old_meta_data.Cfa.cfa_tn
			      and cur_tn = meta_data.Cfa.cfa_tn
			      and old_hashname = Hashtbl.find directory id in
			      let (compat, typetable2) =
				compare_types old_tn old_hashname
				  cur_tn hashname typetable1 in
			      if not compat then
			      failwith ("Multiple definitions with inconsistent types of " ^ id);
                            match meta_data.Cfa.cfa_scls,
                                      old_meta_data.Cfa.cfa_scls with
				| Cexpr.EXTERN, _ ->
				    typetable2
                            | Cexpr.NILSCLS, Cexpr.EXTERN ->
				    ((Hashtbl.replace directory id hashname);
				     typetable2)
				| _ ->
				    typetable2
                          with
                          | Not_found ->
				((Hashtbl.replace directory id hashname);
				 typetable1)
                        end)
                      | Cexpr.REGISTER ->
                        (begin
                          try
			      let _ = find_cfa id directory libs in ();
			      typetable1
                          with
                          | Not_found ->
				((Hashtbl.replace directory id hashname);
				 typetable1)
                        end)
		      | Cexpr.AUTO ->
			  (failwith ("Bad storage class for: " ^ id))
		      | Cexpr.STATIC ->
			  (* No checks, just ignore. *)
			  typetable1
		      | Cexpr.TYPEDEF ->
			  (* No checks, just ignore. *)
			  typetable1
		    end)
		  else
		    typetable1
	      | (Cfa.CFA_OTHERDEF) ->
		  typetable1)
	    o.Cfa.ofile_cfas typetable0) in
	if (!Interp.aiTRACE_linking) then
	  (debug (";; LINKING FILE '" ^ filename ^ "' DONE")) ;
	if (Hashtbl.mem libs hashname) then
	  (failwith ("o files with conflicting random names: " ^ hashname)) ;
	(Hashtbl.replace libs hashname o);
	typetable3)
      String2KeyedMap.empty
      ofiles) in
  let a : Cfa.afile_t = {
    Cfa.afile_start = Sexpr.nil_function_entry ;
    Cfa.afile_libs = libs ;
    Cfa.afile_directory = directory ;
    Cfa.afile_global_predicates = abstpreds
  } in
  (Cfa.import_cutlist a !Sexpr.cut_functions);
  (Cfa.process_undef_functions a);
  a)

(* ================================================================ *)

(* Context for Checking. *)

type check_context_t = {
    chk_afile : Cfa.afile_t;
    chk_options : ps_options_t;
    (* Reserved
    chk_callstack_ofile : Cfa.ofile_t Vector.t;
    chk_callstack_cfa : Ccfa.cfa Vector.t;
    *)
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_check_context (afile : Cfa.afile_t) 
                         (options : ps_options_t)
                         : check_context_t = (
    {
      chk_afile = afile;
      chk_options = options;
    })
(*
let create_check_context (afile : Cfa.afile_t) 
                         (options : ps_options_t)
                         : check_context_t = (
    {
      chk_afile = afile;
      chk_options = options;
    })
*)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_cfa_in_context (ctx : check_context_t)
                        ?(caller = None)
                        (symbol_name : string)
                        : Cfa.cfa =
    Cfa.find_cfa_in_afile ctx.chk_afile ~caller:caller symbol_name
    
let find_cfa_of_function_entry (ctx : check_context_t)
                               (entry : Sexpr.function_entry_t)
                               : Cfa.cfa =
    Cfa.find_cfa_of_function_entry ctx.chk_afile entry

(* {do_check} holds a function (check_context_t -> unit) which is
   an entry point of a checker.  This body function is dummy. *)

let do_check = ref
    (fun (ctx : check_context_t) -> (Printf.fprintf stderr "Check nothing.\n"))

(* Checker's internals:
       + builin_ofile : set of checker's built-in functions
 *)

let builtin_ofile : Cfa.ofile_t =
    {
      Cfa.ofile_hash_name = "no_hash_name";
      Cfa.ofile_file_name = "include/builtin_functions.h";
      Cfa.ofile_cfas = Hashtbl.create 10;
      Cfa.ofile_global_predicates = [];
      Cfa.ofile_contract_of_field = Hashtbl.create 10
    }

let clap_builtin_modulo () : Cfa.cfa = 
    let specs_of_builtin_modulo () = (
        let raw_specs_of_modulo =
            " $requires (a >= 0); $requires(b > 0); $ensures ($result < b); $ensures ($result >= 0);"
        in
        let lspecs =
            Ccommon.parse_spec raw_specs_of_modulo
                               ("include/builtin_functions.h", 9, 3)
                               true
        in
        let scope : (string, Cexpr.expr_t) Hashtbl.t = Hashtbl.create 10 in
        (Hashtbl.add scope "a" (Cexpr.VREF(Cexpr.UNSIGNED_INTEGER, "a", 2, 0)));
        (Hashtbl.add scope "b" (Cexpr.VREF(Cexpr.UNSIGNED_INTEGER, "b", 2, 0)));
        (Hashtbl.add scope
                     Ccommon.return_expr_name
                     (Cexpr.VREF(Cexpr.UNSIGNED_INTEGER,
                                 Ccommon.return_expr_name, 3, 0)));
        let cspecs = Stranslate.cspec_of_lspec lspecs scope in
        match cspecs with
        | Sexpr.SPEC_FUN fspecs ->
            (Cfa.get_agent_of_fspecs fspecs)
        | _ -> assert false)
    in
    (Cfa.CFA_VARDEF
      {
        Cfa.cfa_v_meta_data =
          {
            Cfa.cfa_id = "clap_builtin_modulo";
            Cfa.cfa_tn = Cexpr.FUNCTION_OF (Cexpr.UNSIGNED_INTEGER,
                                            Some [Cexpr.UNSIGNED_INTEGER;
                                                  Cexpr.UNSIGNED_INTEGER]);
            Cfa.cfa_scls = Cexpr.NILSCLS;
            Cfa.cfa_inits = [];
            Cfa.cfa_params = ["a%2"; "b%2"];
            Cfa.cfa_struct_defined_p = false;
            Cfa.cfa_location = None;
            Cfa.cfa_specs = Some (specs_of_builtin_modulo ());
            Cfa.cfa_container = Some builtin_ofile;
            Cfa.cfa_is_cutentry = true
          }
      })

let init_builtin_ofile () = (
    Hashtbl.add builtin_ofile.Cfa.ofile_cfas
                "clap_builtin_modulo"
                (clap_builtin_modulo ()))
