(* la/la_common.ml                                               *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Define common functions to use in la/                                   *)

(** General methods that operate on Cfa, Slicing_path, La_interpolate, Art, 
Lazyabstraction. 
 *)

(*open Form;;*)
open LogicUtil;;
open LogicUtil.Logic;;
   
(* Some auxiliary logging functions *)
let warning (s : string) : unit =
    LogUtil.log LogUtil.LA LogUtil.WARNING s
   
let debug (s : string) : unit =
    LogUtil.log LogUtil.LA LogUtil.DEBUG s

let error (s : string) : unit =
    LogUtil.log LogUtil.LA LogUtil.ERROR s
    
let info (s : string) : unit =
    LogUtil.log LogUtil.LA LogUtil.INFO s

let emerg (s : string) : unit =
    LogUtil.log LogUtil.LA LogUtil.EMERG s


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Find all eq-predicate. 
 - Param: {ul {li {b pred} [predicate type]}}
 - Used by: {!get_strongest_post_eq_condition_for_art_assume} to find all eq 
assume to consider them like assignment. 
 - Pseudo Code: 
{ul {li [Foreach atom-predicate in pred, check if it is eq-predicate then 
save to return]}} 
 - Return: all eq-predicate (predicate list type) 
 - Since: 20090611
 *)
let rec find_eq_predicates (pred : predicate) : predicate list = 
begin
    match pred with
    | And pred_list -> begin
        let eq_preds : predicate list ref = ref [] in
        List.iter 
            (fun a_pred -> 
                let new_eq_preds : predicate list = find_eq_predicates (a_pred) in
                eq_preds := !eq_preds @ new_eq_preds
            )
            pred_list;
        (* return *)
        !eq_preds
      end 
    | Eq (expr1, expr2) -> [pred]
    | Or _ | Not _ | True | False | Lt _ | Leq _ | Atom _ -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Concate level (1) for prime gvars in expression. 
 - Example: prime gvar x -> concated level - gvar x%1.
 - Used by: {!concat_level_for_gvars_in_predicate}.
 - Param: {ul {li {b expr} [(expression type) contains prime gvars]}}
 - Pseudo Code: {ul {li [Search in expr, for each Variable v in expr, 
replace it with Variable v%1]}} 
 - Return: expression contains concated level - gvars
 - Since: 20090611
 *)
let rec concat_level_for_gvars_in_expression (expr : expression) : expression = 
begin
    match expr with
    | Constant f -> expr
    | Variable s -> Variable (s ^ "%1")
    | Application ("sref", str_expr :: field_expr :: []) ->
        let new_str_expr = concat_level_for_gvars_in_expression (str_expr) in
        Application ("sref", new_str_expr :: field_expr :: [])
    | Application ("srefp", pstr_expr :: field_expr :: []) -> 
        let new_pstr_expr = concat_level_for_gvars_in_expression (pstr_expr) in
        Application ("srefp", new_pstr_expr :: field_expr :: [])  
    | Application (s, expr_list) -> 
        let new_expr_list = ref [] in
        List.iter
            (fun an_expr -> 
                let a_new_expr = 
                    concat_level_for_gvars_in_expression (an_expr) in
                new_expr_list := !new_expr_list @ [a_new_expr]
            )
            expr_list;
        Application (s, !new_expr_list)
    | Sum expr_list -> 
        let new_expr_list = ref [] in
        List.iter
            (fun an_expr -> 
                let a_new_expr = 
                    concat_level_for_gvars_in_expression (an_expr) in
                new_expr_list := !new_expr_list @ [a_new_expr]
            )
            expr_list;
        Sum (!new_expr_list)
    | Coeff (f, expr) -> 
        let new_expr = concat_level_for_gvars_in_expression (expr) in
        Coeff (f, new_expr)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Concate level (1) for prime gvars in predicate. 
 - Example: prime gvar x -> concated level - gvar x%1.
 - Param: {ul {li {b pred} [(predicate type) contains prime gvars]}}
 - Pseudo Code: {ul {li [Search in pred, for each Variable v in pred, 
replace it with Variable v%1]}} 
 - Return: predicate contains concated level - gvars
 - Since: 20090611
 *)
let rec concat_level_for_gvars_in_predicate (pred : predicate) : predicate = 
begin
    match pred with
    | True -> True
    | False -> False
    | And pred_list ->
        let new_pred_list = ref [] in
        List.iter 
            (fun a_pred -> 
                let a_new_pred = concat_level_for_gvars_in_predicate (a_pred) in
                new_pred_list := !new_pred_list @ [a_new_pred]
            )
            pred_list;
        And !new_pred_list
    | Or pred_list -> 
        let new_pred_list = ref [] in
        List.iter 
            (fun a_pred -> 
                let a_new_pred = concat_level_for_gvars_in_predicate (a_pred) in
                new_pred_list := !new_pred_list @ [a_new_pred]
            )
            pred_list;
        Or !new_pred_list
    | Not a_pred -> 
        let a_new_pred = concat_level_for_gvars_in_predicate (a_pred) in
        Not a_new_pred
    | Eq (expr1, expr2) -> 
        let new_expr1 = concat_level_for_gvars_in_expression (expr1) in
        let new_expr2 = concat_level_for_gvars_in_expression (expr2) in
        Eq (new_expr1, new_expr2)
    | Lt (expr1, expr2) ->
        let new_expr1 = concat_level_for_gvars_in_expression (expr1) in
        let new_expr2 = concat_level_for_gvars_in_expression (expr2) in
        Lt (new_expr1, new_expr2)
    | Leq (expr1, expr2) ->
        let new_expr1 = concat_level_for_gvars_in_expression (expr1) in
        let new_expr2 = concat_level_for_gvars_in_expression (expr2) in
        Leq (new_expr1, new_expr2)
    | Atom i -> Atom i 
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variable [<varname%level%index>] to [<varname%level>] 
 - Example: prime var x%3%1 -> based var x%3
 - Used by: {ul {li {!vars_to_based_vars}} 
            {li {!create_based_vars_in_expression}}}. 
 - Param: {ul {li {b var} [(string type)]}}
 - Pseudo Code: {ul {li [Find index of last %, then get substring of var
from index 0 to index of last %]}} 
 - Return: sub string of var
 - Since: 20090611
 *)
let var_to_based_var (var : string) : string =
begin
    let rindex = (try String.rindex var '%'
                 with
                 | Not_found -> (*assert false*)(String.length var))
    in
    (* Cut tail to create based var *)
    (try String.sub var 0 rindex
    with
    | Invalid_argument a -> assert false)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variable [<varname%level%index>] to [<varname%level>] 
 - Example: prime var x%3%1 -> based var x%3
 - Param: {ul {li {b vars} [(string list type)]}}
 - Pseudo Code: {ul {li [Foreach var in vars, create based var for it]}} 
 - Return: based vars of vars (string list type)
 - Since: 20090611
 *)
let vars_to_based_vars (vars : string list) : string list = 
begin
    let based_vars : string list ref = ref [] in
    (* Foreach var in vars, create based var for it *)
    List.iter
        (fun a_var ->
             let based_var : string = var_to_based_var (a_var) in
             based_vars := !based_vars @ [based_var]
        ) 
        vars;
    (* Return *)
    !based_vars
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variables by removing index of Variable or of Atom - Application
in an application
 - Example: prime application <*x%3, 1> -> based application *x%3
 - Used by: {!create_based_vars_in_expression}.
 - Param: {ul {li {b an_app} (expression type)}}
 - Pseudo Code: 
{ul {li [<*varname, index> -> <*varname>]}}
{ul {li [<structname.field, index> -> <based_structname.field>]}}
{ul {li [<pointername->field, index> -> <based_pointername->field>]}}
{ul {li [<&varname> -> <&based_varname>]}}
 - Return: based application of an app (expression type)
 - Since: 20090611
 *)
let rec create_based_var_in_application 
(*  21.02.2009. Temporarily to deal with addrof (is_transformed : bool)*)
      (an_app : expression) 
: expression =
begin
    (* check if an_app is an application or not: *)
    match an_app with
    (* 10.3.2009. Xuan has just added to temporarily deal with array *)
    (*| Application ("array", [array_expr; index_expr; version]) -> begin
        (* remain first & second expr *)
        let based_array : expression = 
            create_based_vars_in_expression (is_transformed) (array_expr) in
        let based_index : expression = 
            create_based_vars_in_expression (is_transformed) (index_expr) in
        Application ("array", [based_array; based_index])
      end*)
    | Application ("pref", head :: temp_expr :: tail) -> begin
        (* remain first expr (head) *)
        let new_head : expression = 
            create_based_vars_in_expression (*is_transformed*) (head)  in
        let new_temp_expr : expression = 
            create_based_vars_in_expression (temp_expr)  in
        Application ("pref", [new_head; new_temp_expr])
      end
    | Application ("sref", str :: field :: [_]) -> begin
        (* remain first & second expr (str & field) *)
        let new_str : expression = 
            create_based_vars_in_expression (*is_transformed*) (str)  in
        let new_field = field in
        Application ("sref", new_str :: [new_field])
      end
    | Application ("srefp", points_to_str :: field :: [_]) -> begin
        (* remain first & second expr (str & field) *)
        let new_points_to_str : expression =
            create_based_vars_in_expression (*is_transformed*) (points_to_str) 
        in
        let new_field = field in
        Application ("srefp", new_points_to_str :: [new_field])
      end
    (* 20.2.2009. Xuan has just added to temporarily deal with addrof *)
    | Application ("addrof", [an_expr]) -> begin
        (*if is_transformed = true then
            an_app
        else begin*)
            let based_expr : expression = 
                create_based_vars_in_expression (*is_transformed*) (an_expr) 
            in
            Application ("addrof", [based_expr])
        (*end*)
      end
    (* any-application else, remain the number of expr (expr_list) *)
    | Application (s, expr_list) -> begin
        let new_expr_list : expression list = 
            List.map (create_based_vars_in_expression (*is_transformed*))
                     (expr_list)
        in
        Application (s, new_expr_list)
      end
    (* isn't application *)
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variables by removing index of Variables or Applications in 
in an expression
 - Example: prime application <*x%3, 1> -> based application *x%3
 - Param: {ul {li {b an_expr} (expression type)}}
 - Pseudo Code: {ul {li [foreach var or application in expr, check & remove
level from it]}}
 - Return: based expression (expression type)
 - Since: 20090611
 *)
and create_based_vars_in_expression 
(*  21.02.2009. Temporary to deal with addrof (is_transformed : bool)*)
      (an_expr : expression) 
: expression =
begin
    match an_expr with
    | Constant f -> Constant f
    | Variable s -> Variable (var_to_based_var s)
    | Application (s, expr_list) ->
        create_based_var_in_application (*is_transformed*) (an_expr) 
    | Sum expr_list ->
        let new_expr_list : expression list = 
            List.map (create_based_vars_in_expression (*is_transformed*))
                     (expr_list)
        in
        Sum new_expr_list;
    | Coeff (f, c_expr) ->
        let new_c_expr : expression = 
            create_based_vars_in_expression (*is_transformed*) c_expr in
        Coeff (f, new_c_expr)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variables by removing index of Variable or of Atom - Application
in expressions
 - Example: prime application <*x%3, 1> -> based application *x%3
 - Param: {ul {li {b expr_list} (expression list type)}}
 - Pseudo Code: {ul {li [foreach expr in expr_list, check & remove
level from it]}}
 - Return: based expr_list (expression list type)
 - Since: 20090611
 *)
let create_based_vars_in_expression_list (*is_transformed : bool*)
                                         (expr_list : expression list)
: expression list = 
begin
    let new_expr_list : expression list ref = ref [] in
    (* Foreach expr in expr_list, create based_vars for it *)
    List.iter
        (fun an_expr -> 
            let based_expr : expression =
                create_based_vars_in_expression (*is_transformed*) (an_expr) in
            new_expr_list := !new_expr_list @ [based_expr]; 
        )
        expr_list;
    (* Return *)
    !new_expr_list
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variables by removing index of Variable or of Atom - Application
in a predicate
 - Example: prime application <*x%3, 1> -> based application *x%3
 - Param: {ul {li {b a_pred} (predicate type)}}
 - Pseudo Code: {ul {li [foreach expression in predicate, check & remove
levels from it]}}
 - Return: based predicate (predicate type)
 - Since: 20090611
 *)
let rec create_based_vars_in_predicate 
(*  21.02.2009. Temporary to deal with addrof*) (*is_transformed : bool*)
                                                (a_pred : predicate) 
: predicate =
begin
    match a_pred with
    | True -> True
    | False -> False
    | And pred_list ->
        let new_pred_list : predicate list = 
            List.map (create_based_vars_in_predicate (*is_transformed*))
                     (pred_list)
        in
        And new_pred_list
    | Or pred_list ->
        let new_pred_list : predicate list = 
            List.map (create_based_vars_in_predicate (*is_transformed*))
                     (pred_list)
        in
        Or new_pred_list
    | Not n_pred ->
        let new_p = create_based_vars_in_predicate (*is_transformed*) n_pred in
        Not new_p
    | Eq (expr1, expr2) ->
        let new_expr1 = create_based_vars_in_expression (*is_transformed*) expr1 
        and new_expr2 = create_based_vars_in_expression (*is_transformed*) expr2 
        in
        Eq (new_expr1, new_expr2)
    | Lt (expr1, expr2) ->
        let new_expr1 = create_based_vars_in_expression (*is_transformed*) expr1 
        and new_expr2 = create_based_vars_in_expression (*is_transformed*) expr2  
        in
        Lt (new_expr1, new_expr2)
    | Leq (expr1, expr2) ->
        let new_expr1 = create_based_vars_in_expression (*is_transformed*) expr1 
        and new_expr2 = create_based_vars_in_expression (*is_transformed*) expr2 
        in
        Leq (new_expr1, new_expr2)
    | Atom i -> Atom i
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform variables by removing index of Variable or of Atom - Application
in a list of predicates
 - Example: prime application <*x%3, 1> -> based application *x%3
 - Param: {ul {li {b pred_list} (predicate list type)}}
 - Pseudo Code: {ul {li [foreach predicate in pred_list, check & remove
levels from its expressions]}}
 - Return: based predicate list (predicate list type)
 - Since: 20090611
 *)
let rec create_based_vars_in_predicate_list (*is_transformed : bool*)
                                            (pred_list : predicate list) 
: predicate list =
begin
    let based_pred_list : predicate list ref = ref [] in
    (* Foreach pred in pred_list, create based vars for it *)
    List.iter
        (fun a_pred ->
            let based_pred : predicate = 
                create_based_vars_in_predicate (*is_transformed*) (a_pred) in
            based_pred_list := !based_pred_list @ [based_pred]
        )
        pred_list;
    (* Return *)
    !based_pred_list
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform a list of form_ts to a list of predicates
 - Param: {ul {li {b formlist} (Form.form_t list type)}}
 - Pseudo Code: {ul {li [foreach form_t in formlist, transform it to predicate, 
then update to predlist to return]}}
 - Return: a list of predicates (predicate list type)
 - Since: 20090611
 *)
let predlist_of_formlist (formlist : Form.form_t list) : predicate list = 
begin
    let predlist : predicate list ref = ref [] in
    (* Foreach form in formlist -> translate to predicate *)
    List.iter
        (fun aform ->
            let new_pred : predicate = Form.form_to_predicate (aform) in
            predlist := !predlist @ [new_pred]
        )
        formlist;
    (* Return *)
    !predlist
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform a list of term_ts to a list of expressions
 - Param: {ul {li {b termlist} (Form.term_t list type)}}
 - Pseudo Code: {ul {li [foreach term_t in termlist, transform it to expression, 
then update to expression list to return]}}
 - Return: expression list (expression list type)
 - Since: 20090611
 *)
let exprlist_of_termlist (termlist : Form.term_t list) : expression list = 
begin
    let exprlist : expression list ref = ref [] in
    List.iter
        (fun a_term -> 
            (*let a_term : Form.term_t = Translate.translate_to_term (a_cexpr) in*)
            let an_expr : expression = Form.term_to_expr (a_term) in
            exprlist := !exprlist @ [an_expr];
        )
        termlist;
    (* Return *)
    !exprlist
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compare an expression with an other expression.
 - Param: {ul {li {b expr1} (expression type)}}
          {ul {li {b expr2} (expression type)}}
 - Example: (compare x x) = 0; (compare x y) = 1
 - Pseudo Code: {ul {li [Create a predicate {expr1 == expr2}, then check valid 
to judge the exprs are same or not]}}
 - Return: 0 if the exprs are equal, 1 otherwise
 - Since: 20090611
 *)
let compare_expr_with_expr (expr1 : expression) (expr2 : expression) : int = 
begin
    (* Step1. Create eq *)
    let eq : predicate = Eq (expr1, expr2) in
    (* Step2. Check valid *)
    if (is_valid eq) = true then 0
    else 1
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if 2 predicates are equivalent or not
 - Param: {ul {li {b pred1} (predicate type)}}
          {ul {li {b pred2} (predicate type)}}
 - Pseudo Code: {ul {li [Create 2 predicates: {pred1 => pred2}, 
{pred2 => pred1}; then check valid the predicates to judge the predicates 
are equivalent or not]}}
 - Return: 0 if the pred1 are equivalent to pred2, 1 otherwise
 - Since: 20090611
 *)
let is_equivalent_predicates (pred1 : predicate) (pred2 : predicate) : int =
begin
    (* debug *)
    (*debug ("is_equivalent_predicates");
    debug ("pred1: " ^ (string_of_predicate pred1));
    debug ("pred2: " ^ (string_of_predicate pred2));*)

    (* pred1 <=> pred2, iff ((pred1 => pred2) & (pred2 => pred1) valid *)
    let pred_to_check_valid1 = Or ((Not pred1) :: [pred2]) in
    let pred_to_check_valid2 = Or ((Not pred2) :: [pred1]) in
    let valid1 = is_valid pred_to_check_valid1 in
    let valid2 = is_valid pred_to_check_valid2 in
    let is_equivalent : int = (
        if (valid1 = true) && (valid2 = true) then 0
        else 1
    ) in
    (* debug *)
    (*debug (string_of_int is_equivalent);*)
    (* return *)
    is_equivalent
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if 2 expressions are identical or not
 - Param: {ul {li {b expr1} (expression type)}}
          {ul {li {b expr2} (expression type)}}
 - Pseudo Code: {ul {li [try to match 2 expressions]}}
 - Return: false if can not match 2 expressions, true otherwise
 - Since: 20090611
 *)
let rec is_identical_expressions (expr1 : expression) (expr2 : expression)
: bool = 
begin
    match expr1 with
    | Constant f1 -> (match expr2 with
                     | Constant f2 -> if f1 = f2 then true else false
                     | anything_else -> false)
    | Variable v1 -> (match expr2 with
                     | Variable v2 -> if (String.compare v1 v2) = 0 then true
                                      else false
                     | anything_else -> false)
    | Application (s1, expr_list1) -> begin
        (match expr2 with
        | Application (s2, expr_list2) -> begin
            if (String.compare s1 s2) = 0 then begin
                is_identical_expression_lists (expr_list1) (expr_list2)    
            end else false
          end
        | anything_else -> false)
      end
    | Sum expr_list1 -> begin
        (match expr2 with
        | Sum expr_list2 -> 
            is_identical_expression_lists (expr_list1) (expr_list2)
        | anything_else -> false)
      end
    | Coeff (f1, an_expr1) -> begin
        (match expr2 with
        | Coeff (f2, an_expr2) -> begin
            if f1 = f2 then is_identical_expressions (an_expr1) (an_expr2)
            else false
          end
        | anything_else -> false)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if 2 expression list are identical or not
 - Param: {ul {li {b expr_list1} (expression list type)}}
          {ul {li {b expr_list2} (expression list type)}}
 - Pseudo Code: {ul {li [try to match 2 expression list respectively, then 
foreach tuple of (expr of list1, expr of list2), check if they are identical
or not. ]}}
 - Return: false if can not match 2 expression list, or exist a tuple of 
(expr of list1, expr of list2) not identical, true otherwise
 - Since: 20090611
 *)
and is_identical_expression_lists (expr_list1 : expression list)
                                  (expr_list2 : expression list)
: bool = 
begin
    try
        (let is_identical : bool ref = ref true in
        List.iter2
            (fun e1 e2 -> 
                if !is_identical = true then begin
                    is_identical := is_identical_expressions (e1) (e2)
                end
            )
            expr_list1
            expr_list2;
        (* return *)
        !is_identical
    )
    with
    | Invalid_argument _ -> false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if 2 predicates are identical or not
 - Param: {ul {li {b pred11} (predicate type)}}
          {ul {li {b pred2} (predicate type)}}
 - Pseudo Code: {ul {li [try to match 2 predicates]}}
 - Return: false if can not match 2 predicates, true otherwise
 - Since: 20090611
 *)
let rec is_identical_predicates (pred1 : predicate) (pred2 : predicate)
: bool = 
begin
    match pred1 with
    | True -> (match pred2 with
              | True -> true
              | anything_else -> false)    
    | False -> (match pred2 with
               | False -> true
               | anything_else -> false)
    | And pred_list1 -> begin
        (match pred2 with
        | And pred_list2 -> is_identical_predicate_lists (pred_list1) (pred_list2)
        | anything_else -> false)
      end
    | Or pred_list1 -> begin
        (match pred2 with
        | Or pred_list2 -> is_identical_predicate_lists (pred_list1) (pred_list2)
        | anything_else -> false)
      end
    | Not apred1 -> (match pred2 with
                    | Not apred2 -> is_identical_predicates (apred1) (apred2)
                    | anything_else -> false)
    | Eq (expr11, expr12) -> begin
        (match pred2 with
        | Eq (expr21, expr22) -> begin
            if (is_identical_expressions expr11 expr21) = true then
                is_identical_expressions (expr12) (expr22)
            else false
          end
        | anything_else -> false)
      end
    | Lt (expr11, expr12) -> begin
        (match pred2 with
        | Lt (expr21, expr22) -> begin
            if (is_identical_expressions expr11 expr21) = true then
                is_identical_expressions (expr12) (expr22)
            else false
          end
        | anything_else -> false)
      end
    | Leq (expr11, expr12) -> begin
        (match pred2 with
        | Leq (expr21, expr22) -> begin
            if (is_identical_expressions expr11 expr21) = true then
                is_identical_expressions (expr12) (expr22)
            else false
          end
        | anything_else -> false)
      end
    | Atom i1 -> (match pred2 with
                 | Atom i2 -> if i1 = i2 then true else false
                 | anything_else -> false)  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if 2 predicate list are identical or not
 - Param: {ul {li {b pred_list1} (predicate list type)}}
          {ul {li {b pred_list2} (predicate list type)}}
 - Pseudo Code: {ul {li [try to match 2 expression list respectively, then 
foreach tuple of (pred of list1, pred of list2), check if they are identical
or not. ]}}
 - Return: false if can not match 2 predicate list, or exist a tuple of 
(pred of list1, pred of list2) not identical, true otherwise
 - Since: 20090611
 *)
and is_identical_predicate_lists (pred_list1 : predicate list)
                                 (pred_list2 : predicate list)
: bool = 
begin
    try
        (let is_identical : bool ref = ref true in
        List.iter2
            (fun p1 p2 -> 
                if !is_identical = true then begin
                    is_identical := is_identical_predicates (p1) (p2)     
                end
            )
            pred_list1
            pred_list2;
        (* return *)
        !is_identical
        )
    with
    | Invalid_argument _ -> false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if a expression is member of a list of expressions
 - Param: {ul {li {b expr} (expression type)}}
          {ul {li {b expr_list} (expression list type)}}
 - Pseudo Code: {ul {li [for each expression in expr_list, compare it with
the expr.]}}
 - Return: If expr_list has an expression same with the expr then return true, 
else otherwise
 - Since: 20090611
 *)
let is_mem_of_expr_list (expr : expression) (expr_list : expression list)
: bool = 
begin
    let is_mem : bool ref = ref false in
    (* Foreach expression in expr_list, compare it with input-expr *)
    List.iter 
        (fun an_expr ->
            if !is_mem = false then begin
                if (compare_expr_with_expr (expr) (an_expr)) = 0 then 
                    is_mem := true
            end 
        )
        expr_list;
    (* Return *)
    !is_mem
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let member_of_predicate_list (pred : predicate) (preds : predicate list)
: bool = begin
    List.fold_left (fun mem p ->
                        if mem then mem
                        else if (is_equivalent_predicates p pred) = 0 then
                                 true
                             else
                                 false) 
                   false preds
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec remove_duplicated_exprs (exprs : expression list) : expression list =
begin
    match exprs with
    | e :: tail -> (
        if (is_mem_of_expr_list e tail) = true then
	    remove_duplicated_exprs tail
	else
	    e :: (remove_duplicated_exprs tail))
    | _ -> exprs
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if a predicate is identical with an element of a predicate list
 - Param: {ul {li {b pred} (predicate type)}}
          {ul {li {b pred_list} (predicate list type)}}
 - Pseudo Code: {ul {li [for each predicate in pred_list, check if it is 
identical with the pred or not.]}}
 - Return: if pred_list has a predicate identical with the pred then return true, 
else otherwise.
 - Since: 20090611
 *)
let is_identical_element_of_pred_list (pred : predicate) 
                                      (pred_list : predicate list)
: bool = 
begin
    let is_mem : bool ref = ref false in
    (* Foreach predicate in pred_list, compare it with input pred *)
    List.iter
        (fun a_pred ->
            if !is_mem = false then begin
                if (is_identical_predicates (a_pred) (pred)) = true then 
                    is_mem := true
            end
        )
        pred_list;
    (* return *)
    !is_mem
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Remove an expression form a list of expressions
 - Param: {ul {li {b an_expr} (expression type)}}
          {ul {li {b expr_list} (expression list type)}}
 - Pseudo Code: {ul {li [for each expression in expr_list, check if it is 
same with the an_expr then remove it form expr_list.]}}
 - Return: expression list removed an_expr (expression list type)
 - Since: 20090611
 *)
let remove_expr_from_expr_list (an_expr : expression)
                               (expr_list : expression list)
: expression list = 
begin
    (* Check if an_expr is an element of expr_list *)
    (*let is_mem : bool = is_mem_of_expr_list (an_expr) (expr_list) in
    if is_mem = true then begin*)
        let new_expr_list : expression list ref = ref [] in
        List.iter
            (fun e -> 
                if (compare_expr_with_expr e an_expr) != 0 then
                    new_expr_list := !new_expr_list @ [e]
            )
            expr_list;
        (* return *)
        !new_expr_list                
    (*end else (*assert false*) expr_list*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Merge two predicate lists
 - Param: {ul {li {b pred_list1} (predicate list type)}}
          {ul {li {b pred_list2} (predicate list type)}}
 - Pseudo Code: {ul {li [for each predicate in pred_list1, check if it is
identical with any element of pred_list2 or not. If not, then add it to 
pred_list2 to return]}}
 - Return: merged predicate list (predicate list type)
 - Since: 20090611
 *)
let merge_two_predicate_lists (pred_list1 : predicate list)
                              (pred_list2 : predicate list)
: predicate list = 
begin
    let new_pred_list : predicate list ref = ref pred_list2 in
    (* Foreach predicate n pred_list1, check if it is identical with one *)
    (* element of pred_list2 or not. If not, add it to new_pred_list     *)
    List.iter
        (fun pred1 -> 
            let flag = is_identical_element_of_pred_list pred1 !new_pred_list in 
            if flag = false then new_pred_list := !new_pred_list @ [pred1];
        )
        pred_list1;
    (* Return *)
    !new_pred_list  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Check if a predicate list has an element which is mem of another 
predicate list.
 - Param: {ul {li {b pred_list1} (predicate list type)}}
          {ul {li {b pred_list2} (predicate list type)}}
 - Since: 20090611
 *)
let an_element_of_exprs_is_mem (expr_list_1 : expression list)
                               (expr_list_2 : expression list)
: bool = 
begin
    let is_mem : bool ref = ref false in
    (* Foreach expr in expr_list_1, check if it is mem of expr_list_2 or not *)
    List.iter
        (fun an_expr1 -> 
            if !is_mem = false then begin
                is_mem := is_mem_of_expr_list (an_expr1) (expr_list_2)
            end  
        )
        expr_list_1;
    (* Return *)
    !is_mem
end
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Find all vars & atom applications in an application. Note that atom 
application is pref, or addrof, sref, srefp.
 - Param: {ul {li {b app} (expression type)}}
 - Pseudo Code: {ul {li [Check if app is an atom application then return it,
else, find all vars in expr list of the app]}}
 - Return: list of vars & atom applications (expression list type)
 - Since: 20090611
 *)
(* This function is used in slicing path                                   *)
let rec find_all_vars_in_application (app : expression) : expression list =
begin
    match app with
    | Application ("pref", [pref_expr; temp_expr; index_expr]) -> [app]
    | Application ("pref", [pref_expr; temp_expr]) -> [app]
    | Application ("addrof", [addrof_expr]) -> [app]
    | Application ("sref", [str_expr; field_expr; index_expr]) -> [app]
    | Application ("sref", [str_expr; field_expr]) -> [app]
    | Application ("srefp", [pstr_expr; field_expr; index_expr]) -> [app]
    | Application ("srefp", [pstr_expr; field_expr]) -> [app]
    (* 20090620. Temp *)
    | Application ("array", [memory_expr; index_expr]) -> (*[app]*) begin
        let vars_in_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let vars_in_an_expr = find_all_vars_in_expression (an_expr) in
                vars_in_expr_list := !vars_in_expr_list @ vars_in_an_expr
            )
            [memory_expr; index_expr];
        (* return list *)
        app :: !vars_in_expr_list
      end
    | Application (_, expr_list) -> (begin
        let vars_in_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let vars_in_an_expr = find_all_vars_in_expression (an_expr) in
                vars_in_expr_list := !vars_in_expr_list @ vars_in_an_expr
            )
            expr_list;
        (* return list *)
        !vars_in_expr_list
      end)
    | not_application -> assert false
end
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Find all vars & atom applications in an expression. Note that atom 
application is pref, or addrof, sref, srefp.
 - Param: {ul {li {b expr} (expression type)}}
 - Pseudo Code: {ul {li [Check if expr is an atom application or a variable, 
then return it, else find all vars in expr list of the expr]}}
 - Return: list of vars & atom applications (expression list type)
 - Since: 20090611
 *)
and find_all_vars_in_expression (expr : expression) : expression list =
begin
    match expr with
    | Constant f -> []
    | Variable s -> [Variable s]
    | Application (s, expr_list) -> find_all_vars_in_application (expr)
    | Sum expr_list -> begin
        let vars_in_expr_list = ref [] in
        List.iter
            (fun a_expr ->
                let vars_in_a_expr = find_all_vars_in_expression (a_expr) in
                vars_in_expr_list := !vars_in_expr_list @ vars_in_a_expr
            )
            expr_list;
        (* return list *)
        !vars_in_expr_list
      end
    | Coeff (f, a_expr) -> find_all_vars_in_expression (a_expr)
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Find all vars & atom applications in a list of predicates. Note that atom 
application is pref, or addrof, sref, srefp.
 - Param: {ul {li {b pred_list} (prediate list type)}}
 - Pseudo Code: {ul {li [Foreach predicate in pred_list, find all vars in it]}}
 - Return: list of vars & atom applications (expression list type)
 - Since: 20090611
 *)
let rec find_all_vars_in_predicate_list (pred_list : predicate list)
: expression list =
begin
    let vars_in_pred_list = ref [] in
    List.iter
        (fun a_pred ->
            let vars_in_a_pred = find_all_vars_in_predicate (a_pred) in
            List.iter  
                (fun a_var -> 
                    let is_mem : bool = 
                        is_mem_of_expr_list (a_var) (!vars_in_pred_list) in
                    if is_mem = false then begin
                        vars_in_pred_list := a_var :: !vars_in_pred_list
                    end  
                )
                vars_in_a_pred;
        )
        pred_list;
    (* return *)
    !vars_in_pred_list
end
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Find all vars & atom applications in a predicate. Note that atom 
application is pref, or addrof, sref, srefp.
 - Param: {ul {li {b pred} (prediate type)}}
 - Pseudo Code: {ul {li [Foreach expression in pred, find all vars in it]}}
 - Return: list of vars & atom applications (expression list type)
 - Since: 20090611
 *)
and find_all_vars_in_predicate (pred : predicate) : expression list =
begin
    match pred with
    | True | False -> []
    | And pred_list -> find_all_vars_in_predicate_list (pred_list)
    | Or pred_list -> find_all_vars_in_predicate_list (pred_list)
    | Not a_pred -> find_all_vars_in_predicate (a_pred)
    | Eq (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_vars_in_expression (expr1) in
        let vars_in_expr2 = find_all_vars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Lt (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_vars_in_expression (expr1) in
        let vars_in_expr2 = find_all_vars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Leq (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_vars_in_expression (expr1) in
        let vars_in_expr2 = find_all_vars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Atom i -> assert false
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Get atom_predicates from one predicate. Atom predicate is (reference 
[http://en.wikipedia.org/wiki/First-order_logic]):
 1. Non-logical predicates If P is an n-ary predicate symbol and t1, ..., tn 
are terms then P(t1,...,tn) is a well-formed formula.
 2. Equality: If the equality symbol is considered part of logic, and t1 and 
t2 are terms, then t1 = t2 is a well-formed formula.
 This function is used to simlify support predicates after call "interpolate".  
 - Param: {ul {li {b pred} (prediate type)}}
 - Pseudo Code: {ul {li [try to match pred with type of atom prediate defined
above. If it isn't matched, then get atom predicates from sub - predicates of
it]}}
 - Return: list of atom predicate (predicate list type)
 - Since: 20090611
 *)
let rec get_atom_predicates_from_predicate (pred : predicate) : predicate list =
begin
    match pred with
    | True -> [True]
    | False -> [False]
    | And pred_list -> begin
        let new_atom_preds : predicate list ref = ref [] in
        List.iter
            (fun a_pred ->
                let atom_preds_in_a_pred : predicate list =
                    get_atom_predicates_from_predicate (a_pred) in
                new_atom_preds := !new_atom_preds @ atom_preds_in_a_pred
            )
            pred_list;
        (* return *)
        !new_atom_preds
      end
    | Or pred_list -> begin [Or pred_list]
        (*let new_atom_preds : predicate list ref = ref [] in
        List.iter
            (fun a_pred ->
                 let atom_preds_in_a_pred : predicate list =
                     get_atom_predicates_from_predicate (a_pred) in
                 new_atom_preds := !new_atom_preds @ atom_preds_in_a_pred
            )
            pred_list;*)
        (* return *)
        (*!new_atom_preds*)
      end
    | Not pred_of_not -> (*get_atom_predicates_from_predicate (pred_of_not)*)
        [Not pred_of_not]
    | Eq (expr1, expr2) -> [Eq (expr1, expr2)]
    | Lt (expr1, expr2) -> [Lt (expr1, expr2)]
    | Leq (expr1, expr2) -> [Leq (expr1, expr2)]
    | Atom i -> [Atom i]
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Find all vars & atom applications in a predicate. Note that atom 
application is pref, or addrof, sref, srefp.
 - Param: {ul {li pred (prediate type)}}
 - Used by: module {!Slicing_path}
 - Pseudo Code: {ul {li [Foreach expression in pred, find all vars in it]}}
 - Return: list of vars & atom applications (expression list type)
 - Since: 20090611
 *)
(* This function is used in change_index_of_variables_in_assume            *)
(*let rec find_all_resetvars_in_application (app : expression) : expression list =
begin
    match app with
    (*| Application ("pref", [pref_expr; index_expr]) -> [app]*)
    | Application ("pref", [pref_expr]) -> [app]
    | Application ("addrof", [addrof_expr]) -> [app]
    (*| Application ("sref", [str_expr; field_expr; index_expr]) -> [app]*)
    | Application ("sref", [str_expr; field_expr]) -> [app]
    (*| Application ("srefp", [pstr_expr; field_expr; index_expr]) -> [app]*)
    | Application ("srefp", [pstr_expr; field_expr]) -> [app]
    | Application ("array", [memory_expr; index_expr]) -> [app]
    | Application (_, expr_list) -> begin
        let vars_in_expr_list : expression list ref = ref [] in
        List.iter
            (fun a_expr ->
                let vars_in_a_expr : expression list = 
                    find_all_resetvars_in_expression (a_expr) in
                vars_in_expr_list := !vars_in_expr_list @ vars_in_a_expr
            )
            expr_list;
        (* return list *)
        !vars_in_expr_list
      end
    | not_application -> assert false
end
*)        

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in change_index_of_variables_in_assume            *)
(*and find_all_resetvars_in_expression (expr : expression) : expression list =
begin
    match expr with
    | Constant f -> []
    | Variable s -> [Variable s]
    | Application (s, expr_list) -> find_all_resetvars_in_application (expr)
    | Sum expr_list -> begin
        let vars_in_expr_list : expression list ref = ref [] in
        List.iter
            (fun a_expr ->
                let vars_in_a_expr : expression list = 
                    find_all_resetvars_in_expression (a_expr) in
                vars_in_expr_list := !vars_in_expr_list @ vars_in_a_expr
            )
            expr_list;
        (* return list *)
        !vars_in_expr_list
      end
    | Coeff (f, a_expr) -> find_all_resetvars_in_expression (a_expr)
end
*) 

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in change_index_of_variables_in_assume            *)
(*let rec find_all_resetvars_in_predicate_list (pred_list : predicate list)
: expression list =
begin
    let vars_in_pred_list : expression list ref = ref [] in
    List.iter
        (fun a_pred ->
            let vars_in_a_pred = find_all_resetvars_in_predicate (a_pred) in
            List.iter  
                (fun a_var -> 
                    let is_mem : bool = 
                        is_mem_of_expr_list (a_var) (!vars_in_pred_list) in
                    if is_mem = false then begin
                        vars_in_pred_list := a_var :: !vars_in_pred_list
                    end  
                )
                vars_in_a_pred;
        )
        pred_list;
    (* return *)
    !vars_in_pred_list
end
*)
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in change_index_of_variables_in_assume            *)
(*and find_all_resetvars_in_predicate (pred : predicate) : expression list =
begin
    match pred with
    | True | False -> []
    | And pred_list -> find_all_resetvars_in_predicate_list (pred_list)
    | Or pred_list -> find_all_resetvars_in_predicate_list (pred_list)
    | Not a_pred -> find_all_resetvars_in_predicate (a_pred)
    | Eq (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_resetvars_in_expression (expr1) in
        let vars_in_expr2 = find_all_resetvars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Lt (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_resetvars_in_expression (expr1) in
        let vars_in_expr2 = find_all_resetvars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Leq (expr1, expr2) -> begin
        let vars_in_expr1 = find_all_resetvars_in_expression (expr1) in
        let vars_in_expr2 = find_all_resetvars_in_expression (expr2) in
        vars_in_expr1 @ vars_in_expr2
      end
    | Atom i -> assert false
end*)
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest post condition for an assignment. Pre-condition must be
an equality predicate. 
 - Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
 - Param: {ul {li {b pre_eq_condition} [predicate type]}
              {li {b assignment} [predicate type]}}
 - Pseudo Code: 
{ul {li [If pre-cond is True then return the assignment]} 
    {li [If pre-cond isn't True then replace left of pre_eq_condition in right
of assignment by right of pre_eq_condition. If left of pre_eq_condition is 
identical with left of assignment, then return the replaced assignment, else 
return the replaced assignment & the pre_eq_condition ]}}
 - Return: strongest post eq condition (predicate list type)
 - Since: 20090611
 *)
let get_strongest_post_eq_condition_for_assignment (pre_eq_condition : predicate) 
                                                   (assignment : predicate)
: predicate list = 
begin
    (* Check pre_eq_condition is True or not *)
    if (normalize pre_eq_condition) = True then begin
        (match assignment with
        | Eq (_, _) -> [assignment]
        | True -> [True]
        | _ -> assert false)
    (* If pre_eq_condition isn't True *)
    end else begin
        (* Check assignment *)
        (match assignment with
        | Eq (left_of_ass, right_of_ass) -> begin
            (* Step2. Check if pre_eq_condition is eq or not *)
            (match pre_eq_condition with
            | Eq (left_precond, right_precond) -> begin
                (* Check left_precond if it changed or didn't *)
                let left_precond_changed : bool = 
                    is_identical_expressions (left_of_ass) (left_precond) in
                (* Replace left_precond in right_of_ass by right_precond *)
                let new_right_of_ass : expression =  
                    replace_var_in_expression (right_of_ass) (left_precond) 
                      (right_precond) in
                (* Returns *)
                if left_precond_changed = true then begin
                    [Eq (left_of_ass, new_right_of_ass)]
                end else begin
                    [pre_eq_condition; Eq (left_of_ass, new_right_of_ass)]
                end
              end
            | anything_else -> assert false)      
          end
        (* If input is not an assignment but not condition: *)
        | True -> [pre_eq_condition]
        (* If input is a condition *)
        | anything_else -> begin
            emerg (string_of_predicate assignment);
            assert false
          end);
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest post conditions for an assignment. Pre-conditions must be
equality predicates. 
 - Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
 - Param: {ul {li {b pre_eq_conditions} [predicate list type]}
              {li {b assignment} [predicate type]}}
 - Pseudo Code: 
{ul {li [If pre-conds is null then raise an error]} 
    {li [Else, foreach pre-cond of pre_eq_conditions, replace left of the 
pre-cond in right of assignment by right of the pre-cond. If left of 
pre_eq_condition is identical with left of assignment, then save replaced 
assignment, else save the replaced assignment & the pre_eq_condition. After all,
add the assignment.]}}
 - Return: strongest post eq conditions (predicate list type)
 - Since: 20090611
 *)
let get_strongest_post_eq_conditions_for_assignment 
      (pre_eq_conditions : predicate list)
      (assignment : predicate)
: predicate list = 
begin
    (* Debug *)
    (*debug ("get_strongest_post_eq_conditions_for_assignment");
    debug ("pre_eq_conditions: " ^ (string_of_predicate (And pre_eq_conditions)));
    debug ("assignment: " ^ (string_of_predicate assignment));*)
    (*let _ = read_int () in*)

    (* Check if pre_eq_conditions is null *)
    if pre_eq_conditions = [] then assert false;
    (* Create vars: *)
    let spcs : predicate list ref = ref [] in
    let is_added_ass : bool ref = ref false in
    (* Foreach pre_eq_condition, compute strongest post eq conditions *)
    List.iter
        (fun a_pre_eq_condition ->
            (* Check pre_eq_condition is True or not *)
            if (normalize a_pre_eq_condition) != True then begin 
                (* Check assignment *)
                (match assignment with
                | Eq (left_of_ass, right_of_ass) -> begin
                    (* Step2. Check if pre_eq_condition is eq or not *)
                    (match a_pre_eq_condition with
                    | Eq (left_precond, right_precond) -> begin
                        (* Check left_precond if it changed or didn't *)
                        let left_precond_changed : bool = 
                            is_identical_expressions (left_of_ass) (left_precond) 
                        in
                        (* Replace left_precond in right_of_ass by right_precond *)
                        let new_right_of_ass : expression =  
                            replace_var_in_expression (right_of_ass) 
                              (left_precond) (right_precond) in
                        (* Check if the replacing is successful *)
                        let is_identical : bool = 
                            is_identical_expressions (right_of_ass) (new_right_of_ass)
                        in 
                        (* Returns *)
                        if is_identical = true then begin
                             spcs := !spcs @ [a_pre_eq_condition];
                        end else if left_precond_changed = true then begin
                            spcs := !spcs @ [Eq (left_of_ass, new_right_of_ass)];
                            is_added_ass := true;
                        end else begin
                            spcs := !spcs @ [a_pre_eq_condition]; 
                            spcs := !spcs @ [Eq (left_of_ass, new_right_of_ass)];
                            is_added_ass := true;
                        end
                      end
                    | anything_else -> assert false)      
                  end
                (* If input is not an assignment but not condition: *)
                | True -> spcs := !spcs @ [a_pre_eq_condition]
                (* If input is a condition *)
                | anything_else -> begin 
                    emerg (string_of_predicate assignment); 
                    assert false;
                  end);
            end
        )
        pre_eq_conditions;
    (* Check is_added_ass to add assignment *)
    if !is_added_ass = false then begin
        spcs := !spcs @ [assignment];
    end;
    (* Return *)
    !spcs    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest post conditions for assignments. Pre-conditions must be
equality predicates. 
 - Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
 - Param: {ul {li {b pre_eq_conditions} [predicate list type]}
              {li {b assignments} [predicate list type]}}
 - Pseudo Code: 
{ul {li [If assignments are null, then return pre_eq_conditios]} 
    {li [Else, compute forward. Foreach assignment, compute strongest-post-eq-
conditions with pre-eq-conditions are strongest-post-eq-conditions of the 
before assignment. Note that compute strongest-post-eq-condition for first 
assignment with pre-eq-conditions are the input]}}
 - Return: strongest post eq conditions (predicate list type)
 - Since: 20090611
 *)
let get_strongest_post_eq_conditions_for_assignments 
      (pre_eq_conditions : predicate list)
      (assignments : predicate list)
: predicate list = 
begin
    (* Debug *)
    (*debug ("get_strongest_post_eq_conditions_for_assignments: ");
    debug ("pre_eq_conditions: " ^ (string_of_predicate (And pre_eq_conditions)));
    debug ("assignments: " ^ (string_of_predicate (And assignments)));*)

    if assignments = [] then pre_eq_conditions
    else begin
        let spcs : predicate list ref = ref pre_eq_conditions (*[]*) in
        (*let current_pre_eq_conditions : predicate list ref = ref pre_eq_conditions in*)
        (* Foreach assignment, compute strongest post conditions, then update *)
        (* them with current_pre_eq_conditions                                *)
        List.iter
            (fun (an_assignment : predicate) ->
                (* Step1. Compute strongest post conditions *)
                let new_spcs : predicate list = 
                    get_strongest_post_eq_conditions_for_assignment
                      (!spcs(*current_pre_eq_conditions*)) (an_assignment)
                in
                (* Update new_spcs with current_pre_eq_conditions *)
                (*spcs := merge_two_predicate_list (new_spcs) (!current_pre_eq_conditions);
                current_pre_eq_conditions := !spcs;*)
                spcs := new_spcs;    
            )
            assignments;
        (* Return *)
        !spcs
    end   
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest post conditions for edge of assignments with a 
pre_eq_condition. The pre-condition must be an equality predicate. 
 - Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
 - Param: {ul {li {b pre_eq_condition} [predicate type]}
              {li {b assignments_edge} [predicate type]}}
 - Pseudo Code: 
{ul {li [If the edge is True then return [pre_eq_condition]]} 
    {li [Else, if the edge contains an assignment then call 
{!get_strongest_post_eq_conditions_for_assignment}]}
    {li [Else, if the edge contains a list of assignments then call 
{!get_strongest_post_eq_conditions_for_assignments}]}}
 - Return: strongest post eq conditions (predicate list type)
 - Since: 20090611
 *)
let rec get_strongest_post_eq_condition_for_assignments_edge 
    (pre_eq_condition : predicate)
    (assignments_edge : predicate)
: predicate list = 
begin
    (* Debug *)
    (*debug ("get_strongest_post_eq_condition_for_assignments_edge");
    debug ("pre_eq_condition: " ^ (string_of_predicate pre_eq_condition));
    debug ("assignments_edge: " ^ (string_of_predicate assignments_edge));*)

    match assignments_edge with
    | True -> [pre_eq_condition]
    | Eq (expr1, expr2) -> 
        get_strongest_post_eq_conditions_for_assignment ([pre_eq_condition])
          (assignments_edge)
    | And pred_list -> 
        get_strongest_post_eq_conditions_for_assignments ([pre_eq_condition])
          (pred_list)
    (* in other cases, assert false *)
    | False | Or _ | Not _ | Lt _ | Leq _ | Atom _ -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest post conditions for edge of assignments with a list of
pre_eq_conditions. The pre-eq-conditions must be equality predicates. 
 - Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
 - Param: {ul {li {b pre_eq_conditions} [predicate list type]}
              {li {b assignments_edge} [predicate type]}}
 - Pseudo Code: 
{ul {li [If the edge is True then return pre_eq_conditions]} 
    {li [Else, if the edge contains an assignment then call 
{!get_strongest_post_eq_conditions_for_assignment}]}
    {li [Else, if the edge contains a list of assignments then call 
{!get_strongest_post_eq_conditions_for_assignments}]}}
 - Return: strongest post eq conditions (predicate list type)
 - Since: 20090611
 *)
let rec get_strongest_post_eq_conditions_for_assignments_edge 
    (pre_eq_conditions : predicate list)
    (assignments_edge : predicate)
: predicate list = 
begin
    (* Debug *)
    (*debug ("get_strongest_post_eq_conditions_for_assignments_edge");
    debug ("pre_eq_conditions: " ^ (string_of_predicate (And pre_eq_conditions)));
    debug ("assignments_edge: " ^ (string_of_predicate assignments_edge));*)
    (* Find eq_predicate in assignments edge *)
    let eq_preds : predicate list = find_eq_predicates (assignments_edge) in
    get_strongest_post_eq_conditions_for_assignments (pre_eq_conditions)
          (eq_preds)
    (*match assignments_edge with
    | True -> pre_eq_conditions
    | Eq (expr1, expr2) -> 
        get_strongest_post_eq_conditions_for_assignment (pre_eq_conditions)
          (assignments_edge)
    | And pred_list -> 
        get_strongest_post_eq_conditions_for_assignments (pre_eq_conditions)
          (pred_list)
    (* in other cases, assert false *)
    | False | Or _ | Not _ | Lt _ | Leq _ | Atom _ -> assert false*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Choose eq-predicate which involves to value of var in var_critera. Note
that eq_predicate is a static single assignment (ssa). 
 - Example: 
{ul {li [pred = {x = 1}; var_criteria = {x}; eq-predicate = {x = 1}]}
    {li [pred = {x = 1}; var_criteria = {y}; eq-predicate = null]}}         
 - Param: {ul {li {b pred} [predicate type]}
              {li {b var_criteria} [expression list type]}}
 - Pseudo Code: 
{ul {li [If the pred is an assignment, and left of the assignment is member of
var_critera, then return Some pred, else return None]}} 
 - Return: predicate option type
 - Since: 20090611
 *)
let choose_eq_predicate_with_criteria (pred : predicate)
                                      (var_criteria : expression list)
: predicate option = 
begin
    match pred with
    | Eq (left_expr, right_expr) -> begin
        if (is_mem_of_expr_list left_expr var_criteria) = true then Some pred
        else None
      end
    | anything_else -> None  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Choose eq-predicates which involve to value of vars in var_critera. Note
that eq_predicates is static single assignments (ssa). 
 - Example: 
{ul {li [pred = {x = 1}; var_criteria = {x}; eq-predicate = {x = 1}]}
    {li [pred = {x = 1}; var_criteria = {y}; eq-predicate = null]}}         
 - Param: {ul {li {b pred_list} [predicate list type]}
              {li {b var_criteria} [expression list type]}}
 - Pseudo Code: 
{ul {li [Foreach pred in pred_list, check if it is involve to value of vars in
var_criteria then save it to list to return, else ignore it]}} 
 - Return: predicate list (the result maybe null) 
 - Since: 20090611
 *)
let choose_eq_predicates_with_criteria (pred_list : predicate list)
                                       (var_criteria : expression list)
: predicate list = 
begin
    let new_pred_list : predicate list ref = ref [] in
    (* Foreach pred in pred_list, check if it contains one var_criteria in *)
    (* the left side or not                                                *)
    List.iter
        (fun pred -> 
            let po : predicate option = 
                choose_eq_predicate_with_criteria (pred) (var_criteria) in
            (match po with
            | Some p -> new_pred_list := !new_pred_list @ [p]; ()
            | None -> ())
        )
        pred_list;
    (* return *)
    if !new_pred_list = []  then [True]
    else !new_pred_list
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest-post-eq-condition for an assume edge. Note that an assume
egde maybe contains a predicate of assume, or 2 predicates of assume & reset.  
 - Param: {ul {li {b edge} [predicate type]}
              {li {b pre_eq_condition} [predicate type]}
              {li {b var_criteria} [expression list type]}}
 - Pseudo Code: 
{ul {li [If the edge contains only pred_of_assume, then find all equality 
predicates of the pred_of_assume, then compute strongest-post-eq-condition 
traditionally.]}
    {li [If the edge contains pred_of_assume & pred_of_reset, first we find all 
equality predicates of the pred_of_assumes & pred_of_reset, then concate them
and compute strongest-post-eq-conditions traditionally]}} 
 - Return: strongest-post-eq-conditions (predicate list type) 
 - Since: 20090611
 *)
let get_strongest_post_eq_condition_for_art_assume (edge : predicate)
    (pre_eq_condition : predicate)
    (var_criteria : expression list)
: predicate list = 
begin
    (* Note that edge of ART_ASSUME has type: *)
    (* And [pred_of_assume; pred_of_reset]     *)
    (match edge with
    | And [pred_of_assume; pred_of_reset] -> begin
        (* First, find preds in assume which maybe assignments *)
        let ass_assignments : predicate list = find_eq_predicates (pred_of_assume) in
        let reset_assignments : predicate list = (match pred_of_reset with
                                                 | And pl -> pl
                                                 | p -> [p]) 
        in
        (* Create temp_edge *)
        let temp_edge : predicate = And (ass_assignments @ reset_assignments) in
        (* Compute strongest post conditions and return *)
        let spcs : predicate list = 
            get_strongest_post_eq_condition_for_assignments_edge (pre_eq_condition)
              (temp_edge)
        in
        choose_eq_predicates_with_criteria (spcs) (var_criteria)
      end
    | pred_of_assume -> begin
        (* First, find preds in assume which maybe assignments *)
        let ass_assignments : predicate list = find_eq_predicates (pred_of_assume) in
        (* Create temp_edge *)
        let temp_edge : predicate = And ass_assignments in
        (* Compute strongest post conditions *)
        let spcs : predicate list = 
             get_strongest_post_eq_condition_for_assignments_edge (pre_eq_condition)
              (temp_edge)
        in
        (* Return *)
        choose_eq_predicates_with_criteria (spcs) (var_criteria)
      end)
end                                                    

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Compute strongest-post-eq-conditions for an assume edge. Note that an assume
egde maybe contains a predicate of assume, or 2 predicates of assume & reset.  
 - Param: {ul {li {b edge} [predicate type]}
              {li {b pre_eq_conditions} [predicate list type]}
              {li {b var_criteria} [expression list type]}}
 - Pseudo Code: 
{ul {li [Foreach pre_eq_condition in pre_eq_conditions, compute strongest-
post-eq-condition for it with the edge & var_criteria, then save the result
to list to return]}}
 - Return: strongest-post-eq-conditions (predicate list type) 
 - Since: 20090611
 *)
let get_strongest_post_eq_conditions_for_art_assume (edge : predicate)
    (pre_eq_conditions : predicate list)
    (var_criteria : expression list)
: predicate list = 
begin
    (* Create var to save result *)
    let spcs : predicate list ref = ref [] in
    (* Foreach pre_eq_conditon, compute spcs, add result to spcs to return *)
    List.iter 
        (fun a_pre_eq_condition -> 
            (* Compute new spcs for a_pre_eq_condition *)
            let new_spcs : predicate list = 
                get_strongest_post_eq_condition_for_art_assume (edge) 
                  (a_pre_eq_condition) (var_criteria) in
            (* Add to spcs *)
            spcs := !spcs @ new_spcs 
        )
        pre_eq_conditions;
    (* Return *)
    !spcs
end   

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Replace sub-expressions in an expression by new sub-expressions, 
respectively.  
 - Param: {ul {li {b replacing_exprs} [expression list type]}
              {li {b exprs_to_replace} [expression list type]}
              {li {b current_expr} [expression type]}}
 - Pseudo Code: 
{ul {li [Foreach replacing_expr in replacing_exprs, replace it by corresponding
expr_to_replace of it in exprs_to_replace, then save the result to list to 
return]}}
 - Return: replaced_expr (expression type) 
 - Since: 20090611
 *)
let replace_exprs_in_expression (replacing_exprs : expression list) 
                                (exprs_to_replace : expression list)
                                (current_expr : expression)
: expression = 
begin
    let replaced_expr : expression ref = ref current_expr in
    (* Foreach tuples {replacing_expr, expr_to_replace}, then replace *)
    (try 
        (List.iter2
            (fun replacing_expr expr_to_replace -> 
                replaced_expr := 
                    replace_var_in_expression (!replaced_expr) (replacing_expr)
                      (expr_to_replace);
            )
            replacing_exprs
            exprs_to_replace;
        )
    with Invalid_argument _ -> assert false);
    (* Return *)
    !replaced_expr 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Replace sub-expressions in expressions by new sub-expressions, respectively.  
 - Param: {ul {li {b replacing_exprs} [expression list type]}
              {li {b exprs_to_replace} [expression list type]}
              {li {b current_exprs} [expression list type]}}
 - Pseudo Code: 
{ul {li [Foreach expression in current_exprs, call replace_exprs_in_expression]}}
 - Return: replaced exprs (expression list type) 
 - Since: 20090611
 *)
let replace_exprs_in_expressions (replacing_exprs : expression list) 
                                 (exprs_to_replace : expression list)
                                 (current_exprs : expression list)
: expression list = 
begin
    List.map (replace_exprs_in_expression replacing_exprs exprs_to_replace)
             current_exprs;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Replace sub-expressions in a predicate by new sub-expressions, respectively.  
 - Param: {ul {li {b replacing_exprs} [expression list type]}
              {li {b exprs_to_replace} [expression list type]}
              {li {b current_pred} [predicate type]}}
 - Pseudo Code: 
{ul {li [Foreach replacing_expr in replacing_exprs, replace it in current_pred
 by corresponding expr_to_replace of it in exprs_to_replace, then save the 
result to list to return]}}
 - Return: replaced_pred (predicate type) 
 - Since: 20090611
 *)
let replace_exprs_in_predicate (replacing_exprs : expression list) 
                               (exprs_to_replace : expression list)
                               (current_pred : predicate)
: predicate = 
begin
    let replaced_pred : predicate ref = ref current_pred in
    (* Foreach tuples {replacing_expr, expr_to_replace}, then replace *)
    let rec make_replacement ( replaced_pred : predicate ref)
                             (old_exprs: expression list)
                             (new_exprs: expression list)
                             : unit =
    begin
        match (old_exprs, new_exprs) with
        | ([], _ )                     -> ();
        | (e1 ::exprs1, [])            -> assert false;
        | (e1 ::exprs1, e2 :: exprs2 ) -> (
            replaced_pred := replace_var_in_predicate (!replaced_pred) e1 e2;
            make_replacement replaced_pred exprs1 exprs2;
          )
    end in
    make_replacement replaced_pred replacing_exprs exprs_to_replace;
    (* Return *)
    !replaced_pred    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Replace sub-expressions in predicates by new sub-expressions, respectively.  
 - Param: {ul {li {b replacing_exprs} [expression list type]}
              {li {b exprs_to_replace} [expression list type]}
              {li {b current_preds} [predicate list type]}}
 - Pseudo Code: 
{ul {li [Foreach predicate in current_preds, call replace_exprs_in_predicate]}}
 - Return: replaced_preds (predicate list type)
 - Since: 20090611
 *)
let replace_exprs_in_predicates (replacing_exprs : expression list)
                                (exprs_to_replace : expression list)
                                (current_preds : predicate list)
: predicate list = 
begin
    List.map (replace_exprs_in_predicate (replacing_exprs) (exprs_to_replace))
             current_preds;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Concate variables of an expression & in the demand with a tail.
 - Param: {ul {li {b expr} (expression type)}
              {li {b expr_vars} (expression list type)}
              {li {b tail} (string type)}}
 - Example: [{expression: x%3 + y + 1}; {variables needed concate: y}; 
{tail: %3}; ==> {new expression: x%3 + y%3 + 1}]
 - Pseudo Code: 
{ul {li [Find all variables in the expr.]}
    {li [Foreach variable in the found variables, check if it is mem of 
expr_vars, then concate tail for it and save it to return.]}}
 - Return: concated expression (expression type)
 - Since: 20090611
 *)
let rec concate_vars_with_tail_in_expression (expr : expression)
      (expr_vars : expression list)
      (tail : string)
: expression = 
begin
    match expr with
    | Constant f -> Constant f
    | Variable s -> begin
        let is_mem : bool = is_mem_of_expr_list (Variable s) (expr_vars) in
        if is_mem = true then Variable (s ^ tail)
        else Variable s
      end
    | Application (s, expr_list) -> begin
        let concated_expr_list : expression list ref = ref [] in
        List.iter
            (fun an_expr ->
                let concated_expr : expression = 
                    concate_vars_with_tail_in_expression (an_expr) (expr_vars) 
                      (tail) in
                concated_expr_list := !concated_expr_list @ [concated_expr]
            )
             expr_list;
         Application (s, !concated_expr_list)
      end
    | Sum expr_list -> begin
        let concated_expr_list : expression list ref = ref [] in
        List.iter
            (fun an_expr ->
                let concated_expr : expression = 
                    concate_vars_with_tail_in_expression (an_expr) (expr_vars) 
                      (tail) in
                concated_expr_list := !concated_expr_list @ [concated_expr]
            )
             expr_list;
         Sum !concated_expr_list
      end
    | Coeff (f, an_expr) -> begin
        let concated_expr : expression = 
            concate_vars_with_tail_in_expression (an_expr) (expr_vars) (tail) in
        Coeff (f, concated_expr)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Concate variables of a predicate & in the demand with a tail.
 - Param: {ul {li {b pred} (predicate type)}
              {li {b expr_vars} (expression list type)}
              {li {b tail} (string type)}}
 - Example: [{predicate: x%3 = y + 1}; {variables needed concate: y}; 
{tail: %3}; ==> {new predicate: x%3 = y%3 + 1}]
 - Pseudo Code: 
{ul {li [Find all variables in the pred.]}
    {li [Foreach variable in the found variables, check if it is mem of 
expr_vars, then concate tail for it and save it to return.]}}
 - Return: concated predicate (predicate type)
 - Since: 20090611
 *)
let rec concate_vars_with_tail_in_predicate (pred : predicate)
      (expr_vars : expression list)
      (tail : string)
: predicate = 
begin
    match pred with
    | True -> True
    | False -> False
    | And pred_list -> begin
        let concated_pred_list : predicate list ref = ref [] in
        List.iter
            (fun a_pred -> 
                let concated_pred : predicate = 
                    concate_vars_with_tail_in_predicate (a_pred) (expr_vars) 
                      (tail) 
                in
                concated_pred_list := !concated_pred_list @ [concated_pred]
            )
            pred_list;
        (* Return *)
        And !concated_pred_list
      end
    | Or pred_list -> begin
        let concated_pred_list : predicate list ref = ref [] in
        List.iter
            (fun a_pred -> 
                let concated_pred : predicate = 
                    concate_vars_with_tail_in_predicate (a_pred) (expr_vars) 
                      (tail) 
                in
                concated_pred_list := !concated_pred_list @ [concated_pred]
            )
            pred_list;
        (* Return *)
        Or !concated_pred_list
      end
    | Not a_pred -> begin
        let concated_pred : predicate = 
            concate_vars_with_tail_in_predicate (a_pred) (expr_vars) (tail) in
        Not concated_pred
      end
    | Eq (expr1, expr2) -> begin
        let concated_expr1 : expression = 
            concate_vars_with_tail_in_expression (expr1) (expr_vars) (tail) in
        let concated_expr2 : expression = 
            concate_vars_with_tail_in_expression (expr2) (expr_vars) (tail) in
        Eq (concated_expr1, concated_expr2)
      end
    | Lt (expr1, expr2) -> begin
        let concated_expr1 : expression = 
            concate_vars_with_tail_in_expression (expr1) (expr_vars) (tail) in
        let concated_expr2 : expression = 
            concate_vars_with_tail_in_expression (expr2) (expr_vars) (tail) in
        Lt (concated_expr1, concated_expr2)
      end
    | Leq (expr1, expr2) -> begin
        let concated_expr1 : expression = 
            concate_vars_with_tail_in_expression (expr1) (expr_vars) (tail) in
        let concated_expr2 : expression = 
            concate_vars_with_tail_in_expression (expr2) (expr_vars) (tail) in
        Leq (concated_expr1, concated_expr2)
      end
    | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Transform strings to variables.
 - Param: {ul {li {b string_vars} (string list type)}}
 - Example: [{string: x%3} ==> {variable: Variable x%3}]
 - Pseudo Code: {ul {li [Foreach var in the input, transform to "Variable var"
then save the result to list to return]}}
 - Return: expression list
 - Since: 20090611
 *)
let string_vars_to_exprs (string_vars : string list) : expression list = 
begin
    let expr_list : expression list ref = ref [] in
    (* Foreach string_var, transforms it to expression *)
    List.iter
        (fun string_var ->
            expr_list := !expr_list @ [Variable string_var];
        )
        string_vars;
    (* Return *)
    !expr_list
end 

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** {b Temporarily} nomalize an expression with theories:  
{Pref (Addrof (x)) = x},
{Srefp (Addrof(x), field_expr) = Sref(x, field_expr)}
 - Param: {ul {li {b expr} (expression type)}}
 - Pseudo Code: 
{ul {li [Replace any sub-expression has form *(&varname) with varname.]}}
 - Return: expression
 - Since: 20090611
 *)
let rec normalize_expr_with_theories (expr : expression) : expression = 
begin
    match expr with
    | Constant f -> Constant f
    | Variable s -> Variable s
    | Application ("pref", [Application ("addrof", [an_expr]); _]) -> begin
        let a_new_expr : expression = normalize_expr_with_theories (an_expr) in
        a_new_expr
      end
    | Application ("srefp", [Application ("addrof", [str_expr]); field_expr]) -> 
      begin
        let new_str_expr : expression = normalize_expr_with_theories str_expr in
        let new_field_expr = normalize_expr_with_theories field_expr in
        Application ("sref", [new_str_expr; new_field_expr])
      end
    | Application (s, exprs) -> begin
        let new_exprs : expression list ref = ref [] in
        (* Foreach expression in exprs, normalize it *)
        List.iter   
            (fun an_expr -> 
                let a_new_expr = normalize_expr_with_theories (an_expr) in
                new_exprs := !new_exprs @ [a_new_expr]
            )
            exprs;
        (* Return *)
        Application (s, !new_exprs)
      end 
    | Sum exprs -> begin
        let new_exprs : expression list ref = ref [] in
        (* Foreach expression in exprs, normalize it *)
        List.iter   
            (fun an_expr -> 
                let a_new_expr = normalize_expr_with_theories (an_expr) in
                new_exprs := !new_exprs @ [a_new_expr]
            )
            exprs;
        (* Return *)
        Sum !new_exprs
      end
    | Coeff (f, an_expr) -> begin
        let a_new_expr : expression = normalize_expr_with_theories (an_expr) in
        Coeff (f, a_new_expr)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** {b Temporarily} nomalize an predicate with theories:  
{Pref (Addrof (x)) = x},
{Srefp (Addrof(x), field_expr) = Sref(x, field_expr)}
 - Param: {ul {li {b pred} (predicate type)}}
 - Pseudo Code: 
{ul {li [Replace any sub-expression has form *(&varname) with varname.]}}
 - Return: predicate
 - Since: 20090611
 *)
let rec normalize_pred_with_theories (pred : predicate) : predicate = 
begin
    match pred with
    | True -> True
    | False -> False
    | And preds -> begin
        let new_preds : predicate list ref = ref [] in
        (* Foreach pred in preds, normalize it *)
        List.iter
            (fun a_pred ->
                let a_new_pred = normalize_pred_with_theories (a_pred) in
                new_preds := !new_preds @ [a_new_pred];
            )
            preds;
        (* Return *)
        And !new_preds
      end
    | Or preds -> begin
        let new_preds : predicate list ref = ref [] in
        (* Foreach pred in preds, normalize it *)
        List.iter
            (fun a_pred ->
                let a_new_pred = normalize_pred_with_theories (a_pred) in
                new_preds := !new_preds @ [a_new_pred];
            )
            preds;
        (* Return *)
        Or !new_preds
      end
    | Not a_pred -> begin
        let a_new_pred : predicate = normalize_pred_with_theories (a_pred) in
        Not a_new_pred
      end
    | Eq (expr1, expr2) -> begin
        let new_expr1 : expression = normalize_expr_with_theories (expr1) in
        let new_expr2 : expression = normalize_expr_with_theories (expr2) in
        Eq (new_expr1, new_expr2)
      end
    | Lt (expr1, expr2) -> begin
        let new_expr1 : expression = normalize_expr_with_theories (expr1) in
        let new_expr2 : expression = normalize_expr_with_theories (expr2) in
        Lt (new_expr1, new_expr2)
      end
    | Leq (expr1, expr2) -> begin
        let new_expr1 : expression = normalize_expr_with_theories (expr1) in
        let new_expr2 : expression = normalize_expr_with_theories (expr2) in
        Leq (new_expr1, new_expr2)
      end
    | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** {b Temporarily} nomalize a list of expressions with theories:  
{Pref (Addrof (x)) = x},
{Srefp (Addrof(x), field_expr) = Sref(x, field_expr)}
 - Param: {ul {li {b exprs} (expression list type)}}
 - Pseudo Code: 
{ul {li [Replace any sub-expression has form *(&varname) with varname.]}}
 - Return: expression list
 - Since: 20090611
 *)
let normalize_exprs_with_theories (exprs : expression list)
: expression list = 
begin
    let new_exprs : expression list ref = ref [] in
    (* Foreach expression in exprs, normalize it *)
    List.iter   
        (fun an_expr -> 
            let a_new_expr = normalize_expr_with_theories (an_expr) in
            new_exprs := !new_exprs @ [a_new_expr]
        )
        exprs;
    (* Return *)
    !new_exprs
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** {b Temporarily} nomalize a list of predicates with theories:  
{Pref (Addrof (x)) = x},
{Srefp (Addrof(x), field_expr) = Sref(x, field_expr)}
 - Param: {ul {li {b preds} (predicate list type)}}
 - Pseudo Code: 
{ul {li [Replace any sub-expression has form *(&varname) with varname.]}}
 - Return: normalized preds with the theory (predicate list type)
 - Since: 20090611
 *)
let normalize_preds_with_theories (preds : predicate list) 
: predicate list = 
begin
    let new_preds : predicate list ref = ref [] in
    (* Foreach pred in preds, normalize it *)
    List.iter
        (fun a_pred ->
            let a_new_pred : predicate = normalize_pred_with_theories (a_pred) in
            new_preds := !new_preds @ [a_new_pred];
        )
        preds;
    (* Return *)
    !new_preds
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find existing of arrays in an expression       *)
let rec expr_contains_arrays (expr : expression) : bool = 
begin
    match expr with
    | Constant f -> false
    | Variable v -> false
    | Application ("array", _) -> true
    | Application (s, exprs) -> begin
        let flag : bool ref = ref false in
        List.iter
            (fun e -> if !flag = false then flag := expr_contains_arrays e;)
            exprs;
        !flag
      end
    | Sum exprs -> begin
        let flag : bool ref = ref false in
        List.iter
            (fun e -> if !flag = false then flag := expr_contains_arrays e;)
            exprs;
        !flag
      end
    | Coeff (f, e) -> expr_contains_arrays (e)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find existing of arrays in a predicate         *)
let rec pred_contains_arrays (pred : predicate) : bool = 
begin
    match pred with
    | True -> false
    | False -> false
    | And preds -> begin
        let flag : bool ref = ref false in
        List.iter 
            (fun p -> if !flag = false then flag := pred_contains_arrays p;)
            preds;
        !flag;
      end
    | Or preds -> begin
        let flag : bool ref = ref false in
        List.iter 
            (fun p -> if !flag = false then flag := pred_contains_arrays p;)
            preds;
        !flag;
      end
    | Not p -> pred_contains_arrays (p)
    | Eq (expr1, expr2) -> begin
        let flag1 : bool = expr_contains_arrays (expr1) in
        let flag2 : bool = expr_contains_arrays (expr2) in
        (flag1 || flag2)
      end
    | Lt (expr1, expr2) -> begin
        let flag1 : bool = expr_contains_arrays (expr1) in
        let flag2 : bool = expr_contains_arrays (expr2) in
        (flag1 || flag2)
      end
    | Leq (expr1, expr2) -> begin
        let flag1 : bool = expr_contains_arrays (expr1) in
        let flag2 : bool = expr_contains_arrays (expr2) in
        (flag1 || flag2)
      end
    | Atom i -> false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find arrays in a expression                    *)
let rec find_arrays_in_expr (expr : expression) : expression list = 
begin
    match expr with
    | Constant f -> []
    | Variable v -> []
    | Application ("array", memory_expr :: _) -> [memory_expr]
    | Application (s, exprs) -> begin
        let memories : expression list ref = ref [] in
        List.iter
            (fun e ->
                let new_memories : expression list = find_arrays_in_expr e in
                memories := new_memories @ !memories;
            )
            exprs;
        !memories
      end
    | Sum exprs -> begin
        let memories : expression list ref = ref [] in
        List.iter
            (fun e ->
                let new_memories : expression list = find_arrays_in_expr e in
                memories := new_memories @ !memories;
            )
            exprs;
        !memories
      end
    | Coeff (f, e) -> find_arrays_in_expr (e)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find arrays in a predicate                     *)
let rec find_arrays_in_pred (pred : predicate) : expression list = 
begin
    match pred with
    | True -> []
    | False -> []
    | And preds -> begin
        let memories : expression list ref = ref [] in
        List.iter 
            (fun p -> 
                let new_memories : expression list = find_arrays_in_pred p in
                memories := !memories @ new_memories;
            )
            preds;
        !memories
      end
    | Or preds -> begin
        let memories : expression list ref = ref [] in
        List.iter 
            (fun p -> 
                let new_memories : expression list = find_arrays_in_pred p in
                memories := !memories @ new_memories;
            )
            preds;
        !memories
      end
    | Not p -> find_arrays_in_pred (p)
    | Eq (expr1, expr2) -> begin
        let memories1 : expression list = find_arrays_in_expr (expr1) in
        let memories2 : expression list = find_arrays_in_expr (expr2) in
        memories1 @ memories2
      end
    | Lt (expr1, expr2) -> begin
        let memories1 : expression list = find_arrays_in_expr (expr1) in
        let memories2 : expression list = find_arrays_in_expr (expr2) in
        memories1 @ memories2
      end
    | Leq (expr1, expr2) -> begin
        let memories1 : expression list = find_arrays_in_expr (expr1) in
        let memories2 : expression list = find_arrays_in_expr (expr2) in
        memories1 @ memories2
      end
    | Atom i -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find indexes of arrays in an expression        *)
let rec find_indexes_of_arrays_in_expr (expr : expression) : expression list =
begin
    match expr with
    | Constant f -> []
    | Variable v -> []
    | Application ("array", [memory_expr; index_expr]) -> [index_expr]
    | Application (s, exprs) -> begin
        let indexes : expression list ref = ref [] in
        List.iter
            (fun e ->
                let new_indexes : expression list = 
                    find_indexes_of_arrays_in_expr e in
                indexes := new_indexes @ !indexes;
            )
            exprs;
        !indexes
      end
    | Sum exprs -> begin
        let indexes : expression list ref = ref [] in
        List.iter
            (fun e ->
                let new_indexes : expression list = 
                    find_indexes_of_arrays_in_expr e in
                indexes := new_indexes @ !indexes;
            )
            exprs;
        !indexes
      end
    | Coeff (f, e) -> find_indexes_of_arrays_in_expr (e)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090619. New function to find indexes of arrays in a predicate          *)
let rec find_indexes_of_arrays_in_pred (pred : predicate) : expression list =
begin
    match pred with
    | True -> []
    | False -> []
    | And preds -> begin
        let indexes : expression list ref = ref [] in
        List.iter 
            (fun p -> 
                let new_indexes : expression list = 
                    find_indexes_of_arrays_in_pred p in
                indexes := !indexes @ new_indexes;
            )
            preds;
        !indexes
      end
    | Or preds -> begin
        let indexes : expression list ref = ref [] in
        List.iter 
            (fun p -> 
                let new_indexes : expression list = 
                    find_indexes_of_arrays_in_pred p in
                indexes := !indexes @ new_indexes;
            )
            preds;
        !indexes
      end
    | Not p -> find_indexes_of_arrays_in_pred (p)
    | Eq (expr1, expr2) -> begin
        let indexes1 : expression list = find_indexes_of_arrays_in_expr (expr1) in
        let indexes2 : expression list = find_indexes_of_arrays_in_expr (expr2) in
        indexes1 @ indexes2
      end
    | Lt (expr1, expr2) -> begin
        let indexes1 : expression list = find_indexes_of_arrays_in_expr (expr1) in
        let indexes2 : expression list = find_indexes_of_arrays_in_expr (expr2) in
        indexes1 @ indexes2
      end
    | Leq (expr1, expr2) -> begin
        let indexes1 : expression list = find_indexes_of_arrays_in_expr (expr1) in
        let indexes2 : expression list = find_indexes_of_arrays_in_expr (expr2) in
        indexes1 @ indexes2
      end
    | Atom i -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec get_format_string_of_predicate (tab : string) 
                                       (pred : predicate) 
: string = 
begin
    match pred with
    | True -> tab ^ "True"
    | False -> tab ^ "False"
    | And preds -> begin
        let length : int = List.length preds in
        if length <= 1 then tab ^ (string_of_predicate (And preds))
        else begin
            let format_string : string ref = ref (tab ^ "and [\n") in
            let new_tab : string = tab ^ "\t" in
            List.iter
                (fun p ->
                    let new_fs = get_format_string_of_predicate (new_tab) (p) in
                    format_string := !format_string ^ "\n" ^ new_fs; 
                )
                preds;
            format_string := !format_string ^ "\n" ^ tab ^ "]";
            !format_string
        end        
      end
    | Or preds -> begin
        let length : int = List.length preds in
        if length <= 1 then tab ^ (string_of_predicate (Or preds))
        else begin
            let format_string : string ref = ref (tab ^ "or [\n") in
            let new_tab : string = tab ^ "\t" in
            List.iter
                (fun p ->
                    let new_fs = get_format_string_of_predicate (new_tab) (p) in
                    format_string := !format_string ^ "\n" ^ new_fs; 
                )
                preds;
            format_string := !format_string ^ "\n" ^ tab ^ "]";
            !format_string
        end         
      end
    | Not (And preds) -> begin
        let length : int = List.length preds in
        if length <= 1 then tab ^ (string_of_predicate (Not (And preds)))
        else begin
            let format_string : string ref = ref (tab ^ "not and [\n") in
            let new_tab : string = tab ^ "\t" in
            List.iter
                (fun p ->
                    let new_fs = get_format_string_of_predicate (new_tab) (p) in
                    format_string := !format_string ^ "\n" ^ new_fs; 
                )
                preds;
            format_string := !format_string ^ "\n" ^ tab ^ "]";
            !format_string
        end
      end
    | Not p -> tab ^ (string_of_predicate (Not p)) 
    | Eq (expr1, expr2) | Lt (expr1, expr2) | Leq (expr1, expr2) -> 
        tab ^ (string_of_predicate pred)
    | Atom i -> tab ^ (string_of_predicate (Atom i))
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* last updated: 20090621                                                   *)
let rec flatten_implications (pred : predicate) : predicate list = 
begin
    match pred with
    | True -> [True]
    | False -> [False]
    | And preds -> begin
        let flatten_preds : predicate list ref = ref [] in
        List.iter
            (fun p -> let new_flatten_ps = flatten_implications p in
                      flatten_preds := !flatten_preds @ new_flatten_ps;)
            preds;
        !flatten_preds
      end
    | Or [Not p1; p2] -> flatten_implications p2
    | Or _ | Not _ | Eq _ | Lt _ | Leq _ -> [pred] 
    | Atom i -> [Atom i]
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_addrofs_in_expression (expr : expression) : expression list = 
begin
    (match expr with
    | Constant f -> []
    | Variable v -> []
    | Application ("addrof", _) -> [expr]
    | Application (_, exprs)
    | Sum exprs -> List.concat (List.map find_addrofs_in_expression exprs)
    | Coeff (f, e) -> (find_addrofs_in_expression e))
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_addrofs_in_predicate (pred : predicate) : expression list = 
begin
    match pred with
    | True | False -> []
    | And preds 
    | Or preds -> List.concat (List.map find_addrofs_in_predicate preds) 
    | Not p -> find_addrofs_in_predicate p
    | Eq (expr1, expr2) | Lt (expr1, expr2) | Leq (expr1, expr2) -> begin
        let addrofs_1 : expression list = find_addrofs_in_expression expr1 in
        let addrofs_2 : expression list = find_addrofs_in_expression expr2 in
        addrofs_1 @ addrofs_2
      end         
    | Atom i -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_constraint_for_addrof (addrof : expression) : predicate = 
begin
    (* Check addrof *)
    (match addrof with
    | Application ("addrof", _) -> ()
    | _ -> assert false);
    (* Create constraint & return *)
    Not (Eq (addrof, Constant 0.0))    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let remove_mem (mem : 'a) (list : 'a list) : 'a list =
begin
    let new_list = ref [] in
    List.iter
        (fun an_element -> 
            if an_element != mem then new_list := !new_list @ [an_element]
        )
        list;
    !new_list
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_left_of_addrof_assignment (addrof_asgn : predicate) 
: expression list =
begin
    match addrof_asgn with
    | True | False -> [] 
    | And preds | Or preds -> begin
        let lefts : expression list ref = ref [] in
        List.iter
            (fun p -> let new_lefts = find_left_of_addrof_assignment (p) in
                      lefts := !lefts @ new_lefts;
            )
            preds;
        !lefts
      end
    | Not p -> find_left_of_addrof_assignment p
    | Eq (left_expr, Application ("addrof", _)) -> [left_expr]
    | Eq _ | Lt _ | Leq _ | Atom _ -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_value_of_var_in_pred (pred : predicate)
                                  (var : expression)
: string option = 
begin
    (* debug *)
    (*debug ("find_value_of_var_in_pred: ");
    debug ("pred: " ^ (string_of_predicate pred));
    debug ("var: " ^ (string_of_expression var));*)

    (* Ensure var is Variable *)
    (*(match var with
    | Variable _ -> ()
    | Application ("sref", _) -> ()
    | _ -> assert false);*)
    (* Find value of var in pred *)
    (match pred with
    | True | False -> None
    | And pred_list -> begin
        let result : string option ref = ref None in
        List.iter
            (fun p ->
                if !result = None then begin
                    result := find_value_of_var_in_pred (p) (var);
                end
            )
            pred_list;
        (* Return *)
        !result
      end
    | Or pred_list -> None (* Temporarily return None *) 
    | Not p -> None (* Temporarily return None *)
    | Eq (expr1, expr2) -> begin
        (* Check if expr1 is the var, then expr2 is value of it *)
        (if (compare_expr_with_expr expr1 var) = 0 then begin
            (match expr2 with
            | Variable v -> Some v
            | _ -> None)
        (* Check if expr2 is the var, then expr1 is value of it *)
        end else if (compare_expr_with_expr expr2 var) = 0 then begin
            (match expr1 with
            | Variable v -> Some v
            | _ -> None)
        end else None)
      end
    | Lt _ | Leq _ | Atom _ -> None)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec expr_is_in_alias_expr (expr1 : expression)
                              (expr2 : expression)
: bool =
begin
    (* debug *)
    (*debug "expr_is_in_alias_expr";
    debug (string_of_expression expr1);
    debug (string_of_expression expr2);*)

    match expr2 with 
    | Constant f -> false
    | Variable v -> false
    | Application ("pref", [pointer_expr; temp_expr; index_expr]) -> begin
        (* Check pointer_expr if it is identical with expr1 or not *)
        let is_identical : bool = is_identical_expressions expr1 pointer_expr in 
        (*debug ("is_identical: " ^ (string_of_bool is_identical));*)
        if is_identical = true then true
        else expr_is_in_alias_expr (expr1) (pointer_expr)
      end
    | Application ("pref", _) -> assert false
    | Application ("srefp", [pstr_expr; field_expr; index_expr]) -> begin
        (* Check if pstr_expr is identical with expr1 or not *)
        if (is_identical_expressions expr1 pstr_expr) = true then true
        else expr_is_in_alias_expr (expr1) (pstr_expr)
      end
    | Application ("srefp", _) -> assert false
    | Application (_, exprs) -> begin
        let is_in_alias : bool ref = ref false in
        List.iter
            (fun an_expr -> 
                if !is_in_alias = false then begin
                    is_in_alias := expr_is_in_alias_expr expr1 an_expr;
                end
            )
            exprs;
        !is_in_alias
      end
    | Sum exprs -> begin
        let is_in_alias : bool ref = ref false in
        List.iter
            (fun an_expr -> 
                if !is_in_alias = false then begin
                    is_in_alias := expr_is_in_alias_expr expr1 an_expr;
                end
            )
            exprs;
        !is_in_alias
      end
    | Coeff (f, an_expr) -> begin
        expr_is_in_alias_expr expr1 an_expr
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec expr_is_in_alias_pred (expr : expression)
                              (pred : predicate)
: bool = 
begin
    match pred with
    | True -> false
    | False -> false
    | And preds -> begin
        let is_in_alias : bool ref = ref false in
        List.iter
            (fun a_pred -> 
                if !is_in_alias = false then begin
                    is_in_alias := expr_is_in_alias_pred expr a_pred
                end;
            )
            preds;
        (* Return *)
        !is_in_alias
      end
    | Or preds -> begin
        let is_in_alias : bool ref = ref false in
        List.iter
            (fun a_pred -> 
                if !is_in_alias = false then begin
                    is_in_alias := expr_is_in_alias_pred expr a_pred
                end;
            )
            preds;
        (* Return *)
        !is_in_alias
      end
    | Not p -> begin
        expr_is_in_alias_pred expr p
      end
    | Eq (expr1, expr2) | Lt (expr1, expr2) | Leq (expr1, expr2) -> begin
        let is_in_1 : bool = expr_is_in_alias_expr expr expr1 
        and is_in_2 : bool = expr_is_in_alias_expr expr expr2 in
        is_in_1 || is_in_2
      end
    | Atom _ -> false
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec insert_pred_to_where_expr_appears_in_pred (appearing_expr : expression)
                                                  (inserted_pred : predicate)
                                                  (pred : predicate)
: predicate =
begin
    (* debug *)
    (*debug "insert_pred_to_where_expr_appears_in_pred:";
    debug ("appearing_expr: " ^ (string_of_expression appearing_expr));
    debug ("inserted_pred: " ^ (string_of_predicate inserted_pred));
    debug ("pred:" ^ (string_of_predicate pred));*)

    match pred with
    | True -> True
    | False -> False
    | And preds -> begin
        let temp_fun : predicate -> predicate = 
            insert_pred_to_where_expr_appears_in_pred (appearing_expr)
                                                      (inserted_pred)
        in
        let new_preds : predicate list = List.map (temp_fun) (preds) in
        (* Return *)
        And new_preds 
      end
    (* in case that expression has constraints *)
    | Or [Not condition; pred_of_exprs] -> begin
        let new_pred = insert_pred_to_where_expr_appears_in_pred (appearing_expr)
                                          (inserted_pred) (pred_of_exprs) in
        let is_in_alias : bool = expr_is_in_alias_pred appearing_expr condition in
        if is_in_alias = true then Or [And [Not condition; inserted_pred]; new_pred]
        else Or [False; new_pred] (*Not p*)
      end
    | Or preds -> begin
        let temp_fun : predicate -> predicate = 
            insert_pred_to_where_expr_appears_in_pred (appearing_expr)
                                                      (inserted_pred)
        in
        let new_preds : predicate list = List.map (temp_fun) (preds) in
        (* Return *)
        Or new_preds
      end
    | Not p -> begin
        (* Not p = atom-predicate *)
        (* Check if appearing_expr is in p *)
        let is_in_alias : bool = expr_is_in_alias_pred appearing_expr p in
        (*debug ("is_in_alias: " ^ (string_of_bool is_in_alias));*) 
        if is_in_alias = true then And [Not p; inserted_pred]
        else True (*Not p*)
      end
    (* atom-predicates *)
    | Eq (expr1, expr2) | Lt (expr1, expr2) | Leq (expr1, expr2) -> begin
        (* Check if appearing_expr is in pred *)
        let is_in_alias : bool = expr_is_in_alias_pred appearing_expr pred in
        (*debug ("is_in_alias: " ^ (string_of_bool is_in_alias));*) 
        if is_in_alias = true then And [pred; inserted_pred]
        else True (*pred*)
      end
    | Atom _ -> pred
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec multiply_an_element_by_n (an_element : 'a)
                                 (n : int)
: 'a list = 
begin
    (* debug *)
    (*debug "multiply_an_element_by_n";
    debug ("n: " ^ (string_of_int n));*)

    if n > 1 then an_element :: (multiply_an_element_by_n (an_element) (n - 1))
    else if n = 1 then [an_element]
    else assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec multiply_string_by_n (s : string)
                             (n : int) 
: string = 
begin
    if n = 0 then ""
    else if n = 1 then s
    else if n >= 2 then s ^ (multiply_string_by_n s (n-1))
    else assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let hashtbl_flatten_to_list (hash : ('a, 'b) Hashtbl.t) : ('a * 'b) list = 
    Hashtbl.fold (fun key value ls -> (key, value) :: ls) hash []

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let hashtbl_import_list (hash : ('a, 'b) Hashtbl.t) 
                        (adding_list : ('a * 'b) list)
                        : unit =
    List.iter (fun (key, value) -> Hashtbl.add hash key value) adding_list     

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Remove all bindings of a key in a hashtable *)
let remove_elements_in_hashtbl (hash : ('k, 'v) Hashtbl.t)
                               (key : 'k)
                               : unit = (
    let all_bindings = Hashtbl.find_all hash key in
    List.iter (fun a_binding -> Hashtbl.remove hash key) all_bindings)

let subtract_exprs_from_exprs (exprs1 : expression list)
                              (exprs2 : expression list)
                              : expression list = (
    List.filter (fun e -> not (is_mem_of_expr_list e exprs2)) exprs1)

let member_of_list (e : 'a)
                   (a_list : 'a list)
                   (compare : 'a -> 'a -> int)
                   : bool = (
    List.fold_left (fun mem a ->
                        if mem then true
                        else if ((compare a e) = 0) then true
                             else false)
                   false a_list)

let subtract_list_from_list (list1 : 'a list)
                            (list2 : 'a list)
                            (member_of_list : 'a -> 'a list -> bool)
                            : 'a list = (
    List.filter (fun an_element -> not (member_of_list an_element list2))
                list1)

let merge_two_lists (list1 : 'a list)
                    (list2 : 'a list)
                    (member_of_list : 'a -> 'a list -> bool)
                    : 'a list = (
    let subtraction12 = subtract_list_from_list list1 list2 member_of_list in
    subtraction12 @ list2)

let find_memory
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (expr : expression)
: expression option = (
    try
        let expr_index = Hashtbl.find var_to_index expr in
        let (_, memory_index) = Hashtbl.find int_hash expr_index in
        let memory_expr = Hashtbl.find index_to_var memory_index in
        Some memory_expr
    with Not_found -> None
)

let rec replace_pointer_by_memory_in_expr
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (expr : expression)
: (predicate list) * expression = (
    match expr with
    | Constant _ -> ([], expr)
    | Variable _ -> (
        let memoryop_of_expr =
            find_memory int_hash var_to_index index_to_var expr in
        match memoryop_of_expr with
        | None -> ([], expr)
        | Some memory_of_expr -> (
            ([Eq (expr, Application ("addrof", [memory_of_expr]))],
             Application ("addrof", [memory_of_expr]))
          )
      )
    | Application (sym, exprs) -> (
        let new_preds, new_exprs =
            List.fold_left (fun (new_preds, new_exprs) a_expr ->
                let new_preds_, new_expr_ =
                    replace_pointer_by_memory_in_expr int_hash
                                                      var_to_index
                                                      index_to_var
                                                      a_expr
                in
                (new_preds @ new_preds_, new_exprs @ [new_expr_])
            ) ([], []) exprs
        in
        (new_preds, Application (sym, new_exprs))
      )
    | Sum exprs -> (
        let new_preds, new_exprs =
            List.fold_left (fun (new_preds, new_exprs) a_expr ->
                let new_preds_, new_expr_ =
                    replace_pointer_by_memory_in_expr int_hash
                                                      var_to_index
                                                      index_to_var
                                                      a_expr
                in
                (new_preds @ new_preds_, new_exprs @ [new_expr_])
            ) ([], []) exprs
        in
        (new_preds, Sum new_exprs)
      )
    | Coeff (f, a_expr) -> (
        let new_preds, a_new_expr =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              a_expr
        in
        (new_preds, Coeff (f, a_new_expr))
      )
)

let rec replace_pointer_by_memory_in_pred
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pred : predicate)
: predicate list * predicate = (
    match pred with
    | True
    | False
    | Atom _ -> ([], pred)
    | And preds -> (
        let new_preds1, new_preds2 =
            List.fold_left (fun (new_preds1, new_preds2) a_pred ->
                let new_preds1_, new_pred2_ =
                    replace_pointer_by_memory_in_pred int_hash
                                                      var_to_index
                                                      index_to_var
                                                      a_pred
                in
                (new_preds1 @ new_preds1_, new_preds2 @ [new_pred2_])
            ) ([], []) preds
        in
        (new_preds1, And (new_preds2)))
    | Or preds -> (
        let new_preds1, new_preds2 =
            List.fold_left (fun (new_preds1, new_preds2) a_pred ->
                let new_preds1_, new_pred2_ =
                    replace_pointer_by_memory_in_pred int_hash
                                                      var_to_index
                                                      index_to_var
                                                      a_pred
                in
                (new_preds1 @ new_preds1_, new_preds2 @ [new_pred2_])
            ) ([], []) preds
        in
        (new_preds1, Or (new_preds2)))
    | Not a_pred -> (
        let new_preds, a_new_pred =
            replace_pointer_by_memory_in_pred int_hash
                                               var_to_index
                                               index_to_var
                                               a_pred
        in
        (new_preds, Not a_new_pred)
      )
    | Eq (expr1, expr2) -> (
        let new_preds1, new_expr1 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr1
        in
        let new_preds2, new_expr2 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr2
        in
        (new_preds1 @ new_preds2, Eq (new_expr1, new_expr2))
      )
    | Lt (expr1, expr2) -> (
        let new_preds1, new_expr1 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr1
        in
        let new_preds2, new_expr2 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr2
        in
        (new_preds1 @ new_preds2, Lt (new_expr1, new_expr2))
      )
    | Leq (expr1, expr2) -> (
        let new_preds1, new_expr1 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr1
        in
        let new_preds2, new_expr2 =
            replace_pointer_by_memory_in_expr int_hash
                                              var_to_index
                                              index_to_var
                                              expr2
        in
        (new_preds1 @ new_preds2, Leq (new_expr1, new_expr2))
      )
)

let replace_pointer_by_memory_in_pred
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pred : predicate)
: predicate = (
    let new_preds, new_pred =
        replace_pointer_by_memory_in_pred int_hash
                                          var_to_index
                                          index_to_var
                                          pred
    in
    And (new_preds @ [new_pred])
)