(* la/lazyabstraction.ml                                         *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(** Define functions to analysis CFAs with LazyAbstraction paradigm        *)


(* ------------- LAZY ABSTRACTION - ABSTRACTION FROM PROOFS -------------- 
Reference:                                                             
  [1] Thomas A.Henzinger, Ranjit Jhala, Rupak Majumdar, Kenneth L.McMillan: 
      Abstraction from proofs. 
  [2] Dirk Beyer, Thomas A.Henzinger, Ranjit Jhala, Rupak Majumdar: The 
      software model checker BLAST. 
  [3] Thomas A.Henzinger, Ranjit Jhala, Rupak Majumdar: Lazy Abstraction. 
  [4] Georg WeiBenbacher: An Abstract/Refinement Scheme for Model Checking 
      C Programs.                                                         *)

open Checker;;
open LogicUtil.Logic;;
open Art;;
open Cfa;;
open La_common;;
open Slicing_path;;
open La_interpolate;;
(*module Log = ConfigUtil.Log;;*)

let debug = La_common.debug
let info = La_common.info
let warning = La_common.warning
let error = La_common.error
let emerg = La_common.emerg

type function_entry_t = Sexpr.function_entry_t
type ps_options_t = Checker.ps_options_t
type check_context_t = Checker.check_context_t

let la_options : ps_options_t ref
    = ref {
        Checker.psopt_entries = (Vector.create 4 Sexpr.nil_function_entry); 
        Checker.psopt_iter_max = 5;			(* default value *)
        Checker.psopt_recursive_max = 5;		(* default value *)
        Checker.psopt_timeout = 0.;  	(* default value *)
        Checker.psopt_slice_path = 2;			(* default value *)
        Checker.psopt_cutlist = [];			(* default value *)
        Checker.psopt_la_debug = false;			(* default value *)
        Checker.psopt_dot_output = false;		(* default value *)
        Checker.psopt_xml_location = "errorpath.xml";	(* default value *)
        Checker.psopt_intensive_pointer_checking = true;       (* default value *)
        Checker.psopt_debug_mode = false;               (* default value *)
        Checker.psopt_bmc = true;                      (* default value *)
        Checker.psopt_verbose_components = Vector.create 2 "";
        Checker.psopt_check_in_callees = false;
        Checker.psopt_check_array_index = true;
        Checker.psopt_strategy = Checker.HRBC;
        Checker.psopt_no_assert = false;
        Checker.psopt_no_subtyping_check = false;
        Checker.psopt_temp_dir = Filename.temp_dir_name; (* default value *)
        Checker.psopt_round_coeffs = false;
      };;

type la_error_context_t = {
    ctx_error_path : art_path;
    ctx_error_location : (string * string * string * string * string);
    ctx_error_message : string
  }

let la_error : la_error_context_t ref
    = ref {
        ctx_error_path = create_new_art_path ();
        ctx_error_location = ("", "", "", "", "");
        ctx_error_message = ""
      };;

let check_duplicated_error : bool ref = ref true;;

(* ~~~~~~~~~~~~~ NEW: normalize_overapproximation_list ~~~~~~~~~~~~~~~~~~~ 
Example: {x = 1 /\ x > 0} => normalize = {x = 1}                           *)
let rec normalize_overapproximation_list (app_list : predicate list) 
                                         (remaining_list : predicate list) 
: predicate list = 
begin
    match app_list with
    | [] -> remaining_list
    | a_app :: rapp_list -> begin
        let new_remaining_list = ref remaining_list in
        let implication = Or [Not (And (remaining_list @ rapp_list)); a_app] in
        if (is_valid implication) = false then begin
            new_remaining_list := a_app :: !new_remaining_list
        end; 
        normalize_overapproximation_list (rapp_list) (!new_remaining_list)  
      end      
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~ REGION: COMPUTE OVERAPPROXIMATION' ~~~~~~~~~~~~~~ 
[3, pages 6] For a region (phi, L) and a label l (- E, we construct the over
approximation post^((phi, L), l) of post((phi, L), l) as a boolean      
combination of the predicates in L. We fix a canonical sum-of-product form 
of formulas. We ask, for each disjunct teta of canonical sum-of-product 
form of phi, and each support predicate p (- L, if teta implies pre(p,l), 
and if teta implies pre(!p, l). This gives, for each disjunct teta of phi, 
a conjunction teta' of support predicates, where the predicate p occurs 
positively if teta => pre(p, l) is valid, occurs negatively if 
teta => pre(!p, l) is valid, and does not occur if neither validity holds. 
Let phi' denote the disjunction of all conjunctions teta' so constructed. 
The region post^((phi, L), l) is the formula phi' together with the same 
set of support predicates, i.e., post^((phi, L), l) = (phi', L).The abstract 
predecessor pre^ is computed similarly, using post in place of pre.        *)
let get_overapproximation_for_assignment (sp_preds : predicate list)
                                         (rr_of_father : predicate) 
                                         (edge : predicate) 
: predicate =
begin
    (* region: debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug "_________________________________________";
        debug "get_overapproximation_for_assignment: ";
        debug ("support_predicates: " ^ string_of_predicate (And sp_preds));
        debug ("rr_of_father: " ^ string_of_predicate rr_of_father);
        debug ("edge " ^ string_of_predicate edge);
    end;
    (* create hashtables to change index of variables *)
    let index_hash = Hashtbl.create (100) in
    let int_hash = Hashtbl.create (100) in
    let var_to_index = Hashtbl.create (100) in
    let index_to_var = Hashtbl.create (100) in
    let structs_hash = Hashtbl.create (100) in
    let var_eq_hash = Hashtbl.create (100) in
    let current_vars_in_fun = ref [] in
    let elements_of_arrays : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let uniqueness_hash : (int, expression) Hashtbl.t = Hashtbl.create 0 in
    (* Step1: (Forward) change index *)
    let new_rr_of_father : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays) (true) (false)
          (ref []) (rr_of_father) (uniqueness_hash)
    in
    let new_edge : predicate = 
        change_index_of_variables_in_non_condition_predicate (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (structs_hash) 
          (current_vars_in_fun) (true) (false) (elements_of_arrays) (false) 
          (ref []) (var_eq_hash) (edge) (uniqueness_hash)
    in
    (* debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug ("new_rr_of_father " ^ (string_of_predicate new_rr_of_father));
        debug ("new_edge " ^ (string_of_predicate new_edge));
    end;
    if (is_sat (And [new_rr_of_father; new_edge])) = false then False
    else (
        (* Step2: for each Support-Predicate, compute implications: *)
        let over_app_list = ref [] in
        List.iter
            (fun a_sp_pred -> 
                let a_new_sp_pred : predicate = 
                    change_index_of_variables_in_condition  
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (structs_hash) (current_vars_in_fun) 
                      (elements_of_arrays) (false) (false) (ref []) (a_sp_pred)
                      (uniqueness_hash)
                in
                let remaining_vars : expression list = 
                    find_all_vars_in_predicate (a_new_sp_pred) in
                let new_rr_of_father =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                new_rr_of_father
                in
                let new_edge =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                new_edge
                in

                (* Step3: Check valid *)
                let implied_left : predicate = 
                    normalize (
                        remove_lit_clash ( 
                            And (new_rr_of_father :: [(*sliced_edge*)new_edge]))) 
                in
                let implication1 : predicate = 
                    Or ((Not implied_left) :: [a_new_sp_pred]) in
                let implication2 : predicate = 
                    Or ((Not implied_left) :: [Not a_new_sp_pred]) in
                    
                (* debug *)
                if !la_options.Checker.psopt_la_debug = true then begin
                    debug ("implication1: " ^ string_of_predicate implication1);
                    debug ("implication2: " ^ string_of_predicate implication2);
                end;
                if (is_valid implication1) = true then begin
                    (* debug *)
                    over_app_list := !over_app_list @ [a_sp_pred]
                end else if (is_valid implication2) = true then begin
                    (* debug *)
                    over_app_list := !over_app_list @ [Not a_sp_pred]
                end else (*debug "no implication";*) ()
            )
            sp_preds;
        (* Step4: return *)
        if !over_app_list = [] then True
        else begin

            let normalized_over_app_list = 
                normalize_overapproximation_list (!over_app_list) ([]) in
            (normalize (And normalized_over_app_list))
            (*normalize (And !over_app_list)*)
        end
    )
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_over_approximation_for_reset (sp_preds : predicate list)
                                     (rr_of_father : predicate) 
                                     (reset_edge : predicate)
: predicate =
begin
    (* region: debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug "_________________________________________";
        debug "get_over_approximation_for_reset: ";
        debug ("support_predicates: " ^ string_of_predicate (And sp_preds));
        debug ("rr_of_father: " ^ string_of_predicate rr_of_father); 
        debug ("edge " ^ string_of_predicate reset_edge);
    end;
    (* create hashtables to change index of variables *)
    let index_hash = Hashtbl.create (100) in
    let int_hash = Hashtbl.create (100) in
    let var_to_index = Hashtbl.create (100) in
    let index_to_var = Hashtbl.create (100) in
    let structs_hash = Hashtbl.create (100) in
    let current_vars_in_fun = ref [] in
    let elements_of_arrays : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let uniqueness_hash : (int, expression) Hashtbl.t = Hashtbl.create 0 in
    (* Step1: (Forward) change index *)
    let new_rr_of_father : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays) 
          (true) (false) (ref []) (rr_of_father) (uniqueness_hash)
    in
    let new_edge : predicate = 
        change_index_of_variables_in_reset (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (structs_hash) 
          (current_vars_in_fun) (elements_of_arrays) (true) (false) 
          (ref []) (reset_edge) uniqueness_hash
    in
    (* debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug ("new_rr_of_father " ^ (string_of_predicate new_rr_of_father));
        debug ("new_edge " ^ (string_of_predicate new_edge));
    end;
    if (is_sat (And [new_rr_of_father; new_edge])) = false then False
    else (
        (* Step2: for each Support-Predicate, compute implications: *)
        let over_app_list = ref [] in
        List.iter
            (fun a_sp_pred -> 
                let a_new_sp_pred : predicate = 
                    change_index_of_variables_in_condition  
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (structs_hash) (current_vars_in_fun) 
                      (elements_of_arrays) (false) (false) 
                      (ref []) (a_sp_pred) (uniqueness_hash)
                in
                let remaining_vars : expression list = 
                    find_all_vars_in_predicate (a_new_sp_pred) in
                let new_rr_of_father =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                new_rr_of_father
                in
                let new_edge =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                new_edge
                in

                (* Step3: Check valid *)
                let implied_left : predicate = 
                    normalize (
                        remove_lit_clash ( 
                            And (new_rr_of_father :: [(*sliced_edge*)new_edge]))) 
                in
                let implication1 : predicate = 
                    Or ((Not implied_left) :: [a_new_sp_pred]) in
                let implication2 : predicate = 
                    Or ((Not implied_left) :: [Not a_new_sp_pred]) in
                    
                (* debug *)
                if !la_options.Checker.psopt_la_debug = true then begin
                    debug ("implication1: " ^ string_of_predicate implication1);
                    debug ("implication2: " ^ string_of_predicate implication2);
                end;
                if (is_valid implication1) = true then begin
                    (* debug *)
                    (*debug "implication1 is true";*)
                    over_app_list := !over_app_list @ [a_sp_pred]
                end else if (is_valid implication2) = true then begin
                    (* debug *)
                    (*debug "implication2 is true";*)
                    over_app_list := !over_app_list @ [Not a_sp_pred]
                end else (*debug "no implication";*) ()
            )
            sp_preds;
        (* Step4: return *)
        if !over_app_list = [] then True
        else begin
            let normalized_over_app_list = 
                normalize_overapproximation_list (!over_app_list) ([]) in
            (normalize (And normalized_over_app_list))
            (*normalize (And !over_app_list)*)
        end
    )
end

(* ~~~~~~~~~ get overapproximation for condition ~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_overapproximation_for_condition (sp_preds : predicate list)
                                        (rr_of_father : predicate) 
                                        (condition : predicate) 
: predicate =
begin
    (* region: debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug "_________________________________________";
        debug "get_overapproximation_for_condition: ";
        debug ("sp_preds: " ^ string_of_predicate (And sp_preds));
        debug ("rr_of_father: " ^ string_of_predicate rr_of_father);
        debug ("condition " ^ string_of_predicate condition);
    end;
    (* create hashtables to change index of variables *)
    let index_hash = Hashtbl.create (100) in
    let int_hash = Hashtbl.create (100) in
    let var_to_index = Hashtbl.create (100) in
    let index_to_var = Hashtbl.create (100) in
    let structs_hash = Hashtbl.create (100) in
    let current_vars_in_fun = ref [] in
    let elements_of_arrays : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let uniqueness_hash : (int, expression) Hashtbl.t = Hashtbl.create 0 in
    (* Step1: (Forward) change index *)
    let new_rr_of_father : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays)
          (true) (false) (ref []) (rr_of_father) (uniqueness_hash)
    in
    let new_condition : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays) 
          (false) (false) (ref []) (condition) (uniqueness_hash)
    in
    (* debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug ("new_rr_of_father: " ^ (string_of_predicate new_rr_of_father));
        debug ("new_condition: " ^ (string_of_predicate new_condition));
    end;

    (* Step2: Check if we can pass through condition or not *)
    let left_of_implication = 
        normalize (And [new_rr_of_father; new_condition]) in
    let is_passed = is_sat (left_of_implication) in
    (* debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug ("left_of_implication: " ^ (string_of_predicate left_of_implication));
        debug ("is_sat: " ^ (string_of_bool is_passed));
    end;
    if is_passed = false then False
    else begin
        (* Step3: foreach sp_pred, check if rr_of_father implies *)
        (* sp_pred, or (not sp_pred), or not                     *)
        let over_app_list = ref [] in
        List.iter
            (fun a_sp_of_child -> 
                let a_new_sp_of_child : predicate = 
                    change_index_of_variables_in_condition 
                      (index_hash) (int_hash) (var_to_index)(index_to_var) 
                      (structs_hash) (current_vars_in_fun) 
                      (elements_of_arrays) (false) (false) (ref []) (a_sp_of_child)
                      (uniqueness_hash)
                in
                let left_of_implication =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                left_of_implication
                in
                let implication1 = 
                    Or ((Not left_of_implication) :: [a_new_sp_of_child]) 
                in
                let implication2 =  
                    Or ((Not left_of_implication) :: [Not a_new_sp_of_child]) 
                in
                (* debug *)
                if !la_options.Checker.psopt_la_debug = true then begin
                    debug ("implication1: " ^ string_of_predicate implication1);
                    debug ("implication2: " ^ string_of_predicate implication2);
                end;
                (*let _ = read_int () in*)
                (* Step4: Check valid *)
                if (is_valid implication1) = true then
                    over_app_list := !over_app_list @ [a_sp_of_child]
                else if (is_valid implication2) = true then
                    over_app_list := !over_app_list @ [Not a_sp_of_child]                                 
            )
            sp_preds;
        (* Step5: return *)
        if !over_app_list = [] then True
        else begin
            let normalized_over_app_list = 
                normalize_overapproximation_list (!over_app_list) ([]) in
            (normalize (And normalized_over_app_list))
        end
    end
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_over_approximation_for_assume (sp_preds : predicate list)
                                      (rr_of_father : predicate) 
                                      (assume_edge : predicate)
: predicate = 
begin
    (* region: debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug "_________________________________________";
        debug "get_over_approximation_for_assume: ";
        debug ("sp_preds: " ^ string_of_predicate (And sp_preds));
        debug ("rr_of_father: " ^ string_of_predicate rr_of_father);
        debug ("condition " ^ string_of_predicate assume_edge);
    end;
    (* create hashtables to change index of variables *)
    let index_hash = Hashtbl.create (100) in
    let int_hash = Hashtbl.create (100) in
    let var_to_index = Hashtbl.create (100) in
    let index_to_var = Hashtbl.create (100) in
    let structs_hash = Hashtbl.create (100) in
    let current_vars_in_fun = ref [] in
    let elements_of_arrays : (expression, expression list) Hashtbl.t = 
        Hashtbl.create (0) in
    let uniqueness_hash : (int, expression) Hashtbl.t = Hashtbl.create 0 in
    (* Step1: (Forward) change index *)
    let new_rr_of_father : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays)
          (true) (false) (ref []) (rr_of_father) (uniqueness_hash)
    in
    let new_assume_edge : predicate = 
        change_index_of_variables_in_condition 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (elements_of_arrays)
          (false) (false) (ref []) (assume_edge) (uniqueness_hash)
    in
    (* debug *)
    if !la_options.Checker.psopt_la_debug = true then begin
        debug ("new_rr_of_father: " ^ (string_of_predicate new_rr_of_father));
        debug ("new_assume_edge: " ^ (string_of_predicate new_assume_edge));
    end;
    (* Step2: foreach sp_pred, check if rr_of_father implies *)
    (* sp_pred, or (not sp_pred), or not                     *)
    let left_of_implication : predicate = 
        normalize (And [new_rr_of_father; new_assume_edge]) in
    if (is_sat left_of_implication) = false then False
    else ( 
        let over_app_list = ref [] in
        List.iter
            (fun a_sp_of_child -> 
                let a_new_sp_of_child : predicate = 
                    change_index_of_variables_in_condition 
                      (index_hash) (int_hash) (var_to_index)(index_to_var) 
                      (structs_hash) (current_vars_in_fun) 
                      (elements_of_arrays) (false) (false) 
                      (ref []) (a_sp_of_child) (uniqueness_hash)
                in
                let left_of_implication =
                    La_common.replace_pointer_by_memory_in_pred int_hash
                                                                var_to_index
                                                                index_to_var
                                                                left_of_implication
                in
                let implication1 = 
                    Or ((Not left_of_implication) :: [a_new_sp_of_child]) 
                in
                let implication2 =  
                    Or ((Not left_of_implication) :: [Not a_new_sp_of_child]) 
                in
                (* debug *)
                if !la_options.Checker.psopt_la_debug = true then begin
                    debug ("implication1: " ^ string_of_predicate implication1);
                    debug ("implication2: " ^ string_of_predicate implication2);
                end;
                (* Step4: Check valid *)
                if (is_valid implication1) = true then
                    over_app_list := !over_app_list @ [a_sp_of_child]
                else if (is_valid implication2) = true then
                    over_app_list := !over_app_list @ [Not a_sp_of_child]                                 
            )
            sp_preds;
        (* Step5: return *)
        if !over_app_list = [] then True
        else begin
            let normalized_over_app_list = 
                normalize_overapproximation_list (!over_app_list) ([]) in
            (normalize (And normalized_over_app_list))
        end
    )
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_art_label (current_art : art)
                     (traced_labels : (string list, int) Hashtbl.t)
                     (current_apath : art_path) 
                     (current_cfa_node : cfa_node) 
: string * int =
begin
    let clabel = current_cfa_node.cfa_label in
    (* Transform apath to cpath *)
    let cpath : string list = art_path_to_cfa_path (current_apath) in
    (* Try to find in current_art.art_label_hash *)
    let version_in_art : int =
        (try Hashtbl.find current_art.art_label_hash clabel
        with Not_found -> -1)
    in
    (if version_in_art < 0 then begin
        (* Update information & return *)
        Hashtbl.add (current_art.art_label_hash) (clabel) (0);
        Hashtbl.add (traced_labels) (cpath) (0);
        (clabel, 0)
    end else begin
        (* Try to find in traced_labels *)
        let version_in_trace : int = 
            (try Hashtbl.find traced_labels cpath
            with Not_found -> -1)
        in
        if version_in_trace < 0 then begin
            (* Get version_in_art + 1*)
            Hashtbl.add (current_art.art_label_hash) (clabel) (version_in_art + 1);
            Hashtbl.add (traced_labels) (cpath) (version_in_art + 1);
            (clabel, version_in_art + 1)
        end else begin
            (* Else, get version in trace *)
            (clabel, version_in_trace)
        end
    end); 
    (* BEGIN: temporal comment out *)
    (* name of art_node is label of current_cfa_node *)
    (*let art_name : string = current_cfa_node.cfa_label in
    (* version of art_node is current_max_version + 1 *)
    let max_version : int =
        find_max_version_of_cfa_label_in_art (current_art) (art_name) in
    (* return *)
    (art_name, max_version + 1)*)
    (* END: temporal comment out *)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
   Input: 2 tuple_list (label & interpolant)                               
   Output: true if 2 tuple_list is the same, false otherwise               *)
let is_duplicated_labels_and_interpolants_tuples
      (new_labels_and_interpolants : ((string * int) * predicate) list)
      (old_labels_and_interpolants_tuples : ((string * int), predicate) Hashtbl.t) 
: bool =
begin
    match (List.length new_labels_and_interpolants) with
    | 0 -> false
    | _ -> (
        List.for_all
            (fun (label, new_itp) ->
                 let current_itps =
                     Hashtbl.find_all old_labels_and_interpolants_tuples label
                 in
                 List.exists
                     (fun pred ->
                          match (is_equivalent_predicates pred new_itp) with
                          | 0 -> true
                          | _ -> false)
                     current_itps)
            new_labels_and_interpolants)
end
  
(* ~~~~~~~~~ find all initializing global variables in gvar_cfas ~~~~~~~~~ 
   IMPORTANT: because gvar_cfas are contained in hashtbl, therefore we     
   can't get the order of the initializing. Sometime this causes a        
   failure  processing                                                     *)
let find_initializing_global_variables_of_cfas      
      (gvar_cfas : cfa list) 
: ((string * string * string * string * string) list) * (predicate list) = 
begin 
    let initializing_list : predicate list ref = ref [] in
    let memo_list : (string * string * string * string * string) list ref = ref [] in
    let gvar_cfas = List.filter (function
                                 | CFA_VARDEF v ->
                                     v.cfa_v_meta_data.cfa_scls != Cexpr.EXTERN
                                 | CFA_FUNDEF _
                                 | CFA_OTHERDEF -> assert false)
                                gvar_cfas
    in                                     
    List.iter
        (fun a_gvar_cfa ->
            match a_gvar_cfa with
            | CFA_VARDEF v ->
                let current_inits : predicate list =
                    v.cfa_v_meta_data.cfa_inits
                in
                let memo = (
                    match v.cfa_v_meta_data.cfa_location with
		    | None ->
		        ("", "", "", "", "")
	            | Some (Cexpr.LINE (_, s, real_i, _)) ->
		        (s, (string_of_int real_i), "", "", "")
		    | Some (Cexpr.RANGE (s, i1, i2, i3, i4)) ->
		        (s, (string_of_int i1), (string_of_int i2),
		            (string_of_int i3), (string_of_int i4)))
		in
                (match current_inits with
                | []
                | [True] -> begin
                    let default_init_opt =
                        let tn = v.cfa_v_meta_data.cfa_tn in
                        match tn with
                        | Cexpr.NOTYPE
                        | Cexpr.SPECIFR_TYPE _
                        | Cexpr.QUALIFIED_TYPE _
                        | Cexpr.ATTRIBUTED_TYPE _
                        | Cexpr.TYPEDEF_TYPE _
                        | Cexpr.STRCDEF_TYPE _
                        | Cexpr.ENUMDEF_TYPE _
                        | Cexpr.ELLIPSIS_TYPE
                        | Cexpr.PSEUDO_VA_LIST_TYPE -> None
                        (* Basic Types *)
                        | Cexpr.VOID_TYPE -> None
                        | Cexpr.BOOLEAN
                        | Cexpr.CHARACTER
                        | Cexpr.SIGNED_CHARACTER
                        | Cexpr.UNSIGNED_CHARACTER
                        | Cexpr.SHORT_INTEGER
                        | Cexpr.UNSIGNED_SHORT_INTEGER
                        | Cexpr.INTEGER
                        | Cexpr.UNSIGNED_INTEGER
                        | Cexpr.LONG_INTEGER
                        | Cexpr.UNSIGNED_LONG_INTEGER
                        | Cexpr.LONG_LONG_INTEGER
                        | Cexpr.UNSIGNED_LONG_LONG_INTEGER
                        | Cexpr.SINGLE_FLOAT
                        | Cexpr.DOUBLE_FLOAT
                        | Cexpr.LONG_DOUBLE_FLOAT
                        | Cexpr.SINGLE_COMPLEX
                        | Cexpr.DOUBLE_COMPLEX
                        | Cexpr.LONG_DOUBLE_COMPLEX
                        | Cexpr.SINGLE_IMAGINARY
                        | Cexpr.DOUBLE_IMAGINARY
                        | Cexpr.LONG_DOUBLE_IMAGINARY -> (
                            let new_var =
                                Variable (v.cfa_v_meta_data.cfa_id ^ "%1")
                            in
                            let new_init = Eq (new_var, Constant 0.0) in
                            Some new_init)
                        (* Composit Types *)
                        | Cexpr.POINTER_TO _ -> (
                            let new_var =
                                Variable (v.cfa_v_meta_data.cfa_id ^ "%1")
                            in
                            let new_init = Eq (new_var, Constant 0.0) in
                            Some new_init)
                        | Cexpr.ARRAY_OF _
                        | Cexpr.FUNCTION_OF _
                        | Cexpr.BITFIELD_TYPE _
                        | Cexpr.ENUMERATION_TYPE _
                        | Cexpr.STRUCTURE_TYPE _
                        (* GCC Extension *)
                        | Cexpr.GCC_VA_ARG -> None
                    in
                    match default_init_opt with
                    | None -> ()
                    | Some init -> (
                        initializing_list := !initializing_list @ [init];
                        memo_list := !memo_list @ [memo])
                  end
                | [an_init] -> begin
                    initializing_list := !initializing_list @ [an_init];
                    memo_list := !memo_list @ [memo]
                  end
                | a_list -> begin
                    initializing_list := !initializing_list @ [And current_inits];
                    memo_list := !memo_list @ [memo]
                  end)
            | anything_else -> assert false
        )
        gvar_cfas;
    (* return *)
    (!memo_list, !initializing_list);
end

(* ~~~~~~~~~~~~~~~ classify_cfas_into_funcs_and_gvars ~~~~~~~~~~~~~~~~~~~~ 
   Input: a hashtable contains func_cfas and gvar_cfas                     
   Output: a hashtable contains func_cfas & a hashtable contains gvar_cfas *)
let categorize_symbols (afile : Cfa.afile_t)
: Cfa.cfa list * Cfa.cfa list = 
begin 
    let variables = Hashtbl.create 10
    and functions = Hashtbl.create 10 in
    (* travel through global symbols first *)
    (Hashtbl.iter
        (fun global_name hashstr ->
             try
                 let lib = Hashtbl.find afile.afile_libs hashstr in
                 let cfa = Cfa.find_cfa_in_ofile lib global_name in
                 match cfa with
                 | CFA_VARDEF _ ->
                     Hashtbl.add variables global_name cfa
                 | CFA_FUNDEF _ ->
                     Hashtbl.add functions global_name cfa
                 | CFA_OTHERDEF ->
                     assert false
             with 
             | Not_found -> assert false)
        afile.afile_directory);
    (* now travel through each lib to collect local symbols *)
    (Hashtbl.iter
        (fun _ lib ->
             let local_prefix = lib.ofile_hash_name ^ "%" in
             let is_local name =
                 ExtString.String.starts_with name local_prefix
             in                 
             (Hashtbl.iter
                 (fun name cfa ->
                     match cfa with
                     | CFA_VARDEF _ ->
                         if (is_local name) then (
                             if not (Hashtbl.mem functions name) then
                                 Hashtbl.add variables name cfa)
                     | CFA_FUNDEF _ -> (
                         if not (Hashtbl.mem functions name) then 
                             Hashtbl.add functions name cfa;
                         let synonyms = Hashtbl.find_all variables name in
                         List.iter (fun _ ->
                                        Hashtbl.remove variables name)
                                   synonyms)
                     | CFA_OTHERDEF -> ())
                 lib.ofile_cfas))
          afile.afile_libs);
    let hashtbl_to_list tbl =
        Hashtbl.fold (fun _ x ls -> x::ls) tbl []
    in
    ((hashtbl_to_list variables), (hashtbl_to_list functions))
end
   
(* ~~~~~~~~~~~~~~~~ function find all global variables ~~~~~~~~~~~~~~~~~~~ *)
let find_gvars (gvar_cfas : cfa list) 
: expression list =
begin
    List.map (fun cfa ->
                  let id = match cfa with
                           | CFA_VARDEF v -> v.cfa_v_meta_data.cfa_id
                           | _ -> assert false
                  in (Variable (id ^ "%1")))
             gvar_cfas
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let assume_function_exprs_inequal_zero (func_cfas : cfa list)
: ((string * string * string * string * string) list) * (predicate list) = 
begin
    (* Foreach defined-function in cfas, assume its name != 0 *)    
    List.fold_left
        (fun memos_and_assumes cfa ->
            let memos, assumes = memos_and_assumes in
            (* Check a_cfa if it is defined-function or not *)
            match cfa with
            | CFA_FUNDEF f -> begin
                let meta_data = f.cfa_f_meta_data in
                let fun_expr : expression = 
                    Variable (meta_data.cfa_id ^ "%1")
                in
                let new_assume : predicate =
                    Not (Eq (fun_expr, Constant 0.0))
                in
                let memo = (
		    match meta_data.cfa_location with
		    | None ->
		        ("", "", "", "", "")
	            | Some (Cexpr.LINE (_, s, real_i, _)) ->
		        (s, (string_of_int real_i), "", "", "")
		    | Some (Cexpr.RANGE (s, i1, i2, i3, i4)) ->
		        (s, (string_of_int i1), (string_of_int i2),
		            (string_of_int i3), (string_of_int i4)))
                in
                (memos @ [memo], assumes @ [new_assume])
              end;
            | CFA_VARDEF _
            | CFA_OTHERDEF -> assert false
        )
        ([], [])
        func_cfas
end
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* ~~ preprocess_cfas: insert initializings of gvars into entry of main ~~ 
   IMPORTANT: because gvar_cfas are contained in hashtbl, therefore we     
   can't get the order of the initializing. Sometime this causes a        
   failure  processing                                                     *)
let preprocess_cfas (afile : Cfa.afile_t)
: (expression list) (* gvar list *) =
begin
    (* Collect global variables *)
    let var_cfas, func_cfas = categorize_symbols afile in
    (* Find global variables *)
    let gvars : expression list = find_gvars (var_cfas) in
    (* Find initializing of gvars *)
    let (memo_gvars, initializing_gvars) = 
        find_initializing_global_variables_of_cfas (var_cfas) in
    (* Assume function exprs inequal zero *)
    let (memo_fassumes, funcs_assumes) = 
        assume_function_exprs_inequal_zero func_cfas
    in
    (* Debug *)
    (*debug ("gvars: " ^ (string_of_expression (Sum gvars)));
    debug ("initializing_gvars: " ^ (string_of_predicate (And initializing_gvars)));*)
    (* find main function *)
    let main_cfa : cfa =
        (try Cfa.find_cfa_of_function_entry afile afile.Cfa.afile_start
         with
         | Not_found -> (
             let error_msg =
                 ("Cannot find the entry function `"
                    ^ (Sexpr.string_of_function_entry afile.Cfa.afile_start)
                    ^ "'")
             in
             (error error_msg);
             failwith error_msg))
    in
    (* find entry of main fuction *)
    (match main_cfa with
    | CFA_FUNDEF m_fundef -> begin
        let entry_label = m_fundef.cfa_entry_label in 
        let nodes_hash = m_fundef.cfa_node_hashtbl in 
        let m_root = (try Hashtbl.find (nodes_hash) (entry_label)
                     with | Not_found -> assert false)
        in
        (* Create 2 new node and add to the art tree*)
        let after_assume_fname_label : string = "After_assume_fname" in
        let initializing_gvars_label = "initializing_gvars" in
        let new_assume_m_root : cfa_node = {
            cfa_label = m_root.cfa_label;
            cfa_outgoing_labels = [after_assume_fname_label];
            cfa_incoming_labels = m_root.cfa_incoming_labels;
            cfa_memo = memo_fassumes;
            cfa_node_stmt = CFA_ENTRY (Not (Not (And funcs_assumes)));
            cfa_inputs = m_root.cfa_inputs;
            cfa_outputs = m_root.cfa_inputs;
        } in
        (* replace m_root by new_assume_m_root *)
        Hashtbl.replace (nodes_hash) (entry_label) (new_assume_m_root);
        let after_assume_fname_stmt = (
            match m_root.cfa_node_stmt with
            | CFA_ENTRY pred_of_m_root -> CFA_FORM pred_of_m_root
            | _ -> assert false
        ) in
        let after_assume_fname_node : cfa_node = {
            cfa_label = after_assume_fname_label ;
            cfa_outgoing_labels = [initializing_gvars_label];
            cfa_incoming_labels = [m_root.cfa_label];
            cfa_memo = memo_gvars @ m_root.cfa_memo;
            cfa_node_stmt = after_assume_fname_stmt;
            cfa_inputs = m_root.cfa_inputs;
            cfa_outputs = m_root.cfa_outputs;
        } in
        let flatten_initializing_gvars = ref [] in
        List.iter (
            fun an_init ->
            begin
                match an_init with
                | And preds -> flatten_initializing_gvars :=
                                   !flatten_initializing_gvars @ preds;
                | pred      -> flatten_initializing_gvars :=
                                   !flatten_initializing_gvars @ [pred];
            end
        ) initializing_gvars;
        let initializing_node : cfa_node = {
            cfa_label = initializing_gvars_label;
            cfa_outgoing_labels = m_root.cfa_outgoing_labels;
            cfa_incoming_labels = [after_assume_fname_label];
            cfa_memo = memo_gvars @ m_root.cfa_memo;
            cfa_node_stmt = CFA_FORM (And !flatten_initializing_gvars);
            cfa_inputs = m_root.cfa_inputs;
            cfa_outputs = m_root.cfa_outputs;
        } in
        (* Update after_assume_fname_node & initializing_node *)
        Hashtbl.add (nodes_hash) (initializing_gvars_label) (initializing_node);
        Hashtbl.add (nodes_hash) (after_assume_fname_label) (after_assume_fname_node);
      end
    | anything_else -> assert false); 
    (* return *)
    gvars
end
 
(* ~~~~~~ function: create_temp_art_node to use in create_art_node ~~~~~~~ *)
let create_temp_art_node (current_cfa_node : cfa_node)
                         (current_rr : predicate)
                         (current_sps : predicate list)
                         (current_incoming_label : string * int) 
                         (current_art : art)
                         (current_apath : art_path)
                         (traced_labels : (string list, int) Hashtbl.t)
                         (is_trust : bool)
: art_node =
begin
    StatisticUtil.update_node_explored_number ();
    (* compute some attributes of the needed temp_art_node *)
    let new_art_type = 
        transform_cfa_stmt_to_art_stmt_type (current_cfa_node) in
    let (new_art_name, new_art_version) =
        create_art_label (current_art) (traced_labels) 
        (current_apath) (current_cfa_node) in
    let is_covered_by : string * int = 
        covered_by (current_art) (current_rr) (new_art_name) in
    (*let new_dependent_labels = ref [] in*)
    let is_covered_node = (match is_covered_by with
                          | ("", -1) -> false
                          | _ -> (*new_dependent_labels := [is_covered_by];*) true
    ) in
    (* create temp_art_node *)
    let temp_art_node : art_node =
        create_new_art_node (new_art_type) (new_art_name, new_art_version) 
          ([]) ([]) (current_incoming_label)(current_rr) [] (current_sps) 
          (is_covered_node) (false) (*!new_dependent_labels*)
          (is_trust) (current_cfa_node.cfa_memo)
    in
    (* return *)
    temp_art_node
end
  
(* ~~~~~~~~~~~~~~~ compute_reachable_region_of_art_child ~~~~~~~~~~~~~~~~~ *)
let compute_rr_of_art_child (type_of_father : art_stmt)
                            (rr_of_art_father : predicate)
                            (outgoing_edge : predicate)
                            (sps_of_child : predicate list)
: predicate (* reachable_region of art_child *) =
begin
    (* create some vars to return *)
    let new_rr_of_art_child = (
        match type_of_father with
        | ART_RESET -> 
            get_over_approximation_for_reset (sps_of_child)
              (rr_of_art_father) (outgoing_edge)
        | ART_ASSUME -> 
            get_over_approximation_for_assume (sps_of_child)
              (rr_of_art_father) (outgoing_edge)
        (* Temporarily deal with ART_ASSUME *)
        | ART_ENTRY | ART_FORM | ART_CALL | ART_RETURN ->
            get_overapproximation_for_assignment (sps_of_child)
              (rr_of_art_father) (outgoing_edge);
        | ART_COND ->
            get_overapproximation_for_condition (sps_of_child)
              (rr_of_art_father) (outgoing_edge);
        | ART_ERROR | ART_FAIL | ART_NORETURN -> assert false
    ) in
    (* region: debug *)
    (*debug ("new_rr_of_art_child: \n" ^ string_of_predicate new_rr_of_art_child);*)
    (* return *)
    new_rr_of_art_child
end
  
(* ~~~~~~~~~~~~~~~~~~~ create an error_art_node ~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* When we reach to an error_cfa_node with reachable_region of it is false *)
(* , we create an art_node and return                                      *)
let create_error_art_node (current_cnode : cfa_node)
                          (current_rr : predicate) 
                          (current_sps : predicate list)
                          (current_incoming_label : string * int) 
                          (temp_art : art) 
                          (current_art : art)
                          (traced_labels : (string list, int) Hashtbl.t)
                          (current_apath : art_path) 
: bool (* flag denotes the traverse to create art is successful or not *) 
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *) 
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node 
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *)  =
begin
    let (new_name, new_version) =
        create_art_label (current_art) (traced_labels) 
          (current_apath) (current_cnode) in

    (* create new error_art_node *)
    let new_anode : art_node =
        create_new_art_node (ART_ERROR) (new_name, new_version) ([]) ([])
          (current_incoming_label) (current_rr) [] (current_sps) (false) 
          (false) (*[]*) (true) (current_cnode.cfa_memo)
    in
    (* insert new_anode & information of it into current_art *)
    Hashtbl.add (current_art.art_node_hash) (new_name, new_version) (new_anode);
    Hashtbl.add (current_art.art_version_hash)
                (new_name)
                (new_version, current_rr);
    Hashtbl.add (current_art.art_trust_nodes) (new_name, new_version) new_anode;

    (* insert information of new_anode into temp_art *)
    if !la_options.Checker.psopt_dot_output = true then begin
        Hashtbl.add (temp_art.art_node_hash) (new_name, new_version) (new_anode);
    end;
    (* return *)
    (true, true, Some new_anode, [new_anode.art_label], 
     (*[],*)(("", - 1), []), (Some temp_art, []))
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
Used when we process an art_node, if we go to branch2 of a condition,     
If current-number-of-functioncall < number in "number_of_fc_hash", then 
assign: current-number-of-functioncall = number in "number_of_fc_hash"   *)
(*let insert_to_number_of_fc_hash (a_label: string * int) 
                                (number_of_fc : int)
                                (number_of_fc_hash : (string * int, int) Hashtbl.t) 
=
begin
    try (
        let old_number_of_fc =
            Hashtbl.find (number_of_fc_hash) (a_label) in
        if old_number_of_fc < number_of_fc then begin
            Hashtbl.remove (number_of_fc_hash) (a_label);
            Hashtbl.add (number_of_fc_hash) (a_label) (number_of_fc);
        end
    )
    with
    | Not_found ->
        Hashtbl.add (number_of_fc_hash) (a_label) (number_of_fc)
end
*)    
 
let find_in_number_of_fc_hash (number_of_fc_hash : (string list, int) Hashtbl.t)
                              (current_path : art_path)
                              (current_number_of_fc : int)
: int = 
begin
    (* Get cpath of current art_path *)
    let current_cpath : string list = art_path_to_cfa_path (current_path) in
    (* Try to find in number_of_fc_hash *)
    try (
        Hashtbl.find (number_of_fc_hash) (current_cpath)
    ) with Not_found -> begin
        (* If current_cpath is not in number_of_fc_hash, then insert it into
        number_of_fc_hash, then return                                       *)
        Hashtbl.add (number_of_fc_hash) (current_cpath)
                    (current_number_of_fc + 1);
        (* Return *)
        (current_number_of_fc + 1)
    end
end
(* ~~~~~~~~~~~~~~ update support_predicate foreach cfa_node ~~~~~~~~~~~~~~ *)
let update_cnodes_and_sps (new_labels_and_itps : ((string * int) * predicate) list)
      (current_label_and_itps_tuples : ((string * int), predicate) Hashtbl.t)
      (sps_hash : ((string * int), predicate) Hashtbl.t) =
begin
    (* check if inputs duplicate or not before update *)
    let is_duplicated : bool = 
        is_duplicated_labels_and_interpolants_tuples (new_labels_and_itps) 
          (current_label_and_itps_tuples)
    in
    if is_duplicated = false then begin
        (* update current tuples: *)
        List.iter
            (fun (a_label, an_itp) ->
                 Hashtbl.add current_label_and_itps_tuples a_label an_itp)
            new_labels_and_itps;
        (* update sps_hash: *)
        List.iter
            (fun (a_label, an_itp) ->
                (* get atom_preds of an_itp *)
                let new_atom_preds : predicate list = 
                    get_atom_predicates_from_predicate (an_itp) in
                (* update each atom_pred into sps_hash *)
                let old_atom_preds : predicate list = 
                    try Hashtbl.find_all (sps_hash) (a_label)
                    with Not_found -> []
                in
                List.iter
                    (fun a_new_atom_pred -> 
                        let is_added = ref false in
                        is_added := member_of_predicate_list 
                                        (Not a_new_atom_pred)
                                        (old_atom_preds);
                        if !is_added = false then
                            is_added := member_of_predicate_list
                                            (a_new_atom_pred)
                                            (old_atom_preds);
                        if !is_added = false then
                            Hashtbl.add (sps_hash) (a_label) (a_new_atom_pred)
                    )
                    new_atom_preds
            )
            new_labels_and_itps;
    end else begin 
        emerg "labels_and_interpolants_tuples is duplicate";
        assert false;                                               
        (*raise (ConfigUtil.Message 
                  (Logging.Error,
                      "labels_and_interpolants_tuples is duplicate: \n "));*)
    end
end
  
(* ~~~~~~~~~~~~~ define global variables for create art_node ~~~~~~~~~~~~~ *)
let current_art : art ref = ref (create_new_art (0) ("", - 1));;
let current_path : art_path ref = ref (create_new_art_path ());;
let check_iter_max : bool ref  = ref false;;
let is_duplicate : bool ref = ref false;;
let unsafe_because_iter_max = ref false;;
let exist_feasible_iter_max_paths : bool ref = ref false;;
let current_cfa_fundef : cfa_fundef_t ref = ref (create_nil_cfa_fundef ());;
let cfas_hash : (string, cfa) Hashtbl.t ref = ref (Hashtbl.create (1));;
let sps_hash : ((string * int), predicate) Hashtbl.t ref = 
    ref (Hashtbl.create (0));;
let gvars : expression list ref = ref [];;
(* times for checking the feasibility of iter-max paths, -1 means unlimited times*)
let feasible_iter_max_times : int ref = ref (-1);; 
let current_number_of_fc : int ref = ref 0;;
let current_label_and_itps_tuples : ((string * int), predicate) Hashtbl.t
    = Hashtbl.create 107;;
(*let number_of_fc_hash : (string * int, int) Hashtbl.t ref = 
    ref (Hashtbl.create (0));;*)
let number_of_fc_hash : (string list, int) Hashtbl.t ref = 
    ref (Hashtbl.create (0));;
let traced_labels : (string list, int) Hashtbl.t ref =
    ref (Hashtbl.create (0));;
let hash_to_count_cfa_nodes : (string, int) Hashtbl.t ref = 
    ref (Hashtbl.create (0));;
let fc_pointer_flag : bool ref = ref true;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 20090710. Temporarily find last_condition_label to slicing error path  
   to print                                                                *)
let find_last_cond_label_of_path (path : art_path) : (string * int) = 
begin
    let last_cond_index : int = (List.length (path.art_nodes_in_path)) - 9 in
    try (
        let last_art_cond : art_node =  
            List.nth (path.art_nodes_in_path) (last_cond_index) in
        last_art_cond.art_label
    ) with Invalid_argument _ -> ("", -1) 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_name_func_of_art_call (art_node : art_node)
                              (edge : predicate) : string =
begin
    let get_name_func_in_pred (pred : predicate) : string = (
        match pred with
        | Eq (e1, e2) -> (
            if (is_identical_expressions e1 (Variable "temp_out")) then (
                match e2 with
                | Variable func_name -> func_name
                | _ -> ""
              )
            else ""
          )
        | _ -> ""
      ) in

    if art_node.art_type = ART_CALL then (
        match edge with
        | Eq _ -> get_name_func_in_pred edge
        | And lst -> (
            let tmp_name = ref "" in
            List.iter
                (fun a_pred ->
                    if ((String.compare !tmp_name "") = 0) then ( 
                        let tmp_func_name = get_name_func_in_pred a_pred in
                        tmp_name := tmp_func_name;)
                ) lst;
            (!tmp_name)
          )
        | _ -> ""
      )
    else ""
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let location_of_error_assert (path : art_path) 
: string * string * string * string * string = begin
    let rev_nodes = List.rev path.art_nodes_in_path in
    let rev_edges = List.rev path.art_edges_in_path in
    let location = ref ("", "", "", "", "") in
    let is_found = ref false in
    (try
        List.iter2
            (fun anode aedge -> 
                if !is_found = false then begin (
                    if anode.art_type = ART_COND then begin
                        (location :=
                             try List.hd anode.art_memo
                             with Failure _ -> assert false);
                        is_found := true;
                    end;
                    if anode.art_type = ART_CALL then (begin
                        let func_name =
                            get_name_func_of_art_call anode aedge in
                        if (ExtString.String.ends_with func_name "__assert_fail") then (
                            (location :=
                                try List.hd anode.art_memo
                                with Failure _ -> assert false);
                             is_found := true;)
                    end;))
                end
            ) rev_nodes rev_edges;
    with Invalid_argument _ -> assert false;); 
    !location
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)           
let update_traverse_time (current_outgoing_tuples)
                         (current_traversed_time : int) 
                         (current_cfa_node : cfa_node) 
: unit = 
begin
    (* debug *)
    (*debug ("update_traverse_time: " );*)
    if (List.length current_outgoing_tuples) = 2 then begin
        if current_traversed_time = 0 then begin
            Hashtbl.replace (!hash_to_count_cfa_nodes) 
                        (current_cfa_node.cfa_label) (1) 
        end else begin
            Hashtbl.replace (!hash_to_count_cfa_nodes) 
                            (current_cfa_node.cfa_label) 
                            (current_traversed_time + 1);
        end 
    end;
end  

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let revert_traverse_time (current_outgoing_tuples)
                         (current_traversed_time : int)
                         (current_cfa_node : cfa_node)
: unit =
begin
    if (List.length current_outgoing_tuples) = 2 then begin
        if current_traversed_time <> 0 then begin
            Hashtbl.replace (!hash_to_count_cfa_nodes)
                            (current_cfa_node.cfa_label)
                            (current_traversed_time - 1);
        end;
    end;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* When the checker sees incorrect pointer dereferences, the reported pointers
   usually are a temporary variables created inside the checker. Therefore,
   error messages in such cases dont give users meaningful hints.
   This function, based on assignments along a error path, tries to extract 
   a closest `meaningful' expression of a null pointer, and hopefully gives
   users better error messages *)
let get_closest_expr_of_null_expr (null_expr : expression)
                                  (error_path : art_path)
: expression (* closest expr *) =
begin
    (* get closest expr *)
    let rec get_closest_expr (expr : expression)
                             (edge_list : predicate list)
    : expression (* closest expression *) =
    begin
        match expr with
        | Constant _ -> expr
        | Variable v ->
            if (Cexpr.temporary_label_p v) then
                get_closest_variable_from_edge_list expr edge_list edge_list
            else expr
        | Application (symbol, expr_list) ->
            let new_expr_list =
            get_closest_expr_list expr_list edge_list in
            Application(symbol, new_expr_list)
        | Sum expr_list ->
            let new_expr_list = get_closest_expr_list (expr_list) (edge_list) in
            Sum(new_expr_list)
        | Coeff (f, e) ->
            let e1 = get_closest_expr (e) (edge_list) in
            Coeff(f, e1)
    end
    (* get closest exprs of expr_list *)
    and get_closest_expr_list (expr_list : expression list)
                              (edge_list : predicate list)
    : expression list (* closest expression list *) =
    begin
        match expr_list with
        | [] -> []
        | h::t -> (get_closest_expr (h) (edge_list))
                  ::(get_closest_expr_list (t) (edge_list))
    end
    (* get closest variable from edge list *)
    and get_closest_variable_from_edge_list (expr : expression)
                                            (edge_list : predicate list)
                                            (edge_list_full : predicate list)
    : expression (* closest variable *) =
    begin
        match edge_list with
        | [] -> assert false
        | h::t ->
            (match h with
            | True | False | Not _ | Lt _ | Leq _ | Atom _ ->
                get_closest_variable_from_edge_list expr t edge_list_full
            | And pred_list ->
                let new_edge_list = pred_list @ t in
                get_closest_variable_from_edge_list expr new_edge_list
                    edge_list_full
            | Or pred_list -> assert false
            | Eq (left_expr, right_expr) ->
                if expr = left_expr then
                    get_closest_expr right_expr edge_list_full
                else get_closest_variable_from_edge_list expr t
                    edge_list_full)
    end
    (* filter and reverse edge_list *)
    and filter_and_reverse_edge_list (node_list : art_node list)
                                     (edge_list : predicate list)
    : predicate list =
    begin
        try
            List.fold_left2
            (fun (f_result : predicate list) 
                 (anode : art_node) (an_edge : predicate) ->
            begin
                match anode.art_type with
                | ART_FORM  | ART_CALL | ART_RESET| ART_ENTRY
                | ART_RETURN | ART_ERROR | ART_FAIL | ART_NORETURN ->
                    an_edge::f_result;
                | ART_COND | ART_ASSUME ->
                    f_result;
            end) 
            [] node_list edge_list
        with
        | Invalid_argument _ -> assert false
    end
    (* entry function *)
    in
    let edge_list = filter_and_reverse_edge_list
                    error_path.art_nodes_in_path
                    error_path.art_edges_in_path
    in
    (* left out for a moment
    get_closest_expr null_expr edge_list *)
    null_expr
end

let is_pointer (name : string)
               (afile : Cfa.afile_t)
               ?(caller = None)
               ()
: bool = (
    (try
        match Cfa.find_cfa_in_afile afile
                                    ~caller:caller
                                    name
        with
        | Cfa.CFA_VARDEF vardef -> (
            let meta_data = vardef.Cfa.cfa_v_meta_data in
            match meta_data.cfa_tn with
            | Cexpr.POINTER_TO _ -> true
            | _ -> false)
        | _ -> false
    with Not_found -> false))

(** check if a function pointer has been initialized or not *)
let check_init_func_pointer (path : Art.art_path)
                            (current_cfa_node : Cfa.cfa_node)
                            (afile : Cfa.afile_t)
                            (current_cfa : Cfa.cfa)
: bool * string = (
    match current_cfa_node.cfa_node_stmt with
    | CFA_CALL (exprs, _) -> (
        let func_var = try List.hd exprs
                       with _ -> assert false
        in
        let (func_name, func_level) =
            Cfa.find_name_and_level_of_expr_func func_var
        in
        let pred = And path.Art.art_edges_in_path in
        let is_pointer =
            is_pointer func_name afile ~caller:(Some current_cfa) ()
        in
        if is_pointer then (
            match Cfa.function_pointer_in_current_rr_points_to pred func_var with
            | None -> (false, func_name)
            | Some real_func -> (
                try
                    let real_func_name, _ =
                        Cfa.find_name_and_level_of_expr_func (Variable real_func)
                    in
                    let callee_cfa =
                        Cfa.find_cfa_in_afile afile
                                              ~caller:(Some current_cfa)
                                              real_func_name
                    in
                    match callee_cfa with
                    | Cfa.CFA_FUNDEF _
                    | _ -> (true, "")
                with Not_found -> (true, "")))
        else (true, ""))
    | _ -> (true, ""))

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* List of node itermax *)
let itermax_context_list : la_error_context_t list ref = ref []
let start_time = ref 0.

let rec create_art_stop_node (current_cfa_node : cfa_node)
                             (current_rr : predicate)
                             (current_sps : predicate list) 
                             (current_incoming_label : string * int) 
                             (temp_art : art) 
                             (is_trust : bool)
: bool (* flag denotes the traverse to create art is successful or not *) 
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *) 
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node 
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *)  =
begin
    (* create art_node *)
    let new_anode : art_node = 
        create_temp_art_node (current_cfa_node) (current_rr)
          (current_sps) (current_incoming_label) (!current_art) 
          (!current_path) (!traced_labels) (is_trust)
    in
    let (new_art_name, new_art_version) = new_anode.art_label in
    (* insert new_anode into hashtbls *)
    if is_trust = true then begin
        Hashtbl.add (!current_art.art_version_hash) (new_art_name)
          (new_art_version, current_rr);
          (* Note: a stop_node is never included in trust_nodes since
                   if it is trusted due to is_covered, it cannot be
                   100% trusted, especially if the node is in a loop *)
    end;
    Hashtbl.add (!current_art.art_node_hash) (new_anode.art_label) (new_anode);
    if !la_options.Checker.psopt_dot_output = true then begin
        Hashtbl.add (temp_art.art_node_hash) (new_anode.art_label) (new_anode);
    end;
    (* return *)
    (true, is_trust, Some new_anode, [new_anode.art_label], 
    (*new_anode.art_dependent_labels,*) (("", - 1), []),(Some temp_art, []))
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_art_node_when_reachs_error_label (current_cfa_node : cfa_node)
                                            (current_rr : predicate) 
                                            (current_sps : predicate list)
                                            (current_incoming_label : string * int) 
                                            (temp_art : art)
: bool (* flag denotes the traverse to create art is successful or not *) 
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *) 
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node 
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *) =
begin
    StatisticUtil.update_number_of_preds_in_paths
        (LogicUtil.Logic.count_atom_pred_in_preds !current_path.art_edges_in_path);
    StatisticUtil.update_refinement_number ();
    (* if rr of error_cfa_node is False, then create error_art_node: *)
    if current_rr = False then begin
        create_error_art_node (current_cfa_node) (current_rr)
          (current_sps) (current_incoming_label) (temp_art) (!current_art)
          (!traced_labels) (!current_path)
    (* else, check_sat error_path & compute interpolants to refine: *)
    end else (begin
        let error_label : string * int = 
            create_art_label (!current_art) (!traced_labels) (!current_path)
              (current_cfa_node) in
        let new_art_node : art_node =
            create_new_art_node (ART_ERROR) (error_label) ([]) ([]) 
              (current_incoming_label) (current_rr) [] (current_sps) (false) 
              (false) (*[]*) (false) (current_cfa_node.cfa_memo)
        in
        (* update information about error_node into temp_art *)
        if !la_options.Checker.psopt_dot_output = true then begin
            Hashtbl.add (temp_art.art_node_hash) (error_label) (new_art_node);
            let old_labels = get_art_labels_in_path (!current_path) in 
            temp_art.art_labels_in_error_path <- (old_labels) @ [new_art_node.art_label];
        end;

        (* Try to change index of vars in path, then analysis the path *)
        try (
            let (pivot_label, labels_and_interpolants) = 
                (if !fc_pointer_flag = true then begin
                    let (temp_pivot_label, temp_labels_and_interpolants) = 
                        analysis_function_pointers_in_path
                            (!current_path)
                            (!gvars)
                            (!la_options.Checker.psopt_cutlist)
                    in
                    let fp_duplicated : bool = 
                        is_duplicated_labels_and_interpolants_tuples 
                          (temp_labels_and_interpolants) (current_label_and_itps_tuples)
                    in
                    if (fp_duplicated = true) || 
                        ((compare_art_labels temp_pivot_label ("", -1)) == 0) then begin
                        if !la_options.Checker.psopt_slice_path = 1 then
                            analysis_path_normally (!current_path) (!gvars) 
                        (* 24.3.2009. Try new idea *)
                        else
                            analysis_path_with_WeiBenbacher_paradigm (!current_path) 
                             (!gvars)
                    end else (temp_pivot_label, temp_labels_and_interpolants)
                end else begin
                    if !la_options.Checker.psopt_slice_path = 1 then
                        analysis_path_normally (!current_path) (!gvars) 
                    (* 24.3.2009. Try new idea *)
                    else
                        analysis_path_with_WeiBenbacher_paradigm (!current_path) 
                          (!gvars)
                end)
            in
            (* 27.2.2009. Xuan has added to deal with duplicate *)
            (* check if inputs duplicate or not: *)
            let is_duplicated : bool = 
                is_duplicated_labels_and_interpolants_tuples 
                  (labels_and_interpolants) (current_label_and_itps_tuples)
            in
            if is_duplicated = true then begin
                if !check_duplicated_error = false then begin
                    create_art_stop_node (current_cfa_node) (current_rr)
                      (current_sps) (current_incoming_label) (temp_art) (false)
                end else begin
                    emerg "interpolations is duplicated";
                    is_duplicate := true;
                    (false, false, Some new_art_node, [], (*[],*) (("", -1), []), (None, [temp_art]))
                end;
            (* 27.2.2009. end *)
            end else begin
                debug "functions in the path: ";
                debug (find_all_func_names_in_path !current_path);
                (* 20090710. Print feasible path -- START*)
                let last_cond_label = (find_last_cond_label_of_path !current_path) in
                if ((compare last_cond_label ("", -1)) != 0) &
                   ((compare_art_labels pivot_label ("", -1)) = 0) then begin
(*                     let changed_flatten_path : art_path =  *)
(*                         change_index_of_variables_in_path (!current_path) (!gvars)  *)
(*                           (true) (true) (false) (ref []) in *)
                    (* START-ENCODE *)
                    (* Encode path *)
(*                     let (encoded_edges_in_path, encoded_hash) =  *)
(*                         encode_vars_in_path (changed_flatten_path.art_edges_in_path) in *)
(*                     let encoded_flatten_path : art_path = { *)
(*                         art_nodes_in_path = changed_flatten_path.art_nodes_in_path; *)
(*                         art_edges_in_path = encoded_edges_in_path; *)
(*                     } in *)
                    (* STOP-ENCODE *)
                    (*let sliced_error_path : art_path = 
                        slice_path_with_criteria_labels (encoded_flatten_path) ([last_cond_label])
                    in*) ()
                    (*debug "trace-path: ";*)
                    (*debug (get_string_of_cutting_apath_2 sliced_error_path !current_path);*)
                end;
                (* 20090710. Print feasible path --END *)
                (* return *)
                if (compare_art_labels pivot_label ("", -1)) = 0
                    & (not !exist_feasible_iter_max_paths) then begin
                    la_error :=
                        {
                          ctx_error_path = copy_art_path !current_path;
                          ctx_error_location = 
                              location_of_error_assert !current_path;
                          ctx_error_message = "" (* ??? *)
                        };
                end;
                (false, false, Some new_art_node, [], (*[],*) 
                (pivot_label, labels_and_interpolants), (None, [temp_art]))
            end
        (* Catch error Not_point_to raised from changing index of vars in path *)
        ) with 
        | La_exception.Not_point_to (alabel, null_expr1, null_expr2, need_addrof_exprs) -> 
          begin
(*             let changed_flatten_path : art_path =  *)
(*                 change_index_of_variables_in_path (!current_path) (!gvars)  *)
(*                   (true) (true) (false) (ref []) in *)
            (* START-ENCODE *)
            (* Encode path *)
(*             let (encoded_edges_in_path, encoded_hash) =  *)
(*                 encode_vars_in_path (changed_flatten_path.art_edges_in_path) in *)
(*             let encoded_flatten_path : art_path = { *)
(*                 art_nodes_in_path = changed_flatten_path.art_nodes_in_path; *)
(*                 art_edges_in_path = encoded_edges_in_path; *)
(*             } in *)
            (* Encode null_expr2 & need_addrof_exprs *)
(*             let encoded_null_expr2 = encode_vars_in_expression (null_expr2) (encoded_hash) in *)
            (*let encoded_addr_exprs = encode_vars_in_expressions (need_addrof_exprs) (encoded_hash) in*)
            (* STOP-ENCODE *)
            (*let sliced_error_path : art_path = 
                slice_path_backward_from_label_with_criterion_vars 
                  (encoded_flatten_path) ([encoded_null_expr2] (*@ encoded_addr_exprs*)) 
                  (alabel)
            in*)
            (*debug "trace-path: ";*)
            (*debug (get_string_of_cutting_apath_2 sliced_error_path !current_path);*)
            debug "functions in the path: ";
            let cut_prime_path : art_path = cut_path (!current_path) (alabel) in
            debug (find_all_func_names_in_path cut_prime_path);
            error "-------------------------------------------------";
            let closest_expr = 
                get_closest_expr_of_null_expr null_expr1 (!current_path) in
            error ("In " ^ (string_of_art_label alabel) ^ " : " ^
                          (string_of_expression closest_expr) ^ "  --- (" ^
                          (string_of_expression null_expr2) ^ ") --- " ^
                          " points to nowhere");
            (* return *)
            let enode =
                find_anode_in_path (cut_prime_path.art_nodes_in_path) (alabel) in
            (match enode with 
            | Some enode ->
                if (not !exist_feasible_iter_max_paths) then  
                    la_error :=
                        {
                          ctx_error_path = copy_art_path cut_prime_path;
                          ctx_error_location = 
                              (try List.hd enode.art_memo
                              with Failure _ -> assert false);
                          ctx_error_message =
                              (string_of_expression
                                  (La_common.create_based_vars_in_expression null_expr1))
                              ^ " points to nowhere"
                        }
                else ()
            | None -> assert false); 
            (false, false, Some new_art_node, [], (*[],*) (("", -1), []), (None, [temp_art]))
          end
        | La_exception.Weak_constraints pred_list -> 
          begin
            emerg ("Weak-constraints: ");
            List.iter
                (fun p -> 
                    if (normalize p) != True then
                        emerg (string_of_predicate p);
                )
                pred_list;
            if (not !exist_feasible_iter_max_paths) then
                la_error := {
                              ctx_error_path = copy_art_path !current_path;
                              ctx_error_location = 
                                  location_of_error_assert !current_path;
                              ctx_error_message =
                                  "LogicProver can not prove this path";
                            }
            else ();
            (false, false, Some new_art_node, [], (("", -1), []), (None, [temp_art]))
          end
    end);
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_art_node_when_reachs_iter_max (current_cfa_node : cfa_node)
                                         (current_rr : predicate)
                                         (current_sps : predicate list)
                                         (current_incoming_label : string * int)
                                         (temp_art : art)
: bool (* flag denotes the traverse to create art is successful or not *)
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *)
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *)  =
begin
    try (
        (* check the feasibility of the path reaching iter-max *) 
        let (pivot_label, labels_and_interpolants) = (
            if !la_options.Checker.psopt_slice_path = 1 then
                analysis_path_normally (!current_path) (!gvars)
            else
                analysis_path_with_WeiBenbacher_paradigm (!current_path)
                                                         (!gvars)
        ) in
        let is_feasible = (compare_art_labels pivot_label ("", -1) = 0) in
        if not !exist_feasible_iter_max_paths then
            exist_feasible_iter_max_paths := is_feasible;
        if !exist_feasible_iter_max_paths then (
            la_error :=
              {
                ctx_error_path = copy_art_path !current_path;
                ctx_error_location = 
                    location_of_error_assert !current_path;
                ctx_error_message =
                    "The current iter-max number '"
                    ^ (string_of_int !la_options.Checker.psopt_iter_max)
                    ^ "' is not enough to decide safe/unsafe. "
                    ^ "You may need to increase the iter-max number"
              };
        )
        else ();
        if (!feasible_iter_max_times = 0) then
        begin
            check_iter_max := true;
            if (is_feasible) then
            begin
                let error_label : string * int =
                    create_art_label (!current_art) (!traced_labels) 
                        (!current_path) (current_cfa_node) 
                in
                let error_node : art_node =
                    create_new_art_node (ART_ERROR)
                                        (error_label)
                                        ([])
                                        ([])
                                        (current_incoming_label)
                                        (current_rr)
                                        ([])
                                        (current_sps)
                                        (false)
                                        (false)
                                        (false)
                                        (current_cfa_node.cfa_memo)
                in
                (false, false, Some error_node, [], 
                 (("", -1), []), (None, [temp_art]));
            end
            else 
                create_art_stop_node (current_cfa_node)
                                     (current_rr)
                                     (current_sps)
                                     (current_incoming_label)
                                     (temp_art)
                                     (false);
        end
        else
        begin
            check_iter_max := true;
            (* if feasible_iter_max_times = -1 then always check deeply
               for any iter-max paths, else decrease the number of times
               check deeply for iter-max paths *)
            if (!feasible_iter_max_times > 0) then
                feasible_iter_max_times := !feasible_iter_max_times - 1;
            (* create a stop node for iter-max in order to find out
               another path *)
            create_art_stop_node (current_cfa_node)
                                 (current_rr)
                                 (current_sps)
                                 (current_incoming_label)
                                 (temp_art)
                                 (false)
        end)
     with 
     | La_exception.Not_point_to (alabel, null_expr1, null_expr2, need_addrof_exprs) -> 
          begin
(*             let changed_flatten_path : art_path =  *)
(*                 change_index_of_variables_in_path (!current_path) (!gvars)  *)
(*                   (true) (true) (false) (ref []) in *)
            (* START-ENCODE *)
            (* Encode path *)
(*             let (encoded_edges_in_path, encoded_hash) =  *)
(*                 encode_vars_in_path (changed_flatten_path.art_edges_in_path) in *)
(*             let encoded_flatten_path : art_path = { *)
(*                 art_nodes_in_path = changed_flatten_path.art_nodes_in_path; *)
(*                 art_edges_in_path = encoded_edges_in_path; *)
(*             } in *)
            (* Encode null_expr2 & need_addrof_exprs *)
(*             let encoded_null_expr2 = *)
(*                 encode_vars_in_expression (null_expr2) (encoded_hash) in *)
            (*let encoded_addr_exprs = encode_vars_in_expressions (need_addrof_exprs) (encoded_hash) in*)
            (* STOP-ENCODE *)
            (*let sliced_error_path : art_path = 
                slice_path_backward_from_label_with_criterion_vars 
                  (encoded_flatten_path) ([encoded_null_expr2] (*@ encoded_addr_exprs*)) 
                  (alabel)
            in*)
            (*debug "trace-path: ";*)
            (*debug (get_string_of_cutting_apath_2 sliced_error_path !current_path);*)
            debug "functions in the path: ";
            let cut_prime_path : art_path = cut_path (!current_path) (alabel) in
            debug (find_all_func_names_in_path cut_prime_path);
            error "-------------------------------------------------";
            let closest_expr = 
                get_closest_expr_of_null_expr null_expr1 (!current_path) in
            error ("In " ^ (string_of_art_label alabel) ^ " : " ^
                          (string_of_expression closest_expr) ^ "  --- (" ^
                          (string_of_expression null_expr2) ^ ") --- " ^
                          " points to nowhere");
            (* return *)
            let enode =
                find_anode_in_path (cut_prime_path.art_nodes_in_path) (alabel)
            in
            (match enode with 
            | Some enode ->
                if (not !exist_feasible_iter_max_paths) then  
                    la_error :=
                        {
                          ctx_error_path = copy_art_path cut_prime_path;
                          ctx_error_location = 
                              (try List.hd enode.art_memo
                              with Failure _ -> assert false);
                          ctx_error_message =
                              (string_of_expression
                                  (La_common.create_based_vars_in_expression null_expr1))
                              ^ " points to nowhere"
                        }
                else ();
            | None -> assert false);
            (false, false, None, [], (*[],*) (("", -1), []), (None, [temp_art]))
          end
        | La_exception.Weak_constraints pred_list -> 
          begin
            emerg ("Weak-constraints: ");
            List.iter
                (fun p -> 
                    if (normalize p) != True then
                        emerg (string_of_predicate p);
                )
                pred_list;
            if (not !exist_feasible_iter_max_paths) then
                la_error := {
                              ctx_error_path = copy_art_path !current_path;
                              ctx_error_location = 
                                  location_of_error_assert !current_path;
                              ctx_error_message =
                                  "LogicProver can not prove this path";
                            }
            else ();
            (false, false, None, [], (("", -1), []), (None, [temp_art]))
          end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_art_children
      (ctx : Checker.check_context_t)
      (outgoings : (cfa_node * predicate) list) 
      (aroot : art_node)
      (current_rr : predicate) 
      (cfa_call_stack : (int * cfa_node * cfa_fundef_t) list)
      (temp_cfa_fundef : cfa_fundef_t) 
      (temp_art : art) 
: (string * int)                (* pivot_label *)
  * (((string * int) * predicate) list) (* new_cnodes_and_sps *)
  * art                        (*current_temp_art *)
  (* ((string * int) list)       dependent_labels *)
  * (art list)                 (*error_art_list *)
  * ((string * int) list)       (* child_labels *)
  * bool                        (* repeating_flag *)
  * bool (* flag denotes if the traverse is trust or not *) = 
begin
    (* Step1: define some variable *)
    let repeating_flag : bool ref = ref true in
    let new_cnodes_and_sps = ref [] in
    let pivot_label : (string * int) ref = ref ("", - 1) in
    let current_outgoings = ref outgoings in
    let (root_name, root_version) = aroot.art_label in
    let copy_temp_art : (art ref) = ref temp_art in
    let current_temp_art : (art ref) = ref temp_art in
    let error_art_list : (art list ref) = ref [] in
    (*let dependent_labels : (string * int) list ref = ref [] in*)
    let child_labels : (string * int) list ref = ref [] in
    (*let length_of_outgoings = List.length outgoings in*)
    let aroot_is_trust : bool ref = ref true in
        
    (* step2: create art_children of temp_anode *)
    while (!repeating_flag = true) & (!current_outgoings != [])do
        (* copy some vars, maybe need when refine *)
        let copy_of_number_of_function_call = !current_number_of_fc in
        copy_temp_art := copy_art (!current_temp_art);
        (* get an outgoing *)
        let ((next_cfa_node, edge), tail_of_tuples) =  
            match !current_outgoings with
            | a_outgoing :: tail_of_tuples -> (a_outgoing, tail_of_tuples) 
            | _ -> assert false
        in
        current_outgoings := tail_of_tuples;
        (* update information about aroot & edge into path *)
        !current_path.art_nodes_in_path <- !current_path.art_nodes_in_path @ 
                                           [aroot];
        !current_path.art_edges_in_path <- !current_path.art_edges_in_path @ 
                                           [edge];
        let new_art_label = create_art_label (!current_art) (!traced_labels) (!current_path)
              (next_cfa_node) in
        (* compute sps & rr for child: *)
        let sps_of_child : predicate list =
            try Hashtbl.find_all (!sps_hash) new_art_label
            with | Not_found -> [True]
        in
        let rr_of_child : predicate =
            compute_rr_of_art_child (aroot.art_type) (current_rr) (edge) 
              (sps_of_child) 
        in
        (* create art_child_node for next_cfa_node *)      
        let (flag, is_trust, new_child, new_child_labels, (*new_dependent_labels,*) 
        (new_pivot_label, updated_sps_in_path) ,(new_temp_art, new_error_arts)) =
            create_art_node (ctx) (next_cfa_node) (rr_of_child) (sps_of_child)
              (aroot.art_label) (cfa_call_stack) (!current_temp_art)
        in
        child_labels := !child_labels @ new_child_labels;
        (* check new_error_arts:______________________________________ *)
        if (new_error_arts != [])
           & (!la_options.Checker.psopt_dot_output = true) then begin
            let cp_aroot = copy_art_node (aroot) in
            (match new_child with
            | Some a_new_child -> begin
                cp_aroot.art_outgoing_labels <- 
                    cp_aroot.art_outgoing_labels @ [a_new_child.art_label];
                cp_aroot.art_outgoing_edges <- 
                    cp_aroot.art_outgoing_edges @ [edge];
              end
            | _ -> assert false
            );
            List.iter
                (fun a_temp_art ->
                    Hashtbl.add (a_temp_art.art_node_hash) (cp_aroot.art_label) (cp_aroot)
                )
                new_error_arts;
            error_art_list := !error_art_list @ new_error_arts;
        end;
        (* check result after create art_child_node for next_cfa_node: *)
        (if flag = true then begin
            (* check is_trust to update informations for aroot:  *)
            if is_trust = false then aroot_is_trust := false;
            current_temp_art := (match new_temp_art with
                                | Some t -> t
                                | None -> assert false);
            (*dependent_labels := !dependent_labels @ new_dependent_labels;*)
            (* update child for aroot: *)
            (match new_child with
            | Some a_new_child ->
                aroot.art_outgoing_labels <- aroot.art_outgoing_labels @ 
                                             [a_new_child.art_label];
                aroot.art_outgoing_edges <- aroot.art_outgoing_edges @ 
                                            [edge];
            | _ -> assert false);
        end else begin
            pivot_label := new_pivot_label;
            (*let l2 = List.length !current_outgoings in  
            if (length_of_outgoings = 2) & (l2 = 1) then
                insert_to_number_of_fc_hash (aroot.art_label) 
                  (!current_number_of_fc) (!number_of_fc_hash));*)
            (* check if aroot is pivot_node or not: *)
            if aroot.art_label = new_pivot_label then begin
                aroot.art_is_pivot_node <- true;
                current_outgoings := (next_cfa_node, edge) :: tail_of_tuples;
                update_cnodes_and_sps (updated_sps_in_path) 
                  (current_label_and_itps_tuples) (!sps_hash);
                current_temp_art := !copy_temp_art;
                current_number_of_fc := copy_of_number_of_function_call;
                List.iter (remove_a_version_in_art !current_art)
                          (!child_labels);
                List.iter (fun child_label ->
		               (* Remove from art_node_hash *)
			       try (
                                   let _ = Hashtbl.find
                                               !current_art.art_node_hash
                                               child_label
                                   in
				   Hashtbl.remove !current_art.art_node_hash
                                                  child_label)
                               with Not_found -> ())
                          !child_labels;
            end else begin
                repeating_flag := false;
                new_cnodes_and_sps := updated_sps_in_path;
            end
        end);
        (* temporarily remove aroot *) 
        !current_path.art_nodes_in_path <-
            List.rev (List.tl (List.rev !current_path.art_nodes_in_path));
        !current_path.art_edges_in_path <-
            List.rev (List.tl (List.rev !current_path.art_edges_in_path));
    done; (* stop while*)
    (* recover some vars *)
    current_cfa_fundef := temp_cfa_fundef;
    let old_traversed_time = 
        try Hashtbl.find (!hash_to_count_cfa_nodes) (root_name)
        with
        | Not_found -> 0
    in
    if old_traversed_time > 0 then begin
        Hashtbl.replace (!hash_to_count_cfa_nodes) (root_name) 
          (old_traversed_time - 1); 
    end;
    (* return *)
    (!pivot_label, !new_cnodes_and_sps, !current_temp_art, 
    (*!dependent_labels,*) !error_art_list, !child_labels, 
    !repeating_flag, !aroot_is_trust)
end
      
      
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)           
(*and update_traverse_time (current_outgoing_tuples)
                         (current_traversed_time : int) 
                         (current_cfa_node : cfa_node) 
: unit = 
begin
    (* debug *)
    (*debug ("update_traverse_time: " );*)
    if (List.length current_outgoing_tuples) = 2 then begin
        if current_traversed_time = 0 then begin
            Hashtbl.replace (!hash_to_count_cfa_nodes) 
                        (current_cfa_node.cfa_label) (1) 
        end else begin
            Hashtbl.replace (!hash_to_count_cfa_nodes) 
                            (current_cfa_node.cfa_label) 
                            (current_traversed_time + 1);
        end 
    end;
end*)      
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_art_node_root_of_subtree
      (ctx : check_context_t)
      (current_cfa_node : cfa_node)
      (current_rr : predicate)
      (current_sps : predicate list) 
      (current_incoming_label : string * int)
      (cfa_call_stack : (int * cfa_node * cfa_fundef_t) list)
      (temp_art : art)
      (current_traversed_time : int) 
: bool (* flag denotes the traverse to create art is successful or not *) 
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *) 
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node 
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *)  =
begin
    (* Step1: create art_node_root *)
    let aroot : art_node =
        create_temp_art_node (current_cfa_node) (current_rr) (current_sps) 
          (current_incoming_label) (!current_art) (!current_path)
          (!traced_labels) (false)
    in
    let (new_art_name, new_art_version) = aroot.art_label in
    Hashtbl.add (!current_art.art_version_hash)
                (new_art_name)
                (new_art_version, current_rr);

    let create_new : bool ref = ref true in
    let result = ref (true, true, None, [], (("", -1), []), (Some temp_art, [])) in
    (* Check in art_created_nodes: *)
    (try (
        let created_root : art_node = 
            Hashtbl.find !current_art.art_trust_nodes aroot.art_label in
        let is_eq : int = 
            is_equivalent_predicates (created_root.art_reachable_region) (current_rr) in
        if is_eq = 0 then begin
            create_new := false;
            (* return *)
            result := (true, created_root.art_trust, Some created_root, 
                       [aroot.art_label], (("", -1), []), (Some temp_art, []));
        end
    ) with Not_found -> ());
    if !create_new = false then !result
    else begin 
        (* Step3: update current_number_of_fc *)        
        if aroot.art_type = ART_CALL then begin
            (*current_number_of_fc := !current_number_of_fc + 1;*)
            current_number_of_fc := 
                find_in_number_of_fc_hash (!number_of_fc_hash) (!current_path)
                  (!current_number_of_fc);
        end;
        (* check if function pointer point to nowhere then return UNSAFE *)
        let init_func_pointer, pointer_name =
            check_init_func_pointer !current_path current_cfa_node
                                    ctx.Checker.chk_afile
                                    (Cfa.CFA_FUNDEF !current_cfa_fundef)
        in
        if init_func_pointer = false then (
            if (not !exist_feasible_iter_max_paths) then
                la_error := {
                          ctx_error_path = copy_art_path !current_path;
                          ctx_error_location = 
                              List.hd current_cfa_node.Cfa.cfa_memo;
                          ctx_error_message =
                              "Uninitialized function pointer : " ^ pointer_name;
                    }
            else ();
            (false, true, Some aroot, [], (("", -1), []), (Some temp_art, [])))
        else (
            (* Step4: get outgoing_tupples of current_cfa_node: *)
            let (outgoing_tuples, new_cfa_call_stack, new_cfa_fundef) =
                get_next_nodes_and_edges_of_cfa_node (ctx.Checker.chk_afile)
                                                     (current_cfa_node)
                                                     (!current_cfa_fundef)
                                                     (cfa_call_stack)
                                                     (!current_number_of_fc)
                                                     (current_rr)
            in
            (* Step5: save new_cfa to current_cfa_fundef to use when return *)
            let temp_cfa_fundef = !current_cfa_fundef in
            current_cfa_fundef := new_cfa_fundef;
            
            (* Step6: Update traverse time for cfa-condition-node *)
            update_traverse_time (outgoing_tuples) (current_traversed_time) 
               (current_cfa_node); 
            
            (* Step7: create art_children *)
            let (pivot_label, new_cnodes_and_sps, current_temp_art(*, dependent_labels,*) 
                , error_art_list, child_labels, repeating_flag, is_trust) = 
                create_art_children (ctx) (outgoing_tuples) (aroot) (current_rr) 
                  (new_cfa_call_stack) (temp_cfa_fundef) (temp_art)
            in

            (* Revert traverse time for cfa-condition-node *)
            revert_traverse_time (outgoing_tuples) (current_traversed_time)
              (current_cfa_node);
        
            (* Step8: check result after create_art_children & return *)
            if repeating_flag = false then begin
                (*remove_a_version_in_art (!current_art) (aroot.art_label) 
                  !current_path.art_nodes_in_path;*)
                (* return *)
	        let new_child_labels = aroot.art_label :: child_labels in
                (false, false, Some aroot, new_child_labels, (*[],*) (pivot_label, new_cnodes_and_sps),
                (None (*Some !current_temp_art*), error_art_list))
            end else begin
                (*aroot.art_dependent_labels <- dependent_labels @ 
                                               aroot.art_dependent_labels;*)
                (* Update information "is_trust" for aroot *)
                aroot.art_trust <- is_trust;
                if !la_options.Checker.psopt_dot_output = true then begin
                    Hashtbl.add (current_temp_art.art_node_hash) (aroot.art_label) 
                      (aroot);
                end;
                (* insert information about temp_art_node into current_art *)
(*                 let (new_art_name, new_art_version) = aroot.art_label in *)
(*                 Hashtbl.add (!current_art.art_node_hash) (aroot.art_label) (aroot); *)
                if is_trust = true then begin
                    Hashtbl.add (!current_art.art_trust_nodes)
                                (aroot.art_label)
                                (aroot)
                end;
                (* debug *)
                (*debug "repeating_flag = true";
                debug ("outgoing_tuples : " ^ (string_of_int (List.length outgoing_tuples)));
                debug ("update to art_version_hash in create_art_node_root_of_subtree");
                debug ("name: " ^ new_art_name);
                debug ("rr: " ^ (string_of_predicate current_rr));
                let _ = read_int () in*)
                let new_child_labels = aroot.art_label :: child_labels in
                (* return *)
                (true, aroot.art_trust, Some aroot, new_child_labels, (*aroot.art_dependent_labels,*) 
                (("", -1), []), (Some current_temp_art, error_art_list))
            end;
            )
    end;
end  
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ create_art_node ~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and  create_art_node
      (ctx : check_context_t)
      (current_cfa_node : cfa_node) 
      (current_rr : predicate)
      (current_sps : predicate list) 
      (current_incoming_label : string * int)
      (cfa_call_stack : (int * cfa_node * cfa_fundef_t) list)
      (temp_art : art)
: bool (* flag denotes the traverse to create art is successful or not *) 
  * bool (* flag denotes if the traverse is trust or not *)
  * art_node option (* Some root-of-subtree if the traverse is successful *) 
  * ((string * int) list) (* children_node list of current_node *)
  (* ((string * int) list) dependent_labels which current_node depends on *)
  * ((string * int) * ((string * int) * predicate) list)  (* label of pivot_node 
  and updated support_predicates ~ name of node in path *)
  * (art option * art list) (*Some new_temp_art if we traverse successfully
  , None otherwise) * (error_art_list) *)  =
begin
    (* Step1: check if current_cfa_node is an error_node or not *)
    match current_cfa_node.cfa_node_stmt with
    | CFA_ERROR -> (* current_cfa_node is an error_node *)
        create_art_node_when_reachs_error_label (current_cfa_node) 
          (current_rr) (current_sps) (current_incoming_label) (temp_art)
          
    | anything_else -> (* current_cfa_node is not an error_node *)
        let current_cfa_label = current_cfa_node.cfa_label in       
        let traversed_time : int = 
            try Hashtbl.find (!hash_to_count_cfa_nodes) (current_cfa_label)
            with
            | Not_found -> 0
        in
        let covering_node_label : (string * int) =
            covered_by (!current_art) (current_rr) (current_cfa_label) 
        in
        (* Step2: create stop_node or subtree *)
        if (Art.compare_art_labels covering_node_label ("", -1)) != 0 then begin
            let covered_by_trust =
                try let _ = Hashtbl.find !current_art.art_trust_nodes
                                         covering_node_label
                    in true
                with Not_found -> false
            in
            create_art_stop_node (current_cfa_node)
                                 (current_rr) 
                                 (current_sps)
                                 (current_incoming_label)
                                 (temp_art)
                                 (covered_by_trust)
        end
        else if (normalize current_rr) = False then
            begin
                create_art_stop_node (current_cfa_node)
                                     (current_rr) 
                                     (current_sps)
                                     (current_incoming_label)
                                     (temp_art)
                                     (false)
            end
            else if (traversed_time >= !la_options.Checker.psopt_iter_max) then
                begin
                    create_art_node_when_reachs_iter_max (current_cfa_node)
                                                         (current_rr)
                                                         (current_sps)
                                                         (current_incoming_label)
                                                         (temp_art)
                end
                else begin
                    (match current_cfa_node.cfa_node_stmt with
                     | CFA_CALL ([func], _)
                         when ((find_name_and_level_of_expr_func func)
                               = (Ccommon.raise_itermax_err_fn,
                                  Cexpr.global_scope_level)) -> (
                         let ssa_path =
                             change_index_of_variables_in_path !current_path
                                                               !gvars
                                                               true
                                                               false
                                                               false
                                                               (ref [])
                         in
                         let pred_of_path =
                             normalize (
                                 And (ssa_path.art_edges_in_path @
                                      [current_rr])) in
                         if (is_sat pred_of_path) then (
                             let memo = try List.hd current_cfa_node.cfa_memo
                                        with Not_found -> ("", "", "", "", "")
                             in
                             let itermax_context : la_error_context_t =
                               {
                                 ctx_error_path = copy_art_path !current_path;
                                 ctx_error_location = memo;
                                 ctx_error_message = "";
                               }
                             in
                             itermax_context_list := !itermax_context_list
                                                     @ [itermax_context];)
                         else ()
                      )
                    | _ -> ());
                    create_art_node_root_of_subtree (ctx)
                                                        (current_cfa_node)
                                                        (current_rr)
                                                        (current_sps)
                                                        (current_incoming_label)
                                                        (cfa_call_stack)
                                                        (temp_art)
                                                        (traversed_time)
		    end
end



(* ~~~~~~~~~~~~~~~~~~~~ initialize_global_variables ~~~~~~~~~~~~~~~~~~~~~~ *)
let initialize_global_variables (main_func_entry : Sexpr.function_entry_t)
      (options : Checker.ps_options_t)
      (new_gvars : expression list) 
: unit =
begin
    let main_func_label = main_func_entry.Sexpr.fe_funcname in
    la_options := options; 
    current_art := create_new_art (1) (main_func_label, 1);
    current_path := create_new_art_path ();
    current_cfa_fundef := create_nil_cfa_fundef ();
    sps_hash := Hashtbl.create (0);
    gvars := new_gvars;  
    current_number_of_fc := 0;
    number_of_fc_hash := Hashtbl.create (0);
    traced_labels := Hashtbl.create (0);
    hash_to_count_cfa_nodes := Hashtbl.create (0);
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_error_loc (tab : string)
                        (error_context : la_error_context_t) : string = (
    let (filename, line1, pos1, line2, pos2) = error_context.ctx_error_location
    in
    let tab1 = tab ^ "          " in
    let er_file =
        if (String.compare filename "") = 0 then
            ""
        else
            tab1 ^ "File: " ^ filename ^ "\n"
    in
    let er_line1 =
        if (String.compare line1 "") = 0 then
            ""
        else
            if (String.compare line2 "") <> 0 then
                tab1 ^ "Start line: " ^ line1 ^ "\n"
            else
                tab1 ^ "Line: " ^ line1
    in
    let er_pos1 =
        if (String.compare pos1 "") = 0 then
            ""
        else
            tab1 ^ "Start position: " ^ pos1 ^ "\n"
    in
    let er_line2 =
        if (String.compare line2 "") = 0 then
            ""
        else
        tab1 ^ "End line: " ^ line2 ^ "\n"
    in
    let er_pos2 =
        if (String.compare pos2 "") = 0 then
            ""
        else
            tab1 ^ "End position: " ^ pos2 in
    let string_of_error_loc = er_file ^ er_line1 ^ er_pos1 ^ er_line2 ^ er_pos2
    in
    let call_stack_data =
        string_of_call_stack error_context.ctx_error_path
    in
    (tab ^ "+ Error location: \n" ^ string_of_error_loc ^ "\n"
    ^ tab ^ "+ Call stack: \n"
    ^ tab ^ call_stack_data)
)

(* Print Result & call stack of Error Path *)
let print_checking_result (is_safe : bool)
                          (error_context: la_error_context_t)
                          (consumed_time : float)
                          : unit=
begin
    let complexity_of_checking =
        StatisticUtil.string_of_la_complexity () in
    let tab = "    " in
        let double_tab = tab ^ tab
        and triple_tab = tab ^ tab ^ tab
        in
        let i = ref 0 in
        let string_of_itermax_locs =
            List.fold_left
                (fun string_of_itermax_locs error_context ->
                     (incr i);
                     (string_of_itermax_locs
                      ^ double_tab ^ "+ (" ^ (string_of_int !i) ^ ") : \n"
                      ^ (string_of_error_loc triple_tab error_context)))
                ""
                (!itermax_context_list)
        in
    if (is_safe) then
    begin
        let safe_property =
            if (!itermax_context_list <> []) then
                "UNSAFE"
            else
                "SAFE"
        in
        let message = (" =========RESULT==========\n"
                       ^ "    + Program is " ^ safe_property ^ "!\n"
                       ^ "    + Loop(s) traversed more than iter-max times:\n"
                       ^ string_of_itermax_locs
                       ^ "    + Time consumed by the lazy abstraction: "
                       ^ (string_of_float consumed_time) ^ " s";
        ) in
        info (message ^ "\n" ^ complexity_of_checking);
    end
    else
    begin
        let string_of_error_loc = string_of_error_loc tab error_context in
        let message = (" =========RESULT==========\n"
                      ^ "    + Program is UNSAFE!\n"
                      ^ "    + Error message: " ^ error_context.ctx_error_message ^ "\n"
                      ^ string_of_error_loc
                      ^ "    + Loop(s) traversed more than iter-max times:\n"
                      ^ string_of_itermax_locs
                      ^ "    + Time consumed by the lazy abstraction: "
                      ^ (string_of_float consumed_time) ^ " s")
        in
        error (message ^ "\n" ^ complexity_of_checking);
    end;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_art_from_cfas
      (ctx : check_context_t)
: bool * (art(*option*) list) * art_path =
begin
    start_time := Unix.time ();
    (* preprocess cfas *)
    let new_gvars = preprocess_cfas (ctx.Checker.chk_afile) in
    (* debug *)
    (*debug ("AFTER PREPROCESS:------------------------------------");
    debug ("number of cfas: " ^ (string_of_int (Hashtbl.length !cfas_hash)));
    Hashtbl.iter
        (fun (name, level) a_cfa -> 
            debug (name ^ (string_of_int level));
            debug (string_of_cfa a_cfa);
        )
        !cfas_hash;*)
    (* init global variables of lazyabstraction.ml *) 
    initialize_global_variables (ctx.Checker.chk_afile.Cfa.afile_start)
                                (ctx.Checker.chk_options)
                                (new_gvars); 
    (* find cfa of main_func_label *)
    let main_cfa =         
        try Checker.find_cfa_of_function_entry
                ctx
                ctx.Checker.chk_afile.Cfa.afile_start
        with
        | Not_found -> (
            error ("Entry function `"
                   ^ (Sexpr.string_of_function_entry
                          ctx.Checker.chk_afile.Cfa.afile_start)
                   ^ "' not found!"); 
            assert false)
    in
    let main_cfa_with_assumptions =
        create_main_cfa_with_assumptions main_cfa
    in
    (match main_cfa_with_assumptions with
    | CFA_FUNDEF f -> current_cfa_fundef := f;
    | _ -> assert false);
    (* find root of main_cfa *)    
    let cfa_root : cfa_node = (
        try Hashtbl.find (!current_cfa_fundef.cfa_node_hashtbl)
                         (!current_cfa_fundef.cfa_entry_label)
        with
        | Not_found -> assert false
            (*raise (ConfigUtil.Message (Logging.Error, "entry is not found"))*)
    ) in
    (* start creating art: *)
    let temp_art = create_new_art (10) ("temp_art", 1) in
    let (checking_successed, is_trust, _, _, _, (*_,*) (current_temp_art_option, error_temp_arts)) =
        create_art_node (ctx)
                        (cfa_root)
                        (True)
                        ([True])
                        ("where_call_main", 1) 
                        ([])
                        (temp_art)
    in
    (*determine whether the result is based on iter-max processing or not*)
    let is_successed = (
        (* if reach an error node then program unsafe, unmark the iter-max flag *)
        if (not checking_successed && (not !is_duplicate)) then
        begin
            unsafe_because_iter_max := !exist_feasible_iter_max_paths;
            false;
        end
        (* if reach iter-max and the path reach iter-max is feasible then program is unsafe *)
        else if (!check_iter_max && !exist_feasible_iter_max_paths) then
        begin
            unsafe_because_iter_max := true;
            false;
        end
        (* otherwises program is safe*)
        else
            true;
    ) in
    (* Print run-time *)
    let stop_time : float = Unix.time() in
    (* print checking result *)
    let consumed_time = stop_time -. !start_time in
    print_checking_result is_successed !la_error consumed_time;
    (* Return *)
    (is_successed, !current_art :: error_temp_arts, rr_in_epath !la_error.ctx_error_path !gvars)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Entry point for lazy abstraction *)
let do_lazy_check (ctx : Checker.check_context_t)
                  : unit = 
  (begin
    (* Check for reachability of an error node in the main function *)
    Preprocess.check_reachability_of_error_node ctx;
    (* Dump cfas to dot file *)
    if ctx.chk_options.Checker.psopt_dot_output then 
        Cfa_dumper.dump_cfas ctx.Checker.chk_afile;
    (* Create art *)
    let ((is_successful : bool), arts, (error_path : Art.art_path)) =  
        create_art_from_cfas (ctx)
    in
    (* Dump result to xml file *)
    (Art.save_result_as_xml_file ctx.chk_options.Checker.psopt_xml_location 
                                 is_successful 
                                 error_path
                                 !la_error.ctx_error_location
                                 !la_error.ctx_error_message);
    (* Dump result to dot file *)
    if ctx.chk_options.Checker.psopt_dot_output then 
        Art_dumper.dump_arts ~source:"global" arts;
  end)

(* exception return until checking *)
exception Exit_safe;;
exception Exit_unsafe;;
exception Exit_pointer_null;;
exception Exit_weak_constraint;;
exception Exit_safe_by_duplicating;;
exception Exit_unsafe_because_itermax;;

(** Process when rasing exception La_exception.Not_point_to
    @param alabel : art_label
    @param null_expr1 : expression
    @param null_expr2 : expression
    @param need_addrof_exprs : expression list
*)
let process_exception_pointer_null (alabel : string * int)
                                   (null_expr1 : expression)
                                   (null_expr2 : expression)
                                   (need_addrof_exprs : expression list)
                                   (art_path : Art.art_path)
: unit = (
    debug "functions in the path: ";
    let cut_prime_path : art_path = cut_path (art_path) (alabel) in
    debug (find_all_func_names_in_path cut_prime_path);
    error "-------------------------------------------------";
    let closest_expr = 
        get_closest_expr_of_null_expr null_expr1 (art_path) in
    error ("In " ^ (string_of_art_label alabel) ^ " : " ^
        (string_of_expression closest_expr) ^ "  --- (" ^
        (string_of_expression null_expr2) ^ ") --- " ^
        " points to nowhere");
    (* return *)
    let enode_op =
        find_anode_in_path (cut_prime_path.art_nodes_in_path) (alabel) in
    (match enode_op with 
    | Some enode ->
        if (not !exist_feasible_iter_max_paths) then  
            la_error :=
              {
                ctx_error_path = copy_art_path cut_prime_path;
                ctx_error_location = 
                    (try List.hd enode.art_memo
                    with Failure _ -> assert false);
                ctx_error_message =
                    (string_of_expression
                        (La_common.create_based_vars_in_expression null_expr1))
                        ^ " points to nowhere"
              }
          else ()
    | None -> assert false);
)

let process_exception_weak_constraints (pred_list : predicate list)
                                       (art_path : Art.art_path)
: unit = (
    emerg ("Weak-constraints: ");
    List.iter (fun p -> 
        if (normalize p) != True then
            emerg (string_of_predicate p);
    ) pred_list;
    if (not !exist_feasible_iter_max_paths) then
        la_error := { ctx_error_path = copy_art_path art_path;
                      ctx_error_location = 
                          location_of_error_assert art_path;
                      ctx_error_message =
                          "LogicProver can not prove this path";
                    }
    else ();
)

let analysis_path (art_path : Art.art_path)
                  (gvars : expression list)
                  (psopt_slice_path : int)
: (string * int) * (((string * int) * predicate) list) = (
    if psopt_slice_path = 1 then
        analysis_path_normally art_path gvars
    else analysis_path_with_WeiBenbacher_paradigm art_path gvars
)

(** refine a path in art and update reachable region to art
    @param art : art
    @param error_path : art_path to error node in art
*)
let refine_path (art : Art.art) (error_path : Art.art_path)
: (string * int) * ((string * int) * predicate) list = (
    (* Try to change index of vars in path, then analysis the path *)
    try (
        (* find pivot node and interpolants *)
        let (pivot_label, labels_and_interpolants) =
            (if !fc_pointer_flag = true then (
                let (temp_pivot_label, temp_labels_and_interpolants) = 
                    analysis_function_pointers_in_path
                        (error_path)
                        (!gvars)
                        (!la_options.Checker.psopt_cutlist)
                in
                let fp_duplicated : bool = 
                    is_duplicated_labels_and_interpolants_tuples 
                      (temp_labels_and_interpolants) (current_label_and_itps_tuples)
                in
                if (fp_duplicated = true) || 
                    ((compare_art_labels temp_pivot_label ("", -1)) == 0) then
                    analysis_path error_path
                                  !gvars
                                  !la_options.Checker.psopt_slice_path
                else (temp_pivot_label, temp_labels_and_interpolants))
            else (
                analysis_path error_path
                              !gvars
                              !la_options.Checker.psopt_slice_path
            ))
            in
            (* 27.2.2009. Xuan has added to deal with duplicate *)
            (* check if inputs duplicate or not: *)
            let is_duplicated : bool = 
                is_duplicated_labels_and_interpolants_tuples 
                  (labels_and_interpolants) (current_label_and_itps_tuples)
            in
            (* if interpolants is duplicate then stop checking
               and announce that the program is safe *)
            if is_duplicated = true then begin
                emerg "interpolations is duplicated";
                is_duplicate := true;
                la_error :=
                  {
                    ctx_error_path = copy_art_path error_path;
                    ctx_error_location = 
                        location_of_error_assert error_path;
                    ctx_error_message = "Exit checking as safe by duplicating"
                  };
                raise Exit_safe_by_duplicating;
            (* 27.2.2009. end *)
            end else begin
                debug "functions in the path: ";
                debug (find_all_func_names_in_path error_path);
                (* 20090710. Print feasible path --END *)
                (* if pivot node don't exist in path then the program is unsafe *)
                if (compare_art_labels pivot_label ("", -1)) = 0
                    & (not !exist_feasible_iter_max_paths) then begin
                    la_error :=
                        {
                          ctx_error_path = copy_art_path error_path;
                          ctx_error_location =
                              location_of_error_assert error_path;
                          ctx_error_message = ""
                        };
                    raise Exit_unsafe;
                end;
                (pivot_label, labels_and_interpolants)
            end
        (* Catch error Not_point_to raised from changing index of vars in path *)
        ) with 
        | La_exception.Not_point_to (alabel, null_expr1, null_expr2, need_addrof_exprs) -> ( 
            process_exception_pointer_null alabel
                                           null_expr1
                                           null_expr2
                                           need_addrof_exprs
                                           error_path;
            
            raise Exit_pointer_null;)
        | La_exception.Weak_constraints pred_list -> (
            process_exception_weak_constraints pred_list error_path;
            raise Exit_weak_constraint;)
)

(** get art_path reachs to a node having label is art_label
    @param art : art
    @param art_label : current node which art_path reachs to
    @return art_path reach to art_label
*)
let get_art_path (art : Art.art) (art_label : string * int)
: Art.art_path = (
    (* find art_path reach to node art_label except it *)
    let rec get_art_path_to (art : Art.art) (art_node : Art.art_node)
    : (Art.art_node list) * (predicate list) = (
        try
            let father_of_art_node_label = art_node.Art.art_incoming_label in
            let father_of_art_node = Hashtbl.find art.Art.art_node_hash
                                                  father_of_art_node_label in
            let edge = Art.find_edge_in_art father_of_art_node
                                            art_node.Art.art_label
            in
            let art_nodes, edges = get_art_path_to art father_of_art_node in
            (art_nodes @ [father_of_art_node], edges @ [edge])
        with Not_found -> ([], [])
    )in
    let art_node = try Hashtbl.find art.Art.art_node_hash art_label
                   with Not_found -> assert false
    in
    let art_nodes_to, edges_to = get_art_path_to art art_node in
    let art_nodes = art_nodes_to @ [art_node]
    and edges = edges_to @ [True]
    in
    { art_nodes_in_path = art_nodes;
      art_edges_in_path = edges;
    }
)

(** Compute rr from pivot node to error node
    If art_node is covered then stop
*)
let rec update_reachable_region (temp_art : Art.art)
                                (labels_and_interpolants : ((string * int)
                                                            * predicate) list)
: unit = (
    match labels_and_interpolants with
    | [] -> ()
    | (current_art_node_label, sp) :: remain_labels_and_interpolants -> (
        let current_art_node =
            try Hashtbl.find temp_art.Art.art_node_hash current_art_node_label
            with Not_found -> assert false
        in
        (* update support *)
        let father_label = current_art_node.Art.art_incoming_label in
        current_art_node.Art.art_support_predicates
            <- current_art_node.Art.art_support_predicates @ [sp];
        Hashtbl.replace temp_art.Art.art_node_hash
                        current_art_node_label
                        current_art_node;
        let typ_of_father, rr_of_father, edge =
            try
                let father_node =
                    Hashtbl.find temp_art.Art.art_node_hash father_label in
                (father_node.Art.art_type,
                 father_node.Art.art_reachable_region,
                 Art.find_edge_in_art father_node current_art_node_label)
            with Not_found -> (Art.ART_FORM, True, True)
        in
        (* if rr_of_father & edge ==> sp is valid then rr = sp
           ele if rr_of_father & edge ==> !sp is valid then rr = !sp
           else rr = True *)
        let temp_rr = And [rr_of_father; edge] in
        let pred1 = Or [Not temp_rr; sp]
        and pred2 = Or [Not temp_rr; Not sp]
        in
        let rr =
            compute_rr_of_art_child typ_of_father
                                    rr_of_father
                                    edge
                                    [sp]
        in
        (* update rr to node in art *)
        current_art_node.Art.art_reachable_region <- rr;
        current_art_node.Art.art_old_reachable_regions <-
            current_art_node.Art.art_old_reachable_regions @ [rr];
        (* check cover *)
        let covering_label = Art.covered_by1 temp_art rr current_art_node_label in
        if (is_sat rr) = false then (
            Hashtbl.replace temp_art.Art.art_node_hash
                            current_art_node_label
                            current_art_node;
            Art.remove_childrens_of_node temp_art current_art_node;            
        )
        else if (Art.compare_art_labels covering_label ("", -1)) = 0 then (
            Hashtbl.replace temp_art.Art.art_node_hash
                            current_art_node_label
                            current_art_node;
            update_reachable_region temp_art remain_labels_and_interpolants
        )
        else (
            (* update node as covered node *)
            current_art_node.Art.art_is_cover_node <- true;
            current_art_node.Art.art_covered_by <- covering_label;
            Hashtbl.replace temp_art.Art.art_node_hash
                            current_art_node_label
                            current_art_node;
            (* remove childs of current_node *)
            Art.remove_childrens_of_node temp_art current_art_node;
        )
    )
 )

(** After find interpolations.
    Compute reachable region of node from pivot_node in error path
    @param temp_art :  art
    @param pivot_node : pivot_node
    @param labels_and_interpolants : art_node_label and it's interpolation
*)
let update_reachable_region (temp_art : Art.art)
                            (pivot_node : string * int)
                            (labels_and_interpolants : ((string * int) * predicate) list)
: unit = (
    (* detain interpolants from pivot node to error node *)
    let _, detained_interpolants =
        List.fold_left (fun (b, remained_interpolants) (label_and_interpolant) ->
            let current_label = fst label_and_interpolant in
            let is_pivot_node = Art.compare_art_labels current_label pivot_node
            in
            let b = b or (is_pivot_node = 1) in
            if b then (b, remained_interpolants @ [label_and_interpolant])
            else (b, remained_interpolants)
        ) (false, []) labels_and_interpolants
    in
    update_reachable_region temp_art detained_interpolants;
)

let write_dot_file = ref false
(** find and refine path reach to node error having label is error_label
    update reachable region of nodes in path to temp_art
    @param temp_art : art is being constructed
    @param error_label : label of error art node
*)
let find_and_refine_path (temp_art : Art.art) (error_label : string * int)
: unit = (
    StatisticUtil.update_refinement_number ();
    (* get art_path reachs to error node from temp_art *)
    let art_path = get_art_path temp_art error_label in
    StatisticUtil.update_number_of_preds_in_paths
        (LogicUtil.Logic.count_atom_pred_in_preds art_path.art_edges_in_path);
    let art_nodes_in_path =
        List.map (fun art_node ->
            art_node.Art.art_label
        ) art_path.Art.art_nodes_in_path in
    temp_art.Art.art_labels_in_error_path <- art_nodes_in_path; 
    if !write_dot_file then
        Art_dumper.dump_arts ~source:"global" [temp_art];
    (* find pivot node and interpolants *)
    let pivot_node, labels_and_interpolants = refine_path temp_art art_path
    in
    (* update reachable regions *)
    update_reachable_region temp_art
                            pivot_node
                            (List.rev labels_and_interpolants);
    if !write_dot_file then
        Art_dumper.dump_arts ~source:"global" [temp_art];
)

(** Find branch count of a node label.
    It is equal to number of nodes in path reach to this node
    @param art: art which is being constructed
    @param reviewing_node_label: node is being reviewing in find path
    @param finding_label: cfa label of node which is being found
    @return number of nodes that having cfa label is finding_label
*)
let rec count_branch (art : Art.art)
                     (reviewing_node_label : string * int)
                     (finding_label : string)
: int = (
    try
        let reviewing_node =
            Hashtbl.find art.Art.art_node_hash reviewing_node_label
        in
        let father_of_reviewing_node = reviewing_node.Art.art_incoming_label in
        let branch_count_of_father =
            (count_branch art
                          father_of_reviewing_node
                          finding_label)
        in 
        if (String.compare finding_label (fst reviewing_node_label)) = 0 then
            branch_count_of_father + 1
        else
            branch_count_of_father
    with Not_found -> 0
)

(** Create art_node from cfa_node and add to after current_art_node in
    art which is being constructing *)
let add_new_node_to_art (temp_art : Art.art)
                        (current_art_node_label : string * int)
                        (outgoing : Cfa.cfa_node * LogicUtil.Logic.predicate)
                        (cfa_fundef : Cfa.cfa_fundef_t)
                        (call_stack : (int * cfa_node * cfa_fundef_t) list)
: unit = (
    StatisticUtil.update_node_explored_number ();
    (* Get index of art_node, if cfa_node was added to temp_art then return
       maximum index of it plus 1 else return 0.
       Then, update hash maximun_indexes *)
    let cfa_node, edge = outgoing in
    let cfa_node_label = cfa_node.Cfa.cfa_label in
    let art_node_index =
        try (Hashtbl.find temp_art.Art.art_max_indexes
                          cfa_node_label) + 1
        with Not_found -> 0
    in
    let new_outgoing_label = (cfa_node_label, art_node_index) in
    Hashtbl.replace temp_art.Art.art_max_indexes cfa_node_label art_node_index;
    (* create new art_node *)
    let new_art_node =
        create_new_art_node (transform_cfa_stmt_to_art_stmt_type cfa_node) 
                            new_outgoing_label
                            []
                            []
                            current_art_node_label
                            True
                            []
                            []
                            false
                            false
                            false
                            cfa_node.Cfa.cfa_memo
    in
    (* add art_nodes to art *)
    Hashtbl.add temp_art.Art.art_node_hash
                new_outgoing_label
                new_art_node;
    (* add outgoing_label and outgoing_edge for current_art_node whithout
       art_label_of_root
    *)
    if (Art.compare_art_labels current_art_node_label
                               temp_art.Art.art_label_of_root) = 0 then ()
    else (
        let current_art_node =
            try Hashtbl.find temp_art.Art.art_node_hash current_art_node_label
            with Not_found -> assert false
        in
        current_art_node.Art.art_outgoing_labels <-
            current_art_node.Art.art_outgoing_labels @ [new_outgoing_label];
        current_art_node.Art.art_outgoing_edges <-
            current_art_node.Art.art_outgoing_edges @ [edge];
        Hashtbl.replace temp_art.Art.art_node_hash
                        current_art_node_label
                        current_art_node;
    );
    (match cfa_node.Cfa.cfa_node_stmt with
    (* if added node is error node then refine path *)
    | CFA_ERROR -> (
        find_and_refine_path temp_art new_outgoing_label)
    (* if added node is function call to __raise_itermax then update itermax_context_list *)
    | CFA_CALL ([func], _)
      when ((find_name_and_level_of_expr_func func)
                = (Ccommon.raise_itermax_err_fn,
                   Cexpr.global_scope_level)) -> (
        let path = get_art_path temp_art new_outgoing_label in
        let ssa_path =
            change_index_of_variables_in_path path
                                              !gvars
                                              true
                                              false
                                              false
                                              (ref [])
        in
        let pred_of_path =
            normalize (And (ssa_path.art_edges_in_path)) in
        if (is_sat pred_of_path) then (
            let memo = try List.hd cfa_node.cfa_memo
                       with Not_found -> ("", "", "", "", "")
            in
            let itermax_context : la_error_context_t =
              {
                ctx_error_path = copy_art_path path;
                ctx_error_location = memo;
                ctx_error_message = "";
              }
            in
            itermax_context_list := !itermax_context_list
                                    @ [itermax_context];)
        else ()
      )
    | _ -> (    (* update art_queue *)
        let branch_count = count_branch temp_art
                                        new_outgoing_label
                                        (fst new_outgoing_label) in
        temp_art.Art.art_queue <- temp_art.Art.art_queue
                                  @ [(new_outgoing_label, branch_count, cfa_fundef, call_stack)];
      )
    );
)

(** Create art_node from cfa_node and add to after current_art_node
    in art which is being constructingnew_art_queue
    @param temp_art : art is being constructing
    @param current_art_node_label : label of art_node that is being continued
    @param outgoings : children of current_art_node
    @param cfa_fundef : cfa_fundef containing outgoings
*)
let add_new_nodes_to_art (temp_art : Art.art)
                         (current_art_node_label : string * int)
                         (outgoings : (Cfa.cfa_node * LogicUtil.Logic.predicate) list)
                         (cfa_fundef : Cfa.cfa_fundef_t)
                         (call_stack : (int * cfa_node * cfa_fundef_t) list)
: unit = (
    List.iter (fun outgoing ->
        (* if current_art_node is removed then don't add outgoing to art *)
        if (Hashtbl.mem temp_art.Art.art_node_hash
                        current_art_node_label) = false then
            ()
        else
            add_new_node_to_art temp_art
                                current_art_node_label
                                outgoing
                                cfa_fundef
                                call_stack
    ) outgoings;
)

let art_index = ref 1

let get_real_cfa_label (label : string) : string = (
    try (fst (ExtString.String.split label "*"))
    with _ -> label
)

let check_recurring_func (call_stack : (int * cfa_node * cfa_fundef_t) list)
                         (iter_max : int)
: bool = (
    let rec check_recurring_func (call_stack : (int * cfa_node * cfa_fundef_t) list)
                                 (iter_max : int)
                                 (func_name : string)
                                 (func_count : int) = (
        if func_count >= iter_max then true
        else (
            match call_stack with
            | [] -> false
            | (_, _, cfa_fundef) :: remain_call_stack -> (
                let new_func_name =
                    get_real_cfa_label
                        (Cfa.get_id_of_cfa (Cfa.CFA_FUNDEF cfa_fundef)) in
                let func_count =
                    if (String.compare func_name new_func_name) = 0 then
                        func_count + 1
                    else func_count
                in
                check_recurring_func remain_call_stack iter_max func_name func_count
              )
        )
    ) in
    match call_stack with
    | [] -> false
    | (_, _, cfa_fundef) :: remain_call_stack -> (
        let func_name =
            get_real_cfa_label
                (Cfa.get_id_of_cfa (Cfa.CFA_FUNDEF cfa_fundef)) in
        check_recurring_func remain_call_stack iter_max func_name 1
      ) 
)

let check_itermax (temp_art: Art.art) : unit = (
    let nodes_reach_to_itermax =
        List.fold_left (fun nodes_reach_to_itermax
                           (art_node_label, branch_count, _, call_stack) ->
            if branch_count >= !la_options.Checker.psopt_iter_max
               or (check_recurring_func call_stack
                                        !la_options.Checker.psopt_recursive_max) then
                nodes_reach_to_itermax @ [art_node_label]
            else nodes_reach_to_itermax
        ) [] temp_art.Art.art_queue
    in
    List.iter (fun art_node_label ->
      try
   	    let art_path =
            get_art_path temp_art art_node_label in
        let pivot_label, labels_and_interpolants = refine_path temp_art art_path
        in
        let art_node_op = Art.find_anode_in_path art_path. Art.art_nodes_in_path
                                              art_node_label
        in
        match art_node_op with
        | None -> ()
        | Some art_node -> (
            let is_feasible = (compare_art_labels pivot_label ("", -1) = 0) in
            if is_feasible then (
                let memo = try List.hd art_node.art_memo
                           with Not_found -> ("", "", "", "", "")
                in
                let itermax_context : la_error_context_t =
                  {
                    ctx_error_path = copy_art_path art_path;
                    ctx_error_location = memo;
                    ctx_error_message = "";
                  }
                in
                itermax_context_list := !itermax_context_list
                                        @ [itermax_context];)
            else ())
      with Exit_unsafe -> (
        itermax_context_list := !itermax_context_list @ [!la_error])
    ) nodes_reach_to_itermax;
    if !itermax_context_list <> [] then
        raise Exit_unsafe_because_itermax
    else
        raise Exit_safe
)

(** Find continued_art_node which has branch_count is minimum in art
    Add nodes can be reached from continued_art_node
    @author NhungPT
    @param temp_art : art is being constructing
    @param ctx : contain afile and all options
 *)
let rec create_next_level_of_art (temp_art : Art.art)
                                 (ctx : check_context_t)
: unit = (
    (* find node have branch_count which is the least *)
    let (continued_art_node_label_opt, _, cfa_fundef_opt, call_stack) =
        List.fold_left (fun (art_node_label, min_branch_count, cfa_fundef_opt, call_stack)
                            (label, branch_count, cfa_fundef, a_call_stack) ->
            if (branch_count < !la_options.Checker.psopt_iter_max)
               & (check_recurring_func a_call_stack
                                       !la_options.Checker.psopt_recursive_max) = false
               & (branch_count < min_branch_count or art_node_label = None) then
                (Some label, branch_count, Some cfa_fundef, Some a_call_stack)
            else (art_node_label, min_branch_count, cfa_fundef_opt, call_stack)
        ) (None, 0, None, None) (List.rev temp_art.art_queue) in
    match continued_art_node_label_opt, cfa_fundef_opt, call_stack with
    (* if queue is empty then stop and announce safe *)
    | None, None, None ->
        check_itermax temp_art
    (* else find next nodes from continued_art_node and insert to art *)
    | Some continued_art_node_label, Some cfa_fundef, Some call_stack -> (
        (* Remove continued_art_node from art_queue *)
        temp_art.Art.art_queue <-
            List.filter (fun (label, _, _, _) ->
                (Art.compare_art_labels label continued_art_node_label) <> 0
            ) temp_art.Art.art_queue;
        let current_cfa_node =
            try Hashtbl.find cfa_fundef.cfa_node_hashtbl
                             (fst continued_art_node_label)
            with Not_found -> assert false
        in
        let path_to_continued_node =
            get_art_path temp_art continued_art_node_label
        in
        current_number_of_fc :=
            find_in_number_of_fc_hash (!number_of_fc_hash)
                                      (path_to_continued_node)
                                      (!current_number_of_fc);
        let pred_of_path = And (path_to_continued_node.Art.art_edges_in_path)
        in
        (* find next nodes from continued_art_node *)
        let outgoings, new_call_stack, new_cfa_fundef =
            get_next_nodes_and_edges_of_cfa_node ctx.chk_afile
                                                 current_cfa_node
                                                 cfa_fundef
                                                 call_stack
                                                 !current_number_of_fc 
                                                 pred_of_path
        in
        let outgoings = List.rev outgoings in
        (* check init func pointer*)
        List.iter (fun (cfa_node, pred) ->
            path_to_continued_node.Art.art_edges_in_path <- 
                List.rev (pred :: (List.tl (List.rev path_to_continued_node.Art.art_edges_in_path)));                      
            let init_func_pointer, pointer_name =
                check_init_func_pointer path_to_continued_node
                                        cfa_node
                                        ctx.chk_afile
                                        (Cfa.CFA_FUNDEF cfa_fundef)
            in
            if init_func_pointer = false then (
                la_error := {
                          ctx_error_path = copy_art_path !current_path;
                          ctx_error_location = 
                              List.hd cfa_node.Cfa.cfa_memo;
                          ctx_error_message =
                              "Uninitialized function pointer : " ^ pointer_name;
                    };
                raise Exit_unsafe)
            else ();
        ) outgoings;
        add_new_nodes_to_art temp_art
                             continued_art_node_label
                             outgoings
                             new_cfa_fundef
                             new_call_stack;
        create_next_level_of_art temp_art ctx;
      )
    | _ -> assert false
)

let create_art_from_cfas1 (ctx : check_context_t)
: unit = (
    write_dot_file := ctx.chk_options.Checker.psopt_dot_output;
    (* preprocess cfas *)
    let new_gvars = preprocess_cfas (ctx.Checker.chk_afile) in
    (* init global variables of lazyabstraction.ml *)

    initialize_global_variables (ctx.Checker.chk_afile.Cfa.afile_start)
                                (ctx.Checker.chk_options)
                                (new_gvars); 
    (* find cfa of main_func_label *)
    let main_cfa =         
        try Checker.find_cfa_of_function_entry
                ctx
                ctx.Checker.chk_afile.Cfa.afile_start
        with
        | Not_found -> (
            error ("Entry function `"
                   ^ (Sexpr.string_of_function_entry
                          ctx.Checker.chk_afile.Cfa.afile_start)
                   ^ "' not found!"); 
            assert false)
    in
    let main_cfa_with_assumptions =
        create_main_cfa_with_assumptions main_cfa
    in
    (match main_cfa_with_assumptions with
    | CFA_FUNDEF f -> current_cfa_fundef := f;
    | _ -> assert false);
    (* start creating art: *)
    let temp_art = create_new_art (10) ("temp_art", 1) in
    let current_cfa_node =
        try Hashtbl.find !current_cfa_fundef.Cfa.cfa_node_hashtbl
                         !current_cfa_fundef.cfa_entry_label
        with Not_found -> assert false
    in
    add_new_node_to_art temp_art
                        ("temp_art",1)
                        (current_cfa_node, True)
                        (!current_cfa_fundef)
                        [];
    create_next_level_of_art temp_art ctx;
)

let do_lazy_check1 (ctx : Checker.check_context_t)
                  : unit = 
  (begin
    (* Check for reachability of an error node in the main function *)
    Preprocess.check_reachability_of_error_node ctx;
    (* Dump cfas to dot file *)
    if ctx.chk_options.Checker.psopt_dot_output then 
        Cfa_dumper.dump_cfas ctx.Checker.chk_afile;
    (* Create art *)
    let is_successful = ref false in
    let error_path = Art.create_new_art_path () in
(*    let ((is_successful : bool), arts, (error_path : Art.art_path)) =  *)
    start_time := Unix.time ();
    (try
        create_art_from_cfas1 (ctx)
    with
    | Exit_safe
    | Exit_safe_by_duplicating ->
        is_successful := true
    | Exit_unsafe
    | Exit_pointer_null
    | Exit_weak_constraint
    | Exit_unsafe_because_itermax -> ()
    );
    let consumed_time = (Unix.time ()) -. !start_time in
    print_checking_result !is_successful
                          !la_error
                          consumed_time;
    if (!itermax_context_list <> []) then
        is_successful := false
    else ();
(*    in*)
    (* Dump result to xml file *)
    (Art.save_result_as_xml_file ctx.chk_options.Checker.psopt_xml_location
                                 !is_successful
                                 error_path
                                 !la_error.ctx_error_location
                                 !la_error.ctx_error_message);
(*    (* Dump result to dot file *)                                           *)
(*    if ctx.chk_options.Checker.psopt_dot_output then                        *)
(*        Art_dumper.dump_arts ~source:"global" arts;                         *)
  end)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let do_lazy_check (ctx : Checker.check_context_t)
                  : unit = (
    match ctx.Checker.chk_options.psopt_strategy with
    | Checker.DFS -> do_lazy_check ctx
    | Checker.HRBC -> do_lazy_check1 ctx
)
