(* parser/form.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Var/Term/Form -- Basic Data for SMT *)

(* NOTE: (1) This is a common data representation shared by an
   interpreter and an SMT solver.  (2) This assumes equality (equality
   with uninterpreted functions), and variables and terms can be
   equated. *)

(* MEMO: terms and forms are not cyclic (and can be compared by
   {compare}). *)

open Num

(* Disallows equality. *)

let (=) = (Misc.(=))
let (<>) = (Misc.(<>))

let string_equal = Misc.string_equal

module StringSet = Set.Make (
  struct
  type t = string
  let compare x y = (compare x y)
  end)

(* Variable and Proposition Names. *)

type varname_t = string

module VarSet = Set.Make (
  struct
  type t = varname_t
  let compare x y = (compare x y)
  end)

(* Linear Arithmetic Terms: Terms are pairs of variables and their
   coefficients.  It is used only in linear constraint solver.
   Constant part is stored under the name {linear_term_constant}. *)

module LTermMap = Map.Make (
  struct
  type t = varname_t
  let compare x y = (compare x y)
  end)

type linear_term_t = num LTermMap.t
let linear_term_constant = "#1"

(* Uninterpreted function terms consist of {VAR}, {FUN}, {UIFUN}.
   NOTE: {FUN} appears only in the head of {UIFUN} terms.  {FUN} is
   included in terms to make list-matching easy.  Arithmetic (linear)
   terms consist of {VAR}, {LTERM}, and {ARITH}.  {LTERM}, {ZERO}, and
   {CVT} only appear in processing SMT. *)

type term_t =
    VAR of varname_t
    | FUN of string
    | UIFUN of term_t list
    | ARITH of arith_term_t
    | LTERM of linear_term_t

and arith_term_t =
    NUM of num
    | MUL of term_t * term_t
    | ADD of term_t * term_t
    | NEG of term_t
    | ZERO
    | CVT of num * varname_t

type relop_t =
    EQ | NE | GE | GT | LE | LT | DIVS

type form_t =
    TT | FF
    | PROP of varname_t
    | PRED of relop_t * term_t * term_t
	(* Connectives *)
    | NOT of form_t
    | AND of form_t list
    | OR of form_t list
    | IMP of form_t * form_t
    | IFF of form_t * form_t
	(* Quantifiers *)
    | FORALL of varname_t * form_t
    | EXISTS of varname_t * form_t

(* Compares two forms.  Forms be identical to be equal (ie, not
   commutativity etc).  It is used in {FormSet}.  Note that default
   structural equality may not work, because term includes a {Map}
   with {num} which needs an explicitly-specified proper equality
   test. *)

let rec form_compare (x : form_t) (y : form_t) : int = (
  if (x == y) then
    0
  else
    match (x, y) with
      (TT, TT) ->
	0
    | (FF, FF) ->
	0
    | ((PROP x0), (PROP y0)) ->
	(compare x0 y0)
    | ((PRED (xop, x0, x1)), (PRED (yop, y0, y1))) ->
	let v = (compare xop yop) in
	if (v != 0) then
	  v
	else
	  (Misc.list_map_compare term_compare [x0 ; x1] [y0 ; y1])
    | ((NOT x0), (NOT y0)) ->
	(form_compare x0 y0)
    | ((AND xl), (AND yl)) ->
	(Misc.list_map_compare form_compare xl yl)
    | ((OR xl), (OR yl)) ->
	(Misc.list_map_compare form_compare xl yl)
    | ((IMP (x0, x1)), (IMP (y0, y1))) ->
	(Misc.list_map_compare form_compare [x0 ; x1] [y0 ; y1])
    | ((IFF (x0, x1)), (IFF (y0, y1))) ->
	(Misc.list_map_compare form_compare [x0 ; x1] [y0 ; y1])
    | ((FORALL (xv, x0)), (FORALL (yv, y0))) ->
	let v = (compare xv yv) in
	if (v != 0) then
	  v
	else
	  (form_compare x0 y0)
    | ((EXISTS (xv, x0)), (EXISTS (yv, y0))) ->
	let v = (compare xv yv) in
	if (v != 0) then
	  v
	else
	  (form_compare x0 y0)
    | _ ->
	(* Case of different kind of forms. *)
	let xx = (form_variant_key x) in
	let yy = (form_variant_key y) in
	assert ((xx - yy) != 0) ;
	(xx - yy))

and term_compare (x : term_t) (y : term_t) : int = (
  if (x == y) then
    0
  else
    match (x, y) with
      ((VAR xv), (VAR yv)) ->
	(compare xv yv)
    | ((FUN xf), (FUN yf)) ->
	(compare xf yf)
    | ((UIFUN xl), (UIFUN yl)) ->
	(Misc.list_map_compare term_compare xl yl)
    | ((ARITH xt), (ARITH yt)) ->
	(arith_term_compare xt yt)
    | ((LTERM xt), (LTERM yt)) ->
	(linear_term_compare xt yt)
    | _ ->
	(* Case of different kind of terms. *)
	let xx = (term_variation_key x) in
	let yy = (term_variation_key y) in
	assert ((xx - yy) != 0) ;
	(xx - yy))

and arith_term_compare (x : arith_term_t) (y : arith_term_t) : int = (
  if (x == y) then
    0
  else
    match (x, y) with
      ((NUM xn), (NUM yn)) ->
	(compare_num xn yn)
    | ((MUL (x0, x1)), (MUL (y0, y1))) ->
	(Misc.list_map_compare term_compare [x0 ; x1] [y0 ; y1])
    | ((ADD (x0, x1)), (ADD (y0, y1))) ->
	(Misc.list_map_compare term_compare [x0 ; x1] [y0 ; y1])
    | ((NEG x0), (NEG y0)) ->
	(term_compare x0 y0)
    | (ZERO, ZERO) ->
	0
    | ((CVT (xn, xv)), (CVT (yn, yv))) ->
	let v = (compare_num xn yn) in
	if (v != 0) then
	  v
	else
	  (compare xv yv)
    | _ ->
	(* Case of different kind of terms. *)
	let xx = (arith_term_variation_key x) in
	let yy = (arith_term_variation_key y) in
	assert ((xx - yy) != 0) ;
	(xx - yy))

and linear_term_compare (x : linear_term_t) (y : linear_term_t) : int = (
  (LTermMap.compare compare_num x y))

and term_variation_key (x : term_t) : int = (
  match x with
    (VAR _) ->
      1
  | (FUN _) ->
      2
  | (UIFUN _) ->
      3
  | (ARITH _) ->
      4
  | (LTERM _) ->
      5)

and arith_term_variation_key (x : arith_term_t) : int = (
  match x with
    (NUM _) ->
      1
  | (MUL _) ->
      2
  | (ADD _) ->
      3
  | (NEG _) ->
      4
  | ZERO ->
      5
  | (CVT _) ->
      6)

and form_variant_key (x : form_t) : int = (
  match x with
      TT ->
	1
    | FF ->
	2
    | (PROP _) ->
	3
    | (PRED _) ->
	4
    | (NOT _) ->
	5
    | (AND _) ->
	6
    | (OR _) ->
	7
    | (IMP _) ->
	8
    | (IFF _) ->
	9
    | (FORALL _) ->
	10
    | (EXISTS _) ->
	11)

let form_equal (x : form_t) (y : form_t) : bool = ((form_compare x y) == 0)

let term_equal (x : term_t) (y : term_t) : bool = ((term_compare x y) == 0)

module LTermSet = Set.Make (
  struct
  type t = linear_term_t
  let compare x y = (linear_term_compare x y)
  end)

module FormSet = Set.Make (
  struct
  type t = form_t
  let compare x y = (form_compare x y)
  end)

module FormSetSet = Set.Make (
  struct
  type t = FormSet.t
  let compare x y = (FormSet.compare x y)
  end)

(* Cube is a conjunction of literals. *)

type cube_t = CUBE of FormSet.t

let cube_equal (x : cube_t) (y : cube_t) : bool = (
  match (x, y) with
    ((CUBE xs), (CUBE ys)) ->
      (FormSet.equal xs ys))

let cube_compare (x : cube_t) (y : cube_t) : int = (
  match (x, y) with
    ((CUBE xs), (CUBE ys)) ->
      (FormSet.compare xs ys))

module CubeSet = Set.Make (
  struct
  type t = cube_t
  let compare x y = (cube_compare x y)
  end)

(* Similar to Hashtbl.length or Set.S.cardinal. *)

let linear_term_length (e : linear_term_t) : int = (
  (LTermMap.fold (fun _ _ count -> (count + 1)) e 0))

let rec form_to_string (f : form_t) : string = (
  match f with
    TT -> "#t"
  | FF -> "#f"
  | PROP v ->
      ("P[" ^ (varname_to_string v) ^ "]")
  | PRED (p, t1, t2) ->
      ("(" ^ (relop_to_string p) ^ " "
	^ (String.concat " " (List.map term_to_string [t1 ; t2])) ^ ")")
  | NOT f1 ->
      ("(!" ^ " " ^ (form_to_string f1) ^ ")")
  | AND l ->
      if (l == []) then
	("(&&)")
      else
	("(&& " ^ (String.concat " " (List.map form_to_string l)) ^ ")")
  | OR l ->
      if (l == []) then
	("(||)")
      else
	("(|| " ^ (String.concat " " (List.map form_to_string l)) ^ ")")
  | IMP (f1, f2) ->
      ("(" ^ (form_to_string f1) ^ " -> " ^ (form_to_string f2) ^ ")")
  | IFF (f1, f2) ->
      ("(" ^ (form_to_string f1) ^ " <-> " ^ (form_to_string f2) ^ ")")
  | FORALL (v, f1) ->
      ("(#forall[" ^ (varname_to_string v) ^ "] "
	^ (form_to_string f1) ^ ")")
  | EXISTS (v, f1) ->
      ("(#exists[" ^ (varname_to_string v) ^ "] "
	^ (form_to_string f1) ^ ")"))

and term_to_string (t : term_t) : string = (
  match t with
    LTERM e ->
      if ((linear_term_length e) == 0) then
	"(0)"
      else
	("(" ^
	    (String.concat "+"
		(LTermMap.fold
		    (fun k v r ->
		      if (string_equal k linear_term_constant) then
			((string_of_num v) :: r)
		      else
			(((string_of_num v) ^ "*" ^ k) :: r))
		    e []))
	  ^ ")")
  | VAR v ->
      (varname_to_string v)
  | FUN s ->
      s
  | UIFUN l ->
      ("(" ^ (String.concat " " (List.map term_to_string l)) ^ ")")
  | ARITH ZERO ->
      "0"
  | ARITH NUM i ->
      (string_of_num i)
  | ARITH CVT (i, v) ->
      ("(" ^ (string_of_num i) ^ "*" ^ (varname_to_string v) ^ ")")
  | ARITH MUL (t1, t2) ->
      ("(" ^ (term_to_string t1) ^ "*" ^ (term_to_string t2) ^ ")")
  | ARITH ADD (t1, t2) ->
      ("(" ^ (term_to_string t1) ^ "+" ^ (term_to_string t2) ^ ")")
  | ARITH NEG t1 ->
      ("(- " ^ (term_to_string t1) ^ ")"))

and relop_to_string p = (
  match p with
    EQ -> "=="
  | NE -> "!="
  | GE -> ">="
  | GT -> ">"
  | LE -> "<="
  | LT -> "<"
  | DIVS -> "divides")

and varname_to_string v = v

let cube_to_literal_list (cube : cube_t) : form_t list = (
  match cube with
    (CUBE s) ->
      (FormSet.elements s))

let cube_to_form (cube : cube_t) : form_t = (
  (AND (cube_to_literal_list cube)))

(* NOTE: IT DOES NOT CHECK IF FORMS ARE LITERALS. *)

let form_list_to_string (l : form_t list) : string = (
  ("[" ^ (String.concat ";" (List.map form_to_string l)) ^ "]"))

let term_list_to_string (l : term_t list) : string = (
  ("[" ^ (String.concat ";" (List.map term_to_string l)) ^ "]"))

let cube_to_string (cube : cube_t) : string = (
  match cube with
    (CUBE s) ->
      if (FormSet.is_empty s) then
	(*"(&&)"*) ("#t")
      else
	("(&& " ^ (String.concat " "
		      (List.map form_to_string (FormSet.elements s)))
	  ^ ")"))

let cube_list_to_string (l : cube_t list) : string = (
  ("[" ^ (String.concat ";" (List.map cube_to_string l)) ^ "]"))

let cube_set_to_string (s : CubeSet.t) : string = (
  (cube_list_to_string (CubeSet.elements s)))

(* ================================================================ *)

(* Makes a new variable. *)

let xgensym : string -> string = (
  let counter = ref 10000 in
  (fun (prefix : string) ->
    (incr counter) ;
    assert (not (string_equal prefix "")) ;
    ("#" ^ prefix ^ (string_of_int !counter))))

(* Coerces a num (big_int) to an integer; It may raise {Failure} with
   {int_of_big_int}. *)

let coerce_to_int (i : num) : int = ((int_of_num i) :> int)

(* Makes a new linear term. *)

let term_zero : linear_term_t = (LTermMap.empty)

let term_remove (v : varname_t) (e : linear_term_t) : linear_term_t = (
  (LTermMap.remove v e))

let term_coeff (v : varname_t) (e : linear_term_t) : num = (
  if (LTermMap.mem v e) then
    (LTermMap.find v e)
  else
    (num_of_int 0))

(* Returns a constant term. *)

let term_constant_part (e : linear_term_t) : num = (
  if (LTermMap.mem linear_term_constant e) then
    (LTermMap.find linear_term_constant e)
  else
    (num_of_int 0))

(* Adds or subtracts with regard to the third argument.  Note
   arguments to fold are swapped. *)

let term_add_or_sub (e1 : linear_term_t) (e2 : linear_term_t) (addp : bool) : linear_term_t = (
  (LTermMap.fold
      (fun k v oldmap ->
	if (LTermMap.mem k oldmap) then
	  (begin
	    let x = (if (addp) then v else (minus_num v)) in
	    let c = (add_num (LTermMap.find k oldmap) x) in
	    if (eq_num c (num_of_int 0)) then
	      (LTermMap.remove k oldmap)
	    else
	      (LTermMap.add k c oldmap)
	  end)
	else
	  (begin
	    let x = (if (addp) then v else (minus_num v)) in
	    assert (not (eq_num v (num_of_int 0))) ;
	    (LTermMap.add k x oldmap)
	  end))
      e2
      e1))

(* Adds linear terms. *)

let term_add (e1 : linear_term_t) (e2 : linear_term_t) : linear_term_t = (
  (term_add_or_sub e1 e2 true))

(* Subtracts linear terms. *)

let term_sub (e1 : linear_term_t) (e2 : linear_term_t) : linear_term_t = (
  (term_add_or_sub e1 e2 false))

(* Adds a constant to a linear term. *)

let term_addc_int (e : linear_term_t) (c : int) : linear_term_t = (
  let v = (term_constant_part e) in
  (LTermMap.add linear_term_constant (add_num v (num_of_int c)) e))

(* Multiplies a linear term by a constant. *)

let term_mulc (e : linear_term_t) (c : num) : linear_term_t = (
  if (eq_num c (num_of_int 0)) then
    LTermMap.empty
  else
    (LTermMap.fold
	(fun k v oldmap ->
	  (LTermMap.add k (mult_num c v) oldmap))
	e
	LTermMap.empty))

(* Negates a linear term. *)

let term_neg (e : linear_term_t) : linear_term_t = (
  (term_mulc e (num_of_int (-1))))

(* Adds a constant {m} to a coefficient to variable {k}. *)

let term_add_coeff (e : linear_term_t) (k : varname_t) (m : num) : linear_term_t = (
  if (LTermMap.mem k e) then
    (begin
      let c = (add_num (LTermMap.find k e) m) in
      if (eq_num c (num_of_int 0)) then
	(LTermMap.remove k e)
      else
	(LTermMap.add k c e)
    end)
  else
    (begin
      if (eq_num m (num_of_int 0)) then
	e
      else
	(LTermMap.add k m e)
    end))

(* Removes a term with zero coefficient (never happens). *)

let term_reduce (e : linear_term_t) : linear_term_t = (
  (LTermMap.fold
      (fun k v oldmap ->
	if (not (eq_num v (num_of_int 0))) then
	  (LTermMap.add k v oldmap)
	else
	  oldmap)
      e
      LTermMap.empty))

let term_constantp (e : linear_term_t) : bool = (
  let e2 = (term_reduce e) in
  match (linear_term_length e2) with
    0 -> true
  | 1 -> (LTermMap.mem linear_term_constant e2)
  | _ -> false)

let term_check_constant (e : linear_term_t) : num option = (
  let e2 = (term_reduce e) in
  match (linear_term_length e2) with
    0 ->
      (Some (num_of_int 0))
  | 1 ->
      if (LTermMap.mem linear_term_constant e2) then
	(Some (LTermMap.find linear_term_constant e2))
      else
	None
  | _ ->
      None)

(* ================================================================ *)

let varname_list_to_set (l : varname_t list) : StringSet.t = (
  (List.fold_left
      (fun s v -> (StringSet.add v s))
      (StringSet.empty) l))

let varname_set_to_list (s : StringSet.t) : varname_t list = (
  (StringSet.elements s))

let form_list_to_form_set (l : form_t list) : FormSet.t = (
  (Misc.list_fold FormSet.add l (FormSet.empty)))

let form_set_to_form_list (s : FormSet.t) : form_t list = (
  (FormSet.elements s))

let cube_true : cube_t = (CUBE (FormSet.empty))
let cube_false : cube_t = (CUBE (FormSet.singleton FF))

let cube_true_p (cube : cube_t) : bool = (cube_equal cube cube_true)
let cube_false_p (cube : cube_t) : bool = (cube_equal cube cube_false)

let cube_form_to_literal_list (cube : form_t) : form_t list = (
  match cube with
    TT ->
      (*TRIVIALCASE*) []
  | FF ->
      (*TRIVIALCASE*) [FF]
  | (AND l) -> l
  | _ -> (assert false))

let literal_list_to_cube (l : form_t list) : cube_t = (
  (CUBE (List.fold_left
	    (fun s v -> (FormSet.add v s))
	    (FormSet.empty) l)))

(* NOTE: IT DOES NOT CHECK IF FORMS ARE LITERALS. *)

let form_to_cube (cube : form_t) : cube_t = (
  match cube with
    TT ->
      (*TRIVIALCASE*) cube_true
  | FF ->
      (*TRIVIALCASE*) cube_false
  | (AND l) ->
      (literal_list_to_cube l)
  | _ -> (assert false))

let form_list_to_cube_list (l : form_t list) : cube_t list = (
  (List.map form_to_cube l))

let cube_list_to_cube_set (l : cube_t list) : CubeSet.t = (
  (Misc.list_fold CubeSet.add l (CubeSet.empty)))

let cube_set_to_cube_list (s : CubeSet.t) : cube_t list = (
  (CubeSet.elements s))

let cube_set_to_form_list (s : CubeSet.t) : form_t list = (
  (List.map cube_to_form (CubeSet.elements s)))

let cube_subset (cube0 : cube_t) (cube1 : cube_t) : bool = (
  match (cube0, cube1) with
    ((CUBE s0), (CUBE s1)) ->
      (FormSet.subset s0 s1))

let make_singleton_cube (f : form_t) : cube_t = (
  (CUBE (FormSet.singleton f)))

let singleton_cube_to_literal (cube : cube_t) : form_t = (
  match (cube_to_literal_list cube) with
    [l] ->
      l
  | _ -> (assert false))

let singleton_cube_p (cube : cube_t) : bool = (
  match (cube_to_literal_list cube) with
    [l] ->
      true
  | _ ->
      false)

(* NOTE: NO CHECK OF CUBE-NESS. *)

let cube_form_list_to_cube_set (l : form_t list) : CubeSet.t = (
  (List.fold_left
      (fun s c ->
	match c with
	  TT ->
	    (*AHO*)
	    (CubeSet.add cube_true s)
	| (AND _) ->
	    (CubeSet.add (form_to_cube c) s)
	| _ -> (assert false))
      (CubeSet.empty)
      l))

let cube_list_to_form (l : cube_t list) : form_t = (
  (OR (List.map cube_to_form l)))

(*SLOW*)

let cube_append (cube0 : cube_t) (cube1 : cube_t) : cube_t = (
  (literal_list_to_cube
      ((cube_to_literal_list cube0) @ (cube_to_literal_list cube1))))

let cube_add_literal (f : form_t) (cube : cube_t) : cube_t = (
  if (f == TT) then
    cube
  else if (f == FF) then
    cube_false
  else
    match cube with
      (CUBE (s : FormSet.t)) ->
	(CUBE (FormSet.add f s)))

(* Merges two lists of cubes, removing equal ones. *)

let cube_list_union (l0 : cube_t list) (l1 : cube_t list) : cube_t list = (
  let s0 =
    (List.fold_left
	(fun s cube -> (CubeSet.add cube s))
	(CubeSet.empty)
	l0) in
  let ss =
    (List.fold_left
	(fun s cube -> (CubeSet.add cube s))
	s0
	l1) in
  (cube_set_to_cube_list ss))

let uniqify_form_list (l : form_t list) : form_t list = (
  (form_set_to_form_list (form_list_to_form_set l)))

let uniqify_cube_list (l : cube_t list) : cube_t list = (
  (cube_set_to_cube_list (cube_list_to_cube_set l)))

(* ================================================================ *)

let form_atomp (f : form_t) = (
  match f with
    TT | FF | PROP _ | PRED _ -> true
  | _ -> false)

let form_literalp (f : form_t) = (
  match f with
    TT | FF | PROP _ | PRED _ -> true
  | NOT f1 -> (form_atomp f1)
  | _ -> false)

let form_clausep (f : form_t) = (
  match f with
    TT | FF | PROP _ | PRED _ | NOT _ -> (form_literalp f)
  | OR l ->
      (List.for_all form_literalp l)
  | _ -> false)

let form_in_cnf (f : form_t) = (
  match f with
    AND l -> (List.for_all form_clausep l)
  | _ -> false)

(* ================================================================ *)

(* FORM MANIPULATIONS (NON-OPTIMAL) *)

(* Pseudoly simplifies a form by logicals.  Similar one is in SMT. *)

let rec pseudo_simplify_form (f : form_t) = (
  match f with
    TT | FF ->
      f
  | (PROP _) ->
      f
  | (PRED _) ->
      f
  | (NOT f0) ->
      (pseudo_simplify_form_onelevel (NOT (pseudo_simplify_form f0)))
  | (AND l) ->
      (pseudo_simplify_form_onelevel (AND (List.map pseudo_simplify_form l)))
  | (OR l) ->
      (pseudo_simplify_form_onelevel (OR (List.map pseudo_simplify_form l)))
  | (IMP (f0, f1)) ->
      (IMP ((pseudo_simplify_form f0), (pseudo_simplify_form f1)))
  | (IFF (f0, f1)) ->
      (IFF ((pseudo_simplify_form f0), (pseudo_simplify_form f1)))
  | (FORALL (v0, f0)) ->
      (FORALL (v0, (pseudo_simplify_form f0)))
  | (EXISTS (v0, f0)) ->
      (EXISTS (v0, (pseudo_simplify_form f0))))

and pseudo_simplify_form_onelevel (f : form_t) = (
  match f with
    TT | FF ->
      f
  | (PROP _) -> (assert false)
  | (PRED _) -> (assert false)
  | (NOT (NOT f0)) ->
      f0
  | (NOT f1) ->
      (begin
	match f1 with
	  TT -> FF
	| FF -> TT
	| _ -> (NOT f1)
      end)
  | (AND l) ->
      (begin
	if (List.exists (fun x -> (x == FF)) l) then
	  FF
	else
	  (begin
	    let l2 = (List.filter (fun x -> (x != TT)) l) in
	    if (l2 == []) then
	      TT
	    else
	      (AND l2)
	  end)
      end)
  | (OR l) ->
      (begin
	if (List.exists (fun x -> (x == TT)) l) then
	  TT
	else
	  (begin
	    let l2 = (List.filter (fun x -> (x != FF)) l) in
	    if (l2 == []) then
	      FF
	    else
	      (OR l2)
	  end)
      end)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false))

let rec pseudo_conjoin (f0 : form_t) (f1 : form_t) = (
  match (f0, f1) with
    (TT, _) -> f1
  | (_, TT) -> f0
  | (FF, _) -> FF
  | (_, FF) -> FF
  | (AND l0, AND l1) ->
      (AND (l0 @ l1))
  | (AND l0, _) ->
      (AND (l0 @ [f1]))
  | (_, AND l1) ->
      (AND ([f0] @ l1))
  | _ ->
      (AND [f0 ; f1]))

let rec pseudo_conjoin_forms (l : form_t list) = (
  if (List.mem FF l) then
    FF
  else
    let l2 = (List.filter (fun x -> (x != TT)) l) in
    match l2 with
      [] ->
	TT
    | _ ->
	(AND l2))

let rec pseudo_disjoin (f0 : form_t) (f1 : form_t) = (
  match (f0, f1) with
    (TT, _) -> TT
  | (_, TT) -> TT
  | (FF, _) -> f1
  | (_, FF) -> f0
  | (OR l0, OR l1) ->
      (OR (l0 @ l1))
  | (OR l0, _) ->
      (OR (l0 @ [f1]))
  | (_, OR l1) ->
      (OR ([f0] @ l1))
  | _ ->
      (OR [f0 ; f1]))

let rec pseudo_disjoin_forms (l : form_t list) = (
  if (List.mem TT l) then
    TT
  else
    let l2 = (List.filter (fun x -> (x != FF)) l) in
    match l2 with
      [] ->
	FF
    | _ ->
	(OR l2))

(* Simplifies literal list as a cube (as a conjunction). *)

let rec pseudo_simplify_literal_list (l : form_t list) : form_t list = (
  match l with
    [] ->
      []
  | (hd :: tl) ->
      if (hd == FF) then
	[FF]
      else if (hd == TT) then
	(pseudo_simplify_literal_list tl)
      else
	(begin
	  let x = (pseudo_simplify_literal_list tl) in
	  match x with
	    [FF] ->
	      [FF]
	  | _ ->
	      (hd :: x)
	end))

let rec pseudo_simplify_cube (c : cube_t) : cube_t = (
  let l = (cube_to_literal_list c) in
  let k = (pseudo_simplify_literal_list l) in
  (literal_list_to_cube k))

(* Simplify a list of cubes, where list elements are disjuncts, and
   thus an empty list means false. *)

(* NOTE: IT UNIQIFIES CUBE ELEMENTS. *)
(*(form_set_to_list (form_list_to_form_set (pseudo_simplify_cube_list l)))*)

let rec pseudo_simplify_cube_list (l : cube_t list) : cube_t list = (
  match l with
    [] ->
      []
  | (hd :: tl) ->
      let hx = (pseudo_simplify_cube hd) in
      if (cube_true_p hx) then
	[cube_true]
      else if (cube_false_p hx) then
	(pseudo_simplify_cube_list tl)
      else
	(begin
	  let tx = (pseudo_simplify_cube_list tl) in
	  match tx with
	    [c] when (cube_true_p c) ->
	      [cube_true]
	  | _ ->
	      (hx :: tx)
	end))

(* ================================================================ *)

(* Substitutes variables in a predicate as specified by {varmap}, and
   records reverting substitution (a new predicate as a key) in
   {predmap}. *)

let rec substitute_variables_in_predicate (varmap : (varname_t, varname_t) Hashtbl.t) (predmap : (form_t, form_t) Hashtbl.t) (p : form_t) : form_t = (
  match p with
    TT | FF -> (assert false)
    | (PROP _) -> (assert false)
    | (PRED (op, t0, t1)) ->
	(begin
	  let x0 = (substitute_variables_in_term varmap t0) in
	  let x1 = (substitute_variables_in_term varmap t1) in
	  if (not (t0 == x0 && t1 == x1)) then
	    (begin
	      let newpred = (PRED (op, x0, x1)) in
	      (Hashtbl.replace predmap newpred p) ;
	      newpred
	    end)
	  else
	    p
	end)
    | NOT _ -> (assert false)
    | AND _ | OR _ -> (assert false)
    | IMP _ | IFF _ -> (assert false)
    | FORALL _ | EXISTS _ -> (assert false))

(* Substitutes variables in a cube as {varmap}.  It accepts {TT} as a
   special case (for trivial-sat). *)

and substitute_variables_in_cube (varmap : (varname_t, varname_t) Hashtbl.t) (cube : cube_t) : cube_t = (
  (form_to_cube (substitute_variables_in_form varmap (cube_to_form cube))))

(*AHO*) (* THIS IS THE SAME AS substitute_variables_in_cube_form. *)

and substitute_variables_in_form (varmap : (varname_t, varname_t) Hashtbl.t) (form : form_t) : form_t = (
  match form with
    TT | FF ->
      form
  | (PROP _) ->
      form
  | (PRED (op, t0, t1)) ->
      (begin
	let x0 = (substitute_variables_in_term varmap t0) in
	let x1 = (substitute_variables_in_term varmap t1) in
	if (not (t0 == x0 && t1 == x1)) then
	  (PRED (op, x0, x1))
	else
	  form
      end)
  | (NOT p) ->
      (NOT (substitute_variables_in_form varmap p))
  | (AND l) ->
      (AND (List.map (substitute_variables_in_form varmap) l))
  | (OR l) ->
      (OR (List.map (substitute_variables_in_form varmap) l))
  | IMP _ | IFF _ -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false))

(*GOMI*)(*
and substitute_variables_in_cube_form_ (varmap : (varname_t, varname_t) Hashtbl.t) (cube : form_t) : form_t = (
  match cube with
    TT | FF ->
      (*TRIVIALCASE*) cube
  | (PROP _) ->
      cube
  | (PRED (op, t0, t1)) ->
      (begin
	let x0 = (substitute_variables_in_term varmap t0) in
	let x1 = (substitute_variables_in_term varmap t1) in
	if (not (t0 == x0 && t1 == x1)) then
	  (PRED (op, x0, x1))
	else
	  cube
      end)
  | (NOT p) ->
      (NOT (substitute_variables_in_cube_form_ varmap p))
  | (AND l) ->
      (AND (List.map (substitute_variables_in_cube_form_ varmap) l))
  | (OR l) ->
      (OR (List.map (substitute_variables_in_cube_form_ varmap) l))
  | IMP _ | IFF _ -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false))
*)

and substitute_variables_in_term (varmap : (varname_t, varname_t) Hashtbl.t) (t : term_t) : term_t = (
  match t with
    (VAR s) ->
      if (Hashtbl.mem varmap s) then
	(VAR (Hashtbl.find varmap s))
      else
	t
  | (FUN s) ->
      t
  | (UIFUN l) ->
      (begin
	let x = (List.map (substitute_variables_in_term varmap) l) in
	if (List.exists2 (fun t0 t1 -> (not (t0 == t1))) l x) then
	  (UIFUN x)
	else
	  t
      end)
  | (ARITH a) ->
      (begin
	match a with
	  ZERO ->
	    t
	| (NUM n) ->
	    t
	| (CVT (n, v)) ->
	    if (Hashtbl.mem varmap v) then
	      (ARITH (CVT (n, (Hashtbl.find varmap v))))
	    else
	      t
	| (MUL (t0, t1)) ->
	    let x0 = (substitute_variables_in_term varmap t0) in
	    let x1 = (substitute_variables_in_term varmap t1) in
	    if (not (t0 == x0 && t1 == x1)) then
	      (ARITH (MUL (x0, x1)))
	    else
	      t
	| (ADD (t0, t1)) ->
	    let x0 = (substitute_variables_in_term varmap t0) in
	    let x1 = (substitute_variables_in_term varmap t1) in
	    if (not (t0 == x0 && t1 == x1)) then
	      (ARITH (ADD (x0, x1)))
	    else
	      t
	| (NEG t0) ->
	    let x0 = (substitute_variables_in_term varmap t0) in
	    if (not (t0 == x0)) then
	      (ARITH (NEG x0))
	    else
	      t
      end)
  | (LTERM l) ->
      let appears =
	(LTermMap.fold
	    (fun v n found -> (found || (Hashtbl.mem varmap v)))
	    l false) in
      if (appears) then
	(begin
	  let x =
	    (LTermMap.fold
		(fun v n oldmap ->
		  if (Hashtbl.mem varmap v) then
		    (LTermMap.add v n oldmap)
		  else
		    oldmap)
		l
		(LTermMap.empty)) in
	  (LTERM x)
	end)
      else
	t)

(* Substitutes predicates in a form as {predmap}.  NOTE: It is used to
   rename variables in predicates before/after state transition,
   because variables may have recurrence even after SSA
   transformation. *)

let rec substitute_predicates (predmap : (form_t, form_t) Hashtbl.t) (cube : cube_t) : cube_t = (
  (form_to_cube (substitute_predicates_in_form predmap (cube_to_form cube))))

and substitute_predicates_in_form (predmap : (form_t, form_t) Hashtbl.t) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      f
  | (PROP _) ->
      f
  | (PRED _) ->
      (begin
	if (Hashtbl.mem predmap f) then
	  (Hashtbl.find predmap f)
	else
	  f
      end)
  | (NOT g) ->
      (begin
	let x = (substitute_predicates_in_form predmap g) in
	if (not (x == g)) then
	  (NOT x)
	else
	  f
      end)
  | (AND l) ->
      (begin
	let x = (List.map (substitute_predicates_in_form predmap) l) in
	if (List.exists2 (fun s t -> (not (s == t))) l x) then
	  (AND x)
	else
	  f
      end)
  | (OR l) ->
      (begin
	let x = (List.map (substitute_predicates_in_form predmap) l) in
	if (List.exists2 (fun s t -> (not (s == t))) l x) then
	  (OR x)
	else
	  f
      end)
  | (IMP (f0, f1)) ->
      (begin
	let x0 = (substitute_predicates_in_form predmap f0) in
	let x1 = (substitute_predicates_in_form predmap f1) in
	if (not (f0 == x0 && f1 == x1)) then
	  (IMP (x0, x1))
	else
	  f
      end)
  | (IFF (f0, f1)) ->
      (begin
	let x0 = (substitute_predicates_in_form predmap f0) in
	let x1 = (substitute_predicates_in_form predmap f1) in
	if (not (f0 == x0 && f1 == x1)) then
	  (IFF (x0, x1))
	else
	  f
      end)
  | (FORALL _) ->
      (*AHO*)
      (assert false)
  | (EXISTS _) ->
      (*AHO*)
      (assert false))

(* Filters literals in a {cube}, to extract literals which appear in
   {preds}.  It retains the predicates when {include_or_exclude} is
   true, or removes otherwise.  It accepts {TT} as a special case (for
   trivial-sat). *)

let filter_cube_form_by_predicates (include_or_exclude : bool) (preds : form_t list) (cube : form_t) : form_t = (
  match cube with
    TT | FF ->
      (*TRIVIALCASE*) cube
  | (AND l) ->
      (AND (List.filter
	       (fun f ->
		 let appears =
		   match f with
		     TT | FF -> (assert false)
		   | ((PROP _) as x)
		   | (NOT ((PROP _) as x)) ->
		       (List.mem x preds)
		   | ((PRED _) as x)
		   | (NOT ((PRED _) as x)) ->
		       (List.mem x preds)
		   | NOT _ -> (assert false)
		   | AND _ | OR _ -> (assert false)
		   | IMP _ | IFF _ -> (assert false)
		   | FORALL _ | EXISTS _ -> (assert false) in
		 if (include_or_exclude) then
		   appears
		 else
		   (not appears))
	       l))
  | _ -> (assert false))

let filter_cube_by_predicates (include_or_exclude : bool) (preds : form_t list) (cube : cube_t) : cube_t = (
  let f = (cube_to_form cube) in
  let x = (filter_cube_form_by_predicates include_or_exclude preds f) in
  match x with
    (AND l) ->
      (literal_list_to_cube l)
  | _ -> (assert false))

(* Removes NOT node in a literal. *)

let literal_to_predicate (f : form_t) : form_t = (
  match f with (NOT p) -> p | _ -> f)

let find_predicates_in_literals (preds : form_t list) (literals : form_t list) : form_t list = (
  (List.filter
      (fun f ->(List.mem (literal_to_predicate f) preds))
      literals))

let find_predicates_in_cube (preds : form_t list) (cube : cube_t) : form_t list = (
  let literals = (cube_to_literal_list cube) in
  (find_predicates_in_literals preds literals))

let cube_mem (cube : cube_t) (f : form_t) : bool = (
  match cube with
    (CUBE s) ->
      (FormSet.mem f s))

let cube_remove (cube : cube_t) (f : form_t) : cube_t = (
  match cube with
    (CUBE s) ->
      (CUBE (FormSet.remove f s)))

let remove_predicates_in_cubes (preds : form_t list) (cubes : cube_t list) : cube_t list = (
  (List.map (filter_cube_by_predicates false preds) cubes))

(* Splits a {cube} into two cubes, where literals appear or not appear
   in {abstpreds}.  It keeps the predicates when {include_or_exclude}
   is true, or removes otherwise.  It accepts {TT} as a special case
   (for trivial-sat). *) (*SLOW*)

let split_cube_by_predicates (abstpreds : form_t list) (cube : cube_t) : (cube_t * cube_t) = (
  ((filter_cube_by_predicates true abstpreds cube),
  (filter_cube_by_predicates false abstpreds cube)))

(* Filters {cube} to extract literals which includes only
   appearances of given variables.  It keeps the predicates when
   {include_or_exclude} is true, or removes otherwise.  It accepts
   {TT} as a special case (for trivial-sat). *)

let rec filter_cube_by_variables (include_or_exclude : bool) (vs : StringSet.t) (cube : cube_t) : cube_t = (
  (form_to_cube
      (filter_cube_form_by_variables include_or_exclude vs
	  (cube_to_form cube))))

and filter_cube_form_by_variables (include_or_exclude : bool) (vs : StringSet.t) (cube : form_t) : form_t = (
  match cube with
    TT | FF ->
      (*TRIVIALCASE*) cube
  | (AND l) ->
      (AND
	  (List.filter
	      (fun f ->
		let appears =
		  match f with
		    TT | FF -> (assert false)
		  | (PROP _)
		  | (NOT (PROP _)) ->
		      false
		  | (PRED (_, t0, t1))
		  | (NOT (PRED (_, t0, t1))) ->
		      ((check_variables_in_term vs t0)
			|| (check_variables_in_term vs t1))
		  | NOT _ -> (assert false)
		  | AND _ | OR _ -> (assert false)
		  | IMP _ | IFF _ -> (assert false)
		  | FORALL _ | EXISTS _ -> (assert false) in
		if (include_or_exclude) then
		  appears
		else
		  (not appears))
	      l))
  | _ -> (assert false))

and check_variables_in_term (vs : StringSet.t) (t : term_t) : bool = (
  match t with
    (VAR s) ->
      (StringSet.mem s vs)
  | (FUN s) ->
      false
  | (UIFUN l) ->
      (List.exists (check_variables_in_term vs) l)
  | (ARITH a) ->
      (begin
	match a with
	  ZERO ->
	    false
	| (NUM n) ->
	    false
	| (CVT (n, s)) ->
	    (StringSet.mem s vs)
	| (MUL (t0, t1)) ->
	    ((check_variables_in_term vs t0)
	      || (check_variables_in_term vs t1))
	| (ADD (t0, t1)) ->
	    ((check_variables_in_term vs t0)
	      || (check_variables_in_term vs t1))
	| (NEG t0) ->
	    (check_variables_in_term vs t0)
      end)
  | (LTERM l) ->
      let ls =
	(LTermMap.fold (fun k v ss -> (StringSet.add k ss))
	    l (StringSet.empty)) in
      (not (StringSet.is_empty (StringSet.inter ls vs))))

(* Lists variables appearing in a cube.  NOTE: It accepts {TT} as a
   special case (for trivial-sat). *)

let rec list_variables_in_cube (cube : cube_t) (vs : StringSet.t) : StringSet.t = (
  (list_variables_in_cube_form (cube_to_form cube) vs))

and list_variables_in_cube_form (cube : form_t) (vs : StringSet.t) : StringSet.t = (
  match cube with
    TT | FF ->
      (*TRIVIALCASE*) (StringSet.empty)
  | (AND l) ->
      (List.fold_left
	  (fun vs0 f ->
	    match f with
	      TT | FF -> (assert false)
	    | PROP v -> (assert false)
	    | (PRED (_, t0, t1)) | (NOT (PRED (_, t0, t1))) ->
		(list_variables_in_term t1
		    (list_variables_in_term t0 vs0))
	    | NOT _ -> (assert false)
	    | AND _ | OR _ -> (assert false)
	    | IMP _ | IFF _ -> (assert false)
	    | FORALL _ | EXISTS _ -> (assert false))
	  vs l)
  | _ -> (assert false))

and list_variables_in_term (t : term_t) (vs : StringSet.t) : StringSet.t = (
  match t with
    (VAR s) ->
      (StringSet.add s vs)
  | (FUN s) ->
      vs
  | (UIFUN l) ->
      (List.fold_left
	  (fun vs0 t0 ->
	    (list_variables_in_term t0 vs0))
	  vs l)
  | (ARITH a) ->
      (begin
	match a with
	  ZERO ->
	    vs
	| (NUM n) ->
	    vs
	| (CVT (n, s)) ->
	    (StringSet.add s vs)
	| (MUL (t0, t1)) ->
	    (list_variables_in_term t1
		(list_variables_in_term t0 vs))
	| (ADD (t0, t1)) ->
	    (list_variables_in_term t1
		(list_variables_in_term t0 vs))
	| (NEG t0) ->
	    (list_variables_in_term t0 vs)
      end)
  | (LTERM l) ->
      (LTermMap.fold
	  (fun s v ss ->
	    (StringSet.add s ss))
	  l
	  vs))

(* FOR DEBUGGING *)

let cnf_number_of_clauses (cnf : form_t) = (
  match cnf with
    (AND l) ->
      (List.length l)
  | _ -> (assert false))

let cnf_clause_length (clause : form_t) = (
  match clause with
    (OR l) ->
      (List.length l)
  | _ -> (assert false))

let cnf_average_clause_length (cnf : form_t) = (
  match cnf with
    (AND []) ->
      0
  | (AND l) ->
      let t = (List.fold_left (fun s c -> (s + (cnf_clause_length c))) 0 l) in
      ((t * 1000) / (List.length l))
  | _ -> (assert false))

(* TRANSLATORS TO EXPRESSIONS AND PREDICATES *)

module Logic = LogicUtil.Logic

let get_function_name (term : term_t) : string =
    match term with
    | FUN fun_name -> fun_name
    | _ -> assert false

let rec term_to_expr (term : term_t) : Logic.expression =
    match term with
    | VAR var_name -> 
         Logic.Variable var_name
    | UIFUN terms ->
         begin
            match terms with
            | [] -> assert false
            | h::terms -> 
                let fun_name = get_function_name h in 
                let exprs = List.map term_to_expr terms in
                Logic.Application (fun_name, exprs)  
         end
    | ARITH arith_term -> 
         arith_term_to_expr arith_term
    | FUN _ (* not allowed to appear here *)
    | LTERM _ -> assert false (* LTERM appears improperly *)

and arith_term_to_expr (arith_term : arith_term_t) : Logic.expression =
    match arith_term with
    | NUM n 
        -> Logic.Constant (Num.float_of_num n)
    | MUL (term1, term2) 
        -> begin
             let expr1 = term_to_expr term1 
             and expr2 = term_to_expr term2 in
             match expr1, expr2 with
             | Logic.Constant f, _ -> Logic.Coeff (f, expr2)
             | _, Logic.Constant f -> Logic.Coeff (f, expr1)
             | _ -> Logic.Application ("MUL", [expr1; expr2])
           end
    | ADD (term1, term2) 
        -> let expr1 = term_to_expr term1 
           and expr2 = term_to_expr term2 in
           Logic.Sum [expr1; expr2]
    | NEG term 
        -> Logic.Coeff(-1.0, term_to_expr term)
    | ZERO 
    | CVT _ -> assert false (* ZERO & CVT appear improperly *)

let rec form_to_predicate (input : form_t) :  Logic.predicate = 
    match input with
    | TT 
        -> Logic.True
    | FF 
        -> Logic.False
    | PROP var 
        -> assert false (* temporarily not considered *)
    | PRED (op, term1, term2) 
        -> begin
           let expr1 = term_to_expr term1 
           and expr2 = term_to_expr term2 in
           match op with
           | EQ -> Logic.Eq (expr1, expr2)
           | NE -> Logic.Not (Logic.Eq(expr1,expr2))
           | GE -> Logic.Not (Logic.Lt(expr1,expr2))
           | GT -> Logic.Not (Logic.Leq(expr1,expr2))
           | LE -> Logic.Leq(expr1, expr2)
           | LT -> Logic.Lt(expr1, expr2)
           | DIVS -> assert false (* temporarily not considered *)
        end
    | NOT form 
        -> Logic.Not (form_to_predicate form)
    | AND forms
        -> let predicates = List.map form_to_predicate forms in
           Logic.And predicates
    | OR forms
        -> let predicates = List.map form_to_predicate forms in
           Logic.Or predicates
    | IMP (form1, form2)
        -> let pred1 = form_to_predicate form1
           and pred2 = form_to_predicate form2 in
           Logic.Or [Logic.Not pred1; pred2]
    | IFF (form1, form2)
        (* stupid way to handle *)
        -> let pred1 = form_to_predicate (IMP(form1, form2))
           and pred2 = form_to_predicate (IMP(form1, form2)) in
           Logic.And [pred1; pred2]
    | FORALL _
    | EXISTS _ 
        -> assert false  (* temporarily not considered *)
