(* parser/interp.ml 2009-01-26 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Abstract Interpreter/Checker *)

(* TODO: COVERING TEST OF JUNCTIONS NEEDS PREDICATES ON GLOBALS. *)

(* MEMO: (1) Handling Loops: Label which is a target from multiple
   gotoes records a state set, which is used to check a invaliant or
   to find a fixedpoint. *)

(* MEMO: Structure reference is converted to an UI function:
   ("sref%slot" x) where a slot name is not unique but is unique to
   each value/variable... (not sure, may be). *)

(* MEMO: Predicates used for predicate abstraction are quantifier
   free. *)

let aiTRACE_dump_form = ref false
let aiTRACE_interp = ref true
let aiTRACE_interp_smt = ref false
let aiTRACE_linking = ref true
let aiOPTION_merge_cubes = ref false
(*let aiTRACE_global_state = ref true*)

let info = Cexpr.info
let debug = Cexpr.debug

(* Flag which indicates whether to deeply check pointer accessses *)
let intensive_pointer_checking = ref false

let structural_equal = (=)

open Num
open Form
open Smt
open Translate

module Sat = Minisat
module StringSet = Form.StringSet
module FormSet = Form.FormSet
(*module StringInt2Set = Csyntax.StringInt2Set*)

(* Cube with a marker string.  A marker string is used to associate
   each cube with a calling state. *)

type markedcube_t = (string * cube_t)

let markedcube_compare ((xs, xc) : markedcube_t) ((ys, yc) : markedcube_t) : int = (
  let v = (String.compare xs ys) in
  if (v != 0) then
    v
  else
    (cube_compare xc yc))

module MarkedCubeSet = Set.Make (
  struct
  type t = markedcube_t
  let compare x y = (markedcube_compare x y)
  end)

type expr_t = Cexpr.expr_t
type stmt_t = Cexpr.stmt_t

let string_equal = Misc.string_equal

let pseudo_entry_label : string = Csyntax.pseudo_entry_label
let pseudo_return_label : string = Csyntax.pseudo_return_label
let pseudo_elsewhere_label : string = Csyntax.pseudo_elsewhere_label

(* State flowing-through, a pair of a cubeset and a conjunction of
   forms representing statements. *)

type flowin_state_t = (markedcube_t list * form_t)

(* Evaluation Context.  {ctx_function_memoizers} returns a memoizer
   for a function (a pair of a function name and a file name). *)

type program_context_t = {
  ctx_afile : afile_t ;
  ctx_global_abstpreds : (string, form_t list) Hashtbl.t ;
  ctx_function_memoizers : ((string * string), function_memoizer_t) Hashtbl.t
}

(* Record of Function Memoization.  {fmem_memoizer} maps a calling
   state to a returning state.  {fmem_recurse} records recursive call
   state. *)

and function_memoizer_t = {
  fmem_memoizer : (cube_t, cube_t list) Hashtbl.t ;
  mutable fmem_recurse : bool option
}

(* Function Evaluation Context.  {fctx_return_context} is a
   continuation.  {fctx_calling_global_states} is a list of global
   calling states.  {fctx_calling_cubes} is a local state after
   mapping to versioned variables to evaluate in a function.
   {fctx_junction_covers} is a mapping from a label to a covered
   states (a list of marked cubes).  {fctx_current_slices} records
   dropped global states not accessed/modified in this function
   execution.  {fctx_result_cubes} is a collected resulting states,
   which have markers on the caller side.
   {fctx_result_cubes_collecting} is a temporary collector of result
   states, which are later merged to {fctx_result_cubes}.
   {fctx_remaining_contexts} records choices of else-parts of
   branches. *)

and function_context_t = {
  fctx_fdef : fundef_t ;
  fctx_return_context : eval_next_t ;
  fctx_calling_global_states : markedcube_t list ;
  fctx_calling_cubes : markedcube_t list ;
  fctx_junction_covers : (string, markedcube_t list) Hashtbl.t ;
  mutable fctx_current_slices : (string, markedcube_t list) Hashtbl.t ;
  mutable fctx_result_cubes : markedcube_t list ;
  mutable fctx_result_cubes_collecting : markedcube_t list ;
  mutable fctx_remaining_contexts : eval_next_t list
}

(* {EVALCONT} is of ({fctx}, {label}, {sourcelabel}, {flowform},
   {cc}), which is a tuple of a current function-context, a label to
   resume, a label of current block (used only in junction blocks to
   choose phi), a flow form to be passed to the continuation, and a
   next continuation.  {CONTRET} is for function calls, and {CONTALT}
   is for conditional choices.  *)

and eval_next_t =
    EVALSTOP
    | EVALCONT of eval_context_t

and eval_context_t = {
  ectx_fctx : function_context_t ;
  ectx_current_slices : (string, markedcube_t list) Hashtbl.t option ;
  ectx_label : string ;
  ectx_source_label : string ;
  ectx_flowform : flowin_state_t
}

let xTRACE (s : string) : unit = (if (true) then (debug (";; " ^ s)))
let xPRINT (f : bool) (s0 : string) (s1 : string) : unit =
  (if (f) then (debug (";; " ^ s0 ^ s1)))
let aiTRACEPRINT (s : string) : unit = (
  if (!aiTRACE_interp) then (debug (";; " ^ s)))

let ite (c : bool) (e0 : 'a) (e1 : 'a) : 'a = (if c then e0 else e1)

let funcname (fdef : fundef_t) : string = (
  let decl : Cexpr.declaration_t = fdef.fundef_declaration in
  (fst (decl.Cexpr.decl_id)))

let markedcube_to_string ((m, c) : markedcube_t) : string = (
  ((cube_to_string c) ^ ":" ^ m))

let markedcube_list_to_string (l : markedcube_t list) : string = (
  ("[" ^ (String.concat ";" (List.map markedcube_to_string l)) ^ "]"))

let aiTRACEPRINT_flowing_state (s : flowin_state_t) (dropped : (string, markedcube_t list) Hashtbl.t) : unit = (
  let (cubes, statements) = s in
  (aiTRACEPRINT
      ("- states=" ^ (markedcube_list_to_string cubes))) ;
  (aiTRACEPRINT
      ("- slices =["
	^ (String.concat ";"
	      (Misc.hashtbl_map
		  (fun (newmark : string) (l : markedcube_t list) ->
		    (String.concat ";"
			(List.map
			    (fun cube ->
			      (newmark ^ "->" ^ (markedcube_to_string cube)))
			    l)))
		  dropped))
	^ "]")) ;
  (aiTRACEPRINT
      ("- constraints=" ^ (form_to_string statements))))

let caller_callee_mark_map_string (dropped : (string, markedcube_t list) Hashtbl.t) : string = (
  (String.concat ";"
      (Misc.hashtbl_map
	  (fun (newmark : string) (l : markedcube_t list) ->
	    (String.concat ";"
		(List.map
		    (fun ((oldmark : string), cube) ->
		      (oldmark ^ "->" ^ newmark))
		    l)))
	  dropped)))

(* ================================================================ *)

let get_memoizer (gctx : program_context_t) (funname : (string * string)) : function_memoizer_t = (
  if (not (Hashtbl.mem gctx.ctx_function_memoizers funname)) then
    (begin
      let memo = {
	  fmem_memoizer = (Hashtbl.create 10) ;
	  fmem_recurse = None
      } in
      (Hashtbl.replace gctx.ctx_function_memoizers funname memo)
    end) ;
  (Hashtbl.find gctx.ctx_function_memoizers funname))

let memoize_values (memoizer : (cube_t, cube_t list) Hashtbl.t) (callcube : cube_t) (cubes : cube_t list) : unit = (
  (aiTRACEPRINT
      ("memoize call=" ^ (cube_to_string callcube)
	^ ";value=" ^ (cube_list_to_string cubes))) ;
  (Hashtbl.replace memoizer callcube cubes))

(* Check memoized state of a function call.  It returns triple of a
  marker (of a calling state), a calling state, and a list of result
  states. *)

let check_memoization (gctx : program_context_t) (fn : (string * string)) (callcubes : markedcube_t list) : (string * cube_t * cube_t list option) list = (
  let memo : function_memoizer_t = (get_memoizer gctx fn) in
  let recurse : bool option = (memo.fmem_recurse) in
  (List.map
      (fun ((mark : string), (cube : cube_t)) ->
	if (not (Hashtbl.mem memo.fmem_memoizer cube)) then
	  (begin
	    (* UNCOVERED.  Record bottom.  Record non-recursion. *)
	    (Hashtbl.replace memo.fmem_memoizer cube []) ;
	    if (memo.fmem_recurse == None) then
	      memo.fmem_recurse <- (Some false) ;
	    (mark, cube, None)
	  end)
	else
	  (begin
	    (*assert (memo.fmem_recurse != None)*)
	    let results = (Hashtbl.find memo.fmem_memoizer cube) in
	    if (structural_equal recurse None) then
	      (begin
		(* COVERED.  (Recurse already done). *)
		(mark, cube, (Some results))
	      end)
	    else if (structural_equal recurse (Some false)) then
	      (begin
		(* COVERED.  First recurse, set flag true. *)
		memo.fmem_recurse <- (Some true) ;
		(mark, cube, (Some []))
	      end)
	    else
	      (begin
		(* COVERED.  In recurse. *)
		assert (structural_equal recurse (Some true)) ;
		(mark, cube, (Some []))
	      end)
	  end))
      callcubes))

(* ================================================================ *)

(* INTERPRETER *)

(* Merges two lists of marked cubes, removing equal ones. *)

let markedcube_list_union (l0 : markedcube_t list) (l1 : markedcube_t list) : markedcube_t list = (
  let ss0 : MarkedCubeSet.t =
    (List.fold_left
	(fun ss c -> (MarkedCubeSet.add c ss))
	(MarkedCubeSet.empty)
	l0) in
  let ss1 : MarkedCubeSet.t =
    (List.fold_left
	(fun ss c -> (MarkedCubeSet.add c ss))
	ss0
	l1) in
  (MarkedCubeSet.elements ss1))

let uniqify_markedcube_list (l : markedcube_t list) : markedcube_t list = (
  (markedcube_list_union l []))

(* Makes a set difference; (A - B) (non-symmetric difference). *)

let markedcube_list_difference (l0 : markedcube_t list) (l1 : markedcube_t list) : markedcube_t list = (
  let ss0 : MarkedCubeSet.t =
    (List.fold_left
	(fun ss c -> (MarkedCubeSet.add c ss))
	(MarkedCubeSet.empty)
	l0) in
  let ss1 : MarkedCubeSet.t =
    (List.fold_left
	(fun ss c -> (MarkedCubeSet.add c ss))
	(MarkedCubeSet.empty)
	l1) in
  (MarkedCubeSet.elements (MarkedCubeSet.diff ss0 ss1)))

let trace_print_infeasible_path (printlabel : string) (form : form_t) : unit = (
  (aiTRACEPRINT
      (printlabel ^ " INFEASIBLE PATH: " ^ (form_to_string form))))

let make_empty_covering (bbset : (string, form_block_t) Hashtbl.t) : (string, markedcube_t list) Hashtbl.t = (
  let covermap : (string, markedcube_t list) Hashtbl.t = (Hashtbl.create 10) in
  (Hashtbl.iter
      (fun label b ->
	match b with
	  { bb_body = BBJUNC body } ->
	    (Hashtbl.replace covermap label [])
	| _ ->
	    ())
      bbset) ;
  covermap)

let find_function (afile : afile_t) (f : Cexpr.vartag_t) (call : term_t list) : definition_t = (
  assert (call != []) ;
  match f with
    (id, level, sequence) when (not (string_equal id "")) ->
      if (not (Hashtbl.mem afile.afile_directory id)) then
	(failwith ("Function '" ^ id ^ "' not found")) ;
      let filename : string = (Hashtbl.find afile.afile_directory id) in
      if (not (Hashtbl.mem afile.afile_libs filename)) then
	(failwith ("Library '" ^ filename ^ "' not found")) ;
      let ofile : ofile_t = (Hashtbl.find afile.afile_libs filename) in
      if (not (Hashtbl.mem ofile.ofile_defs id)) then
	(failwith ("Function '" ^ id ^ "' not found")) ;
      (Hashtbl.find ofile.ofile_defs id)
  | _ ->
      (*AHO*)
      (failwith ("FUNCTION POINTER CASE: ALIAS LIST BE GIVEN")))

(* Splits cubes in two parts of related/unrelated literals with regard
   to the given {predicates}.  It returns related literals in cubes,
   and remaining literals in a hashtable.  Each cube is associated to
   a new marker, and a hashtable is keyed by the markers to dropped
   literals.  *)

let split_and_mark_states (predicates : form_t list) (cubes : markedcube_t list) : (markedcube_t list * (string, markedcube_t list) Hashtbl.t) = (
  let dropped : (string, markedcube_t list) Hashtbl.t = (Hashtbl.create 10) in
  let cubemap : (cube_t, string) Hashtbl.t = (Hashtbl.create 10) in
  (List.iter
      (fun ((oldmark : string), (cube : cube_t)) ->
	let ((cin : cube_t), (cex : cube_t)) =
	  (split_cube_by_predicates predicates cube) in
	if (Hashtbl.mem cubemap cin) then
	  (begin
	    let newmark = (Hashtbl.find cubemap cin) in
	    let tl = (Hashtbl.find dropped newmark) in
	    (Hashtbl.replace dropped newmark ((oldmark, cex) :: tl))
	  end)
	else
	  (begin
	    let newmark = (xgensym "K") in
	    (Hashtbl.replace cubemap cin newmark) ;
	    (Hashtbl.replace dropped newmark [(oldmark, cex)])
	  end))
      cubes) ;
  let newcores : markedcube_t list =
    (Hashtbl.fold
	(fun cube mark tl -> ((mark, cube) :: tl))
	cubemap []) in
  (newcores, dropped))

(* Recovers dropped (split) part of each cube.  It returns a list of
   recovered cubes, in which a marker string still remains. *)

let rec recover_split_states (dropped : (string, markedcube_t list) Hashtbl.t) (cubes : markedcube_t list) : markedcube_t list = (
  (List.flatten
      (List.map
	  (fun ((newmark : string), (corepart : cube_t)) ->
	    let droppedparts : markedcube_t list =
	      (Hashtbl.find dropped newmark) in
	    (List.map
		(fun ((oldmark : string), (droppedpart : cube_t)) ->
		  (oldmark, (cube_append corepart droppedpart)))
		droppedparts))
	  cubes)))

(* Finds global predicates which include global variables appearing in
   a function. *)

let filter_global_predicates (globals : varname_t list) (globalpreds : (string, form_t list) Hashtbl.t) : form_t list = (
  (List.flatten
      (List.map
	  (Misc.hashtbl_find_safe [] globalpreds)
	  globals)))

(* Makes a transition.  It calls an SMT solver, and copies marker
   strings from pre-states to post-states.  The {once} version calls
   SMT solver at once, and the {each} verson calls SMT solver for each
   state.  THE {each} VERSION IS FASTER. *)

let rec solve_feasible_cubes (states : markedcube_t list) (constraints : form_t) (predicates : form_t list) (printlabel : string) (message : string) : markedcube_t list = (
  if (!aiOPTION_merge_cubes) then
    (solve_feasible_cubes_once
	states constraints predicates printlabel message)
  else
    (solve_feasible_cubes_each
	states constraints predicates printlabel message))

and solve_feasible_cubes_once (states : markedcube_t list) (constraints : form_t) (predicates : form_t list) (printlabel : string) (message : string) : markedcube_t list = (
  let positivesnegatives =
    (AND (List.map (fun f -> (OR [f ; (NOT f)])) predicates)) in
  let truestates : cube_t list = (List.map snd states) in
  let evalform : form_t =
    (pseudo_conjoin (cube_list_to_form truestates)
	(pseudo_conjoin constraints positivesnegatives)) in
  xTRACE ("evalform=" ^ (form_to_string evalform)) ;
  let feasibles : cube_t list =
    (form_list_to_cube_list
	(smt_solve_satisfying_cubes evalform)) in
  let nextstates : markedcube_t list =
    (uniqify_markedcube_list
	(List.flatten
	    (List.map
		(filter_and_mark_state_as_origins states predicates)
		feasibles))) in

  if (!aiTRACE_interp_smt) then
    (aiTRACEPRINT
	(printlabel ^ " " ^ message ^ ":\n"
	  ^ ";; - form=" ^ (form_to_string evalform) ^ "\n"
	  ^ ";; - feasibles=" ^ (cube_list_to_string feasibles) ^ "\n"
	  ^ ";; - nextstates=" ^ (markedcube_list_to_string nextstates))) ;

  nextstates)

and solve_feasible_cubes_each (states : markedcube_t list) (constraints : form_t) (predicates : form_t list) (printlabel : string) (message : string) : markedcube_t list = (
  let positivesnegatives =
    (AND (List.map (fun f -> (OR [f ; (NOT f)])) predicates)) in
  let nextstates : markedcube_t list =
    (uniqify_markedcube_list
	(List.flatten
	    (List.map
		(fun ((mark, state) : markedcube_t) ->
		  let evalform : form_t =
		    (pseudo_conjoin (cube_to_form state)
			(pseudo_conjoin constraints positivesnegatives)) in
		  xTRACE ("evalform=" ^ (form_to_string evalform)) ;
		  let feasibles : cube_t list =
		    (form_list_to_cube_list
			(smt_solve_satisfying_cubes evalform)) in
		  let filterednewstates : cube_t list =
		    (List.map
			(filter_cube_by_predicates true predicates)
			feasibles) in
		  (List.map (fun cube -> (mark, cube)) filterednewstates))
		states))) in

  if (!aiTRACE_interp_smt) then
    (aiTRACEPRINT
	(printlabel ^ " " ^ message ^ ":\n"
	    (*^ ";; - form=" ^ (form_to_string evalform) ^ "\n"*)
	    (*^ ";; - feasibles=" ^ (cube_list_to_string feasibles) ^ "\n"*)
	  ^ ";; - nextstates=" ^ (markedcube_list_to_string nextstates))) ;

  nextstates)

(* Filters a state with predicates, and marks the state as of one of
   the old states. *)

and filter_and_mark_state_as_origins (oldstates : markedcube_t list) (predicates : form_t list) (newstate : cube_t) : markedcube_t list = (
  let origins : markedcube_t list =
    (List.filter
	(fun (_, (origin : cube_t)) -> (cube_subset origin newstate))
	oldstates) in
  assert (origins != []) ;
  let filterednewstate : cube_t =
    (filter_cube_by_predicates true predicates newstate) in
  (List.map
      (fun ((mark : string), _) ->
	(mark, filterednewstate))
      origins))

(* Evaluates a block of {blocklabel}.  {sourcelabel} is the previous
  block. *)

let rec eval_path (gctx : program_context_t) (fctx : function_context_t) (blocklabel : string) (sourcelabel : string) (flowform : flowin_state_t) : unit = (
  assert (not (string_equal blocklabel pseudo_elsewhere_label)) ;
  (*let afile = gctx.ctx_afile in*)
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  let b : form_block_t = (Hashtbl.find bbset blocklabel) in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  match b with
    {
      bb_label = label ;
      bb_memo = _ ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = [_] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = (BBHEAD _)
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " HEAD"))*)
	(*(eval_path gctx fctx outgoing label flowform)*)
	(assert false)
      end)
  | {
      bb_label = _ ;
      bb_memo = _ ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = [_] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = (BBTAIL _)
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " TAIL"))*)
	assert (string_equal outgoing pseudo_elsewhere_label) ;
	(eval_path_tail gctx fctx b flowform)
      end)
  | {
      bb_label = _ ;
      bb_memo = _ ;
      bb_outgoing_labels = [_] ;
      bb_incoming_labels = [_] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = (BBSTMT _)
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " STMT"))*)
	(eval_path_statement gctx fctx b sourcelabel flowform)
      end)
  | {
      bb_label = _ ;
      bb_memo = _ ;
      bb_outgoing_labels = [_] ;
      bb_incoming_labels = incomings ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBJUNC {
	bbjunc_phis = phis ;
	bbjunc_phifix = _ ;
	bbjunc_predicates = _ ;
	bbjunc_predmap = _
      }
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " JUNC"))*)
	assert (List.mem sourcelabel incomings) ;
	let (cubes, statements) = flowform in
	let form = (Hashtbl.find phis sourcelabel) in
	(check_at_junction gctx fctx b cubes
	    (pseudo_conjoin statements form))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = _ ;
      bb_incoming_labels = [_] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBCOND {
	bbcond_cond = (cond, thengo, elsego)
      }
    } ->
      (begin
	let (cubes, statements) = flowform in
	let f2 = (pseudo_conjoin statements cond) in
	let f3 = (pseudo_conjoin statements (NOT cond)) in

	(aiTRACEPRINT
	    (printlabel ^ " taking path cond-then near " ^ ""(*memo*) ^ ":")) ;
	(aiTRACEPRINT_flowing_state (cubes, f2)
	    fctx.fctx_current_slices) ;

	let cc = EVALCONT {
	  ectx_fctx = fctx ;
	  ectx_current_slices = (Some fctx.fctx_current_slices) ;
	  ectx_label = elsego ;
	  ectx_source_label = label ;
	  ectx_flowform = (cubes, f3)
	} in
	fctx.fctx_remaining_contexts <-
	  (fctx.fctx_remaining_contexts @ [cc]) ;
	(eval_path gctx fctx thengo label (cubes, f2))
      end)
  | {
      bb_label = _ ;
      bb_memo = _ ;
      bb_outgoing_labels = [_] ;
      bb_incoming_labels = [_] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = (BBCALL _)
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " CALL"))*)
	(eval_path_call gctx fctx b sourcelabel flowform)
      end)
  | _ -> (assert false))

(* Interprets a function at the toplevel. *)

and eval_path_function (gctx : program_context_t) (fctx : function_context_t) (blocklabel : string) (sourcelabel : string) (callstates : markedcube_t list) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  let printlabel = ("(" ^ fn ^ "/" ^ sourcelabel ^ ")") in
  (aiTRACEPRINT
      (printlabel ^ " next.state="
	^ (markedcube_list_to_string callstates))) ;
  (eval_path gctx fctx blocklabel sourcelabel (callstates, TT)))

and eval_path_statement (gctx : program_context_t) (fctx : function_context_t) (b : form_block_t) (sourcelabel : string) (flowform : flowin_state_t) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  let decl : Cexpr.declaration_t = fdef.fundef_declaration in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  match b with
    {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = [incoming] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBSTMT {
	bbstmt_form = form ;
	bbstmt_assert = assertform ;
	bbstmt_assume = assumeform ;
	bbstmt_invariant = invariantform
      }
    } ->
      (begin
	let (cubes, statements) = flowform in
	let f1 = (pseudo_conjoin statements form) in

	(aiTRACEPRINT
	    (printlabel ^ " statement near " ^ (*memo*)"" ^ ":")) ;
	(aiTRACEPRINT_flowing_state (cubes, f1)
	    fctx.fctx_current_slices) ;

	if (assertform != TT) then
	  (begin
	    let truecubes = (List.map snd cubes) in
	    let fx = (pseudo_conjoin
			 (cube_list_to_form truecubes)
			 f1) in
	    let ok = (check_at_assertion gctx fctx decl bbset
			 b fx assertform) in
	      if (ok) then
		(aiTRACEPRINT ("* ASSERTION OK near at " ^ (*memo*)""
				^ " in " ^ printlabel)) ;
	  end) ;
	let f2 = (pseudo_conjoin f1 assertform) in
	let f3 = (pseudo_conjoin f2 assumeform) in
	assert (not (string_equal outgoing pseudo_elsewhere_label)) ;
	(eval_path gctx fctx outgoing label (cubes, f3))
      end)
  | _ -> (assert false))

(* Once runs an SMT, and calls {eval_path_head} on a new function with
   (recovered) global states. *)

and eval_path_call (gctx : program_context_t) (fctx : function_context_t) (b : form_block_t) (sourcelabel : string) (flowform : flowin_state_t) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  let afile = gctx.ctx_afile in
  match b with
    {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = [incoming] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBCALL {
	bbcall_fn = (target, _, _) as fname ;
	bbcall_call = (call : term_t list) ;
	bbcall_global_reversions = _ ;
	bbcall_globals_inmap = _ ;
	bbcall_globals_outmap = voutmap ;
	bbcall_predicates = (localpreds : form_t list)
      }
    } ->
      (begin
	let fn : string = (funcname fdef) in
	let printlabel = ("(" ^ fn ^ "/" ^ label ^ ")") in

	(aiTRACEPRINT
	    (printlabel ^ " call '" ^ target ^ "' near " ^ (*memo*)"" ^ ":" )) ;
	(aiTRACEPRINT_flowing_state flowform
	    fctx.fctx_current_slices) ;

	let (statecubes, statements) = flowform in
	let abstpreds = localpreds in
	let feasibles : markedcube_t list =
	  (solve_feasible_cubes statecubes statements abstpreds
	      printlabel "check states at call") in

	if (feasibles == []) then
	  (begin
	    (* Unsatisfied; infeasible path. *)
	    let truecubes = (List.map snd statecubes) in
	    let evalform = (pseudo_conjoin
			       (cube_list_to_form truecubes)
			       statements) in
	    (trace_print_infeasible_path printlabel evalform) ;
	    let tailb = (Hashtbl.find bbset pseudo_return_label) in
	    (eval_path_return gctx fctx tailb)
	  end)
	else
	  (begin
	    let states : markedcube_t list =
	      (uniqify_markedcube_list
		  (List.map
		      (fun ((mark : string), (cube : cube_t)) ->
			let cube1 =
			  (substitute_variables_in_cube voutmap
			      cube
			      (*GOMI*)(*(filter_cube_by_predicates
					true abstpreds cube)*)) in
			(mark, (pseudo_simplify_cube cube1)))
		      feasibles)) in
	    let globalstates : markedcube_t list =
	      (recover_split_states fctx.fctx_current_slices states) in

	    let def : definition_t = (find_function afile fname call) in
	    match def with
	      (FUNDEF newfdef) ->
		(begin
		  (eval_path_head gctx newfdef call
		      abstpreds globalstates
		      (EVALCONT {
			ectx_fctx = fctx ;
			ectx_current_slices = None ;
			ectx_label = label ;
			ectx_source_label = sourcelabel ;
			ectx_flowform = flowform
		      }))
		end)
	    | _ ->
		(Cexpr.generalerror ("No defintion for '" ^ target ^ "'"))
	  end)
      end)
  | _ -> (assert false))

and eval_path_head (gctx : program_context_t) (fdef : fundef_t) (callterms : term_t list) (callerabstpreds : form_t list) (callingstates : markedcube_t list) (cc : eval_next_t) : unit = (
  let fn : string = (funcname fdef) in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  let b = (Hashtbl.find bbset pseudo_entry_label) in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  match b with
    {
      bb_label = label ;
      bb_memo = _ ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = [incoming] ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBHEAD {
	bbhead_pre = (assertform, assumeform, invariantform) ;
	bbhead_predicates = (localpreds : form_t list) ;
	bbhead_predmap = predmap ;
	bbhead_params = params ;
	bbhead_paramsmap = paramsmap ;
	bbhead_globals_inmap = vinmap
      }
    } ->
      (begin
	let globalpreds : form_t list =
	  (filter_global_predicates (Misc.hashtbl_keys vinmap)
	      gctx.ctx_afile.afile_global_predicates) in
	let abstpreds : form_t list =
	  (uniqify_form_list (globalpreds @ localpreds)) in

	(aiTRACEPRINT
	    (printlabel ^ " function entry:"
	      ^ "\n;; - calling.states="
	      ^ (markedcube_list_to_string callingstates)
	      ^ "\n;; - predicates="
	      ^ (form_list_to_string abstpreds))) ;

	(* Pickup related global states to both caller and callee. *)

	let unionpreds : form_t list =
	  (uniqify_form_list (abstpreds @ callerabstpreds)) in
	let (states1, dropped1) =
	  (split_and_mark_states unionpreds callingstates) in

	(* Evaluate parameter passing. *)

	let parsform : form_t = (match_arguments params (List.tl callterms)) in
	let feasibles : markedcube_t list =
	  (solve_feasible_cubes states1 parsform abstpreds
	      printlabel "check states at parameter matching") in

	if (feasibles == []) then
	  (begin
	    (* Unsatisfied; infeasible path. *)
	    let truecubes = (List.map snd states1) in
	    let evalform = (pseudo_conjoin
			       (cube_list_to_form truecubes)
			       parsform) in
	    (trace_print_infeasible_path printlabel evalform) ;
	    let dummyfctx = {
	      fctx_fdef = fdef ;
	      fctx_return_context = cc ;
	      fctx_calling_global_states = [] ;
	      fctx_calling_cubes = [] ;
	      fctx_junction_covers = (make_empty_covering bbset) ;
	      fctx_current_slices = (Hashtbl.create 10) ;
	      fctx_result_cubes = [] ;
	      fctx_result_cubes_collecting = [] ;
	      fctx_remaining_contexts = []
	    } in
	    let tailb = (Hashtbl.find bbset pseudo_return_label) in
	    (eval_path_return gctx dummyfctx tailb)
	  end)
	else
	  (begin

	    (* Recover calling states after removing caller states. *)

	    (*GOMI*)(*
	    let states : markedcube_t list =
	      (List.map
		  (fun ((mark : string), (cube : cube_t)) ->
		    let cube1 =
		      (filter_cube_by_predicates
			  true abstpreds cube) in
		    (mark, cube1))
		  feasibles) in
	    *)
	    let truecallingstates : markedcube_t list =
	      (recover_split_states dropped1 feasibles) in

	    (* Check memoized state of a function call. *)

	    let (resultpairs : (string * cube_t * cube_t list option) list) =
	      (check_memoization gctx (fn, fdef.fundef_hash_name)
		  truecallingstates) in
	    let someuncovered =
	      (List.exists
		  (fun (_, _, (results : cube_t list option)) ->
		    (results == None))
		  resultpairs) in

	    if (not someuncovered) then
	      (begin
		(aiTRACEPRINT (printlabel ^ " CALL COVERED")) ;

		let nextstates : markedcube_t list =
		  (List.flatten
		      (List.map
			  (fun (mark, call, (results : cube_t list option)) ->
			    (List.map
				(fun cube -> (mark, cube))
				(Misc.some_of_safe [] results)))
			  resultpairs)) in
		(eval_path_continue gctx cc nextstates)
	      end)
	    else
	      (begin

		(* Mark states with key strings, drop callers markers. *)

		let uncoveredcallingstates : markedcube_t list =
		  (Misc.list_filtering_map
		      (fun (oldmark, callcube, results) ->
			if (results == None) then
			  (Some (oldmark, callcube))
			else
			  None)
		      resultpairs) in

		let (states2, dropped2) =
		  (split_and_mark_states abstpreds uncoveredcallingstates) in
		let mappedcallingstates : markedcube_t list =
		  (List.map
		      (fun ((mark : string), (cube : cube_t)) ->
			let cube1 : cube_t =
			  (substitute_variables_in_cube paramsmap
			      (substitute_variables_in_cube vinmap cube)) in
			(mark, (pseudo_simplify_cube cube1)))
		      states2) in
		assert (mappedcallingstates != []) ;

		(aiTRACEPRINT
		    (printlabel ^ " function prologue:"
		      ^ "\n;; - callee.states="
		      ^ (markedcube_list_to_string mappedcallingstates)
		      ^ "\n;; - dropped.states="
		      ^ (markedcube_list_to_string
			    (List.flatten
				(Misc.hashtbl_map (fun s l -> l) dropped2)))
		      ^ "\n;; - predicates="
		      ^ (form_list_to_string abstpreds)
		      ^ "\n;; - caller/callee.map="
		      ^ (caller_callee_mark_map_string dropped2))) ;

		(* Make new function state. *)

		let newfctx = {
		  fctx_fdef = fdef ;
		  fctx_return_context = cc ;
		  fctx_calling_global_states = uncoveredcallingstates ;
		  fctx_calling_cubes = mappedcallingstates ;
		  fctx_junction_covers = (make_empty_covering bbset) ;
		  fctx_current_slices = dropped2 ;
		  fctx_result_cubes = [] ;
		  fctx_result_cubes_collecting = [] ;
		  fctx_remaining_contexts = []
		} in
		(eval_path gctx newfctx outgoing label
		    (mappedcallingstates, TT))
	      end)
	  end)
      end)
  | _ -> (assert false))

and eval_path_tail (gctx : program_context_t) (fctx : function_context_t) (b : form_block_t) (flowinform : flowin_state_t) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  (*let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in*)
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  match b with
    {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = _ ;
      bb_incoming_labels = _ ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBTAIL {
	bbtail_predicates = (localpreds : form_t list) ;
	bbtail_globals_outmap = voutmap
      }
    } ->
      (begin
	(*(aiTRACEPRINT (printlabel ^ " function tail:"))*)

	let globalpreds : form_t list =
	  (filter_global_predicates (Misc.hashtbl_values voutmap)
	      gctx.ctx_afile.afile_global_predicates) in

	let abstpreds : form_t list =
	  (uniqify_form_list (globalpreds @ localpreds)) in

	(* Rename versioned globals to globals. *)

	let (cubes0, statements0) = flowinform in
	let cubes : markedcube_t list =
	  (List.map
	      (fun ((mark : string), (cube : cube_t)) ->
		let cube1 = (substitute_variables_in_cube voutmap cube) in
		(mark, cube1))
	      cubes0) in
	let statements = (substitute_variables_in_form voutmap statements0) in
	let feasibles : markedcube_t list =
	  (solve_feasible_cubes cubes statements abstpreds
	      printlabel "check states at tail") in

	if (feasibles == []) then
	  (begin
	    (* It is an infeasible (return) path when unsatisfied. *)
	    let truecubes = (List.map snd cubes) in
	    let evalform = (pseudo_conjoin
			       (cube_list_to_form truecubes)
			       statements) in
	    (trace_print_infeasible_path printlabel evalform) ;
	    (eval_path_return gctx fctx b)
	  end)
	else
	  (begin
	    let states : markedcube_t list =
	      (List.map
		  (fun ((mark : string), (cube : cube_t)) ->
		    let cube1 =
		      (substitute_variables_in_cube voutmap
			  cube
			  (*GOMI*)(*(filter_cube_by_predicates
			    true abstpreds cube)*)) in
		    (mark, cube1))
		  feasibles) in
	    let globalstates : markedcube_t list =
	      (recover_split_states fctx.fctx_current_slices states) in

	    fctx.fctx_result_cubes_collecting <-
	      (fctx.fctx_result_cubes_collecting @ globalstates) ;
	    (eval_path_return gctx fctx b)
	  end)
      end)
  | _ -> (assert false))

(* Returns from a function at a tail block. *)

and eval_path_return (gctx : program_context_t) (fctx : function_context_t) (b : form_block_t) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  if (fctx.fctx_remaining_contexts != []) then
    (eval_path_resume_to_choices gctx fctx)
  else
    (begin
      match b with
	{
	  bb_label = label ;
	  bb_memo = memo ;
	  bb_outgoing_labels = _ ;
	  bb_incoming_labels = _ ;
	  bb_inputs = _ ;
	  bb_outputs = _ ;
	  bb_body = BBTAIL {
	    bbtail_globals_outmap = gvarmap
	  }
	} ->
	  (begin
	    let newcubes : markedcube_t list =
	      (markedcube_list_union
		  fctx.fctx_result_cubes
		  fctx.fctx_result_cubes_collecting) in
	    let uncovered =
	      ((List.length newcubes)
		> (List.length fctx.fctx_result_cubes)) in
	    fctx.fctx_result_cubes <- newcubes ;

	    (* Check memoized state of a function call. *)

	    let memo : function_memoizer_t =
	      (get_memoizer gctx (fn, fdef.fundef_hash_name)) in
	    let recurse : bool option = (memo.fmem_recurse) in
	    if ((not uncovered) || (structural_equal recurse (Some false))) then
	      (begin
		(* Covered (recursion meets fixpoint), or no recursion. *)
		(aiTRACEPRINT (printlabel ^ " function return:")) ;
		memo.fmem_recurse <- None ;
		(*SLOW*)
		(List.iter
		    (fun ((mark0 : string), (call : cube_t)) ->
		      let results : cube_t list =
			(Misc.list_filtering_map
			    (fun ((mark1 : string), (result : cube_t)) ->
			      if (string_equal mark0 mark1) then
				(Some result)
			      else
				None)
			    fctx.fctx_result_cubes) in
		      (memoize_values memo.fmem_memoizer call results))
		    fctx.fctx_calling_global_states) ;
		(eval_path_continue gctx fctx.fctx_return_context
		    fctx.fctx_result_cubes)
	      end)
	    else
	      (begin
		(* Uncovered with recursion, eval it again for fixpoint. *)
		(aiTRACEPRINT (printlabel ^ " call resuming:")) ;
		assert (structural_equal recurse (Some true)) ;
		let b2 = (Hashtbl.find bbset pseudo_entry_label) in
		match b2 with
		  {
		    bb_label = label ;
		    bb_outgoing_labels = [outgoing] ;
		    bb_incoming_labels = _ ;
		    bb_body = (BBHEAD _)
		  } ->
		    (* Make new function context. *)
		    let newfctx = {
		      fctx with
			fctx_junction_covers = (make_empty_covering bbset) ;
			fctx_result_cubes_collecting = [] ;
			fctx_remaining_contexts = []
		    } in
		    (eval_path_function gctx newfctx outgoing label
			(fctx.fctx_calling_cubes))
		| _ -> (assert false)
	      end)
	  end)
      | _ -> (assert false)
    end))

(* Follows more choices in conditionals. *)

and eval_path_resume_to_choices (gctx : program_context_t) (fctx0 : function_context_t) : unit = (
  let (cc : eval_next_t) = (List.hd fctx0.fctx_remaining_contexts) in
  fctx0.fctx_remaining_contexts <- (List.tl fctx0.fctx_remaining_contexts) ;
  match cc with
    EVALSTOP -> (assert false)
  | (EVALCONT { ectx_current_slices = None } ) -> (assert false)
  | (EVALCONT {
      ectx_fctx = fctx ;
      ectx_current_slices = (Some dropped) ;
      ectx_label = label ;
      ectx_source_label = sourcelabel ;
      ectx_flowform = flowform
    }) ->
      (begin
	(* NOTE: USE IDENTITY-EQUAL BECUASE {fctx} INCLUDES RECURSIVE DATA. *)
	assert (fctx == fctx0) ;
	(begin
	  let fdef : fundef_t = fctx.fctx_fdef in
	  let decl : Cexpr.declaration_t = fdef.fundef_declaration in
	  let (fn, _) = (decl.Cexpr.decl_id) in
	  let printlabel = ("(" ^ fn ^ "/" ^ sourcelabel ^ ")") in
	  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
	  let b = (Hashtbl.find bbset sourcelabel) in
	  (*let memo = b.bb_memo in*)
	  (aiTRACEPRINT
	      (printlabel ^ " taking path cond-else near " ^ (*memo*)"" ^ ":")) ;
	  (aiTRACEPRINT_flowing_state flowform
	      dropped)
	end) ;
	fctx.fctx_current_slices <- dropped ;
	(eval_path gctx fctx label sourcelabel flowform)
      end))

(* Continues caller at return. *)

and eval_path_continue (gctx : program_context_t) (cc : eval_next_t) (globalstates : markedcube_t list) : unit = (
  match cc with
    EVALSTOP ->
      (info "* DONE")
  | (EVALCONT { ectx_current_slices = Some _ } ) -> (assert false)
  | (EVALCONT {
      ectx_fctx = savedfctx ;
      ectx_current_slices = None ;
      ectx_label = label ;
      ectx_source_label = sourcelabel ;
      ectx_flowform = _
    }) ->
      let fdef : fundef_t = savedfctx.fctx_fdef in
      let fn : string = (funcname fdef) in
      let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
      let b : form_block_t = (Hashtbl.find bbset label) in
      let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
      match b with
	{
	  bb_label = label0 ;
	  bb_memo = memo ;
	  bb_outgoing_labels = [outgoing] ;
	  bb_incoming_labels = [incoming] ;
	  bb_body = BBCALL {
	    bbcall_fn = _ ;
	    bbcall_call = _ ;
	    bbcall_global_reversions = callfix ;
	    bbcall_globals_inmap = vinmap ;
	    bbcall_globals_outmap = _ ;
	    bbcall_predicates = (localpreds : form_t list)
	  }
	} ->
	  (begin
	    let globalpreds : form_t list =
	      (filter_global_predicates (Misc.hashtbl_keys vinmap)
		  gctx.ctx_afile.afile_global_predicates) in
	    let cubes : markedcube_t list =
	      (List.map
		  (fun ((mark : string), (cube : cube_t)) ->
		    let cube1 : cube_t =
		      (substitute_variables_in_cube vinmap cube) in
		    (mark, cube1))
		  globalstates) in

	    let abstpreds : form_t list =
	      (uniqify_form_list (globalpreds @ localpreds)) in
	    let (states, dropped) =
	      (split_and_mark_states abstpreds cubes) in
	    let newstates = (add_equaters_of_globals callfix states) in

	    (aiTRACEPRINT
		(printlabel ^ " continue near " ^ (*memo*)"" ^ ":"
		  ^ "\n;; - results="
		  ^ (markedcube_list_to_string globalstates)
		  ^ "\n;; - states="
		  ^ (markedcube_list_to_string newstates))) ;

	    savedfctx.fctx_current_slices <- dropped ;
	    (eval_path gctx savedfctx outgoing label (newstates, TT))
	  end)
      | _ -> (assert false))

and check_at_assertion (gctx : program_context_t) (fctx : function_context_t) (decl : Cexpr.declaration_t) (bbset : (string, form_block_t) Hashtbl.t) (b : form_block_t) (f0 : form_t) (f1 : form_t) : bool = (
  let (fn, _) = (decl.Cexpr.decl_id) in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  let ff = (Form.pseudo_conjoin f0 (NOT f1)) in
  (aiTRACEPRINT
      (printlabel ^ " checking assertion as unsatisfiable:"
	^ "\n;; - form=" ^ (form_to_string ff))) ;
  let ((r : Sat.sat_result_t), (cube : form_t list)) = (smt_solve ff) in
  match r with
    Sat.SAT ->
      (aiTRACEPRINT
	  (printlabel ^ " ASSERTION FAILED" ^ " near " ^ (*b.bb_memo*)""
	    ^ " with cube: " ^ (form_to_string (AND cube)))) ;
      (print_backtrace fctx) ;
      (failwith "ASSERTION FAILED")
  | Sat.UNSAT ->
      true
  | Sat.UNKNOWN ->
      (failwith "ASSERTION FAILED (UNKNOWN)"))

and check_at_junction (gctx : program_context_t) (fctx : function_context_t) (b : form_block_t) (states : markedcube_t list) (constraints : form_t) : unit = (
  let fdef : fundef_t = fctx.fctx_fdef in
  let fn : string = (funcname fdef) in
  let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
  let printlabel = ("(" ^ fn ^ "/" ^ b.bb_label ^ ")") in
  match b with
    {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = [outgoing] ;
      bb_incoming_labels = incomings ;
      bb_inputs = _ ;
      bb_outputs = _ ;
      bb_body = BBJUNC {
	bbjunc_phis = phis ;
	bbjunc_phifix = phifix ;
	bbjunc_predicates = (localpreds : form_t list) ;
	bbjunc_predmap = predmap
      }
    } ->
      (begin
	let abstpreds = (localpreds) in
	let feasibles : markedcube_t list =
	  (solve_feasible_cubes states constraints abstpreds
	      printlabel "check states at junction") in

	if (feasibles == []) then
	  (begin
	    (* It is an infeasible path when unsatisfied. *)
	    let truecubes = (List.map snd states) in
	    let evalform = (pseudo_conjoin
			       (cube_list_to_form truecubes)
			       constraints) in
	    (trace_print_infeasible_path printlabel evalform) ;
	    let tailb = (Hashtbl.find bbset pseudo_return_label) in
	    (eval_path_return gctx fctx tailb)
	  end)
	else
	  (begin
	    let statecubes : markedcube_t list =
	      (uniqify_markedcube_list
		  (List.map
		      (fun ((mark : string), (cube : cube_t)) ->
			let cube1 =
			  (substitute_predicates predmap
			      cube
			      (*GOMI*)(*(filter_cube_by_predicates
				true abstpreds cube)*)) in
			(mark, cube1))
		      feasibles)) in

	    let oldcover : markedcube_t list =
	      (Hashtbl.find fctx.fctx_junction_covers label) in
	    let uncovereds : markedcube_t list =
	      (check_covered fctx.fctx_current_slices
		  statecubes oldcover) in

	    (aiTRACEPRINT
		(printlabel ^ " junction near " ^ (*memo*)"" ^ ":"
		  ^ "\n;; - states="
		  ^ (markedcube_list_to_string statecubes)
		  ^ "\n;; - predicates="
		  ^ (form_list_to_string
			(List.map (substitute_predicates_in_form predmap)
			    abstpreds))
		  ^ (ite (uncovereds == [])
			("\n;; - junction covered")
			("\n;; - junction uncovered; uncovereds="
			  ^ (markedcube_list_to_string uncovereds))))) ;

	    if (uncovereds == []) then
	      (begin
		(* Covered: Do nothing. *)
		let tailb = (Hashtbl.find bbset pseudo_return_label) in
		(eval_path_return gctx fctx tailb)
	      end)
	    else
	      (begin
		let oldmarkedstatecubes : markedcube_t list =
		  (remark_to_original fctx.fctx_current_slices statecubes) in
		let newcover : markedcube_t list =
		  (markedcube_list_union oldmarkedstatecubes oldcover) in
		(Hashtbl.replace fctx.fctx_junction_covers label newcover) ;
		(eval_path gctx fctx outgoing label (uncovereds, TT))
	      end)
	  end)
      end)
  | _ -> (assert false))

and print_backtrace (fctx : function_context_t) : unit = (
  let cc : eval_next_t = fctx.fctx_return_context in
  match cc with
    EVALSTOP ->
      ()
  | EVALCONT {
      ectx_fctx = (nextfctx : function_context_t) ;
      ectx_current_slices = _ ;
      ectx_label = (label : string) ;
      ectx_source_label = _ ;
      ectx_flowform = (flowform : flowin_state_t)
    } ->
      (begin
	let fdef : fundef_t = nextfctx.fctx_fdef in
	let fn : string = (funcname fdef) in
	(*let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in*)
	let (cubes, statements) = flowform in
	(debug
	    ("backtrace: " ^ fn ^ "/" ^ label
	      ^ " states=" ^ (markedcube_list_to_string cubes)
	      ^ " constraints=" ^ (form_to_string statements))) ;
	(print_backtrace nextfctx)
      end))

(* Returns states uncovered in old states {cover}.  It only checks
   covering condition by subset relation (ie, not implication).  NOTE:
   It distinguishes marker difference, because different markers
   correspond to different calling states and covering are
   distinguished. *)

and check_covered (dropped : (string, markedcube_t list) Hashtbl.t) (statecubes : markedcube_t list) (cover : markedcube_t list) : markedcube_t list = (
  let coverset : MarkedCubeSet.t =
    (List.fold_left
	(fun ss c -> (MarkedCubeSet.add c ss))
	(MarkedCubeSet.empty)
	cover) in
  let markmap : (string, string list) Hashtbl.t = (Hashtbl.create 10) in
  (Hashtbl.iter
      (fun (mark0 : string) (cubes : markedcube_t list) ->
	let oldmarks : string list = (List.map fst cubes) in
	(Hashtbl.replace markmap mark0 oldmarks))
      dropped) ;
  (List.filter
      (fun ((mark0 : string), (cube : cube_t)) ->
	let oldmarks : string list = (Hashtbl.find markmap mark0) in
	(List.exists
	    (fun m ->
	      (not (MarkedCubeSet.mem (m, cube) coverset)))
	    oldmarks))
      statecubes))

and remark_to_original (dropped : (string, markedcube_t list) Hashtbl.t) (statecubes : markedcube_t list) : markedcube_t list = (
  let markmap : (string, string list) Hashtbl.t = (Hashtbl.create 10) in
  (Hashtbl.iter
      (fun (mark0 : string) (cubes : markedcube_t list) ->
	let oldmarks : string list = (List.map fst cubes) in
	(Hashtbl.replace markmap mark0 oldmarks))
      dropped) ;
  (List.flatten
      (List.map
	  (fun ((mark0 : string), (cube : cube_t)) ->
	    let oldmarks : string list = (Hashtbl.find markmap mark0) in
	    (List.map (fun m -> (m, cube)) oldmarks))
	  statecubes)))

(* Makes assignments of globals.  It filters globals which appear in
   the modified list of a call.  {globalsfix} is a list of globals (of
   the same globals with different version numbers) which are equated
   when the globals are not in the modified list.  {globalstates} is a
   list of globals modified in a function call.  NOTE: It accepts that
   {globalstates} includes {TT} as a special case (for
   trivial-sat). *)

and add_equaters_of_globals (globalsfix : (term_t * term_t) list) (globalstates : markedcube_t list) : markedcube_t list = (
  let globalset : StringSet.t =
    (List.fold_left
	(fun ss (t0, t1) ->
	  match (t0, t1) with
	    ((VAR s0), (VAR s1)) ->
	      (StringSet.add s1 ss)
	  | _ -> (assert false))
	(StringSet.empty) globalsfix) in
  (List.map
      (fun (cube : markedcube_t) ->
	(add_equaters_of_globals_to_cube globalsfix globalset cube))
      globalstates))

and add_equaters_of_globals_to_cube (globalsfix : (term_t * term_t) list) (globalset : StringSet.t) (cube : markedcube_t) : markedcube_t = (
  let ((mark : string), (truecube : cube_t)) = cube in
  let modifiedset : StringSet.t =
    (list_variables_in_cube truecube (StringSet.empty)) in
  let modifiedglobalset : StringSet.t =
    (StringSet.inter modifiedset globalset) in
  let equaters : form_t list =
    (Misc.list_filtering_map
	(fun (t0, t1) ->
	  match (t0, t1) with
	    ((VAR s0), (VAR s1)) ->
	      if (StringSet.mem s1 modifiedglobalset) then
		None
	      else
		(Some (PRED (EQ, (VAR s0), (VAR s1))))
	  | _ -> (assert false))
	globalsfix) in
  (mark, (cube_append truecube (literal_list_to_cube equaters))))

and match_arguments (pars : string list) (args : term_t list) : form_t = (
  (match_arguments_loop pars args []))

and match_arguments_loop (pars : string list) (args : term_t list) (l : form_t list): form_t = (
  match (pars, args) with
    ([], []) ->
      (AND l)
  | ((p :: pl), (a :: al)) ->
      if (not (Cexpr.ellipsis_parameter_p p)) then
	(match_arguments_loop pl al (l @ [(PRED (EQ, (VAR p), a))]))
      else
	(begin
	  assert (pl == []) ;
	  let varargs =
	    (List.map2
		(fun i v ->
		  let x = (VAR (Csyntax.pseudo_varargs ^ (string_of_int i))) in
		  (PRED (EQ, x, v)))
		(Misc.list_int_sequence (List.length args))
		args) in
	  (AND (l @ varargs))
	end)
  | _ -> (assert false))

(* ================================================================ *)

(* Collector of definitions in a file scope. *)

let parsed_definitions : ((string, definition_t) Hashtbl.t) option ref =
  (ref None)

(* Unique name for current file.  It is appended to static variable
   names. *)

let file_hash_name : string ref = ref ""

(* Constructs a definition block.  This is a parser callback. *)
(*****************************************************************************)
let make_assert_true () = (
    let renamed_assert, level =
        try Hashtbl.find !Csyntax.file_scope_static_variable_name_map
	                 ("artificial_assert", 1)
        with
        | Not_found -> "artificial_assert", 1
    in
    Cexpr.STMT (Cexpr.CALL (Cexpr.INTEGER,
                 Cexpr.OPER (Cexpr.POINTER_TO (
                                 Cexpr.FUNCTION_OF (Cexpr.INTEGER,None)),
                             Cexpr.CAST (false),
                             Cexpr.VREF (
                                 Cexpr.FUNCTION_OF (Cexpr.INTEGER,None),
                                 renamed_assert,
                                 level,0),
                             Cexpr.NIL,Cexpr.NIL),
                  [ Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (1), "1")) ],
                None));

)

(* Check whether an expression contains a pointer dereference *)
let rec has_pointer_dereference_in_expr (expr : expr_t) : bool =
begin
    match expr with
    | Cexpr.NIL                   -> false
    | Cexpr.PSEUDO_NONDET_VALUE _ -> false
    | Cexpr.OPER (_, _, e1, e2, e3) ->
          has_pointer_dereference_in_exprs [e1; e2; e3]
    | Cexpr.SETF (_, _, e1, e2)   -> has_pointer_dereference_in_exprs [e1; e2]
    | Cexpr.CALL (_, e, es, _)       -> has_pointer_dereference_in_exprs (e :: es)
    | Cexpr.VREF (_, _, _, _)   -> false
    | Cexpr.PREF _                -> true
    | Cexpr.SREF (typ, _, e, _)   ->
          if (Cexpr.pointer_type_p typ) then
              true
          else
              has_pointer_dereference_in_expr e;
    | Cexpr.DSGN_AREF (typ, e1, e2) -> has_pointer_dereference_in_exprs [e1; e2]
    | Cexpr.ADDROF (typ, e)         -> has_pointer_dereference_in_expr e;
    | Cexpr.SIZEOF _                -> false
    | Cexpr.PROGN _                 -> false
    | Cexpr.CONST_I _               -> false
    | Cexpr.CONST_F _               -> false
    | Cexpr.STRING _                -> false
    | Cexpr.CLITERAL (_, init)  -> has_pointer_dereference_in_initializers init;
    | Cexpr.ASMSEQUENCE es          -> has_pointer_dereference_in_exprs es
    | Cexpr.PHI (typ, str_e_list)   ->
          if (Cexpr.pointer_type_p typ) then
              true
          else
          begin
              let (strs, es) = List.split str_e_list in
              has_pointer_dereference_in_exprs es
          end
    | Cexpr.PSEUDO_MEMORY _       -> false
    | Cexpr.INIT_FILL _           -> false
    | Cexpr.PRAGMA _              -> false
    | Cexpr.QUANTIFIER (_, es, e) -> has_pointer_dereference_in_exprs (es @ [e])
    | Cexpr.PSEUDO_PROLOGUE       -> false
    | Cexpr.PSEUDO_EPILOGUE       -> false
    | Cexpr.PSEUDO_ACSL_RESULT _  -> assert false
    | Cexpr.PSEUDO_ACSL_OLD _     -> assert false
    | Cexpr.ALL_FIELD _           -> assert false
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK _
    | Cexpr.PSEUDO_ACSL_UNIQUE _  -> assert false
    | Cexpr.PSEUDO_ACSL_VALID _   -> assert false
    (*| _                         -> false*)
end

and has_pointer_dereference_in_exprs (exprs : expr_t list) : bool =
begin
    List.exists (
        fun expr -> has_pointer_dereference_in_expr expr;
    ) exprs
end

(* Check whether initializers contain a pointer dereference *)
and has_pointer_dereference_in_initializers (init : Cexpr.initializers_t)
                                            : bool =
begin
    match init with
   | Cexpr.INIT_EXPR e -> has_pointer_dereference_in_expr e
   | Cexpr.INIT_LIST inits ->
         List.exists (
             fun ini -> has_pointer_dereference_in_initializers ini;
         ) inits;
   | Cexpr.INIT_DSGN (es, ini) -> (
         has_pointer_dereference_in_exprs es
         or has_pointer_dereference_in_initializers ini;)
end

(* Insert auxiliary assert(true) after stmts which:
       + dereference pointers
       + call to a function pointer *)
let rec insert_auxiliary_asserts_to_stmts (stmts : Cexpr.stmt_t list)
                                          : Csyntax.stmt_t list =
begin
    List.flatten(
        List.map insert_auxiliary_asserts_to_stmt stmts
    )
end

and insert_auxiliary_asserts_to_stmt (stmt : Cexpr.stmt_t)
                                     : Csyntax.stmt_t list =
begin
    let create_assert_true_stmts () : stmt_t list =
    begin
        let returnlabel = (Cexpr.make_new_label ()) in
        let tn = Cexpr.NOTYPE in
        let ctx : Csyntax.flatten_context_t = {
            Csyntax.ctx_file_hash_name = !file_hash_name ;
            Csyntax.ctx_function_name = "" ;
            Csyntax.ctx_return_type = tn ;
            Csyntax.ctx_return_label = returnlabel ;
            Csyntax.ctx_break_label = None ;
            Csyntax.ctx_continue_label = None ;
            Csyntax.ctx_static_variable_name_map = Hashtbl.create 10;
            (*(Hashtbl.copy !file_scope_static_variable_name_map) ;*)
            Csyntax.ctx_case_label_map = [] ;
            Csyntax.ctx_static_variables = [] ;
	    Csyntax.ctx_referred_labels = []
        } in
        Csyntax.flatten_stmt_body ctx (make_assert_true ())
    end in
    match stmt with
    | Cexpr.DEFUN (header, body) -> (
          let new_body = (
              match (insert_auxiliary_asserts_to_stmt body) with
              | [Cexpr.BEGIN (stmts)] -> Cexpr.BEGIN (stmts)
              | _ -> assert false
          ) in
          [Cexpr.DEFUN(header, new_body)]
      )
    | Cexpr.ASGN (expr1, expr2) -> (
          if !intensive_pointer_checking then
          begin
             (* insert an assert(true) after a pointer dereference *)
             if (has_pointer_dereference_in_exprs [expr1; expr2]) then
             begin
                 let assert_true_stmts = create_assert_true_stmts () in
                 [stmt] @ assert_true_stmts
             end
             else
                 [stmt]
          end
          else
              [stmt]
      )
    | Cexpr.CALLASGN (Cexpr.CALL (_, expr, _, _), _, _) -> (
          let ty = (Cexpr.type_of_expr expr) in
          (* insert an assert(true) after a call to a function pointer *)
          if (Ccommon.is_function_pointer ty expr) then
          begin
              let assert_true_stmts = create_assert_true_stmts () in
              [stmt] @ assert_true_stmts
          end
          else
              [stmt]
      )
    | Cexpr.BEGIN stmts ->
          [Cexpr.BEGIN(insert_auxiliary_asserts_to_stmts stmts)]
    | Cexpr.DEFVAR _
    | Cexpr.STMT _
    | Cexpr.IF _
    | Cexpr.GOTO _
    | Cexpr.LABEL _
    | Cexpr.NOTE _
    (*| Cexpr.DEFVAR _*)
    | Cexpr.ASSUME _
    | Cexpr.RESET _
    | Cexpr.ANNOTATION _ -> [stmt]
    | _ -> ((Cexpr.dump_stmts [stmt]);assert false)
end

(******************************************************************************)

let rec translate_declaration (decl : Cexpr.stmt_t) : unit = (
  (*let decl2 = (Csyntax.presyntax_statement decl) in*)
  if (!Csyntax.bbTRACE_dump_parsed) then
    (Cexpr.dump_stmts [decl]) ;
  let ((defs : Cexpr.stmt_t list), (referredlabels : (string * string) list)) =
    (Csyntax.flatten_syntax !file_hash_name decl) in
  (* BinhLT *)
  (*print_endline "before";Cexpr.dump_stmts defs; Scommon.pause();*)
  let defs = (* 20090702 - BinhLT*)
      match (!Scommon.process_specs) with
      | true  -> insert_auxiliary_asserts_to_stmts defs
      | false -> defs
  in
  (*print_endline "after";Cexpr.dump_stmts defs; Scommon.pause();*)
(*   let defs =
       match (!Scommon.process_specs) with
       | true ->
           (Sinstrument.insert_assumes ((* Binh.LT 22/05/2009
               Sinstrument.remove_function_pointer_called_stmts*) defs))
       | false -> defs
   in
*)
  (* BinhLT *)
  (List.iter
      (fun (def0 : Cexpr.stmt_t) ->
	if (!Csyntax.bbTRACE_dump_flatten) then
	  (Cexpr.dump_stmts [def0]) ;
	let def1 = (Csyntax.breakup_blocks def0) in
	if (!Csyntax.bbTRACE_dump_breakup) then
	  (Csyntax.dump_stmt_blocks def1) ;
	(begin
	  match def1 with
	    Csyntax.CFUNDEF {
	      Csyntax.cfdef_declaration = decl ;
	      Csyntax.cfdef_blocks = bbset
	    } ->
	      let (id, level) = (decl.Cexpr.decl_id) in
	      (Csyntax.collapse_blocks bbset referredlabels) ;
	      if (!Csyntax.bbTRACE_dump_collapse) then
		(Csyntax.dump_stmt_blocks def1) ;
	      let vartypes : ((string * int), Cexpr.typenode_t) Hashtbl.t =
		(Hashtbl.create 10) in
	      (Csyntax.find_variable_usage bbset vartypes) ;
	      (Csyntax.propagate_variable_usage bbset) ;
	      if (!Csyntax.bbTRACE_dump_variables) then
		(Csyntax.dump_stmt_blocks def1) ;
	      (Csyntax.bind_new_variables bbset decl vartypes) ;
	      (Csyntax.fixate_minor_on_blocks bbset) ;
	      if (!Csyntax.bbTRACE_dump_ssa_binding) then
		(Csyntax.dump_stmt_blocks def1) ;
        (* Binh - mat thong tin *)
	      let def2 = (translate_definition !file_hash_name def1) in
	      if (!aiTRACE_dump_form) then
		(dump_form_blocks def2) ;
	      (*(Hashtbl.replace !parsed_definitions id def2)*)
	      (record_definition id def2)
	  | Csyntax.CVARDEF {
	      Csyntax.cvdef_declaration = decl
	    } ->
	      let (id, _) = (decl.Cexpr.decl_id) in
	      let def2 = (translate_definition !file_hash_name def1) in
	      if (!aiTRACE_dump_form) then
		(dump_form_blocks def2) ;
	      (*(Hashtbl.replace !parsed_definitions id def2)*)
	      (record_definition id def2)
	  | _ ->
	      ()
	end))
      defs))

(* Mostly emulates the function translate_declaration
   (except for skipping the call to function Csyntax.bind_new_variables)
   Used to translate functional specs attached in declaration_t
   into corresponding blocks *)
and translate_specs (decl : Cexpr.stmt_t) : unit = (
  let ((defs : Cexpr.stmt_t list), (referredlabels : (string * string) list)) =
    (Csyntax.flatten_syntax !file_hash_name decl) in
  (List.iter
      (fun (def0 : Cexpr.stmt_t) ->
	let def1 = (Csyntax.breakup_blocks def0) in
	(begin
	  match def1 with
	    Csyntax.CFUNDEF {
	      Csyntax.cfdef_declaration = decl ;
	      Csyntax.cfdef_blocks = bbset
	    } ->
	      let (id, level) = (decl.Cexpr.decl_id) in
	      (Csyntax.collapse_blocks bbset referredlabels) ;
	      let vartypes : ((string * int), Cexpr.typenode_t) Hashtbl.t =
		(Hashtbl.create 10) in
	      (Csyntax.find_variable_usage bbset vartypes) ;
	      (Csyntax.propagate_variable_usage bbset) ;
	      (Csyntax.fixate_minor_on_blocks bbset) ;
	      let def2 = (translate_definition !file_hash_name def1) in
	      (record_definition id def2)
	  | _ ->
              (assert false)
	end))
      defs))

(* MEMO: {fundef_filename} is filled later (and it is set empty). *)

and translate_definition (hashname : string) (def : Csyntax.cdefinition_t) : definition_t = (
  match def with
    (Csyntax.CFUNDEF fundef) ->
      let decl = fundef.Csyntax.cfdef_declaration in
      let fn : string = (fst (decl.Cexpr.decl_id)) in
      let oldbbset = (fundef.Csyntax.cfdef_blocks) in
      let predsets : (string, expr_t list) Hashtbl.t = (Hashtbl.create 10) in
      (scan_abstraction_predicates fn oldbbset predsets) ;
      let bbset : (string, form_block_t) Hashtbl.t = (Hashtbl.create 10) in
      (translate_blocks bbset oldbbset decl predsets) ;
      let globalpreds =
	(List.map
	    translate_to_form
	    (Hashtbl.find predsets pseudo_elsewhere_label)) in
      (FUNDEF {
	fundef_hash_name = hashname ;
	fundef_declaration = decl ;
	fundef_bbset = bbset ;
	fundef_global_preds = globalpreds
      })
  | (Csyntax.CVARDEF vardef) ->
      (VARDEF {
	vardef_declaration = vardef.Csyntax.cvdef_declaration ;
      })
  | (Csyntax.COTHERDEF s) ->
      (OTHERDEF s))

(* NOTE: No checks, because type checks are done in the parser. *)

and record_definition (id : string) (def : definition_t) : unit = (
  if (not (Hashtbl.mem (Misc.some_of !parsed_definitions) id)) then
    (Hashtbl.replace (Misc.some_of !parsed_definitions) id def)
  else
    (begin
      let olddef = (Hashtbl.find (Misc.some_of !parsed_definitions) id) in
      match (olddef, def) with
	((FUNDEF _), (FUNDEF d1)) ->
	  if (Cattr.attribute_check_in_declaration "$force_link"
		 d1.fundef_declaration) then
	    (begin
	      (Cexpr.warn_in_parse ("Overriding " ^ id)) ;
	      (Hashtbl.replace (Misc.some_of !parsed_definitions) id def)
	    end)
	  else
	    (Cexpr.syntaxerror ("Duplicate defintion '" ^ id ^ "'"))
      | ((FUNDEF _), (VARDEF _)) ->
	  ()
      | ((FUNDEF _), (OTHERDEF _)) ->
	  ()
      | ((VARDEF _), (FUNDEF _)) ->
	  (Hashtbl.replace (Misc.some_of !parsed_definitions) id def)
      | ((VARDEF _), (VARDEF _)) ->
	  ()
      | ((VARDEF _), (OTHERDEF _)) ->
	  ()
      | ((OTHERDEF _), (FUNDEF _)) ->
	  (Hashtbl.replace (Misc.some_of !parsed_definitions) id def)
      | ((OTHERDEF _), (VARDEF _)) ->
	  (Hashtbl.replace (Misc.some_of !parsed_definitions) id def)
      | ((OTHERDEF _), (OTHERDEF _)) ->
	  ()
    end))

(* ================================================================ *)

(* TOPLEVELS OF TRANSLATER/INTERPRETER *)

(* Parses and translates a source file.  {filename} is of a file
   output from preprocessor, while {truefilename} is the one
   passed. *)

let parse_source_file (filename : string) (truefilename : string) : ofile_t = (
  let name = (Filename.chop_extension (Filename.basename truefilename)) in
  assert ((String.length name) > 0) ;
  let hash = (Printf.sprintf "%04x" ((Random.bits ()) land 0xFFFF)) in
  if (!aiTRACE_linking) then
    (debug (";; READING FILE '" ^ truefilename ^ "'...")) ;
  let defs : (string, definition_t) Hashtbl.t = (Hashtbl.create 10) in
  file_hash_name := (name ^ "%" ^ hash) ;
  parsed_definitions := (Some defs) ;
  Cexpr.file_typebody_table := (Some (Hashtbl.create 40)) ;
  Csyntax.file_scope_static_variable_name_map := (Hashtbl.create 10) ;
  (Clex.cparse filename) ;
  if (!aiTRACE_linking) then
    (debug (";; READING FILE '" ^ truefilename ^ "' DONE")) ;
  (*GOMI*)
  (* Add file name in the FUNDEF. *)
  (*
  (List.iter
      (fun name ->
	let def = (Hashtbl.find defs name) in
	match def with
	  (FUNDEF d) ->
	    (Hashtbl.replace defs name
		(FUNDEF {d with fundef_hash_name = filename }))
	| (VARDEF d) ->
	    ()
	| (OTHERDEF s) ->
	    ())
      (Misc.hashtbl_keys defs)) ;
  *)
  let globalpreds =
    (uniqify_form_list
	(Hashtbl.fold
	    (fun n def tl ->
	      match def with
		(FUNDEF d) ->
		  (d.fundef_global_preds @ tl)
	      | (VARDEF d) ->
		  tl
	      | (OTHERDEF s) ->
		  tl)
	    defs [])) in
  let o : ofile_t = {
    ofile_hash_name = !file_hash_name ;
    ofile_file_name = truefilename ;
    ofile_defs = defs ;
    ofile_global_predicates = globalpreds
  } in
  o)

(* Gathers definitions of files and makes an archive. *)

let package_definitions (afile : afile_t option) (ofiles : ofile_t list) : afile_t = (
  let abstpreds : (string, form_t list) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).afile_global_predicates
    else
      (Hashtbl.create 10) in
  let libs : (string, ofile_t) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).afile_libs
    else
      (Hashtbl.create 10) in
  let directory : (string, string) Hashtbl.t =
    if (afile != None) then
      (Misc.some_of afile).afile_directory
    else
      (Hashtbl.create 10) in
  (* Collect abstraction predicates. *)
  let globalpreds =
    (uniqify_form_list
	(List.flatten
	    (List.map (fun o -> o.ofile_global_predicates) ofiles))) in
  (List.iter
      (fun p ->
	let vs : StringSet.t = (find_free_variables (StringSet.empty) p) in
	(StringSet.iter
	    (fun (v : string) ->
	      (Hashtbl.replace abstpreds v
		  (if (Hashtbl.mem abstpreds v) then
		    ([p] @ (Hashtbl.find abstpreds v))
		    else
		      [p])))
	    vs))
      globalpreds) ;
  if (!aiTRACE_linking) then
    (debug (";; global.predicates="
		     ^ (form_list_to_string globalpreds))) ;
  (* Put files into archive. *)
  (List.iter
      (fun (o : ofile_t) ->
	let filename = o.ofile_file_name in
	if (!aiTRACE_linking) then
	  (debug (";; LINKING FILE '" ^ filename ^ "'...")) ;
	let hashname = o.ofile_hash_name in
	(Hashtbl.iter
	    (fun (id : string) (def : definition_t) ->
	      match def with
		FUNDEF {
		  fundef_declaration = (decl : Cexpr.declaration_t)
		} ->
		  (begin
		    if (!aiTRACE_linking) then
		      (debug (";; looking at symbol(f)="
				       ^ (Cexpr.decl_to_simple_string decl))) ;
		    match decl.Cexpr.decl_scls with
		      Cexpr.NILSCLS | Cexpr.EXTERN ->
			if (Hashtbl.mem directory id) then
			  (failwith ("multiply defined: " ^ id)) ;
			(Hashtbl.replace directory id hashname)
		    | _ ->
			()
		  end)
	      | VARDEF {
		  vardef_declaration = (decl : Cexpr.declaration_t)
		} ->
		  (* MEMO: Empty name is for structure definition. *)
		  if (not (string_equal id "")) then
		    (begin
		      if (!aiTRACE_linking) then
			(debug (";; looking at symbol(v)="
					 ^ (Cexpr.decl_to_simple_string decl))) ;
		      match decl.Cexpr.decl_scls with
			Cexpr.NILSCLS ->
			  if (Hashtbl.mem directory id) then
			    (failwith ("multiply defined: " ^ id))
			  else
			    (Hashtbl.replace directory id hashname)
		      | Cexpr.EXTERN ->
			  (* No checks, just ignore. *)
			  ()
		      | Cexpr.AUTO | Cexpr.REGISTER ->
			  (failwith ("Bad storage class for: " ^ id))
		      | Cexpr.STATIC ->
			  (* No checks, just ignore. *)
			  ()
		      | Cexpr.TYPEDEF ->
			  (* No checks, just ignore. *)
			  ()
		    end)
	      | (OTHERDEF (stmt : stmt_t)) ->
		  ())
	    o.ofile_defs) ;
	if (!aiTRACE_linking) then
	  (debug (";; LINKING FILE '" ^ filename ^ "' DONE")) ;
	if (Hashtbl.mem libs hashname) then
	  (failwith ("o files with conflicting random names: " ^ hashname)) ;
	(Hashtbl.replace libs hashname o))
      ofiles) ;
  let a : afile_t = {
    afile_start = "" ;
    afile_libs = libs ;
    afile_directory = directory ;
    afile_global_predicates = abstpreds
  } in
  a)

let rec interpret_program (gctx : program_context_t) : unit = (
  let afile = gctx.ctx_afile in
  match afile with
    {
      afile_start = (start : string) ;
      afile_libs = (libs : (string, ofile_t) Hashtbl.t) ;
      afile_directory = (directory : (string, string) Hashtbl.t)
    } ->
      (begin
	if (not (Hashtbl.mem directory start)) then
	  (failwith ("Procedure '" ^ start ^ "' not defined")) ;
	let hashname = (Hashtbl.find directory start) in
	let o : ofile_t = (Hashtbl.find libs hashname) in
	match o with
	  {
	    ofile_file_name = (name : string) ;
	    ofile_defs = (defs : (string, definition_t) Hashtbl.t)
	  } ->
	    (begin
	      let def = (Hashtbl.find defs start) in
	      match def with
		(FUNDEF d) ->
		  (eval_from_start gctx def pseudo_entry_label TT)
	      | (VARDEF d) ->
		  (failwith "start symbol is a variable")
	      | (OTHERDEF s) ->
		  (failwith "start symbol is not a function")
	    end)
      end))

(* Fakes a function call (a pseudo toplevel). *)

and eval_from_start (gctx : program_context_t) (xdef : definition_t) (label : string) (f : form_t) : unit = (
  match xdef with
    (FUNDEF fdef) ->
      (begin
	let bbset : (string, form_block_t) Hashtbl.t = fdef.fundef_bbset in
	let b = (Hashtbl.find bbset pseudo_entry_label) in
	match b with
	  {
	    bb_label = l ;
	    bb_body = BBHEAD {
	      bbhead_params = params
	    }
	  } ->
	    (debug
		("* START from '" ^ (funcname fdef) ^ "'")) ;

	    assert (string_equal l pseudo_entry_label) ;
	    let args = (List.map (fun p -> (VAR ("#fake:" ^ p))) params) in
	    let callterms : term_t list = ([(VAR "main")] @ args) in
	    let mark = (xgensym "K") in
	    let states : markedcube_t = (mark, cube_true) in
	    (eval_path_head gctx
		fdef callterms [] [states] EVALSTOP)
	| _ -> (assert false)
      end)
  | (VARDEF vdef) ->
      ()
  | (OTHERDEF s) ->
      ())

let _ = (Cexpr.process_declaration := translate_declaration)
