(* parser/minisat.ml 2008-07-01 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* SAT Solver *)

(* This follows the "MiniSAT" paper "An Extensible SAT-solver
   [extended version 1.2]" by Niklas Een and Niklas Sorensson.  It
   includes major SAT techniques: conflict clause learning,
   backjumping or non-chronological backtracking, variable state
   independent decaying sum (VSIDS), two-watched literals, restarts,
   dynamic largest individual sum (DLIS) (???), ... *)

(* NOTE: (1) Variables are assigned with one-less number (zero-origin)
   of the CNF problem description (one-origin), when reading a problem
   from a file. *)

(* MEMO: (1) (invariant) Each clause appears twice in the watcher
   list.  (2) Ocaml arrays (in 32bit implementation) are limited to
   4,194,303 elements. *)

let xTRACE (s : string) = (if (false) then (print_endline (";; SMT " ^ s)))

type sat_result_t = SAT | UNSAT | UNKNOWN

(* Takes the body of {Some}. *)

let some_of v = (
  match v with
    Some e -> e
  | None -> raise (Invalid_argument "None is given to some_of."))

type searchparams_t = {
  params_var_decay : float ;
  params_cla_decay : float
}

(* Assignments, {LBottom} for unknown. *)

type lbool_t = LTrue | LFalse | LBottom

let lbool_of_bool (x : bool) = (
  if (x) then LTrue else LFalse)

let lbool_neg (x : lbool_t) = (
  if (x = LTrue) then LFalse else if (x = LFalse) then LTrue else LBottom)

(* Variables and literals.  Literals are encoded as integers, by
   concatenating the value of a variable with LSB as a sign.  Negative
   literals have sign=false and LSB=1.  MEMO: Boolean sign is negated
   from the one in the original paper. *)

type var_t = int
type lit_t = int

let null_literal = -1

let lit_make (v : var_t) (s : bool) = ((v lsl 1) + (if (s) then 0 else 1))

let lit_var (p : lit_t) : var_t = (p lsr 1)

let lit_neg (p : lit_t) : lit_t = (p lxor 1)

let lit_sign (p : lit_t) : bool = (not ((p land 1) = 1))

(*let lit_index (p : lit_t) : int = (p)*)

(* {vec} is an Array with a fill-pointer. *)

type 'a vec = {
  mutable vector_size : int ;
  vector_nullelem : 'a ;
  mutable v : 'a array
}

let vec_minimum_size = 64

let vec_size (v : 'a vec) = (v.vector_size)

let vec_create (size : int) (pad : 'a) = (
  let a = (Array.make (max size vec_minimum_size) pad) in
  {vector_size = size ; vector_nullelem = pad ; v = a})

let rec vec_pow2 x y = (if (x >= y) then x else (vec_pow2 (x * 2) y))

let vec_growto (v : 'a vec) (size : int) (pad : 'a) = (
  if ((Array.length v.v) < size) then
    (begin
      let s = (vec_pow2 vec_minimum_size size) in
      let a1 = (Array.make s pad) in
      (Array.blit v.v 0 a1 0 v.vector_size) ;
      v.v <- a1 ;
    end))

let vec_shrinkto (v : 'a vec) (n : int) = (
  assert (n <= v.vector_size) ;
  if (n < v.vector_size) then
    (begin
      for i = n to (v.vector_size - 1) do (
	v.v.(i) <- v.vector_nullelem
      ) done ;
      v.vector_size <- n
    end))

(* Clears a vector (resets entries for releasing). *)

let vec_clear (v : 'a vec) = (
  for i = 0 to (v.vector_size - 1) do (
    (v.v.(i) <- v.vector_nullelem)
  ) done ;
  v.vector_size <- 0)

(* Adds a new element to a vector. *)

let vec_push (v : 'a vec) (e : 'a) = (
  if ((Array.length v.v) < (v.vector_size + 1)) then
    (vec_growto v (v.vector_size + 1) e)
  else
    v.v.(v.vector_size) <- e ;
  v.vector_size <- (v.vector_size + 1))

(* Removes the last element. *)

let vec_pop (v : 'a vec) = (v.vector_size <- (v.vector_size - 1))

let vec_last (v : 'a vec) = (v.v.(v.vector_size - 1))

(* Same as vec_pop, but it returns the last entry. *)

let vec_last_pop (v : 'a vec) = (
  (v.vector_size <- (v.vector_size - 1)) ;
  (v.v.(v.vector_size)))

(* Swaps two entries of a vector. *)

let vec_swap (v : 'a vec) i j = (
  let t = v.v.(i) in
  v.v.(i) <- v.v.(j) ;
  v.v.(j) <- t)

(* Makes a vector by taking over the contents the argument.  The
   original vector be emptied. *)

let vec_steal (v : 'a vec) = (
  let n = {
    vector_size = v.vector_size ;
    vector_nullelem = v.vector_nullelem ;
    v = v.v
  } in
  v.vector_size <- 0 ;
  v.v <- (Array.make 0 v.vector_nullelem) ;
  n)

(*
let vec_at (v : 'a vec) (i : int) = (v.v.(i))
*)

let vec_dup (v : 'a vec) = (
  {
    vector_size = v.vector_size ;
    vector_nullelem = v.vector_nullelem ;
    v = (Array.sub v.v 0 (vec_size v))
  })

(*
let vec_copyto (v : 'a vec) (copy : 'a vec) = (
  (vec_growto copy v.vector_size v.vector_nullelem) ;
  (Array.blit v.v 0 copy.v 0 v.vector_size))
*)

(* Vector fold_left : ('a -> 'b -> 'a) -> 'a -> 'b vec -> 'a *)

(*
let vec_fold_left (f : ('a -> 'b -> 'a)) (x : 'b) (v : 'b vec) = (
  let rec vec_fold_left_loop = (fun f x v n i ->
    if (i = n) then
      x
    else
      (vec_fold_left_loop f (f x v.v.(i)) v n (i + 1))) in
  (vec_fold_left_loop f x v (vec_size v) 0))
*)

(* Finds an element and returns its index.  Or, it returns -1 or
   raises {Not_found} whe {checkp} is true. *)

let vec_find_element e (v : 'a vec) (checkp : bool)= (
  let rec vec_find_element_loop = (fun e v checkp n i -> (
    if (i = n) then
      if (checkp) then
	raise Not_found
      else
	-1
    else if (v.v.(i) = e) then
      i
    else
      (vec_find_element_loop e v checkp n (i + 1)))) in
  (vec_find_element_loop e v checkp (vec_size v) 0))

(* Finds an entry with {f} to true.  It returns an index, or an index
   equals to the size of vector if not found. *)

let vec_find (f : ('a -> bool)) (v : 'a vec) = (
  let rec vec_find_loop = (fun f v n i -> (
    if (i = n) then
      i
    else if (f (v.v.(i))) then
      i
    else
      (vec_find_loop f v n (i + 1)))) in
  (vec_find_loop f v (vec_size v) 0))

let vec_findi (f : (int -> 'a -> bool)) (v : 'a vec) = (
  let rec vec_findi_loop = (fun f v n i -> (
    if (i = n) then
      i
    else if (f i (v.v.(i))) then
      i
    else
      (vec_findi_loop f v n (i + 1)))) in
  (vec_findi_loop f v (vec_size v) 0))

(* Removes an element {e} (which equals) from a vector {v}.  It raises
   {Not_found} exception if {checkp} is true, or just ignores. *)

let vec_remove e (v : 'a vec) (checkp : bool) = (
  let i = (vec_find_element e v checkp) in
  if (i != -1) then
    (begin
      (Array.blit v.v (i + 1) v.v i (v.vector_size - i - 1)) ;
      v.vector_size <- (v.vector_size - 1)
    end))

(* Replaces a whole vector, with a given content {a} and size {s}. *)

let vec_replace (v : 'a vec) a s = (
  v.v <- a ;
  v.vector_size <- s)

let rec vec_sorted_insert_loop (v : 'a vec) e e_compare n i = (
  if (i = n) then
    (vec_push v e)
  else if ((e_compare e v.v.(i)) = 0 && e = v.v.(i)) then
    ()
  else if ((e_compare e v.v.(i)) <= 0) then
    (vec_sorted_insert_loop v e e_compare n (i + 1))
  else
    (begin
      (Array.blit v.v i v.v (i + 1) (v.vector_size - i)) ;
      v.v.(i) <- e ;
      v.vector_size <- (v.vector_size + 1)
    end))

(* MEMO: It finds an entry in [i,i+range). *)

let rec vec_sorted_insert_binary (v : 'a vec) e e_compare n i range = (
  let threshold = 128 in
  let r2 = (range / 2) in
  if (range <= threshold) then
    (vec_sorted_insert_loop v e e_compare n i)
  else if ((e_compare e v.v.(i + r2)) = 0) then
    (vec_sorted_insert_loop v e e_compare n (i + r2))
  else if ((e_compare e v.v.(i + r2)) < 0) then
    (vec_sorted_insert_binary v e e_compare n i r2)
  else
    (vec_sorted_insert_binary v e e_compare n (i + r2 + 1) (range - r2 - 1)))

(* Inserts an entry to a sorted vector.  Duplicate entries (modulo
   '=') are NOT inserted. *)

let vec_sorted_insert (v : 'a vec) e (e_compare : 'a -> 'a -> int) = (
  (vec_sorted_insert_binary v e e_compare (vec_size v) 0 (vec_size v)))

(* Vector iteri: ((int -> 'a -> unit) -> 'a vec -> unit) *)

let vec_iteri (f : int -> 'a -> unit) (v : 'a vec) = (
  for i = 0 to ((vec_size v) - 1) do (
    (f i v.v.(i))
  ) done)

(*
let vec_iteri_from (f : int -> 'a -> unit) (v : 'a vec) start = (
  for i = start to ((vec_size v) - 1) do (
    (f i v.v.(i))
  ) done)
*)

(* Vector for_all: (('a -> bool) -> 'a vec -> bool). *)

let vec_forall (f : 'a -> bool) (v : 'a vec) = (
  let rec vec_forall_loop f v n i = (
    if (i = n) then
      true
    else if (not (f v.v.(i))) then
      false
    else
      (vec_forall_loop f v n (i + 1))) in
  (vec_forall_loop f v (vec_size v) 0))

let vec_exists (f : 'a -> bool) (v : 'a vec) = (
  let rec vec_exists_loop f v n i = (
    if (i = n) then
      false
    else if (f v.v.(i)) then
      true
    else
      (vec_exists_loop f v n (i + 1))) in
  (vec_exists_loop f v (vec_size v) 0))

(* Vector filter: (('a -> bool) -> 'a vec -> 'a vec). *)

let vec_filter (f : 'a -> bool) (v : 'a vec) = (
  let w = (vec_create 0 v.vector_nullelem) in
  (vec_iteri (fun i e -> if (f e) then (vec_push w e)) v) ;
  w)

(* Utility: Vec and List Operations. *)

let vec_mapi (f : int -> 'a -> 'b) (v : 'a vec) : 'b list = (
  let rec vec_mapi_loop f l v n i = (
    if (i = n) then
      l
    else
      (vec_mapi_loop f (l @ [(f i v.v.(i))]) v n (i + 1))) in
  (vec_mapi_loop f [] v (vec_size v) 0))

let vec_from_list (l : 'a list) (nullelem : 'a) : 'a vec = (
  let w = (vec_create 0 nullelem) in
  (List.iter (fun e -> (vec_push w e)) l) ;
  w)

let vec_to_list (v : 'a vec) : 'a list = (vec_mapi (fun i e -> e) v)

(* Queue Instance *)

type 'a queue_t = 'a Queue.t

let queue_create () = (Queue.create ())
let queue_insert q x = (Queue.add x q)
let queue_dequeue q = (Queue.take q)
let queue_clear q = (Queue.clear q)
let queue_size q = (Queue.length q)

(* Clause/Constraint *)

(* {cla_no} field is for info, problem clause number. *)

type clause_t = {
  cla_no : int ;
  cla_learntp : bool ;
  mutable cla_activity : float ;
  cla_literals : lit_t vec
}

let null_literal_vector = {
  vector_size = 0 ;
  vector_nullelem = 0 ;
  v = (Array.make 0 0)
}

let null_literal_vector_p (v : 'a vec) : bool = (
  (v.vector_size == 0 && v.vector_nullelem == 0))

let null_clause = {
  cla_no = -1 ;
  cla_learntp = false ;
  cla_activity = 0.0 ;
  cla_literals = null_literal_vector
}

let clause_compare c0 c1 = (
  let v = (c0.cla_activity -. c1.cla_activity) in
  (if (v = 0.0) then 0 else if (v > 0.0) then 1 else -1))

(* Variable ordering.  It is a sorted list (by activity) of variables.
   NOTE: Calculation of variable ordering refers to {s.sol_assigns}
   and {s.sol_activity} in {sovler_t} to make a choice. *)

type varorder_t = {
  ord_choice : var_t vec
}

(* Solver body.
   PARAMETERS
   {sol_cla_inc}: Increment of clause activity.
   {sol_cla_decay}: Decay factor of clause activity.
   {sol_var_inc}: Increment of variable activity.
   {sol_var_decay}: Decay factor of variable activity.
   LIMITER
   {sol_max_learnts}: Limit of number of learnt clauses.
   {sol_max_conflicts}: Limit of number of conflicts in one major loop.
   CONSTRAINTS
   {sol_constrs}: List of problem clauses (constraints).
   {sol_learnts}: List of learnt clauses.
   ASSIGNMENTS
   {sol_root_level}: Level separating incremental and decision assumptions.
   {sol_assigns}: Assignment on each variable.
   {sol_level}: Guess/decision level of assignment of each variable.
   {sol_reason}: Constraint which implied the value of each variable.
   {sol_trail}: List of assignments in chronological order.
   {sol_trail_limit}: Position in trail separating guess/decision levels.
   PROPAGATION
   {sol_prop_queue}: Propagation queue.
   {sol_watches}: List of clauses watching on each literal.
   {sol_undos}: List of clauses undone in backtracking of each variable.
   VARIABLE ORDERING
   {sol_order}: Variable ordering.
   {sol_activity}: Activity measurement of each variable. *)

type solver_t = {
  (* PARAMETERS *)
  mutable sol_cla_inc : float ;
  mutable sol_cla_decay : float ;
  mutable sol_var_inc : float ;
  mutable sol_var_decay : float ;
  (* LIMITER (UPDATED WHILE LOOPING) *)
  mutable sol_max_learnts : int ;
  mutable sol_max_conflicts : int ;
  (* CONSTRAINTS *)
  sol_constrs : clause_t vec ;
  sol_learnts : clause_t vec ;
  (* ASSIGNMENTS *)
  mutable sol_root_level : int ;
  sol_assigns : lbool_t vec ;
  sol_level : int vec ;
  sol_reason : clause_t vec ;
  sol_trail : lit_t vec ;
  sol_trail_limit : int vec ;
  (* PROPAGATION *)
  sol_prop_queue : lit_t queue_t ;
  sol_watches : clause_t vec vec ;
  sol_undos : clause_t vec vec ;
  (* VARIABLE ORDERING *)
  sol_order : varorder_t ;
  sol_activity : float vec
}

let sat_result_to_string (r : sat_result_t) = (
  match r with SAT -> "SAT" | UNSAT -> "UNSAT" | UNKNOWN -> "UNKNOWN")

let lbool_to_string (s : solver_t) (v : lbool_t) = (
  match v with LTrue -> "LTrue" | LFalse -> "LFalse" | LBottom -> "LBottom")

(* NOTE: LITERALS (VARIABLES) ARE PRINTED WITH ONE-ADDED FOR MATCHING
   THE CNF FORMAT. *)

let literal_to_string (s : solver_t) (p : lit_t) = (
  ((if (lit_sign p) then "" else "-") ^ (string_of_int ((lit_var p) + 1))))

let clause_to_string (s : solver_t) (c : clause_t) = (
  if (c = null_clause) then
    "C(null)"
  else
    (begin
      let ss = ref "C(" in
      if (c.cla_no != -1) then
	ss := (!ss ^ "[" ^ (string_of_int c.cla_no) ^ "] ") ;
      (vec_iteri
	  (fun i p ->
	    ss := (!ss ^ (if (i = 0) then "" else " ")
		    ^ (literal_to_string s p)))
	  c.cla_literals) ;
      ss := (!ss ^ ")") ;
      !ss
    end))

(*
let literals_to_string (s : solver_t) (ps : lit_t vec) = (
  let ss = ref "L(" in
  (vec_iteri
      (fun i p ->
	ss := (!ss ^ (if (i = 0) then "" else " ")
		^ (literal_to_string s p)))
      ps) ;
  (!ss ^ ")"))
*)

let varorder_create () = (
  {ord_choice = (vec_create 0 0)})

(* Compares activities, with tie-break small variable number being
   greater. *)

let varorder_compare (s : solver_t) (x0 : var_t) (x1 : var_t) = (
  let v = (s.sol_activity.v.(x0) -. s.sol_activity.v.(x1)) in
  if (v = 0.0) then
    (x0 - x1)
  else if (v > 0.0) then
    1
  else
    -1)

let varorder_new_var (s : solver_t) (o : varorder_t) (x : var_t) = (
  (vec_sorted_insert o.ord_choice x (varorder_compare s)))

let varorder_update (s : solver_t) (o : varorder_t) (x : var_t) = (
  (vec_remove x o.ord_choice false) ;
  (vec_sorted_insert o.ord_choice x (varorder_compare s)))

let varorder_update_all (s : solver_t) (o : varorder_t) = (
  let v = o.ord_choice in
  let sorted = (Array.sub v.v 0 (vec_size v)) in
  (Array.sort (varorder_compare s) sorted) ;
  (vec_replace v sorted (vec_size v)))

let varorder_select (s : solver_t) (o : varorder_t) = (
  (vec_last_pop o.ord_choice))

let varorder_undo (s : solver_t) (o : varorder_t) (x : var_t) = (
  (vec_sorted_insert o.ord_choice x (varorder_compare s)))

let solver_create () = (
  {
    sol_constrs = (vec_create 0 null_clause) ;
    sol_learnts = (vec_create 0 null_clause) ;
    sol_cla_inc = 0.0 ;
    sol_cla_decay = 0.0 ;
    sol_activity = (vec_create 0 0.0) ;
    sol_var_inc = 0.0 ;
    sol_var_decay = 0.0 ;
    sol_max_learnts = 0 ;
    sol_max_conflicts = 0 ;
    sol_order = (varorder_create ()) ;
    sol_watches = (vec_create 0 (vec_create 0 null_clause)) ;
    sol_undos = (vec_create 0 (vec_create 0 null_clause)) ;
    sol_prop_queue = (queue_create ()) ;
    sol_assigns = (vec_create 0 LBottom) ;
    sol_level = (vec_create 0 0) ;
    sol_reason = (vec_create 0 null_clause) ;
    sol_trail = (vec_create 0 0) ;
    sol_trail_limit = (vec_create 0 0) ;
    sol_root_level = 0
  })

(* Returns the number of variables in use. *)

let n_variables (s : solver_t) = (vec_size s.sol_assigns)

let solver_var_rescale_activity (s : solver_t) = (
  for x = 0 to ((n_variables s) - 1) do (
    (s.sol_activity.v.(x) <- (s.sol_activity.v.(x) *. 1e-100))
  ) done ;
  s.sol_var_inc <- (s.sol_var_inc *. 1e-100))

let solver_var_bump_activity (s : solver_t) (x : var_t) = (
  (s.sol_activity.v.(x) <- (s.sol_activity.v.(x) +. s.sol_var_inc)) ;
  if (s.sol_activity.v.(x) > 1e100) then
    (solver_var_rescale_activity s) ;
  (varorder_update s s.sol_order x))

let solver_var_decay_activity (s : solver_t) = (
  s.sol_var_inc <- (s.sol_var_inc *. s.sol_var_decay))

let solver_clause_rescale_activity (s : solver_t) = (
  (vec_iteri
      (fun i c -> c.cla_activity <- (c.cla_activity *. 1e-100))
      s.sol_constrs) ;
  s.sol_cla_inc <- (s.sol_cla_inc *. 1e-100))

let solver_clause_bump_activity (s : solver_t) (c : clause_t) = (
  (c.cla_activity <- (c.cla_activity +. s.sol_var_inc)) ;
  if (c.cla_activity > 1e100) then
    (solver_clause_rescale_activity s))

let solver_clause_decay_activity (s : solver_t) = (
  s.sol_cla_inc <- (s.sol_cla_inc *. s.sol_cla_decay))

let solver_decay_activities (s : solver_t) = (
  (solver_var_decay_activity s) ;
  (solver_clause_decay_activity s))

(*let solver_n_assigns (s : solver_t) = (vec_size s.sol_trail)*)
(*let solver_n_constraints (s : solver_t) = (vec_size s.sol_constrs)*)
(*let solver_n_learnts (s : solver_t) = (vec_size s.sol_learnts)*)
(*let solver_var_value (s : solver_t) (x : var_t) = (s.sol_assigns.v.(x))*)

(* Returns an assignment to a literal. *)

let solver_lit_value (s : solver_t) (p : lit_t) = (
  let v = (s.sol_assigns.v.(lit_var p)) in
  if (lit_sign p) then v else (lbool_neg v))

(* Returns a decision level. Decision level is the number of
   guesses/decisions currently made. *)

let solver_decision_level (s : solver_t) = (vec_size s.sol_trail_limit)

(* Creates a new variable and returns it.  It also assigns a slot for
   it in vectors in the solver.  It pushes twice in watches for
   positive/negative literals. *)

let solver_new_variable (s : solver_t) : var_t = (
  let v = (n_variables s) in
  (vec_push s.sol_assigns LBottom) ;
  (vec_push s.sol_level (-1)) ;
  (vec_push s.sol_reason null_clause) ;
  (vec_push s.sol_watches (vec_create 0 null_clause)) ;
  (vec_push s.sol_watches (vec_create 0 null_clause)) ;
  (vec_push s.sol_undos (vec_create 0 null_clause)) ;
  (vec_push s.sol_activity 0.0) ;
  (varorder_new_var s s.sol_order v) ;
  v)

(* Removes a clause {c} from the watcher list.  MEMO: The 1st and 2nd
   literals make a unit, and a clause may be enteried for them. *)

let clause_remove (s : solver_t) (c : clause_t) = (
  if (c <> null_clause) then
    (begin
      let lits = c.cla_literals in
      (vec_remove c (s.sol_watches.v.(lit_neg lits.v.(0))) true) ;
      (vec_remove c (s.sol_watches.v.(lit_neg lits.v.(1))) true)
    end))

let rec clause_simplify_loop (s : solver_t) (lits : lit_t vec) n i j = (
  if (i = n) then
    (begin
      (vec_shrinkto lits j) ;
      false
    end)
  else
    (begin
      match (solver_lit_value s lits.v.(i)) with
	LTrue ->
	  true
      | LFalse ->
	  (clause_simplify_loop s lits n (i + 1) j)
      | LBottom ->
	  lits.v.(j) <- lits.v.(i) ;
	  (clause_simplify_loop s lits n (i + 1) (j + 1))
    end))

(* Shrinks a clause destructively only to not-assigned literals.  It
   is only called at the toplevel.  It returns true if a clause is
   trivial (includes a true literal). *)

let clause_simplify (s : solver_t) (c : clause_t) = (
  if ((vec_size c.cla_literals) = 0) then
    true
  else
    (clause_simplify_loop s (c.cla_literals) (vec_size c.cla_literals) 0 0))

(* Assigns and enqueues a new assignment {p}=true for propagation.  It
   returns true on consistent assignment or false on conflicting
   assignment.  A given {reason} clause is one implies {p}, or a
   null-clause.  (It is named {enqueue} in the paper). *)

let solver_assign (s : solver_t) (p : lit_t) (reason : clause_t) = (
  match (solver_lit_value s p) with
    LFalse ->
      (* Conflicting assignment. *)
      false
  | LTrue ->
      (* Existing consistent assignment. *)
      true
  | LBottom ->
      (begin
	let v = (lit_var p) in
	s.sol_assigns.v.(v) <- (lbool_of_bool (lit_sign p)) ;
	s.sol_level.v.(v) <- (solver_decision_level s) ;
	s.sol_reason.v.(v) <- reason ;
	(vec_push s.sol_trail p) ;
	(queue_insert s.sol_prop_queue p) ;
	true
      end))

(* Makes a new guess/decision with a literal {p}=true, and records a
   decision point (for backtracking) as a trail-limit. *)

let solver_assume (s : solver_t) (p : lit_t) = (
  (vec_push s.sol_trail_limit (vec_size s.sol_trail)) ;
  (solver_assign s p null_clause))

let rec clause_find_nonfalse_loop (s : solver_t) (lits : lit_t vec) n i = (
  if (i = n) then
    i
  else if ((solver_lit_value s lits.v.(i)) != LFalse) then
    i
  else
    (clause_find_nonfalse_loop s lits n (i + 1)))

(* Propagates a clause in the watcher list of {p} (ie, a clause
   includes {neg(p)} at 0th or 1st position).  It returns true on
   consistent assignment, or false on conflicting assignment. *)

let clause_propagate (s : solver_t) (c : clause_t) (p : lit_t) = (
  xTRACE ("propagate: " ^ (clause_to_string s c)) ;
  let lits = c.cla_literals in
  (* Swap positions for the false literal be at lits[1]. *)
  if (lits.v.(0) = (lit_neg p)) then
    (vec_swap lits 0 1) ;
  if ((solver_lit_value s lits.v.(0)) = LTrue) then
    (begin
      (* Return the watcher. *)
      (vec_push (s.sol_watches.v.(p)) c) ;
      true
    end)
  else
    let i = (clause_find_nonfalse_loop s lits (vec_size lits) 2) in
    if (i = (vec_size lits)) then
      (begin
	(* Return the watcher. *)
	(vec_push (s.sol_watches.v.(p)) c) ;
	(* Clause is a unit, so assign one. *)
	(solver_assign s (lits.v.(0)) c)
      end)
    else
      (begin
	(* Insert the watcher in the new watcher list. *)
	(vec_swap lits 1 i) ;
	(vec_push (s.sol_watches.v.(lit_neg lits.v.(1))) c) ;
	true
      end))

(* Lists reasons which are negation of literals in a clause.  {p} is
   null, or a head of the clause {c}. *)

let clause_list_reason (s : solver_t) (c : clause_t) (p : lit_t) = (
  assert ((p = null_literal) || (p = c.cla_literals.v.(0))) ;
  let reason : lit_t vec = (vec_create 0 0) in
  (vec_iteri
      (fun i q ->
	if (not (i = 0 && p != null_literal)) then
	  (vec_push reason (lit_neg q)))
      c.cla_literals) ;
  assert (vec_forall (fun q -> ((solver_lit_value s q) = LTrue)) reason) ;
  if (c.cla_learntp) then
    (solver_clause_bump_activity s c) ;
  reason)

let rec clause_new_simplify_loop s lits n i j = (
  if (i = n) then
    (begin
      (vec_shrinkto lits j) ;
      false
    end)
  else if ((solver_lit_value s lits.v.(i)) = LTrue) then
    (* See a true literal. *)
    (begin
      (vec_shrinkto lits 0) ;
      true
    end)
  else if ((solver_lit_value s lits.v.(i)) = LFalse) then
    (* Skip a false literal. *)
    (clause_new_simplify_loop s lits n (i + 1) j)
  else if (i = (n - 1)) then
    (* Copy the last literal to the output. *)
    (begin
      lits.v.(j) <- lits.v.(i) ;
      (clause_new_simplify_loop s lits n (i + 1) (j + 1))
    end)
  else if (lits.v.(i) = lits.v.(i + 1)) then
    (* Skip a duplicate. *)
    (clause_new_simplify_loop s lits n (i + 1) j)
  else if ((lits.v.(i) lor 1) = lits.v.(i + 1)) then
    (* See a positive/negative pair. *)
    (begin
      (vec_shrinkto lits 0) ;
      true
    end)
  else
    (* Copy the literal to output. *)
    (begin
      lits.v.(j) <- lits.v.(i) ;
      (clause_new_simplify_loop s lits n (i + 1) (j + 1))
    end))

(* Checks trivially satisfied cases.  It is of occurrence of any true
   literal or both positive and negative literals.  It returns true if
   it is trivially satisfied.  It DESTRUCTIELY simplifies a clause by
   removing false and duplicate literals.  And thus, it returns an
   empty literal vector, which is trivially unsatisfied.  It requires
   the literals are sorted.  It is only called at the toplevel, for
   not-learnt clauses.  *)

let clause_new_simplify (s : solver_t) (ps : lit_t vec) = (
  let sorted = (Array.sub ps.v 0 (vec_size ps)) in
  (Array.sort compare sorted) ;
  (vec_replace ps sorted (Array.length sorted)) ;
  (clause_new_simplify_loop s ps (vec_size ps) 0 0))

(* Returns an index of the maximum of decision levels of literals. *)

let clause_maxi_of_level (s : solver_t) (ps : lit_t vec) = (
  let maxiv = ref (0, -1) in
  (Array.iteri
      (fun i p ->
	if (s.sol_level.v.(lit_var p) > (fst !maxiv)) then
	  maxiv := (s.sol_level.v.(lit_var p), i))
      ps.v) ;
  (snd !maxiv))

(* Creates a new clause and adds it to watcher lists.  It returns
   UNKNOWN when it is consistent, or SAT/UNSAT when it is trivially
   satisfied/unsatisfied.  It returns a null-clause when the length of
   the clause (after simplification) is zero or one.  NOTE: The input
   {ps} is destructed, and the literals are sorted, when it is not a
   learnt-clause. *)

let clause_new (s : solver_t) (ps : lit_t vec) (learntp : bool) (id : int) : (sat_result_t * clause_t) = (
  (*let ps = (vec_dup ps) in*)
  let trivialsat = if (learntp) then false else (clause_new_simplify s ps) in
  if (trivialsat) then
    (SAT, null_clause)
  else if ((vec_size ps) = 0) then
    (UNSAT, null_clause)
  else if ((vec_size ps) = 1) then
    (* Assign a variable as a fact. *)
    let ok = (solver_assign s ps.v.(0) null_clause) in
    if (ok) then
      (SAT, null_clause)
    else
      (UNSAT, null_clause)
  else
    (begin
      let c : clause_t = {
	cla_no = id ;
	cla_learntp = learntp ;
	cla_activity = 0.0 ;
	cla_literals = ps
      } in
      let lits = c.cla_literals in
      if (learntp) then
	(begin
	  (* Set the 2nd literal with the highest decision level. *)
	  let i = (clause_maxi_of_level s lits) in
	  (vec_swap lits 1 i) ;
	  (solver_clause_bump_activity s c) ;
	  (vec_iteri
	      (fun i p -> (solver_var_bump_activity s (lit_var p)))
	      lits)
	end) ;
      (* Add the created clause to watcher lists. *)
      (vec_push (s.sol_watches.v.(lit_neg lits.v.(0))) c) ;
      (vec_push (s.sol_watches.v.(lit_neg lits.v.(1))) c) ;
      (UNKNOWN, c)
    end))

(* Propagates clauses in the watcher list until a conflict, and
   returns -1 or an index of a conflicting clause. *)

let rec solver_propagate_loop (s : solver_t) watch p n i conflict = (
  if (i = n) then
    conflict
  else if (conflict = -1) then
    let ok = (clause_propagate s (watch.v.(i)) p) in
    (solver_propagate_loop s watch p n (i + 1) (if (ok) then -1 else i))
  else
    (begin
      (vec_push (s.sol_watches.v.(p)) (watch.v.(i))) ;
      (solver_propagate_loop s watch p n (i + 1) conflict)
    end))

(* Propagates literals in the propagation queue.  It returns a
   null-clause when a clause is consistent, or a conflicting clause
   when conflicting.  It empties the watcher list on {p} once, which
   is later filled. *)

let rec solver_propagate (s : solver_t) = (
  if ((queue_size s.sol_prop_queue) = 0) then
    null_clause
  else
    (begin
      let p : lit_t = (queue_dequeue s.sol_prop_queue) in
      xTRACE ("dequeue: L(" ^ (literal_to_string s p) ^ ")") ;
      let p_watch : clause_t vec = (vec_steal (s.sol_watches.v.(p))) in
      let conflict = (solver_propagate_loop s p_watch p
			 (vec_size p_watch) 0 (-1)) in
      if (conflict = -1) then
	(solver_propagate s)
      else
	(begin
	  (queue_clear s.sol_prop_queue) ;
	  p_watch.v.(conflict)
	end)
    end))

(* Undoes on clause when unbinding {p}. *)

let clause_undo (s : solver_t) (c : clause_t) (p : lit_t) = (
  (*nothing*))

(* Pops off the last literal in the trail, and undoes on it. *)

let solver_undo_one (s : solver_t) = (
  let p : lit_t = (vec_last_pop s.sol_trail) in
  let x = (lit_var p) in
  s.sol_assigns.v.(x) <- LBottom ;
  s.sol_level.v.(x) <- -1 ;
  s.sol_reason.v.(x) <- null_clause ;
  (varorder_undo s s.sol_order x) ;
  while ((vec_size s.sol_undos.v.(x)) > 0) do (
    let c = (vec_last_pop s.sol_undos.v.(x)) in
    (clause_undo s c p) ;
  ) done)

(* Unbinds literals in the trail until meeting a {seen} literal.  It
   returns a pair of a clause and a literal (where the clause causes
   to bind the literal). *)

let rec solver_analyze_unbind_to_some_reason_loop (s : solver_t) seen = (
  let p = (vec_last s.sol_trail) in
  let c = (s.sol_reason.v.(lit_var p)) in
  (solver_undo_one s) ;
  if (seen.v.(lit_var p)) then
    (c, p)
  else
    (solver_analyze_unbind_to_some_reason_loop s seen))

let rec solver_analyze_loop (s : solver_t) (conflict : clause_t) (p : lit_t) seen n_vars_to_resolve learnt_clause backtrack_level = (
  assert (conflict != null_clause) ;
  (* Trace the reason for p. *)
  let p_reason : lit_t vec = (clause_list_reason s conflict p) in
  (vec_iteri
      (fun i (q : lit_t) ->
	if (not (seen.v.(lit_var q))) then
	  (begin
	    seen.v.(lit_var q) <- true ;
	    let qlev = (s.sol_level.v.(lit_var q)) in
	    if (qlev = (solver_decision_level s)) then
	      (* Variables of the current level be resolved. *)
	      incr n_vars_to_resolve
	    else if (qlev > 0) then
	      (* Only variables deduced (not given constraint). *)
	      (begin
		(vec_push learnt_clause (lit_neg q)) ;
		backtrack_level := (max !backtrack_level qlev)
	      end)
	  end))
      p_reason) ;
  assert (!n_vars_to_resolve > 0) ;
  (* Select a next literal. *)
  let (conflict1, p1) = (solver_analyze_unbind_to_some_reason_loop s seen) in
  (decr n_vars_to_resolve) ;
  if (!n_vars_to_resolve = 0) then
    p1
  else
    (solver_analyze_loop (s : solver_t) conflict1 p1 seen
	n_vars_to_resolve learnt_clause backtrack_level))

(* Analyzes a conflict, then makes and returns a learnt-clause and a
   backtrack-level.  The 1st literal in the learnt-clause is the
   asserting literal at the backtrack-level. *)

let solver_analyze (s : solver_t) (conflict : clause_t) = (
  (* Keep room in the 1st entry of {learnt_clause}. *)
  let learnt_clause : lit_t vec = (vec_create 1 0) in
  let backtrack_level = ref 0 in
  let n_vars_to_resolve = ref 0 in
  let seen : bool vec = (vec_create (n_variables s) false) in
  let p = (solver_analyze_loop s conflict null_literal seen
	      n_vars_to_resolve learnt_clause backtrack_level) in
  learnt_clause.v.(0) <- (lit_neg p) ;
  (* NOTE: OMIT SOME SIMPLIFICATION OF CONFLICT CLAUSE... *)
  (learnt_clause, !backtrack_level))

(* Undoes to the last guess/decision point (trail-limit). *)

let solver_cancel (s : solver_t) = (
  let limit = (vec_last_pop s.sol_trail_limit) in
  while (((vec_size s.sol_trail) > limit)) do (
    (solver_undo_one s)
  ) done)

(* Undoes to the specified depth of decision (number of guesses). *)

let solver_cancel_until (s : solver_t) (level : int) = (
  while ((solver_decision_level s) > level) do (
    (solver_cancel s)
  ) done)

(* Sorts clauses with increasing order.  MEMO: It copies once to use
   {Array.sort}. *)

let sort_on_activity (cs : clause_t vec) = (
  cs.v <- (Array.sub cs.v 0 (vec_size cs)) ;
  (Array.sort clause_compare cs.v))

(* Checks if a learnt-clause is currently used as a reason of some
   assignment. *)

let clause_locked (s : solver_t) (c : clause_t) = (
  (s.sol_reason.v.(lit_var c.cla_literals.v.(0)) = c))

let rec solver_reduce_db_loop (s : solver_t) (lim : float) n i j = (
  if (i = n) then
    (vec_shrinkto s.sol_learnts j)
  else
    (begin
      let c = (s.sol_learnts.v.(i)) in
      if (clause_locked s c) then
	(begin
	  (s.sol_learnts.v.(j) <- s.sol_learnts.v.(i)) ;
	  (solver_reduce_db_loop s lim n (i + 1) (j + 1))
	end)
      else if (i < (n / 2)) then
	(begin
	  (clause_remove s c) ;
	  (solver_reduce_db_loop s lim n (i + 1) j)
	end)
      else if (c.cla_activity < lim) then
	(begin
	  (clause_remove s c) ;
	  (solver_reduce_db_loop s lim n (i + 1) j)
	end)
      else
	(begin
	  (s.sol_learnts.v.(j) <- s.sol_learnts.v.(i)) ;
	  (solver_reduce_db_loop s lim n (i + 1) (j + 1))
	end)
    end))

(* Removes learnt-clauses with low activity. *)

let solver_reduce_db (s : solver_t) = (
  if ((vec_size s.sol_learnts) - (vec_size s.sol_trail) >= s.sol_max_learnts) then
    (begin
      let lim = (s.sol_cla_inc /. (float_of_int (vec_size s.sol_learnts))) in
      (sort_on_activity s.sol_learnts) ;
      (solver_reduce_db_loop s lim (vec_size s.sol_learnts) 0 0)
    end))

let rec solver_simplify_db_loop (s : solver_t) (cs : clause_t vec) n i j = (
  if (i = n) then
    (vec_shrinkto cs j)
  else if (clause_simplify s cs.v.(i)) then
    (begin
      (clause_remove s cs.v.(i)) ;
      (solver_simplify_db_loop s cs n (i + 1) j)
    end)
  else
    (begin
      cs.v.(j) <- cs.v.(i) ;
      (solver_simplify_db_loop s cs n (i + 1) (j + 1))
    end))

(* Calls {clause_simplify} on constraint-clauses and learnt-clauses.
   It returns false if conflicting.  It is only called at the
   toplevel. *)

let solver_simplify_db (s : solver_t) = (
  if ((solver_propagate s) != null_clause) then
    false
  else
    (begin
      (solver_simplify_db_loop s s.sol_learnts (vec_size s.sol_learnts) 0 0) ;
      (solver_simplify_db_loop s s.sol_constrs (vec_size s.sol_constrs) 0 0) ;
      true
    end))

(* Records a learnt-clause of given literals.  NOTE: A clause cannot
   be trivially unsatisfiable, but it can be trivially satisfiable due
   to the already inferred assignments. *)

let solver_record (s : solver_t) (ps : lit_t vec) = (
  let (trivialsat, c) = (clause_new s ps true (-1)) in
  assert (trivialsat <> UNSAT) ;
  if (c != null_clause) then
    (vec_push s.sol_learnts c) ;
  let ok = (solver_assign s ps.v.(0) c) in
  assert (ok = true))

(* Searches one solution.  It returns {LTrue,Some} when a model found,
   {LFalse,None} a model not found, or {LBottom,None} when number of
   conflicts meets a limit.  MEMO: It makes only positive assumptions;
   Negative one are assumed from conflict. *)

let rec solver_search_loop (s : solver_t) (conflict_count : int) = (
  let conflict : clause_t = (solver_propagate s) in
  if (conflict != null_clause) then
    (begin
      (* Conflict. *)
      if ((solver_decision_level s) = s.sol_root_level) then
	(UNSAT, None)
      else if (conflict_count >= s.sol_max_conflicts) then
	(begin
	  (solver_cancel_until s s.sol_root_level) ;
	  (UNKNOWN, None)
	end)
      else
	(begin
	  xTRACE ("conflict: " ^ (clause_to_string s conflict)) ;
	  let (learnt_clause, backtrack_level) = (solver_analyze s conflict) in
	  (solver_cancel_until s (max backtrack_level s.sol_root_level)) ;
	  (solver_record s learnt_clause) ;
	  (solver_decay_activities s) ;
	  (solver_search_loop s (conflict_count + 1))
	end)
    end)
  else
    (begin
      (* No conflict. *)
      if ((solver_decision_level s) = 0) then
	(* {solver_simplify_db} never returns false: *)
	ignore (solver_simplify_db s) ;
      (solver_reduce_db s) ;
      if ((vec_size s.sol_trail) = (n_variables s)) then
	(begin
	  (* Model found.  Copy the found model and reset. *)
	  let model = (vec_create (n_variables s) false) in
	  (vec_iteri
	      (fun i assign -> model.v.(i) <- (assign = LTrue))
	      s.sol_assigns) ;
	  (solver_cancel_until s s.sol_root_level) ;
	  (SAT, Some model)
	end)
      else
	(begin
	  (* Make new (positive) assumption, ignore conflict. *)
	  let p : lit_t = (lit_make (varorder_select s s.sol_order) true) in
	  let ok = (solver_assume s p) in
	  if (ok) then
	    xTRACE ("assume: L(" ^ (literal_to_string s p) ^ ")") ;
	  (solver_search_loop s conflict_count)
	end)
    end))

let rec solver_solve_assume_loop (s : solver_t) (lits : lit_t vec) n i = (
  if (i = n) then
    true
  else
    let ok = (solver_assume s lits.v.(i)) in
    let conflict = (if (ok) then (solver_propagate s) else null_clause) in
    if (conflict != null_clause) then
      (solver_solve_assume_loop s lits n (i + 1))
    else
      (begin
	(solver_cancel_until s 0) ;
	false
      end))

(* Solves with literals as assumptions.  It assigns literals as given
   assumptions, or resets all assumptions when conflict. *)

let solver_solve (s : solver_t) (assumps : lit_t vec) = (
  (* Push incremental assumptions. *)
  let ok = (solver_solve_assume_loop s assumps (vec_size assumps) 0) in
  if (not ok) then
    (begin
      xTRACE ((sat_result_to_string UNSAT)) ;
      (UNSAT, None)
    end)
  else
    (begin
      (* Solve. *)
      let params = {params_var_decay = 0.95 ; params_cla_decay = 0.999} in
      s.sol_var_decay <- (1.0 /. params.params_var_decay) ;
      s.sol_cla_decay <- (1.0 /. params.params_cla_decay) ;
      s.sol_root_level <- (solver_decision_level s) ;
      let n_conflicts = ref 100.0 in
      let n_learnts = ref ((float_of_int (vec_size s.sol_constrs)) /. 3.0) in
      let status = ref UNKNOWN in
      let model = ref None in
      while (!status = UNKNOWN) do (
	s.sol_max_learnts <- (int_of_float !n_learnts) ;
	s.sol_max_conflicts <- (int_of_float !n_conflicts) ;
	let (s, m) = (solver_search_loop s 0) in
	status := s ;
	model := m ;
	n_conflicts := (!n_conflicts *. 1.5) ;
	n_learnts := (!n_learnts *. 1.1) ;
      ) done ;
      (solver_cancel_until s 0) ;
      xTRACE ((sat_result_to_string !status)) ;
      (!status, !model)
    end))

(* Convers a model (output of {solver_solve}) to a vector of negations
   of literals.  It is used as a reject clause to enumerate all
   solutions. *)

let solver_reject (s : solver_t) (model : bool vec) : lit_t vec = (
  let r = (vec_create 0 0) in
  (vec_iteri
      (fun (i : int) (v : bool) -> (vec_push r (lit_make i (not v))))
      model) ;
  r)

(* OPTIONAL; NOT USED NOW. *)

(*
let solver_static_var_order = (
  (* Clear activity: *)
  for (int x = 0; x < (n_variables s) ; i++) do (
    s.sol_activity.v.(x) <- 0
  ) done ;
  (* Do simple variable activity heuristic: *)
  for (int i = 0; i < clauses.size(); i++) do (
    Clause c = clauses[i] ;
    double add = pow2(- size(c)) ;
    for (int j = 0; j < size(c) j++) do (
      activity[var(c[j])] += add
    ) done ;
  ) done ;
  (* Calculate the initial "heat" of all clauses: *)
  (* Map literal to list of clause indices *)
  Vec<Vec<int>> occurs(2* (n_variables s)) ;
  (* Pairs of heat and clause index *)
  Vec<Pair<double, int>> heat(clauses.size()) ;
  for (int i = 0; i < clauses.size(); i++) do (
    Clause c = clauses[i] ;
    double sum = 0 ;
    for (int j = 0; j < size(c); j++) do (
      occurs[index(c[j])].push(i) ;
      sum += activity[var(c[j])]
    ) done ;
    heat[i] = Pair new (sum, i)
  ) done ;
  (* Bump heat for clauses whose variables occur in other hot clauses: *)
  double iter_size = 0 ;
  for (int i = 0; i < clauses.size(); i++) do (
    Clause c = clauses[i] ;
    for (int j = 0; j < size(c); j++) do (
      iter_size += occurs[index(c[j])].size()
    ) done ;
  ) done ;
  int iterations = min((int)((double)literals / iter_size * 100), 10) ;
  double disapation = 1.0 / iterations ;
  for (int c = 0; c < iterations; c++) do (
    for (int i = 0; i < clauses.size() i++) do (
      Clause c = clauses[i] ;
      for (int j = 0; j < size(c); j++) do (
	Vec<int> os = occurs[index(c[j])] ;
	for (int k = 0; k < os.size(); k++) do (
	  heat[i].fst += heat[os[k]].fst * disapation
	) done ;
      ) done ;
    ) done ;
  ) done ;
  (* Set activity according to hot clauses: *)
  sort(heat) ;
  for (int i = 0; i < (n_variables s); i++) do (
    activity[i] = 0 ;
    double extra = 1e200 ;
    for (int i = 0; i < heat.size() i++) do (
      Clause& c = clauses[heat[i].snd] ;
      for (int j = 0; j < size(c) j++) do (
	if (activity[var(c[j])] = 0) then
	  activity[var(c[j])] = extra ;
	extra *= 0.995
      ) done ;
    ) done ;
  ) done ;
  (order_update_all) ;
  var_inc = 1)
*)

(* Makes variables up to a variable {x}. *)

let claim_var (solver : solver_t) (x : var_t) = (
  for i = (n_variables solver) to x do (
    ignore (solver_new_variable solver)
  ) done)

(* Makes a new problem clause of literals {ps}.  It returns UNKNOWN
   normally, SAT on a trivially satisfied clause case, or UNSAT on a
   trivially unsatisfied case.  NOTE: SAT and UNKNOWN can be treated
   in the same way. *)

let claim_clause (solver : solver_t) (ps : lit_t vec) : sat_result_t = (
  let id = (vec_size solver.sol_constrs) in
  let (trivialsat, c) = (clause_new solver ps false id) in
  match trivialsat with
    UNSAT ->
      UNSAT
  | SAT ->
      SAT
  | UNKNOWN ->
      assert (c <> null_clause) ;
      xTRACE ("claim clause: " ^ (clause_to_string solver c)) ;
      (vec_push solver.sol_constrs c) ;
      UNKNOWN)

(* **** Utilities for Testing **** *)

let dump_clauses (s : solver_t) = (
  (vec_iteri
      (fun i c -> (print_endline ((clause_to_string s c))))
      s.sol_constrs))

let dump_learnts (s : solver_t) = (
  (vec_iteri
      (fun i c -> (print_endline ("[" ^ (string_of_int i) ^ "]="
				   ^ (clause_to_string s c))))
      s.sol_learnts))

let dump_assigns (s : solver_t) = (
  (vec_iteri
      (fun i a -> (print_endline ("[" ^ (string_of_int i) ^ "]="
				   ^ (lbool_to_string s a))))
      s.sol_assigns))

let print_model (model : bool vec) = (
  (print_string ";; model: ") ;
  for x = 0 to ((vec_size model) - 1) do (
    (print_string ((string_of_int
		       (if (model.v.(x)) then (x + 1) else (- (x + 1))))
		    ^ " "))
  ) done ;
  (print_endline "0"))

(* Checks the model.  IT DOES NOT CHECK TRIVIAL CLUASES, BECAUSE THEY
   ARE NOT IN THE CONSTRAINT LIST. *)

let check_model (s : solver_t) (model : bool vec) (verbose : bool) = (
  if (verbose) then
    (print_model model) ;
  if ((n_variables s) != (vec_size model)) then
    (begin
      if (verbose) then
	print_endline (";; check " ^ "fail") ;
      false
    end)
  else
    (begin
      let ok =
	(vec_forall
	    (fun c ->
	      if (verbose) then
		print_endline (";; check: " ^ (clause_to_string s c));
	      (vec_exists
		  (fun p ->
		    let b = (model.v.(lit_var p)) in
		    (if (lit_sign p) then b else (not b)))
		  c.cla_literals))
	    s.sol_constrs) in
      if (verbose) then
	print_endline (";; check " ^ (if ok then "ok" else "fail")) ;
      ok
    end))

let rec dump_all_models (solver : solver_t) = (
  let (sat, model) = (solver_solve solver (vec_create 0 0)) in
  if (sat = SAT) then
    (begin
      (print_model (some_of model)) ;
      let ok1 = (check_model solver (some_of model) false) in
      if (not ok1) then
	(failwith "SAT GENERATES A BAD SOLUTION") ;
      let reject = (solver_reject solver (some_of model)) in
      let trivialsat = (claim_clause solver reject) in
      (* Generated results can be trivially unsatisfiable. *)
      if (trivialsat <> UNSAT) then
	(dump_all_models solver)
    end))

let make_model_from_literal_list (size : int) (l : int list) = (
  let model = (vec_create size false) in
  let index = ref 1 in
  (List.iter
      (fun e ->
	assert ((abs e) = !index) ;
	incr index ;
	if (e > 0) then
	  model.v.((abs e) - 1) <- true
	else
	  model.v.((abs e) - 1) <- false)
      l) ;
  model)

(* ================================================================ *)

(* DIMACS CNF READER *)

(* INPUT *)
(* <input> ::= <comment>* <problem> <clause>* EOF *)
(* Comment:	'c' <anything> <nl> *)
(* Problem:	'p' 'cnf' <#variables> <#clauses> <nl> *)
(* Clause: 	[-]<var> ... '0' *)
(* Where [-]<var> is non-zero integer, *)
(* and clauses can span multiple lines. *)

(* OUTPUT *)
(* <output> ::= <comment>* <solution> <timing> <variable>* <clause>* *)
(* Commnet: 	'c' <commnet> *)
(* Solution: 	's' {'max'|'cnf'} {'0'|'1'|'-1'} <#variables> <#clauses> *)
(* Where ('0' for UNSAT, '1' for SAT, '-1' for UNDEC) *)
(* Timing:	't' <same-line-as-solution> <cpusecs> "0" *)
(* Variable:	'v' [-]<variable> *)
(* Clause:	'c' [-]<clause> *)

(* MEMO: Peeking by "%0c" scanf format only works on Scanf.Scanning --
   not work with input channels. *)

(* Note that each variable in a CNF file is assigned one-less number
   in the solver, because variable number is one-origin in a CNF
   file. *)

(* Reads CNF in DIMACS from a file {f}.  ERROR CHECKING IS NONE.  It
   does not read preamble at all, following the practice.  It finishes
   reading at seeing non-decimal or EOF. *)

let read_cnf (f : string) = (
  let s = (Scanf.Scanning.from_file f) in
  let rec read_cnf_peek s = (
    (* No handling of eof. *)
    (Scanf.bscanf s "%0c" (fun x -> x))) in
  let rec read_cnf_input s = (
    (* No handling of eof. *)
    (Scanf.bscanf s "%c" (fun x -> x))) in
  let rec read_cnf_skip_to_eol s = (
    (* No handling of eof. *)
    (Scanf.bscanf s "%[^\n]"
	(fun x -> ()))) in
  let rec read_cnf_skip_whitespaces s = (
    (* No handling of eof. *)
    let c = (read_cnf_peek s) in
    match c with
      ' ' | '\t' | '\n' | '\r' ->
	ignore (read_cnf_input s) ;
	(read_cnf_skip_whitespaces s)
    | _ ->
	()) in
  let rec read_cnf_skip_preamble s = (
    (read_cnf_skip_whitespaces s) ;
    let c = (read_cnf_peek s) in
    if (c = 'c' || c = 'p') then
      (begin
	(read_cnf_skip_to_eol s) ;
	(read_cnf_skip_preamble s)
      end)) in
  let clauses = ref [] in
  try (
    (read_cnf_skip_preamble s) ;
    let literals = ref [] in
    while true do (
      let d = (Scanf.bscanf s " %d" (fun d -> d)) in
      if (d != 0) then
	literals := (!literals @ [d])
      else if (!literals = []) then
	raise End_of_file
      else
	(begin
	  let ps = (vec_create 0 0) in
	  (List.iter
	      (fun x -> (vec_push ps (lit_make ((abs x) - 1) (x > 0))))
	      !literals) ;
	  literals := [] ;
	  assert ((vec_size ps) <> 0) ;
	  clauses := (!clauses @ [ps]) ;
	end)
    ) done ;
    !clauses
  ) with
    End_of_file ->
      !clauses
  | Scanf.Scan_failure s ->
      !clauses)

(* Claims a CNF {cnf} read by {read_cnf} to the solver. *)

let claim_cnf (solver : solver_t) (cnf : lit_t vec list) = (
  (List.iter
      (fun ps ->
	(vec_iteri (fun i p -> (claim_var solver (lit_var p))) ps) ;
	let trivialsat = (claim_clause solver ps) in
	if (trivialsat = SAT) then
	  (*AHO*) (print_endline "warning: TRIVIAL SAT") ;
	if (trivialsat = UNSAT) then
	  (*AHO*) (print_endline "warning: TRIVIAL UNSAT"))
      cnf))

(* Writes out a SAT solution to a file {f}. *)

let write_cnf (sat : bool) (model : bool vec option) (f : string) = (
  let s = (open_out f) in
  if (sat = true) then
    (begin
      let m = (some_of model) in
      (output_string s "SAT\n") ;
      for x = 0 to ((vec_size m) - 1) do (
	(output_string s
	    (string_of_int (if (m.v.(x)) then (x + 1) else (- (x + 1))))) ;
	(output_string s " ")
      ) done ;
      (output_string s "0\n")
    end)
  else
    (begin
      (output_string s "UNSAT\n")
    end) ;
  (flush s) ;
  (close_out s))

let solve_cnf fi fo = (
  let solver = (solver_create ()) in
  let cnf = (read_cnf fi) in
  (claim_cnf solver cnf) ;
  let (sat, model) = (solver_solve solver (vec_create 0 0)) in
  (write_cnf (sat = SAT) model fo) ;
  if (sat = SAT) then
    let m = (some_of model) in
    ignore (check_model solver m true))

(* It tries all assignments for checking SAT solutions. *)

let rec stupid_sat (cnf : lit_t vec list) : unit = (
  let rec stupid_sat_check (assign : int) (p : lit_t) = (
    (((assign land (1 lsl (lit_var p))) <> 0) = (lit_sign p))) in
  let rec stupid_sat_clause (assign : int) (ps : lit_t list) = (
    (List.exists (stupid_sat_check assign) ps)) in
  let stupid_sat_cnf (assign : int) (cnf : lit_t list list) = (
    (List.for_all (stupid_sat_clause assign) cnf)) in
  let rec stupid_sat_loop (cnf : lit_t list list) (n : int) (assign : int) = (
    if (assign = n) then
      ()
    else
      (begin
	if (stupid_sat_cnf assign cnf) then
	  (begin
	    print_endline ("SAT: " ^ (string_of_int assign))
	  end) ;
	(stupid_sat_loop cnf n (assign + 1))
      end)) in
  let stupid_sat_max_var (cnf : lit_t list list) = (
    (List.fold_left
	(fun m ps -> (List.fold_left (fun n p -> (max n (lit_var p))) m ps))
	0 cnf)) in
  let c = (List.map vec_to_list cnf) in
  let v = (stupid_sat_max_var c) in
  (stupid_sat_loop c (1 lsl (v + 1)) 0))

(*
  open Minisat ;;
  let solver = (solver_create ()) ;;
  let cnf = (read_cnf "uf20-011.cnf") ;;
  (claim_cnf solver cnf) ;;
  (dump_all_models solver) ;;
  let x = [-1;2;3;-4;5;6;7;-8;9;10;11;12;-13;14;15;-16;-17;18;-19;20];;
  let (sat, model) = (solver_solve solver (vec_create 0 0)) ;;
  (check_model solver (some_of model)) ;;
  let trivialsat = (claim_clause solver (solver_reject solver (some_of model)) true) ;;
*)

(*
let _ = (solve_cnf "uf20-011.cnf" "uf20-011.out") ;;
let main () = (
  let files = ref [] : string list ref in
  let specs = [
    ("-sample", Arg.Unit (fun () -> ()), "for sample.\n") ;
  ] in
  let usage = "usage: ...\n" in
  let argfun = (fun a -> (files := (a :: (!files)))) in
  (Arg.parse specs argfun usage))
*)
