(* parser/smt.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Stupid Decision Procedure *)

(* NOTE: VERY SLOW.  Rewrite needed with efficient algorithms. *)

(* MEMO: Calling a SAT solver does not utilize restarting assignment
   (solver argument {assumeps}) yet. *)

(* MEMO: Abstraction to propositions is redundant, it is replaced
   simply by assigning a SAT variable directly.  Using two steps is
   just for tracing. *)

(* MEMO: Generated names (using {xgensym}) have prefix "P" for
   predicates, "Q" for equality predicates between UIF and arithmetic,
   and "U" for generated variables. *)

let smtTRACE_sat = ref false
let smtTRACE_cnf = ref true
let smtTRACE_result_cube = ref false

(* ================================================================ *)

open Num
open Form

module Sat = Minisat
module LTermMap = Form.LTermMap
module VarBag = Misc.Bag (Form.VarSet)
module FormBag = Misc.Bag (Form.FormSet)

let xTRACE (s : string) = (if (false) then (print_endline (";; SMT " ^ s)))
let xDEBUG (s : string) = (if (false) then (print_endline (";; SMT " ^ s)))
let xPRINT (f : bool) (s0 : string) (s1 : string) =
  (if (f) then (print_endline (";; SMT " ^ s0 ^ s1)))

(* Tests eq-ness, to make visibly distinguishable from '='. *)

let eql = (==)

(* NOTE: THERE IS {gcd_big_int} IN {big_int} BUT NO {lcm_big_int}. *)

let rec gcd (a : num) (b : num) : num = (
  if (lt_num a b) then
    (gcd b a)
  else
    let r = (mod_num a b) in
    if (eq_num r (num_of_int 0)) then
      b
    else
      (gcd b r))

let rec lcm (a : num) (b : num) : num = (
  (div_num (mult_num a b) (gcd a b)))

let rec list_lcm (l : num list) : num = (
  (List.fold_left lcm (num_of_int 1) l))

(* ================================================================ *)

(* Equality (TE) (with uninterpreted function) *)

(* Congruence Closure Procedure.  This part follows the algorithm in
   the paper "Simplify: A Theorem Prover for Program Checking" by
   Detlefs, Nelson, and Saxe; and the disjoint-set forests algorithm
   by Galler and Fischer. *)

(* Toplevel entry is {eqlsat_solve}. *)

(* Equality management table.  {enode_tbl_t} maps a term f(x1,x2...)
   to an enode.  {enode_var_tbl} maps a variable/function to an enode.
   {enode_next_id} is a counter for assigning an enode ID. *)

type enode_table_t = {
  enode_tbl : ((int list), enode_t) Hashtbl.t ;
  enode_var_tbl : (term_t, enode_t) Hashtbl.t ;
  mutable enode_next_id : int
}

and enode_t = {
  enode_id : int ;
  mutable enode_root : enode_t ;
  mutable enode_size_at_root : int ;
  mutable enode_parents_at_root : enode_t list ;
  mutable enode_forbids_at_root : enode_t list ;
  enode_subterms : enode_t list ;
  enode_term : term_t
  (*tz_unknown : tz_unknown_t *)
}

(* Printer for {enode_t}.  It prints a pair of id and root-id.  Note
   that the root-id may not be a proper representative node. *)

let enode_printer (f : Format.formatter) (n : enode_t) : unit = (
  (Format.fprintf f "<enode_t[%d/%d]>" n.enode_id n.enode_root.enode_id))
(*#install_printer enode_printer;;*)

(* Finds a representative of a set.  MEMO: It updates the root link
   only to the immediate parent, not to the representative (thus, not
   optimal). *)

let rec find_root (n : enode_t) : enode_t = (
  if (eql n.enode_root n) then
    n
  else
    begin
      if (n.enode_root != n.enode_root.enode_root) then
	n.enode_root <- n.enode_root.enode_root ;
      (find_root n.enode_root)
    end)

let make_union_at_root (r : enode_t) (s : enode_t) : unit = (
  s.enode_root <- r ;
  r.enode_size_at_root <- (s.enode_size_at_root + r.enode_size_at_root) ;
  r.enode_parents_at_root <- (s.enode_parents_at_root
			       @ r.enode_parents_at_root) ;
  r.enode_forbids_at_root <- (s.enode_forbids_at_root
			       @ r.enode_forbids_at_root) ;
  s.enode_size_at_root <- 0 ;
  s.enode_parents_at_root <- [] ;
  s.enode_forbids_at_root <- [])

(* Makes union of the two sets. *)

let make_union (n0 : enode_t) (n1 : enode_t) : unit = (
  let (r0, r1) = ((find_root n0), (find_root n1)) in
  let (r, s) =
    if (r0.enode_size_at_root > r1.enode_size_at_root) then
      (r0, r1)
    else
      (r1, r0) in
  (make_union_at_root r s))

let make_enode_table () : enode_table_t = (
  {
    enode_tbl = (Hashtbl.create 40) ;
    enode_var_tbl = (Hashtbl.create 40) ;
    enode_next_id = 1
  })

let make_enode_raw (tbl : enode_table_t) (t : term_t) (children : enode_t list) : enode_t = (
  let id = tbl.enode_next_id in
  tbl.enode_next_id <- (id + 1) ;
  let rec n = {
    enode_id = id ;
    enode_root = n ;
    enode_size_at_root = 1 ;
    enode_parents_at_root = [] ;
    enode_forbids_at_root = [] ;
    enode_subterms = children ;
    enode_term = t
  } in
  n)

(* Makes an enode for a term.  It returns a representative node. *)

let rec make_enode (tbl : enode_table_t) (t : term_t) : enode_t = (
  match t with
    (LTERM _) -> (assert false)
  | (ARITH _) ->
      (failwith "arithmetic term appears in t-of-equality")
  | (VAR _) | (FUN _) ->
      if (Hashtbl.mem tbl.enode_var_tbl t) then
	(find_root (Hashtbl.find tbl.enode_var_tbl t))
      else
	(begin
	  let n = (make_enode_raw tbl t []) in
	  (Hashtbl.add tbl.enode_var_tbl t n) ;
	  n
	end)
  | (UIFUN l) ->
      let subterms = (List.map (make_enode tbl) l) in
      let k = (List.map (fun c -> c.enode_id) subterms) in
      if (Hashtbl.mem tbl.enode_tbl k) then
	(find_root (Hashtbl.find tbl.enode_tbl k))
      else
	(begin
	  let n = (make_enode_raw tbl t subterms) in
	  (List.iter
	      (fun c ->
		(c.enode_parents_at_root <- (n :: c.enode_parents_at_root)))
	      (List.tl subterms)) ;
	  (Hashtbl.add tbl.enode_tbl k n) ;
	  n
	end))

let deassociate_enode (tbl : enode_table_t) (n : enode_t) : unit = (
  let k = (List.map (fun c -> (find_root c).enode_id) n.enode_subterms) in
  (Hashtbl.remove tbl.enode_tbl k))

(* Reassociates a node, called when some of its subterms get new
   equivalence association.  It returns a new implied equivalence as a
   list, when it is associated to. *)

let associate_enode (tbl : enode_table_t) (n : enode_t) : (enode_t * enode_t) list = (
  let k = (List.map (fun c -> (find_root c).enode_id) n.enode_subterms) in
  if (Hashtbl.mem tbl.enode_tbl k) then
    [((Hashtbl.find tbl.enode_tbl k), n)]
  else
    (begin
      (Hashtbl.add tbl.enode_tbl k n) ;
      []
    end))

(* Chooses a larger one as a representative for the two of the
   equivalence classes. *)

let choose_new_root (r0 : enode_t) (r1 : enode_t) : (enode_t * enode_t) = (
  if (r0.enode_size_at_root > r1.enode_size_at_root) then
    (r0, r1)
  else
    (r1, r0))

(* Checks satisfiability of conjunctions equality/disequality. *)

let rec eqlsat_solve (l : form_t list) : Sat.sat_result_t = (
  let tbl = (make_enode_table ()) in
  let l1 = (eqlsat_assert_literals tbl l) in
  if (List.for_all (fun r -> (r == Sat.SAT)) l1) then
    Sat.SAT
  else
    Sat.UNSAT)

(* Adds equality between nodes. *)

and eqlsat_assert_eq (tbl : enode_table_t) (l : (enode_t * enode_t) list) : Sat.sat_result_t = (
  match l with
    [] ->
      Sat.SAT
  | ((n0, n1) :: tl) ->
      let (r0, r1) = ((find_root n0), (find_root n1)) in
      if ((List.mem r1 r0.enode_forbids_at_root)
	   || (List.mem r0 r1.enode_forbids_at_root)) then
	Sat.UNSAT
      else
	(begin
	  let (r, s) = (choose_new_root r0 r1) in
	  let ups = s.enode_parents_at_root in
	  (List.iter (deassociate_enode tbl) ups) ;
	  (make_union_at_root r s) ;
	  let eqs = (List.flatten (List.map (associate_enode tbl) ups)) in
	  (eqlsat_assert_eq tbl (tl @ eqs))
	end))

(* Adds disequality between nodes. *)

and eqlsat_assert_neq (tbl : enode_table_t) (n0 : enode_t) (n1 : enode_t) : Sat.sat_result_t = (
  let (r0, r1) = ((find_root n0), (find_root n1)) in
  if (eql r0 r1) then
    Sat.UNSAT
  else
    (begin
      (r0.enode_forbids_at_root <- (r1 :: r0.enode_forbids_at_root)) ;
      (r1.enode_forbids_at_root <- (r0 :: r1.enode_forbids_at_root)) ;
      Sat.SAT
    end))

and eqlsat_assert_literals (tbl : enode_table_t) (l : form_t list) : Sat.sat_result_t list = (
  (List.map
      (fun p ->
	(match p with
	  PRED (EQ, t1, t2) | NOT PRED (NE, t1, t2) ->
	    let n1 = (make_enode tbl t1) in
	    let n2 = (make_enode tbl t2) in
	    (eqlsat_assert_eq tbl [(n1, n2)]) ;
	| PRED (NE, t1, t2) | NOT PRED (EQ, t1, t2) ->
	    let n1 = (make_enode tbl t1) in
	    let n2 = (make_enode tbl t2) in
	    (eqlsat_assert_neq tbl n1 n2) ;
	| _ -> (assert false)))
      l))

(* (TRACE) Displays a representative of equivalence. *)

let enode_term (n : enode_t) = (term_to_string (find_root n).enode_term)
let enode_subs (n : enode_t) = ((find_root n).enode_subterms)

(* ================================================================ *)

(* Linear Integer (Presburger) Arithmetic (TZ) *)

(* The Cooper's Algorithm.  This part follows the description as
   presented in the paper "Complete Integer Decision Procedures as
   Derived Rules in HOL" by Norrish. *)

(* This handles linear combination on integer.  Toplevel entry is
   {liasat_solve}. *)

(* CONVERSION TO A NORMAL FORM: universals eliminated; negations only
   before equals, divides, or existentials; relations only with LT.
   Thus, leafs are: "cx<e", "e<cx", "cx=e", "!(cx=e)", "(c|dx+e)", or
   "!(c|dx+e)", where c and d are positive constant numerals. *)

(* **** Simple Normalization of Terms **** *)

(* Converts a term to a linear_term. *)

let rec additive_form (t : term_t) : linear_term_t = (
  (additive_form_loop (num_of_int 1) t LTermMap.empty))

(* Checks if the term is a coefficient (constant).  It returns {None}
   if term is non-constant, or {(Some_(m*t+a))}. *)

and additive_form_coeff (m : num) (a : num) (t : term_t) : num option = (
  match t with
    (LTERM e) -> (assert false)
  | (VAR w) ->
      None
  | (FUN _) -> (assert false)
  | (UIFUN _) -> (failwith "function appears in t-of-integer")
  | (ARITH (CVT (i, v))) -> (assert false)
  | (ARITH ZERO) ->
      (Some a)
  | (ARITH (NUM i)) ->
      (Some (add_num (mult_num m i) a))
  | (ARITH (ADD (t1, t2))) ->
      let v1 = (additive_form_coeff m a t1) in
      if (v1 == None) then
	None
      else
	(additive_form_coeff m (Sat.some_of v1) t2)
  | (ARITH (MUL (t1, t2))) ->
      let v1 = (additive_form_coeff (num_of_int 1) (num_of_int 0) t1) in
      if (v1 == None) then
	None
      else
	(additive_form_coeff (mult_num m (Sat.some_of v1)) a t2)
  | (ARITH (NEG t1)) ->
      (additive_form_coeff (minus_num m) a t1))

and additive_form_coeff_int (m : int) (a : int) (t : term_t) : num option = (
  (additive_form_coeff (num_of_int m) (num_of_int a) t))

(* Adds {(m*t)} to {e}. *)

and additive_form_loop (m : num) (t : term_t) (e : linear_term_t) : linear_term_t = (
  match t with
    (LTERM e2) ->
      (term_add e (term_mulc e2 m))
  | (VAR v) ->
      (term_add_coeff e v m)
  | (FUN _) -> (failwith "function appears in t-of-integer")
  | (UIFUN _) -> (failwith "function appears in t-of-integer")
  | (ARITH (CVT (i, v))) ->
      (term_add_coeff e v (mult_num i m))
  | (ARITH ZERO) ->
      e
  | (ARITH (NUM i)) ->
      (term_add_coeff e linear_term_constant (mult_num m i))
  | (ARITH (ADD (t1, t2))) ->
      (additive_form_loop m t2 (additive_form_loop m t1 e))
  | (ARITH (MUL (t1, t2))) ->
      (begin
	let c1 = (additive_form_coeff_int 1 0 t1) in
	let c2 = (additive_form_coeff_int 1 0 t2) in
	match (c1, c2) with
	  (None, None) -> (failwith "non-linear in t-of-integer")
	| (Some i, None) ->
	    (additive_form_loop (mult_num m i) t2 e)
	| (None, Some j) ->
	    (additive_form_loop (mult_num m j) t1 e)
	| (Some i, Some j) ->
	    (term_add_coeff e linear_term_constant (mult_num m (mult_num i j)))
      end)
  | (ARITH (NEG t1)) ->
      (additive_form_loop (minus_num m) t1 e))

(* **** Simple Normalization of Forms **** *)

(* Finds free variables, treating {vs} as bound variables. *)

let rec find_free_variables (vs : StringSet.t) (f : form_t) : StringSet.t = (
   match f with
     TT | FF ->
       vs
   | (PROP _) ->
       vs
   | (PRED (p, t1, t2)) ->
       (List.fold_left find_free_variables_term vs [t1 ; t2])
   | (NOT f1) ->
       (find_free_variables vs f1)
   | (AND l) ->
       (List.fold_left find_free_variables vs l)
   | (OR l) ->
       (List.fold_left find_free_variables vs l)
   | (IMP (f1, f2)) ->
       (find_free_variables (find_free_variables vs f1) f2)
   | (IFF (f1, f2)) ->
       (find_free_variables (find_free_variables vs f1) f2)
   | (FORALL (v, f1)) ->
       (find_free_variables (StringSet.add v vs) f1)
   | (EXISTS (v, f1)) ->
       (find_free_variables (StringSet.add v vs) f1))

and find_free_variables_term (vs : StringSet.t) (t : term_t) : StringSet.t = (
  match t with
    (LTERM e) ->
      (LTermMap.fold (fun k v ss -> (StringSet.add k ss)) e vs)
  | (VAR v) ->
      (StringSet.add v vs)
  | (FUN _) ->
      vs
  | (UIFUN l) ->
      (List.fold_left find_free_variables_term vs l)
  | (ARITH (CVT (i, v))) ->
      (StringSet.add v vs)
  | (ARITH ZERO) ->
      vs
  | (ARITH (NUM i)) ->
      vs
  | (ARITH (MUL (t1, t2))) ->
      (find_free_variables_term (find_free_variables_term vs t1) t2)
  | (ARITH (ADD (t1, t2))) ->
      (find_free_variables_term (find_free_variables_term vs t1) t2)
  | (ARITH (NEG t1)) ->
      (find_free_variables_term vs t1))

(* Normalizes a form, with propagating negation, and eliminating
   forall, implies, and iff.  After normalization, forms consists of
   {OR}, {AND}, {EXISTS}, {NOT(EXISTS)}, {PRED:EQ(zero)},
   {(PRED:NE(zero))}, {PRED:LT(zero)}, {PRED:DIVS}, {NOT(PRED:DIVS)}.
   {DIVS} has positive divider. *)

let rec normalize_form (positive : bool) (f : form_t) : form_t = (
  match f with
    TT ->
      if (positive) then TT else FF
  | FF ->
      if (positive) then FF else TT
  | (PROP _) ->
      (failwith "proposition appears in t-of-equality")
  | (PRED (p, t1, t2)) ->
      (normalize_pred positive f)
  | (NOT f1) ->
      if (positive) then
	(normalize_form false f1)
      else
	(normalize_form true f1)
  | (AND l) ->
      if (positive) then
	(AND (List.map (normalize_form true) l))
      else
	(OR (List.map (normalize_form false) l))
  | (OR l) ->
      if (positive) then
	(OR (List.map (normalize_form true) l))
      else
	(AND (List.map (normalize_form false) l))
  | (IMP (f1, f2)) ->
      if (positive) then
	(OR [(normalize_form false f1) ;
	     (normalize_form true f2)])
      else
	(AND [(normalize_form true f1) ;
	      (normalize_form false f2)])
  | (IFF (f1, f2)) ->
      if (positive) then
	(AND [(normalize_form true (IMP (f1, f2))) ;
	      (normalize_form true (IMP (f2, f1)))])
      else
	(OR [(normalize_form false (IMP (f1, f2))) ;
	     (normalize_form false (IMP (f2, f1)))])
  | (FORALL (v, f)) ->
      if (positive) then
	(NOT (EXISTS (v, (normalize_form false f))))
      else
	(EXISTS (v, (normalize_form false f)))
  | (EXISTS (v, f)) ->
      if (positive) then
	(EXISTS (v, (normalize_form true f)))
      else
	(NOT (EXISTS (v, (normalize_form true f)))))

and normalize_pred (positive : bool) (f : form_t) : form_t = (
  match f with
    (PRED (EQ, t1, t2)) ->
      let e = (term_sub (additive_form t1) (additive_form t2)) in
      if (not positive) then
	(PRED (NE, (LTERM e), (ARITH ZERO)))
      else
	(PRED (EQ, (LTERM e), (ARITH ZERO)))
  | (PRED (NE, t1, t2)) ->
      (normalize_pred (not positive) (PRED (EQ, t1, t2)))
  | (PRED (LT, t1, t2)) ->
      if (not positive) then
	(normalize_pred true (PRED (GE, t1, t2)))
      else
	(begin
	  let e = (term_sub (additive_form t1) (additive_form t2)) in
	  (PRED (LT, (LTERM e), (ARITH ZERO)))
	end)
  | (PRED (GT, t1, t2)) ->
      if (not positive) then
	(normalize_pred true (PRED (LE, t1, t2)))
      else
	(begin
	  let e = (term_sub (additive_form t2) (additive_form t1)) in
	  (PRED (LT, (LTERM e), (ARITH ZERO)))
	end)
  | (PRED (GE, t1, t2)) ->
      if (not positive) then
	(normalize_pred true (PRED (LT, t1, t2)))
      else
	(begin
	  (OR [(normalize_pred true (PRED (GT, t1, t2))) ;
	       (normalize_pred true (PRED (EQ, t1, t2)))])
	end)
  | (PRED (LE, t1, t2)) ->
      if (not positive) then
	(normalize_pred true (PRED (GT, t1, t2)))
      else
	(begin
	  (OR [(normalize_pred true (PRED (LT, t1, t2))) ;
	       (normalize_pred true (PRED (EQ, t1, t2)))])
	end)
  | (PRED (DIVS, t1, t2)) ->
      if (not positive) then
	(NOT (normalize_pred true f))
      else
	(begin
	  let e1 = (additive_form t1) in
	  if (not (term_constantp e1)) then
	    (failwith "predicate divides with non-constant") ;
	  let d = (term_constant_part e1) in
	  let e =
	    if (gt_num d (num_of_int 0)) then
	      (additive_form t2)
	    else
	      (additive_form (ARITH (NEG t2)))
	  in
	  (PRED (DIVS, (ARITH (NUM (abs_num d))), (LTERM e)))
	end)
  | TT | FF -> (assert false)
  | PROP _ -> (assert false)
  | NOT _ -> (assert false)
  | AND _ -> (assert false)
  | OR _ -> (assert false)
  | IMP _ -> (assert false)
  | IFF _ -> (assert false)
  | FORALL _ -> (assert false)
  | EXISTS _ -> (assert false))

(* Splits terms into a coefficient-variable pair and other terms.  It
   returns (c,s,e), where the term equals to {(c*v-e)} with {s} is the
   sign of {c}, {c} is positive, and {e} is negated.  It returns with
   {c} zero, when terms do not include a given variable. *)

let split_term_for_variable (v : varname_t) (e : linear_term_t) : (num * bool * linear_term_t) = (
  if (not (LTermMap.mem v e)) then
    ((num_of_int 0), true, e)
  else if (eq_num (LTermMap.find v e) (num_of_int 0)) then
    ((num_of_int 0), true, e)
  else
    (begin
      let c = (term_coeff v e) in
      let x = (term_remove v e) in
      if (gt_num c (num_of_int 0)) then
	(c, true, (term_neg x))
      else
	((minus_num c), false, x)
    end))

(* Extracts terms with a given variable from predicates and its
   argument terms.  It normalizes into the forms: {cx<e}, {cx>e},
   {cx=e}, {!cx=e}, {d|cx+e}, {!d|cx+e}.  Or, the cases when a
   variable is not included: {t<0}, {t>0}, {t=0}, {!t=0}, {d|t},
   {!d|t}. *)

let rec pivoting_form_for_variable (v : varname_t) (f : form_t) : form_t = (
  match f with
    FF | TT ->
      (*AHO*)(*(assert false)*)
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (PRED _) ->
      (pivoting_pred_for_variable v f)
  | (NOT f1) ->
      (NOT (pivoting_form_for_variable v f1))
  | (AND l) ->
      (AND (List.map (pivoting_form_for_variable v) l))
  | (OR l) ->
      (OR (List.map (pivoting_form_for_variable v) l)))

and pivoting_pred_for_variable (v : varname_t) (f : form_t) : form_t = (
  match f with
    (PRED (EQ, (LTERM t1), (ARITH ZERO))) ->
      (begin
	let (c, s, e) = (split_term_for_variable v t1) in
	if (eq_num c (num_of_int 0)) then
	  f
	else
	  (PRED (EQ, (ARITH (CVT (c, v))), (LTERM e)))
      end)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) ->
      (begin
	let (c, s, e) = (split_term_for_variable v t1) in
	if (eq_num c (num_of_int 0)) then
	  f
	else
	  (PRED (NE, (ARITH (CVT (c, v))), (LTERM e)))
      end)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) ->
      (begin
	let (c, s, e) = (split_term_for_variable v t1) in
	if (eq_num c (num_of_int 0)) then
	  f
	else if (s) then
	  (PRED (LT, (ARITH (CVT (c, v))), (LTERM e)))
	else
	  (PRED (GT, (ARITH (CVT (c, v))), (LTERM e)))
      end)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) ->
      (begin
	let (c, s, e) = (split_term_for_variable v t1) in
	if (eq_num c (num_of_int 0)) then
	  f
	else if (s) then
	  (PRED (GT, (ARITH (CVT (c, v))), (LTERM e)))
	else
	  (PRED (LT, (ARITH (CVT (c, v))), (LTERM e)))
      end)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) ->
      (begin
	let (c, s, e) = (split_term_for_variable v t2) in
	if (eq_num c (num_of_int 0)) then
	  f
	else
	  (PRED (DIVS, (ARITH (NUM d)),
		(ARITH (ADD ((ARITH (CVT (c, v))), (LTERM e))))))
      end)
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, _, _)) -> (assert false)
  | (PRED (NE, _, _)) -> (assert false)
  | (PRED (LT, _, _)) -> (assert false)
  | (PRED (GT, _, _)) -> (assert false)
  | (PRED (DIVS, _, _)) -> (assert false)
  | TT | FF ->
      (*AHO*)(*(assert false)*)
      f
  | PROP _ -> (assert false)
  | NOT _ -> (assert false)
  | AND _ -> (assert false)
  | OR _ -> (assert false)
  | IMP _ -> (assert false)
  | IFF _ -> (assert false)
  | FORALL _ -> (assert false)
  | EXISTS _ -> (assert false))

(* Takes LCM of coefficients of a variable.  It takes a form output
   from {pivoting_form_for_variable}. *)

let rec find_coeff_lcm (f : form_t) : num = (
  match f with
    TT | FF ->
      (num_of_int 1)
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (find_coeff_lcm f1)
  | (AND l) | (OR l) ->
      (list_lcm (List.map find_coeff_lcm l))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) -> (num_of_int 1)
  | (PRED (EQ, (ARITH (CVT (c, v))), (LTERM e))) -> c
  | (PRED (NE, (ARITH (CVT (c, v))), (LTERM e))) -> c
  | (PRED (LT, (ARITH (CVT (c, v))), (LTERM e))) -> c
  | (PRED (GT, (ARITH (CVT (c, v))), (LTERM e))) -> c
  | (PRED (DIVS, (ARITH (NUM d)),
	 (ARITH (ADD ((ARITH (CVT (c, v))), (LTERM e)))))) -> c
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Removes coefficients of a variable by dividing LCM of coefficients.
   It takes a form output from {pivoting_form_for_variable}. It
   generates a form of unit-coefficient. *)

let rec mutliply_coeff_by_lcm (m : num) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (NOT (mutliply_coeff_by_lcm m f1))
  | (AND l) ->
      (AND (List.map (mutliply_coeff_by_lcm m) l))
  | (OR l) ->
      (OR (List.map (mutliply_coeff_by_lcm m) l))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) -> f
  | (PRED (EQ, (ARITH (CVT (c, v))), (LTERM e))) ->
      (PRED (EQ, (VAR v), (LTERM (term_mulc e (quo_num m c)))))
  | (PRED (NE, (ARITH (CVT (c, v))), (LTERM e))) ->
      (PRED (NE, (VAR v), (LTERM (term_mulc e (quo_num m c)))))
  | (PRED (LT, (ARITH (CVT (c, v))), (LTERM e))) ->
      (PRED (LT, (VAR v), (LTERM (term_mulc e (quo_num m c)))))
  | (PRED (GT, (ARITH (CVT (c, v))), (LTERM e))) ->
      (PRED (GT, (VAR v), (LTERM (term_mulc e (quo_num m c)))))
  | (PRED (DIVS, (ARITH (NUM d)),
	 (ARITH (ADD ((ARITH (CVT (c, v))), (LTERM e)))))) ->
      (PRED (DIVS, (ARITH (NUM (mult_num d (quo_num m c)))),
	    (ARITH (ADD ((VAR v), (LTERM (term_mulc e (quo_num m c))))))))
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Takes a divider.  It takes a normalized form (after converting
   coefficient to one). *)

let rec find_divs_lcm (f : form_t) : num = (
  match f with
    TT | FF ->
      (num_of_int 1)
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (find_divs_lcm f1)
  | (AND l) | (OR l) ->
      (list_lcm (List.map find_divs_lcm l))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) -> (num_of_int 1)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) -> (num_of_int 1)
  | (PRED (EQ, (VAR v), (LTERM e))) -> (num_of_int 1)
  | (PRED (NE, (VAR v), (LTERM e))) -> (num_of_int 1)
  | (PRED (LT, (VAR v), (LTERM e))) -> (num_of_int 1)
  | (PRED (GT, (VAR v), (LTERM e))) -> (num_of_int 1)
  | (PRED (DIVS, (ARITH (NUM d)), (ARITH (ADD ((VAR v), (LTERM e)))))) -> d
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Removes coefficients to the variable {v} in terms in a normalized
   form (with quantification outermost).  It calls
   {mutliply_coeff_by_lcm} anyway to simplify the LHS of GT/LT. *)

let factor_coefficient_with_lcm (v : varname_t) (f : form_t) : form_t = (
  let f2 = (pivoting_form_for_variable v f) in
  let m = (find_coeff_lcm f2) in
  if (not (eq_num m (num_of_int 1))) then
    (begin
      let f3 = (mutliply_coeff_by_lcm m f2) in
      let f4 = (PRED (DIVS, (ARITH (NUM m)),
		     (ARITH (ADD ((VAR v), (LTERM (term_zero))))))) in
      (AND [f3 ; f4])
    end)
  else
    (mutliply_coeff_by_lcm (num_of_int 1) f2))

(* Makes upper/lower bounds as A-set/B-set. *)

let rec make_aset_bset ((a, b) : (LTermSet.t * LTermSet.t)) (f : form_t) : (LTermSet.t * LTermSet.t) = (
  match f with
    TT | FF -> (a, b)
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (make_aset_bset (a, b) f1)
  | (AND l) ->
      (List.fold_left make_aset_bset (a, b) l)
  | (OR l) ->
      (List.fold_left make_aset_bset (a, b) l)
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) -> (a, b)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) -> (a, b)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) -> (a, b)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) -> (a, b)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) -> (a, b)
  | (PRED (EQ, (VAR v), (LTERM e))) ->
      ((LTermSet.add (term_addc_int e 1) a),
      (LTermSet.add (term_addc_int e (-1)) b))
  | (PRED (NE, (VAR v), (LTERM e))) ->
      ((LTermSet.add e a), (LTermSet.add e b))
  | (PRED (LT, (VAR v), (LTERM e))) ->
      ((LTermSet.add e a), b)
  | (PRED (GT, (VAR v), (LTERM e))) ->
      (a, (LTermSet.add e b))
  | (PRED (DIVS, (ARITH (NUM d)), (ARITH (ADD ((VAR v), (LTERM e)))))) ->
      (a, b)
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

let rec infinity_form (positive : bool) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (NOT (infinity_form positive f1))
  | (AND l) ->
      (AND (List.map (infinity_form positive) l))
  | (OR l) ->
      (OR (List.map (infinity_form positive) l))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) -> f
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) -> f
  | (PRED (EQ, (VAR v), (LTERM e))) ->
      FF
  | (PRED (NE, (VAR v), (LTERM e))) ->
      TT
  | (PRED (LT, (VAR v), (LTERM e))) ->
      (if (positive) then FF else TT)
  | (PRED (GT, (VAR v), (LTERM e))) ->
      (if (positive) then TT else FF)
  | (PRED (DIVS, (ARITH (NUM d)), (ARITH (ADD ((VAR v), (LTERM e)))))) ->
      f
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Substitutes a variable {v} with a term {c}, with calling reducer.
   Reducer be either {reduce_onelevel_form} for getting a simplified
   form, or {identity_form} for debugging. *)

let rec substitute_and_reduce (reducer : form_t -> form_t) (v : varname_t) (c : linear_term_t) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (NOT (substitute_and_reduce reducer v c f1))
  | (AND l) ->
      (reducer (AND (List.map (substitute_and_reduce reducer v c) l)))
  | (OR l) ->
      (reducer (OR (List.map (substitute_and_reduce reducer v c) l)))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t2))) ->
      (reducer f)
  | (PRED (EQ, (VAR v), (LTERM e))) ->
      (reducer (PRED (EQ, (LTERM (term_sub c e)), (ARITH ZERO))))
  | (PRED (NE, (VAR v), (LTERM e))) ->
      (reducer (PRED (NE, (LTERM (term_sub c e)), (ARITH ZERO))))
  | (PRED (LT, (VAR v), (LTERM e))) ->
      (reducer (PRED (LT, (LTERM (term_sub c e)), (ARITH ZERO))))
  | (PRED (GT, (VAR v), (LTERM e))) ->
      (reducer (PRED (GT, (LTERM (term_sub c e)), (ARITH ZERO))))
  | (PRED (DIVS, (ARITH (NUM d)), (ARITH (ADD ((VAR v), (LTERM e)))))) ->
      (reducer (PRED (DIVS, (ARITH (NUM d)), (LTERM (term_add e c)))))
  | (PRED (EQ, _, _)) -> (assert false)
  | (PRED (NE, _, _)) -> (assert false)
  | (PRED (LT, _, _)) -> (assert false)
  | (PRED (GT, _, _)) -> (assert false)
  | (PRED (DIVS, _, _)) -> (assert false))

let identity_form (f : form_t) = f

let reduce_onelevel_form (f : form_t) : form_t = (
  match f with
    TT | FF ->
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (begin
	match f1 with
	  TT -> FF
	| FF -> TT
	| _ -> f
      end)
  | (AND l) ->
      (begin
	if (List.exists (fun x -> (x == FF)) l) then
	  FF
	else
	  (begin
	    let l2 = (List.filter (fun x -> (x != TT)) l) in
	    if (l2 == []) then
	      TT
	    else
	      (AND l2)
	  end)
      end)
  | (OR l) ->
      (begin
	if (List.exists (fun x -> (x == TT)) l) then
	  TT
	else
	  (begin
	    let l2 = (List.filter (fun x -> (x != FF)) l) in
	    if (l2 == []) then
	      FF
	    else
	      (OR l2)
	  end)
      end)
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) ->
      (begin
	match (term_check_constant t1) with
	  None -> f
	| Some v -> (if (eq_num v (num_of_int 0)) then TT else FF)
      end)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) ->
      (begin
	match (term_check_constant t1) with
	  None -> f
	| Some v -> (if (eq_num v (num_of_int 0)) then FF else TT)
      end)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) ->
      (begin
	match (term_check_constant t1) with
	  None -> f
	| Some v -> (if (lt_num v (num_of_int 0)) then TT else FF)
      end)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) ->
      (begin
	match (term_check_constant t1) with
	  None -> f
	| Some v -> (if (gt_num v (num_of_int 0)) then TT else FF)
      end)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t1))) ->
      (begin
	match (term_check_constant t1) with
	  None -> f
	| Some v -> (if (eq_num (mod_num v d) (num_of_int 0)) then TT else FF)
      end)
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Reduces a form by simplifying predicates and its constant
   arithmetic terms.  Call it after substitution. *)

let rec reduce_form (f : form_t) : form_t = (
  let reducer = reduce_onelevel_form in
  match f with
    TT | FF ->
      f
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false)
  | (NOT f1) ->
      (reducer (NOT (reduce_form f1)))
  | (AND l) ->
      (reducer (AND (List.map reduce_form l)))
  | (OR l) ->
      (reducer (OR (List.map reduce_form l)))
  | (PRED (LE, _, _)) -> (assert false)
  | (PRED (GE, _, _)) -> (assert false)
  | (PRED (EQ, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (NE, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (LT, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (GT, (LTERM t1), (ARITH ZERO))) ->
      (reducer f)
  | (PRED (DIVS, (ARITH (NUM d)), (LTERM t1))) ->
      (reducer f)
  | PRED (EQ, _, _) -> (assert false)
  | PRED (NE, _, _) -> (assert false)
  | PRED (LT, _, _) -> (assert false)
  | PRED (GT, _, _) -> (assert false)
  | PRED (DIVS, _, _) -> (assert false))

(* Makes a list of forms by replacing a variable with existential
   candidate values, where {positive=true} is for the A-set,
   {positive=false} is for the B-set. *)

let make_eliminate_preds (v : varname_t) (f2 : form_t) (positive : bool) (delta : int) (tset : LTermSet.t) : form_t = (
  let reducer = (if (true) then reduce_onelevel_form else identity_form) in
  let forms = ref [] in
  let pinf = (infinity_form positive f2) in
  for i = 1 to delta do (
    let c = (if (positive) then (-i) else i) in
    let t = (term_zero) in
    forms := ((substitute_and_reduce reducer v (term_addc_int t c) pinf)
	       :: !forms)
  ) done ;
  for i = 1 to delta do (
    let c = (if (positive) then (-i) else i) in
    (LTermSet.iter
	(fun t ->
	  forms := ((substitute_and_reduce reducer v (term_addc_int t c) f2)
		     :: !forms))
	tset)
  ) done ;
  (* Reduce result because it is unreduced when no {delta} is specified. *)
  if (delta <= 0) then
    (reduce_form (OR !forms))
  else
    (reducer (OR !forms)))

(* Eliminates an existential with {v} (for innermost one). *)

let eliminate_innermost_existential (v : varname_t) (f : form_t) : form_t = (
  (*let f1 = (normalize_form true f) in*)
  let f2 = (factor_coefficient_with_lcm v f) in
  let delta = (coerce_to_int (find_divs_lcm f2)) in
  let (aset, bset) = (make_aset_bset ((LTermSet.empty), (LTermSet.empty)) f2) in
  if ((LTermSet.cardinal aset) > (LTermSet.cardinal bset)) then
    (make_eliminate_preds v f2 false delta bset)
  else
    (make_eliminate_preds v f2 true delta aset))

(* Eliminates existentials from innermost ones. *)

let rec eliminate_existential (f : form_t) : form_t = (
  match f with
    TT | FF -> (assert false)
  | PROP _ -> (assert false)
  | IMP _ | IFF _ | FORALL _ -> (assert false)
  | (PRED _) ->
      f
  | (NOT f1) ->
      (NOT (eliminate_existential f1))
  | (AND l) ->
      (AND (List.map eliminate_existential l))
  | (OR l) ->
      (OR (List.map eliminate_existential l))
  | (EXISTS (v, f1)) ->
      let f2 = (eliminate_existential f1) in
      (eliminate_innermost_existential v f2))

(* Adds universal quantifications outermost, then checks its
   satisfiability. *)

let satisfy_forall (f : form_t) : form_t = (
  let vs = (StringSet.elements (find_free_variables (StringSet.empty) f)) in
  let f1 = (normalize_form false f) in
  let f2 = (List.fold_left (fun ff v -> (EXISTS (v, ff))) f1 vs) in
  (reduce_onelevel_form (eliminate_existential (NOT f2))))

(* Adds existential quantifications outermost, then checks its
   satisfiability. *)

let satisfy_exists (f : form_t) : form_t = (
  let vs = (StringSet.elements (find_free_variables (StringSet.empty) f)) in
  let f1 = (normalize_form true f) in
  let f2 = (List.fold_left (fun ff v -> (EXISTS (v, ff))) f1 vs) in
  (eliminate_existential f2))

let liasat_solve (l : form_t list) : Sat.sat_result_t = (
  if (l == []) then
    Sat.SAT
  else
    (begin
      match (satisfy_exists (AND l)) with
	TT -> Sat.SAT
      | FF -> Sat.UNSAT
      | f ->
	  xDEBUG ("liasat_solve("
		   ^ (String.concat "&&" (List.map form_to_string l))
		   ^ ")=(" ^ (Form.form_to_string f) ^ ")") ;
	  Sat.UNKNOWN
    end))

(* ================================================================ *)

(* CNF Conversion *)

(* NOTE: This is VERY SLOW.  Conversion is done by distributivity, and
   the results have expensive sizes.  It shall be simplified by
   equi-satisfiability. *) (*SLOW*)

let rec convert_to_cnf (f : form_t) : form_t = (
  (pseudo_reduce_cnf
      (convert_to_cnf_form (convert_to_cnf_propagate_not true f))))

(* Propagates not in the form.  It removes {IMP} and {IFF}, and some
   trivial forms of {AND} and {OR}. *)

and convert_to_cnf_propagate_not (positive : bool) (f : form_t) : form_t = (
  match f with
    TT ->
      (if (positive) then TT else FF)
  | FF ->
      (if (positive) then FF else TT)
  | (PROP _) | (PRED _) ->
      (if (positive) then f else (NOT f))
  | (NOT f1) ->
      (convert_to_cnf_propagate_not (not positive) f1)
  | (AND l) ->
      if (positive) then
	(Form.pseudo_conjoin_forms
	    (List.map (convert_to_cnf_propagate_not true) l))
      else
	(Form.pseudo_disjoin_forms
	    (List.map (convert_to_cnf_propagate_not false) l))
  | (OR l) ->
      if (positive) then
	(Form.pseudo_disjoin_forms
	    (List.map (convert_to_cnf_propagate_not true) l))
      else
	(Form.pseudo_conjoin_forms
	    (List.map (convert_to_cnf_propagate_not false) l))
  | (IMP (f1, f2)) ->
      if (positive) then
	(Form.pseudo_disjoin
	    (convert_to_cnf_propagate_not false f1)
	    (convert_to_cnf_propagate_not true f2))
      else
	(Form.pseudo_conjoin
	    (convert_to_cnf_propagate_not true f1)
	    (convert_to_cnf_propagate_not false f2))
  | (IFF (f1, f2)) ->
      if (positive) then
	(Form.pseudo_conjoin
	    (convert_to_cnf_propagate_not true (IMP (f1, f2)))
	    (convert_to_cnf_propagate_not true (IMP (f2, f1))))
      else
	(Form.pseudo_disjoin
	    (convert_to_cnf_propagate_not false (IMP (f1, f2)))
	    (convert_to_cnf_propagate_not false (IMP (f2, f1))))
  | FORALL _ -> (failwith "quantifiers not allowed")
  | EXISTS _ -> (failwith "quantifiers not allowed"))

and convert_to_cnf_form (f : form_t) : form_t = (
  match f with
    TT | FF | (PROP _) | (PRED _) | (NOT _) ->
      (AND [(OR [f])])
  | (AND l) ->
      let x = (List.map convert_to_cnf_form l) in
      (AND (List.flatten (List.map convert_to_cnf_conjuncts x)))
  | (OR l) ->
      let x = (List.map convert_to_cnf_form l) in
      (List.fold_left
	  convert_to_cnf_distribute_or
	  (AND [OR [FF]]) x)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false))

(* Makes a clause list for appending (flattening) for conjuncts. *)

and convert_to_cnf_conjuncts (f : form_t) : form_t list = (
  match f with
    TT | FF | PROP _ | PRED _ | NOT _ ->
      [f]
  | (AND l) ->
      l
  | (OR [TT]) ->
      []
  | (OR _) ->
      [f]
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false))

(* Distributes {f1} over {f2} both in CNF.  It first reduces CNF to
   clauses, and {f1} and {f2} may be clauses, in recursive calls.  It
   works as ((A&B)|(C&D))->(((A&B)|C)&((A&B)|D)). *)

and convert_to_cnf_distribute_or (f1 : form_t) (f2 : form_t) : form_t = (
  match f1 with
    TT | FF | PROP _ | PRED _ | NOT _ -> (assert false)
  | (AND l) ->
      (pseudo_reduce_cnf
	  (AND (List.flatten
		   (List.map convert_to_cnf_conjuncts
		       (List.map (convert_to_cnf_distribute_or f2) l)))))
  | (OR _) ->
      (convert_to_cnf_clause_or f1 f2)
  | IMP _ | IFF _ | FORALL _ | EXISTS _ -> (assert false))

(* Disjoins two clauses. *)

and convert_to_cnf_clause_or (f1 : form_t) (f2 : form_t) : form_t = (
  assert (form_clausep f1) ;
  assert (form_clausep f2) ;
  match (f1, f2) with
    ((OR [TT]), (OR _)) ->
      (OR [TT])
  | ((OR _), (OR [TT])) ->
      (OR [TT])
  | ((OR []), (OR _)) ->
      f2
  | ((OR [FF]), (OR _)) ->
      f2
  | ((OR _), (OR [])) ->
      f1
  | ((OR _), (OR [FF])) ->
      f1
  | ((OR l1), (OR l2)) ->
      (OR (l1 @ l2))
  | ((OR l1), _) -> (assert false)
      (*AHO*) (*(OR (l1 @ [f2]))*)
  | (_, (OR l2)) -> (assert false)
      (*AHO*) (*(OR ([f1] @ l2))*)
  | _ -> (assert false))

(* Pseudoly makes CNF smaller. *)

and pseudo_reduce_cnf (f : form_t) : form_t = (
  match f with
    (AND [(OR [TT])]) ->
      f
  | (AND conj) ->
      let cs : FormSetSet.t =
	(Misc.list_fold
	    (fun clause ss ->
	      let fs = (pseudo_clause_to_form_set clause) in
	      (FormSetSet.add fs ss))
	    conj FormSetSet.empty) in
      let cl : form_t list =
	(FormSetSet.fold
	    (fun fs tl ->
	      ((pseudo_form_set_to_clause fs) :: tl))
	    cs []) in
      (AND cl)
  | _ -> (assert false))

and pseudo_clause_to_form_set (clause : form_t) : FormSet.t = (
  match clause with
    (OR disj) ->
      (Misc.list_fold
	  (fun e ss -> (FormSet.add e ss))
	  disj FormSet.empty)
  | _ -> (assert false))

and pseudo_form_set_to_clause (fs : FormSet.t) : form_t = (
  (OR (FormSet.elements fs)))

(* Appends two CNF. *)

let concat_cnf (f1 : form_t) (f2 : form_t) : form_t = (
  match (f1, f2) with
    ((AND [OR [TT]]), (AND _)) ->
      f2
  | ((AND _), (AND [OR [TT]])) ->
      f1
  | ((AND l1), (AND l2)) ->
      (AND (l1 @ l2))
  | _ -> (assert false))

(* ================================================================ *)

(* Combination of Theories *)

(* Nelson-Oppen Procedure.  This part follows the description in the
   paper "Combining Decision Procedures" by Manna and Zarba, in LNCS
   2757 (2003). *)

(* Toplevel entry is {smt_loop}. *)

(* {UFORM}, {AFORM}, {VFORM} are marker tags for uninterpreted
   function, arithmetic, and variable forms.  Variable forms are
   created when combining two forms of different types. *)

type form_type_t = UFORM | AFORM | VFORM

(* Mapping between propositions and predicates.  Each predicate is
   assigned a new proposition for making a SAT problem.  NOTE: Mapping
   has both directions of mapping: prop->pred and pred->prop.  {umap}
   is for UIF, {amap} is arithmetical, {emap} is equational. *)

type prop_pred_maps_t = {
  propabst_umap : (form_t, form_t) Hashtbl.t ;
  propabst_amap : (form_t, form_t) Hashtbl.t ;
  propabst_emap : (form_t, form_t) Hashtbl.t
}

(* {sat_info_cnf} is just a record of a problem CNF. *)

type smt_t = {
  sat_solver : Sat.solver_t ;
  prop_svar_map : (form_t, Sat.var_t) Hashtbl.t ;
  svar_prop_map : (Sat.var_t, form_t) Hashtbl.t ;
  prop_pred_maps : prop_pred_maps_t ;
  (*positive_propositions : FormSet.t ;*)
  mutable sat_last_result : form_t list ;
  mutable sat_info_cnf : form_t
}

(* Variable abstraction data.  It makes only variables are shared
   among theories.  {varabst_added_vars} records added variables.
   {varabst_shared_uvars} records variables appearing-in UI functions.
   {varabst_shared_avars} records variables appearing-in arithmetic
   terms.  {varabst_added_ueqls} records equality of variable
   abstractions for UI functions and proposition abstractions.
   {varabst_added_aeqls} records equality of variable abstractions for
   arithemetic.  {varabst_shared_props}.  Appearing-in means used-in
   an expression or compared-with in a predicate.  NOTE: Use of
   hashtables (without indicated hash function) are safe because terms
   do not include linear-terms here. *)

type varabst_t = {
  varabst_added_vars : (term_t, varname_t) Hashtbl.t ;
  varabst_shared_uvars : VarBag.t ;
  varabst_shared_avars : VarBag.t ;
  varabst_added_ueqls : FormBag.t ;
  varabst_added_aeqls : FormBag.t ;
  varabst_shared_props : FormBag.t
}

let make_propabst () : prop_pred_maps_t = (
  {
    propabst_umap = (Hashtbl.create 40) ;
    propabst_amap = (Hashtbl.create 40) ;
    propabst_emap = (Hashtbl.create 40)
  })

let make_varabst () : varabst_t = (
  {
    varabst_added_vars = (Hashtbl.create 10) ;
    varabst_shared_uvars = (VarBag.create ()) ;
    varabst_shared_avars = (VarBag.create ()) ;
    varabst_added_ueqls = (FormBag.create ()) ;
    varabst_added_aeqls = (FormBag.create ()) ;
    varabst_shared_props = (FormBag.create ())
  })

let rec dump_prop_pred_maps (d : prop_pred_maps_t) : unit = (
  let dump_map (h : (form_t, form_t) Hashtbl.t) : unit = (
    (Hashtbl.iter
	(fun k v ->
	  match k with
	    (PROP _) ->
	      (print_endline ("k=" ^ (form_to_string k)
			       ^ "/v=" ^ (form_to_string v)))
	  | _ ->
	      ())
	h)) in
  (print_endline ("propabst_umap:")) ;
  (dump_map d.propabst_umap) ;
  (print_endline ("propabst_amap:")) ;
  (dump_map d.propabst_amap) ;
  (print_endline ("propabst_emap:")) ;
  (dump_map d.propabst_emap))

(* Makes a variable and proposition abstraction for a form {f}.  It
   returns a CNF and a mapping between propositions and predicates in
   {d}. *)

let rec make_abstraction_form (d : prop_pred_maps_t) (f : form_t) : form_t = (
  let (vs, ps, uf, af) = (variable_abstraction_form f) in
  let ef = (make_equality_combinations vs) in
  let ux = (convert_to_cnf (prop_abstraction_form d.propabst_umap uf)) in
  let ax = (convert_to_cnf (prop_abstraction_form d.propabst_amap af)) in
  let ex = (prop_abstraction_form d.propabst_emap ef) in
  (*AHOAHO*) if (false) then
    (begin
      (print_endline ("var-abst vars=" ^ (String.concat ";" vs))) ;
      (print_endline ("var-abst props=" ^ (form_list_to_string ps))) ;
      (print_endline ("var-abst ef=" ^ (form_to_string ef))) ;
      (print_endline ("var-abst uf=" ^ (form_to_string uf))) ;
      (print_endline ("var-abst af=" ^ (form_to_string af))) ;
      (print_endline ("var-abst ex=" ^ (form_to_string ex))) ;
      (print_endline ("var-abst ux=" ^ (form_to_string ux))) ;
      (print_endline ("var-abst ax=" ^ (form_to_string ax))) ;
      (dump_prop_pred_maps d) ;
    end) ;
  (*(pseudo_reduce_cnf)*)
  (concat_cnf ex (concat_cnf ux ax)))

and variable_abstraction_form (f : form_t) : (varname_t list * form_t list * form_t * form_t) = (
  let d : varabst_t = (make_varabst ()) in
  let (xt, f1) = (var_abst_form d f) in
  let vs = (VarSet.elements
	       (VarSet.inter
		   (VarBag.as_set d.varabst_shared_avars)
		   (VarBag.as_set d.varabst_shared_uvars))) in
  let ps = (FormBag.elements d.varabst_shared_props) in
  match xt with
    UFORM | VFORM ->
      (vs, ps,
      (AND ((FormBag.contents d.varabst_added_ueqls) @ [f1])),
      (AND (FormBag.contents d.varabst_added_aeqls)))
  | AFORM ->
      (vs, ps,
      (AND (FormBag.contents d.varabst_added_ueqls)),
      (AND ((FormBag.contents d.varabst_added_aeqls) @ [f1]))))

(* Makes a variable abstraction form, and returns a pair of a form and
   its type. *)

and var_abst_form (d : varabst_t) (f : form_t) : (form_type_t * form_t) = (
  match f with
    TT | FF | (PROP _) ->
      (VFORM, f)
  | FORALL _ | EXISTS _ -> (assert false)
  | (PRED (p, t1, t2)) ->
      (begin
	let (xt, (x1, x2)) = (var_abst_two_terms d t1 t2) in
	match xt with
	  UFORM ->
	    (begin
	      match p with
		EQ ->
		  (UFORM, (PRED (EQ, x1, x2)))
	      | NE ->
		  (UFORM, (NOT (PRED (EQ, x1, x2))))
	      | GE | GT | LE | LT | DIVS ->
		  (failwith "arith rel-op with uninterpreted functions")
	    end)
	| AFORM ->
	    (AFORM, (PRED (p, x1, x2)))
	| VFORM ->
	    (begin
	      match p with
		EQ ->
		  (VFORM, (PRED (EQ, x1, x2)))
	      | NE ->
		  (VFORM, (NOT (PRED (EQ, x1, x2))))
	      | GE | GT | LE | LT | DIVS ->
		  (AFORM, (PRED (p, x1, x2)))
	    end)
      end)
  | (NOT f1) ->
      let (t1, x1) = (var_abst_form d f1) in
      (t1, (NOT x1))
  | (AND l) ->
      let (xt, xl) = (var_abst_combine_forms d l) in
      (xt, (AND xl))
  | (OR l) ->
      let (xt, xl) = (var_abst_combine_forms d l) in
      (xt, (OR xl))
  | (IMP (f1, f2)) ->
      (begin
	let (xt, xl) = (var_abst_combine_forms d [f1 ; f2]) in
	match xl with
	  [x1 ; x2] -> (xt, (IMP (x1, x2)))
	| _ -> (assert false)
      end)
  | (IFF (f1, f2)) ->
      (begin
	let (xt, xl) = (var_abst_combine_forms d [f1 ; f2]) in
	match xl with
	  [x1 ; x2] -> (xt, (IFF (x1, x2)))
	| _ -> (assert false)
      end))

and var_abst_two_terms (d : varabst_t) (t1 : term_t) (t2 : term_t) : (form_type_t * (term_t * term_t)) = (
  match (t1, t2) with
    ((LTERM _), _) | (_, (LTERM _)) -> (assert false)
  | ((FUN _), _) | (_, (FUN _)) -> (assert false)
  | ((VAR v1), (VAR v2)) ->
      (VFORM, (t1, t2))
  | ((VAR v), (UIFUN _)) ->
      (VarBag.add d.varabst_shared_uvars v) ;
      (UFORM, (t1, (var_abst_term UFORM d t2)))
  | ((UIFUN _), (VAR v)) ->
      (VarBag.add d.varabst_shared_uvars v) ;
      (UFORM, ((var_abst_term UFORM d t1), t2))
  | ((VAR v), (ARITH _)) ->
      (VarBag.add d.varabst_shared_avars v) ;
      (AFORM, (t1, (var_abst_term AFORM d t2)))
  | ((ARITH _), (VAR v)) ->
      (VarBag.add d.varabst_shared_avars v) ;
      (AFORM, ((var_abst_term AFORM d t1), t2))
  | ((UIFUN _), (UIFUN _)) ->
      (UFORM, ((var_abst_term UFORM d t1), (var_abst_term UFORM d t2)))
  | ((ARITH _), (ARITH _)) ->
      (AFORM, ((var_abst_term AFORM d t1), (var_abst_term AFORM d t1)))
  | ((UIFUN _), (ARITH _)) ->
      (AFORM, ((var_abst_term VFORM d t1), (var_abst_term AFORM d t2)))
  | ((ARITH _), (UIFUN _)) ->
      (AFORM, ((var_abst_term AFORM d t1), (var_abst_term VFORM d t2))))

and var_abst_combine_forms (d : varabst_t) (l : form_t list) : (form_type_t * form_t list) = (
  let x = (List.map (var_abst_form d) l) in
  let ux = (List.map snd (List.filter (fun (t, f) -> (t == UFORM)) x)) in
  let ax = (List.map snd (List.filter (fun (t, f) -> (t == AFORM)) x)) in
  let vx = (List.map snd (List.filter (fun (t, f) -> (t == VFORM)) x)) in
  if (ux != [] && ax != []) then
    (begin
      let ux2 =
	(List.map
	    (fun f ->
	      let p = (PROP (xgensym "Q")) in
	      (FormBag.add d.varabst_shared_props p) ;
	      (FormBag.add d.varabst_added_ueqls (IFF (p, f))) ;
	      p)
	    ux) in
      (AFORM, (ux2 @ ax @ vx))
    end)
  else if (ax != []) then
    (begin
      assert (ux == []) ;
      (AFORM, (ax @ vx))
    end)
  else if (ux != []) then
    (begin
      assert (ax == []) ;
      (UFORM, (ux @ vx))
    end)
  else
    (begin
      assert (ux == [] && ax == []) ;
      (VFORM, vx)
    end))

(*GOMI*)(*
(* NOTE: It prefers an arithmetic form, when forms are different
   class. *)
and var_abst_merge (d : varabst_t) (f : form_type_t * form_t) (l : form_type_t * form_t list) : (form_type_t * form_t list) = (
  match (f, l) with
    ((UFORM, f1), (UFORM, l1)) ->
      (UFORM, (f1 :: l1))
  | ((AFORM, f1), (AFORM, l1)) ->
      (AFORM, (f1 :: l1))
  | ((VFORM, f1), (VFORM, l1)) ->
      (VFORM, (f1 :: l1))
  | ((UFORM, f1), (VFORM, l1)) ->
      (UFORM, (f1 :: l1))
  | ((VFORM, f1), (UFORM, l1)) ->
      (UFORM, (f1 :: l1))
  | ((AFORM, f1), (VFORM, l1)) ->
      (AFORM, (f1 :: l1))
  | ((VFORM, f1), (AFORM, l1)) ->
      (AFORM, (f1 :: l1))
  | ((UFORM, f1), (AFORM, l1)) ->
      let p = (PROP (xgensym "Q")) in
      (FormBag.add d.varabst_shared_props p) ;
      (FormBag.add d.varabst_added_ueqls (IFF (p, f1))) ;
      (AFORM, (p :: l1))
  | ((AFORM, f1), (UFORM, l1)) ->
      let lx = (List.map
		   (fun f2 ->
		     let p = (PROP (xgensym "Q")) in
		     (FormBag.add d.varabst_shared_props p) ;
		     (FormBag.add d.varabst_added_ueqls (IFF (p, f2))) ;
		     p)
		   l1) in
      (AFORM, (f1 :: lx)))
*)

(* Makes a variable abstraction of a term for a wanted type with
   recording added variables in {d}. *)

and var_abst_term (wantterm : form_type_t) (d : varabst_t) (t : term_t) : term_t = (
  match t with
    (LTERM e) -> (assert false)
  | (ARITH (CVT _)) | (ARITH ZERO) -> (assert false)
  | (VAR v) ->
      (begin
	match wantterm with
	  UFORM ->
	    (VarBag.add d.varabst_shared_uvars v) ;
	    t
	| AFORM ->
	    (VarBag.add d.varabst_shared_avars v) ;
	    t
	| VFORM ->
	    t
      end)
  | (FUN _) ->
      if (wantterm == UFORM) then
	t
      else
	(assert false)
  | (UIFUN l) ->
      if (wantterm == UFORM) then
	(UIFUN (List.map (var_abst_term UFORM d) l))
      else
	(make_var_abst_term d t)
  | (ARITH (NUM i)) ->
      if (wantterm == AFORM) then
	t
      else
	(make_var_abst_term d t)
  | (ARITH (MUL (t1, t2))) ->
      if (wantterm == AFORM) then
	(ARITH (MUL ((var_abst_term AFORM d t1), (var_abst_term AFORM d t2))))
      else
	(make_var_abst_term d t)
  | (ARITH (ADD (t1, t2))) ->
      if (wantterm == AFORM) then
	(ARITH (ADD ((var_abst_term AFORM d t1), (var_abst_term AFORM d t2))))
      else
	(make_var_abst_term d t)
  | (ARITH (NEG t1)) ->
      if (wantterm == AFORM) then
	(ARITH (NEG (var_abst_term AFORM d t1)))
      else
	(make_var_abst_term d t))

(* Abstracts a term with (possibly) a new variable.  *)

and make_var_abst_term (d : varabst_t) (t : term_t) : term_t = (
  let wantterm : form_type_t =
    match t with
      (VAR _) | (FUN _) -> (assert false)
    | (LTERM _) | (ARITH (CVT _)) | (ARITH ZERO) -> (assert false)
    | (UIFUN _) ->
	UFORM
    | (ARITH _) ->
	AFORM in
  let a = (var_abst_term wantterm d t) in
  if (Hashtbl.mem d.varabst_added_vars a) then
    (begin
      let v = (Hashtbl.find d.varabst_added_vars a) in
      (VAR v)
    end)
  else
    (begin
      let v = (xgensym "U") in
      (Hashtbl.add d.varabst_added_vars a v) ;
      (VarBag.add d.varabst_shared_uvars v) ;
      (VarBag.add d.varabst_shared_avars v) ;
      match wantterm with
	VFORM -> (assert false)
      | UFORM ->
	  (FormBag.add d.varabst_added_ueqls (PRED (EQ, (VAR v), a))) ;
	  (VAR v)
      | AFORM ->
	  (FormBag.add d.varabst_added_aeqls (PRED (EQ, (VAR v), a))) ;
	  (VAR v)
    end))

and prop_abstraction_form (proppred : (form_t, form_t) Hashtbl.t) (f : form_t) : form_t = (
  match f with
    TT | FF | PROP _ ->
      f
  | FORALL _ | EXISTS _ -> (assert false)
  | (PRED (p, t1, t2)) ->
      if (Hashtbl.mem proppred f) then
	(Hashtbl.find proppred f)
      else
	(begin
	  let p = (PROP (xgensym "P")) in
	  (Hashtbl.add proppred f p) ;
	  (Hashtbl.add proppred p f) ;
	  p
	end)
  | (NOT f1) ->
      (NOT (prop_abstraction_form proppred f1))
  | (AND l) ->
      (AND (List.map (prop_abstraction_form proppred) l))
  | (OR l) ->
      (OR (List.map (prop_abstraction_form proppred) l))
  | (IMP (f1, f2)) ->
      (IMP ((prop_abstraction_form proppred f1),
	    (prop_abstraction_form proppred f2)))
  | (IFF (f1, f2)) ->
      (IFF ((prop_abstraction_form proppred f1),
	   (prop_abstraction_form proppred f2))))

(* Makes all combinations of equals/disequals, and returns them in
   CNF.  It stupidly makes all combinations without respect to
   transitivity. *)

and make_equality_combinations (vs : varname_t list) : form_t = (
  let rec make_all_combinations (vs : varname_t list) : form_t list = (
    match vs with
      [] ->
	[]
    | (v1 :: tl) ->
	((List.map (fun v2 -> (PRED (EQ, (VAR v1), (VAR v2)))) tl)
	  @ (make_all_combinations tl))) in
  let rec make_pos_neg_combinations (l : form_t list) : form_t list = (
    match l with
      [] ->
	[]
    | [hd] ->
	[(OR [hd ; (NOT hd)])]
    | (hd :: tl) ->
	((OR [hd ; (NOT hd)]) :: (make_pos_neg_combinations tl))) in
  let l1 = (make_all_combinations vs) in
  (AND (make_pos_neg_combinations l1)))

(* Makes a deabstraction form.  It removes predicates which do not
   appear in the mapping. *)

let rec prop_deabstraction_form (removep : bool) (h : (form_t, form_t) Hashtbl.t) (f : form_t) : form_t = (
  match f with
    TT | FF -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | (NOT f1) ->
      (prop_deabstraction_literal removep h false f1)
  | (PROP _) ->
      (prop_deabstraction_literal removep h true f)
  | (AND l) ->
      let x1 = (List.map (prop_deabstraction_form removep h) l) in
      let x2 = (List.filter (fun x -> (x != TT)) x1) in
      (AND x2)
  | (OR l) ->
      let x1 = (List.map (prop_deabstraction_literal removep h true) l) in
      let x2 = (List.filter (fun x -> (x != TT)) x1) in
      match x2 with
	[] -> TT
      | [ f1 ] -> f1
      | _ -> (OR x2))

and prop_deabstraction_literal (removep : bool) (h : (form_t, form_t) Hashtbl.t) (positive : bool) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      (*AHOAHO*)
      f
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | AND _ | OR _ -> (assert false)
  | (PROP _) ->
      if (Hashtbl.mem h f) then
	if (positive) then
	  (Hashtbl.find h f)
	else
	  (NOT (Hashtbl.find h f))
      else if (removep) then
	TT
      else
	if (positive) then
	  f
	else
	  (NOT f)
  | (NOT f1) ->
      (prop_deabstraction_literal removep h (not positive) f1))

(* Recovers the predicates from propositions for a SAT solution.  It
   removes predicates which do not appear in the mapping. *)

let rec make_deabstraction_cube (removep : bool) (d : prop_pred_maps_t) (l : form_t list) : form_t list = (
  let map_insert (h : (form_t, form_t) Hashtbl.t) (map : (form_t, form_t) Hashtbl.t) : unit = (
    (Hashtbl.iter
	(fun k v ->
	  assert (not (Hashtbl.mem h k)) ;
	  (Hashtbl.replace h k v))
	map)) in
  let proppred : (form_t, form_t) Hashtbl.t = (Hashtbl.create 10) in
  (map_insert proppred d.propabst_emap) ;
  (map_insert proppred d.propabst_umap) ;
  (map_insert proppred d.propabst_amap) ;
  (List.filter
      (fun x -> (x != TT))
      (List.map (prop_deabstraction_form removep proppred) l)))

(* Makes a deabstraction form for printing traces. *)

let rec make_deabstraction_form (d : prop_pred_maps_t) (f : form_t) : form_t = (
  match f with
    TT | FF ->
      (*AHOAHO*)
      f
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | (AND l) ->
      (AND (make_deabstraction_cube false d l))
  | (OR l) ->
      (OR (make_deabstraction_cube false d l))
  | (PROP _) ->
      (List.hd (make_deabstraction_cube false d [f]))
  | (NOT _) ->
      (List.hd (make_deabstraction_cube false d [f])))

(* **** SAT Solver Interface **** *)

(* Makes a SAT solver instance. *)

let make_smt () : smt_t = (
  (*
  let positiveset : FormSet.t =
    (List.fold_left
	(fun ss f -> (FormSet.add f ss))
	(FormSet.empty) positives) in
  *)
  {
    sat_solver = (Sat.solver_create ()) ;
    prop_svar_map = (Hashtbl.create 40) ;
    svar_prop_map = (Hashtbl.create 40) ;
    prop_pred_maps = (make_propabst ()) ;
    (*positive_propositions = positiveset ;*)
    sat_last_result = [] ;
    sat_info_cnf = FF
  })

let rec sat_literal (s : smt_t) (f : form_t) : Sat.lit_t = (
  match f with
    TT | FF -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | AND _ | OR _ -> (assert false)
  | (PROP _) ->
      if (Hashtbl.mem s.prop_svar_map f) then
	(Sat.lit_make (Hashtbl.find s.prop_svar_map f) true)
      else
	(begin
	  let v = (Sat.solver_new_variable s.sat_solver) in
	  (Hashtbl.add s.prop_svar_map f v) ;
	  (Hashtbl.add s.svar_prop_map v f) ;
	  (Sat.lit_make v true)
	end)
  | (NOT f1) ->
      let p = (sat_literal s f1) in
      (Sat.lit_neg p))

(* Claims a problem in CNF to a SAT solver.  It returns UNKNOWN
   normally, or UNSAT on a trivial unsatisfiable case.  It returns SAT
   on a trivial satisfiable empty conjunction case. *)

let rec sat_claim_cnf (s : smt_t) (f : form_t) : Sat.sat_result_t = (
  (* Record CNF for just information. *)
  s.sat_info_cnf <- f ;
  match f with
    TT | FF -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | PROP _  | NOT _ -> (assert false)
  | (OR l) -> (assert false)
  | (AND []) ->
      Sat.SAT
  | (AND l) ->
      (sat_claim_clause_loop s l))

and sat_claim_clause_loop (s : smt_t) (l : form_t list) : Sat.sat_result_t = (
  match l with
    [] ->
      Sat.UNKNOWN
  | (f :: tl) ->
      let trivialsat = (sat_claim_clause s f) in
      match trivialsat with
	Sat.UNSAT ->
	  xTRACE ("TRIVIAL UNSAT") ;
	  Sat.UNSAT
      | Sat.SAT ->
	  (*xTRACE ("TRIVIAL SAT") ;*)
	  (sat_claim_clause_loop s tl)
      | Sat.UNKNOWN ->
	  (sat_claim_clause_loop s tl))

(* Claims a clause to a SAT solver. *)

and sat_claim_clause (s : smt_t) (f : form_t) : Sat.sat_result_t = (
  match f with
    TT | FF -> (assert false)
  | FORALL _ | EXISTS _ -> (assert false)
  | PRED _ -> (assert false)
  | IMP _ | IFF _ -> (assert false)
  | PROP _  | NOT _ -> (assert false)
  | (AND l) -> (assert false)
  | (OR []) ->
      (*AHOAHO*)
      Sat.UNSAT
  | (OR [FF]) ->
      (*AHOAHO*)
      Sat.UNSAT
  | (OR l) ->
      let ps = (Sat.vec_from_list
		   (List.map (sat_literal s) l)
		   Sat.null_literal) in
      (Sat.claim_clause s.sat_solver ps))

let make_reject (s : smt_t) (reject : form_t list) : Sat.lit_t Sat.vec = (
  if (reject == []) then
    Sat.null_literal_vector
  else
    (begin
      let ps = (Sat.vec_create 0 0) in
      (List.iter
	  (fun p ->
	    match p with
	      (PROP _) ->
		let v = (Hashtbl.find s.prop_svar_map p) in
		(Sat.vec_push ps (Sat.lit_make v false))
	    | (NOT ((PROP _) as q)) ->
		let v = (Hashtbl.find s.prop_svar_map q) in
		(Sat.vec_push ps (Sat.lit_make v true))
	    | _ -> (assert false))
	  reject) ;
      ps
    end))

(* Runs a SAT solver, and returns a list of propositions.  It takes a
   reject cube {lastform}, whose negation (a clause) is added before
   running a solver.  MEMO: It does not utilize restarting assignment
   (solver argument {assume}) yet. *)

(* MEMO: Record the last sat result, to avoid converting back a reject
   later. *)

let sat_solve (s : smt_t) (lastform : form_t list) : form_t list = (
  if (lastform != []) then
    xTRACE ("reject-cube=" ^ (form_to_string (AND lastform))) ;
  let reject = (make_reject s lastform) in
  (* Check the last generated result is trivially unsatisfiable. *)
  let trivialsat =
    if (Sat.null_literal_vector_p reject) then
      Sat.SAT
    else
      (Sat.claim_clause s.sat_solver reject) in
  if (trivialsat == Sat.UNSAT) then
    (begin
      xTRACE ("UNSAT (trivial)") ;
      []
    end)
  else
    (begin
      let (ok, m) = (Sat.solver_solve s.sat_solver Sat.null_literal_vector) in
      assert (ok != Sat.UNKNOWN) ;
      if (ok == Sat.UNSAT) then
	[]
      else
	(Sat.vec_mapi
	    (fun (i : int) (v : bool) ->
	      if (v) then
		(Hashtbl.find s.svar_prop_map i)
	      else
		(NOT (Hashtbl.find s.svar_prop_map i)))
	    (Sat.some_of m))
    end))

(* Recovers the equality predicates from propositions of a SAT
   solution. *)

let eqlsat_preds_from_sat_solution (d : prop_pred_maps_t) (l : form_t list) : form_t list = (
  let l1 = (List.filter (fun x -> (x != TT))
	       (List.map (prop_deabstraction_form true d.propabst_emap) l)) in
  let l2 = (List.filter (fun x -> (x != TT))
	       (List.map (prop_deabstraction_form true d.propabst_umap) l)) in
  (l1 @ l2))

(* Recovers the linear integer arithmetics predicates from
   propositions of a SAT solution. *)

let liasat_preds_from_sat_solution (d : prop_pred_maps_t) (l : form_t list) : form_t list = (
  let l1 = (List.filter (fun x -> (x != TT))
	       (List.map (prop_deabstraction_form true d.propabst_emap) l)) in
  let l2 = (List.filter (fun x -> (x != TT))
	       (List.map (prop_deabstraction_form true d.propabst_amap) l)) in
  (l1 @ l2))

(* Checks if literals of a SAT result are positively constrained. *)

(*
  {positives} gives a constraint of positive propositions.
  else if (not (check_positive_constraint s.positive_propositions sat)) then
    (smt_loop s sat)
*)

(*GOMI*)(*
let rec check_positive_constraint (positives : FormSet.t) (sat : form_t list) : bool = (
  match sat with
    [] ->
      true
  | (f :: tl) ->
      (begin
	match f with
	  TT | FF -> (assert false)
	| FORALL _ | EXISTS _ -> (assert false)
	| PRED _ -> (assert false)
	| IMP _ | IFF _ -> (assert false)
	| (PROP _) ->
	    (check_positive_constraint positives tl)
	| (NOT ((PROP _) as p)) ->
	    if (FormSet.mem p positives) then
	      false
	    else
	      (check_positive_constraint positives tl)
	| (NOT _) -> (assert false)
	| (AND l) -> (assert false)
	| (OR l) -> (assert false)
      end))
*)

(* Runs SMT (toplevel).  It takes a reject cube, whose negation (a
   cluase) is added to a SAT problem set. *)

let rec smt_loop (s : smt_t) (reject : form_t list) : (Sat.sat_result_t * (form_t list)) = (
  let sat = (sat_solve s reject) in
  s.sat_last_result <- sat ;
  if (sat == []) then
    (Sat.UNSAT, [])
  else
    (begin
      let l1 = (eqlsat_preds_from_sat_solution (s.prop_pred_maps) sat) in
      let ok1 = (eqlsat_solve l1) in
      if (!smtTRACE_sat) then
	(print_endline (";; SMT equality-sat: ("
			 ^ (String.concat "&&"
			       (List.map form_to_string l1))
			 ^ ")=" ^ (Sat.sat_result_to_string ok1))) ;
      match ok1 with
	Sat.UNKNOWN -> (assert false)
      | Sat.SAT ->
	  (begin
	    let l2 = (liasat_preds_from_sat_solution (s.prop_pred_maps) sat) in
	    let ok2 = (liasat_solve l2) in
	    if (!smtTRACE_sat) then
	      (print_endline (";; SMT linear-sat: ("
			       ^ (String.concat "&&"
				     (List.map form_to_string l2))
			       ^ ")=" ^ (Sat.sat_result_to_string ok2))) ;
	    match ok2 with
	      Sat.UNKNOWN -> (assert false)
	    | Sat.SAT ->
		let solution = (make_deabstraction_cube false
				   (s.prop_pred_maps) sat) in
		(Sat.SAT, solution)
	    | Sat.UNSAT ->
		xTRACE ("linear-sat unsat") ;
		(smt_loop s sat)
	  end)
      | Sat.UNSAT ->
	  xTRACE ("equality-sat unsat") ;
	  (smt_loop s sat)
    end))

(* Runs SMT.  It returns all satisfying cubes.  It returns a list of
   cubes ([] if nothing), or [TT] if trivial-sat.  {positives} gives a
   constraint of positive propositions. *)

let rec smt_solve_satisfying_cubes (f : form_t) : form_t list = (
  let s = (make_smt ()) in
  let cnf = (make_abstraction_form s.prop_pred_maps f) in
  let trivialsat = (sat_claim_cnf s cnf) in
  (xPRINT (!smtTRACE_cnf) "smt_solve cnf: "
      ("cnf(" ^ (string_of_int (cnf_number_of_clauses cnf))
	^ "," ^ (string_of_int (cnf_average_clause_length cnf))
	^ "," ^ (string_of_int (Sat.n_variables s.sat_solver))
	^ ")="
	^ (form_to_string (make_deabstraction_form (s.prop_pred_maps) cnf)))) ;
  match trivialsat with
    Sat.SAT ->
      (*AHO*) xDEBUG ("TRIVIAL SAT") ;
      [TT]
  | Sat.UNSAT ->
      []
  | Sat.UNKNOWN ->
      (smt_solve_cubes_loop s []))

and smt_solve_cubes_loop (s : smt_t) (foundcubes : form_t list ) : form_t list = (
  let (r, c) = (smt_loop s s.sat_last_result) in
  match r with
    Sat.SAT ->
      if (!smtTRACE_result_cube) then
	(print_endline (";; SMT SAT-CUBE=" ^ (form_to_string (AND c)))) ;
      (smt_solve_cubes_loop s (foundcubes @ [(AND c)]))
  | Sat.UNSAT ->
      foundcubes
  | Sat.UNKNOWN ->
      (failwith "NOT-SAT (UNKNOWN)"))

(* Runs SMT.  It simply returns satisfiability. *)

let smt_solve (f : form_t) : (Sat.sat_result_t * (form_t list)) = (
  let s = (make_smt ()) in
  let cnf = (make_abstraction_form s.prop_pred_maps f) in
  let trivialsat = (sat_claim_cnf s cnf) in
  (xPRINT (!smtTRACE_cnf) "smt_solve cnf: "
      ("cnf(" ^ (string_of_int (cnf_number_of_clauses cnf))
	^ "," ^ (string_of_int (cnf_average_clause_length cnf))
	^ "," ^ (string_of_int (Sat.n_variables s.sat_solver))
	^ ")="
	^ (form_to_string (make_deabstraction_form (s.prop_pred_maps) cnf)))) ;
  match trivialsat with
    Sat.SAT ->
      (*AHO*) xDEBUG ("TRIVIAL SAT") ;
      (Sat.SAT, [])
  | Sat.UNSAT ->
      (Sat.UNSAT, [])
  | Sat.UNKNOWN ->
      (smt_loop s []))
