(* parser/specs/sdump.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Scommon
(* 
    dump Cexpr.stmt_t list to text
*)

let tab_string = "    "
let curr_tab = ref ""

let normalize_id id =
    ExtString.String.replace_chars (function
                                        '#' -> "_"
                                      | _ as c -> String.make 1 c)
                                   id

(* this set contain declaration of all struct, for dumping C file purpose *)
module StructsSet = Set.Make (
    struct
        type t = Cexpr.structure_tag_t
        let compare x y = (compare x y)
    end
)
(* global variable for dump C file purpose *)
let structs_set = ref StructsSet.empty 
let dump_c_file  = ref false

(*
   dump typenode_t
*)
let rec string_of_typenode (typ : Cexpr.typenode_t) = 
   match typ with
   | Cexpr.NOTYPE -> "/* NOTYPE */"
   (* Pseudo Types *)
   | Cexpr.SPECIFR_TYPE string_list -> "/* SPECIFR_TYPE */"
   | Cexpr.QUALIFIED_TYPE (typenode, type_qualifier) -> (
       let a_qualifier = (
         match type_qualifier with
         | Cexpr.CONSTANT -> "const"
         | Cexpr.VOLATILE -> "volatile"
         | Cexpr.RESTRICT -> "restrict"
         | Cexpr.INLINE -> "inline"  
         | Cexpr.NILQUAL -> ""  
	     | Cexpr.CONSTANT_VOLATILE -> "const volatile" 
         | Cexpr.CONSTANT_RESTRICT -> "constant restrict"
		 | Cexpr.VOLATILE_RESTRICT  -> "volatile restrict"
         | Cexpr.CONSTANT_VOLATILE_RESTRICT -> "const volatile restrict"
         | _ -> assert false
       ) in
       if (a_qualifier <> "") then
         a_qualifier ^ " " ^ (string_of_typenode typenode)
       else
         string_of_typenode typenode;
     ) 
   | Cexpr.ATTRIBUTED_TYPE (typenode, gcc_attribute_list) -> 
       string_of_typenode typenode
   | Cexpr.TYPEDEF_TYPE (typenode, string)     -> "/* TYPEDEF_TYPE */"
   | Cexpr.STRCDEF_TYPE (tyn,_)                -> string_of_typenode tyn
   | Cexpr.ENUMDEF_TYPE (typenode, string_expr_list)  -> (
       let enum_data = (
           List.map (
               fun enum_item ->
               begin
                   let (name, value) = enum_item in
                   let str = (string_of_expr value false) in
                   if (str = "" ) then
                       "    " ^ name
                   else
                       "    " ^ name ^ " = " ^ str
               end
           ) string_expr_list
       ) in
       let enum_data_str = String.concat ",\n" enum_data in
       "enum " ^ string_of_typenode typenode ^ " {\n" ^ enum_data_str ^ "\n}"
     )
   | Cexpr.ELLIPSIS_TYPE                       -> assert false
   | Cexpr.PSEUDO_VA_LIST_TYPE                 -> "/* PSEUDO_VA_LIST_TYPE */"
   (*Basic types *)
   | Cexpr.VOID_TYPE                    -> "void"
   | Cexpr.BOOLEAN                      -> "bool"             
   | Cexpr.CHARACTER                    -> "char"
   | Cexpr.SIGNED_CHARACTER             -> "char"
   | Cexpr.UNSIGNED_CHARACTER           -> "char"
   | Cexpr.SHORT_INTEGER                -> "short int"
   | Cexpr.UNSIGNED_SHORT_INTEGER       -> "unsigned short int"
   | Cexpr.INTEGER                      -> "int"
   | Cexpr.UNSIGNED_INTEGER             -> "unsigned int"
   | Cexpr.LONG_INTEGER                 -> "long int"
   | Cexpr.UNSIGNED_LONG_INTEGER        -> "unsigned long int"
   | Cexpr.LONG_LONG_INTEGER            -> "long long int"
   | Cexpr.UNSIGNED_LONG_LONG_INTEGER   -> "unsigned long long int"
   | Cexpr.SINGLE_FLOAT                 -> "float"
   | Cexpr.DOUBLE_FLOAT                 -> "double"
   | Cexpr.LONG_DOUBLE_FLOAT            -> "long double"
   | Cexpr.SINGLE_COMPLEX               -> "float _Complex"
   | Cexpr.DOUBLE_COMPLEX               -> "double _Complex"
   | Cexpr.LONG_DOUBLE_COMPLEX          -> "long double _Complex" 
   | Cexpr.SINGLE_IMAGINARY             -> "float _Imaginary"
   | Cexpr.DOUBLE_IMAGINARY             -> "double _Imaginary"
   | Cexpr.LONG_DOUBLE_IMAGINARY        -> "long double _Imaginary"
   (*Composit types *)
   | Cexpr.FUNCTION_OF (ty,_)           ->  string_of_typenode ty
   | Cexpr.STRUCTURE_TYPE (struct_type) -> (
       let name =
           if (String.length struct_type.Cexpr.strc_tag) == 0 then
               (normalize_id struct_type.Cexpr.strc_unique_id)
           else
               (normalize_id struct_type.Cexpr.strc_tag)
       in
       let prefix =
           match struct_type.Cexpr.strc_struct_or_union_p with
           | true -> "struct"
           | false -> "union"
       in
       prefix ^ " " ^ name)
   | Cexpr.POINTER_TO tyn -> (string_of_typenode tyn) ^ "*"
   | Cexpr.ARRAY_OF (typenode, expr_option) -> string_of_typenode typenode
   | Cexpr.BITFIELD_TYPE (typenode, expr)   -> "/* BITFIELD_TYPE */"
   | Cexpr.ENUMERATION_TYPE enumeration_tag -> enumeration_tag.Cexpr.enum_tag
   (* GCC Extension *)
   | Cexpr.GCC_VA_ARG -> "/* GCC_VA_ARG */"
   | _ -> assert false

and string_of_point_to_type (ty : Cexpr.typenode_t) : string = (
    match ty with
    | Cexpr.POINTER_TO typ -> (string_of_point_to_type typ)
    | Cexpr.FUNCTION_OF (tyn, _) -> (string_of_typenode tyn)
    | _ -> (string_of_typenode ty)
)

and string_of_struct (struct_type : Cexpr.structure_tag_t ) (dump_c_file : bool) = (
    let tag = struct_type.Cexpr.strc_tag in
    let unique_id = struct_type.Cexpr.strc_unique_id in
    let table = struct_type.Cexpr.strc_typebody_table in
    let prefix =
        match struct_type.Cexpr.strc_struct_or_union_p with
        | true -> "struct"
        | false -> "union"
    in
    let type_body = (
        try
           Hashtbl.find table unique_id
        with
        | Not_found -> assert false
    ) in
    let new_id =
        if (String.length tag) == 0 then
            (normalize_id unique_id)
        else
            (normalize_id tag)
    in
    match type_body with
    | Cexpr.STRC_BODY (str_type) -> (
        let components = str_type.Cexpr.strc_components in
        if (dump_c_file) then
            if (StructsSet.mem struct_type !structs_set) then
                prefix ^ " " ^ new_id
            else 
            begin
                structs_set := StructsSet.add struct_type !structs_set;
                prefix ^ " " ^ new_id
                ^ " {\n" ^ (string_of_components components) ^ "}"
            end    
        else
            prefix ^ " " ^ new_id
            ^ " {\n" ^ (string_of_components components) ^ "}"
     )
    | Cexpr.ENUM_BODY (str_type) -> ""
    | _ -> assert false
)

(* dump the parameter of an array to a string 
     input: name & typenode of an array *)
and string_of_array_parameter (typenode : Cexpr.typenode_t) : string = (
  match typenode with
  | Cexpr.ARRAY_OF(ty, ex_o) -> (
      let str_temp = (  
        match ex_o with
        | None -> "[]"
        | Some ex_temp -> "[" ^ (string_of_expr ex_temp false) ^ "]"
      ) in
      let array_parameter = (
        match ty with
        | Cexpr.ARRAY_OF (_ , _) -> str_temp ^ (string_of_array_parameter ty)   
        | _ -> str_temp
      ) in 
      array_parameter;)
  | _ -> assert false
)    

(* dump an array to a string *)        
and string_of_array (name: string) (typenode : Cexpr.typenode_t) : string = (
  match typenode with
  | Cexpr.ARRAY_OF(ty, ex_o) -> 
      (string_of_typenode typenode) ^ " " ^ name ^ 
      (string_of_array_parameter typenode) 
  | _ -> assert false
)

(*
   dump struct/union field
*)
and string_of_components components = (
  match components with
  | [] -> ""
  | head::tail -> (
      let head_str = ref "" in
      head_str := !head_str ^ tab_string;
      let  name , ty = head in 
      tab_string ^ (string_of_typenode ty) ^ " " ^ name ^ ";\n";

      (match ty with
      | Cexpr.ARRAY_OF _ -> (
          head_str := (!head_str) ^ (string_of_array name ty)
        )
      | Cexpr.STRCDEF_TYPE(Cexpr.STRUCTURE_TYPE (struct_type), _) -> 
          head_str := (!head_str)
                      ^ (string_of_struct struct_type !dump_c_file)
                      ^ " " ^ (normalize_id name);
      | Cexpr.STRUCTURE_TYPE (struct_type) -> 
          head_str := (!head_str)
                      ^ (string_of_struct struct_type !dump_c_file)
                      ^ " " ^ (normalize_id name);
      | Cexpr.POINTER_TO typ -> (
          let new_typ = Cexpr.strip_off_pseudo_type typ true in
          match new_typ with
          | Cexpr.FUNCTION_OF (func_typ, para_typ_opt) -> (
              let para_decls = (
                  match para_typ_opt with
                  | None -> ""
                  | Some typs -> (
                      let para_strs =
                          List.map (
                              fun a_typ -> string_of_typenode a_typ
                          ) typs in
                      String.concat ", " para_strs;
                    )
              ) in
              head_str := (!head_str) ^ string_of_typenode (func_typ)
                          ^ " (*" ^ name ^ ")" ^ "(" ^ para_decls ^ ")"
            );
          | _ -> head_str := (!head_str) ^ (string_of_typenode ty)
                             ^ " " ^ name
        )
      | _ -> head_str := (!head_str) ^ (string_of_typenode ty) ^ " " ^ name);

      !head_str ^ ";\n" ^ (string_of_components tail);
    );
)

(* dump an operator expression to a string 
   input: 
     - an operator expression.
     - a flag indicate operator is an array or not. (for dumping arrays purpose)  
*)
and string_of_expr_oper  (expr : Cexpr.expr_t) (is_array : bool)= 
   match expr with 
   | Cexpr.OPER(typ,op,ex1,ex2,ex3) -> (
           let has_pa1 = match ex1 with
                         | Cexpr.OPER _ -> true
                         | _ -> false
           in
           let has_pa2 = match ex2 with
                         | Cexpr.OPER _ -> true
                         | _ -> false
           in
           let has_pa3 = match ex3 with
                         | Cexpr.OPER _ -> true
                         | _ -> false
           in
	   match op with
           | Cexpr.INCR_PRE   -> "++"^(string_of_expr ex1 has_pa1)
           | Cexpr.DECR_PRE   -> "--"^(string_of_expr ex1 has_pa1)
           | Cexpr.INCR_POST  -> (string_of_expr ex1 has_pa1)^"++ "
           | Cexpr.DECR_POST  -> (string_of_expr ex1 has_pa1)^"-- "
           | Cexpr.ADD        -> (
               if (((Ccommon.is_array_oper ex1) = true)
                   && (is_array = true)) then
                 (string_of_expr ex1 false) ^ 
                      "[" ^ (string_of_expr ex2 has_pa2) ^ "]"
               else
                 (string_of_expr ex1 has_pa1) ^ " + " ^ 
                     (string_of_expr ex2 has_pa2)
             )
           | Cexpr.SUB        -> (string_of_expr ex1 has_pa1)^" - "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.MUL        -> (string_of_expr ex1 has_pa1)^" * "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.DIV        -> (string_of_expr ex1 has_pa1)^" / "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.MOD        -> (string_of_expr ex1 has_pa1)^" % "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.LSH        -> (string_of_expr ex1 has_pa1)^" + "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.RSH        -> (string_of_expr ex1 has_pa1)^" + "^
                                 (string_of_expr ex2 has_pa1)
           | Cexpr.AND        -> (string_of_expr ex1 has_pa1)^" & "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.IOR        -> (string_of_expr ex1 has_pa1)^" | "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.XOR        -> (string_of_expr ex1 has_pa1)^" ^ "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.CAST tf     -> (
                                  match tf with
                                  |  true -> 
                                      let str_ty = string_of_typenode typ in
                                      ("(" ^ str_ty ^ ")" ^
                                      (string_of_expr ex1 has_pa1))
                                  | false -> (string_of_expr ex1 has_pa1)
                                 )
           | Cexpr.NEG        -> "-" ^ (string_of_expr ex1 has_pa1)
           | Cexpr.BITNOT     -> "~"^(string_of_expr ex1 has_pa1)
           | Cexpr.LOGNOT     -> "!"^(string_of_expr ex1 has_pa1)
           | Cexpr.EQ         -> (string_of_expr ex1 has_pa1)^" == "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.NE         -> (string_of_expr ex1 has_pa1)^" != "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.GE         -> (string_of_expr ex1 has_pa1)^" >= "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.GT         -> (string_of_expr ex1 has_pa1)^" > "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.LE         -> (string_of_expr ex1 has_pa1)^" <= "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.LT         -> (string_of_expr ex1 has_pa1)^" < "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.COND       -> (string_of_expr ex1 has_pa1)^"? "^
                                 (string_of_expr ex2 has_pa2)^":"^
                                 (string_of_expr ex3 has_pa3) 
           | Cexpr.OROR       -> (string_of_expr ex1 has_pa1)^" || "^
                                 (string_of_expr ex2 has_pa2) 
           | Cexpr.ANDAND     -> (string_of_expr ex1 has_pa1)^" && "^
                                 (string_of_expr ex2 has_pa2)
           | Cexpr.COMMA      -> (string_of_expr ex1 has_pa1)^" , "^
                                 (string_of_expr ex2 has_pa2)
           | _                -> assert false
     )
   | _ -> assert false

(* 
  dump expr list 
*)
and string_of_exprs (exprs : Cexpr.expr_t list) (has_pa : bool) = 
    match exprs with
     | [] -> ""
     | [expr] -> (string_of_expr expr has_pa)
     | head :: tail -> ((string_of_expr head has_pa)^","^
                         (string_of_exprs tail has_pa))

(*
  dump expr
*)
and string_of_expr (expr : Cexpr.expr_t) (has_pa : bool) = (
    let res = (
    match expr with
    | Cexpr.NIL -> ""
    | Cexpr.PSEUDO_NONDET_VALUE (typenode, str) -> " /* PSEUDO_NONDET_VALUE */ "
    (* Arithmetic/Logical/Conditional *)
    | Cexpr.OPER (typenode, op, ex1, _, _) -> (
          match (typenode, op, ex1) with
          (* dump the init a string, ex: char ch[10] = "hello" *)
          | (Cexpr.CHARACTER, Cexpr.CAST _, Cexpr.CONST_I (_, a_num, _)) ->
              Num.string_of_num a_num
          | _ -> string_of_expr_oper expr false   
      )
(*    | Cexpr.OPER _ -> (string_of_expr_oper expr false)*)
    (* Assignment/Increment/Decrement *)
    | Cexpr.SETF (typ, opt, expr1, expr2) -> (
         match opt with
         | Cexpr.INCR_PRE  -> "++"^(string_of_expr expr1 has_pa)
         | Cexpr.DECR_PRE -> "--"^(string_of_expr expr1 has_pa)
         | Cexpr.INCR_POST -> (string_of_expr expr1 has_pa)^"++"
         | Cexpr.DECR_POST-> (string_of_expr expr1 has_pa)^"--"
         | _ ->  ((string_of_expr expr1 has_pa)^" = "^
                  (string_of_expr expr2 has_pa))
      )
    (* Primaries *)
    | Cexpr.CALL (typ, ex, exs, _) -> 
        (string_of_expr ex has_pa) ^ "(" ^ (string_of_exprs exs has_pa) ^ ")"
    | Cexpr.VREF (typ, str, num1, num2) -> str
    | Cexpr.PREF (typ, _, ex, _) -> (
        match ex with
        | Cexpr.OPER _ -> (
            if (Ccommon.is_array_oper ex) then
              string_of_expr_oper ex true
            else
              "*(" ^ (string_of_expr_oper ex has_pa) ^ ")" )
        | _ -> ("*"^(string_of_expr ex has_pa))
      )
    | Cexpr.SREF(typ,pdef,ex,str) -> (
        let typ_ex = Cexpr.type_of_expr ex in
        let main_expr = string_of_expr ex has_pa in
        let str = normalize_id str in
        if pdef then
            main_expr ^ "->" ^ str
        else
            main_expr ^ "." ^ str
      )
    | Cexpr.DSGN_AREF (typenode, expr1, expr2) -> " /* DSGN_AREF */ "
    | Cexpr.ADDROF (_, ex) -> (
        match ex with
        | Cexpr.OPER _ -> ("&("^(string_of_expr ex has_pa)^")")
        | _ -> ("&"^(string_of_expr ex has_pa))
      )
    | Cexpr.SIZEOF (typ1, typ2) -> "sizeof(" ^ (string_of_typenode typ2) ^ ")"
    | Cexpr.PROGN (typenode, stmt_list) -> " /* PROGN */"
    (* Value Node *)
    | Cexpr.CONST_I(typ, num, str) -> (Num.string_of_num num)
    | Cexpr.CONST_F (typ, numf, str) -> (string_of_float numf)
    | Cexpr.STRING (wide, str, _) ->
        if wide then
            "L\"" ^ (String.escaped str) ^ "\""
        else
            "\"" ^ (String.escaped str) ^ "\""
    | Cexpr.CLITERAL (typenode, initializers) -> " /* CLITERAL */ "
    | Cexpr.ASMSEQUENCE expr_list -> " /* ASMSEQUENCE */ "
    (* Internal Use *)
    | Cexpr.PHI (typenode, string_expr_list) -> " /* PHI */ "
    | Cexpr.PSEUDO_MEMORY (typenode1, typenode2) -> " /* PSEUDO_MEMORY */ "
    | Cexpr.INIT_FILL _ -> " /* INIT_FILL */ "
    (* Pragma Extension *)
    | Cexpr.PRAGMA (str) -> "pragma" ^ str
    | Cexpr.QUANTIFIER (str, expr_list, expr) -> " /* QUANTIFIER */ "
    (* Pseudo Node (not appear in parser) *)
    | Cexpr.PSEUDO_PROLOGUE -> " /* PSEUDO_PROLOGUE */ "
    | Cexpr.PSEUDO_EPILOGUE -> " /* PSEUDO_EPILOGUE */ "
    | Cexpr.PSEUDO_ACSL_RESULT _ -> " /* PSEUDO_ACSL_RESULT */ "
    | Cexpr.PSEUDO_ACSL_OLD _ -> " /* PSEUDO_ACSL_OLD */ "
    | Cexpr.PSEUDO_ACSL_UNIQUE _ -> " /* PSEUDO_ACSL_UNIQUE */ "
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK _ -> " /* PSEUDO_ACSL_UNIQUEBLOCK */ "
    | Cexpr.ALL_FIELD (_, ptr) -> " /* ALL_FIELD "
                                  ^ (string_of_expr ptr has_pa)  ^ " */ "
    | Cexpr.PSEUDO_ACSL_VALID _ -> " /* PSEUDO_ACSL_VALID */ "
    | _ -> assert false )
    in
    if has_pa then
        "("^res^")"
    else
        res
)

let string_of_init (init : Cexpr.expr_t) : string = (
    match init with
    | Cexpr.SETF(_, Cexpr.SET, ex1, ex2) -> (
          let init_value = (string_of_expr ex2 false) in
          match ex1 with
          (* init of struct variable*)
          | Cexpr.SREF (_, false, _, str) -> "." ^ str ^ " = " ^ init_value
          | Cexpr.SREF (_, true, _, str) -> "->" ^ str ^ " = " ^ init_value
          (* init of other variable type.
             FIXME: check here if the init dumping is wrong *)
          | _ -> init_value
      )
    | Cexpr.SETF(_, Cexpr.INCR_PRE, _, ex2) -> ("++"^(string_of_expr ex2 false))    
    | Cexpr.SETF(_, Cexpr.DECR_PRE, _, ex2) -> ("--"^(string_of_expr ex2 false))
    | Cexpr.SETF(_, Cexpr.INCR_POST, _, ex2) -> ((string_of_expr ex2 false)^"++")
    | Cexpr.SETF(_, Cexpr.DECR_POST, _, ex2) -> ((string_of_expr ex2 false)^"--")
    | _ -> ""
)   

let string_of_inits (inits : Cexpr.expr_t list) 
                    (typ: Cexpr.typenode_t) 
                    : string = (
    let strings = (List.map string_of_init inits) in
    let strings = (List.filter (fun s -> (String.length s) > 0) strings) in
    let string_of_array_or_struct_inits strings =
        match strings with
        | [] -> ""
        | _ -> " = { "^ (String.concat ", " strings) ^ " }"
    in
    match typ with
    (* get string of an character init *)
    | Cexpr.CHARACTER -> (
          match strings with
          | []    -> ""
          | [str] -> " = '" ^ str ^ "'"
          | _     -> assert false  
      ) 
    (* get string of a string or an array init *)
    | Cexpr.ARRAY_OF _ ->
          string_of_array_or_struct_inits strings
    | _ when (Cexpr.structure_or_union_type_p typ) ->
          string_of_array_or_struct_inits strings
    (* get string of other data inits *)
    | _ -> (
          match strings with
          | [] -> ""
          | [str] -> " = " ^ str
          | _ -> " = { "^ (String.concat ", " strings) ^ " }"
      )
)

(* dump defvar *)
let string_of_header (decl : Cexpr.declaration_t) = (
    let decl_text = ref "" in
    let (id,level) = decl.Cexpr.decl_id in
    let typ = decl.Cexpr.decl_tn in
    let scls = decl.Cexpr.decl_scls in
    (match scls with
    | Cexpr.TYPEDEF -> (
          if (id <> "") then
          begin
              decl_text := (!decl_text) ^ "typedef ";
              match typ with
              | Cexpr.STRCDEF_TYPE(Cexpr.STRUCTURE_TYPE (struct_type), _)
              | Cexpr.STRUCTURE_TYPE (struct_type) ->
                    decl_text := (!decl_text)
                                 ^ (string_of_struct struct_type !dump_c_file)
                                 ^ " " ^ id;
              | _ -> (
                    if (Ccommon.is_pointer typ) = true then
                        decl_text := (!decl_text) ^ (string_of_typenode typ)
                                     ^ " *" ^ id
                    else
                        decl_text := (!decl_text) ^ (string_of_typenode typ)
                                     ^ " " ^ id
                )
          end
          else assert false
      )
    | _ -> (
        let store_name = (
            match scls with
            | Cexpr.NILSCLS  
            | Cexpr.AUTO -> ""
            | Cexpr.EXTERN -> "extern "
            | Cexpr.REGISTER -> "register "
            | Cexpr.STATIC -> "static "
            | Cexpr.TYPEDEF -> assert false
          ) in
          if (store_name <> "") then  
            decl_text := (!decl_text) ^ store_name ^ " "
          else
            decl_text := (!decl_text) ^ store_name;
             
          match typ with
          | Cexpr.ARRAY_OF(ty, ex_o) -> 
              decl_text := (!decl_text) ^ (string_of_array id typ)
              
          (* Temporarily ignore for later deal with "struct" declaration *)
          | Cexpr.STRCDEF_TYPE(Cexpr.STRUCTURE_TYPE (struct_type), _) -> (
              if (id <> "") then
                (* `foo_t s;' in which foo_t is a typedefed type *)
                decl_text := (!decl_text) ^
                             (string_of_struct struct_type !dump_c_file) ^ 
                             " " ^ id
              else
                (* declaration of a struct *)
                decl_text := (!decl_text) ^ 
                             (string_of_struct struct_type !dump_c_file)  
            )
          | _ -> (
              if((Ccommon.is_pointer typ) = true) then (
                if  ((Ccommon.is_function_type typ) = true) then
                  decl_text := (!decl_text) ^ 
                               (string_of_point_to_type typ) ^ "(*" ^ id ^ ")"
                else 
                  decl_text := (!decl_text) ^ (string_of_typenode typ) ^ " " ^ id
              )
              else
                decl_text := (!decl_text) ^ (string_of_typenode typ) ^ " " ^ id)
            )
    );
    let inits = decl.Cexpr.decl_inits in
    let params = decl.Cexpr.decl_params in
    let _ = ( 
      match (Ccommon.get_function_name decl) with
      | None -> (
          (* dump DEFVAR *)
          decl_text := (!decl_text) ^ (string_of_inits inits typ);
        )
      | Some _ -> (
          (* DUMP DEFFUN *)  
          let typ = (
            match (Ccommon.get_function_type typ) with
            | Some t -> t
            | None -> assert false
          ) in ( 
            match typ with
            | Cexpr.FUNCTION_OF(_, tys) -> (
                let param_types = (
                  match tys with
                  | None -> []
                  | Some ls -> ls
                ) in
                (* in case of function declaration with full parameters name*)
                let param_text = ref "" in
                if ((List.length params) = (List.length param_types)) then
                  param_text := (
                    List.fold_left2 (
                      fun s name _type -> 
                        let param_txt =
                          if (Cexpr.ellipsis_type_p _type) then
                            Cexpr.ellipsis_string
                          else
                            (string_of_typenode _type) ^ " " ^ name
                        in
                        match String.length s with
                        | 0 -> param_txt
                        | _ -> s ^ ", " ^ param_txt)
                      "" params param_types
                  )
                 else
                 (* in case of function declaration lack of parameters name*)
                   param_text := (
                     List.fold_left (
                       fun s _type -> 
                         let param_txt = (string_of_typenode _type ^ " " ) in (
                           match String.length s with
                           | 0 -> param_txt
                           | _ -> s ^ ", " ^ param_txt
                         )
                     ) "" param_types
                   );
                 decl_text := !decl_text ^ "(" ^ !param_text ^ ")");
            | _ -> ()
          );
          (* get attribute of function declaration*)
          let typ = decl.Cexpr.decl_tn in (
            match typ with
            | Cexpr.ATTRIBUTED_TYPE (typenode, gcc_attribute_list) -> (
                let gcc_attributes = ref "" in 
                List.iter (
                  fun a_gcc_attribute -> (
                    match a_gcc_attribute with
                    | Cexpr.ATTR (a_string, a_expr) -> 
                        gcc_attributes := !gcc_attributes ^ 
                            " __attribute__ (( "  ^ a_string ^ " )) "
                  )
                ) gcc_attribute_list;
                decl_text := !decl_text ^ !gcc_attributes;
              )
            | _ -> ()
          );
        ) 
    ) in
    !decl_text) 

let string_of_for_stat_init (stmt : Cexpr.stmt_t) : string = (
    match stmt with
    | Cexpr.STMT(ex) -> (
              match ex with 
              | Cexpr.PSEUDO_PROLOGUE 
              | Cexpr.PSEUDO_EPILOGUE -> ""
              | _ -> (string_of_expr ex false)
              ) 
    | Cexpr.DEFVAR (decl) -> (string_of_header decl)
    | Cexpr.NOTE _ -> ""
    | _ -> assert false 
)

let string_of_for_stat_inits (stmts : Cexpr.stmt_t list) : string = (
    let temps = (List.map string_of_for_stat_init stmts) in
    let temps = List.filter (
        fun a_str -> if (a_str = "") then false else true 
    ) temps in
    String.concat "," temps 
)
(* dump scope stack *)
let string_of_compacted_scope_stack (scope_stack : (string, Cexpr.binding_t) Hashtbl.t)
: string = (
    Hashtbl.fold (fun _ binding result ->
        result ^ "\n" ^ (Cexpr.binding_to_string binding)
    ) scope_stack ""
)
let string_of_scope_stacks (scope_stacks : (string, Cexpr.binding_t) Hashtbl.t *
                                          (string, Cexpr.binding_t) Hashtbl.t)
: string = (
    "name_scope_stack:" ^ "\n"
    ^ string_of_compacted_scope_stack (fst scope_stacks)
    ^ "tag_scope_stack:" ^ "\n"
    ^ string_of_compacted_scope_stack (snd scope_stacks)
    
)
let string_of_scope_stack (scope_stack : (string, Cexpr.binding_t) Hashtbl.t list)
: string = (
    String.concat "\n-------------------------------"
                (List.map string_of_compacted_scope_stack scope_stack)
)

(*
  dump stmt_t
*)
let rec string_of_stmt (stmt : Cexpr.stmt_t) = 
    let res = ref "" in
    (match stmt with
    | Cexpr.DEFUN( header, body) -> (
                  let old_tab = !curr_tab in
                  curr_tab := (!curr_tab)^(tab_string);
                  res := !res ^ "\n" ^ (string_of_header header) ^
                             " {\n"^string_of_stmt body^"}\n\n";
                  curr_tab := old_tab
        )
    | Cexpr.SPEC (specs) 
    	   -> (res:= !res
                   ^ (!curr_tab) ^ "/*@ " ^ (Sexpr.string_of_lspecs specs) ^ 
                              (!curr_tab) ^"@*/\n")
    | Cexpr.ASGN (ex1, ex2) -> res := !res ^ 
                                      (!curr_tab) ^ 
                                      ((string_of_expr ex1 false)^" = "^
                                      (string_of_expr ex2 false)^";\n")
    | Cexpr.GOTO (goto) -> res := !res ^ (!curr_tab) ^ 
                                  ("goto "^goto^";\n")
    | Cexpr.LABEL (label) -> (
             let label_str = Cexpr.lable_to_string label in
             res := !res ^ !curr_tab ^ label_str ^ "\n")
    | Cexpr.IF(ex, stmtst, stmtsf) -> (
             let old_tab = !curr_tab in
             curr_tab := (!curr_tab)^(tab_string);
             res := !res ^ (old_tab)^ ( "if ( "^
                    (string_of_expr ex false)^" ){\n"^
                    (string_of_stmts stmtst) );
             if (List.length stmtsf) > 0 then
                 res := !res ^ (old_tab)^"} else {\n"^
                        (string_of_stmts stmtsf)
             ;
             res := !res ^ (old_tab)^"}\n" ;
             curr_tab := old_tab
       )
    | Cexpr.STMT(ex) -> (
              match ex with 
              | Cexpr.PSEUDO_PROLOGUE 
              | Cexpr.PSEUDO_EPILOGUE -> ()
              | _ -> res := !res ^ (!curr_tab) ^ (string_of_expr ex false)^";\n"
              ) 
    | Cexpr.DEFVAR (decl) -> res := !res ^ (!curr_tab) ^
                                    (string_of_header decl) ^ ";\n" 
    | Cexpr.BEGIN (stmts) -> res := !res ^ (string_of_stmts stmts)
    | Cexpr.SWITCH (expr, stmts) -> (
             res := !res ^ (!curr_tab) ^ "switch (" ^ 
                    (string_of_expr expr false) ^ "){\n";
             let old_tab = !curr_tab in
             curr_tab := (!curr_tab) ^ tab_string;
             res := !res ^ (string_of_stmts stmts) ;
             curr_tab := old_tab;
             res := !res ^ (!curr_tab) ^"}\n"
       )
    | Cexpr.DOWHILE (boo, expr, stmts ) -> (
           match boo with
           | false -> (
                  res := !res ^ !curr_tab ^ "do {\n";
                  let old_tab = !curr_tab in
                  curr_tab := (!curr_tab) ^ tab_string;
                  res := !res ^ (string_of_stmts stmts) ;
                  curr_tab := old_tab;
                  res := !res ^ !curr_tab ^ "}\n"; 
                  res := !res ^ !curr_tab ^ "while (" ^ 
                         (string_of_expr expr false) ^ ");\n"; 
              ) 
           | true -> (
                  res := !res ^ !curr_tab ^ "while (" ^ 
                         (string_of_expr expr false) ^ ") {\n"; 
                  let old_tab = !curr_tab in
                  curr_tab := (!curr_tab) ^ tab_string;
                  res := !res ^ (string_of_stmts stmts) ;
                  curr_tab := old_tab;
                  res := !res ^ !curr_tab ^ "}\n"; 

              )
       )
    | Cexpr.FOR (stmt_inits, expr_cond, expr, stmts ) -> (
          res := !res ^ (!curr_tab) ^ "for (";
          let old_tab = !curr_tab in
          curr_tab := (!curr_tab) ^ tab_string;
          res := !res ^ (string_of_for_stat_inits stmt_inits);
          res := !res ^  "; ";
          res := !res ^ (string_of_expr expr_cond false)^ "; " ^ 
                 (string_of_expr expr false) ^ ") {\n";
          res := !res ^ (string_of_stmts stmts);
          curr_tab := old_tab;
          res := !res ^ !curr_tab ^ "}\n";

       )
    | Cexpr.BREAK -> res := !res ^ (!curr_tab) ^ "break;\n"
    | Cexpr.RETURN (temp) -> (
        let expr_name = (
          match temp with
          | None -> ""
          | Some(expr) -> string_of_expr expr false
        )
        in
        res := !res ^ (!curr_tab) ^ "return " ^ expr_name ^ ";\n"
       )
    | Cexpr.CONTINUE -> res := !res ^ (!curr_tab) ^ "continue;\n"
    | Cexpr.ANNOTATION (ano, expr) -> (
         match ano with
         | "$assert" -> res := !res ^ (!curr_tab) ^ "assert("^
                               (string_of_expr expr false)^");\n";
         | "$assume" -> res := !res ^ (!curr_tab) ^ "assume("^
                               (string_of_expr expr false)^");\n";
         | _ -> ()
       )
    | Cexpr.ASSUME (expr) -> ( res := !res ^ (!curr_tab) ^ "assume("^
                               (string_of_expr expr false)^");\n";)
    | _ -> ()  
   );
   !res

(*
  dump definition list
*)

and string_of_stmts (defs : Cexpr.stmt_t list) = (
   match defs with
   | [] -> ""
   | head::tail -> (
        let stmt_of_head = string_of_stmt head in 
        stmt_of_head ^ (string_of_stmts tail)
    )
)

(* save intermediate result to file *)
let save_inter_result (output_filename : string) (decls : Cexpr.stmt_t list) = (
  (debug (";; WRITING INTERMEDIATE RESULT TO FILE '"
          ^ output_filename ^ "'..."));
  (* init all the global variables for dumping C file purpose *)
  structs_set := StructsSet.empty;
  dump_c_file := true;
  (* save intermediate result to file *)
  try (
    let output_file = open_out output_filename in
    Printf.fprintf output_file "%s\n" (string_of_stmts decls);
    close_out output_file;
    (debug (";; WRITING TO FILE '" ^ output_filename ^ "' DONE")) ;
  )
  with 
  | Sys_error msg ->  
        raise (Sexpr.Specs_error 
            (";; WRITING TO FILE '" ^ output_filename ^ "' FAILED\n"
                 ^ ";; ERROR: " ^ msg, None))
  | Assert_failure (file_name, line, column) ->  
        let error_location = "line: \"" ^ (string_of_int line) ^ 
                        "\" in file :\"" ^ file_name ^ "\"" in
        error (";; WRITING TO FILE '" ^ output_filename ^ "' FAILED\n"
                           ^ ";; ERROR at " ^ error_location);
  (* clear all the global variables *)          
  structs_set := StructsSet.empty;
  dump_c_file := false;
)

let rec string_of_ofile_defs (ofile_defs : (string * Cexpr.stmt_t) list) : string = (  
    List.fold_left
        (fun (curr_str : string) (a_oflie_def : (string * Cexpr.stmt_t)) ->
            let (name, st) = a_oflie_def in
            curr_str ^
            "\n    =======================================================" ^
            "\n    " ^ name ^ " : " ^ (string_of_stmt st)
        )
        "" ofile_defs
)

and string_of_contract_of_field
        (f_pointer_field_hashtbl : ((string * string), string) Hashtbl.t)
        (contract_of_field_hashtbl :
            ((string * string), (((Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t)
                           * Cexpr.expr_t list)) Hashtbl.t)
: string = (
    Hashtbl.fold                                 
        (fun (struct_var, struct_field) struct_name curr_str ->
            let (string_func_con, string_paras) =
                try
                    let (func_con, paras) =
                        Hashtbl.find contract_of_field_hashtbl
                                     (struct_name, struct_field)
                    in
                    let func_con_str =
                        (Sexpr.string_of_function_contract Cexpr.full_expr_to_string
                                                           Cexpr.full_expr_to_string
                                                           func_con) in
                    let paras_str =
                        List.fold_left
                            (fun curr_str a_para ->
                                (curr_str ^ "\n        " ^
                                 (string_of_expr a_para false))
                            ) "" paras
                    in (func_con_str, paras_str)
                with Not_found -> ("", "")
            in

            (curr_str ^
             "\n    =======================================================" ^
             "\n    struct name : " ^ struct_name ^ "    field name : " ^
             struct_field ^ "\n    struct var name : " ^ struct_var ^
             "\n    parameter : " ^ string_paras ^
             "\n    function contract of field : " ^ string_func_con)
        ) f_pointer_field_hashtbl ""
)

and string_of_ofile (ofile : Cexpr.ofile_t) : string = (
    ("\nOfile hash name : " ^ ofile.Cexpr.ofile_hash_name ^
     "\nOfile file name : " ^ ofile.Cexpr.ofile_file_name ^
     "\nOfile definition : " ^ (string_of_ofile_defs !(ofile.Cexpr.ofile_defs)) ^
     "\nOfile function contract of field : " ^
     (string_of_contract_of_field ofile.Cexpr.f_pointer_field_hashtbl
                                  ofile.Cexpr.contract_of_field_hashtbl))
)
