(* parser/specs/sexpr.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)
(* --------------------------------------- *)

(* Location memo *)
type local_pos_t = int * int 	(* line number * character index from bol *)

type global_pos_t = string 	(* filename *)
                    * int 	(* line number *)
                    * int 	(* character index from bol *)

type range_t = string           (* filename       *)
               * local_pos_t    (* start position *)
               * local_pos_t    (* end position   *)

(* C source file name, line number, character number from bol of the specs
   Used to specify exact locations of specifications *)
let current_pos : global_pos_t ref = ref ("(noname)", 0, 0)
let entered_newline = ref false

(* identifier *)
type identifier_t =
     string        

(* Spec *)

(* loop statement contracts *)
and 'pred loop_statement_t = 
    | STMT_RETURN of identifier_t list * 'pred
    | STMT_CONTINUE of identifier_t list * 'pred
    | STMT_BREAK of identifier_t list * 'pred
    | STMT_LOOP_INV of identifier_t list * 'pred

(* statement contracts *)
and ('pred, 'locs) statement_contract_t =                  
    | STMT_ASSUME of 'pred (*assume_clause_t*)
    | STMT_RESET of 'locs assign_t list (*reset_clause_t*)
    | STMT_ASSERT of identifier_t list * 'pred
    | STMT_INIT of string list
    | STMT_RELEASE of string list    

(* behavior *)
and ('pred, 'locs) behavior_t = {
    b_name : identifier_t;
    b_assumes : 'pred list;
    b_conditions : 'pred func_condition_t list;
    b_assigns : 'locs assign_t list
}

(* prohibit *)
and prohibit_t = identifier_t (* prohibit(f) *)

(* assign *)
and 'locs zone_t =
    | Location of 'locs
    | Nothing

and 'locs assign_t = 'locs zone_t * ('locs zone_t list)

(* function condition *)
and 'pred func_condition_t = position_t * 'pred

(* position of function's condition *)
and position_t =  
    | PRE_CONDITION 
    | POST_CONDITION of bool    (* determine if the post-condition could have
                                   full-effect in the caller-side *)

and access_cond_t =
    | READ of position_t
    | WRITE of position_t

and 'pred var_condition_t = access_cond_t * 'pred
   
and context_t = 
    | HARD
    | SOFT
    | PROCESS
    | ANY

and ('pred, 'locs) function_contract_t = {
    fspec_conditions : 'pred func_condition_t list;
    fspec_behaviors : ('pred, 'locs) behavior_t list;
    fspec_assigns : 'locs assign_t list;
    fspec_prohibits : prohibit_t list;
    fspec_local_invs : (identifier_t * 'pred) list;
    fspec_context : context_t;
    fspec_is_handler : bool
}

and inv_strength_t =
    | STRONG_INV
    | WEAK_INV

and except_t = 
    | EXCEPT_WHOLE of string	(*except ( name ) *)
    | EXCEPT_TAIL of string	(* except (tail ( name ) ) *)
    | EXCEPT_HEAD of string	(* except (head ( name ) ) *)

(* global invariant *)
and 'pred global_invariant_t = {
    ginv_strength : inv_strength_t;
    ginv_name : identifier_t;
    ginv_expr : 'pred;
    ginv_exceptions : except_t list option
}

(* type invariant *)
and ('pred, 'logic_type) type_invariant_t = {
    tinv_strength : inv_strength_t;
    tinv_name : identifier_t;
    tinv_type : 'logic_type;
    tinv_param_name : identifier_t;
    tinv_expr : 'pred;
    tinv_exceptions : except_t list option
}
(*
and ghost_code_t = 
    | GHOST_DEF of ghost_define_t 
    | GHOST_ASG of ghost_assign_t 

and ghost_define_t =  ctype_t * 
                      identifier * 
                      init_ghost_code_t option   
and ghost_assign_t = term * term 	 (* ghost a = b; *)

and init_ghost_code_t = term 
 *)  
and 'logic_type model_field_t = 
      (* name_struct , fun_init_name , requires contract or not *)
    | INIT_MODEL_FIELD  of string * string * bool 
    | RELEASE_MODEL_FIELD  of string              (* name_struct *)
    (* name_struct , name_field, type_of_field *)
    | DECL_FIELD of  string * 
                     string * 
                     'logic_type
    (* curr_struct_name * init_fun_name * term_constant *)
    (*
    | NOT_NULL_MODEL_FIELD of string * string * term
    (* curr_struct_name * init_fun_name * term_constant *)
    | CONSTANT_MODEL_FIELD of string * string * term*)

and ('pred, 'locs, 'logic_type) spec_t =
    | SPEC_FUN of ('pred, 'locs) function_contract_t
    | SPEC_VAR of 'pred var_condition_t list
    | SPEC_STMT of ('pred, 'locs) statement_contract_t list
    | STMT_LOOP of 'pred loop_statement_t list
    | SPEC_GLOBAL_INV of 'pred global_invariant_t
    | SPEC_TYPE_INV of ('pred, 'logic_type) type_invariant_t
(*     | SPEC_GHOST of ghost_code_t list       (* ghost code *) *)
    | SPEC_MODEL_FIELD of 'logic_type model_field_t
    | SPEC_ENUM of string list
    | SPEC_ANYTHING                       (* other - when wrong syntax *)

(* ------------------------------- *)
(* Parameterized string_of_* functions *)
let string_of_strings = Scommon.string_of_strings

let string_of_ids = string_of_strings

let string_of_pos pos =
    let ln, cnum_from_bol = pos in
    "(" ^ (string_of_int ln) ^ ", " ^ (string_of_int cnum_from_bol) ^ ")"

let string_of_range (range : range_t) : string =
    let fn, startpos, endpos = range in
    "(" ^ fn
    ^ "; " ^ (string_of_pos startpos)
    ^ "; " ^ (string_of_pos endpos)
    ^ ")"

let string_of_func_conditions (string_of_pred : 'pred -> string)
                              (fun_conds : 'pred func_condition_t list)
                              : string = (
    let pres = List.fold_left (fun conds condition ->
                                   let pos, expr = condition in
                                   match pos with
                                   | PRE_CONDITION -> expr :: conds
                                   | _ -> conds) [] fun_conds
    and posts = List.fold_left (fun conds condition ->
                                    let pos, expr = condition in
                                    match pos with
                                    | POST_CONDITION _ -> expr :: conds
                                    | _ -> conds) [] fun_conds in
    let pre_res1 = string_of_strings 
                       (List.map string_of_pred pres) in
    let pre_res = "<Pre-condition>"
                  ^ pre_res1
		  ^ "</Pre-condition>" in
    let post_res1 = string_of_strings
                       (List.map string_of_pred posts) in
    let post_res = "<Post-condition>" 
                  ^ post_res1
		  ^ "</Post-condition>"
    in
    (pre_res ^ "\n" ^ post_res ^ "\n")
)

let string_of_zone (string_of_loc : 'loc -> string)
                   (zone : 'loc zone_t) : string = (
    match zone with
    | Location expr -> "Location: " ^ (string_of_loc expr)
    | Nothing -> "Nothing")

let string_of_assign (string_of_loc : 'loc -> string)
                     (assign : 'loc assign_t) : string = (
    let zone1, zones = assign in
    let res = List.map (string_of_zone string_of_loc) zones
    in 
    "<"
    ^ (string_of_zone string_of_loc zone1)
    ^ " : "
    ^ (string_of_strings res)
    ^ ">")

let string_of_assigns (string_of_loc : 'loc -> string)
                      (assigns : 'loc assign_t list)
                      : string = (
    let res = List.map (string_of_assign string_of_loc) assigns in
    "<Reset-variable>"
    ^ (string_of_strings res)
    ^ "</Reset-variable>\n"
)

let string_of_behavior_cond (string_of_pred : 'pred -> string)
                            (beha_cond : 'pred list)
                            : string = (
    "<Assumes>\n"
    ^ (string_of_strings (List.map string_of_pred beha_cond))
    ^ "\n</Assumes>\n"
)

let string_of_behavior (string_of_pred : 'pred -> string)
                       (string_of_loc : 'loc -> string)
                       (behav : ('pred, 'loc) behavior_t)
                       : string = (
    match behav with
      {
        b_name = id;
        b_assumes = beh_cond;
        b_conditions = fun_conds;
        b_assigns = assigns
      } ->
        "<Behavior name=\"" ^ id ^ "\">\n"
        ^ (string_of_behavior_cond string_of_pred beh_cond)
        ^ (string_of_func_conditions string_of_pred fun_conds)
        ^ (string_of_assigns string_of_loc assigns)
        ^ "\n</Behavior>\n"
)

let string_of_behaviors (string_of_pred : 'pred -> string)
                        (string_of_loc : 'loc -> string)
                        (behavs : ('pred, 'loc) behavior_t list)
                        : string = (
    let res = 
        List.map (string_of_behavior string_of_pred string_of_loc) behavs
    in
    String.concat " " res
(*    string_of_strings res*)
)
let string_of_prohibits pros = (
    if (List.length pros)>0 then
    "prohibits("^
    (string_of_ids pros)^")"
    else ""
)

let string_of_local_invariant (string_of_pred : 'pred -> string)
                              (inv : identifier_t * 'pred)
                              : string = (
    let name, expr = inv in
    "invariant "
    ^ name
    ^ " {" ^ (string_of_pred expr) ^ "}"
)

let string_of_local_invariants (string_of_pred : 'pred -> string)
                               (invs : (identifier_t * 'pred) list)
                               : string = (
    let res = (List.map (string_of_local_invariant string_of_pred) invs) in
    (String.concat ";\n" res) 
)

let string_of_function_contract (string_of_pred : 'pred -> string)
                                (string_of_loc : 'loc -> string)
                                (fun_s : ('pred, 'loc) function_contract_t)
                                : string = (
    match fun_s with
      {
        fspec_conditions = fun_conds ;
        fspec_behaviors = behavs ;
        fspec_assigns = assigns ;
        fspec_prohibits = pros ;
        fspec_local_invs = invs ;
        fspec_context = _ ;
        fspec_is_handler = _
      } ->
        (string_of_func_conditions string_of_pred fun_conds)
        ^ (string_of_assigns string_of_loc assigns)
        ^ (string_of_behaviors string_of_pred string_of_loc behavs)
        ^ (string_of_prohibits pros)
        ^ (string_of_local_invariants string_of_pred invs)
)

let string_of_var_specs (string_of_pred : 'pred -> string)
                        (spec_vars : 'pred var_condition_t list)
: string =
begin
    let (read_pres, read_posts, write_pres, write_posts) =
        List.fold_left (fun (rpres, rposts, wpres, wposts) condition ->
                           let acc_pos, expr = condition in
                           match acc_pos with
                           | READ PRE_CONDITION ->
                               (expr :: rpres, rposts, wpres, wposts)
                           | READ POST_CONDITION _ ->
                               (rpres, expr :: rposts, wpres, wposts)
                           | WRITE PRE_CONDITION ->
                               (rpres, rposts, expr :: wpres, wposts)
                           | WRITE POST_CONDITION _ ->
                               (rpres, rposts, wpres, expr :: wposts)
                       ) ([], [], [], []) spec_vars in
    let read_res = (
        let read_pre_res1 =
            match read_pres with
            | [] -> ""
            | _ ->
                string_of_strings
                    (List.map string_of_pred read_pres) in
        let read_post_res1 =
            match read_posts with
            | [] -> ""
            | _ ->
                string_of_strings
                    (List.map string_of_pred read_posts) in
        ("<Read-condition>\n" ^ "    <Pre-condition>" 
        ^ read_pre_res1 ^ "</Pre-condition>\n" ^ "    <Post-condition>"
        ^ read_post_res1 ^ "</Post-condition>\n"
        ^ "</Read-condition>\n")) in
    let write_res =
        let write_pre_res1 =
            match write_pres with
            | [] -> ""
            | _ ->
                string_of_strings 
                    (List.map string_of_pred write_pres) in
        let write_post_res1 =
            match write_posts with
            | [] -> ""
            | _ ->
                string_of_strings
                    (List.map string_of_pred write_posts) in
            ("<Write-condition>\n" ^ "    <Pre-condition>" 
            ^ write_pre_res1 ^ "</Pre-condition>\n" ^ "    <Post-condition>"           
            ^ write_post_res1 ^ "</Post-condition>\n"
            ^ "</Write-condition>\n") in
    (read_res ^ write_res) 
end

let string_of_loop (string_of_pred : 'pred -> string)
                   (s : 'pred loop_statement_t)
                   : string = (
    match s with
    | STMT_RETURN (ids, expr) ->(
        if (List.length ids) > 0 then
            "for " ^ (string_of_ids ids)
            ^ ": return (" ^ (string_of_pred expr) ^ ");\n"
        else string_of_pred expr
     ) 
    | STMT_CONTINUE (ids, expr) -> (
        if (List.length ids) > 0 then
            "for " ^ (string_of_ids ids)
            ^  ": continue (" ^ (string_of_pred expr) ^ ");\n"
        else string_of_pred expr
     )
    | STMT_BREAK (ids, expr) -> (
        if (List.length ids) > 0 then
            "for " ^ (string_of_ids ids)
            ^ ": break (" ^ (string_of_pred expr) ^ ");\n"
        else string_of_pred expr
     )
    | STMT_LOOP_INV (ids, expr) -> (
        if (List.length ids) > 0 then
            "for " ^ (string_of_ids ids)
            ^ ": loop invariant (" ^ (string_of_pred expr) ^ ");\n"
        else string_of_pred expr)
)

let string_of_loops (string_of_pred : 'pred -> string)
                    (ss : 'pred loop_statement_t list)
                    : string = (
    let res = (List.map (string_of_loop string_of_pred) ss) in
    string_of_strings res
)

let string_of_statement_contract (string_of_pred : 'pred -> string)
                                 (string_of_loc : 'loc -> string)
                                 (s_spec : ('pred, 'loc) statement_contract_t)
                                 : string = (
    match s_spec with
    | STMT_ASSUME tes -> (
        "assume (" ^ (string_of_pred tes) ^ ")\n"
    )
    | STMT_RESET tes -> (
        let res = List.map (string_of_assign string_of_loc) tes in
        "reset (" ^ (string_of_strings res) ^ ")\n"
    )
    | STMT_ASSERT (ids, expr) -> (
        if (List.length ids) > 0 then
            "for " ^ (string_of_ids ids)
            ^ ": assert (" ^ (string_of_pred expr) ^ ");\n"
        else string_of_pred expr
     )
     | _ -> ""
)

let string_of_global_invariant (string_of_pred : 'pred -> string)
                               (g_spec : 'pred global_invariant_t)
                               : string = (
    let {
          ginv_strength = strength;
          ginv_name = id;
          ginv_expr = expr;
          ginv_exceptions = excs_o
        } = g_spec in
    let temp = match excs_o with
               | None -> ""
               | Some ecxs -> (
                     let r = 
                     List.map(
                         fun exc ->
                             match exc with
                             | EXCEPT_WHOLE str -> str  
                             | EXCEPT_TAIL str -> "tail(" ^ str ^ ")"
                             | EXCEPT_HEAD str -> "head(" ^ str ^ ")")
                         ecxs
                     in
                     "except (" ^ (string_of_strings r) ^ ")"
                )
    and s_strength = match strength with
                     | STRONG_INV -> "Strong"
                     | WEAK_INV -> "Weak"
    in
    s_strength
    ^ "global invariant" ^ id ^ "{"
    ^ (string_of_pred expr) ^ "}" ^ temp
)

let string_of_type_invariant (string_of_pred : 'pred -> string)
                             (string_of_logictype : 'logic_type -> string)
                             (t_spec : ('pred, 'logic_type) type_invariant_t)
                             : string = (
    let {
          tinv_strength = strength;
          tinv_name = id1;
          tinv_type = ty;
          tinv_param_name = id2;
          tinv_expr = expr;
          tinv_exceptions = excs_o
        } = t_spec in
    let temp = match excs_o with
               | None -> ""
               | Some ecxs -> (
                     let r = List.map
                                  (function
                                  | EXCEPT_WHOLE str -> str
                                  | EXCEPT_TAIL str -> "tail(" ^ str ^ ")"
                                  | EXCEPT_HEAD str -> "head(" ^ str ^ ")")
                                  ecxs
                     in
                     "except (" ^ (string_of_strings r) ^ ")"
                )
    and s_strength = match strength with
                     | WEAK_INV -> "Weak"
                     | STRONG_INV -> "Strong"
    in
    s_strength
    ^ "type invariant" ^ id1 ^ "(" ^ (string_of_logictype ty) ^ " "
    ^ id2 ^ "){"
    ^ (string_of_pred expr) ^ "}" ^ temp
)

let string_of_specs (string_of_pred : 'pred -> string)
                    (string_of_loc : 'loc -> string)
                    (string_of_logictype : 'logic_type -> string)
                    (specs : ('pred, 'loc, 'logic_type) spec_t) : string = (
    match specs with
    | SPEC_FUN fun_s ->
        (string_of_function_contract string_of_pred string_of_loc fun_s)
        ^ "\n\n"
    | SPEC_STMT statement_specs ->
        let res = (List.map (string_of_statement_contract string_of_pred
                                                          string_of_loc)
                            statement_specs)
        in
        (string_of_strings res) ^ "\n\n"
    | STMT_LOOP loop_specs ->
        (string_of_loops string_of_pred loop_specs) ^ "\n\n"
    | SPEC_GLOBAL_INV g_spec ->
        (string_of_global_invariant string_of_pred g_spec)
        ^ "\n\n"
    | SPEC_TYPE_INV t_spec ->
        (string_of_type_invariant string_of_pred string_of_logictype t_spec)
        ^ "\n\n"
(*     | SPEC_GHOST ghost_specs -> (string_of_ghost_codes ghost_specs)^"\n\n" *)
    | SPEC_MODEL_FIELD _ ->
        "model field;\n\n"
    | SPEC_ANYTHING    ->
        "anything;\n\n"
    | SPEC_ENUM strs ->
        "enum("^ (string_of_strings strs)^ ")\n\n"
    | SPEC_VAR var_specs ->
        (string_of_var_specs string_of_pred var_specs)
        ^ "\n\n"
)


(***************************************************************)
(* Logical types *)

(** Describes a location in a source file *)
type location = Lexing.position * Lexing.position

exception Specs_error of (string                        (* error message *)
                          * global_pos_t option)        (* position of error *)

exception Not_well_formed of (location * string)

(** Various kinds of integers *)
type ikind =
    IBool       (** [_Bool] *)
  | IChar       (** [char] *)
  | ISChar      (** [signed char] *)
  | IUChar      (** [unsigned char] *)
  | IInt        (** [int] *)
  | IUInt       (** [unsigned int] *)
  | IShort      (** [short] *)
  | IUShort     (** [unsigned short] *)
  | ILong       (** [long] *)
  | IULong      (** [unsigned long] *)
  | ILongLong   (** [long long] (or [_int64] on Microsoft Visual C) *)
  | IULongLong  (** [unsigned long long] (or [unsigned _int64] on Microsoft
                    Visual C) *)

(** Various kinds of floating-point numbers*)
and fkind =
    FFloat      (** [float] *)
  | FDouble     (** [double] *)
  | FLongDouble (** [long double] *)

(** logic types. *)
type logic_type =
  | LTvoid (** C void *)
  | LTinteger (** mathematical integers. *)
  | LTreal (** mathematical real. *)
  | LTint of ikind (** C integral type.*)
  | LTfloat of fkind (** C floating-point type *)
  | LTarray of logic_type (** C array *)
  | LTpointer of logic_type (** C pointer *)
  | LTenum of string (** C enum *)
  | LTstruct of string (** C struct *)
  | LTunion of string (** C union *)
  | LTnamed of string (** user-defined type. *)

(** logic constants. *)
type constant =
    IntConstant of string (** integer constant *)
  | FloatConstant of string (** real constant *)
  | StringConstant of string (** string constant *)
  | WStringConstant of string (** wide string constant *)

(** comparison operators. *)
type relation = Lt | Gt | Le | Ge | Eq | Neq

(** arithmetic and logic binary operators. *)
type binop = Badd | Bsub | Bmul | Bdiv | Bmod | Bbw_and | Bbw_or | Bbw_xor |
             Blshift | Brshift

(** unary operators. *)
type unop = Uminus | Ustar | Uamp | Ubw_not

(** logical expression. The distinction between locations, terms and
    predicate is done during typing.
*)
type lexpr = {
  lexpr_node : lexpr_node; (** kind of expression. *)
  lexpr_loc : location (** position in the source code. *)
}

(* PL is for Parsed Logic *)
(** kind of expression. *)
and lexpr_node =
    (* both terms and predicates *)
  | PLvar of string (** a variable *)
  | PLapp of string * lexpr list (** an application. *)
      (* terms *)
  | PLconstant of constant (** a constant. *)
  | PLunop of unop * lexpr (** unary operator. *)
  | PLbinop of lexpr * binop * lexpr (** binary operator. *)
  | PLdot of lexpr * string (** field access ({t a.x}) *)
  | PLarrow of lexpr * string (** field access ({t a->x})*)
  | PLarrget of lexpr * lexpr (** array access. *)
  | PLold of lexpr (** expression refers to pre-state of a function. *)
  | PLresult (** value returned by a function. *)
  | PLnull (** null pointer. *)
  | PLcast of logic_type * lexpr (** cast. *)
  | PLsizeof of logic_type (** sizeof a type. *)
  | PLsizeofE of lexpr (** sizeof the type of an expression. *)
      (* predicates *)
  | PLfalse (** false (either as a term or a predicate. *)
  | PLtrue (** true (either as a term or a predicate. *)
  | PLrel of lexpr * relation * lexpr (** comparison operator. *)
  | PLand of lexpr * lexpr (** conjunction. *)
  | PLor of lexpr * lexpr (** disjunction. *)
  | PLxor of lexpr * lexpr (** logical xor. *)
  | PLimplies of lexpr * lexpr (** implication. *)
  | PLiff of lexpr * lexpr (** equivalence. *)
  | PLnot of lexpr (** negation. *)
  | PLif of lexpr * lexpr option * lexpr (** conditional operator. *)
  | PLvalid of lexpr (** pointer is valid. *)
  | PLuniqueloc of lexpr (** unique memory location. *)
  | PLuniqueblock of lexpr * lexpr (** unique memory block. *)
  | PLfield of lexpr (** struct_var: every field of struct_var *)

type lspec_t = (lexpr, lexpr, logic_type) spec_t
type lfunction_contract_t = (lexpr, lexpr) function_contract_t
type lglobal_invariant_t = lexpr global_invariant_t
type ltype_invariant_t = (lexpr, logic_type) type_invariant_t
type lloop_statement_t = lexpr loop_statement_t
type lstatement_contract_t = (lexpr, lexpr) statement_contract_t 

type lbehavior_t = (lexpr, lexpr) behavior_t
type lassign_t = lexpr assign_t
type lzone_t = lexpr zone_t
type lfunc_condition_t = lexpr func_condition_t

(* --------------------------------------- *)

(* Artificial empty function contract *)
let empty_fun_contract : ('pred, 'locs) function_contract_t = {
    fspec_conditions = [];
    fspec_behaviors = [];
    fspec_assigns = [];
    fspec_prohibits = [];
    fspec_local_invs = [];
    fspec_context = ANY;
    fspec_is_handler = false
}

let (function_contract : (string, lfunction_contract_t) Hashtbl.t) 
        = Hashtbl.create 10

(**********************************************************)

type name2contract_t = string * lfunction_contract_t

let (funcs : name2contract_t list ref) = ref []
let (fields : name2contract_t list ref) = ref []
let current_fun_contract = ref empty_fun_contract
let current_function_init = ref ""
let current_name_struct = ref ""

let add_contract (name : string) 
                 (func_contract : lfunction_contract_t) 
                 (funcs1 : name2contract_t list): name2contract_t list = (
    if (func_contract != empty_fun_contract) then
       (name, func_contract)::(funcs1)
    else funcs1
)

let contract_of_name2contract (fun_contract : name2contract_t) 
                             : lfunction_contract_t = (
    let _, c = fun_contract in
    c
)

let find_contracts (name : string)
                   (funs_contract : name2contract_t list)
                   : lfunction_contract_t list = (
    let temp = 
        List.filter (
            fun t ->
                let (n, c) = t in
                if (Misc.string_equal n name) then true
                else false
        )funs_contract
    in
    List.map contract_of_name2contract temp
)

let add_curr_component (str : string) : unit = (
    fields := add_contract str 
                           !current_fun_contract 
                           !fields;
    current_fun_contract := empty_fun_contract
)

(**********************************************************)
(* Auxiliary utility functions *)

let make_local_position (rel_pos : Lexing.position) : local_pos_t =
       (rel_pos.Lexing.pos_lnum, rel_pos.Lexing.pos_cnum)

let make_range (rel_startpos : Lexing.position)
               (rel_endpos : Lexing.position)
               : range_t =
    assert(rel_startpos.Lexing.pos_fname == rel_endpos.Lexing.pos_fname);
    let new_startpos = make_local_position rel_startpos 
    and new_endpos = make_local_position rel_endpos
    in (rel_startpos.Lexing.pos_fname, new_startpos, new_endpos)

let make_global_pos (filename : string)
                    (linenumber : int)
                    (cnumber : int)
                    : global_pos_t = (
    (filename, linenumber, cnumber)
)

(* Strips off location information *)
let strip_off_loc_info (ls_with_loc : ('a * range_t option) list)
                   : 'a list =
    List.map fst ls_with_loc

let make_function_contract (conditions : 'pred func_condition_t list)
                           (behaviors : ('pred, 'locs) behavior_t list)
                           (assigns : 'locs assign_t list)
                           (prohibits : prohibit_t list)
                           (local_invariants : (identifier_t * 'pred) list)
                           (context : context_t)
                           (is_handler : bool)
                           : ('pred, 'locs) function_contract_t =
    {
      fspec_conditions = conditions;
      fspec_behaviors = behaviors;
      fspec_assigns = assigns;
      fspec_prohibits = prohibits;
      fspec_local_invs = local_invariants;
      fspec_context = context;
      fspec_is_handler = is_handler
    }

let make_behavior id assumes conditions assigns =
    {
      b_name = id;
      b_assumes = assumes;
      b_conditions = conditions;
      b_assigns = assigns
    }

let make_global_invariant strength id expr exceptions =
    {
      ginv_strength = strength;
      ginv_name = id;
      ginv_expr = expr;
      ginv_exceptions = exceptions
    }

let make_type_invariant strength id tn paramname expr exceptions =
    {
      tinv_strength = strength;
      tinv_name = id;
      tinv_type = tn;
      tinv_param_name = paramname;
      tinv_expr = expr;
      tinv_exceptions = exceptions
    }

let create_false_location () : location = (
    let position = {
        Lexing.pos_fname = "";
        Lexing.pos_lnum = -1;
        Lexing.pos_bol = -1;
        Lexing.pos_cnum = -1;
    } in
    (position, position)
)

let create_lexpr_without_loc (lexpr_node : lexpr_node) : lexpr = (
    let loc = create_false_location () in
    {
        lexpr_node = lexpr_node;
        lexpr_loc = loc;
    }
)

let pre_conds_of_func_conds (fcs : 'pred func_condition_t list)
                        : 'pred list = (
    List.fold_left (fun (pre_conds : 'pred list)
                        (func_cond : 'pred func_condition_t) ->
                        (begin
			    let p, cond = func_cond in
                            match p with
			    | PRE_CONDITION -> pre_conds @ [cond]
			    | _ -> pre_conds
			end))
		   []
		   fcs
) 

let post_conds_of_func_conds (fcs : 'pred func_condition_t list)
                           : 'pred list = (
    List.fold_left (fun pre_conds func_cond -> 
                        (begin
			    let p, cond = func_cond in
                            match p with
			    | POST_CONDITION _ -> pre_conds @ [cond]
			    | _ -> pre_conds
			end))
		   []
		   fcs
)		   

(******************************)
let rec check_result_list_lexpr (lexprs : lexpr list) = (
   match lexprs with
   | [] -> false
   | head :: tail -> if (check_result_lexpr head) then true
                      else (check_result_list_lexpr tail)
)

and check_result_lexpr (lexpr : lexpr) = 
(begin
    match lexpr.lexpr_node with
    | PLresult -> true
    | PLapp (_, lexprs) ->
      check_result_list_lexpr lexprs
    | PLunop (_, lexpr1)
    | PLdot (lexpr1, _)
    | PLarrow (lexpr1, _)
    | PLold lexpr1
    | PLcast (_, lexpr1)
    | PLsizeofE lexpr1
    | PLnot lexpr1
    | PLvalid lexpr1
    | PLuniqueloc lexpr1
    | PLuniqueblock (lexpr1, _) ->
      check_result_lexpr (lexpr1)
    | PLbinop (lexpr1, _,lexpr2)
    | PLarrget (lexpr1, lexpr2)
    | PLrel (lexpr1, _, lexpr2)
    | PLand (lexpr1, lexpr2)
    | PLor (lexpr1, lexpr2)
    | PLxor (lexpr1, lexpr2)
    | PLimplies (lexpr1, lexpr2)
    | PLiff (lexpr1, lexpr2) ->
      (check_result_lexpr lexpr1) || (check_result_lexpr lexpr2)
    | PLif (lexpr1, lexpr2_op, lexpr3) ->
      (match lexpr2_op with
      | None -> 
        (check_result_lexpr lexpr1) || (check_result_lexpr lexpr3)
      | Some lexpr2 ->
        (check_result_lexpr lexpr1) || (check_result_lexpr lexpr2) ||
        (check_result_lexpr lexpr3)
      )
    | PLfalse
    | PLtrue
    | PLvar _
    | PLconstant _
    | PLnull
    | PLfield _
    | PLsizeof _ ->
      false
end)

exception Found_result;;

(* Find construct '$result' in fun_condition_t.
   Return true if this construct is found.
   Otherwise return false.
*)
let find_result_in_func_cond (func_cond : lfunc_condition_t) : bool = (
    let lexpr = snd func_cond in
    check_result_lexpr lexpr;
)

(* Find construct '$result' in fun_condition_t list.
   Raise exception Found_result if this construct is found.
*)
let rec find_result_in_func_conds (func_conds : lfunc_condition_t list) 
: unit = (
    List.iter ( fun a_func_cond ->
        if (find_result_in_func_cond a_func_cond) then 
            raise Found_result;
    ) func_conds;
)

(* Find construct '$result' in behavior.
   Raise exception Found_result if this construct is found.
*)
let rec find_result_in_behavior_spec (behavior_spec : lbehavior_t) 
: unit = (
    match behavior_spec with
      {
        b_name = _;
        b_assumes = assumes;
        b_conditions = conditions;
        b_assigns = assigns
      } ->
    check_result_list_lexpr assumes;
    find_result_in_func_conds conditions;
    check_result_assigns assigns;
)

(* Find construct '$result' in behavior list.
   Raise exception Found_result if this construct is found.
*)
and find_result_in_behavior_specs (behavior_specs : lbehavior_t list) 
: unit = (
    List.iter find_result_in_behavior_spec behavior_specs;
)

(* Find construct result in spec function_contract.
   Return true if this construct is found.
   Otherwise return false
 *)
and check_result_assigns (assigns : lassign_t list)
: unit = (
    List.iter ( fun (zone, zones) ->
        List.iter (fun a_zone ->
           match a_zone with
           | Location lexpr ->
             (match check_result_lexpr lexpr with
             | true ->
               raise Found_result
             | false -> ())
           | Nothing -> () 
        ) (zone :: zones)
    ) assigns;
)

let find_result_in_spec_func (spec : lfunction_contract_t) : bool = (
    match spec with
      {
        fspec_conditions = func_conds;
        fspec_behaviors = behavior_specs;
        fspec_assigns = assigns;
        fspec_prohibits = _;
        fspec_local_invs = ident_lexpr_list;
        fspec_context = _;
        fspec_is_handler = _
      } ->
        try
            find_result_in_func_conds func_conds;
            find_result_in_behavior_specs behavior_specs;
            check_result_assigns assigns;
            check_result_list_lexpr
            (List.map (snd) ident_lexpr_list);
            false
        with Found_result -> true
)

(************************************************)
let merge_2_fun_contract s1 s2 = 
(
    let (fun_conds1, behavs1, inout1, p1, is1) = s1
    and (fun_conds2, behavs2, inout2, p2, is2) = s2 in
    let fun_conds = fun_conds1 @ fun_conds2
    and behavs = behavs1 @ behavs2
    and inout =  inout1 @ inout2
    and p = p1 @ p2
    and is = is1 @ is2
    in (fun_conds, behavs, inout, p, is)
) 
(* add_list_spec_contracts *)     
let merge_fun_contracts (specs : ('pred, 'locs, 'logic_type) spec_t list)
: ('pred, 'locs, 'logic_type) spec_t = (
    let merge_two_fun_contracts spec1 spec2 = 
        match spec1, spec2 with
        | SPEC_FUN { 
                     fspec_conditions = fun_conds1;
                     fspec_behaviors = behavs1;
                     fspec_assigns = inout1;
                     fspec_prohibits = p1;
                     fspec_local_invs = is1;
                     fspec_context = c1;
                     fspec_is_handler = ishandler1
                   },
          SPEC_FUN { 
                     fspec_conditions = fun_conds2;
                     fspec_behaviors = behavs2;
                     fspec_assigns = inout2;
                     fspec_prohibits = p2;
                     fspec_local_invs = is2;
                     fspec_context = c2;
                     fspec_is_handler = ishandler2
                   } ->
            let fun_conds = fun_conds1 @ fun_conds2
            and behavs = behavs1 @ behavs2
            and inout = inout1 @ inout2
            and p = p1 @ p2
            and is = is1 @ is2
            in 
            SPEC_FUN (make_function_contract fun_conds
                                             behavs
                                             inout
                                             p
                                             is
                                             c1
                                             ishandler1)
        | _ -> spec1
    in
    List.fold_left merge_two_fun_contracts (SPEC_FUN(empty_fun_contract)) specs
)

(**********************************************************)

let string_of_ikind i = (
    match i with
    | IBool -> "bool"                   (** [_Bool] *)
    | IChar -> "char"                   (** [char] *)
    | ISChar -> "signed char"           (** [signed char] *)
    | IUChar -> "unsigned char"         (** [unsigned char] *)
    | IInt -> "int"                     (** [int] *)
    | IUInt -> "unsigned int"           (** [unsigned int] *)
    | IShort -> "short"                 (** [short] *)
    | IUShort -> "unsigned short"       (** [unsigned short] *)
    | ILong -> "long"                   (** [long] *)
    | IULong -> "unsigned long"         (** [unsigned long] *)
    | ILongLong -> "long long"          
        (** [long long] (or [_int64] on Microsoft Visual C) *)
    | IULongLong -> "unsigned long long"  
        (** [unsigned long long] (or [unsigned _int64] on Microsoft Visual C) *)
)

let string_of_fkind f = (
    match f with
    | FFloat -> "float"                 (** [float] *)
    | FDouble -> "double"               (** [double] *)
    | FLongDouble -> "long double"      (** [long double] *)
)
 
let string_of_constant (const : constant) : string = (
    match const with
    | IntConstant s                     (** integer constant *)
    | FloatConstant s                   (** real constant *)
    | StringConstant s                  (** string constant *)
    | WStringConstant s                 (** wide string constant *)
        -> s)

let string_of_rel_op (op : relation) : string = (
    match op with
    | Lt -> "<"
    | Gt -> ">"
    | Le -> "<="
    | Ge -> ">="
    | Eq -> "=="
    | Neq -> "!=")

let string_of_binop (op : binop) : string = (
    match op with
    | Badd -> "+"
    | Bsub -> "-"
    | Bmul -> "*"
    | Bdiv -> "/"
    | Bmod -> "%"
    | Bbw_and -> "&"
    | Bbw_or -> "|"
    | Bbw_xor -> "^"
    | Blshift -> "<<"
    | Brshift -> ">>")

let string_of_unop (op : unop) : string = (
    match op with
    | Uminus -> "-"
    | Ustar -> "*"
    | Uamp -> "&"
    | Ubw_not -> "!")

let rec string_of_logic_type (ltype : logic_type) : string = (
    match ltype with
    | LTvoid -> "void"                          (** C void *)
    | LTinteger -> "integer"                    (** mathematical integers. *)
    | LTreal -> "real"                          (** mathematical real. *)
    | LTint i -> string_of_ikind i              (** C integral type.*)
    | LTfloat f -> string_of_fkind f            (** C floating-point type *)
    | LTarray ltype -> string_of_logic_type ltype (** C array *)
    | LTpointer ltype -> string_of_logic_type ltype (** C pointer *)
    | LTenum e -> "enum " ^ e                   (** C enum *)
    | LTstruct s -> "struct " ^ s               (** C struct *)
    | LTunion u -> "union" ^ u                  (** C union *)
    | LTnamed s -> s)

let rec string_of_expr_node (exprn : lexpr_node) : string = (
    match exprn with
    | PLvar s -> s
    | PLapp (func_name, params) -> (
        let s = List.fold_left (fun s a ->
            s ^ " " ^ (string_of_expr_node a.lexpr_node)
        ) "" params in       
        func_name ^ "(" ^ s ^ ")")
    | PLconstant const -> string_of_constant const
    | PLunop (op, expr) -> (
        let s_op = string_of_unop op
        and s_expr = string_of_lexpr expr in
        s_op ^ "(" ^ s_expr ^ ")")
    | PLbinop (expr1, op, expr2) -> (
        let s_op = string_of_binop op
        and s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " " ^ s_op ^ " " ^ s_expr2 ^ ")")
    | PLdot (expr, field_name) -> (
        let s_expr = string_of_lexpr expr in
        s_expr ^ "." ^ field_name)
    | PLarrow (expr, field_name) -> (
        let s_expr = string_of_lexpr expr in
        s_expr ^ "->" ^ field_name)
    | PLarrget (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        s_expr1 ^ "[" ^ s_expr2 ^ "]")
    | PLold expr -> (
        let s_expr = string_of_lexpr expr in
        "$old(" ^ s_expr ^ ")")
    | PLresult -> "$result"
    | PLnull -> "$null"
    | PLcast (ltype, expr) -> (
        let s_type = string_of_logic_type ltype
        and s_expr = string_of_lexpr expr in
        "(" ^ s_type ^ ")(" ^ s_expr ^ ")")
    | PLsizeof ltype -> (
        let s_type = string_of_logic_type ltype in
        "sizeof(" ^ s_type ^ ")")
    | PLsizeofE expr -> (
        let s_expr = string_of_lexpr expr in
        "sizeof(" ^ s_expr ^ ")")
      (* predicates *)
    | PLfalse -> "$false"
    | PLtrue -> "$true" 
    | PLrel (expr1, rel_op, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        let s_rel_op = string_of_rel_op rel_op in
        "(" ^ s_expr1 ^ " " ^ s_rel_op ^ " " ^ s_expr2 ^ ")")
    | PLand (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " && " ^ s_expr2 ^ ")")
    | PLor (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " || " ^ s_expr2 ^ ")")
    | PLxor (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " ^^ " ^ s_expr2 ^ ")")
    | PLimplies (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " ==> " ^ s_expr2 ^ ")")
    | PLiff (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2 in
        "(" ^ s_expr1 ^ " <==> " ^ s_expr2 ^ ")")
    | PLnot expr -> (
        let s_expr = string_of_lexpr expr in
        "!(" ^ s_expr ^ ")")
    | PLif (expr1, expr2_opt, expr3) -> (
        let s_expr1 = string_of_lexpr expr1        
        and s_expr2 = match expr2_opt with
                      | Some expr2 -> string_of_lexpr expr2
                      | None -> ""
        and s_expr3 = string_of_lexpr expr3 in
        "(" ^ s_expr1 ^ "?" ^ s_expr2 ^ ":" ^ s_expr3 ^ ")")
    | PLvalid expr -> (
        let s_expr = string_of_lexpr expr in
        "$valid(" ^ s_expr ^ ")")
    | PLfield expr -> (
        (string_of_lexpr expr) ^ ".*")
    | PLuniqueloc expr -> (
        let s_expr = string_of_lexpr expr in
        "$unique(" ^ s_expr ^ ")")
    | PLuniqueblock (expr1, expr2) -> (
        let s_expr1 = string_of_lexpr expr1
        and s_expr2 = string_of_lexpr expr2
        in
        "$unique(" ^ s_expr1 ^ ", " ^ s_expr2 ^ ")"))


and string_of_lexpr (expr : lexpr) : string = (
    match expr with
    | { 
        lexpr_node = exprn;
        lexpr_loc = _ ;
      } -> (
        let s = string_of_expr_node exprn in
        s))

(* ---------------------------------------------- *)
let check_old_in_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLold _ -> true
    | _ ->  false  
end

let check_result_in_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLresult -> true
    | _ ->  false  
end

let check_unique_in_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLuniqueloc _
    | PLuniqueblock _ -> true
    | _ ->  false  
end

let check_field_in_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLfield _ -> true
    | _ -> false 
end

let check_zero_in_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLrel (lex, Eq, _) -> (check_field_in_lexpr lex)
    | _ ->  false  
end

let check_unique_isnt_in_logic_lexpr (cond : lexpr) : bool =
begin
    match cond.lexpr_node with
    | PLbinop (lex1, _, lex2)
    | PLarrget (lex1, lex2)
    | PLrel (lex1, _, lex2) ->
        (check_unique_in_lexpr lex1) || (check_unique_in_lexpr lex2)
    | PLarrow (lex, _)
    | PLdot (lex, _)
    | PLfield lex
    | PLuniqueloc lex
    | PLuniqueblock (lex, _)
    | PLold lex
    | PLvalid lex
    | PLunop (_, lex) ->
        check_unique_in_lexpr lex 
    | _ ->
        false
end

let get_first_loc_var_spec (lspec : lspec_t) : global_pos_t option =
begin
    match lspec with
    | SPEC_VAR var_spec -> (
        let first_lex_op =
            match var_spec with
            | [] -> None
            | var_cond :: remains -> Some (snd var_cond) in
        let pos_op =
            match first_lex_op with
            | None -> None
            | Some lex -> (
                let loc = fst lex.lexpr_loc in
                let pos =
                    (loc.Lexing.pos_fname, loc.Lexing.pos_lnum,
                     loc.Lexing.pos_cnum) in
                Some pos
            ) in
        (pos_op)
    )
    | _ -> assert false
end

let old_in_old_error_msg = "Found_old_in_old"

let check_old_in_old (cond : lexpr) : unit =
begin
    if (check_old_in_lexpr cond) then
        raise (Specs_error (old_in_old_error_msg, None))
end

(* Check for construct '\old(...)' and construct '\result' in pre-conditions *)
let check_validity_of_pre_condition (do_check : lexpr -> bool) (cond : lexpr) 
: unit = 
begin
    if (do_check cond) then (
        let loc = fst cond.lexpr_loc in
        let pos =
            (loc.Lexing.pos_fname, loc.Lexing.pos_lnum, loc.Lexing.pos_cnum)
        in
        raise (Specs_error ("Construct "
                            ^ (string_of_lexpr cond)
                            ^ " appears in a pre-condition.",
                            Some pos)))
end

(* Visit lexpr tree in pre_oder to process 'process_lexpr' in every node *)
let rec visit_lexpr (cond : lexpr) (process_lexpr : lexpr -> unit) : unit =
begin
    process_lexpr cond;
    match cond.lexpr_node with
    | PLapp (_, lexprs) ->
        List.iter (fun lex -> visit_lexpr lex process_lexpr) lexprs
    | PLold lex
    | PLunop (_, lex)
    | PLdot (lex, _)
    | PLarrow (lex, _)
    | PLcast (_, lex)
    | PLsizeofE lex
    | PLnot lex
    | PLvalid lex
    | PLuniqueloc lex
    | PLuniqueblock (lex, _) ->
        visit_lexpr lex process_lexpr
    | PLbinop (lex1, _, lex2)
    | PLarrget (lex1, lex2)
    | PLrel (lex1, _, lex2)
    | PLand (lex1, lex2)
    | PLor (lex1, lex2)
    | PLxor (lex1, lex2)
    | PLimplies (lex1, lex2)
    | PLiff (lex1, lex2) -> (
        visit_lexpr lex1 process_lexpr;
        visit_lexpr lex2 process_lexpr
    )
    | PLif (lex1, lex2_op, lex3) -> (
        match lex2_op with
        | None -> (
            visit_lexpr lex1 process_lexpr;
            visit_lexpr lex3 process_lexpr
        )
        | Some lex2 -> (
            visit_lexpr lex1 process_lexpr;
            visit_lexpr lex2 process_lexpr;
            visit_lexpr lex3 process_lexpr
        )
    )
    | PLresult
    | PLfalse
    | PLtrue
    | PLvar _
    | PLconstant _
    | PLnull
    | PLfield _
    | PLsizeof _ -> ()
end

let string_of_lspecs : lspec_t -> string =
    string_of_specs string_of_lexpr
                    string_of_lexpr
                    string_of_logic_type

let dump_specs specs = print_string (string_of_lspecs specs)

(* 
   get ID of term
*)
let get_name_of_binary_op (bin_op : binop) : string = (
    match bin_op with
    | Badd -> "plus"
    | Bsub -> "minus"
    | Bmul -> "mult"
    | Bdiv -> "div"
    | Bmod -> "mod"
    | Bbw_and -> "and"
    | Bbw_or -> "or"
    | Bbw_xor -> "hat"
    | Blshift
    | Brshift -> assert false (* temporary do not process *) 
)
let get_name_of_unary_op (u_op : unop): string = (
    match u_op with
    | Uminus -> "neg"
    | Ustar -> "pref_of"
    | Uamp -> "addr_of"
    | Ubw_not -> "not" 
)

let rec get_name_of_term (term : lexpr) : string = (
     match term.lexpr_node with
    | PLvar id -> id
    | PLunop (unop, lexpr1) ->
      (get_name_of_unary_op unop) ^ "_" ^ (get_name_of_term lexpr1) 
    | PLbinop (lexpr1, binop, lexpr2) ->
      (get_name_of_term lexpr1) ^ "_" ^
      (get_name_of_binary_op binop) ^ "_" ^
      (get_name_of_term lexpr2)
    | PLdot (lexpr1, str) ->
      ((get_name_of_term lexpr1) ^ "." ^ str)
    | PLarrow (lexpr1, str) ->
      ((get_name_of_term lexpr1) ^ "->" ^ str)
    | PLarrget (lexpr1, lexpr2) ->
      (get_name_of_term lexpr1) ^ "_" ^
      (get_name_of_term lexpr2)
    | PLresult -> "result"
    | PLconstant const ->
      (string_of_constant const)
    | PLnull
    | PLfalse
    | PLtrue
    | PLsizeof _
    | PLsizeofE _
    | PLold _ -> ""
    | _ -> assert false (* temporary do not process *)
(*   | TERM_ADDR ter                  -> "addr_of_"^(get_name_of_term ter)*)

)
(****************************************)
(* 
   get var name of lexpr 
*)
let rec get_var_name_of_lexpr (lexpr : lexpr) : string list = (
    match lexpr.lexpr_node with
    | PLvar id -> [id]
    | PLapp (_, lexprs) -> (
      List.fold_left (fun strs a_lexpr ->
          strs @ (get_var_name_of_lexpr a_lexpr)
      ) [] lexprs;
      )
    | PLunop (_, lexpr1)
    | PLdot (lexpr1, _)
    | PLarrow (lexpr1, _)
    | PLold lexpr1
    | PLcast (_, lexpr1)
    | PLsizeofE lexpr1
    | PLnot lexpr1
    | PLvalid lexpr1 ->
      get_var_name_of_lexpr lexpr1
    | PLuniqueloc lexpr1
    | PLuniqueblock (lexpr1, _) ->
      get_var_name_of_lexpr lexpr1
    | PLbinop (lexpr1, _, lexpr2)
    | PLarrget (lexpr1, lexpr2)
    | PLrel (lexpr1, _, lexpr2)
    | PLand (lexpr1, lexpr2)
    | PLor (lexpr1, lexpr2)
    | PLxor (lexpr1, lexpr2)
    | PLimplies (lexpr1, lexpr2)
    | PLiff (lexpr1, lexpr2) ->
      (get_var_name_of_lexpr lexpr1) @ (get_var_name_of_lexpr lexpr2)
    | PLif (lexpr1, lexpr2_op, lexpr3) ->
      (match lexpr2_op with
      | None ->
        (get_var_name_of_lexpr lexpr1) @ (get_var_name_of_lexpr lexpr3)
      | Some lexpr2 ->
        (get_var_name_of_lexpr lexpr1) @ (get_var_name_of_lexpr lexpr2)
        @ (get_var_name_of_lexpr lexpr3))
    | PLfalse -> ["false"]
    | PLtrue -> ["true"]
    | PLresult
    | PLnull
    | PLconstant _
    | PLfield _
    | PLsizeof _ ->
      []
)

let rec get_name_of_predicate (pred : lexpr) : string list = (
    match pred.lexpr_node with
    | PLfalse -> ["false"]
    | PLtrue -> ["true"]
    | PLrel (lexpr1, _, lexpr2)
    | PLand (lexpr1, lexpr2) 
    | PLor (lexpr1, lexpr2)
    | PLxor (lexpr1, lexpr2)
    | PLimplies (lexpr1, lexpr2)
    | PLiff (lexpr1, lexpr2) ->
      (get_var_name_of_lexpr lexpr1) @ 
      (get_var_name_of_lexpr lexpr2)
    | PLnot lexpr1 ->
      get_name_of_predicate lexpr1
    | PLif (lexpr1, lexpr2_op, lexpr3) ->
      (match lexpr2_op with
      | None -> (
        (get_name_of_predicate lexpr1) @ (get_name_of_predicate lexpr3))
      | Some lexpr2 -> (
        (get_name_of_predicate lexpr1) @ (get_name_of_predicate lexpr2) @
        (get_name_of_predicate lexpr3))
      )
    | PLvalid lexpr1 ->
      get_var_name_of_lexpr lexpr1
    | PLcast _
    | PLapp _
    | PLconstant _ -> assert false (* temporary do not process *)
    | _ -> assert false (* temporary do not process *)
)
  
(* 
   get var name of list_predicate  
*)
and get_name_of_list_predicate (preds : lexpr list) = 
    List.flatten(List.map get_name_of_predicate preds)

(*************************************)
(*********************************************)
let rec change_term_to_sref (var_name : string)
                            (term : lexpr) : lexpr = (
    let var = { lexpr_loc = term.lexpr_loc;
                lexpr_node = PLvar (var_name)} in
    let new_lexpr_node = (
    match term.lexpr_node with
    | PLvar (field) -> PLdot (var, field)
    | PLapp (str, lexprs) -> (
      let new_lexprs = (
          List.map (
              fun lexpr -> 
                  change_term_to_sref var_name lexpr
          ) lexprs
      ) in
      PLapp (str, new_lexprs)
      )
    | PLunop (unop, lexpr) ->
      PLunop (unop, change_term_to_sref var_name lexpr)
    | PLbinop (lexpr1, binop, lexpr2) ->
      PLbinop (change_term_to_sref var_name lexpr1, binop,
               change_term_to_sref var_name lexpr2)
    | PLold lexpr1 ->
      PLold (change_term_to_sref var_name lexpr1)
    | PLsizeofE lexpr1 ->
      PLsizeofE (change_term_to_sref var_name lexpr1)
    | PLif (lexpr1, lexpr2_op, lexpr3) ->
      let new_lexpr2_op = (
          match lexpr2_op with
          | None -> None
          | Some lexpr2 ->
            Some (change_term_to_sref var_name lexpr2) 
      ) in
      PLif (change_term_to_sref var_name lexpr1,
            new_lexpr2_op,
            change_term_to_sref var_name lexpr3)

    | PLfalse
    | PLtrue
    | PLresult
    | PLsizeof _
    | PLconstant _ ->
      term.lexpr_node
    | _ ->
      raise (Specs_error ((string_of_lexpr term)
	                     ^ " unsupported in model field", None))
    ) in
      {
        lexpr_node = new_lexpr_node;
        lexpr_loc = term.lexpr_loc
      }
 )


(***********************************************)
(* get name of old_variable*)
let rec get_old_term_from_function_contract 
            (func_conds: lfunc_condition_t list)
            (behavs : lbehavior_t list) 
            : lexpr list = (
    let b_func_conds = (
        List.flatten(
            List.map(
                function
                  {
                    b_name = _;
                    b_assumes = _;
                    b_conditions = b_conds;
                    b_assigns = _
                  } -> b_conds )
            behavs)) 
    in 
    let total_func_conds = (func_conds @ b_func_conds) in
    let lexprs = List.map snd  total_func_conds in
    List.flatten (List.map get_old_term_from_lexpr lexprs)    
)

and get_old_term_from_lexpr (lexpr : lexpr) : lexpr list = (
    match lexpr.lexpr_node with
    | PLold lexpr1 ->
        [lexpr1]
    | PLapp (_, lexprs) ->
        List.fold_left 
            (fun lexs a_lexpr ->
                lexs @ (get_old_term_from_lexpr a_lexpr)
            ) [] lexprs;
    | PLarrget (lexpr1, lexpr2)
    | PLrel (lexpr1, _, lexpr2)
    | PLand (lexpr1, lexpr2)
    | PLor (lexpr1, lexpr2)
    | PLxor (lexpr1, lexpr2)
    | PLimplies (lexpr1, lexpr2)
    | PLiff (lexpr1, lexpr2)
    | PLbinop (lexpr1, _, lexpr2) ->
        (get_old_term_from_lexpr lexpr1) @ (get_old_term_from_lexpr lexpr2)
    | PLunop (_, lexpr1)
    | PLdot (lexpr1, _)
    | PLcast (_, lexpr1)
    | PLsizeofE lexpr1
    | PLvalid lexpr1
    | PLuniqueloc lexpr1
    | PLuniqueblock (lexpr1, _)
    | PLnot (lexpr1)
    | PLarrow (lexpr1, _) ->
        (get_old_term_from_lexpr lexpr1)
    | PLif (lexpr1, lexpr2_op, lexpr3) -> (
        match lexpr2_op with
        | None ->
            (get_old_term_from_lexpr lexpr1) @ (get_old_term_from_lexpr lexpr3)
        | Some lexpr2 ->
            (get_old_term_from_lexpr lexpr1) @ (get_old_term_from_lexpr lexpr2)
            @ (get_old_term_from_lexpr lexpr3)
    )
    | PLfield _
    | PLvar _
    | PLconstant _
    | PLresult
    | PLnull
    | PLsizeof _
    | PLfalse
    | PLtrue ->
      []
)

(* Data structure and auxiliary functions for handling cutlist *)
type file_location_t = {
    fl_abs_path : string;
    fl_filename : string
}

type function_entry_t = {
    fe_filelocation_opt : file_location_t option;
    fe_funcname : string;
    fe_raw_data : string;
    fe_location : (string * int) option
}

let nil_function_entry = {
    fe_filelocation_opt = None;
    fe_funcname = "";
    fe_raw_data = "";
    fe_location = None
}

let (cut_functions : function_entry_t list ref) = ref []

(* Utility functions of type file_location_t *)

let make_file_location (path : string) : file_location_t = (
    if (Sys.file_exists path) then (
        let dir_name = Filename.dirname path
        and file_name = Filename.basename path
        in
        let current_dir = Sys.getcwd () in
        (Sys.chdir dir_name);
        let temp = Sys.getcwd () in
        (Sys.chdir current_dir);
        {
          fl_abs_path = temp;
          fl_filename = file_name
        })
    else (
        raise Not_found))

let string_of_file_location (fl : file_location_t) : string =
    Filename.concat fl.fl_abs_path fl.fl_filename

let same_file fl1 fl2 =
    (string_of_file_location fl1) = (string_of_file_location fl2)

let is_in_cutlist (cut_func_lst : function_entry_t list)
                  (file_loc : file_location_t)
                  (func_name : string)
: bool = 
begin
    List.exists 
        (fun a_cut_func ->
            match a_cut_func.fe_filelocation_opt with
            | None -> (String.compare a_cut_func.fe_funcname func_name = 0)
            | Some fe_file_loc -> 
                 (same_file fe_file_loc file_loc) &&
                 (String.compare a_cut_func.fe_funcname func_name = 0)         
        )
        cut_func_lst
end

let unique_error_msg = "Invalid use of construct $unique"

let zero_error_message = "Invalid use of construct $zero"

(** Check the invalid exist of a structure in a lexpr, such as $unique, $rezo,
    ...    
*)
let check_validity_of_a_structure (do_check : lexpr -> bool) 
                                  (err_message : string)
                                  (cond : lexpr) : unit =
begin
    if (do_check cond) then
        let loc = fst cond.lexpr_loc in
        let pos =
            (loc.Lexing.pos_fname, loc.Lexing.pos_lnum, loc.Lexing.pos_cnum)
        in
        raise (Specs_error (err_message, Some pos))
end

let check_validity_of_post_condition (cond : lexpr) : unit =
begin
    (check_validity_of_a_structure check_unique_in_lexpr 
                                   unique_error_msg
                                   cond);
    (check_validity_of_a_structure check_zero_in_lexpr 
                                   zero_error_message
                                   cond);
end

let check_validity_of_function_contract (spec : lspec_t) : unit =
begin
    match spec with
    | SPEC_FUN func_cont -> (
        let { fspec_conditions = conds;
              fspec_behaviors = behavs;
              fspec_assigns = _;
              fspec_prohibits = _;
              fspec_local_invs = _;
              fspec_context = _;
              fspec_is_handler = _;
            } = func_cont in
        let b_func_conds = (
            List.flatten(
                List.map(
                    function
                        {
                            b_name = _;
                            b_assumes = _;
                            b_conditions = b_conds;
                            b_assigns = _
                        } -> b_conds )
                    behavs)) 
        in 
        let total_func_conds = (conds @ b_func_conds) in
        let post_conds = 
            List.filter 
                (fun a_func_cond -> 
                    match (fst a_func_cond) with
                    | PRE_CONDITION -> false
                    | POST_CONDITION _ -> true   
                )
                total_func_conds
        in
        let lexprs = List.map snd post_conds in   
        List.iter (fun a_lexpr ->
                      visit_lexpr a_lexpr
                                  check_validity_of_post_condition)
                  lexprs             
    )
    | SPEC_VAR _
    | SPEC_STMT _
    | STMT_LOOP _
    | SPEC_GLOBAL_INV _
    | SPEC_TYPE_INV _
    | SPEC_MODEL_FIELD _
    | SPEC_ENUM _
    | SPEC_ANYTHING -> assert false         
end
    
let identical (fl : file_location_t) (filename : string) = (
    let new_fl = make_file_location filename in
    same_file fl new_fl)
   
(* Parses a cutlist file *)
    
let make_function_entry (filename : string option)
                        (funcname : string)
                        (raw_data : string)
                        ?(loc=None) () =
    match filename with
    | Some fn ->
        {
          fe_filelocation_opt = Some (make_file_location fn);
          fe_funcname = funcname;
          fe_raw_data = raw_data;
          fe_location = loc
        }
    | None ->
        {
          fe_filelocation_opt = None;
          fe_funcname = funcname;
          fe_raw_data = raw_data;
          fe_location = loc
        }

let parse_function_entry (input : string) ?(loc=None) () : function_entry_t = (
    let filenameopt, funcname =
        try
            let filename, funcname =
                ExtString.String.split input "#"
            in (Some filename, funcname)
        with
        | ExtString.Invalid_string ->
            None, input
    in
    make_function_entry filenameopt funcname input ~loc:loc ())

let parse_cutlist_file (file_name : string) : function_entry_t list = (
    if not (Sys.file_exists file_name) then (
        Scommon.error ("Cutlist file not found : `" ^ file_name ^ "'");
        failwith "File not found");
    if (Sys.is_directory file_name) then (
        Scommon.error ("Cutlist file not found : `"
                       ^ file_name
                       ^ "'. Found a directory");
        failwith "File not found");
    let cutlist_dir = Filename.dirname file_name
    and cutlist_file = Filename.basename file_name
    and current_dir = Sys.getcwd ()
    in
    let cut_entries = ref [] in
    let chan = open_in file_name in
    (Sys.chdir cutlist_dir);
    let ln = ref 0 in
    let _ =
    try
        while true do
            let str = input_line chan in
            (incr ln);
            let str = ExtString.String.strip ~chars:"\r\t" str in
            if (String.length str) > 0 then (
                try
                    let new_centry =
                        parse_function_entry str ~loc:(Some (file_name, !ln)) ()
                    in
                    cut_entries := !cut_entries @ [new_centry]
                with
                | Not_found ->
                    Scommon.warning (file_name ^ ":" ^ (string_of_int !ln)
                                     ^ ": ignore an invalid cutlist entry '"
                                     ^ str
                                     ^ "'"))
        done;
    with End_of_file -> close_in chan in
    (Sys.chdir current_dir);
    !cut_entries)


let string_of_function_entry (entry : function_entry_t) : string = (
    match entry with
    | {
        fe_filelocation_opt = None;
        fe_funcname = funcname
      } -> funcname
    | {
        fe_filelocation_opt = Some fl;
        fe_funcname = funcname
      } -> (string_of_file_location fl) ^ "#" ^ funcname)

let is_cut_function (cutentry : function_entry_t) : bool = (
    List.mem cutentry !cut_functions)

let rec struct_get_field_in_pred (pred : lexpr) : lexpr list = (
    match pred.lexpr_node with
  | PLvar _
  | PLconstant _
  | PLresult
  | PLnull
  | PLsizeof _
  | PLfalse
  | PLtrue -> []
  | PLapp (_, lexprs) -> (
      List.flatten (
          List.map (fun a_lexpr ->
              struct_get_field_in_pred a_lexpr
          ) lexprs)
    )
  | PLunop (_, lexpr1)
  | PLdot (lexpr1, _)
  | PLarrow (lexpr1, _)
  | PLold (lexpr1)
  | PLcast (_, lexpr1)
  | PLsizeofE (lexpr1)
  | PLnot lexpr1
  | PLvalid lexpr1
  | PLuniqueloc lexpr1
  | PLuniqueblock (lexpr1, _) -> (
      struct_get_field_in_pred lexpr1;
    )
  | PLfield lexpr1 -> [lexpr1]
  | PLbinop (lexpr1, _, lexpr2)
  | PLarrget (lexpr1, lexpr2)
  | PLrel (lexpr1, _, lexpr2)
  | PLand (lexpr1, lexpr2)
  | PLor (lexpr1, lexpr2)
  | PLxor (lexpr1, lexpr2)
  | PLimplies (lexpr1, lexpr2)
  | PLiff (lexpr1, lexpr2) -> (
      (struct_get_field_in_pred lexpr1) @
      (struct_get_field_in_pred lexpr2)
    )
  | PLif (lexpr1, lexpr_op, lexpr2) -> (
      (struct_get_field_in_pred lexpr1) @
      (match lexpr_op with
       | None -> []
       | Some lexpr3 -> struct_get_field_in_pred lexpr3
      ) @
      (struct_get_field_in_pred lexpr2)
    )
)

let struct_get_field_in_func_conds func (func_conds : 'pred func_condition_t list)
: 'expr list = (
    List.flatten (
        List.map (fun (_, pred) ->
            func pred;
        ) func_conds)
)

let struct_get_field_in_behav func (behav : ('pred, 'locs) behavior_t)
: 'expr list = (
    let { b_name = _;
          b_assumes = preds;
          b_conditions = func_conds;
          b_assigns = _ } = behav in
    (struct_get_field_in_func_conds func func_conds) @
    (List.flatten (
         List.map (fun pred ->
             func pred
         ) preds
     ))
)

let struct_get_field_in_behavs func (behavs : ('pred, 'locs) behavior_t list)
: 'expr list = (
    List.flatten (
        List.map (fun behav ->
            struct_get_field_in_behav func behav
        ) behavs)
)

let struct_get_field_in_local_invs func (local_invs : (identifier_t * 'pred) list)
: 'expr list = (
    List.flatten (
        List.map (fun (_, pred) ->
            func pred
        ) local_invs)
)

let struct_get_field func (spec : ('pred, 'pred, 'loc) spec_t) : 'expr list = (
     match spec with
    | SPEC_FUN { fspec_conditions = func_conds;
                 fspec_behaviors = behavs;
                 fspec_assigns = _;
                 fspec_prohibits = prohibits;
                 fspec_local_invs = local_invs;
                 fspec_context = _;
                 fspec_is_handler = _
               } ->
        (struct_get_field_in_func_conds func func_conds) @
        (struct_get_field_in_behavs func behavs) @
        (struct_get_field_in_local_invs func local_invs)
     | _ -> []
 )
