(* parser/translate.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Translator to Formulas *)

(* TODO: (1) Return block (tail) needs predicates (for results). (2)
   Globals have no scope (file/global). *)

(* MEMO: (1) Handling Loops: Label which is a target from multiple
   gotoes records a state set, which is used to check a invaliant or
   to find a fixedpoint. *)

(* MEMO: Structure reference is converted to an UI function:
   ("sref%slot" x) where a slot name is not unique but is unique to
   each value/variable... (not sure, may be). *)

(* MEMO: Predicates used for predicate abstraction are quantifier
   free. *)

open Num
open Form

module StringSet = Form.StringSet
module FormSet = Form.FormSet
module StringInt2Set = Csyntax.StringInt2Set

let string_equal = Misc.string_equal
let debug = Cexpr.debug

let trTRACE_forms = ref false

type expr_t = Cexpr.expr_t
type stmt_t = Cexpr.stmt_t
type stmt_block_t = Csyntax.stmt_block_t

let pseudo_entry_label = Csyntax.pseudo_entry_label
let pseudo_return_label = Csyntax.pseudo_return_label
let pseudo_elsewhere_label = Csyntax.pseudo_elsewhere_label

(* Basic Block.  {bb_inputs} and {bb_outputs} are copied from
   {blk_inputs} and {blk_outputs} of {stmt_block_t}. *)

type form_block_t = {
  bb_label : string ;
(*  bb_memo : string ;*)
  bb_memo : stmt_t list ;
  bb_outgoing_labels : string list ;
  bb_incoming_labels : string list ;
  bb_inputs : StringInt2Set.t ;
  bb_outputs : StringInt2Set.t ;
  bb_body : bb_body_t
}

and bb_body_t =
    BBHEAD of bbhead_t
    | BBTAIL of bbtail_t
    | BBSTMT of bbstmt_t
    | BBCOND of bbcond_t
    | BBJUNC of bbjunc_t
    | BBCALL of bbcall_t
    | BBASSUME of bbassume_t
    | BBASSIGN of bbassign_t

and bbassume_t = {
  bbassume : form_t ;
}

and bbassign_t = {
  bbassign : Cexpr.expr_t list ;
}

(* Function Entry.  {bbhead_predicates} is a set of abstraction
   predicates on parameters and globals (thus, predicates on not
   versioned variables).  {bbhead_predmap} maps a predicate on
   parameters to a predicate on versioned variables used in the body.
   {bbhead_params} is a list of parameter names (eg, "v%2%0").
   {bbhead_paramsmap} maps a parameter to a versioned variable locally
   referenced (eg, "v%2%0" to "v%2%1").  Parameters are given with
   zero versions to avoid name conflicts in forms at calls.
   {bbhead_globals_inmap} maps an unversioned global to a locally
   referenced one (eg, "g%1%0" to "g%1%i"). *)

and bbhead_t = {
  bbhead_pre : (form_t * form_t * form_t) ;
  bbhead_predicates : form_t list ;
  bbhead_predmap : (form_t, form_t) Hashtbl.t ;
  bbhead_params : string list ;
  bbhead_paramsmap : (varname_t, varname_t) Hashtbl.t ;
  bbhead_globals_inmap : (varname_t, varname_t) Hashtbl.t
}

(* Function Return.  {bbtail_predicates} are prediates on results.  IT
   IS CURRENTLY ALWAYS EMPTY.  {bbtail_globals_outmap} maps locally
   referenced globals to unversioned ones (eg, from "g%1%i" to
   "g%1%0"). *)

and bbtail_t = {
  bbtail_predicates : form_t list ;
  bbtail_globals_outmap : (varname_t, varname_t) Hashtbl.t
}

(* Statements.  {bbstmt_form} is a formula representing statements. *)

and bbstmt_t = {
  bbstmt_form : form_t ;
  bbstmt_assert : form_t ;
  bbstmt_assume : form_t ;
  bbstmt_invariant : form_t
}

(* Junction.  States are checked with predicates on pre-phi variables
   and then they are mapped to predicates on post-phi variables.
   {bbjunc_phis} maps a label to an equation between post-phi and
   pre-phi variables.  {bbjunc_phifix} is a list of pairs of post-phi
   and pre-phi variables.  {bbjunc_predicates} is a set of abstraction
   predicates.  Predicates are on pre-phi variables.  {bbjunc_predmap}
   maps a predicate on pre-phi variables to a corresponding predicate
   on post-phi variables.  MEMO: 'pre-phi' variables are on the
   arguments of a phi-function, and 'post-phi' variables are on the
   result. *)

and bbjunc_t = {
  bbjunc_phis : (string, form_t) Hashtbl.t ;
  bbjunc_phifix : (term_t * term_t) list ;
  bbjunc_predicates : form_t list ;
  bbjunc_predmap : (form_t, form_t) Hashtbl.t
}

(* Conditional.  It consists of a form and two labels. *)

and bbcond_t = {
  bbcond_cond : form_t * string * string
}

(* Function Call.  {bbcall_fn} is a function name.  {bbcall_call} is a
   list of calling terms, including one at a function position.
   {bbcall_global_reversions} is a list of globals (of the same
   globals with different version numbers) which are equated when the
   globals are not in the modified list.  {bbcall_globals_inmap} maps
   an unversioned global to a locally referenced global at the point
   (eg, from "g%1%0" to "g%1%i").  {bbcall_globals_outmap} maps in
   reverse.  {bbcall_predicates} is a set of abstraction
   predicates. *)

and bbcall_t = {
  bbcall_fn : (string * int * int) ;
  bbcall_call : term_t list ;
  bbcall_global_reversions : (term_t * term_t) list ;
  bbcall_globals_inmap : (varname_t, varname_t) Hashtbl.t ;
  bbcall_globals_outmap : (varname_t, varname_t) Hashtbl.t ;
  bbcall_predicates : form_t list;
  bbcall_spec : Cexpr.cspec_t option;
}

and definition_t =
    FUNDEF of fundef_t
    | VARDEF of vardef_t
    | OTHERDEF of stmt_t

and fundef_t = {
  fundef_hash_name : string ;
  fundef_declaration : Cexpr.declaration_t ;
  fundef_bbset : (string, form_block_t) Hashtbl.t ;
  fundef_global_preds : form_t list
}

and vardef_t = {
  vardef_declaration : Cexpr.declaration_t ;
}

(* Parsed and translated definitions in file.  {ofile_defs} holds
   function/variable definitions of a file.  {ofile_global_predicates}
   is a list of predicates on globals, which is used to trace global
   states.  {ofile_hash_name} is a generated file name, which is used
   as a unique key for storing ofiles in an afile.  *)

and ofile_t = {
  ofile_hash_name : string ;
  ofile_file_name : string ;
  ofile_defs : (string, definition_t) Hashtbl.t ;
  ofile_global_predicates : form_t list
}

(* Set of translated files (in archives or executables).
   {afile_start} is a start symbol.  {afile_libs} holds a set of file
   definitions.  {afile_directory} maps an external name to a
   definition file name.  {afile_global_predicates} maps a global
   variable name to a list of its related abstraction predicates. *)

and afile_t = {
  afile_libs : (string, ofile_t) Hashtbl.t ;
  afile_directory : (string, string) Hashtbl.t ;
  afile_start : string ;
  afile_global_predicates : (string, form_t list) Hashtbl.t
}

let expr_list_to_string (l : Cexpr.expr_t list) : string = (
  ("[" ^ (String.concat ";" (List.map Cexpr.expr_to_string l)) ^ "]"))

let rec dump_form_blocks (def : definition_t) : unit = (
  match def with
    (FUNDEF d) ->
      let decl = d.fundef_declaration in
      let bbset = d.fundef_bbset in
      let globalpreds = d.fundef_global_preds in
      (debug ("FUNDEF=" ^ (Cexpr.decl_to_string decl))) ;
      (debug ("BLOCKS=" ^ (string_of_int (Hashtbl.length bbset)))) ;
      (debug ("GLOBALPREDS="
		       ^ (form_list_to_string globalpreds))) ;
      (dump_form_blocks_loop bbset [pseudo_entry_label] StringSet.empty)
  | (VARDEF d) ->
      let decl = d.vardef_declaration in
      (debug ("VARDEF=" ^ (Cexpr.decl_to_string decl)))
  | (OTHERDEF s) ->
      (debug ("OTHERDEF=" ^ (Cexpr.stmt_to_string s))))

and dump_form_blocks_loop (bbset : (string, form_block_t) Hashtbl.t) (labels : string list) (visited : StringSet.t) = (
  match labels with
    [] ->
      ()
  | (label :: tl) ->
      if ((Hashtbl.length bbset) == 0) then
	()
      else if (string_equal label pseudo_elsewhere_label) then
	(dump_form_blocks_loop bbset tl visited)
      else if (StringSet.mem label visited) then
	(dump_form_blocks_loop bbset tl visited)
      else if (not (Hashtbl.mem bbset label)) then
	(assert false)
      else
	(begin
	  let b = (Hashtbl.find bbset label) in
	  (dump_form_block b) ;
	  let newvisited = (StringSet.add label visited) in
	  (dump_form_blocks_loop bbset (b.bb_outgoing_labels @ tl) newvisited)
	end))

and dump_form_block (b : form_block_t) = (
  let variable_set_to_string (s : StringInt2Set.t) : string = (
    (String.concat ";"
	(List.map Cexpr.vartag_to_string
	    (Csyntax.variable_set_to_list s)))) in
  match b with
    {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBHEAD {
	bbhead_pre = pre ;
	bbhead_predicates = abstpreds ;
	bbhead_predmap = predmap ;
	bbhead_params = params ;
	bbhead_paramsmap = paramsmap ;
	bbhead_globals_inmap = gvarmap
      }
    } ->
      (begin
	(debug ("BB:HEAD{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	(debug ("|abstpreds="
			 ^ (form_list_to_string abstpreds))) ;
	(*(debug ("|globals=" ^ (String.concat ";" gvars))) ;*)
	(Hashtbl.iter
	    (fun v0 v1 ->
	      (debug ("|gvarmap=" ^ v0 ^ "->" ^ v1)))
	    gvarmap) ;
	(Hashtbl.iter
	    (fun v0 v1 ->
	      (debug ("|paramsmap=" ^ v0 ^ "->" ^ v1)))
	    paramsmap) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBTAIL {
	bbtail_predicates = abstpreds ;
	bbtail_globals_outmap = gvarmap
      }
    } ->
      (begin
	(debug ("BB:TAIL{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	(debug ("|abstpreds="
			 ^ (form_list_to_string abstpreds))) ;
	(Hashtbl.iter
	    (fun v0 v1 ->
	      (debug ("|gvarmap=" ^ v0 ^ "->" ^ v1)))
	    gvarmap) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBSTMT {
	bbstmt_form = form ;
	bbstmt_assert = assertform ;
	bbstmt_assume = assumeform ;
	bbstmt_invariant = invariantform
      }
    } ->
      (begin
	(debug ("BB:STMT{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	if (form != TT) then
	  (debug ("|form=" ^ (form_to_string form))) ;
	if (assertform != TT) then
	  (debug ("|assert=" ^ (form_to_string assertform))) ;
	if (assumeform != TT) then
	  (debug ("|assume=" ^ (form_to_string assumeform))) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBASSUME {
         bbassume = form ;
      }
    } ->
      (begin
	(debug ("BB:ASSUME{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	if (form != TT) then
	  (debug ("|form=" ^ (form_to_string form))) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBASSIGN {
         bbassign = exs ;
      }
    } ->
    (begin
	(debug ("BB:ASSIGN{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
    (List.iter (
        fun ex -> 
            (debug (Cexpr.expr_to_string_body 0 [] ex))
    )exs);
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
     end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBCOND {
	bbcond_cond = (cond, thengo, elsego)
      }
    } ->
      (begin
	(begin
	  match outgoings with
	    [then_out_label ; else_out_label] ->
	      (debug ("BB:COND{" ^ "(label=" ^ label
			       ^ "|then=" ^ then_out_label
			       ^ "|else=" ^ else_out_label
			       ^ ("|incoming=" ^ (String.concat ";" incomings))
			       ^ ")")) ;
(*	      (debug ("|memo=" ^ memo)) ;*)
	  | _ -> (assert false)
	end) ;
	(debug ("|cond=" ^ "IF(" ^ (form_to_string cond)
			 ^ ";" ^ thengo ^ ";" ^ elsego^ ")")) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBJUNC {
	bbjunc_phis = phis ;
	bbjunc_phifix = phifix ;
	bbjunc_predicates = abstpreds ;
	bbjunc_predmap = predmap
      }
    } ->
      (begin
	(debug ("BB:JUNC{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	(Hashtbl.iter
	    (fun label f ->
	      (debug ("|phi=" ^ label ^ ":" ^ (form_to_string f))))
	    phis) ;
	(debug ("|phifix="
			   ^ (String.concat ";"
				 (List.map
				     (fun (v0, v1) ->
				       ("(" ^ (term_to_string v0) ^ "/"
					 ^ (term_to_string v1) ^ ")"))
				     phifix)))) ;
	(debug ("|abstpreds="
			 ^ (form_list_to_string abstpreds))) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end)
  | {
      bb_label = label ;
      bb_memo = memo ;
      bb_outgoing_labels = outgoings ;
      bb_incoming_labels = incomings ;
      bb_inputs = ivars ;
      bb_outputs = ovars ;
      bb_body = BBCALL {
	bbcall_call = call ;
	bbcall_global_reversions = callfix ;
	bbcall_globals_inmap = gvarmap ;
	bbcall_globals_outmap = _ ;
	bbcall_predicates = abstpreds
      }
    } ->
      (begin
	assert (call != []) ;
	(debug ("BB:CALL{" ^ "(label=" ^ label
			 ^ ("|outgoing=" ^ (String.concat ";" outgoings))
			 ^ ("|incoming=" ^ (String.concat ";" incomings))
			 ^ ")")) ;
(*	(debug ("|memo=" ^ memo)) ;*)
	(debug ("|call=" ^ "("
			 ^ (String.concat "," (List.map term_to_string call))
			 ^ ")")) ;
	(debug ("|callfix="
			   ^ (String.concat ";"
				 (List.map
				     (fun (v0, v1) ->
				       ("(" ^ (term_to_string v0) ^ "/"
					 ^ (term_to_string v1) ^ ")"))
				     callfix)))) ;
	(Hashtbl.iter
	    (fun v0 v1 ->
	      (debug ("|gvarmap=" ^ v0 ^ "->" ^ v1)))
	    gvarmap) ;
	(debug ("|abstpreds="
			 ^ (form_list_to_string abstpreds))) ;
	(debug ("|ivars=" ^ (variable_set_to_string ivars))) ;
	(debug ("|ovars=" ^ (variable_set_to_string ovars))) ;
	(debug ("}"))
      end))

(* ================================================================ *)

(* PICKUP OF ABSTRACTION PREDICATES *)

(* Scans blocks for abstraction predicates, and stores the found
   predicates under a label in mapping {predsets}.  Predicates on
   globals are set under the label {pseudo_elsewhere_label}.  NOTE:
   Predicates are taken from annotations and conditionals.  Predicates
   from invariant annotations (eg, "$invariant" and "$check") ignore
   versions of variables.  MEMO: It is called before translation to
   formulas for ease of handling of versions. *)

let rec scan_abstraction_predicates (fn : string) (bbset : (string, stmt_block_t) Hashtbl.t) (predsets : (string, expr_t list) Hashtbl.t) : unit = (
  let predvars : (expr_t * StringInt2Set.t) list =
    (collect_predicates bbset) in
  if (!trTRACE_forms) then
    (debug (";; predicates.in(" ^ fn ^ ")="
		     ^ (expr_list_to_string (List.map fst predvars)))) ;
  (scan_abstraction_predicates_on_globals predvars predsets) ;
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	(scan_abstraction_predicates_to_block b predvars predsets))
      (Misc.hashtbl_keys bbset)))

(* Collects predicates on globals, and stores them under the label
   {pseudo_elsewhere_label}. *)

and scan_abstraction_predicates_on_globals (predvars : (expr_t * StringInt2Set.t) list) (predsets : (string, expr_t list) Hashtbl.t) : unit = (
  let preds =
    (Misc.list_filtering_map
	(fun ((e : expr_t), (ss : StringInt2Set.t)) ->
	  if (StringInt2Set.for_all Cexpr.varid_global_p ss) then
	    let varmap : (Cexpr.vartag_t, Cexpr.vartag_t) Hashtbl.t =
	      (make_variable_version_map false ss) in
	    (Some (Cexpr.substitute_variables_in_expr varmap e))
	  else
	    None)
	predvars) in
  assert (not (Hashtbl.mem predsets pseudo_elsewhere_label)) ;
  (Hashtbl.replace predsets pseudo_elsewhere_label preds))

and scan_abstraction_predicates_to_block (b : stmt_block_t) (predvars : (expr_t * StringInt2Set.t) list) (predsets : (string, expr_t list) Hashtbl.t) : unit = (
  match b with
    {
      Csyntax.blk_type = (Csyntax.BBHEAD|Csyntax.BBJUNC|Csyntax.BBCALL) ;
      Csyntax.blk_label = label ;
      Csyntax.blk_memo = _ ;
      Csyntax.blk_outgoing_labels = _ ;
      Csyntax.blk_incoming_labels = _ ;
      Csyntax.blk_stmts = _ ;
      Csyntax.blk_phis = phis ;
      Csyntax.blk_phifix = phifix ;
      Csyntax.blk_inputs = ivars ;
      Csyntax.blk_outputs = ovars
    } ->
      (* Make a set of unversioned variables. *)
      let varmap : ((string * int * int), (string * int * int)) Hashtbl.t =
	(make_variable_version_map true ovars) in
      let uvars : StringInt2Set.t =
	(Hashtbl.fold
	    (fun (id, level, sequence) _ ss ->
	      assert (sequence == 0) ;
	      (StringInt2Set.add (id, level, 0) ss))
	    varmap (StringInt2Set.empty)) in
      (* Filter predicates by liveness of variables. *)
      let preds =
	(List.flatten
	    (List.map
		(fun ((e : expr_t), (ss : StringInt2Set.t)) ->
		  if (StringInt2Set.subset ss ovars) then
		    [e]
		  else if (StringInt2Set.subset ss uvars) then
		    [(Cexpr.substitute_variables_in_expr varmap e)]
		  else
		    [])
		predvars)) in
      assert (not (Hashtbl.mem predsets label)) ;
      (Hashtbl.replace predsets label preds)
  | _ ->
      ())

(* Makes a mapping from an unversioned variable to a versioned
   variable (or vice versa).  It is used to check subset relation
   between sets of variables.  NOTE: Unversioned variables (non-SSA)
   can match any versioned variables. *)

and make_variable_version_map (unversioned_to_versioned : bool) (s : StringInt2Set.t) : ((string * int * int), (string * int * int)) Hashtbl.t = (
  let tbl : ((string * int * int), (string * int * int)) Hashtbl.t =
    (Hashtbl.create 10) in
  (StringInt2Set.iter
      (fun (id, level, sequence) ->
	(* Never duplicate unversioned variables *)
	if (unversioned_to_versioned) then
	  (begin
	    assert (not (Hashtbl.mem tbl (id, level, 0))) ;
	    (Hashtbl.replace tbl (id, level, 0) (id, level, sequence))
	  end)
	else
	  (begin
	    assert (not (Hashtbl.mem tbl (id, level, sequence))) ;
	    (Hashtbl.replace tbl (id, level, sequence) (id, level, 0))
	  end))
      s) ;
  tbl)

(* Collects predicates appearing in all expressions, and returns pairs
   of an expression and its included variables.  Note that invariants
   have expressions with unversioned variables. *)

and collect_predicates (bbset : (string, stmt_block_t) Hashtbl.t) : (expr_t * StringInt2Set.t) list = (
  let predlist : expr_t list =
    (List.flatten
	(List.map
	    (fun label ->
	      let b : stmt_block_t = (Hashtbl.find bbset label) in
	      match b.Csyntax.blk_type with
		Csyntax.BBHEAD ->
		  []
	      | Csyntax.BBTAIL
	      | Csyntax.BBASME
	      | Csyntax.BBSTMT ->
		  let l = b.Csyntax.blk_stmts in
		  (List.flatten
		      (List.map extract_predicates_in_stmt l))
	      | Csyntax.BBCOND ->
		  let l = b.Csyntax.blk_stmts in
		  (List.flatten
		      (List.map extract_predicates_in_stmt l))
	      | Csyntax.BBJUNC ->
		  []
	      | Csyntax.BBASSIGN
	      | Csyntax.BBCALL ->
		  [])
	    (Misc.hashtbl_keys bbset))) in
  (List.map
      (fun e ->
	(e, (list_free_variables_in_expr e (StringInt2Set.empty))))
      predlist))

and extract_predicates_in_stmt (s : stmt_t) : expr_t list = (
  match s with
    (Cexpr.DEFUN _) | (Cexpr.DEFVAR _) | (Cexpr.SPEC _) -> (assert false)
  | (Cexpr.STMT (Cexpr.CALL (tn, e0, ee, _)))
  | (Cexpr.ASGN (_, (Cexpr.CALL (tn, e0, ee, _)))) -> (assert false)
      (*
      ((extract_predicates_in_expr e0)
	@ (List.flatten (List.map extract_predicates_in_expr ee)))
      *)
  | (Cexpr.ASSUME e0) -> (extract_predicates_in_expr e0)
  | (Cexpr.RESET es) -> (List.flatten (List.map extract_predicates_in_expr es))
  | (Cexpr.STMT e0) ->
      (extract_predicates_in_expr e0)
  | (Cexpr.ASGN (_, e1)) ->
      (extract_predicates_in_expr e1)
  | (Cexpr.CALLASGN ((Cexpr.CALL (tn, e0, ee, _)), assignments, _)) ->
      ((extract_predicates_in_expr e0)
	@ (List.flatten (List.map extract_predicates_in_expr ee))
	@ (List.flatten (List.map extract_predicates_in_stmt assignments)))
  | (Cexpr.CALLASGN _) -> (assert false)
  | (Cexpr.IF (e0, [(Cexpr.GOTO _)], [(Cexpr.GOTO _)])) ->
      (extract_predicates_in_expr e0)
  | Cexpr.IF _ | Cexpr.BEGIN _ | Cexpr.SWITCH _ -> (assert false)
  | Cexpr.DOWHILE _ | Cexpr.FOR _ -> (assert false)
  | Cexpr.BREAK | Cexpr.CONTINUE -> (assert false)
  | Cexpr.LABEL _ | Cexpr.GOTO _ | Cexpr.RETURN _ -> (assert false)
  | (Cexpr.ANNOTATION (s, e0)) ->
      (extract_predicates_in_expr e0)
  | (Cexpr.ASM _) ->
      []
  | (Cexpr.NOTE _) ->
      [])

(* NOTE: IT SHOULD WORK COHERENTLY AS {translate_to_form}. It is
   because both generates predicate forms.  It works on expressions
   instead of forms, because it needs variables as triplets for
   handling version matching. *)

and extract_predicates_in_expr (e : expr_t) : expr_t list = (
  match e with
    Cexpr.NIL -> (assert false)
  | (Cexpr.PSEUDO_NONDET_VALUE (tn, _)) ->
      []
  | (Cexpr.OPER (tn, op, e0, e1, e2)) ->
      (begin
	match op with
	  Cexpr.SET -> (assert false)
	| Cexpr.INCR_PRE | Cexpr.DECR_PRE -> (assert false)
	| Cexpr.INCR_POST | Cexpr.DECR_POST -> (assert false)
	| Cexpr.ADD | Cexpr.SUB | Cexpr.MUL | Cexpr.DIV | Cexpr.MOD
	| Cexpr.LSH | Cexpr.RSH | Cexpr.AND | Cexpr.IOR | Cexpr.XOR ->
	    ((extract_predicates_in_expr e0)
	      @ (extract_predicates_in_expr e1))
	| (Cexpr.CAST _) ->
	    (extract_predicates_in_expr e0)
	| Cexpr.NEG | Cexpr.BITNOT ->
	    (extract_predicates_in_expr e0)
	| Cexpr.LOGNOT ->
	    (extract_predicates_in_expr e0)
	| Cexpr.EQ | Cexpr.NE | Cexpr.GE | Cexpr.GT | Cexpr.LE | Cexpr.LT ->
	    (begin
	      (* Some normalization to relationals in SMT. *)
	      if (false) then
		[e]
	      else
		(begin
		  (*AHOAHO*) (* segv occurs *)
		  match op with
		    Cexpr.EQ -> [e]
		  | Cexpr.NE -> [(Cexpr.OPER (tn, Cexpr.EQ, e0, e1, e2))]
		  | Cexpr.GE -> [(Cexpr.OPER (tn, Cexpr.GT, e1, e0, e2))]
		  | Cexpr.GT -> [(Cexpr.OPER (tn, Cexpr.GT, e0, e1, e2))]
		  | Cexpr.LE -> [(Cexpr.OPER (tn, Cexpr.GT, e0, e1, e2))]
		  | Cexpr.LT -> [(Cexpr.OPER (tn, Cexpr.GT, e1, e0, e2))]
		  | _ -> (assert false)
		end)
	    end)
	| Cexpr.COND ->
	    ((extract_predicates_in_expr e0)
	      @ (extract_predicates_in_expr e1)
	      @ (extract_predicates_in_expr e2))
	| Cexpr.OROR | Cexpr.ANDAND
	| Cexpr.COMMA
	| Cexpr.IMP | Cexpr.IFF ->
	    ((extract_predicates_in_expr e0)
	      @ (extract_predicates_in_expr e1))
	| Cexpr.PSEUDO_AMP | Cexpr.PSEUDO_STR -> (assert false)
	| Cexpr.PSEUDO_PLS | Cexpr.PSEUDO_MNS -> (assert false)
	| Cexpr.PSEUDO_TLD | Cexpr.PSEUDO_BAN -> (assert false)
      end)
  | (Cexpr.SETF _) -> (assert false)
  | (Cexpr.CALL (tn, e0, ee, _)) -> (assert false)
  | (Cexpr.VREF (tn, id, level, sequence)) ->
      (*(Cexpr.OPER (tn, Cexpr.NE, e, Cexpr.constant_zero, NIL))*)
      []
  | (Cexpr.PREF (tn, _, e0, _)) ->
      []
  | (Cexpr.SREF (tn, b, e0, s)) ->
      []
  | (Cexpr.ADDROF (tn, e0)) ->
      []
  | (Cexpr.SIZEOF (tn, t0)) ->
      []
  | (Cexpr.PROGN (tn, l)) -> (assert false)
  | (Cexpr.CONST_I _) ->
      []
  | (Cexpr.CONST_F _) ->
      []
  | (Cexpr.STRING _) ->
      []
  | (Cexpr.CLITERAL (tn, l)) -> (assert false)
  | (Cexpr.DSGN_AREF _) -> (assert false)
  | (Cexpr.ASMSEQUENCE _) -> (assert false)
  | (Cexpr.PHI _) -> (assert false)
  | (Cexpr.PSEUDO_MEMORY _) ->
      []
  | (Cexpr.INIT_FILL _) ->
      []
  | (Cexpr.PRAGMA s) -> (assert false)
  | (Cexpr.QUANTIFIER (s, vs, e0)) -> (assert false)
  | (Cexpr.PSEUDO_PROLOGUE) ->
      []
  | (Cexpr.PSEUDO_EPILOGUE) ->
      []
  | (Cexpr.PSEUDO_ACSL_RESULT _)
  | (Cexpr.PSEUDO_ACSL_OLD _)
  | (Cexpr.ALL_FIELD _) -> assert false
  | (Cexpr.PSEUDO_ACSL_UNIQUE _)
  | (Cexpr.PSEUDO_ACSL_UNIQUEBLOCK _) ->
      [])

and list_free_variables_in_expr (e : expr_t) (vs : StringInt2Set.t) : StringInt2Set.t = (
  match e with
    Cexpr.NIL -> (assert false)
  | (Cexpr.PSEUDO_NONDET_VALUE (tn, _)) ->
      vs
  | (Cexpr.OPER (tn, op, e0, e1, e2)) ->
      (begin
	match op with
	  Cexpr.SET -> (assert false)
	| Cexpr.INCR_PRE | Cexpr.DECR_PRE -> (assert false)
	| Cexpr.INCR_POST | Cexpr.DECR_POST -> (assert false)
	| Cexpr.ADD | Cexpr.SUB | Cexpr.MUL | Cexpr.DIV | Cexpr.MOD
	| Cexpr.LSH | Cexpr.RSH | Cexpr.AND | Cexpr.IOR | Cexpr.XOR ->
	    (list_free_variables_in_expr e0
		(list_free_variables_in_expr e1 vs))
	| (Cexpr.CAST _) ->
	    (list_free_variables_in_expr e0 vs)
	| Cexpr.NEG | Cexpr.BITNOT ->
	    (list_free_variables_in_expr e0 vs)
	| Cexpr.LOGNOT ->
	    (list_free_variables_in_expr e0 vs)
	| Cexpr.EQ | Cexpr.NE | Cexpr.GE | Cexpr.GT | Cexpr.LE | Cexpr.LT ->
	    (list_free_variables_in_expr e1
		(list_free_variables_in_expr e0 vs))
	| Cexpr.COND ->
	    (list_free_variables_in_expr e2
		(list_free_variables_in_expr e1
		    (list_free_variables_in_expr e0 vs)))
	| Cexpr.OROR | Cexpr.ANDAND
	| Cexpr.COMMA
	| Cexpr.IMP | Cexpr.IFF ->
	    (list_free_variables_in_expr e1
		(list_free_variables_in_expr e0 vs))
	| Cexpr.PSEUDO_AMP | Cexpr.PSEUDO_STR -> (assert false)
	| Cexpr.PSEUDO_PLS | Cexpr.PSEUDO_MNS -> (assert false)
	| Cexpr.PSEUDO_TLD | Cexpr.PSEUDO_BAN -> (assert false)
      end)
  | (Cexpr.SETF _) -> (assert false)
  | (Cexpr.CALL (tn, e0, ee, _)) -> (assert false)
  | (Cexpr.VREF (tn, id, level, sequence)) ->
      (StringInt2Set.add (id, level, sequence) vs)
  | (Cexpr.PREF (tn, _, e0, e1)) ->
      (list_free_variables_in_expr e0
          (list_free_variables_in_expr e1 vs))
  | (Cexpr.SREF (tn, _, e0, s)) ->
      (list_free_variables_in_expr e0 vs)
  | (Cexpr.ADDROF (tn, e0)) ->
      (list_free_variables_in_expr e0 vs)
  | (Cexpr.SIZEOF (tn, t0)) ->
      vs
  | (Cexpr.PROGN (tn, l)) -> (assert false)
  | (Cexpr.CONST_I _) ->
      vs
  | (Cexpr.CONST_F _) ->
      vs
  | (Cexpr.STRING _) ->
      vs
  | (Cexpr.CLITERAL (tn, l)) -> (assert false)
  | (Cexpr.DSGN_AREF _) -> (assert false)
  | (Cexpr.ASMSEQUENCE _) -> (assert false)
  | (Cexpr.PHI _) -> (assert false)
  | (Cexpr.PSEUDO_MEMORY (tn, t0)) ->
      vs
  | (Cexpr.INIT_FILL _) ->
      vs
  | (Cexpr.PRAGMA s) -> (assert false)
  | (Cexpr.QUANTIFIER (s, vs, e0)) -> (assert false)
  | (Cexpr.PSEUDO_PROLOGUE) ->
      vs
  | (Cexpr.PSEUDO_EPILOGUE) ->
      vs
  | (Cexpr.PSEUDO_ACSL_RESULT (tn, e0)) ->
      (list_free_variables_in_expr e0 vs)
  | (Cexpr.PSEUDO_ACSL_OLD (tn, e0, e1)) ->
      (list_free_variables_in_expr e0
                                   (list_free_variables_in_expr e1 vs))
  | (Cexpr.ALL_FIELD _) -> assert false
  | (Cexpr.PSEUDO_ACSL_UNIQUE (tn, e0)) ->
      (list_free_variables_in_expr e0 vs)
  | (Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1)) ->
      (list_free_variables_in_expr e0
                                   (list_free_variables_in_expr e1 vs)))

(*GOMI*)(*
and replace_variables_in_expr (varmap : ((string * int * int), (string * int * int)) Hashtbl.t) (e : expr_t) : expr_t = (
  match e with
    Cexpr.NIL -> (assert false)
  | (Cexpr.OPER (tn, op, e0, e1, e2)) ->
      (begin
	match op with
	  Cexpr.SET -> (assert false)
	| Cexpr.INCR_PRE | Cexpr.DECR_PRE -> (assert false)
	| Cexpr.INCR_POST | Cexpr.DECR_POST -> (assert false)
	| Cexpr.ADD | Cexpr.SUB | Cexpr.MUL | Cexpr.DIV | Cexpr.MOD
	| Cexpr.LSH | Cexpr.RSH | Cexpr.AND | Cexpr.IOR | Cexpr.XOR ->
	    assert (e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    let x1 = (replace_variables_in_expr varmap e1) in
	    (Cexpr.OPER (tn, op, x0, x1, Cexpr.NIL))
	| (Cexpr.CAST _) ->
	    assert (e1 == Cexpr.NIL && e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    (Cexpr.OPER (tn, op, x0, Cexpr.NIL, Cexpr.NIL))
	| Cexpr.NEG | Cexpr.BITNOT ->
	    assert (e1 == Cexpr.NIL && e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    (Cexpr.OPER (tn, op, x0, Cexpr.NIL, Cexpr.NIL))
	| Cexpr.LOGNOT ->
	    assert (e1 == Cexpr.NIL && e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    (Cexpr.OPER (tn, op, x0, Cexpr.NIL, Cexpr.NIL))
	| Cexpr.EQ | Cexpr.NE | Cexpr.GE | Cexpr.GT | Cexpr.LE | Cexpr.LT ->
	    assert (e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    let x1 = (replace_variables_in_expr varmap e1) in
	    (Cexpr.OPER (tn, op, x0, x1, Cexpr.NIL))
	| Cexpr.COND ->
	    let x0 = (replace_variables_in_expr varmap e0) in
	    let x1 = (replace_variables_in_expr varmap e1) in
	    let x2 = (replace_variables_in_expr varmap e2) in
	    (Cexpr.OPER (tn, op, x0, x1, x2))
	| Cexpr.OROR | Cexpr.ANDAND
	| Cexpr.COMMA
	| Cexpr.IMP | Cexpr.IFF ->
	    assert (e2 == Cexpr.NIL) ;
	    let x0 = (replace_variables_in_expr varmap e0) in
	    let x1 = (replace_variables_in_expr varmap e1) in
	    (Cexpr.OPER (tn, op, x0, x1, Cexpr.NIL))
	| Cexpr.PSEUDO_AMP | Cexpr.PSEUDO_STR -> (assert false)
	| Cexpr.PSEUDO_PLS | Cexpr.PSEUDO_MNS -> (assert false)
	| Cexpr.PSEUDO_TLD | Cexpr.PSEUDO_BAN -> (assert false)
      end)
  | (Cexpr.SETF _) -> (assert false)
  | (Cexpr.CALL (tn, e0, ee)) -> (assert false)
  | (Cexpr.VREF (tn, id, level, sequence)) ->
      let (i, l, s) = (Hashtbl.find varmap (id, level, sequence)) in
      (Cexpr.VREF (tn, i, l, s))
  | (Cexpr.PREF (tn, _, e0)) ->
      let x0 = (replace_variables_in_expr varmap e0) in
      (Cexpr.PREF (tn, x0))
  | (Cexpr.SREF (tn, _, e0, s)) ->
      let x0 = (replace_variables_in_expr varmap e0) in
      (Cexpr.SREF (tn, b, x0, s))
  | (Cexpr.ADDROF (tn, e0)) ->
      let x0 = (replace_variables_in_expr varmap e0) in
      (Cexpr.ADDROF (tn, x0))
  | (Cexpr.SIZEOF (tn, t0)) ->
      e
  | (Cexpr.CONST_I _) ->
      e
  | (Cexpr.CONST_F _) ->
      e
  | (Cexpr.STRING _) ->
      e
  | (Cexpr.DSGN_AREF _) -> (assert false)
  | (Cexpr.ASMSEQUENCE _) -> (assert false)
  | (Cexpr.PHI _) -> (assert false)
  | (Cexpr.PSEUDO_MEMORY (tn, t0)) ->
      e
  | (Cexpr.PRAGMA s) -> (assert false)
  | (Cexpr.QUANTIFIER (s, vs, e0)) -> (assert false)
  | (Cexpr.PSEUDO_PROLOGUE) ->
      e
  | (Cexpr.PSEUDO_EPILOGUE) ->
      e)
*)

(* ================================================================ *)

(* TRANSLATE EXPRS TO FORMS *)

let rec translate_blocks (bbset : (string, form_block_t) Hashtbl.t) (oldbbset : (string, stmt_block_t) Hashtbl.t) (decl : Cexpr.declaration_t) (predsets : (string, expr_t list) Hashtbl.t) : unit = (
  (Hashtbl.iter
      (fun label b ->
	assert (string_equal label b.Csyntax.blk_label) ;
	let newblk = (translate_stmts b decl predsets) in
	assert (not (Hashtbl.mem bbset label)) ;
	(Hashtbl.replace bbset label newblk))
      oldbbset) ;
  ())

and translate_stmts (oldb : stmt_block_t) (decl : Cexpr.declaration_t) (predsets : (string, expr_t list) Hashtbl.t) : form_block_t = (
  let fn : string = (fst (decl.Cexpr.decl_id)) in
  match oldb.Csyntax.blk_type with
    Csyntax.BBHEAD ->
      (begin
	let label = oldb.Csyntax.blk_label in
	let paramlist : string list = decl.Cexpr.decl_params in
	(* Map temporary versioned parameters/globals to true ones. *)
	let varmap : (varname_t, varname_t) Hashtbl.t = (Hashtbl.create 10) in
	(List.iter
	    (fun v ->
	      if (not (Cexpr.ellipsis_parameter_p v)) then
		(Hashtbl.replace varmap
		    (Cexpr.vartag_to_string (v, 2, 1))
		    (Cexpr.vartag_to_string (v, 2, 0))))
	    paramlist) ;
	(List.iter
	    (fun (id, level, sequence) ->
	      if (level == Cexpr.global_scope_level) then
		(Hashtbl.replace varmap
		    (Cexpr.vartag_to_string (id, level, sequence))
		    (Cexpr.vartag_to_string (id, level, 0))))
	    (Csyntax.variable_set_to_list oldb.Csyntax.blk_inputs)) ;
	let predexprs : expr_t list =
	  (Misc.hashtbl_find_safe [] predsets label) in
	let predforms : form_t list =
	  (uniqify_form_list (List.map translate_to_form predexprs)) in
	let predmap = (Hashtbl.create 10) in
	let predsmapped =
	  (List.map
	      (substitute_variables_in_predicate varmap predmap)
	      predforms) in
	let params : varname_t list =
	  (List.map
	      (fun v ->
		if (not (Cexpr.ellipsis_parameter_p v)) then
		  (Cexpr.vartag_to_string (v, 2, 0))
		else
		  Cexpr.ellipsis_string)
	      paramlist) in
	let paramsmap : (varname_t, varname_t) Hashtbl.t =
	  (Hashtbl.create 10) in
	(List.iter
	    (fun (id, level, sequence) ->
	      if (level == Cexpr.parameter_scope_level) then
		(Hashtbl.replace paramsmap
		    (Cexpr.vartag_to_string (id, level, 0))
		    (Cexpr.vartag_to_string (id, level, sequence))))
	    (Csyntax.variable_set_to_list oldb.Csyntax.blk_inputs)) ;
	let vinmap : (varname_t, varname_t) Hashtbl.t = (Hashtbl.create 10) in
	(List.iter
	    (fun (id, level, sequence) ->
	      if (level == Cexpr.global_scope_level) then
		(Hashtbl.replace vinmap
		    (Cexpr.vartag_to_string (id, level, 0))
		    (Cexpr.vartag_to_string (id, level, sequence))))
	    (Csyntax.variable_set_to_list oldb.Csyntax.blk_inputs)) ;
	if (!trTRACE_forms) then
	  (debug (";; predicates.at(" ^ fn ^ "/" ^ label ^ ")="
			   ^ (form_list_to_string predforms))) ;
	{
	  bb_label = oldb.Csyntax.blk_label ;
	  bb_memo = oldb.Csyntax.blk_memo ;
	  bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
	  bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
	  bb_inputs = oldb.Csyntax.blk_inputs ;
	  bb_outputs = oldb.Csyntax.blk_outputs ;
	  bb_body = BBHEAD {
	    bbhead_pre = (TT, TT, TT) ;
	    bbhead_predicates = predsmapped ;
	    bbhead_predmap = predmap ;
	    bbhead_params = params ;
	    bbhead_paramsmap = paramsmap ;
	    bbhead_globals_inmap = vinmap
	  }
	}
      end)
  | Csyntax.BBTAIL ->
      (begin
	assert (Cexpr.epilogue_statement_p oldb.Csyntax.blk_stmts) ;
	let label = oldb.Csyntax.blk_label in
	let voutmap : (varname_t, varname_t) Hashtbl.t = (Hashtbl.create 10) in
	(List.iter
	    (fun (id, level, sequence) ->
	      (Hashtbl.replace voutmap
		  (Cexpr.vartag_to_string (id, level, sequence))
		  (Cexpr.vartag_to_string (id, level, 0))))
	    (Csyntax.variable_set_to_list oldb.Csyntax.blk_outputs)) ;
	(*AHO*) (* CURRENTLY ALWAYS EMPTY. *)
	let predexprs : expr_t list =
	  (Misc.hashtbl_find_safe [] predsets label) in
	let predforms : form_t list =
	  (uniqify_form_list (List.map translate_to_form predexprs)) in
	if (!trTRACE_forms) then
	  (debug (";; predicates.at(" ^ fn ^ "/" ^ label ^ ")="
			   ^ (form_list_to_string predforms))) ;
	{
	  bb_label = oldb.Csyntax.blk_label ;
	  bb_memo = oldb.Csyntax.blk_memo ;
	  bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
	  bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
	  bb_inputs = oldb.Csyntax.blk_inputs ;
	  bb_outputs = oldb.Csyntax.blk_outputs ;
	  bb_body = BBTAIL {
	    bbtail_predicates = predforms ;
	    bbtail_globals_outmap = voutmap
	  }
	}
      end)
  | Csyntax.BBSTMT ->
      (begin
	match oldb.Csyntax.blk_stmts with
          []
	| ((Cexpr.STMT _) :: _)
	| ((Cexpr.ASGN (_, Cexpr.NIL)) :: _)
	| ((Cexpr.ASGN (_, _)) :: _)
	| ((Cexpr.ANNOTATION _) :: _) ->
	    (begin
	      let l = oldb.Csyntax.blk_stmts in
	      let f0 = (pseudo_simplify_form
			   (AND (List.map translate_stmt_simple l))) in
	      {
		bb_label = oldb.Csyntax.blk_label ;
		bb_memo = oldb.Csyntax.blk_memo ;
		bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
		bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
		bb_inputs = oldb.Csyntax.blk_inputs ;
		bb_outputs = oldb.Csyntax.blk_outputs ;
		bb_body = BBSTMT {
		  bbstmt_form = f0 ;
		  bbstmt_assert = (translate_annotation "$assert" l) ;
		  bbstmt_assume = (translate_annotation "$assume" l) ;
		  bbstmt_invariant = (translate_annotation "$invariant" l)
		}
	      }
	    end)
    (*| ((Cexpr.ASSUME _ ) :: _) -> 
        (begin
	      let l = oldb.Csyntax.blk_stmts in
	      let f0 = ((AND (List.map translate_stmt_simple l))) in
	      {
              bb_label = oldb.Csyntax.blk_label ;
              bb_memo = oldb.Csyntax.blk_memo ;
              bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
              bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
              bb_inputs = oldb.Csyntax.blk_inputs ;
              bb_outputs = oldb.Csyntax.blk_outputs ;
              bb_body = BBSTMT {
                  bbstmt_form = f0 ;
                  bbstmt_assert = (translate_annotation "$assert" l) ;
                  bbstmt_assume = (translate_annotation "$assume" l) ;
                  bbstmt_invariant = (translate_annotation "$invariant" l)
              }
	      }
        end)*)
	| _ -> (assert false)
      end)
  (* BinhLT added - begin *)
  | Csyntax.BBASSIGN ->
      (begin
	    match oldb.Csyntax.blk_stmts with
        |((Cexpr.CALLASGN ( _, _, rs)) :: _) -> 
         (begin
	      {
              bb_label = oldb.Csyntax.blk_label ;
              bb_memo = oldb.Csyntax.blk_memo ;
              bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
              bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
              bb_inputs = oldb.Csyntax.blk_inputs ;
              bb_outputs = oldb.Csyntax.blk_outputs ;
              bb_body = BBASSIGN {
                  bbassign = rs ;
              }
	      }
        end)
        | ((Cexpr.RESET (ex)) :: _) -> (begin
	      {
              bb_label = oldb.Csyntax.blk_label ;
              bb_memo = oldb.Csyntax.blk_memo ;
              bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
              bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
              bb_inputs = oldb.Csyntax.blk_inputs ;
              bb_outputs = oldb.Csyntax.blk_outputs ;
              bb_body = BBASSIGN {
                            bbassign = ex ;
                        }
              }
            
        end)
	   | _ -> (assert false)
     end)
  | Csyntax.BBASME  ->
      (begin
	    match oldb.Csyntax.blk_stmts with
        | ((Cexpr.ASSUME _ ) :: _) -> 
         (begin
	      let l = oldb.Csyntax.blk_stmts in
	      let f0 = ((AND (List.map translate_stmt_simple l))) in
	      {
              bb_label = oldb.Csyntax.blk_label ;
              bb_memo = oldb.Csyntax.blk_memo ;
              bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
              bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
              bb_inputs = oldb.Csyntax.blk_inputs ;
              bb_outputs = oldb.Csyntax.blk_outputs ;
              bb_body = BBASSUME {
                  bbassume = f0 ;
              }
	      }
        end)
	   | _ -> (assert false)
     end)
  (* BinhLT added - end *)
  | Csyntax.BBCOND ->
      (begin
	match oldb.Csyntax.blk_stmts with
	  [(Cexpr.IF (e0, [(Cexpr.GOTO thengo)], [(Cexpr.GOTO elsego)]))] ->
	    (begin
	      let x0 = (translate_to_form e0) in
	      let cond = (x0, thengo, elsego) in
	      {
		bb_label = oldb.Csyntax.blk_label ;
		bb_memo = oldb.Csyntax.blk_memo ;
		bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
		bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
		bb_inputs = oldb.Csyntax.blk_inputs ;
		bb_outputs = oldb.Csyntax.blk_outputs ;
		bb_body = BBCOND {
		  bbcond_cond = cond
		}
	      }
	    end)
	| _ -> (assert false)
      end)
  | Csyntax.BBJUNC ->
      (begin
	assert (oldb.Csyntax.blk_stmts == []) ;
	assert ((List.length oldb.Csyntax.blk_incoming_labels) > 1) ;
	let label = oldb.Csyntax.blk_label in
	let phis = (translate_phis oldb.Csyntax.blk_phis
		       oldb.Csyntax.blk_incoming_labels) in
	let phifix =
	  (List.map
	      (fun e ->
		match e with
		  (Cexpr.ASGN (phivar, tmpvar)) ->
		    ((translate_to_term phivar), (translate_to_term tmpvar))
		| _ -> (assert false))
	      oldb.Csyntax.blk_phifix) in
	let predexprs : expr_t list =
	  (Misc.hashtbl_find_safe [] predsets label) in
	let predforms : form_t list =
	  (uniqify_form_list (List.map translate_to_form predexprs)) in
	(* Map variable names to temporary ones. *)
	let varmap : (varname_t, varname_t) Hashtbl.t = (Hashtbl.create 10) in
	(List.iter
	    (fun (v0, v1) ->
	      match (v0, v1) with
		((VAR n0), (VAR n1)) ->
		  (Hashtbl.replace varmap n0 n1)
	      | _ -> (assert false))
	    phifix) ;
	let predmap : (form_t, form_t) Hashtbl.t = (Hashtbl.create 10) in
	let predsmapped =
	  (List.map
	      (substitute_variables_in_predicate varmap predmap)
	      predforms) in
	if (!trTRACE_forms) then
	  (debug (";; predicates.at(" ^ fn ^ "/" ^ label ^ ")="
			   ^ (form_list_to_string predforms))) ;
	{
	  bb_label = oldb.Csyntax.blk_label ;
	  bb_memo = oldb.Csyntax.blk_memo ;
	  bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
	  bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
	  bb_inputs = oldb.Csyntax.blk_inputs ;
	  bb_outputs = oldb.Csyntax.blk_outputs ;
	  bb_body = BBJUNC {
	    bbjunc_phis = phis ;
	    bbjunc_phifix = phifix ;
	    bbjunc_predicates = predsmapped ;
	    bbjunc_predmap = predmap
	  }
	}
      end)
  | Csyntax.BBCALL ->
      (begin
	let label : string = oldb.Csyntax.blk_label in
	let predexprs : expr_t list =
	  (Misc.hashtbl_find_safe [] predsets label) in
	let predforms : form_t list =
	  (uniqify_form_list (List.map translate_to_form predexprs)) in
  (* debug *)
  (*List.iter
    (fun a_stmt -> debug(Cexpr.stmt_to_string a_stmt))
    oldb.Csyntax.blk_stmts;
  read_int();*)
  
	match oldb.Csyntax.blk_stmts with
   (* 19.02.2009. Xuan has just changed to get outgoing *)
	 (* [(Cexpr.CALLASGN ((Cexpr.CALL (tn, e0, ee)), _))] ->*)
   [Cexpr.CALLASGN ((Cexpr.CALL (tn, e0, ee, spec_op)), 
                    ((Cexpr.ASGN (Cexpr.VREF (_, name, level,version), _)) 
                   :: _), _)
   ] ->
	    (begin
	      let vinmap : (varname_t, varname_t) Hashtbl.t =
		(Hashtbl.create 10) in
	      let voutmap : (varname_t, varname_t) Hashtbl.t =
		(Hashtbl.create 10) in
	      (List.iter
		  (fun (id, level, sequence) ->
		    (Hashtbl.replace vinmap
			(Cexpr.vartag_to_string (id, level, 0))
			(Cexpr.vartag_to_string (id, level, sequence))) ;
		    (Hashtbl.replace voutmap
			(Cexpr.vartag_to_string (id, level, sequence))
			(Cexpr.vartag_to_string (id, level, 0))))
		  (Csyntax.variable_set_to_list oldb.Csyntax.blk_outputs)) ;
	      let callfix =
		      (List.map
  		      (fun e ->
	    	      match e with
			         (Cexpr.ASGN (nvar, ovar)) ->
			           ((translate_to_term nvar), (translate_to_term ovar))
		           | _ -> (assert false)
            )
	          oldb.Csyntax.blk_callfix) 
        in
	      let t0 = (translate_to_term e0) in
	      let tt = (List.map translate_to_term ee) in
	      let ex = (Cexpr.ignore_cast e0) in
	      if (!trTRACE_forms) then
		(debug (";; predicates.at(" ^ fn ^ "/" ^ label ^ ")="
				 ^ (form_list_to_string predforms))) ;
	      match ex with
		(Cexpr.VREF (tn, id, level1, sequence)) ->
		  {
		    bb_label = oldb.Csyntax.blk_label ;
		    bb_memo = oldb.Csyntax.blk_memo ;
		    bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
		    bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
		    bb_inputs = oldb.Csyntax.blk_inputs ; 
        (* 19.02.2009. Xuan has just changed *)
        (*bb_outputs = oldb.Csyntax.blk_outputs ;*) 
		    bb_outputs = Csyntax.variable_list_to_set [name, level, version];
		    bb_body = BBCALL {
		      bbcall_call = (t0 :: tt) ;
		      bbcall_global_reversions = callfix ;
		      bbcall_fn = (id, level1, sequence) ;
		      bbcall_globals_inmap = vinmap ;
		      bbcall_globals_outmap = voutmap ;
		      bbcall_predicates = predforms;
		      bbcall_spec = spec_op;
		    }
		  }
	      | _ ->
		  {
		    bb_label = oldb.Csyntax.blk_label ;
		    bb_memo = oldb.Csyntax.blk_memo ;
		    bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
		    bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
		    bb_inputs = oldb.Csyntax.blk_inputs ;
		    (* 19.02.2009. Xuan has just changed *)
        (*bb_outputs = oldb.Csyntax.blk_outputs ;*) 
        bb_outputs = Csyntax.variable_list_to_set [name, level, version];
		    bb_body = BBCALL {
		      bbcall_call = (t0 :: tt) ;
		      bbcall_global_reversions = callfix ;
		      bbcall_fn = ("", 0, 0) ;
		      bbcall_globals_inmap = vinmap ;
		      bbcall_globals_outmap = voutmap ;
		      bbcall_predicates = predforms;
		      bbcall_spec = spec_op;
		    }
		  }
	    end)
  | (* 19.02.2009. Xuan has just changed to get outgoing *)
     (* [(Cexpr.CALLASGN ((Cexpr.CALL (tn, e0, ee)), _))] ->*)
     [(Cexpr.CALLASGN ((Cexpr.CALL (tn, e0, ee, spec_op)), _, _))] ->
        (begin 
          let vinmap : (varname_t, varname_t) Hashtbl.t =
        (Hashtbl.create 10) in
          let voutmap : (varname_t, varname_t) Hashtbl.t =
        (Hashtbl.create 10) in
          (List.iter
          (fun (id, level, sequence) ->
            (Hashtbl.replace vinmap
            (Cexpr.vartag_to_string (id, level, 0))
            (Cexpr.vartag_to_string (id, level, sequence))) ;
            (Hashtbl.replace voutmap
            (Cexpr.vartag_to_string (id, level, sequence))
            (Cexpr.vartag_to_string (id, level, 0))))
          (Csyntax.variable_set_to_list oldb.Csyntax.blk_outputs)) ;
          let callfix =
              (List.map
              (fun e ->
                  match e with
                     (Cexpr.ASGN (nvar, ovar)) ->
                       ((translate_to_term nvar), (translate_to_term ovar))
                   | _ -> (assert false)
            )
              oldb.Csyntax.blk_callfix) 
        in
          let t0 = (translate_to_term e0) in
          let tt = (List.map translate_to_term ee) in
          let ex = (Cexpr.ignore_cast e0) in
          if (!trTRACE_forms) then
        (debug (";; predicates.at(" ^ fn ^ "/" ^ label ^ ")="
                 ^ (form_list_to_string predforms))) ;
          match ex with
        (Cexpr.VREF (tn, id, level, sequence)) ->
          {
            bb_label = oldb.Csyntax.blk_label ;
            bb_memo = oldb.Csyntax.blk_memo ;
            bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
            bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
            bb_inputs = oldb.Csyntax.blk_inputs ; 
            (* 19.02.2009. Xuan has just changed *)
            (*bb_outputs = oldb.Csyntax.blk_outputs ;*) 
            bb_outputs = Csyntax.variable_list_to_set [];
            bb_body = BBCALL {
              bbcall_call = (t0 :: tt) ;
              bbcall_global_reversions = callfix ;
              bbcall_fn = (id, level, sequence) ;
              bbcall_globals_inmap = vinmap ;
              bbcall_globals_outmap = voutmap ;
              bbcall_predicates = predforms ;
              bbcall_spec = spec_op;
            }
          }
          | _ ->
          {
            bb_label = oldb.Csyntax.blk_label ;
            bb_memo = oldb.Csyntax.blk_memo ;
            bb_outgoing_labels = oldb.Csyntax.blk_outgoing_labels ;
            bb_incoming_labels = oldb.Csyntax.blk_incoming_labels ;
            bb_inputs = oldb.Csyntax.blk_inputs ;
            (* 19.02.2009. Xuan has just changed *)
            (*bb_outputs = oldb.Csyntax.blk_outputs ;*) 
            bb_outputs = Csyntax.variable_list_to_set [];
            bb_body = BBCALL {
              bbcall_call = (t0 :: tt) ;
              bbcall_global_reversions = callfix ;
              bbcall_fn = ("", 0, 0) ;
              bbcall_globals_inmap = vinmap ;
              bbcall_globals_outmap = voutmap ;
              bbcall_predicates = predforms ;
              bbcall_spec = spec_op;
            }
          }
        end)
	| _ -> (assert false)
      end))

and translate_stmt_simple (s : stmt_t) : form_t = (
  match s with
    Cexpr.DEFVAR _ | Cexpr.DEFUN _ | Cexpr.SPEC _ -> (assert false)
  | (Cexpr.STMT (Cexpr.CALL (tn, e0, ee, _))) -> (assert false)
  | (Cexpr.STMT e0) ->
      (*AHO*)
      (debug (";; IGNORE EXPR: "
		       ^ (Cexpr.stmt_to_string_body 1 s))) ;
      TT
  | (Cexpr.ASGN (e0, (Cexpr.CALL (tn, e1, ee, _)))) -> (assert false)
  | (Cexpr.ASGN (e0, Cexpr.NIL)) ->
      (*AHO*)
      (debug (";; IGNORE NULL: "
		       ^ (Cexpr.stmt_to_string_body 1 s))) ;
      TT
  | (Cexpr.ASGN (e0, (Cexpr.PSEUDO_NONDET_VALUE (tn, _)))) ->
      (*AHO*)
      (debug (";; IGNORE NONDET ASSIGNENT: "
		       ^ (Cexpr.stmt_to_string_body 1 s))) ;
      TT
  | (Cexpr.ASGN (e0, e1)) ->
      let t0 = (translate_to_term e0) in
      let t1 = (translate_to_term e1) in
      (PRED (EQ, t0, t1))
  | (Cexpr.ASSUME e) -> (translate_to_form e)
  | (Cexpr.RESET es) -> AND ((List.map translate_to_form es))
  | (Cexpr.CALLASGN _) -> Cexpr.dump_stmts [s];(assert false)
  | Cexpr.IF _ | Cexpr.BEGIN _ | Cexpr.SWITCH _ -> (assert false)
  | Cexpr.DOWHILE _ | Cexpr.FOR _ -> (assert false)
  | Cexpr.BREAK | Cexpr.CONTINUE -> (assert false)
  | Cexpr.LABEL _ | Cexpr.GOTO _ | Cexpr.RETURN _ -> (assert false)
  | (Cexpr.ANNOTATION (_, _)) ->
      (* Just ignore annotations. *)
      TT
  | (Cexpr.ASM e0) ->
      (failwith "NO ASM YET")
  | Cexpr.NOTE _ -> (assert false))

(* Extracts annotations (assert or assume). *)

and translate_annotation (wantform : string) (l : stmt_t list) : form_t = (
  assert (List.mem wantform [ "$assert" ; "$assume" ; "$invariant" ]) ;
  let x =
    (List.map
	(fun (s : stmt_t) ->
	  match s with
	    Cexpr.DEFVAR _ | Cexpr.DEFUN _ | Cexpr.SPEC _ -> (assert false)
	  | (Cexpr.STMT e0) ->
	      TT
      | (Cexpr.ASSUME _) -> TT
      | (Cexpr.RESET _) -> TT
	  | (Cexpr.ASGN (e0, e1)) ->
	      TT
	  | (Cexpr.CALLASGN _) -> (assert false)
	  (*| (Cexpr.IF (e0, [(Cexpr.GOTO thengo)], [(Cexpr.GOTO elsego)])) ->
	    TT*)
	  | Cexpr.IF _ | Cexpr.BEGIN _ | Cexpr.SWITCH _ -> (assert false)
	  | Cexpr.DOWHILE _ | Cexpr.FOR _ -> (assert false)
	  | Cexpr.BREAK | Cexpr.CONTINUE -> (assert false)
	  | Cexpr.LABEL _ | Cexpr.GOTO _ | Cexpr.RETURN _ -> (assert false)
	  | (Cexpr.ANNOTATION (a, e0)) ->
	      if (string_equal a wantform) then
		(translate_to_form e0)
	      else
		TT
	  | (Cexpr.ASM e0) ->
	      (failwith "NO ASM YET")
	  | (Cexpr.NOTE _) -> (assert false))
	l) in
  (pseudo_simplify_form (AND x)))

and translate_to_form (e : expr_t) : form_t = (
  let arithzero = (ARITH (NUM (num_of_int 0))) in
  match e with
    Cexpr.NIL -> (assert false)
  | (Cexpr.PSEUDO_NONDET_VALUE (tn, _)) -> (assert false)
  (*GOMI*)(*
  | (Cexpr.OPER (tn, op, (Cexpr.PSEUDO_NONDET_VALUE (_, _)), _, _))
  | (Cexpr.OPER (tn, op, _, (Cexpr.PSEUDO_NONDET_VALUE (_, _)), _)) ->
      (debug (";; IGNORE NONDET EXPRESSION: "
		       ^ (Cexpr.expr_to_string_body 1 [] e))) ;
      (PROP (xgensym "R"))
  *)
  | (Cexpr.OPER (tn, Cexpr.LOGNOT, e0, _, _)) ->
      let f0 = (translate_to_form e0) in
      (NOT f0)
  | (Cexpr.OPER (tn, ((Cexpr.EQ|Cexpr.NE) as op), e0, e1, _))
  | (Cexpr.OPER (tn, ((Cexpr.GE|Cexpr.GT) as op), e0, e1, _))
  | (Cexpr.OPER (tn, ((Cexpr.LE|Cexpr.LT) as op), e0, e1, _)) ->
      (begin
	let t0 = (translate_to_term e0) in
	let t1 = (translate_to_term e1) in
	match op with
	  Cexpr.EQ -> (PRED (EQ, t0, t1))
	| Cexpr.NE -> (PRED (NE, t0, t1))
	| Cexpr.GT -> (PRED (GT, t0, t1))
	| Cexpr.GE -> (PRED (GE, t0, t1))
	| Cexpr.LT -> (PRED (LT, t0, t1))
	| Cexpr.LE -> (PRED (LE, t0, t1))
	| _ -> (assert false)
      end)
  | (Cexpr.OPER (tn, ((Cexpr.OROR|Cexpr.ANDAND) as op), e0, e1, _)) ->
      (begin
	let f0 = (translate_to_form e0) in
	let f1 = (translate_to_form e1) in
	match op with
	  Cexpr.OROR -> (OR [f0 ; f1])
	| Cexpr.ANDAND -> (AND [f0 ; f1])
	| _ -> (assert false)
      end)
  | (Cexpr.OPER _) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.SETF _) -> (assert false)
  | (Cexpr.VREF (tn, id, level, sequence)) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.PREF (tn, _, e0, _)) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.SREF (tn, _, e0, s)) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.CALL (tn, e0, ee, _)) -> (assert false)
  | (Cexpr.ADDROF (tn, e0)) ->
      TT
  | (Cexpr.SIZEOF (tn, t0)) ->
      TT
  | (Cexpr.PROGN (tn, l)) -> (assert false)
  | (Cexpr.PSEUDO_MEMORY (tn, t0)) ->
      TT
  | (Cexpr.INIT_FILL _) ->
      TT
  | (Cexpr.CONST_I _) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.CONST_F _) ->
      let t0 = (translate_to_term e) in
      (PRED (NE, t0, arithzero))
  | (Cexpr.STRING _) ->
      TT
  | (Cexpr.CLITERAL (tn, l)) -> (assert false)
  | (Cexpr.QUANTIFIER (s, vs, e0)) ->
      (*AHO*)
      (translate_to_form e0)
  | (Cexpr.PSEUDO_PROLOGUE) ->
      TT
  | (Cexpr.PSEUDO_EPILOGUE) ->
      TT
  | (Cexpr.ASMSEQUENCE ee) -> (assert false)
  | (Cexpr.PHI _) -> (assert false)
  | (Cexpr.DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (Cexpr.PRAGMA s) -> (assert false)
  | (Cexpr.PSEUDO_ACSL_RESULT _) -> (assert false)
  | (Cexpr.PSEUDO_ACSL_OLD _ ) -> (assert false)
  | (Cexpr.ALL_FIELD _) -> (assert false)
  | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK _ -> (assert false)
  | (Cexpr.PSEUDO_ACSL_UNIQUE (tn, e0)) -> (assert false)
)

and translate_to_term (e : expr_t) : term_t = (
  let arithzero = (ARITH (NUM (num_of_int 0))) in
  (* Try constant-folding before translating the expression.
     If there is anything wrong while constant-folding, go back
     to translating the original expression *)
  let e = try Cexpr.fold_constant_integer e
          with _ -> e
  in
  match e with
    Cexpr.NIL -> (assert false)
  | (Cexpr.PSEUDO_NONDET_VALUE (tn, _)) ->
      (*AHO*)
      (debug (";; NONDET EXPR: "
		       ^ (Cexpr.expr_to_string_body 1 [] e))) ;
      (VAR (xgensym "R"))
  | (Cexpr.OPER (tn, (Cexpr.CAST _), e0, _, _)) ->
      (translate_to_term e0)
  | (Cexpr.OPER (tn, ((Cexpr.NEG|Cexpr.BITNOT) as op), e0, _, _)) ->
      (begin
	let t0 = (translate_to_term e0) in
	match op with
	  Cexpr.NEG -> (ARITH (NEG t0))
	| Cexpr.BITNOT -> (UIFUN [(FUN "bitnot") ; t0])
	| _ -> (assert false)
      end)
  | (Cexpr.OPER (tn, ((Cexpr.ADD|Cexpr.SUB|Cexpr.MUL) as op), e0, e1, _))
  | (Cexpr.OPER (tn, ((Cexpr.DIV|Cexpr.MOD) as op), e0, e1, _))
  | (Cexpr.OPER (tn, ((Cexpr.LSH|Cexpr.RSH) as op), e0, e1, _))
  | (Cexpr.OPER (tn, ((Cexpr.AND|Cexpr.IOR|Cexpr.XOR) as op), e0, e1, _)) ->
      (begin
	let t0 = (translate_to_term e0) in
	let t1 = (translate_to_term e1) in
	match op with
	  Cexpr.ADD -> (ARITH (ADD (t0, t1)))
	| Cexpr.SUB -> (ARITH (ADD (t0, (ARITH (NEG t1)))))
	| Cexpr.MUL -> (ARITH (MUL (t0, t1)))
	| Cexpr.DIV -> (UIFUN [(FUN "div") ; t0 ; t1])
	| Cexpr.MOD -> (UIFUN [(FUN "mod") ; t0 ; t1])
	| Cexpr.LSH -> (UIFUN [(FUN "lsh") ; t0 ; t1])
	| Cexpr.RSH -> (UIFUN [(FUN "rsh") ; t0 ; t1])
	| Cexpr.AND -> (UIFUN [(FUN "and") ; t0 ; t1])
	| Cexpr.IOR -> (UIFUN [(FUN "ior") ; t0 ; t1])
	| Cexpr.XOR -> (UIFUN [(FUN "xor") ; t0 ; t1])
	| _ -> (assert false)
      end)
  | (Cexpr.OPER (tn, Cexpr.LOGNOT, e0, _, _)) -> (assert false)
  | (Cexpr.OPER (tn, (Cexpr.EQ|Cexpr.NE), e0, e1, _)) -> (assert false)
  | (Cexpr.OPER (tn, (Cexpr.GE|Cexpr.GT), e0, e1, _)) -> (assert false)
  | (Cexpr.OPER (tn, (Cexpr.LE|Cexpr.LT), e0, e1, _)) -> (assert false)
  | (Cexpr.OPER (tn, ((Cexpr.ANDAND|Cexpr.OROR) as op), e0, e1, _)) ->
      (begin
	(* ANDAND/OROR is for value and arguments are zero/one. *)
	let t0 = (translate_to_term e0) in
	let t1 = (translate_to_term e1) in
	match op with
	  Cexpr.ANDAND -> (UIFUN [(FUN "and") ; t0 ; t1])
	| Cexpr.OROR -> (UIFUN [(FUN "ior") ; t0 ; t1])
	| _ -> (assert false)
      end)
  | (Cexpr.OPER (tn, (Cexpr.COND), e0, e1, _)) ->
      (* {Csyntax.foolish_translation_p} never generates {COND}. *)
      (assert false)
  | (Cexpr.OPER _) -> (assert false)
  | (Cexpr.SETF _) -> (assert false)
  | (Cexpr.VREF (tn, id, level, sequence)) ->
      if (Cexpr.enumerator_p tn id) then
        (ARITH (NUM (Cexpr.get_enumerator_value tn id))) 
      else
        (VAR (Cexpr.expr_to_string e))
  | (Cexpr.PREF (tn, _, e0, e1)) ->
      let t0 = (translate_to_term e0)
      and t1 = match e1 with
               | Cexpr.NIL -> (UIFUN [(FUN "NIL")])
               | _ -> (translate_to_term e1)
      in
      (UIFUN [(FUN "pref") ; t0 ; t1])
  (* 20.02.2009. Xuan has just changed UIFUN *)
  (*| (Cexpr.SREF (tn, _, e0, s)) ->
    let t0 = (translate_to_term e0) in
    (UIFUN [(FUN ("sref%" ^ s)) ; t0])*)
  | Cexpr.SREF (tn, b, e0, s) ->
    begin
      let tyn = Cexpr.type_of_expr e0 in
      let name = (*"sref%"*) 
        match tyn with
          | Cexpr.POINTER_TO _ -> "srefp"
          | _ -> "sref"
      in
      let t0 = (translate_to_term e0) in
      (UIFUN [(FUN (name)); t0; VAR (s ^ "%0%0")])
      (* end change *)
    end
  | (Cexpr.CALL (tn, e0, ee, _)) -> (assert false)
  | (Cexpr.ADDROF (tn, e0)) ->
      let t0 = (translate_to_term e0) in
      (UIFUN [(FUN ("addrof")) ; t0])
  | (Cexpr.SIZEOF (tn, t0)) ->
      (*AHO*)
      (*(UIFUN [(FUN ("sizeof" ^ (Cexpr.type_to_string t0)))])*)
      let e0 = (Cexpr.fold_constant_integer e) in
      (translate_to_term e0)
  | (Cexpr.PROGN (tn, l)) -> (assert false)
  | (Cexpr.CONST_I (tn, n, s)) ->
      (ARITH (NUM n))
  | (Cexpr.CONST_F (tn, n, s)) ->
      (UIFUN [(FUN ("constf%" ^ s))])
  | (Cexpr.STRING (b, s, _)) ->
      (UIFUN [(FUN ("memory"(* Xuan commented 20090710 string% ^ s*)))])
  | (Cexpr.CLITERAL (tn, l)) -> (assert false)
  | (Cexpr.PHI _) -> (assert false)
  | (Cexpr.QUANTIFIER (s, vs, e0)) -> (assert false)
  | (Cexpr.PSEUDO_MEMORY (tn, t0)) ->
      (UIFUN [(FUN ("memory"))])
  | (Cexpr.INIT_FILL _) ->
      arithzero
  | (Cexpr.PSEUDO_PROLOGUE) ->
      arithzero
  | (Cexpr.PSEUDO_EPILOGUE) ->
      arithzero
  | (Cexpr.ASMSEQUENCE ee) -> (assert false)
  | (Cexpr.DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (Cexpr.PRAGMA s) -> (assert false)
  | (Cexpr.PSEUDO_ACSL_RESULT (tn, e0)) ->
      let t0 = (translate_to_term e0) in
      t0
  | (Cexpr.PSEUDO_ACSL_OLD (tn, e0, e1)) -> (translate_to_term e0)
  | (Cexpr.ALL_FIELD _) -> assert false
  | (Cexpr.PSEUDO_ACSL_UNIQUE (tn, e0)) -> (
      let t0 = (translate_to_term e0) in
      (UIFUN [(FUN ("unique_addr")); t0])
    )
  | (Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1)) -> (
      let e11 = Cexpr.fold_constant_integer e1 in
      (* still need to check whether e11 is a constant,
         and raise an error otherwise *)
      let t0 = (translate_to_term e0)
      and t1 = (translate_to_term e11) in
      (UIFUN [(FUN ("unique_block")); t0; t1])
    )
)
(* MEMO: It puts TRUE when phi mapping is empty. *)

and translate_phis (stmts : stmt_t list) (labels : string list) : (string, form_t) Hashtbl.t = (
  let insert_conjunct (tbl : (string, form_t) Hashtbl.t) (label : string) (e : form_t) : unit = (
    assert (Hashtbl.mem tbl label) ;
    let f = (Hashtbl.find tbl label) in
    (Hashtbl.replace tbl label (Form.pseudo_conjoin f e))) in
  let newtbl = (Hashtbl.create 10) in
  (List.iter
      (fun label ->
	(Hashtbl.replace newtbl label TT))
      labels) ;
  (List.iter
      (fun s ->
	(begin
	  match s with
	    (Cexpr.ASGN (v, (Cexpr.PHI (_, phimap)))) ->
	      (begin
		let t0 = (translate_to_term v) in
		(List.iter
		    (fun (label, e) ->
		      if (e == Cexpr.NIL) then
			(begin
			  (*AHO*)
			  (debug ("IGNORE NIL PHI CHOICE: " ^ label))
			end)
		      else
			(begin
			  if (not (Cexpr.identical_variable_p v e)) then
			    let t1 = (translate_to_term e) in
			    (insert_conjunct newtbl label (PRED (EQ, t0, t1)))
			end))
		    phimap)
	      end)
	  | _ -> (assert false)
	end))
      stmts) ;
  assert ((Hashtbl.length newtbl) != 0) ;
  newtbl)
