type exit_cause_t = NONDET | TIMEOUT | NORMAL

let exit_cause = ref NONDET
let mutex = Mutex.create ()

let exit_by (cause : exit_cause_t) : unit =
    (assert (cause != NONDET));
    let _ =
        if (!exit_cause = NONDET) then begin
            (Mutex.lock mutex); 
            (exit_cause := cause);
            (Mutex.unlock mutex)
            end
    in
    ()

let normal_exit () = (
    (Mutex.lock mutex);
    let b = (!exit_cause == NORMAL) in
    (Mutex.unlock mutex);
    b)

let timeout_exit () = (
    (Mutex.lock mutex);
    let b = (!exit_cause == TIMEOUT) in
    (Mutex.unlock mutex);
    b)

type la_complexity_t = {
    mutable refinement_number : int;
    mutable node_explored_number : int;
    mutable number_of_preds_in_paths : int list;
    mutable solver_invocation_number : int;
}

let la_complexity : la_complexity_t = {
    refinement_number = 0;
    node_explored_number = 0;
    number_of_preds_in_paths = [];
    solver_invocation_number = 0;
}

let update_refinement_number () : unit = (
    la_complexity.refinement_number <- la_complexity.refinement_number + 1;
)

let update_node_explored_number () : unit = (
    la_complexity.node_explored_number <- la_complexity.node_explored_number + 1;
)

let update_number_of_preds_in_paths (number_of_preds : int) : unit =  (
    la_complexity.number_of_preds_in_paths <-
        la_complexity.number_of_preds_in_paths @ [number_of_preds]
)

let update_solver_invocation_number () : unit = (
    la_complexity.solver_invocation_number <-
        la_complexity.solver_invocation_number + 1;
)

let get_max_pred_number (number_of_preds_in_paths : int list) : int = (
    List.fold_left (fun max_number number ->
        if max_number > number then max_number
        else number
    ) (0) number_of_preds_in_paths;
)

let get_min_pred_number (number_of_preds_in_paths : int list) : int = (
    let start_number = try List.hd number_of_preds_in_paths
                       with Failure _ -> (0)
    in
    List.fold_left (fun min_number number ->
        if min_number < number then min_number
        else number
    ) start_number number_of_preds_in_paths;
)

let get_average_pred_number (number_of_pred_in_paths : int list) : float = (
    let sum = List.fold_left (fun sum number ->
                  sum + number
              ) 0 number_of_pred_in_paths in
    let len = List.length number_of_pred_in_paths in
    if len = 0 then (0.0)
    else ((float_of_int sum) /. (float_of_int len))
)

let get_deviation_pred_number (number_of_pred_in_paths : int list) : float = (
    let average : float = get_average_pred_number number_of_pred_in_paths in
    let square_difference =
        List.fold_left (fun result number ->
            let difference : float = (float_of_int number) -. average in
            result +. (difference *. difference)
        ) 0.0 number_of_pred_in_paths in
    sqrt (square_difference /. average)
)

let string_of_la_complexity () : string = (
    " =========COMPLEXITY_OF_CHECKING==========\n"
    ^ "    + Total number of nodes explored: "
    ^ (string_of_int la_complexity.node_explored_number) ^ "\n"
    ^ "    + Statistics of the number of predicates per each path: \n"
    ^ "        * Max: "
    ^ (string_of_int
          (get_max_pred_number la_complexity.number_of_preds_in_paths)) ^ "\n"
    ^ "        * Min: "
    ^ (string_of_int
          (get_min_pred_number la_complexity.number_of_preds_in_paths)) ^ "\n"
    ^ "        * Average: "
    ^ (string_of_float
          (get_average_pred_number la_complexity.number_of_preds_in_paths)) ^ "\n"
    ^ "        * Deviation: "
    ^ (string_of_float
          (get_deviation_pred_number la_complexity.number_of_preds_in_paths)) ^ "\n"
    ^ "    + Number of refinements: "
    ^ (string_of_int la_complexity.refinement_number) ^ "\n"
    ^ "    + Number of solver invocation: "
    ^ (string_of_int la_complexity.solver_invocation_number)
)
