(* la/cfa.ml                                                     *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Define Control Flow Automation (CFA) & functions to make Abstract       *)
(* Reachability Tree form CFAs                                             *)

open LogicUtil.Logic;;
open La_common;;

let check_pointer : bool ref = ref true;;

(* The counter which is to implement uniqueness of a memory location *)
let unique_counter = ref 0

(* type of cfa_node *)
type cfa_stmt =
    | CFA_FORM of predicate
    | CFA_COND of (predicate * string * string)
    | CFA_CALL of (expression list) * ((agent_fspec_t * old_term_t) option)
    | CFA_ASSUME of predicate
    | CFA_RESET of (expression list)
    | CFA_ENTRY of predicate
    | CFA_RETURN
    (* because we don't use statement of error_node, we remove it *)
    | CFA_ERROR
    (* inherit of form_t, but currently don't use *)
    | CFA_FAIL

(* type cfa_attribute_t doesn't use at present *)
(* type cfa_attribute_t =
  | CFA_NO_ATTR
  | CFA_NORETURN_ATTR
  | CFA_CONST_ATTR
  | CFA_NOTHROW_ATTR *)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and cfa_node = {
    cfa_label : string;
    cfa_outgoing_labels : string list;
    cfa_incoming_labels : string list;
    (* address in C program: *)
    cfa_memo : (*string*) (string * string * string * string * string) list;
    (* cfa_memo includes filename * startline * startposision * endline * endposition *)
    cfa_node_stmt : cfa_stmt;
    cfa_inputs : string list;
    cfa_outputs : string list;
}

and agent_expr_t = {
    agent_assign_nodes : (string, cfa_node) Hashtbl.t;
    agent_temp_varid : string;
    agent_expr_loc : Lexing.position * Lexing.position
}

and agent_fspec_t = (agent_expr_t, agent_expr_t) Sexpr.function_contract_t

and old_term_t = (string, cfa_node) Hashtbl.t
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~ REGION: DEFINE CFA ~~~~~~~~~~~~~~~~~~~~~~~~~ *)
type cfa_meta_data_t = {
    cfa_id : string;
    cfa_tn : Cexpr.typenode_t; 
    cfa_scls : Cexpr.storage_class_t; 
    (* cfa_inits is used to initialize for global varriables *)
    cfa_inits : predicate list;
    cfa_params : string list; (* parameters of funcstion *)
    cfa_struct_defined_p : bool; (* currently don't use *)
    (*cfa_attributes : cfa_attribute_t list*)
    cfa_location : Cexpr.note_t option;
    cfa_specs : (agent_fspec_t * old_term_t) option;
    mutable cfa_container : ofile_t option;
    mutable cfa_is_cutentry : bool
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and cfa_vardef_t = {
    cfa_v_meta_data : cfa_meta_data_t;
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and cfa_fundef_t = {
    cfa_entry_label : string;
    cfa_f_meta_data : cfa_meta_data_t;
    cfa_node_hashtbl : (string, cfa_node) Hashtbl.t;
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and cfa =
    | CFA_FUNDEF of cfa_fundef_t
    | CFA_VARDEF of cfa_vardef_t
    | CFA_OTHERDEF

(**
   Parsed and translated definitions in file.  {ofile_defs} holds
   function/variable definitions of a file.  {ofile_global_predicates}
   is a list of predicates on globals, which is used to trace global
   states.  {ofile_hash_name} is a generated file name, which is used
   as a unique key for storing ofiles in an afile. 
   
   Imitate Translate.ofile_t. 
*)
and ofile_t = {
    ofile_hash_name : string;
    ofile_file_name : string;
    ofile_cfas : (string, cfa) Hashtbl.t;
    ofile_global_predicates : Form.form_t list;
    ofile_contract_of_field : (string * string, ((agent_fspec_t * old_term_t)
                                                 * string list)) Hashtbl.t
}

(**
   Set of translated files (in archives or executables).
   {afile_start} is a start symbol.  {afile_libs} holds a set of file
   definitions.  {afile_directory} maps an external name to a
   definition file name.  {afile_global_predicates} maps a global
   variable name to a list of its related abstraction predicates.

     Imitate Translate.ofile_t. 
*)

and afile_t = {
    afile_libs : (string, ofile_t) Hashtbl.t;
    afile_directory : (string, string) Hashtbl.t;
    afile_start : Sexpr.function_entry_t;
    afile_global_predicates : (string, Form.form_t list) Hashtbl.t
}

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let replace_incoming_nodes (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (label : string)
                           (label_to_be_excluded : string)
                           (labels_to_be_included : string list)
                           : unit = (
    let node =
        try Hashtbl.find cfa_node_hash label
        with Not_found -> assert false
    in
    let incoming_labels = node.cfa_incoming_labels in
    let new_incoming_labels =
        List.filter (fun label -> label <> label_to_be_excluded)
                    incoming_labels
    in
    let new_incoming_labels =
        new_incoming_labels @ labels_to_be_included in
    let new_node =
        { node with cfa_incoming_labels = new_incoming_labels }
    in
    (Hashtbl.replace cfa_node_hash label new_node))

let replace_incoming_nodes (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (labels : string list)
                           (label_to_be_excluded : string)
                           (labels_to_be_included : string list)
                           : unit =
    List.iter (fun label ->
                   replace_incoming_nodes cfa_node_hash
                                          label
                                          label_to_be_excluded
                                          labels_to_be_included)
              labels

let replace_outgoing_nodes (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (label : string)
                           (label_to_be_excluded : string)
                           (labels_to_be_included : string list)
                           : unit = (
    let node =
        try Hashtbl.find cfa_node_hash label
        with Not_found -> assert false
    in
    let outgoings = node.cfa_outgoing_labels in
    let new_outgoings =
        List.filter (fun outgoing -> outgoing <> label_to_be_excluded)
                    outgoings in
    let new_outgoings =
        new_outgoings @ labels_to_be_included in
    let new_node =
        { node with cfa_outgoing_labels = new_outgoings}
    in
    (Hashtbl.replace cfa_node_hash label new_node))

let replace_outgoing_nodes (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (labels : string list)
                           (label_to_be_excluded: string)
                           (labels_to_be_included : string list)
                           : unit =
    List.iter (fun label ->
                   replace_outgoing_nodes cfa_node_hash
                                          label
                                          label_to_be_excluded
                                          labels_to_be_included)
              labels

let insert_cfa_hash_above (sub_cfa : (string, cfa_node) Hashtbl.t)
                          (main_cfa : (string, cfa_node) Hashtbl.t)
                          (above_this_label : string)
                          : (string, cfa_node) Hashtbl.t = (
    let above_this_node =
        try Hashtbl.find main_cfa above_this_label
        with Not_found -> assert false
    in
    let sub_cfa_nodes = Hashtbl.copy sub_cfa
    and main_cfa_nodes = Hashtbl.copy main_cfa in
    let entry_node_name = "#Lentry"
    and return_node_name = "#Lreturn" in
    let sub_entry_node =
        try Hashtbl.find sub_cfa_nodes entry_node_name
        with Not_found -> assert false
    and sub_return_node =
        try Hashtbl.find sub_cfa_nodes return_node_name
        with Not_found -> assert false
    in
    let outgoings_of_sub_entry =
        sub_entry_node.cfa_outgoing_labels
    and incomings_of_sub_return =
        sub_return_node.cfa_incoming_labels
    in
    let fathers_of_above_this_node =
        above_this_node.cfa_incoming_labels
    in
    (* replace outgoings of above_this_node *)
    let new_above_this_node =
        { above_this_node with cfa_incoming_labels = incomings_of_sub_return }
    in
    (Hashtbl.replace main_cfa_nodes
                     above_this_node.cfa_label
                     new_above_this_node);
    (* replace incomings of nodes pointed by sub entry node *)
    (replace_incoming_nodes sub_cfa_nodes 
                            outgoings_of_sub_entry
                            entry_node_name
                            fathers_of_above_this_node);
    (* replace outgoings of sub return node *)
    (replace_outgoing_nodes sub_cfa_nodes
                            incomings_of_sub_return
                            return_node_name
                            [new_above_this_node.cfa_label]);
    (* replace outgoings of nodes pointed by above_this_node *)
    (replace_outgoing_nodes main_cfa_nodes
                            fathers_of_above_this_node
                            above_this_node.cfa_label
                            outgoings_of_sub_entry);
    (* remove entry and return node of sub cfa *)
    (Hashtbl.remove sub_cfa_nodes entry_node_name);
    (Hashtbl.remove sub_cfa_nodes return_node_name);
    (Hashtbl.iter (fun node_name a_node ->
                       Hashtbl.add main_cfa_nodes node_name a_node)
                  sub_cfa_nodes);
    main_cfa_nodes
)

let insert_cfa_node_below (node : cfa_node)
                          (main_cfa_nodes : (string, cfa_node) Hashtbl.t)
                          (below_this_label : string)
                          : (string, cfa_node) Hashtbl.t = (
    let below_this_node =
        try Hashtbl.find main_cfa_nodes below_this_label
        with Not_found -> assert false
    in
    let children_labels = below_this_node.cfa_outgoing_labels in
    let new_below_this_node =
        { below_this_node with cfa_outgoing_labels = [node.cfa_label] }
    in
    let new_node =
        { node with cfa_incoming_labels = [below_this_node.cfa_label];
                    cfa_outgoing_labels = children_labels }
    in
    let new_main_cfa_nodes = Hashtbl.copy main_cfa_nodes in
    (Hashtbl.add new_main_cfa_nodes new_node.cfa_label new_node);
    (Hashtbl.replace new_main_cfa_nodes
                     new_below_this_node.cfa_label
                     new_below_this_node);
    (replace_incoming_nodes new_main_cfa_nodes
                            children_labels
                            below_this_node.cfa_label
                            [node.cfa_label]);
    new_main_cfa_nodes) 

let insert_cfa_node_below_to_cfa (node : cfa_node)
                                 (main_cfa : cfa)
                                 (below_this_label : string)
: cfa = (
    match main_cfa with
    | CFA_FUNDEF f -> (
        let cfa_node_hash = f.cfa_node_hashtbl in
        let new_cfa_node_hash = insert_cfa_node_below node cfa_node_hash
                                                      below_this_label in
        CFA_FUNDEF {f with cfa_node_hashtbl = new_cfa_node_hash}
      )
    | _ -> assert false
)

(** get cfa_node_hashtbl of cfa if it is fundef else raise exception *)
let find_node_in_cfa (cfa : cfa) (node_name : string) : cfa_node = (
    match cfa with
    | CFA_FUNDEF f -> (
        let cfa_nodes = f.cfa_node_hashtbl in
        try Hashtbl.find cfa_nodes node_name
        with Not_found -> assert false
      )
    | _ -> assert false
)

let insert_stmt_below_to_cfa (node_stmt : cfa_stmt)
                             (label : string)
                             (main_cfa : cfa)
                             (below_this_label : string)
: cfa = (
    let below_node = find_node_in_cfa main_cfa below_this_label in
    let new_node = {below_node with cfa_label = label;
                                    cfa_node_stmt = node_stmt} in
    insert_cfa_node_below_to_cfa new_node main_cfa below_this_label
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_initial_assumptions (specs_opt : (agent_fspec_t * old_term_t) option) = (
    match specs_opt with
    | Some ({
             Sexpr.fspec_conditions = fun_conditions;
             Sexpr.fspec_behaviors = _;
             Sexpr.fspec_assigns = _;
             Sexpr.fspec_prohibits = _;
             Sexpr.fspec_local_invs = _;
             Sexpr.fspec_context = _;
             Sexpr.fspec_is_handler = _
           }, _) -> (
        let preconds : agent_expr_t list
            = List.fold_left
                       (fun (preconds : agent_expr_t list)
                            (cond : agent_expr_t Sexpr.func_condition_t) ->
                            let pos, pred = cond in
                            match pos with
                            | Sexpr.PRE_CONDITION -> preconds @ [pred]
                            | _ -> preconds)
                       [] fun_conditions
        in
        preconds)
    | None -> [])

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let create_main_cfa_with_assumptions (acfa : cfa) : cfa = (
    let insert_an_assumption (cfa_nodes : (string, cfa_node) Hashtbl.t)
                             (below_this_label : string)
                             (condition : agent_expr_t)
                             : ((string, cfa_node) Hashtbl.t * string) = (
        let {
              agent_assign_nodes = sub_cfa_nodes;
              agent_temp_varid = varid;
              agent_expr_loc = (start_pos, end_pos)
            } = condition
        in
        (assert (start_pos.Lexing.pos_fname = end_pos.Lexing.pos_fname));
        let assume_node_label = "Assuming_" ^ varid in
        let assumption = (Not(Eq(Variable varid, Constant 0.)))
        in
        let assume_node = 
            {
              cfa_label = assume_node_label;
              cfa_outgoing_labels = [];
              cfa_incoming_labels = [];
              cfa_memo = [(start_pos.Lexing.pos_fname,
                           (string_of_int start_pos.Lexing.pos_lnum),
                           (string_of_int start_pos.Lexing.pos_cnum),
                           (string_of_int end_pos.Lexing.pos_lnum),
                           (string_of_int end_pos.Lexing.pos_cnum))];
              cfa_node_stmt = CFA_ASSUME assumption;
              cfa_inputs = [];
              cfa_outputs = [];
            }
        in
        let cfa_nodes = insert_cfa_node_below assume_node
                                              cfa_nodes
                                              below_this_label
        in
        let cfa_nodes = insert_cfa_hash_above sub_cfa_nodes
                                              cfa_nodes
                                              assume_node_label
        in
        (cfa_nodes, assume_node_label))
    in
    match acfa with
    | CFA_FUNDEF {
                   cfa_entry_label = entry_label;
                   cfa_f_meta_data = meta_data;
                   cfa_node_hashtbl = cfa_nodes;
                 } -> (
        let specs_opt = meta_data.cfa_specs in
        let assumptions = get_initial_assumptions specs_opt in
        let new_cfa_nodes, _ =
            List.fold_left (fun cfa_and_label assumption ->
                                let cfa_hash, label = cfa_and_label in
                                (insert_an_assumption cfa_hash
                                                      label
                                                      assumption))
                           (cfa_nodes, entry_label)
                           assumptions
        in
        (CFA_FUNDEF { 
                      cfa_entry_label = entry_label;
                      cfa_f_meta_data = meta_data;
                      cfa_node_hashtbl = new_cfa_nodes
                    }))
    | CFA_VARDEF _
    | CFA_OTHERDEF -> assert false)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_id_of_cfa (a_cfa : cfa) : string = (
    match a_cfa with
    | CFA_FUNDEF fundef ->
        fundef.cfa_f_meta_data.cfa_id
    | CFA_VARDEF vardef ->
        vardef.cfa_v_meta_data.cfa_id
    | CFA_OTHERDEF ->
        assert false )

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Raise exception Not_found if we can not find any cfa with 
    the input name (cfa_name) *)

let find_cfa_in_ofile (ofile : ofile_t) (id : string) : cfa =
    Hashtbl.find ofile.ofile_cfas id

let find_cfa_in_afile (afile : afile_t) (id : string) : cfa = (
    let hashstr = Hashtbl.find afile.afile_directory id in
    let ofile = Hashtbl.find afile.afile_libs hashstr in
    find_cfa_in_ofile ofile id)
    
let find_cfa_in_afile (afile : afile_t)
                      ?(caller = None)
                      (symbol_name : string)
                      : cfa = (begin
    match caller with
    | None ->
        find_cfa_in_afile afile symbol_name
    | Some cfa -> (
        let container =
            let container_opt =
                match cfa with
                | CFA_VARDEF v ->
                    v.cfa_v_meta_data.cfa_container
                | CFA_FUNDEF f ->
                    f.cfa_f_meta_data.cfa_container
                | CFA_OTHERDEF ->
                    (assert false)
            in
            match container_opt with
            | Some ofile -> ofile
            | None -> (assert false)
        in
        (* specify whether the symbol is a local one *)
        let is_local =
            ExtString.String.starts_with symbol_name
                                         (container.ofile_hash_name ^ "%")
        in
        if is_local then
            find_cfa_in_ofile container symbol_name
        else
            find_cfa_in_afile afile symbol_name)
end)

let find_cfa_of_function_entry (afile : afile_t)
                               (entry : Sexpr.function_entry_t)
                               : cfa = (
    match entry with
    | {
        Sexpr.fe_filelocation_opt = None;
        Sexpr.fe_funcname = funcname
      } ->
        find_cfa_in_afile afile funcname
    | {
        Sexpr.fe_filelocation_opt = Some fileloc;
        Sexpr.fe_funcname = funcname
      } -> (
        (* first, looks for the corresponding lib *)
        let candidate_libs =
            Hashtbl.fold
                (fun _ ofile candidates ->
                     try
                         if (Sexpr.identical fileloc ofile.ofile_file_name) then
                             ofile :: candidates
                         else
                             candidates
                     with
                     (* tolerate with the Not_found exception, which happens
                        when it does not find the path of a lib.
                        Apparently, the exception occurs with the lib of
                        builtin-functions *)
                     | Not_found -> candidates)
                afile.afile_libs
                []
        in
        (* then looks into the corresponding lib *)
        match candidate_libs with
        | [] ->
            raise Not_found
        | [lib] -> (
            try Hashtbl.find lib.ofile_cfas funcname
            with
            | Not_found ->
               let local_name = lib.ofile_hash_name ^ "%" ^ funcname in
               Hashtbl.find lib.ofile_cfas local_name)
        | _ ->
            assert false))

let create_cutentry_hashstr (fileloc : Sexpr.file_location_t)
                            (funcname : string)
                            : string =
    ((Sexpr.string_of_file_location fileloc) ^ "#" ^ funcname)


let mem (cutlist : (string, Sexpr.function_entry_t option) Hashtbl.t)
        (cfa : cfa)
        : bool = (
    let mem (cutlist : (string, Sexpr.function_entry_t option) Hashtbl.t)
            filename
            funcname = (
        try
            let new_fl = Sexpr.make_file_location filename in
            let hashstr = create_cutentry_hashstr new_fl funcname in
            if (Hashtbl.mem cutlist hashstr) then (
                (Hashtbl.replace cutlist hashstr None);
                true)
            else
                false
        with
        | Not_found -> false)
    in
    let meta_data =
        match cfa with
        | CFA_VARDEF v -> v.cfa_v_meta_data
        | CFA_FUNDEF f -> f.cfa_f_meta_data
        | CFA_OTHERDEF -> assert false
    in
    let lib =
        match meta_data.cfa_container with
        | Some ofile -> ofile
        | None -> assert false
    in
    let is_local ofile name =
        let prefix = ofile.ofile_hash_name ^ "%" in
        ExtString.String.starts_with name prefix
    in
    let id = meta_data.cfa_id in
    match meta_data.cfa_location with
    | Some (Cexpr.LINE (_, fn, _, _))
    | Some (Cexpr.RANGE (fn, _, _, _, _)) -> (
        let real_id =
            if (is_local lib id) then
                let n = try String.rindex id '%'
                        with Not_found -> -1
                in
                let length = String.length id in
                String.sub id (n+1) (length-n-1)
            else
                id
        in
        mem cutlist fn real_id)
    | None ->
        false)

let create_cutlist (afile : afile_t)
                   (cutlist : Sexpr.function_entry_t list)
                   : (string, Sexpr.function_entry_t option) Hashtbl.t = (
    let get_meta_data cfa =
        match cfa with
        | CFA_VARDEF v -> v.cfa_v_meta_data
        | CFA_FUNDEF f -> f.cfa_f_meta_data
        | CFA_OTHERDEF -> assert false
    in
    let new_cutlist = Hashtbl.create 10 in
    List.iter
        (fun cut_entry ->
            match cut_entry with
            | {
                Sexpr.fe_filelocation_opt = None;
                Sexpr.fe_funcname = funcname;
                Sexpr.fe_raw_data = raw_data;
                Sexpr.fe_location = locopt
              } -> (
                try
                    let cfa = find_cfa_in_afile afile funcname in
                    let meta_data = get_meta_data cfa in
                    let filename =
                        match meta_data.cfa_location with
                        | Some (Cexpr.LINE (_, fn, _, _))
                        | Some (Cexpr.RANGE (fn, _, _, _, _)) -> fn
                        | None -> assert false
                    in
                    let fl = Sexpr.make_file_location filename in
                    let hashstr = create_cutentry_hashstr fl funcname in
                    Hashtbl.add new_cutlist hashstr (Some cut_entry)
                with
                | Not_found -> (
                    let fn, ln =
                        match locopt with
                        | Some (fn, ln) -> fn, ln
                        | None -> assert false
                    in                        
                    Scommon.warning (fn ^ ":" ^ (string_of_int ln)
                             ^ ": cannot find the cutlist entry: '"
                             ^ raw_data
                             ^ "'")))
            | {
                Sexpr.fe_filelocation_opt = Some fl;
                Sexpr.fe_funcname = funcname
              } -> (
                let hashstr = create_cutentry_hashstr fl funcname in
                Hashtbl.add new_cutlist hashstr (Some cut_entry)))
        cutlist;
    new_cutlist)
 
let import_cutlist (afile : afile_t)
                   (cutlist : Sexpr.function_entry_t list)
                   : unit = (
    let import_cutlist_to_cfa (cfa : cfa) (cutlist) : unit =
        if (mem cutlist cfa) then
            match cfa with
            | CFA_VARDEF v -> v.cfa_v_meta_data.cfa_is_cutentry <- true
            | CFA_FUNDEF f -> f.cfa_f_meta_data.cfa_is_cutentry <- true
            | _ -> ()
    in
    let import_cutlist_to_ofile (ofile : ofile_t) (cutlist) : unit =
        Hashtbl.iter (fun _ cfa -> import_cutlist_to_cfa cfa cutlist)
                     ofile.ofile_cfas
    in
    let cutlist = create_cutlist afile cutlist in
    (Hashtbl.iter (fun _ ofile -> import_cutlist_to_ofile ofile cutlist)
                  afile.afile_libs);
    (Hashtbl.iter (fun _ pseudo_visited ->
                       match pseudo_visited with
                       | None -> ()
                       | Some {
                                Sexpr.fe_filelocation_opt = _;
                                Sexpr.fe_funcname = _;
                                Sexpr.fe_raw_data = raw_data;
                                Sexpr.fe_location = Some (fn, ln)
                              } ->
                           Scommon.warning (fn ^ ":" ^ (string_of_int ln)
                                    ^ ": cannot find the cutlist entry: '"
                                    ^ raw_data
                                    ^ "'")
                       | _ -> assert false)
                  cutlist))

let process_undef_functions (afile : afile_t) : unit = (
    let process_undef_function (cfa:cfa) : unit = (
        match cfa with
        | CFA_VARDEF v -> (
            let meta_data = v.cfa_v_meta_data in
            let tn = meta_data.cfa_tn
            and id = meta_data.cfa_id
            in
            if (Cexpr.function_type_p tn)
               && (not meta_data.cfa_is_cutentry) then (
                let real_id =
                    let n = try String.rindex id '%'
                            with Not_found -> -1
                    in
                    let length = String.length id in
                    String.sub id (n+1) (length-n-1)
                in
                let location_str =
                    match meta_data.cfa_location with
                    | None -> ""
                    | Some (Cexpr.LINE (_, fn, ln, _))
                    | Some (Cexpr.RANGE (fn, ln, _, _, _)) ->
                        fn ^ ":" ^ (string_of_int ln) ^ ": "
                in
                Scommon.warning (location_str
                                 ^ "undefined function '" ^ real_id ^ "'"
                                 ^ "; regarded as a cutlist entry.");
                v.cfa_v_meta_data.cfa_is_cutentry <- true))
        | CFA_FUNDEF _
        | CFA_OTHERDEF -> ())
    in
    let process_undef_funcs_in_ofile (ofile : ofile_t) : unit = (
        Hashtbl.iter (fun symbol_name cfa ->
                          try
                              let definition_cfa =
                                  find_cfa_in_afile afile
                                                    ~caller:(Some cfa)
                                                    symbol_name
                              in
                              process_undef_function definition_cfa
                          with
                          | Not_found -> (
                              if (String.length symbol_name) > 0 then (
                                  (* unfound symbol must be a typedef'ed name
                                     or an CFA_OTHERDEF *)
                                  match cfa with
                                  | CFA_VARDEF v -> (
                                      let scls = v.cfa_v_meta_data.cfa_scls in
                                      match scls with
                                      | Cexpr.TYPEDEF -> ()
                                      | _ -> (
                                          emerg ("Cannot find a variable : '"
                                                 ^ symbol_name ^ "'");
                                          assert false))
                                  | CFA_FUNDEF f -> (
                                      emerg ("Cannot find a function : "
                                             ^ symbol_name);
                                      assert false)
                                  | CFA_OTHERDEF -> ())))
                     ofile.ofile_cfas)
    in
    Hashtbl.iter (fun _ ofile ->
                        process_undef_funcs_in_ofile ofile)
                 afile.afile_libs)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* preprocess to remove the ellipsis "..." in cfa_params list *)
let preprocess_cfa_params (params : string list) : string list =
begin
    let new_cfa_params = ref [] in
    List.iter (
        fun a_para ->
        begin
            let sub_str = (try String.sub a_para 0 3
                           with _ -> "") in
            if (sub_str <> "...") then
                new_cfa_params := !new_cfa_params @ [a_para];
        end
    ) params;
    !new_cfa_params;
end

(*============================================================*)

let range_op_to_memo (range : Sexpr.range_t option)
: (string * string * string * string * string) = begin
    (match range with
    | None -> ("", "", "", "", "")
    | Some (filename, (startline, startnum), (endline, endnum)) -> 
        (filename, string_of_int startline, string_of_int startnum,
                   string_of_int endline, string_of_int endnum))
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let insert_edge_into_cfa_fundef (a_cfa : cfa) 
      (incoming_label : string)
      (outgoing_label : string)
      (new_stmt : cfa_stmt)
      (new_label : string)
      (new_memo : (string * string * string * string * string) list)
: cfa =
begin
    (* Debug *)
    (*debug ("insert_edge_into_cfa_fundef");
    debug ("incoming_label: " ^ incoming_label);
    debug ("outgoing_label: " ^ outgoing_label);
    debug ("new_label: " ^ new_label);*)

    (* Check a_cfa_fundef *)
    let a_cfa_fundef = (match a_cfa with
                       | CFA_FUNDEF fd -> fd
                       | anything_else -> assert false)
    in
    (* Create new cfa_node *)
    let new_cfa_node_hashtbl : (string, cfa_node) Hashtbl.t = 
        Hashtbl.copy (a_cfa_fundef.cfa_node_hashtbl) in
    let new_cfa_node : cfa_node = {
        cfa_label = new_label;
        cfa_outgoing_labels = [outgoing_label];
        cfa_incoming_labels = [incoming_label];
        cfa_memo = (*"temporary"*) new_memo;
        cfa_node_stmt = new_stmt;
        cfa_inputs = [];
        cfa_outputs = [];
    } in
    (* Add new_cfa_node to new_cfa_node_hashtbl *)
    (* 20090626 *)
    (*Hashtbl.add (new_cfa_node_hashtbl) (new_label) (new_cfa_node);*)
    Hashtbl.replace (new_cfa_node_hashtbl) (new_label) (new_cfa_node);
    (* Rename outgoing_label of incoming_node *)
    (try
        let incoming_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (incoming_label) in
        let new_incoming_node : cfa_node = {
            cfa_label = incoming_node.cfa_label;
            cfa_outgoing_labels = (begin
                let new_cfa_outgoing_labels : string list ref = ref [] in
                List.iter
                    (fun l -> 
                        if ((String.compare (l) (outgoing_label)) = 0) then begin
                            new_cfa_outgoing_labels := 
                                !new_cfa_outgoing_labels @ [new_label];
                        end else begin
                            new_cfa_outgoing_labels := 
                                !new_cfa_outgoing_labels @ [l];
                        end
                    )
                    incoming_node.cfa_outgoing_labels;
                !new_cfa_outgoing_labels;
            end);
            cfa_incoming_labels = incoming_node.cfa_incoming_labels;
            cfa_memo = incoming_node.cfa_memo;
            cfa_node_stmt = incoming_node.cfa_node_stmt;
            cfa_inputs = incoming_node.cfa_inputs;
            cfa_outputs = incoming_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (incoming_label) (new_incoming_node);
    with 
    | Not_found -> assert false);
    (* Rename incoming_label of outgoing_node *)
    (try
        let outgoing_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (outgoing_label) in
        let new_outgoing_node : cfa_node = {
            cfa_label = outgoing_node.cfa_label;
            cfa_outgoing_labels = outgoing_node.cfa_outgoing_labels;
            cfa_incoming_labels = (begin
                let new_cfa_incoming_labels : string list ref = ref [] in
                List.iter
                    (fun l -> 
                        if ((String.compare (l) (incoming_label)) = 0) then begin
                            new_cfa_incoming_labels := 
                                !new_cfa_incoming_labels @ [new_label];
                        end else begin
                            new_cfa_incoming_labels := 
                                !new_cfa_incoming_labels @ [l];
                        end
                    )
                    outgoing_node.cfa_incoming_labels;
                !new_cfa_incoming_labels;
            end);
            cfa_memo = outgoing_node.cfa_memo;
            cfa_node_stmt = outgoing_node.cfa_node_stmt;
            cfa_inputs = outgoing_node.cfa_inputs;
            cfa_outputs = outgoing_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (outgoing_label) (new_outgoing_node);
    with 
    | Not_found -> assert false);
    (* Create new cfa_fundef_t *)
    let new_cfa_fundef : cfa_fundef_t = {
        cfa_entry_label = a_cfa_fundef.cfa_entry_label;
        cfa_f_meta_data = a_cfa_fundef.cfa_f_meta_data;
        cfa_node_hashtbl = new_cfa_node_hashtbl;
    } in
    (* Return *)
    CFA_FUNDEF new_cfa_fundef
end

let insert_edge_into_cfa_fundef1 (a_cfa : cfa) 
      (incoming_label : string)
      (outgoing_label : string)
      (new_stmt : cfa_stmt)
      (new_label : string)
      (new_memo : (string * string * string * string * string) list)
: cfa =
begin
    (* Debug *)
    (*debug ("insert_edge_into_cfa_fundef");
    debug ("incoming_label: " ^ incoming_label);
    debug ("outgoing_label: " ^ outgoing_label);
    debug ("new_label: " ^ new_label);*)

    (* Check a_cfa_fundef *)
    let a_cfa_fundef = (match a_cfa with
                       | CFA_FUNDEF fd -> fd
                       | anything_else -> assert false)
    in
    (* Create new cfa_node *)
    let new_cfa_node_hashtbl : (string, cfa_node) Hashtbl.t = 
        Hashtbl.copy (a_cfa_fundef.cfa_node_hashtbl) in
    let new_cfa_node : cfa_node = {
        cfa_label = new_label;
        cfa_outgoing_labels = [outgoing_label];
        cfa_incoming_labels = [incoming_label];
        cfa_memo = (*"temporary"*) new_memo;
        cfa_node_stmt = new_stmt;
        cfa_inputs = [];
        cfa_outputs = [];
    } in
    (* Add new_cfa_node to new_cfa_node_hashtbl *)
    (* 20090626 *)
    (*Hashtbl.add (new_cfa_node_hashtbl) (new_label) (new_cfa_node);*)
    Hashtbl.replace (new_cfa_node_hashtbl) (new_label) (new_cfa_node);
    (* Rename outgoing_label of incoming_node *)
    (try
        let incoming_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (incoming_label) in
        let new_incoming_node : cfa_node = {
            cfa_label = incoming_node.cfa_label;
            cfa_outgoing_labels =
                    incoming_node.cfa_outgoing_labels @ [new_label];
            cfa_incoming_labels = incoming_node.cfa_incoming_labels;
            cfa_memo = incoming_node.cfa_memo;
            cfa_node_stmt = incoming_node.cfa_node_stmt;
            cfa_inputs = incoming_node.cfa_inputs;
            cfa_outputs = incoming_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (incoming_label) (new_incoming_node);
    with 
    | Not_found -> assert false);
    (* Rename incoming_label of outgoing_node *)
    (try
        let outgoing_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (outgoing_label) in
        let new_outgoing_node : cfa_node = {
            cfa_label = outgoing_node.cfa_label;
            cfa_outgoing_labels = outgoing_node.cfa_outgoing_labels;
            cfa_incoming_labels =
                    outgoing_node.cfa_incoming_labels @ [new_label];
            cfa_memo = outgoing_node.cfa_memo;
            cfa_node_stmt = outgoing_node.cfa_node_stmt;
            cfa_inputs = outgoing_node.cfa_inputs;
            cfa_outputs = outgoing_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (outgoing_label) (new_outgoing_node);
    with 
    | Not_found -> assert false);
    (* Create new cfa_fundef_t *)
    let new_cfa_fundef : cfa_fundef_t = {
        cfa_entry_label = a_cfa_fundef.cfa_entry_label;
        cfa_f_meta_data = a_cfa_fundef.cfa_f_meta_data;
        cfa_node_hashtbl = new_cfa_node_hashtbl;
    } in
    (* Return *)
    CFA_FUNDEF new_cfa_fundef
end



(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Blocks:
If (assert_pred = true) then goto outgoing_node (outgoing_label)
Else, Call assert(0), then outgoing = elsewhere
*)
let inline_assert_func_into_cfa_fundef (a_cfa : cfa)
                                       (incoming_label : string)
                                       (outgoing_label : string)
                                       (funcall_number : int)
                                       (require_num : int)
                                       (assert_pred : predicate)
                                       (range_op : Sexpr.range_t option)
: cfa = 
begin
    (* Check cfa_fundef *)
    let a_cfa_fundef = (match a_cfa with
                       | CFA_FUNDEF fd -> fd
                       | anything_else -> assert false)
    in 
    (* Create new cfa_node_hashtbl *)
    let new_cfa_node_hashtbl : (string, cfa_node) Hashtbl.t = 
        Hashtbl.copy (a_cfa_fundef.cfa_node_hashtbl) in
    let funcall_string : string = string_of_int funcall_number in
    let require_num_string : string = string_of_int require_num in
    let new_memo = [range_op_to_memo range_op] in
    (* Insert assert(0) *)
    let assert_node : cfa_node = {
        cfa_label = "##assert_false_call-" ^ funcall_string ^ "-" ^ require_num_string;
        cfa_outgoing_labels = ["##elsewhere"];
        cfa_incoming_labels = ["##else_requires-" ^ funcall_string ^ "-" ^ require_num_string];
        cfa_memo = new_memo (*"temporary"*);
        cfa_node_stmt = CFA_ERROR;
        cfa_inputs = [];
        cfa_outputs = [];
    } in 
    Hashtbl.add (new_cfa_node_hashtbl) (assert_node.cfa_label) (assert_node);
    (* Insert if-else block *)
    let if_else_node : cfa_node = {
        cfa_label = "##check_requires-" ^ funcall_string ^ "-" ^ require_num_string;
        cfa_outgoing_labels = [outgoing_label; assert_node.cfa_label];
        cfa_incoming_labels = [incoming_label];
        cfa_memo = new_memo (*"temporary"*);
        cfa_node_stmt = CFA_COND (assert_pred, outgoing_label, assert_node.cfa_label);
        cfa_inputs = [];
        cfa_outputs = [];
    } in 
    Hashtbl.add (new_cfa_node_hashtbl) (if_else_node.cfa_label) (if_else_node);
    (* Rename outgoing_label of incoming_node *)
    (try
        let incoming_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (incoming_label) in
        let new_incoming_node : cfa_node = {
            cfa_label = incoming_node.cfa_label;
            cfa_outgoing_labels = (begin
                let new_cfa_outgoing_labels : string list ref = ref [] in
                List.iter
                    (fun l -> 
                        if ((String.compare (l) (outgoing_label)) = 0) then begin
                            new_cfa_outgoing_labels := 
                                !new_cfa_outgoing_labels @ [if_else_node.cfa_label];
                        end else begin
                            new_cfa_outgoing_labels := !new_cfa_outgoing_labels @ [l];
                        end
                    )
                    incoming_node.cfa_outgoing_labels;
                !new_cfa_outgoing_labels;
            end);
            cfa_incoming_labels = incoming_node.cfa_incoming_labels;
            cfa_memo = incoming_node.cfa_memo;
            cfa_node_stmt = incoming_node.cfa_node_stmt;
            cfa_inputs = incoming_node.cfa_inputs;
            cfa_outputs = incoming_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (incoming_label) (new_incoming_node);
    with 
    | Not_found -> assert false);
    (* Rename incoming_label of outgoing_node *)
    (try
        let outgoing_node : cfa_node = 
            Hashtbl.find (new_cfa_node_hashtbl) (outgoing_label) in
        let new_outgoing_node : cfa_node = {
            cfa_label = outgoing_node.cfa_label;
            cfa_outgoing_labels = outgoing_node.cfa_outgoing_labels;
            cfa_incoming_labels = (begin
                let new_cfa_incoming_labels : string list ref = ref [] in
                List.iter
                    (fun l -> 
                        if ((String.compare (l) (incoming_label)) = 0) then begin
                            new_cfa_incoming_labels := 
                                !new_cfa_incoming_labels @ [if_else_node.cfa_label];
                        end else begin
                            new_cfa_incoming_labels := !new_cfa_incoming_labels @ [l];
                        end
                    )
                    outgoing_node.cfa_incoming_labels;
                !new_cfa_incoming_labels;
            end);
            cfa_memo = outgoing_node.cfa_memo;
            cfa_node_stmt = outgoing_node.cfa_node_stmt;
            cfa_inputs = outgoing_node.cfa_inputs;
            cfa_outputs = outgoing_node.cfa_outputs;
        } in
        Hashtbl.replace (new_cfa_node_hashtbl) (outgoing_label) (new_outgoing_node);
    with 
    | Not_found -> assert false);
    (* Create new cfa_fundef *)
    let new_cfa_fundef : cfa_fundef_t = {
        cfa_entry_label = a_cfa_fundef.cfa_entry_label;
        cfa_f_meta_data = a_cfa_fundef.cfa_f_meta_data;
        cfa_node_hashtbl = new_cfa_node_hashtbl;
    } in
    (* Return *)
    CFA_FUNDEF new_cfa_fundef
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let get_string_of_cfa_attribute_t (s_attribute : cfa_attribute_t) 
  : string =
    begin
        match a_cfa_attribute with
        | CFA_NO_ATTR -> "CFA_NO_ATTR"
        | CFA_NORETURN_ATTR -> "CFA_NORETURN_ATTR"
        | CFA_CONST_ATTR -> "CFA_CONST_ATTR"
        | CFA_NOTHROW_ATTR -> "CFA_NOTHROW_ATTR"
    end*)

let rec agent_expr_to_string (agent_expr : agent_expr_t) : string = (
    let {
          agent_assign_nodes = cfa_nodes;
          agent_temp_varid = varid;
          agent_expr_loc = _
        } = agent_expr in
    let cfa_node_list = Hashtbl.fold (fun _ cfa_node cfa_node_list ->
        cfa_node_list @ [cfa_node];
    ) cfa_nodes [] in
    (varid) ^ "\n"
    ^ (String.concat "\n" (List.map get_string_of_cfa_node cfa_node_list)))

and agent_fspecs_to_string (agent_fspecs : agent_fspec_t) =
    Sexpr.string_of_function_contract agent_expr_to_string
                                      agent_expr_to_string
                                      agent_fspecs

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used for cfa_dumper.ml                                 *)
and get_string_of_cfa_meta_data (a_meta_data : cfa_meta_data_t) : string =
begin
    let string_to_return = ref "\\n" in
    (* get string of cfa_id *)
    let name = a_meta_data.cfa_id in
    string_to_return := !string_to_return ^ 
                        "cfa_id: " ^ name ^ "%1";
    (* get string of cfa_tn *)
    string_to_return := !string_to_return ^ "\\ncfa_tn: " ^ 
                        Cexpr.type_to_string (a_meta_data.cfa_tn);
    (* get string of cfa_scls *)
    string_to_return := !string_to_return ^ "\\ncfa_scls: " ^ 
                        Cexpr.scls_to_string (a_meta_data.cfa_scls);
    (* get string of cfa_inits *)
    string_to_return := !string_to_return ^ "\\ncfa_inits: ";
    List.iter
        (fun a_pred ->
            string_to_return := !string_to_return ^ "\\n\\t" ^ 
                                (string_of_predicate a_pred)
        )
        a_meta_data.cfa_inits;
    (* get string of cfa_params *)
    string_to_return := !string_to_return ^ "\\ncfa_params: ";
    List.iter
        (fun a_param ->
            string_to_return := !string_to_return ^ "--" ^ a_param)
        a_meta_data.cfa_params;
    (* get string of cfa_struct_defined_p *)
    string_to_return := !string_to_return ^ "\\ncfa_struct_defined_p: " ^
                        (string_of_bool a_meta_data.cfa_struct_defined_p);
    (* get string of cfa_attributes *)
    (*string_to_return := !string_to_return ^ "\\ncfa_attributes: ";
    List.iter
        (fun a_attribute ->
            let string_of_a_attribute = 
                get_string_of_cfa_attribute_t (a_attribute) in
            string_to_return := 
                !string_to_return ^ "--" ^ (string_of_a_attribute))
        a_meta_data.cfa_attributes;*)
    string_to_return := (
        match a_meta_data.cfa_specs with
        | None -> !string_to_return
        | Some (specs, _) ->
            !string_to_return
            ^ "\\ncfa_specs : "
            ^ (agent_fspecs_to_string specs));
    (* return *)
    !string_to_return
end

and get_string_of_cfa_vardef (a_cfa_vardef : cfa_vardef_t) : string =
    begin
        "CFA_VARDEF" ^ get_string_of_cfa_meta_data (a_cfa_vardef.cfa_v_meta_data)
    end

and get_string_of_cfa_f_meta_data (a_f_meta_data : cfa_meta_data_t) : string =
    begin
        "CFA_FUNDEF" ^ get_string_of_cfa_meta_data (a_f_meta_data)
    end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and get_string_of_meta_data_of_cfa (a_cfa : cfa) : string =
    begin
        match a_cfa with
        | CFA_FUNDEF cfa_fundef ->
            get_string_of_cfa_f_meta_data (cfa_fundef.cfa_f_meta_data)
        | CFA_VARDEF cfa_vardef -> get_string_of_cfa_vardef (cfa_vardef)
        | CFA_OTHERDEF -> assert false
    end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_nil_cfa_fundef () =
    begin
        (* create meta data *)
        let nil_meta_data : cfa_meta_data_t = {
            cfa_id = "";
            cfa_tn = Cexpr.NOTYPE;
            cfa_scls = Cexpr.NILSCLS;
            cfa_inits = [];
            cfa_params = [];
            cfa_struct_defined_p = false;
            (*cfa_attributes = [];*)
	    cfa_location = None;
            cfa_specs = None;
            cfa_container = None;
            cfa_is_cutentry = false
        } in
        (* create node_hashtbl *)
        let cfa_node_hashtbl = Hashtbl.create (0) in
        (* create nil_cfa_fundef *)
        let nil_cfa_fundef : cfa_fundef_t = {
            cfa_entry_label = "";
            cfa_f_meta_data = nil_meta_data;
            cfa_node_hashtbl = cfa_node_hashtbl;
        } in
        (* return *)
        nil_cfa_fundef
    end

and create_nil_ofile () = (
  {
    ofile_hash_name = "";
    ofile_file_name = "";
    ofile_cfas = Hashtbl.create 0;
    ofile_global_predicates = [];
    ofile_contract_of_field = Hashtbl.create 0;
  }
)

(* ~~~~~~~~~~~~~~~~ get string of information of cfa_fundef ~~~~~~~~~~~~~~ *)
and get_string_of_cfa_node (c_node : cfa_node) : string =
begin
    let outgoing_labels = String.concat ";" c_node.cfa_outgoing_labels in
    let incoming_labels = String.concat ";" c_node.cfa_incoming_labels in
    let edge_string = 
        (match c_node.cfa_node_stmt with
        | CFA_FORM pred	-> "CFA_FORM\t" ^ string_of_predicate (pred)
        | CFA_COND (pred, s1, s2)	-> "CFA_COND\t" ^ string_of_predicate (pred)
        | CFA_ASSUME pred -> "CFA_ASSUME\t" ^ string_of_predicate (pred)       
        | CFA_RESET expr_list -> begin
            let string_of_expr_list = ref "" in
            List.iter
                (fun a_expr ->
                    string_of_expr_list := !string_of_expr_list ^ "   " ^ 
                                           string_of_expression (a_expr)
                )
                expr_list;
            "CFA_RESET\t"  ^ !string_of_expr_list
          end 
        | CFA_CALL (expr_list, _) -> begin
            let string_of_expr_list = ref "" in
            List.iter
                (fun a_expr ->
                    string_of_expr_list := !string_of_expr_list ^ "   " ^ 
                                           string_of_expression (a_expr)
                )
                expr_list;
            "CFA_CALL\t" ^ !string_of_expr_list
          end
        | anything_else -> "")
    in
    "out: " ^ outgoing_labels ^ " -- in: " ^ incoming_labels ^ " -- edge: " ^
    edge_string
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_cfa_fundef_t (fundef : cfa_fundef_t) : string =
begin
    let sfundef = ref "" in
    Hashtbl.iter
        (fun name cnode ->
            sfundef := 
                !sfundef ^ "\n" ^ name ^ " : " ^ get_string_of_cfa_node (cnode);
        )
        fundef.cfa_node_hashtbl;
    (* return *)
    !sfundef
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_cfa (a_cfa : cfa) : string = 
begin
     match a_cfa with
    | CFA_FUNDEF cfa_fundef -> string_of_cfa_fundef_t (cfa_fundef)
    | CFA_VARDEF cfa_vardef ->
        get_string_of_cfa_meta_data (cfa_vardef.cfa_v_meta_data)
    | anything_else -> ""
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_string_of_cfa_hash (cfa_hash : ((string * int), cfa) Hashtbl.t)
: string =
    begin
        let string_of_cfa_hash = ref "" in
        Hashtbl.iter
            (fun (_, _) a_cfa ->
                match a_cfa with
                | CFA_FUNDEF cfa_fundef ->
                    string_of_cfa_hash := !string_of_cfa_hash ^ "\n" ^ 
                                          (string_of_cfa_fundef_t (cfa_fundef))
                | CFA_VARDEF cfa_vardef ->
                    string_of_cfa_hash := !string_of_cfa_hash ^ "\n" ^ 
                      (get_string_of_cfa_meta_data (cfa_vardef.cfa_v_meta_data))
                | anything_else -> ()
            )
            cfa_hash;
        (* return *)
        !string_of_cfa_hash
    end
 
(* ~~~~~~~~~~~~~~~~ change_entry_and_return_cfa_label ~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used to change entry & return label of a cfa when a    *)
(* statement call that function                                            *)
let change_entry_and_return_cfa_label (new_name_tail : string) 
                                      (old_label : string)
: string =
    begin
        (* debug *)
        (*debug ("change_entry_and_return_cfa_label: ");
        debug ("new_name_tail: " ^ new_name_tail);
        debug ("old_label: " ^ old_label);
        read_int();*)
        
        if (String.compare old_label "#Lentry") = 0 then 
            "#Lentry" ^ new_name_tail
        else if (String.compare old_label "#Lreturn") = 0 then 
            "#Lreturn" ^ new_name_tail
        else old_label
    end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_entry_and_return_labels (old_labels : string list)
                                   (new_name_tail : string) 
: string list =
    begin
        (* foreach label, changes it *)
        let new_labels = 
            List.map (change_entry_and_return_cfa_label (new_name_tail))
                     old_labels
        in
        new_labels
    end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_entry_and_return_labels_in_cfa_stmt (c_stmt : cfa_stmt)
                                               (new_name_tail : string) 
: cfa_stmt =
    begin
        match c_stmt with
        | CFA_COND (pred, label1, label2) ->
            let new_label1 =
                change_entry_and_return_cfa_label (new_name_tail) (label1) in
            let new_label2 =
                change_entry_and_return_cfa_label (new_name_tail) (label2) in
            CFA_COND (pred, new_label1, new_label2)
        | anything_else -> anything_else
    end
  
  
(* ~~~~~~~~ change_entry_and_return_labels_in_cfa_function ~~~~~~~~~~~~~~~ 
 entry_label = entry_label + func_name + level of function;              
 return_label = return_label + func_name + level of function               *)
let change_entry_and_return_labels_in_cfa_function (cfa_fundef : cfa) : cfa =
begin
    (* Step1: check cfa_fundef *)
    match cfa_fundef with
    | CFA_FUNDEF func -> begin
        (* Step2: create new name tail *)
        let name_of_func = func.cfa_f_meta_data.cfa_id in
        let new_name_tail = "%" ^ name_of_func ^ "%1" in
        (* create new Hash: *)
        let new_cnhashtbl : (string, cfa_node) Hashtbl.t = Hashtbl.create (0) in
        (* Step3: foreach cfa_node of cfa_fundef, change label *)
        Hashtbl.iter
            (fun a_label a_node ->
                let new_label = 
                    change_entry_and_return_cfa_label (new_name_tail)
                                                      (a_node.cfa_label) 
                in
                let new_outgoing_labels = 
                    change_entry_and_return_labels (a_node.cfa_outgoing_labels) 
                                                   (new_name_tail)
                in            
                let new_incoming_labels = 
                    change_entry_and_return_labels (a_node.cfa_incoming_labels) 
                                                   (new_name_tail)
                in
                let new_cfa_node_stmt = 
                    change_entry_and_return_labels_in_cfa_stmt (a_node.cfa_node_stmt) 
                                                               (new_name_tail)
                in  
                (* create new node of a_node *)               
                let a_new_node : cfa_node = {
                    cfa_label = new_label;
                    cfa_outgoing_labels = new_outgoing_labels;
                    cfa_incoming_labels = new_incoming_labels;
                    cfa_memo = a_node.cfa_memo;
                    cfa_node_stmt = new_cfa_node_stmt;
                    cfa_inputs = a_node.cfa_inputs;
                    cfa_outputs = a_node.cfa_outputs;
                } in
                (* update new_node to hashtbl to return *)
                Hashtbl.add (new_cnhashtbl) (a_new_node.cfa_label) (a_new_node)
            )
            func.cfa_node_hashtbl;
        (* Step4: create new_func *)
        let new_func : cfa_fundef_t = {
            cfa_entry_label = func.cfa_entry_label ^ new_name_tail;
            cfa_f_meta_data = func.cfa_f_meta_data;
            cfa_node_hashtbl = new_cnhashtbl;
        } in
        (* return *)
        CFA_FUNDEF new_func
      end
    | _ -> assert false (*raise (ConfigUtil.Message (Logging.Error, "isn't cfa_fundef"))*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let transform_bbstmt_to_cfa_node_stmt (bbstmt : Translate.bbstmt_t) : cfa_stmt = 
begin
    (* 24.02.2009. Xuan has just added to get "assume" *)
    (* Step1. Get pred of form: *)
    let pred_of_bbstmt_form : predicate = 
        Form.form_to_predicate (bbstmt.Translate.bbstmt_form) in
    let based_pred_of_bbstmt_form : predicate = 
        create_based_vars_in_predicate (*true*) (pred_of_bbstmt_form) in
    let normalized_pred : predicate = 
        normalize_pred_with_theories (based_pred_of_bbstmt_form) in
    (*let preds_list_of_form : predicate list = 
        (match based_pred_of_bbstmt_form with
        | And [something] -> [something]
        | anything_else -> [anything_else]) in*)
    (* Step2. Get pred of assume: *)
    (*let pred_of_bbstmt_assume : predicate = 
        predicate_of_form (bbstmt.Translate.bbstmt_assume) in
    let based_pred_of_bbstmt_assume : predicate =
        create_based_vars_in_predicate (true) (pred_of_bbstmt_assume) in
    let normalized_pred_of_assume : predicate = 
        Not (Not ((*normalize*) (based_pred_of_bbstmt_assume))) in*)
    (* Step3. Return: *)
    (*if based_pred_of_bbstmt_assume != True then 
        let new_pred = And (normalized_pred_of_assume :: preds_list_of_form) in
        CFA_FORM (new_pred)
    else CFA_FORM (based_pred_of_bbstmt_form) *)
    CFA_FORM normalized_pred  
end
                                      
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let transform_bbcond_to_cfa_node_stmt (bbcond : Translate.bbcond_t) : cfa_stmt = 
begin
    (match bbcond.Translate.bbcond_cond with
    | (cond_form, true_label, false_label) -> begin
        let pred_of_cond_form = Form.form_to_predicate (cond_form) in
        let based_pred_of_cond_form : predicate = 
            create_based_vars_in_predicate (*true*) (pred_of_cond_form) in
        let normalized_cond_pred : predicate = 
            normalize_pred_with_theories (based_pred_of_cond_form) in
        (*CFA_COND (based_pred_of_cond_form, true_label, false_label)*)
        CFA_COND (normalized_cond_pred, true_label, false_label)
      end)
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec transform_bbcall_to_cfa_node_stmt (bbcall : Translate.bbcall_t) : cfa_stmt =
begin
    let expression_list = ref [] in
    (* transform bbcall_call *)
    match bbcall.Translate.bbcall_call with
    | [] -> assert false
    | name_of_fun :: arguments_list -> begin
        List.iter
            (fun a_term ->
                let expr_of_a_term : expression = Form.term_to_expr (a_term) in
                (* 20.02.2009. Xuan comment: *)
                let based_expr_of_a_term : expression = 
                    create_based_vars_in_expression (*true*) (expr_of_a_term) in
                expression_list := !expression_list @ (*[expr_of_a_term]*)
                                   [based_expr_of_a_term]
                (* 20.02.2009. Xuan comment: [based_expr_of_a_term]*)
            )  
            arguments_list;
        (* get name and level of fun *)
        let func_expr = Form.term_to_expr (name_of_fun) in
        let based_func_expr = create_based_vars_in_expression func_expr in
        expression_list := based_func_expr :: !expression_list;
        let normalized_exprs = normalize_exprs_with_theories (!expression_list) in
        (*(match bbcall.Translate.bbcall_fn with
        | (s, i1, i2) -> 
            (*let new_name_of_fun = s ^ "%" ^ (string_of_int i1) ^ 
                                  "%" ^ (string_of_int i2) in*)
            (* debug: *)
            (*debug ("name_of_fun: " ^ (Form.term_to_string name_of_fun) ^ 
              " " ^ (new_name_of_fun));
            read_int();*)
            (*let expr_of_name_of_fun = Variable new_name_of_fun in*)
            (* 20090627. bbcall_fn maybe null, therefor get name_of_fun *)
        );*)
        (* return *)
        (*CFA_CALL !expression_list*)
        let agent_fspec =
            match bbcall.Translate.bbcall_spec with
            | None -> None
            | Some (Sexpr.SPEC_FUN fspecs) ->  Some (get_agent_of_fspecs fspecs)
            | _ -> assert false
        in
        CFA_CALL (normalized_exprs, agent_fspec)
      end
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and transform_bbassume_to_cfa_node_stmt (bbassume : Translate.bbassume_t)
: cfa_stmt =
begin
    let pred_of_bbassume : predicate = 
        Form.form_to_predicate (bbassume.Translate.bbassume) in
    let based_pred_of_bbassume : predicate = 
        create_based_vars_in_predicate (*true*) (pred_of_bbassume) in
    let normalized_pred_of_bbassume : predicate = 
        normalize_pred_with_theories (based_pred_of_bbassume) in
    CFA_ASSUME (normalized_pred_of_bbassume)
    (*CFA_ASSUME (based_pred_of_bbassume)*)
    (*CFA_ASSUME (pred_of_bbassume)*)
end
        
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and transform_bbassign_to_cfa_node_stmt (bbassign : Translate.bbassign_t)
: cfa_stmt =
begin
    (* Transform to term *) 
    let term_vars : Form.term_t list = 
        List.map (Translate.translate_to_term) (bbassign.Translate.bbassign)
    in
    (* Transform to expression *)
    let expr_vars : expression list = List.map (Form.term_to_expr) (term_vars) in
    (* Create based expression vars *)
    let based_expr_vars : expression list ref = ref [] in
    List.iter
        (fun an_expr_var -> 
            let new_expr_var : expression = 
                create_based_vars_in_expression (*true*) (an_expr_var) in
            let normalized_expr_var : expression = 
                normalize_expr_with_theories (new_expr_var) in
            based_expr_vars := !based_expr_vars @ [normalized_expr_var]
            (*based_expr_vars := !based_expr_vars @ [new_expr_var]*)
        )
        expr_vars;
    (* Return *)
    CFA_RESET (!based_expr_vars)
    (*CFA_RESET (expr_vars)*)
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and transform_fblk_form_to_cfa_node_stmt (bb_body : Translate.bb_body_t)
                                         (bb_label : string) 
                                         (error_label : string) 
: cfa_stmt =
begin
    (* Step1: check if bb_label is error label or not  *)
    if (String.compare (bb_label) (error_label)) = 0 then CFA_ERROR
    else begin
        (* Step2: transform bb_body *)
        match bb_body with
        | Translate.BBHEAD bbhead -> CFA_ENTRY True
  	    | Translate.BBTAIL bbtail -> CFA_RETURN
        | Translate.BBSTMT bbstmt -> transform_bbstmt_to_cfa_node_stmt (bbstmt)
	      | Translate.BBCOND bbcond -> transform_bbcond_to_cfa_node_stmt (bbcond)
	      | Translate.BBJUNC bbjunc -> begin
            (* don't use phis, therefor return CFA_FORM true *)
            CFA_FORM True
          end
  	    | Translate.BBCALL bbcall -> transform_bbcall_to_cfa_node_stmt (bbcall)
        | Translate.BBASSUME bbassume -> 
            transform_bbassume_to_cfa_node_stmt (bbassume)
        | Translate.BBASSIGN bbassign -> 
            transform_bbassign_to_cfa_node_stmt (bbassign)
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and transform_bb_memo_to_cfa_memo (bb_memo : Cexpr.stmt_t list)
: (string * string * string * string * string) list =
begin
    let new_cfa_memo : (string * string * string * string * string) list ref = ref [] in
    List.iter
        (fun a_bb_memo -> 
             match a_bb_memo with
             | Cexpr.NOTE n -> begin
                 (match n with
                 | Cexpr.LINE (b, fn, real_i, cpp_i) -> 
                     new_cfa_memo := !new_cfa_memo @ [fn, (string_of_int real_i), "", "", ""]
                 | Cexpr.RANGE (fn, stl, stp, endl, endp) -> 
                     new_cfa_memo := !new_cfa_memo @ [fn, string_of_int stl, string_of_int stp,
                                                          string_of_int endl, string_of_int endp]) 
               end 
             | _ -> assert false
        )
        bb_memo;
    (* Return *)
    !new_cfa_memo
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and transform_basic_block_to_cfa_node (bbt : Translate.form_block_t)
                                      (error_label : string) 
: cfa_node =
begin
    let a_cfa_node : cfa_node = {
        cfa_label = bbt.Translate.bb_label;
        cfa_outgoing_labels = bbt.Translate.bb_outgoing_labels;
        cfa_incoming_labels = bbt.Translate.bb_incoming_labels;
        cfa_memo = (*bbt.Translate.bb_memo;*)
            transform_bb_memo_to_cfa_memo (bbt.Translate.bb_memo);
        cfa_node_stmt =
            transform_fblk_form_to_cfa_node_stmt (bbt.Translate.bb_body) 
              (bbt.Translate.bb_label) error_label;
        cfa_inputs =
            (let new_inputs = ref [] in
            Csyntax.StringInt2Set.iter
                (fun (name, level, index) ->
                    let string_of_level = string_of_int level in
                    new_inputs := !new_inputs @ [name ^ "%" ^ string_of_level]
                )
                bbt.Translate.bb_inputs;
            !new_inputs);
        cfa_outputs =
            (let new_outputs = ref [] in
            Csyntax.StringInt2Set.iter
                (fun (name, level, index) ->
                    (*let string_of_level = string_of_int level in*)
                    new_outputs := !new_outputs @ [name ^ "%0"(* ^ string_of_level*)]
                )
                bbt.Translate.bb_outputs;
            (if !new_outputs = [] then (
                match bbt.Translate.bb_body with
                | Translate.BBCALL _ -> (
                    let make_unique_name () : string =
                        "_V" ^ (string_of_int (Random.bits ()))
                    in
                    new_outputs := [(make_unique_name ()) ^ "%0"]
                  )
                | _ -> ()
            )
            else ());
            !new_outputs);
    } in 
    (* return *)
    a_cfa_node
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let transform_attribute_t_to_cfa_attribute_t 
      (attributes : Cexpr.attribute_t list) 
: cfa_attribute_t list =
    begin
        let new_cfa_attributes = ref [] in
        List.iter
            (fun a_attri ->
                let a_cfa_attri = (match a_attri with
                                  | Cexpr.NO_ATTR -> CFA_NO_ATTR
                                  | Cexpr.NORETURN_ATTR -> CFA_NORETURN_ATTR
                                  | Cexpr.CONST_ATTR -> CFA_CONST_ATTR
                                  | Cexpr.NOTHROW_ATTR -> CFA_NOTHROW_ATTR)
                in
                new_cfa_attributes := !new_cfa_attributes @ [a_cfa_attri]
            )
            attributes;
        !new_cfa_attributes
    end*)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let concat_level_for_gvars_in_cfa_inits (pred_list : predicate list) 
: predicate list = 
    begin
        let new_pred_list = ref [] in
        List.iter 
            (fun a_pred -> 
                let a_new_pred = concat_level_for_gvars_in_predicate (a_pred) in
                new_pred_list := !new_pred_list @ [a_new_pred]
            )
            pred_list;
        (* return *)
        !new_pred_list
    end*)

and translate_stmts_to_nodes (stmts : Cexpr.stmt_t list)
: (string, cfa_node) Hashtbl.t = (
    match stmts with
    | [] -> Hashtbl.create 0
    | _ -> (
        let make_unique_name () : string =
            "_V" ^ (string_of_int (Random.bits ()))
        in
        let body = Cexpr.BEGIN stmts in
        let header = { 
                       Cexpr.decl_id = (make_unique_name (), 0);
                       Cexpr.decl_tn = (Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, None));
                       Cexpr.decl_scls = Cexpr.NILSCLS ;
                       Cexpr.decl_inits = [];
                       Cexpr.decl_params = [];
                       Cexpr.decl_struct_defined_p = false;
                       Cexpr.decl_location = None;
                       Cexpr.decl_specs = None
                     } in
        let stmt = Cexpr.DEFUN (header, body) in
        let defs : (string, Translate.definition_t) Hashtbl.t = (Hashtbl.create 10) 
        in
        Interp.parsed_definitions := (Some defs) ;
        Interp.translate_specs stmt;
        let new_cfa_node_hashtbl = Hashtbl.create 1 in
        Hashtbl.iter
            (fun _ def ->
                Translate.dump_form_blocks def;
                match def with
                | Translate.FUNDEF fundef -> (
                    Hashtbl.iter
                        (fun label a_basic_block ->
                            let new_cfa_node =
                                transform_basic_block_to_cfa_node a_basic_block
                                                                  "__ERROR"
                            in
                            Hashtbl.add (new_cfa_node_hashtbl)
                                        (label)
                                        (new_cfa_node))
                        fundef.Translate.fundef_bbset)
                | _ -> assert false)
            defs;
        new_cfa_node_hashtbl
    )
)

and get_agent_of_full_expr (full_expr : Cexpr.full_expr_t)
: agent_expr_t = (
    let make_unique_name () : string =
        "_V" ^ (string_of_int (Random.bits ()))
    in
    let expr = full_expr.Cexpr.expr_node in
    let temp_var_name = (Cexpr.gensym "V") in
    let temp_var = (Cexpr.VREF (Cexpr.INTEGER, temp_var_name, 3, 0)) in
    let defvar = {
                   Cexpr.decl_id = (Cexpr.expr_to_string temp_var, 3);
                   Cexpr.decl_tn = Cexpr.INTEGER;
                   Cexpr.decl_scls = Cexpr.NILSCLS;
                   Cexpr.decl_inits =
                       [Cexpr.SETF (Cexpr.INTEGER, Cexpr.SET, temp_var, expr)];
                   Cexpr.decl_params = [];
                   Cexpr.decl_struct_defined_p = false ;
                   Cexpr.decl_specs = None;
                   Cexpr.decl_location = None;
                } in
    let new_cfa_node_hashtbl =
        translate_stmts_to_nodes [Cexpr.DEFVAR defvar] in
    let result =
    { 
      agent_assign_nodes = new_cfa_node_hashtbl;
      agent_temp_varid = temp_var_name ^ "%3";
      agent_expr_loc = full_expr.Cexpr.expr_loc;
    } in
    result
)

and get_agent_of_full_exprs (full_exprs : Cexpr.full_expr_t list)
: agent_expr_t list = (
    List.map get_agent_of_full_expr full_exprs;
)

and get_agent_of_conds (conds : Cexpr.full_expr_t Sexpr.func_condition_t list)
: agent_expr_t Sexpr.func_condition_t list = (
    List.map (fun (pos, full_expr) ->
        (pos, get_agent_of_full_expr full_expr);
    ) conds;
)

and get_agent_of_zone (zone : Cexpr.full_expr_t Sexpr.zone_t)
: agent_expr_t Sexpr.zone_t  = (
    match zone with
    | Sexpr.Location full_expr ->
        let stmt = Cexpr.RESET [full_expr.Cexpr.expr_node] in
        let new_cfa_node_hashtbl =
            translate_stmts_to_nodes [stmt] in
        let agent_expr =
          { 
            agent_assign_nodes = new_cfa_node_hashtbl;
            agent_temp_varid = (Cexpr.expr_to_string full_expr.Cexpr.expr_node);
            agent_expr_loc = full_expr.Cexpr.expr_loc;
          } in
        Sexpr.Location agent_expr;
    | Sexpr.Nothing -> Sexpr.Nothing
)

and get_agent_of_zones (zones : (Cexpr.full_expr_t Sexpr.zone_t) list)
: (agent_expr_t Sexpr.zone_t) list  = (
    List.map get_agent_of_zone zones
)

and get_agent_of_assign (assign : Cexpr.full_expr_t Sexpr.assign_t)
: agent_expr_t Sexpr.assign_t = (
    let zone, zones = assign in
    (get_agent_of_zone zone, get_agent_of_zones zones)
)

and get_agent_of_assigns (assigns : (Cexpr.full_expr_t Sexpr.assign_t) list)
: (agent_expr_t Sexpr.assign_t) list = (
    List.map get_agent_of_assign assigns;
)

and get_agent_of_behavior (behav : ('pred, 'locs) Sexpr.behavior_t)
: (agent_expr_t, agent_expr_t) Sexpr.behavior_t = (
    match behav with
      { 
        Sexpr.b_name = id;
        Sexpr.b_assumes = full_exprs;
        Sexpr.b_conditions = func_conds;
        Sexpr.b_assigns = assigns;
      } ->
        {
          Sexpr.b_name = id;
          Sexpr.b_assumes = get_agent_of_full_exprs full_exprs;
          Sexpr.b_conditions = get_agent_of_conds func_conds;
          Sexpr.b_assigns = get_agent_of_assigns assigns;
        }
)

and get_agent_of_behaviors (behavs : ('pred, 'locs) Sexpr.behavior_t list)
: (agent_expr_t, agent_expr_t) Sexpr.behavior_t list = (
    List.map get_agent_of_behavior behavs;    
)

and get_agent_of_local_invs (local_invs : (Sexpr.identifier_t *
                                           Cexpr.full_expr_t) list)
: (Sexpr.identifier_t * agent_expr_t) list = (
    List.map (fun (id, full_expr) ->
                  (id, get_agent_of_full_expr full_expr))
             local_invs
)

and get_agent_of_fspecs
      (fspec : (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t)
      : agent_fspec_t * old_term_t = (
    let stmts_of_init_old_terms =
        List.map (fun (tn, expr1, expr2) ->
            let defvar = {
                           Cexpr.decl_id = (Cexpr.expr_to_string expr2, 3);
                           Cexpr.decl_tn = tn;
                           Cexpr.decl_scls = Cexpr.NILSCLS;
                           Cexpr.decl_inits =
                               [Cexpr.SETF (Cexpr.INTEGER, Cexpr.SET, expr2, expr1)];
                           Cexpr.decl_params = [];
                           Cexpr.decl_struct_defined_p = false ;
                           Cexpr.decl_specs = None;
                           Cexpr.decl_location = None;
                        } in
            (Cexpr.DEFVAR defvar)
        ) (Ccommon.get_old_term_from_function_contract fspec) in
    let nodes_of_init_old_terms =
        translate_stmts_to_nodes stmts_of_init_old_terms in
    let fspec_replaced_old_term =
        Ccommon.replace_old_term_in_function_contract fspec in
    match fspec_replaced_old_term with
      { 
        Sexpr.fspec_conditions = conds;
        Sexpr.fspec_behaviors = behavs;
        Sexpr.fspec_assigns = assigns;
        Sexpr.fspec_prohibits = prohis;
        Sexpr.fspec_local_invs = local_invs;
        Sexpr.fspec_context = context;
        Sexpr.fspec_is_handler = handler;
      } ->
        ({
           Sexpr.fspec_conditions = get_agent_of_conds conds;
           Sexpr.fspec_behaviors = get_agent_of_behaviors behavs;
           Sexpr.fspec_assigns = get_agent_of_assigns assigns;
           Sexpr.fspec_prohibits = prohis;
           Sexpr.fspec_local_invs = get_agent_of_local_invs local_invs;
           Sexpr.fspec_context = context;
           Sexpr.fspec_is_handler = handler;
         },
        nodes_of_init_old_terms))

let translate_contract_of_field
    (f_pointer_field_hashtbl : (string * string, string) Hashtbl.t)
    (contract_of_field_hashtbl : (string * string,
                                 (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t
                                     * Cexpr.expr_t list) Hashtbl.t)
: (string * string, (agent_fspec_t * old_term_t) * string list) Hashtbl.t = (
    let result : (string * string, (agent_fspec_t * old_term_t) * string list) Hashtbl.t =
        Hashtbl.create 2 in
    Hashtbl.iter (fun (struct_var_name, field_name) struct_name ->
        let fspecs = Hashtbl.find_all contract_of_field_hashtbl
                                      (struct_name, field_name)
        in
        List.iter (fun fspec ->
            let agent_fspec = get_agent_of_fspecs (fst fspec) in
            let params = List.map Ccommon.get_name_of_expr (snd fspec) in
            Hashtbl.add result
                        (struct_var_name, field_name)
                        (agent_fspec, params)
        ) fspecs
    ) f_pointer_field_hashtbl;
    result
)

let replace_node_incomings (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (label : string) (replace_label : string)
                           (insert_labels : string list)
: unit = (
    let node = try Hashtbl.find cfa_node_hash label
               with Not_found -> assert false in
    let incomings = node.cfa_incoming_labels in
    let new_incomings = List.filter (
                             fun incoming ->
                                 incoming <> replace_label
                         ) incomings in
    let new_incomings = new_incomings @ insert_labels in
    let new_node = {node with cfa_incoming_labels = new_incomings} in
    Hashtbl.replace cfa_node_hash label new_node
)

let replace_nodes_incomings (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                            (labels : string list) (replace_label : string)
                            (insert_labels : string list)
: unit = (
    List.iter (
        fun label ->
            replace_node_incomings cfa_node_hash label replace_label
                                   insert_labels
    ) labels;
)

let replace_node_outgoings (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                           (label : string) (replace_label: string)
                           (insert_labels : string list)
: unit = (
    let node = try Hashtbl.find cfa_node_hash label
               with Not_found -> assert false in
    let outgoings = node.cfa_outgoing_labels in
    let new_outgoings =
        List.filter (
            fun outgoing ->
                outgoing <> replace_label
        ) outgoings in
    let new_outgoings = new_outgoings @ insert_labels in
    let new_node = {node with cfa_outgoing_labels = new_outgoings} in
    Hashtbl.replace cfa_node_hash label new_node
)

let replace_nodes_outgoings (cfa_node_hash : (string, cfa_node) Hashtbl.t)
                            (labels : string list) (replace_label: string)
                            (insert_labels : string list)
: unit = (
    List.iter (
        fun label ->
            replace_node_outgoings cfa_node_hash label replace_label
                                   insert_labels
    ) labels;
)

(** insert sub_cfa to after behind_node in main_cfa *)
let inline_insert_cfa_hash (sub_cfa : (string, cfa_node) Hashtbl.t)
                           (main_cfa : (string, cfa_node) Hashtbl.t)
                           (behind_node : cfa_node)
: (string, cfa_node) Hashtbl.t = (
    let outgoings_of_behind_node = behind_node.cfa_outgoing_labels in
    let sub_cfa_nodes = Hashtbl.copy sub_cfa in
    let main_cfa_nodes = Hashtbl.copy main_cfa in
    let entry_node_name, return_node_name =
        Hashtbl.fold (fun node_name node (entry_node_name, return_node_name) ->
            if ExtString.String.starts_with node_name "#Lentry" then
                (node_name, return_node_name)
            else if ExtString.String.starts_with node_name "#Lreturn" then
                (entry_node_name, node_name)
            else (entry_node_name, return_node_name)
        ) sub_cfa_nodes ("", "") in
    let sub_entry_node = try Hashtbl.find sub_cfa_nodes entry_node_name
                          with Not_found -> assert false in
    let sub_return_node = try Hashtbl.find sub_cfa_nodes return_node_name
                          with Not_found -> assert false in
    let outgoings_of_sub_entry = sub_entry_node.cfa_outgoing_labels in
    let incomings_of_sub_return = sub_return_node.cfa_incoming_labels in
    (* replace incomings of nodes pointed by sub entry node *)
    replace_nodes_incomings sub_cfa_nodes outgoings_of_sub_entry entry_node_name
                            [behind_node.cfa_label];
    (* replace outgoings of sub return node *)
    replace_nodes_outgoings sub_cfa_nodes incomings_of_sub_return return_node_name
                            outgoings_of_behind_node;
    (* replace incomings of nodes pointed by behind_node *)
    replace_nodes_incomings main_cfa_nodes outgoings_of_behind_node
                            behind_node.cfa_label incomings_of_sub_return;
    (* replace outgoings of behind_node *)
    let new_behind_node =
        { behind_node with cfa_outgoing_labels = outgoings_of_sub_entry } in
    Hashtbl.replace main_cfa_nodes behind_node.cfa_label new_behind_node;
    (* remove entry and return node of sub cfa *)
    Hashtbl.remove sub_cfa_nodes entry_node_name;
    Hashtbl.remove sub_cfa_nodes return_node_name;
    Hashtbl.iter (
        fun node_name a_node ->
            Hashtbl.add main_cfa_nodes node_name a_node
    ) sub_cfa_nodes;
    main_cfa_nodes
)

(** insert sub_cfa to after behind_node has name is behind_node_name in
    cfa *)
let inline_insert_cfa (sub_cfa : (string, cfa_node) Hashtbl.t)
                      (cfa : cfa) (behind_node_name : string)
: cfa = (
    match cfa with
    | CFA_FUNDEF defun -> (
        let behind_node = find_node_in_cfa cfa behind_node_name in
        let cfa_nodes = defun.cfa_node_hashtbl in
        let new_cfa_nodes = inline_insert_cfa_hash sub_cfa cfa_nodes behind_node in
        let new_defun = {defun with cfa_node_hashtbl = new_cfa_nodes} in
        CFA_FUNDEF new_defun
      )
    | _ -> assert false
)

(** divide function conditions into preconditions and postcondition *)
let divide_conds (conds : agent_expr_t Sexpr.func_condition_t list)
: agent_expr_t list * (* preconditions *)
  agent_expr_t list (* postconditions *) = (
    let pre_conds = List.filter (
                        fun (pos, _) ->
                            match pos with
                            | Sexpr.PRE_CONDITION -> true
                            | _ -> false
                    ) conds in
    let post_conds = List.filter (
                         fun (pos, _) ->
                             match pos with
                             | Sexpr.POST_CONDITION true -> true
                             | _ -> false
                     ) conds in
    (List.map snd pre_conds, List.map snd post_conds)
)

(** divide behavior into precondition and postcondition *)
let divide_behavior (behav : ('pred, 'locs) Sexpr.behavior_t)
: agent_expr_t list * (* precondition *)
  agent_expr_t list (* postcondition *) = (
    match behav with
      { Sexpr.b_name = _;
        Sexpr.b_assumes = agent_exprs;
        Sexpr.b_conditions = func_conds;
        Sexpr.b_assigns = _;
      } -> ([], []))

(** divide behaviors into preconditions and postconditions *)
let divide_behaviors (behavs : ('pred, 'locs) Sexpr.behavior_t list)
: agent_expr_t list *
  agent_expr_t list = (
    List.fold_left (
        fun (pre_conds, post_conds) behav ->
            let pre_conds1, post_conds1 = divide_behavior behav in
            (pre_conds @ pre_conds1, post_conds @ post_conds1)
    )([], []) behavs;
)

(** get resets in assigns *)
let divide_assigns (assigns : agent_expr_t Sexpr.assign_t list)
: agent_expr_t list = (
    List.fold_left (
        fun agent_exprs (zone, zones) ->
            match zone with
            | Sexpr.Location agent_expr -> agent_exprs @ [agent_expr]
            | Sexpr.Nothing -> agent_exprs
    ) [] assigns;
)

(** divide agent_fspec into preconditions, resets and postconditions *)
let divide_fspec (agent_fspec : agent_fspec_t)
: agent_expr_t list * (* asserts *)
  agent_expr_t list * (* resets *)
  agent_expr_t list (* ensures *) = (
    match agent_fspec with
      {
        Sexpr.fspec_conditions = conds;
        Sexpr.fspec_behaviors = behavs;
        Sexpr.fspec_assigns = assigns;
        Sexpr.fspec_prohibits = prohis;
        Sexpr.fspec_local_invs = local_invs;
        Sexpr.fspec_context = context;
        Sexpr.fspec_is_handler = handler;
      } ->
    let pre_conds, post_conds = divide_conds conds in
    let b_pre_conds, b_post_conds = divide_behaviors behavs in
    (pre_conds @ b_pre_conds, divide_assigns assigns, post_conds @ b_post_conds)
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let transform_declaration_t_to_cfa_meta_data (decl : Cexpr.declaration_t)
: cfa_meta_data_t =
begin
    (* Add level to decl_params *)
    let new_cfa_params = ref [] in
    let (_, level_of_func) = decl.Cexpr.decl_id in
    let string_of_level_of_params = string_of_int (level_of_func + 1) in
    List.iter
        (fun a_decl_param ->
            let a_cfa_param = a_decl_param ^ "%" ^ string_of_level_of_params in
            new_cfa_params := !new_cfa_params @ [a_cfa_param]
        )
        decl.Cexpr.decl_params;
    (*let new_cfa_attri =
        transform_attribute_t_to_cfa_attribute_t (decl.decl_attributes) in*)
          
    (* 05.1.2008. transform decl_inits to expression *)
    let new_cfa_inits : predicate list ref = ref [] in
    List.iter
        (fun (a_decl_init : Cexpr.expr_t)-> 
            (* 20.02.2009. Xuan has just added to get inits for gvars *)
            (match a_decl_init with
            | (Cexpr.PROGN (tn, stmt_list)) -> begin
                let form_list = 
                    List.map (Translate.translate_stmt_simple) (stmt_list) in
                let predicate_list = List.map (Form.form_to_predicate) (form_list) in
                let based_predicate_list : predicate list = 
                    List.map create_based_vars_in_predicate (*true*) (predicate_list)
                in
                let normalized_preds : predicate list = 
                    normalize_preds_with_theories (based_predicate_list) in
                let inits_to_add : predicate list = 
                    (* 20090706.Temporarily don't concat level for gvars there *)
                    (*List.map (concat_level_for_gvars_in_predicate) 
                             (based_predicate_list)*)
                    normalized_preds
                in
                new_cfa_inits := inits_to_add @ !new_cfa_inits
              end
            | anything_else -> assert false);
        )
        decl.Cexpr.decl_inits;
    (* get location *)
    let loc = (
        match decl.Cexpr.decl_location with
        | None -> None
        | Some location -> (
            let fn, ln, cn = location in
            Some (Cexpr.LINE (true, fn, ln, cn))
        )
    ) in
    (* transform *)
    let agent_fspecs =
        match decl.Cexpr.decl_specs with
        | None -> None
        | Some (Sexpr.SPEC_FUN fspecs) -> (
            let agent_fspecs, old_terms = get_agent_of_fspecs fspecs in
            Some (agent_fspecs, old_terms))
        | Some _ -> (* assert false *) None
    in
    let new_cfa_meta_data : cfa_meta_data_t = {
        cfa_id = fst decl.Cexpr.decl_id;
        cfa_tn = decl.Cexpr.decl_tn;
        cfa_scls = decl.Cexpr.decl_scls;
        cfa_inits = !new_cfa_inits;
        cfa_params = !new_cfa_params; (* da chuyen param x thanh x%i *)
        cfa_struct_defined_p = decl.Cexpr.decl_struct_defined_p;
        (*cfa_attributes = new_cfa_attri;*)
	cfa_location = loc;
        cfa_specs = agent_fspecs;
        cfa_container = None;
        cfa_is_cutentry = false
    } in
    (* return *)
    new_cfa_meta_data
end




(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let transform_definition_t_to_cfa (def_t : Translate.definition_t)
                                  (error_label : string)
                                  : string * cfa =
begin
    match def_t with
    | Translate.FUNDEF fundef ->
        let fundef_declaration = fundef.Translate.fundef_declaration in
        let new_cfa_meta_data = 
            transform_declaration_t_to_cfa_meta_data fundef_declaration
        in
        (* transform blocks to cfa_nodes *)
        let new_cfa_node_hashtbl =
            Hashtbl.create (Hashtbl.length (fundef.Translate.fundef_bbset)) in
        Hashtbl.iter
            (fun label a_basic_block ->
                let new_cfa_node =
                    transform_basic_block_to_cfa_node a_basic_block error_label 
                in
                Hashtbl.add (new_cfa_node_hashtbl) (label) (new_cfa_node)
            )
            fundef.Translate.fundef_bbset;
        (* create cfa_funcdef *)
        let new_cfa_fundef_t : cfa_fundef_t = {
            cfa_entry_label = "#Lentry";
            cfa_f_meta_data = new_cfa_meta_data;
            cfa_node_hashtbl = new_cfa_node_hashtbl;
        } in
        let new_cfa_fundef = CFA_FUNDEF new_cfa_fundef_t in
        (* change labels: "#:entry" & "#:return" of new_cfa_fundef *)
        let new_cfa = 
            change_entry_and_return_labels_in_cfa_function (new_cfa_fundef) in
        let (new_name_of_cfa, new_level_of_cfa) = 
            fundef.Translate.fundef_declaration.Cexpr.decl_id in
        (* return *)
        (new_name_of_cfa, new_cfa)
        
    | Translate.VARDEF vardef ->
        let vardef_declaration = vardef.Translate.vardef_declaration in
        let new_cfa_meta_data =
            transform_declaration_t_to_cfa_meta_data vardef_declaration
        in
        (* create cfa_vardef *)
        let new_cfa_vardef_t : cfa_vardef_t = {
            cfa_v_meta_data = new_cfa_meta_data;
        } in
        let new_cfa_vardef = CFA_VARDEF (new_cfa_vardef_t) in
        let (new_name_of_cfa, new_level_of_cfa) = 
          vardef.Translate.vardef_declaration.Cexpr.decl_id in
        (new_name_of_cfa, new_cfa_vardef)
        
    | Translate.OTHERDEF stmt -> ("", CFA_OTHERDEF)
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used when concat-function need to check if a variables *)
(* is local or global variable, then only concatenate for local variable   *)
let find_level_of_variable (var : string) : int =
begin
    let left_index = try String.rindex (var) ('%')
                     with
                     | Not_found -> -1 
    in
    if left_index = -1 then 100
    else begin
  	    let right_index = String.length (var) in
	      let string_of_level =
	          try String.sub (var) (left_index + 1) (right_index - left_index - 1)
	          with
	          | Invalid_argument i -> assert false
	      in
  	    (* return *)
	      try int_of_string (string_of_level)
	      with
	      | Failure something -> 
	          begin
	              (* debug *)
	              (* emerg (string_of_level);
	              assert false*) 0
	          end
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* If level-of-variable > level-of-function, then it is able to concat     *)
(* NOTE: do not consider nested-function at the moment                     *)
let is_able_to_concat (var : string) (level_of_function : int) : bool =
    begin
        if (String.compare (var) ("#ACC%0")) = 0 then true
        else begin
            let level_of_var : int = find_level_of_variable (var) in
            if level_of_var > level_of_function then true
            else false
        end
    end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let concat_tail_with_a_symbol (s : symbol) 
                              (new_tail_of_var_name : string)
                              (level_of_function : int) 
: symbol =
    begin
        if (is_able_to_concat (s) (level_of_function)) = true then
            s ^ new_tail_of_var_name
        else s
    end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec concat_tail_with_var_name_in_expression (expr : expression)
                                                (new_tail : string) 
                                                (func_level : int) 
: expression =
begin
    match expr with
    | Constant c -> Constant c
    | Variable sym ->
        let new_sym = concat_tail_with_a_symbol (sym) (new_tail) (func_level) in
        Variable (new_sym)
    | Application ("sref", [str_expr; Variable field_var]) -> 
        let new_str_expr : expression = 
            concat_tail_with_var_name_in_expression (str_expr) (new_tail) 
              (func_level) in
        Application ("sref", [new_str_expr; Variable field_var])
    | Application ("srefp", [pstr_expr; Variable field_var]) -> 
        let new_pstr_expr : expression = 
            concat_tail_with_var_name_in_expression (pstr_expr) (new_tail) 
              (func_level) in
        Application ("srefp", [new_pstr_expr; Variable field_var])
    | Application (sym, expr_list) ->
        let new_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let a_new_expr : expression = 
                    concat_tail_with_var_name_in_expression (an_expr) (new_tail) 
                      (func_level) in
                new_expr_list := !new_expr_list @ [a_new_expr]
            )
            expr_list;
        Application (sym, !new_expr_list)
    | Sum expr_list ->
        let new_expr_list = ref [] in
        List.iter
            (fun an_expr ->
                let a_new_expr : expression =
                  concat_tail_with_var_name_in_expression (an_expr)
                    (new_tail) (func_level) in
                new_expr_list := !new_expr_list @ [a_new_expr]
            )
            expr_list;
        Sum !new_expr_list
    | Coeff (c, c_expr) ->
        let new_c_expr : expression =
            concat_tail_with_var_name_in_expression (c_expr) (new_tail) 
              (func_level) in
        Coeff (c, new_c_expr)
  end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec concat_tail_with_var_name_in_predicate (pred : predicate)
                                               (new_tail : string) 
                                               (func_level : int) 
: predicate =
begin
    match pred with
    | True -> True
    | False -> False
    | And pred_list ->
        let new_pred_list = ref [] in
        List.iter
            (fun a_pred ->
                let a_new_pred =
                  concat_tail_with_var_name_in_predicate (a_pred) (new_tail) 
                    (func_level) in
                new_pred_list := !new_pred_list @ [a_new_pred]
            )
            pred_list;
        And (!new_pred_list)
    | Or pred_list ->
        let new_pred_list = ref [] in
        List.iter
            (fun a_pred ->
                let a_new_pred =
                  concat_tail_with_var_name_in_predicate (a_pred) (new_tail) 
                    (func_level) in
                new_pred_list := !new_pred_list @ [a_new_pred]
            )
            pred_list;
        Or (!new_pred_list)
    | Not n_pred ->
        let new_n_pred =
            concat_tail_with_var_name_in_predicate (n_pred) (new_tail) 
              (func_level) in
        Not new_n_pred
    | Eq (expr1, expr2) ->
        let new_expr1 =
            concat_tail_with_var_name_in_expression (expr1) (new_tail)
              (func_level) in
        let new_expr2 =
            concat_tail_with_var_name_in_expression (expr2) (new_tail)
              (func_level) in
        Eq (new_expr1, new_expr2)
    | Lt (expr1, expr2) ->
        let new_expr1 =
            concat_tail_with_var_name_in_expression (expr1) (new_tail)
              (func_level) in
        let new_expr2 =
            concat_tail_with_var_name_in_expression (expr2) (new_tail)
              (func_level) in
        Lt (new_expr1, new_expr2)
    | Leq (expr1, expr2) ->
        let new_expr1 =
            concat_tail_with_var_name_in_expression (expr1) (new_tail)
              (func_level) in
        let new_expr2 =
            concat_tail_with_var_name_in_expression (expr2) (new_tail)
              (func_level) in
        Leq (new_expr1, new_expr2)
    | Atom interger -> Atom interger
  end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec concat_tail_in_agent_expr (agent_expr : agent_expr_t)
                              (new_tail : string)
                              (func_level : int)
: agent_expr_t = (
    let {
          agent_assign_nodes = cfa_nodes;
          agent_temp_varid = varid;
          agent_expr_loc = loc
        } = agent_expr in
    let new_cfa_nodes = Hashtbl.create (Hashtbl.length cfa_nodes) in
    let new_varid = concat_tail_with_a_symbol varid new_tail func_level in
    Hashtbl.iter (fun _ cfa_node ->
        let new_cfa_node = concat_tail_in_cfa_node cfa_node new_tail func_level in
        Hashtbl.add new_cfa_nodes new_cfa_node.cfa_label new_cfa_node
    ) cfa_nodes;
    {
      agent_assign_nodes = new_cfa_nodes;
      agent_temp_varid = new_varid;
      agent_expr_loc = loc
    }
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_in_zone (zone : agent_expr_t Sexpr.zone_t)
                        (new_tail : string)
                        (func_level : int)
: agent_expr_t Sexpr.zone_t = (
    match zone with
    | Sexpr.Location agent_expr ->
        Sexpr.Location (concat_tail_in_agent_expr agent_expr new_tail func_level)
    | Sexpr.Nothing -> Sexpr.Nothing
)

and concat_tail_in_behavior (behavior : (agent_expr_t,
                                         agent_expr_t) Sexpr.behavior_t)
                             (new_tail : string)
                             (func_level : int)
: (agent_expr_t, agent_expr_t) Sexpr.behavior_t = (
    let { Sexpr.b_name = id;
          Sexpr.b_assumes = agent_exprs;
          Sexpr.b_conditions = func_conds;
          Sexpr.b_assigns = assigns; } = behavior in
    let new_agent_exprs =
        List.map (fun agent_expr ->
            concat_tail_in_agent_expr agent_expr new_tail func_level
        ) agent_exprs in
    let new_func_conds =
        List.map (fun (pos, agent_expr) ->
            (pos, concat_tail_in_agent_expr agent_expr new_tail func_level)
        ) func_conds in
    let new_assigns =
        List.map (fun (zone, zones) ->
            (concat_tail_in_zone zone new_tail func_level,
             List.map (fun a_zone ->
                 concat_tail_in_zone a_zone new_tail func_level
             ) zones)
        ) assigns in
    {
      Sexpr.b_name = id;
      Sexpr.b_assumes = new_agent_exprs;
      Sexpr.b_conditions = new_func_conds;
      Sexpr.b_assigns = new_assigns; }
)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_in_agent_fspec (agent_fspec : agent_fspec_t)
                               (new_tail : string)
                               (func_level : int)
: agent_fspec_t = (
    let { Sexpr.fspec_conditions = func_conds;
          Sexpr.fspec_behaviors = behaviors;
          Sexpr.fspec_assigns = assigns;
          Sexpr.fspec_prohibits = prohibits;
          Sexpr.fspec_local_invs = local_invs;
          Sexpr.fspec_context = context;
          Sexpr.fspec_is_handler = handler } = agent_fspec in
    let new_func_conds =
        List.map (fun (pos, agent_expr) ->
            (pos, concat_tail_in_agent_expr agent_expr new_tail func_level)
        ) func_conds in
    let new_behaviors =
        List.map (fun behavior ->
            concat_tail_in_behavior behavior new_tail func_level
        ) behaviors in
    let new_assigns =
        List.map (fun (zone, zones) ->
            (concat_tail_in_zone zone new_tail func_level,
             List.map (fun a_zone ->
                 concat_tail_in_zone a_zone new_tail func_level
             ) zones)
        ) assigns in
    let new_local_invs =
        List.map (fun (id, agent_expr) ->
            (id, concat_tail_in_agent_expr agent_expr new_tail func_level)
        ) local_invs in
    {
      Sexpr.fspec_conditions = new_func_conds;
      Sexpr.fspec_behaviors = new_behaviors;
      Sexpr.fspec_assigns = new_assigns;
      Sexpr.fspec_prohibits = prohibits;
      Sexpr.fspec_local_invs = new_local_invs;
      Sexpr.fspec_context = context;
      Sexpr.fspec_is_handler = handler
    }
)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_in_agent_fspec_op (agent_fspec_op : (agent_fspec_t
                                                     * old_term_t) option)
                                  (new_tail : string)
                                  (func_level : int)
: (agent_fspec_t * old_term_t) option = (
    match agent_fspec_op with
    | None -> None
    | Some (agent_fspec, old_term) -> (
        let new_agent_fspec =
            concat_tail_in_agent_fspec agent_fspec
                                       new_tail
                                       func_level in
        let new_old_term = Hashtbl.create (Hashtbl.length old_term) in
        Hashtbl.iter (fun _ cfa_node ->
            let new_cfa_node = concat_tail_in_cfa_node cfa_node new_tail func_level in
            Hashtbl.add new_old_term new_cfa_node.cfa_label new_cfa_node
        ) old_term;
        Some (new_agent_fspec, new_old_term))
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_in_cfa_stmt (c_stmt : cfa_stmt)
                            (new_tail : string) 
                            (func_level : int) 
: cfa_stmt =
begin
    match c_stmt with
    | CFA_ENTRY pred -> 
        let new_pred : predicate =
            concat_tail_with_var_name_in_predicate (pred) (new_tail)(func_level) 
        in
        CFA_ENTRY new_pred
    | CFA_FORM pred ->
        let new_pred : predicate =
            concat_tail_with_var_name_in_predicate (pred) (new_tail)(func_level) 
        in
        CFA_FORM new_pred
    | CFA_COND (pred, s1, s2) ->
        let new_pred : predicate =
            concat_tail_with_var_name_in_predicate (pred) (new_tail)(func_level) 
        in
        let new_s1 = s1 ^ new_tail in
        let new_s2 = s2 ^ new_tail in
        CFA_COND (new_pred, new_s1, new_s2)
    | CFA_CALL (expr_list, agent_fspec_op) -> begin
        (* 20090711. Concate fucname & argus with tail --START *)
        let new_expr_list = ref [] in
        List.iter 
            (fun an_expr -> 
                let a_new_expr : expression = 
                    concat_tail_with_var_name_in_expression (an_expr)
                      (new_tail) (func_level)
                in
                new_expr_list := !new_expr_list @ [a_new_expr];
            )
            expr_list;
        CFA_CALL (!new_expr_list, agent_fspec_op)
        (* 20090711. Concate fucname & argus with tail --END *)
        (*match expr_list with
        | [] -> assert false
        | name_of_func :: argu_list -> begin
            let new_argu_list = ref [] in
            List.iter
                (fun an_expr ->
                    let a_new_expr : expression =
                        concat_tail_with_var_name_in_expression (an_expr)
                            (new_tail) (func_level) 
                    in
                    new_argu_list := !new_argu_list @ [a_new_expr]
                )
                argu_list;
            let new_expr_list = name_of_func :: !new_argu_list in
            CFA_CALL new_expr_list
          end*)
      end
    | CFA_ASSUME pred -> begin
        let new_pred : predicate =
            concat_tail_with_var_name_in_predicate (pred) (new_tail)(func_level) 
        in
        CFA_ASSUME new_pred
      end
    | CFA_RESET expr_list -> begin
        let new_expr_list = ref [] in
        List.iter
            (fun an_expr -> 
                let a_new_expr : expression = 
                    concat_tail_with_var_name_in_expression (an_expr)
                      (new_tail) (func_level)
                in
                new_expr_list := !new_expr_list @ [a_new_expr]
            )
            expr_list;
        (* return *)
        CFA_RESET !new_expr_list
      end
    | CFA_FAIL -> CFA_FAIL
    | CFA_RETURN -> CFA_RETURN
    | CFA_ERROR -> CFA_ERROR
  end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_with_var_name_in_string_list (slist : string list)
                                             (new_tail : string) 
                                             (func_level : int) 
: string list =
begin
    let new_slist = ref [] in
    List.iter
        (fun a_string ->
            let a_new_string = 
                concat_tail_with_a_symbol (a_string) (new_tail) (func_level) 
            in
            new_slist := !new_slist @ [a_new_string]
        )
        slist;
    (* return *)
    !new_slist
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_with_cfa_label (new_tail : string)
                               (clabel : string) 
: string =
begin
    (* "#:elsewhere" is a sign to know when return -> don't change it  *)
    if (String.compare (clabel) ("#Lelsewhere")) = 0 then clabel
    else clabel ^ new_tail
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and concat_tail_in_cfa_node (c_node : cfa_node)
                            (new_tail: string) 
                            (func_level : int) 
: cfa_node =
begin
    (* Step1: concat tail for some attributes *)
    let new_cfa_node_stmt =
        concat_tail_in_cfa_stmt (c_node.cfa_node_stmt) (new_tail) (func_level) in
    let new_cfa_inputs =
        concat_tail_with_var_name_in_string_list (c_node.cfa_inputs)
          (new_tail) (func_level) in
    let new_cfa_outputs =
        concat_tail_with_var_name_in_string_list (c_node.cfa_outputs)
          (new_tail) (func_level) in
    (* Step2: create new_cfa_node: *)
    let new_cfa_node : cfa_node = {
        cfa_label = concat_tail_with_cfa_label (new_tail) (c_node.cfa_label);
        cfa_outgoing_labels = List.map (concat_tail_with_cfa_label (new_tail)) 
                                       (c_node.cfa_outgoing_labels);
        cfa_incoming_labels = List.map (concat_tail_with_cfa_label (new_tail)) 
                                       (c_node.cfa_incoming_labels);
        cfa_memo = c_node.cfa_memo;
        cfa_node_stmt = new_cfa_node_stmt;
        cfa_inputs = new_cfa_inputs;
        cfa_outputs = new_cfa_outputs;
    } in
    (* Step3: return *)
    new_cfa_node
  end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let concat_tail_with_var_name_in_meta_data_of_cfa_fundef
      (meta_data : cfa_meta_data_t) (new_tail : string) : cfa_meta_data_t =
begin
    (* concat tail with cfa_params *)
    let new_cfa_params = List.map (fun a_param -> a_param ^ new_tail)
                                  (meta_data.cfa_params)
    in
    (* create new_meta_data *)
    let new_meta_data : cfa_meta_data_t = {
        cfa_id = meta_data.cfa_id ^ new_tail;
        cfa_tn = meta_data.cfa_tn;
        cfa_scls = meta_data.cfa_scls;
        cfa_inits = meta_data.cfa_inits;
        cfa_params = new_cfa_params;
        cfa_struct_defined_p = meta_data.cfa_struct_defined_p;
        (*cfa_attributes = meta_data.cfa_attributes;*)
	cfa_location = meta_data.cfa_location;
        cfa_specs = meta_data.cfa_specs;
        cfa_container = meta_data.cfa_container;
        cfa_is_cutentry = meta_data.cfa_is_cutentry
    } in
    (* return *)
    new_meta_data
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let concat_tail_with_var_name_in_cfa_function (cfa_fundef : cfa)
                                              (new_tail : string) 
                                              (func_level : int) 
: cfa =
begin
    (* debug *)
    (*debug ("concat_tail_with_var_name_in_cfa_function");
    debug new_tail;
    let _ = read_int () in*)
    
    match cfa_fundef with
    | CFA_FUNDEF func -> begin
        (* create new_cmeta_data: *)
        let new_cmeta_data =
            concat_tail_with_var_name_in_meta_data_of_cfa_fundef
              (func.cfa_f_meta_data) (new_tail) in
        (* create new cfa_node_hashtbl: *)
        let new_chashtbl = 
            Hashtbl.create (Hashtbl.length func.cfa_node_hashtbl) in
        Hashtbl.iter
            (fun a_label a_node ->
                let a_new_node =
                    concat_tail_in_cfa_node (a_node) (new_tail) (func_level) in
                Hashtbl.add (new_chashtbl) (a_new_node.cfa_label) (a_new_node)
            )
            func.cfa_node_hashtbl;
        (* create new_fundef: *)
        let new_fundef : cfa_fundef_t = {
            cfa_entry_label = func.cfa_entry_label ^ new_tail;
            cfa_f_meta_data = new_cmeta_data;
            cfa_node_hashtbl = new_chashtbl;
        } in
        (* return *)
        CFA_FUNDEF new_fundef
      end
    | anything_else -> assert false
        (*raise (ConfigUtil.Message (Logging.Error, "isn't cfa_fundef"))*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_static_funcnames_in_ofile (ofile : ofile_t) : string list = (begin
    Hashtbl.fold
       (fun cfa_name a_cfa static_funnames ->
            match a_cfa with
            | CFA_FUNDEF fundef -> (
                if fundef.cfa_f_meta_data.cfa_scls = Cexpr.STATIC then
                    cfa_name :: static_funnames
                else
                    static_funnames)
            | _ -> static_funnames)
       ofile.ofile_cfas []
end)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let concat_static_funcnames_in_normal_cfa_node (cnode : cfa_node) 
                                               (static_funcnames : string list)
                                               (new_tail : string)
                                               : cfa_node = (begin
    let new_cnode = {
        cfa_label = cnode.cfa_label;
        cfa_outgoing_labels = cnode.cfa_outgoing_labels;
        cfa_incoming_labels = cnode.cfa_incoming_labels;
        cfa_memo = cnode.cfa_memo;
        cfa_node_stmt = (
            match cnode.cfa_node_stmt with
            | CFA_CALL ((Variable funcname) :: arguments, agent_fspec_op) ->
                (* Because funname has type "name%level", 
                   we must remove level before checking *)
                let based_funcname = var_to_based_var funcname in 
                if (List.mem based_funcname static_funcnames) then
                    CFA_CALL ((Variable (based_funcname ^ new_tail ^ "%1"))
                              :: arguments, agent_fspec_op)
                else 
                    cnode.cfa_node_stmt
            | _ -> cnode.cfa_node_stmt
          );
        cfa_inputs = cnode.cfa_inputs;
        cfa_outputs = cnode.cfa_outputs;
    } in
    new_cnode
end)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let concat_static_funcnames_in_normal_cfa (a_cfa : cfa)
                                          (static_funcnames : string list)
                                          (new_tail : string)
                                          : cfa = (begin
    let new_cfa =
        match a_cfa with
        | CFA_FUNDEF func_def -> begin
            let new_func_def = {
                cfa_entry_label = func_def.cfa_entry_label;
                cfa_f_meta_data = func_def.cfa_f_meta_data;
                cfa_node_hashtbl = (
                    let new_node_hashtbl = 
                        Hashtbl.create (Hashtbl.length func_def.cfa_node_hashtbl) in
                    Hashtbl.iter 
                        (fun key cnode ->
                            let new_cnode = 
                                concat_static_funcnames_in_normal_cfa_node 
                                  cnode static_funcnames 
                                  new_tail 
                            in
                            Hashtbl.add new_node_hashtbl key new_cnode
                        )
                        func_def.cfa_node_hashtbl;
                    new_node_hashtbl
                );
            } in
            CFA_FUNDEF new_func_def
          end
        (* Else, do nothing *)
        | CFA_VARDEF _
        | CFA_OTHERDEF -> a_cfa
    in
    new_cfa
end)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_variable (var : string) 
                             (is_left_var : bool)
                             (index_hash : (expression, int) Hashtbl.t) 
                             (current_vars_in_fun : expression list ref)
: symbol =
begin
    (* check if var is in index_hash or not *)
    let index_of_var = try Hashtbl.find (index_hash) (Variable var)
                       with
                       | Not_found -> - 1
    in
    if index_of_var = - 1 then begin
        Hashtbl.add (index_hash) (Variable var) (0);
        let new_var = var ^ "%" ^ string_of_int (0) in
        current_vars_in_fun := (Variable new_var) :: !current_vars_in_fun;
        new_var
    end else begin
        if is_left_var = true then begin
            Hashtbl.remove (index_hash) (Variable var);
            Hashtbl.add (index_hash) (Variable var) (index_of_var + 1);
            let old_var = var ^ "%" ^ (string_of_int index_of_var) in
            let new_var = var ^ "%" ^ (string_of_int (index_of_var + 1)) in
            current_vars_in_fun := 
                remove_expr_from_expr_list (Variable old_var) (!current_vars_in_fun);
            current_vars_in_fun := (Variable new_var) :: !current_vars_in_fun;
            new_var
        end else begin
            var ^ "%" ^ string_of_int (index_of_var)
        end
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_pref (pref : expression) 
                         (is_left_pref : bool)
                         (index_hash : (expression, int) Hashtbl.t) 
                         (structs_hash : (expression, expression list) Hashtbl.t)
                         (current_vars_in_fun : expression list ref)
: expression =
begin
    (* Check if pref s Application ("pref", ...) or not *)
    match pref with
    | Application ("pref", expr_list) -> begin
        let index_of_pref = try Hashtbl.find (index_hash) (pref)
                            with
                            | Not_found -> - 1
        in
        if index_of_pref = - 1 then begin
            Hashtbl.add (index_hash) (pref) (0);
            Application ("pref", expr_list @ [Constant 0.0]);
        end else begin
            if is_left_pref = true then begin
                Hashtbl.remove (index_hash) (pref);
                Hashtbl.add (index_hash) (pref) (index_of_pref + 1);
                let new_index = float_of_int (index_of_pref + 1) in
                Application ("pref", expr_list @ [Constant new_index]);
            end else begin
                let new_index = float_of_int (index_of_pref) in
                Application ("pref", expr_list @ [Constant new_index]);
            end
        end
      end
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_sref (sref : expression) 
                         (is_left_sref : bool)
                         (index_hash : (expression, int) Hashtbl.t) 
                         (structs_hash : (expression, expression list) Hashtbl.t)
                         (current_vars_in_fun : expression list ref)
: expression =
begin
    (* debug *)
    (*debug ("change_index_of_sref: " ^ (string_of_expression sref));*)
    (* Check if sref s Application ("sref", ...) or not *)
    match sref with
    | Application ("sref", [sref_str; sref_field]) -> begin
        let index_of_sref = try Hashtbl.find (index_hash) (sref)
                            with
                            | Not_found -> - 1
        in
        if index_of_sref = - 1 then begin
            Hashtbl.add (index_hash) (sref) (0);
            let new_sref = 
                Application ("sref", [sref_str; sref_field; Constant 0.0]) in
            current_vars_in_fun := (new_sref) :: !current_vars_in_fun;
            new_sref
        end else begin
            if is_left_sref = true then begin
                Hashtbl.replace (index_hash) (sref) (index_of_sref + 1);
                let old_index = Constant (float_of_int index_of_sref) in
                let new_index = Constant (float_of_int (index_of_sref + 1)) in 
                let old_sref = Application ("sref", [sref_str; sref_field; old_index]) in
                let new_sref = Application ("sref", [sref_str; sref_field; new_index]) in
                current_vars_in_fun := 
                    remove_expr_from_expr_list (old_sref) (!current_vars_in_fun);
                current_vars_in_fun := (new_sref) :: !current_vars_in_fun;
                new_sref
            end else begin
                 let new_index = Constant (float_of_int (index_of_sref)) in
                 Application ("sref",[sref_str; sref_field; new_index]);
            end  
        end
      end
    | anything_else -> assert false
 end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_srefp (srefp : expression) 
                          (is_left_srefp : bool)
                          (index_hash : (expression, int) Hashtbl.t) 
                          (structs_hash:(expression, expression list) Hashtbl.t)
                          (current_vars_in_fun : expression list ref)
: expression =
begin
    (* Check if srefp s Application ("srefp", ...) or not *)
    match srefp with
    | Application ("srefp", [srefp_strp; srefp_field]) -> begin
        let index_of_srefp = try Hashtbl.find (index_hash) (srefp)
                             with
                             | Not_found -> - 1
        in
        if index_of_srefp = - 1 then begin
            Hashtbl.add (index_hash) (srefp) (0);
            Application ("srefp", [srefp_strp; srefp_field; Constant 0.0]);
        end else begin
            if is_left_srefp = true then begin
                Hashtbl.replace (index_hash) (srefp) (index_of_srefp + 1);
                let new_index = Constant (float_of_int (index_of_srefp + 1)) in 
                Application ("srefp", [srefp_strp; srefp_field; new_index]);
            end else begin
                let new_index = Constant (float_of_int (index_of_srefp)) in 
                Application ("srefp", [srefp_strp; srefp_field; new_index]);
            end
          end
        end
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used when we deal with array. We don't use it now      *)
let change_index_of_memory_expr (memory_expr : expression) 
                                (is_left_expr : bool) 
                                (index_hash : (expression, int) Hashtbl.t)
                                (structs_hash : (expression, expression list) Hashtbl.t)
                                (current_vars_in_fun : expression list ref)
: expression = 
begin
    match memory_expr with
    (*| Application ("string%", [])*)
    | Application ("memory", []) -> begin
        (* find index of memory in index_hash (if it has) *)
        let index_of_memory = try Hashtbl.find (index_hash) (memory_expr) 
                                   with
                                   | Not_found -> -1
        in
        if index_of_memory = -1 then begin
            Hashtbl.add (index_hash) (memory_expr) (0);
            Application ("memory", [Constant 0.0])
        end else begin
            (* Condition: memory isn't in left_expr *)
            if is_left_expr = true then assert false
            else begin
                Hashtbl.replace (index_hash) (memory_expr) (index_of_memory + 1);
                let new_index = Constant (float_of_int (index_of_memory + 1)) in
                Application ("memory", [new_index]);
            end
        end
      end
    | anything_else -> assert false
end
  
(* ~~~~~~~~~~~~~~~~~~~~ find index of var in int_list ~~~~~~~~~~~~~~~~~~~~ *)
(* Note: if index of expr is not found in var_to_index, then update        *)
(* information about it into var_to_index, index_to_var & int_hash         *)
let rec find_in_var_to_index (int_hash : (int, (int list) * int) Hashtbl.t)
                             (var_to_index : (expression, int) Hashtbl.t)
                             (index_to_var : (int, expression) Hashtbl.t)
                             (expr : expression) 
: int =
begin
    (* Debug *)
    (*debug ("in function find_in_var_to_index: " ^ (string_of_expression expr));*)
    (*read_int ();*)
    
    match expr with
    | Constant f -> (-1)
		| Variable var -> begin
        let index_of_var : int =
            try Hashtbl.find (var_to_index) (expr)
            with
            | Not_found -> begin
                let new_index = Hashtbl.length (int_hash) in
                Hashtbl.add (int_hash) (new_index) ([], -1);
                Hashtbl.add (var_to_index) (expr) (new_index);
                Hashtbl.add (index_to_var) (new_index) (expr);
                new_index
              end
        in
        index_of_var  
      end
    | Application ("sref", [str; field; version]) -> begin
        (* Check str: *)
        let _ = 
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var) (str) 
        in
        (* Check expr (sref): *)
        let index_of_sref : int =
            try Hashtbl.find (var_to_index) (expr)
            with
            | Not_found -> begin
                let new_index : int = Hashtbl.length (int_hash) in
                Hashtbl.add (int_hash) (new_index) ([], -1);
                Hashtbl.add (var_to_index) (expr) (new_index);
                Hashtbl.add (index_to_var) (new_index) (expr);
                new_index  
              end
        in
        (* return *)
        index_of_sref
      end
    | Application ("srefp", [strp; field; version]) -> begin
        (* Check strp: *)
        let _ = 
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var) (strp) 
        in
        (* Check expr (srefp): *)
        let index_of_srefp : int =
            try Hashtbl.find (var_to_index) (expr)
            with
            | Not_found -> begin
                let new_index : int = Hashtbl.length (int_hash) in
                Hashtbl.add (int_hash) (new_index) ([], -1);
                Hashtbl.add (var_to_index) (expr) (new_index);
                Hashtbl.add (index_to_var) (new_index) (expr);
                new_index  
              end
        in
        (* return *)
        index_of_srefp
      end
    | Application ("array", [memory_expr; index_expr]) -> begin
        (* 11.3.2009. process array *)
        (* IMPORTANT: memory_expr is not a var --> do not check it??? -> no *)
        let _ = 
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var)
              (memory_expr) 
        in
        (* Check index_expr: *)
        let _ = 
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var)
              (index_expr) 
        in
        (* Check array-element: *)
        let index_of_array_i : int =
            try Hashtbl.find (var_to_index) (expr)
            with
            | Not_found -> begin
                let new_index_of_array_i : int = Hashtbl.length (int_hash) in
                Hashtbl.add (int_hash) (new_index_of_array_i) ([], -1);
                Hashtbl.add (var_to_index) (expr) (new_index_of_array_i);
                Hashtbl.add (index_to_var) (new_index_of_array_i) (expr); 
                new_index_of_array_i  
              end
        in
        (* return *)
        index_of_array_i
      end
		| Application (name_of_app, expr_list) -> begin
        let temp_fun : expression -> int = 
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var) in
        let _ = List.map (temp_fun) (expr_list) in
        let index_of_expr : int =
            try Hashtbl.find (var_to_index) (expr)
            with
            | Not_found -> begin
                let new_index_of_expr : int = Hashtbl.length (int_hash) in
                Hashtbl.add (int_hash) (new_index_of_expr) ([], -1);
                Hashtbl.add (var_to_index) (expr) (new_index_of_expr);
                Hashtbl.add (index_to_var) (new_index_of_expr) (expr);
                new_index_of_expr
              end
        in
        (* return *)
        index_of_expr
      end
    | Sum expr_list -> begin
        let temp_fun : expression -> int =
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var) in
        let _ = List.map (temp_fun) (expr_list) in
        (* return *)
        (-1)
      end
	  | Coeff (f, expr) ->
        find_in_var_to_index (int_hash) (var_to_index) (index_to_var) (expr)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* List of applications:                                                    *)
(*     div, mod, lsh, rsh, and, ior, xor, pref, sref,                       *)
(*     srefp addrof, sizeof, constf%, string%, memory                       *)
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let check_uniqueness : expression list ref = ref []

let rec change_index_of_variables_in_application (app : expression)
      (is_left_var : bool) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)  
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: expression =
begin
    match app with
    | Application ("array", [memory_expr; index_expr]) -> begin
        let new_memory_expr : expression = 
            change_index_of_variables_in_expression (is_left_var)
              (index_hash) (int_hash) (var_to_index) (index_to_var)
              (structs_hash) (current_vars_in_fun) (memory_expr)
              (uniqueness_hash)
        in
        let new_index_expr : expression = 
            change_index_of_variables_in_expression (false) 
              (index_hash) (int_hash) (var_to_index) (index_to_var)
              (structs_hash) (current_vars_in_fun) (index_expr)
              (uniqueness_hash)
        in
        Application ("array", [new_memory_expr; new_index_expr])
        (* 09.03.2009. Temporarily deal with array *)
        (*let new_memory_expr : expression = 
            change_index_of_variables_in_expression ((*is_left_var*) false)
              (index_hash) (int_hash) (var_to_index) (index_to_var)
              (structs_hash) (current_vars_in_fun) (memory_expr)
        in
        let new_index_expr : expression = 
            change_index_of_variables_in_expression (false) 
              (index_hash) (int_hash) (var_to_index) (index_to_var)
              (structs_hash) (current_vars_in_fun) (index_expr)
        in
        (* return *)
        let new_array = Application ("array", [new_memory_expr; new_index_expr]) in
        let old_version = try Hashtbl.find (index_hash) (app)
                          with
                          | Not_found -> - 1
        in
        if old_version < 0 then begin
            Hashtbl.add (index_hash) (app) (0);
            Application ("array",[new_memory_expr; new_index_expr; Constant 0.0]);
        end else begin
            if is_left_var = true then begin
                Hashtbl.replace (index_hash) (app) (old_version + 1);
                let new_version = Constant (float_of_int (old_version + 1)) in
                Application ("array", [new_memory_expr; new_index_expr; new_version]);
            end else begin
                let new_version = Constant (float_of_int (old_version)) in
                Application ("array", [new_memory_expr; new_index_expr; new_version]);
            end
        end;*)
        (* 09.03.2009. End deal with array *)
	    end
    | Application ("memory", []) -> begin
        let new_memory_expr : expression = 
            change_index_of_memory_expr (app) (is_left_var) (index_hash)
              (structs_hash) (current_vars_in_fun)
        in
        (* return *)
        new_memory_expr
      end 
    | Application ("addrof", [var_expr]) -> begin
        (* Note that we always get addr of first-new-var (var%0) *)
        let new_var_expr : expression =
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (*(Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (Hashtbl.create (0))*) 
              (*structs_hash*) (current_vars_in_fun) (var_expr)
              (uniqueness_hash)
        in
        (* return *)
        Application ("addrof", [new_var_expr])
      end
    | Application ("pref", (pref_expr) :: temp_expr :: []) -> begin
	      let new_pref_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (pref_expr)
              (uniqueness_hash)
        in
	      let new_temp_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (temp_expr)
              (uniqueness_hash)
        in
	      let new_application = Application ("pref", [new_pref_expr;
                                                    new_temp_expr]) in
        (if (List.mem new_temp_expr !check_uniqueness) = false then
            check_uniqueness := !check_uniqueness @ [new_temp_expr]
        else ());
        (* change index & return *) 
	      change_index_of_pref (new_application) (is_left_var) (index_hash)
          (structs_hash) (current_vars_in_fun)
      end
    | Application ("sref", (sref_str) :: (sref_field) :: []) -> begin
        (* Change index of sref_str: *)
	      let new_sref_str : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (sref_str)
              (uniqueness_hash)
        in
	      let new_application = 
	          Application ("sref", (new_sref_str) :: [sref_field])
	      in
        (* Find in structs_hash: *)
	      let is_in_structs_hash = ref true in  
        let sref_fields_list : expression list = 
	          try Hashtbl.find (structs_hash) (new_sref_str) 
		        with 
		        | Not_found -> is_in_structs_hash := false; []
        in
        (* Update information into structs_hash: *)
        if (!is_in_structs_hash = true) then begin
            let is_in_fields_list : bool = 
                is_mem_of_expr_list (sref_field) (sref_fields_list) in
            if is_in_fields_list = false then begin
                Hashtbl.replace (structs_hash) (new_sref_str) 
                  (sref_field :: sref_fields_list)
            end
        end else begin
            Hashtbl.add (structs_hash) (new_sref_str) ([sref_field]);
        end;
        (* change index of sref & return *)
	      change_index_of_sref (new_application) (is_left_var) (index_hash)
          (structs_hash) (current_vars_in_fun)
      end
    | Application ("srefp", (srefp_strp) :: (srefp_field) :: []) -> begin
        (* Check srefp_strp *)
	      let new_srefp_strp : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (srefp_strp)
              (uniqueness_hash)
        in
	      let new_application =
	          Application ("srefp", (new_srefp_strp) :: [srefp_field]) in
        (* change index of srefp & return *)
	      change_index_of_srefp (new_application) (is_left_var) (index_hash)
          (structs_hash) (current_vars_in_fun)
      end
    (* Temp: anything else, only change index of arguments *)
    | Application (any_string, expr_list) -> begin
        let temp_fun (expr : expression) : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) expr
              (uniqueness_hash)
        in
        let new_expr_list = List.map (temp_fun) (expr_list) in
        (* return *)        
	      Application (any_string, new_expr_list)
      end
    | anything_else -> assert false
  end
  

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and change_index_of_variables_in_expression (is_left_var : bool) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: expression =
begin
    (* Expr is right of an assignment: *)
    if is_left_var = false then begin
        match expr with
        | Constant f -> Constant f
        | Variable s -> begin
	          let new_s = 
                change_index_of_variable (s) (false) (index_hash) 
                  (current_vars_in_fun) in
            let _ = 
                find_in_var_to_index (int_hash) (var_to_index)
                  (index_to_var) (Variable new_s) in
	          Variable new_s
          end
        | Application (s, expr_list) ->
            change_index_of_variables_in_application (expr) (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (uniqueness_hash)
        | Sum expr_list -> begin
            let temp_fun (expr : expression) : expression = 
                change_index_of_variables_in_expression (false) (index_hash) 
                  (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (expr)
                  (uniqueness_hash)
            in
            let new_expr_list = List.map (temp_fun) (expr_list) in
	          Sum new_expr_list
          end
        | Coeff (a_float, an_expr) -> begin
	          let a_new_expr : expression =
	              change_index_of_variables_in_expression (false) 
                  (index_hash) (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (an_expr)
                  (uniqueness_hash)
	          in
	          Coeff (a_float, a_new_expr)
          end
    (* Expr is left of an assignment: *)
    end else begin
        match expr with
        | Constant f -> assert false
        | Variable s ->
            let new_s = 
                change_index_of_variable (s) (true) (index_hash) 
                  (current_vars_in_fun)
            in
            Variable new_s
        | Application (s, expr_list) ->
            change_index_of_variables_in_application (expr) (true) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (uniqueness_hash)
      
        (* Followings: are not in left of an assignment *)
        | Sum expr_list -> assert false
        | Coeff (_, _) -> assert false
    end
  end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let update_elements_of_arrays_in_condition_pref
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t) 
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (pref_expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: expression (* new array_expr *) =
begin
    match pref_expr with
    | Application ("pref", [(Sum (pointer_expr :: index_exprs));
                            temp_expr;
                            version_expr]) ->
      (begin
        (* Check if pointer_expr points to a memory or not: *)
        let index_of_pointer : int = 
            try Hashtbl.find (var_to_index) (pointer_expr)
            with
            | Not_found -> -1
        in
        if index_of_pointer >= 0 then begin
            let (points_to_pointer, pointer_points_to) = 
                try Hashtbl.find (int_hash) (index_of_pointer)
                with
                | Not_found -> assert false
            in
            if pointer_points_to >= 0 then begin
                (* Find first element of memory which the pointer points to *)
                let first_element : expression = 
                    try Hashtbl.find (index_to_var) (pointer_points_to)
                    with
                    | Not_found -> assert false
                in
                (* Find memory *)
                let memory_expr = (match first_element with
                                  | Application ("array", [m; fi]) -> m
                                  | anything_else -> assert false)
                in
                (* Find old_elements of the memory_expr *)
                let based_memory_expr : expression = 
                    create_based_vars_in_expression (*false*) (memory_expr) in
                let new_memory_expr : expression = 
                    change_index_of_variables_in_expression (false) 
                      (index_hash) (int_hash) (var_to_index) (index_to_var) 
                      (structs_hash) (current_vars_in_fun) (based_memory_expr)
                      (uniqueness_hash)
                in
                let old_elements : expression list = 
                    try Hashtbl.find (elements_of_arrays) (based_memory_expr)
                    with
                    | Not_found -> []
                in
                let i = match index_exprs with
                        | [index_expr] -> index_expr
                        | _ -> Sum index_exprs
                in 
                (* Update information *)
                (if old_elements = [] then begin
                    Hashtbl.add (elements_of_arrays) (based_memory_expr) ([i]);
                end else begin
                    let is_mem : bool = is_mem_of_expr_list (Sum index_exprs) (old_elements) in
                    if is_mem = false then begin
                        Hashtbl.replace (elements_of_arrays) (based_memory_expr)
                          (i :: old_elements)
                    end
                end);
                (* Return *)
                Application ("array", [new_memory_expr; i])
            end else begin
                Application ("pref", [(Sum (pointer_expr :: index_exprs));
                                      temp_expr;
                                      version_expr])
            end  
        end else begin
            Application ("pref", [(Sum (pointer_expr :: index_exprs));
                                  temp_expr;
                                  version_expr])
        end;
      end)
    | anything_else -> assert false
end
 
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec update_elements_of_arrays_in_condition_expression 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t) 
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (cond_expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: expression (* new cond_expr *) = 
begin
    (* debug *)
    (*debug ("update_elements_of_arrays_in_condition_expression");
    debug (string_of_expression cond_expr);*)
    
    match cond_expr with
    | Constant f -> Constant f
    | Variable v -> Variable v
    | Application ("pref", [(Sum (pointer_expr :: index_exprs));
                            temp_expr;
                            version_expr]) -> 
        update_elements_of_arrays_in_condition_pref (index_hash) (int_hash)
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (elements_of_arrays) (cond_expr) (uniqueness_hash)
    | Application ("array", [memory_expr; index_expr]) -> begin
        let based_memory_expr : expression = 
            create_based_vars_in_expression (*false*) (memory_expr) in
        let old_elements : expression list = 
            try Hashtbl.find (elements_of_arrays) (based_memory_expr)
            with
            | Not_found -> []
        in
        (* Update information *)
        (if old_elements = [] then begin
            Hashtbl.add (elements_of_arrays) (based_memory_expr) ([index_expr])          
        end else begin
            let is_mem : bool = is_mem_of_expr_list (index_expr) (old_elements) in
            if is_mem = true then begin
                ()
            end else begin
                Hashtbl.replace (elements_of_arrays) (based_memory_expr)
                  (index_expr :: old_elements)
            end
        end);
        (* Return *)
        Application ("array", [memory_expr; index_expr])
      end
    | Application (s, expr_list) -> begin
        let temp_fun (expr : expression) : expression =
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr)
              (uniqueness_hash)
        in
        let new_expr_list : expression list = List.map (temp_fun) (expr_list) in
        Application (s, new_expr_list)
      end  
    | Sum expr_list -> begin
        let temp_fun (expr : expression) : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr)
              (uniqueness_hash)
        in
        let new_expr_list : expression list = List.map (temp_fun) (expr_list) in
        Sum new_expr_list
      end
    | Coeff (f, an_expr) -> begin
        let new_expr : expression = 
            update_elements_of_arrays_in_condition_expression  (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (an_expr)
              (uniqueness_hash)
        in
        Coeff (f, new_expr)
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec update_elements_of_arrays_in_condition_predicate 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t) 
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (cond : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate (* new cond *) = 
begin
    (* debug *)
    (*debug ("update_elements_of_arrays_in_condition_predicate");
    debug (string_of_predicate cond);*)
    match cond with
    | True -> True
    | False -> False
    | And pred_list -> begin
        let temp_fun (expr : predicate) : predicate = 
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr)
              (uniqueness_hash)
        in
        let new_pred_list : predicate list = List.map (temp_fun) (pred_list) in                
        (* Return *)
        And new_pred_list 
      end 
    | Or pred_list -> begin
        let temp_fun (pred : predicate) : predicate = 
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (pred) (uniqueness_hash)
        in
        let new_pred_list : predicate list = List.map (temp_fun) (pred_list) in 
        (* Return *)
        Or new_pred_list 
      end
    | Not a_pred -> begin
        let new_pred : predicate =
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (a_pred) (uniqueness_hash)
        in
        Not new_pred
      end
    | Eq (expr1, expr2) -> begin
        let new_expr1 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr1)
              (uniqueness_hash) in
        let new_expr2 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr2)
              (uniqueness_hash) in
        Eq (new_expr1, new_expr2)  
      end
    | Lt (expr1, expr2) -> begin
        let new_expr1 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr1)
              (uniqueness_hash) in
        let new_expr2 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr2)
              (uniqueness_hash) in
        Lt (new_expr1, new_expr2)  
      end 
    | Leq (expr1, expr2) -> begin
        let new_expr1 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr1)
              (uniqueness_hash) in
        let new_expr2 : expression = 
            update_elements_of_arrays_in_condition_expression (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (expr2)
              (uniqueness_hash) in
        Leq (new_expr1, new_expr2)  
      end
    | Atom i -> Atom i      
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_index_of_expr_in_index_hash (expr : expression)
                                     (index_hash : (expression, int) Hashtbl.t) 
: int =
begin
    (match expr with
    | Constant f -> -1
    | Variable v -> (try Hashtbl.find (index_hash) (Variable v)
                    with Not_found -> - 1)
    | Application ("pref", [e; temp_expr]) -> begin 
        let new_e : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (e)
              (Hashtbl.create 0) in
        let new_temp_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (temp_expr)
              (Hashtbl.create 0) in
        (try Hashtbl.find (index_hash) (Application ("pref", [new_e; new_temp_expr]))
        with Not_found -> -1)
      end
    | Application ("sref", [struct_expr; field_expr]) -> begin
        let new_struct_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (struct_expr)
              (Hashtbl.create 0) in
        (try Hashtbl.find (index_hash) (Application ("sref", [new_struct_expr; field_expr]))
        with Not_found -> -1)
      end
    | Application ("srefp", [pstruct_expr; field_expr]) -> begin
        let new_pstruct_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (pstruct_expr)
              (Hashtbl.create 0) in
        (try Hashtbl.find (index_hash) (Application ("srefp", [new_pstruct_expr; field_expr]))
        with Not_found -> -1)
      end
    | Application (_, _) | Sum _ | Coeff (_, _) -> -1)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec preprocess_pref_for_change_index 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pref : expression)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (is_left : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate list) (* condition *) * expression (* new pref *) = 
begin
    match pref with
    | Application ("pref", pref_expr :: temp_expr :: []) -> begin
        (* Step1. preprocess pref_expr *)
        let (conditions_1, new_pref_expr) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (pref_expr) 
              (check_pointer) (null_pointer_exceptions) (is_left)
              (uniqueness_hash)
        in
        let (conditions_2, new_temp_expr) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (temp_expr) 
              (check_pointer) (null_pointer_exceptions) (is_left)
              (uniqueness_hash)
        in
        (* Step2. change index of pref_expr *)
        let changed_pref_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (Hashtbl.create 0)
              (ref []) (new_pref_expr)
              (uniqueness_hash)
        in
        (* Step3. check in int_hash *)
        let index_of_pref_expr : int = 
            try Hashtbl.find (var_to_index) (changed_pref_expr)
            with 
            | Not_found -> (*assert false*) -1
        in
        let (points_to_pref_expr, pref_expr_points_to) = 
            try Hashtbl.find (int_hash) (index_of_pref_expr)
            with
            | Not_found -> (*assert false*) ([], -1)
        in
        (* Step4. Check if pref_expr points to anywhere or not *)
        if pref_expr_points_to >= 0 then begin
            (* Find replaced_expr: *)
            let replaced_expr : expression = 
                try Hashtbl.find (index_to_var) (pref_expr_points_to)
                with
                | Not_found -> assert false
            in
            (* Note that we always get addr of first-new-var (var%0) *)
            (* Create first-changed-index-expr *)
            let based_replaced_expr : expression = 
                create_based_vars_in_expression (replaced_expr) in
            let first_replaced_expr : expression = 
                change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                  (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                  (Hashtbl.create (0)) (ref []) (based_replaced_expr)
                  (Hashtbl.create 0)
            in
            (* Create condition: *)
            let new_condition : predicate = 
                Eq(changed_pref_expr, Application("addrof", [first_replaced_expr])) 
            in
            (* Step5. return *)
            let based_of_replaced_expr : expression = 
                create_based_vars_in_expression (*false*) (replaced_expr) in
            (new_condition :: conditions_1, based_of_replaced_expr)
        end else begin
            (* Step5. return *)
            if (check_pointer = true) && (is_left = true) then begin
                (* error ((string_of_expression pref_expr) ^ " points to nowhere");
                assert false*)
                (* Find left_of_addrof_assignment *)
                let laddrof_constraints : expression list = 
                    find_left_of_addrof_assignment (And conditions_1) in
                null_pointer_exceptions := 
                    (pref_expr, changed_pref_expr, laddrof_constraints) ::
                    !null_pointer_exceptions;
                (*raise (La_exception.Not_point_to (("", -1), 
                                                  pref_expr, 
                                                  changed_pref_expr,
                                                  lefts_of_addrof_constraints));*)
                (conditions_1, Application ("pref", [new_pref_expr;
                                                     new_temp_expr]))
            end else (conditions_1, Application("pref", [new_pref_expr;
                                                         new_temp_expr]))           
        end
      end
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and preprocess_srefp_for_change_index 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (srefp : expression)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (is_left : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate list) (* condition *) * expression (* new expr *) = 
begin
    (* debug *)
    (*debug ("preprocess_srefp_for_change_index: " ^ (string_of_expression srefp));*)
    match srefp with
    | Application ("srefp", pstruct :: field :: []) -> (begin
        (* Step1. preprocess pstruct: *)
        let (conditions_1, new_pstruct) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (pstruct) (check_pointer) 
              (null_pointer_exceptions) (is_left) (uniqueness_hash)
        in
        (* Step2. change index of new_pstruct *)
        let changed_new_pstruct : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (Hashtbl.create 0)
              (ref []) (new_pstruct)
              (uniqueness_hash)
        in
        (* Step3. check in int_hash *)
        let index_of_pstruct : int =
            try Hashtbl.find (var_to_index) (changed_new_pstruct)
            with
            | Not_found -> (*assert false*) -1
        in
        let (points_to_pstruct, pstruct_points_to) = 
            try Hashtbl.find (int_hash) (index_of_pstruct)
            with
            | Not_found -> (*assert false*) ([], -1)
        in
        (* Step4. Check if pstruct is points to anywhere or not *)
        if pstruct_points_to >= 0 then begin
            (* debug ("pstruct_points_to >= 0");*)
            (* Find replaced_expr: *)
            let replaced_expr : expression = 
                try Hashtbl.find (index_to_var) (pstruct_points_to)
                with
                | Not_found -> assert false
            in
            (* Note that we always get addr of first-new-var (var%0) *)
            (* Create first-changed-index-expr *)
            let based_replaced_expr : expression = 
                create_based_vars_in_expression (replaced_expr) in
            let first_replaced_expr : expression = 
                change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                  (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                  (Hashtbl.create (0)) (ref []) (based_replaced_expr)
                  (Hashtbl.create 0)
            in
            (* Create condition: *)
            let new_condition : predicate = 
                Eq (changed_new_pstruct, Application ("addrof",[first_replaced_expr])) 
            in
            (* Step5. return *)
            let based_of_replaced_expr : expression = 
                create_based_vars_in_expression (*false*) (replaced_expr) in
            let new_expr_of_srefp : expression = 
                Application ("sref", [based_of_replaced_expr; field]) in
            (new_condition :: conditions_1, new_expr_of_srefp)
        end else begin 
            (* debug ("pstruct_points_to < 0");
            debug ("check_pointer: " ^ (string_of_bool check_pointer));
            debug ("is_left: " ^ (string_of_bool is_left));*)
            (* Step5. return *)
            if (check_pointer = true) && (is_left = true) then begin
                (* error ((string_of_expression pstruct) ^ " points to nowhere");*)
                (*assert false*)
                (* Find left_of_addrof_assignment *)
                let laddrof_constraints : expression list = 
                    find_left_of_addrof_assignment (And conditions_1) in
                (*raise (La_exception.Not_point_to (("", -1), 
                                                  pstruct, 
                                                  changed_new_pstruct,
                                                  lefts_of_addrof_constraints));*)
                null_pointer_exceptions := 
                    (pstruct, changed_new_pstruct, laddrof_constraints) ::
                    !null_pointer_exceptions;
                (conditions_1, Application ("srefp", [new_pstruct; field]))
            end else (conditions_1, Application("srefp", new_pstruct :: [field]))
        end
      end)
    | anything_else -> assert false 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Example1:                                                               *)
(*     Before preprocessing: {*x + 1}                                      *)
(*     After preprocessing:  condition = {x = &z}                          *)
(*                           new expr = {z + 1}                            *)
(* Example2:                                                               *)
(*     Before preprocessing: {p->x + y}                                    *)
(*     After preprocessing:  condition: = {p = &s}                         *)
(*                           new assignment = {s.x + y}                    *)
and preprocess_expr_for_change_index
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (expr : expression)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (is_left : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate list) (* condition *) * expression (* new expr *) = 
begin
    (* debug *)
    (*debug ("preprocess_expr_for_change_index: " ^ (string_of_expression expr));*)
    match expr with
    | Constant f -> ([], Constant f)
    | Variable s -> ([], Variable s)
    | Application ("pref", pref_expr :: temp_expr :: []) ->
        preprocess_pref_for_change_index (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (expr) (check_pointer)
          (null_pointer_exceptions) (is_left) (uniqueness_hash)
    | Application ("srefp", pstruct :: field :: []) -> 
        preprocess_srefp_for_change_index (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (expr) (check_pointer)
          (null_pointer_exceptions) (is_left) (uniqueness_hash)
    | Application ("$function_call$", index_expr :: func_name :: parameters) -> begin
        let para_conditions : (predicate list) ref = ref [] in
        let new_parameters : expression list ref = ref [] in
        (* Foreach parameter, preprocess it *)
        List.iter
            (fun para -> 
                let (current_conditions, new_para) = 
                    preprocess_expr_for_change_index (index_hash) (int_hash)
                      (var_to_index) (index_to_var) (para) (check_pointer)
                      (null_pointer_exceptions) (true) (uniqueness_hash)
                in
                para_conditions := !para_conditions @ current_conditions;
                new_parameters := !new_parameters @ [new_para];
            )
            parameters;
        (* Preprocess func-name *)
        let (func_conditions, new_func_name) = 
            preprocess_expr_for_change_index (index_hash) (int_hash)
              (var_to_index) (index_to_var) (func_name) (check_pointer)
              (null_pointer_exceptions) (true) (uniqueness_hash) in
        (* Return *)
        let conditions = func_conditions @ !para_conditions in
        (conditions, Application ("$function_call$", index_expr :: 
                                  new_func_name :: !new_parameters))
      end
    | Application (string_app, expr_list) -> begin
        let conditions : (predicate list) ref = ref [] in
        let new_expr_list : expression list ref = ref [] in
        (* foreach expr in expr_list, preprocess it *)
        List.iter
            (fun an_expr -> 
                let (current_conditions, new_expr) = 
                    preprocess_expr_for_change_index 
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (an_expr) (check_pointer) (null_pointer_exceptions)
                      (true) (uniqueness_hash)
                in
                conditions := !conditions @ current_conditions;
                new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        (* return *)
        (!conditions, Application (string_app, !new_expr_list))
      end
    | Sum expr_list -> begin
        let conditions : (predicate list) ref = ref [] in
        let new_expr_list : expression list ref = ref [] in
        (* foreach expr in expr_list, preprocess it *)
        List.iter
            (fun an_expr -> 
                let (current_conditions, new_expr) = 
                    preprocess_expr_for_change_index 
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (an_expr) (check_pointer) (null_pointer_exceptions) 
                      (true) (uniqueness_hash)
                in
                conditions := !conditions @ current_conditions;
                new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        (* return *)
        (!conditions, Sum !new_expr_list)
      end
    | Coeff (f, an_expr) -> begin
        let (conditions, new_expr) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (an_expr) (check_pointer)
              (null_pointer_exceptions) (true) (uniqueness_hash)
        in
        (* return: *)
        (conditions, Coeff(f, new_expr))
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Example1:                                                               *)
(*     Before preprocessing: {*x = &y}                                     *)
(*     After preprocessing:  condition = {x = &z}                          *)
(*                           new assignment = {z = &y}                     *)
(* Example2:                                                               *)
(*     Before preprocessing: {p->x = &y}                                   *)
(*     After preprocessing:  condition: = {p = &s}                         *)
(*                           new assignment = {s.x = &y}                   *)
(* 20090715. NOTE. Check pointers in left & right of an assignment         *)
and preprocess_pred_for_change_index
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pred : predicate)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate list)(* condition *) * predicate (* new assignment *) = 
begin
    match pred with
    | True | False -> ([], pred)
    | And preds -> begin
        let conditions : predicate list ref = ref [] in
        let new_preds : predicate list ref = ref [] in
        List.iter
            (fun p -> 
                let (cs, np) = 
                    preprocess_pred_for_change_index (index_hash) (int_hash)
                      (var_to_index) (index_to_var) (p) (check_pointer)
                      (null_pointer_exceptions) (uniqueness_hash)
                in
                conditions := cs @ !conditions;
                new_preds := !new_preds @ [np];
            )
            preds;
        (!conditions, And !new_preds) 
      end 
    | Or preds -> begin
        let conditions : predicate list ref = ref [] in
        let new_preds : predicate list ref = ref [] in
        List.iter
            (fun p -> 
                let (cs, np) = 
                    preprocess_pred_for_change_index (index_hash) (int_hash)
                      (var_to_index) (index_to_var) (p) (check_pointer)
                      (null_pointer_exceptions) (uniqueness_hash)
                in
                conditions := cs @ !conditions;
                new_preds := !new_preds @ [np];
            )
            preds;
        (!conditions, Or !new_preds) 
      end
    | Not p -> begin
        let (cs, preprocesed_p) = 
            preprocess_pred_for_change_index (index_hash) (int_hash)
              (var_to_index) (index_to_var) (p) (check_pointer) 
              (null_pointer_exceptions) (uniqueness_hash) in
        (cs, Not preprocesed_p)
      end
    | Eq (left_expr, right_expr) -> begin
        (* debug *)
        (*debug ("preprocess eq : " ^ (string_of_predicate (pred)));*)
        (* Preprocess right_expr: *)
        let (right_conditions, new_right_expr) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (right_expr) 
              (check_pointer) (null_pointer_exceptions)(*false*) (true)
              (uniqueness_hash)
        in
        (* Preprocess left_expr: *)
        let (left_conditions, new_left_expr) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (left_expr) 
              (check_pointer) (null_pointer_exceptions) (true) (uniqueness_hash)
        in  
        (* Compute & return: *)
        let new_assignment : predicate = Eq (new_left_expr, new_right_expr) in
        let conditions : predicate list = left_conditions @ right_conditions in
        (conditions, new_assignment)
      end
    | Lt (expr1, expr2) -> begin
        (* Preprocess expr1: *)
        let (conditions_1, new_expr1) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr1) 
              (check_pointer) (null_pointer_exceptions) (*false*) (true)
              (uniqueness_hash)
        in
        (* Preprocess expr2: *)
        let (conditions_2, new_expr2) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr2) 
              (check_pointer) (null_pointer_exceptions) (true) (uniqueness_hash)
        in  
        (* Compute & return: *)
        let new_p : predicate = Lt (new_expr1, new_expr2) in
        let conditions : predicate list = conditions_1 @ conditions_2 in
        (conditions, new_p)
      end
    | Leq (expr1, expr2) -> begin
        (* Preprocess expr1: *)
        let (conditions_1, new_expr1) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr1) 
              (check_pointer) (null_pointer_exceptions) (*false*) (true)
              (uniqueness_hash)
        in
        (* Preprocess expr2: *)
        let (conditions_2, new_expr2) = 
            preprocess_expr_for_change_index (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr2) 
              (check_pointer) (null_pointer_exceptions) (true) (uniqueness_hash)
        in  
        (* Compute & return: *)
        let new_p : predicate = Leq (new_expr1, new_expr2) in
        let conditions : predicate list = conditions_1 @ conditions_2 in
        (conditions, new_p)
      end
    | Atom i -> ([], Atom i)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Note that expr has changed index. We only check pointers in sref & pref *)
(*let rec check_pointers_in_expression (int_hash : (int, (int list) * int) Hashtbl.t)
                                     (var_to_index : (expression, int) Hashtbl.t)
                                     (index_to_var : (int, expression) Hashtbl.t)
                                     (index_hash : (expression, int) Hashtbl.t)
                                     (expr : expression)
: expression =
begin
    match expr with
    | Constant _ | Variable _ -> expr
    | Application ("pref", [pointer_expr; version_expr]) -> begin
        (* Check pointer_expr *)
        let new_pointer_expr : expression = 
            check_pointers_in_expression (int_hash) (var_to_index) 
              (index_to_var) (index_hash) (pointer_expr) in
        (* Find where the pointer points to *)
        let index_of_pointer : int = try Hashtbl.find var_to_index new_pointer_expr
                                     with Not_found -> assert false
        in
        let (points_to_pointer, pointer_points_to) =
            try Hashtbl.find int_hash index_of_pointer
            with Not_found -> assert false
        in
        (* Check if the pointer points to smwhere or not *)
        (* The pointer points to nowhere *)
        (if pointer_points_to < 0 then begin
            (* error ((string_of_expression pointer_expr) ^ " points to nowhere");
            assert false;*)
            let based_pointer_expr : expression = create_based_vars_in_expression (pointer_expr) in
            (*let based_new_pointer_expr : expression = 
                create_based_vars_in_expression (new_pointer_expr) in*)
            raise (La_exception.Not_point_to (("", -1), based_pointer_expr, new_pointer_expr, []));
        (* Pointer points to somewhere *)
        end else begin
            (* Find pointed_expr to return *)
            let pointed_expr : expression = try Hashtbl.find index_to_var pointer_points_to
                                            with Not_found -> assert false
            in
            pointed_expr
        end);
      end
    | Application ("srefp", [pointer_expr; field_expr; version_expr]) -> begin
        (* Check pointer_expr *)
        let new_pointer_expr : expression = 
            check_pointers_in_expression (int_hash) (var_to_index)
              (index_to_var) (index_hash) (pointer_expr) in
        (* Find where the pointer points to *)
        let index_of_pointer : int = try Hashtbl.find var_to_index new_pointer_expr
                                     with Not_found -> assert false
        in
        let (points_to_pointer, pointer_points_to) = 
            try Hashtbl.find int_hash index_of_pointer
            with Not_found -> assert false
        in
        (* Check if the pointer is points to smwhere or not *)
        (* If the pointer points to nowhere *)
        (if pointer_points_to < 0 then begin
            (* error ((string_of_expression pointer_expr) ^ " points to nowhere");
            assert false;*)
            let based_pointer_expr : expression = create_based_vars_in_expression (pointer_expr) in
            (*let based_new_pointer_expr : expression = 
                create_based_vars_in_expression (new_pointer_expr) in*)
            raise (La_exception.Not_point_to (("", -1), based_pointer_expr, new_pointer_expr, []));
            (*raise (La_exception.Not_point_to (("", -1), pointer_expr, new_pointer_expr));*)
        (* If the pointer points to somewhere *)
        end else begin
            (* Find pointed_expr *)
            let pointed_expr : expression = try Hashtbl.find index_to_var pointer_points_to
                                            with Not_found -> assert false
            in
            (* Create based sref to return *)
            let based_pointed_expr : expression = create_based_vars_in_expression pointed_expr in
            let based_sref : expression = Application ("sref", [based_pointed_expr; field_expr]) in
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (Hashtbl.create (0))
              (ref []) (based_sref) 
        end);
      end
    | Application (s, exprs) -> begin
        let new_exprs : expression list = 
            List.map (check_pointers_in_expression int_hash var_to_index index_to_var index_hash) 
                     exprs 
        in
        Application (s, new_exprs)
      end
    | Sum exprs -> begin
        let new_exprs : expression list = 
            List.map (check_pointers_in_expression int_hash var_to_index index_to_var index_hash) 
                     exprs 
        in
        Sum new_exprs
      end
    | Coeff (f, e) -> begin
        let new_e : expression = 
            check_pointers_in_expression (int_hash) (var_to_index) (index_to_var) 
              (index_hash) (e) in
        Coeff (f, new_e)
      end
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let rec check_pointers_in_predicate (int_hash : (int, (int list) * int) Hashtbl.t)
                                    (var_to_index : (expression, int) Hashtbl.t)
                                    (index_to_var : (int, expression) Hashtbl.t)
                                    (index_hash : (expression, int) Hashtbl.t)
                                    (pred : predicate)
: unit = 
begin
    (* debug *)
    (* debug ("check_pointers_in_predicate: " ^ (string_of_predicate pred));
    let _ = read_int () in*)

    match pred with 
    | True | False -> ()
    | And preds -> 
        let _ = 
            List.map (check_pointers_in_predicate (int_hash) (var_to_index) 
                                                  (index_to_var) (index_hash))
                     preds
        in
        ()
    | Or preds ->
        let _ = 
            List.map (check_pointers_in_predicate (int_hash) (var_to_index) 
                                                  (index_to_var) (index_hash))
                     preds
        in
        ()
    | Not p -> 
        let _ = 
            check_pointers_in_predicate (int_hash) (var_to_index) (index_to_var)
              (index_hash) (p)
        in
        ()
    | Eq (expr1, expr2) | Lt (expr1, expr2) | Leq (expr1, expr2) -> begin
        let _ = 
            check_pointers_in_expression (int_hash) (var_to_index) (index_to_var)
              (index_hash) (expr1)
        in
        let _ =
            check_pointers_in_expression (int_hash) (var_to_index) (index_to_var)
              (index_hash) (expr2)
        in
        ()
      end
    | Atom i -> ()
end
*)
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_for_expr_eq_addrof_condition (an_expr : expression) 
      (addrof_expr : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (adding_pref_flag : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate =
begin
    (* Step2. change index: *)
    (*let new_addrof_expr : expression =
        change_index_of_variables_in_expression (false) (index_hash)
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (addrof_expr)
    in*)
    let first_addrof_expr : expression = 
        change_index_of_variables_in_expression (false) (*index_hash*)
          (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0))
          (Hashtbl.create (0)) (Hashtbl.create (0))
          (*structs_hash*) (current_vars_in_fun) (addrof_expr)
          (Hashtbl.create 0)
    in
    let a_new_expr : expression =
        change_index_of_variables_in_expression (false) (index_hash)
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (an_expr)
          (uniqueness_hash)
    in
    (* Step3. Update information: *)
    let index_of_an_expr : int =
        find_in_var_to_index (int_hash) (var_to_index) (index_to_var) (a_new_expr)
    in
    let index_of_addrof_expr : int =
        find_in_var_to_index (int_hash) (var_to_index) (index_to_var) 
          (first_addrof_expr)
    in
    (* Step4. Update information "an_expr points to addrof_expr" to int_hash *)
    let (points_to_an_expr, an_expr_points_to) =
        (try Hashtbl.find (int_hash) (index_of_an_expr)
        with
        | Not_found -> assert false)
    in
    let (points_to_addrof_expr, addrof_expr_points_to) =
        try Hashtbl.find (int_hash) (index_of_addrof_expr)
        with
        | Not_found -> assert false
    in
    Hashtbl.replace (int_hash) (index_of_an_expr) 
      (points_to_an_expr, index_of_addrof_expr);
    if (List.mem (index_of_an_expr) (points_to_addrof_expr)) = false then begin
        let new_points_to_addrof_expr = index_of_an_expr :: points_to_addrof_expr in 
        Hashtbl.replace (int_hash) (index_of_addrof_expr)
           (new_points_to_addrof_expr, addrof_expr_points_to);
    end;
    
    (* Step5. Check adding_pref_flag & return *)
    if adding_pref_flag = false then begin  
        (* return *)
        let new_addrof = Application ("addrof", [first_addrof_expr]) in
        (* 20090703 Create new addrof_constraint *)
        let addrof_constraint : predicate =
            Not (Eq (Application ("addrof", [first_addrof_expr]), Constant 0.0)) in
        And [addrof_constraint; Eq (a_new_expr, new_addrof)]
    end else begin
        (* Step6. change index: *)
        let new_pref_expr : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (Application ("pref", [an_expr; Application ("NIL", [])]))
              (uniqueness_hash)
        in
        (* Step7. Update information: *)
        (*let _ =
            find_in_var_to_index (int_hash) (var_to_index) (index_to_var) (new_pref_expr)
        in*)
        (* Step8. Create new expression to return*)
        let new_pref_eq : predicate = Eq (new_pref_expr, first_addrof_expr) in
        let new_addrof_eq : predicate = 
            Eq (a_new_expr, Application ("addrof", [first_addrof_expr])) in
        (* 20090703 Create new addrof_constraint *)
        let addrof_constraint : predicate =
            Not (Eq (Application ("addrof", [first_addrof_expr]), Constant 0.0)) in
        (* Step9. Return *)
        And [addrof_constraint; new_addrof_eq; new_pref_eq]
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec change_index_of_variables_in_eq_condition_predicate 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (adding_pref_flag : bool)
      (pred: predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate =
begin
    match pred with
    | Eq (Variable a, Application ("addrof", [addrof_expr])) ->
        change_index_for_expr_eq_addrof_condition (Variable a) (addrof_expr) 
          (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (adding_pref_flag) (uniqueness_hash)
    | Eq (Application ("sref", exprs), Application ("addrof", [addrof_expr])) ->
        change_index_for_expr_eq_addrof_condition (Application ("sref", exprs)) 
          (addrof_expr) (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (adding_pref_flag)
          (uniqueness_hash)
    | Eq (Application ("addrof", [addrof_expr]), Variable a) ->
        change_index_for_expr_eq_addrof_condition (Variable a) (addrof_expr)
          (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (adding_pref_flag) (uniqueness_hash)
    | Eq (Application ("addrof", [addrof_expr]), Application ("sref", exprs)) ->
        change_index_for_expr_eq_addrof_condition (Application ("sref", exprs)) 
          (addrof_expr) (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (adding_pref_flag)
          (uniqueness_hash)
    | Eq (expr1, expr2) -> begin
        let new_expr1 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (expr1)
              (uniqueness_hash)
        in
        let new_expr2 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (expr2)
              (uniqueness_hash)
        in
        (* Create new eq *)
        let new_eq : predicate = Eq (new_expr1, new_expr2) in
        (* Create constraints for eq *)
        let addrofs : expression list = find_addrofs_in_predicate new_eq in
        let addrof_constraints = List.map create_constraint_for_addrof addrofs in 
        (* Return *)
        (match (new_eq :: addrof_constraints) with
        | [an_element] -> an_element
        | _ -> And (new_eq :: addrof_constraints))
      end
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec change_index_of_variables_in_condition_predicate 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t) 
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (adding_pref_flag : bool)
      ?(negative : bool = false)
      (pred : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate =
begin
    match pred with
    | True -> True
    | False -> False
    | And pred_list -> begin
        let temp_fun (pred : predicate) : predicate = 
           change_index_of_variables_in_condition_predicate (index_hash) 
             (int_hash) (var_to_index) (index_to_var) (structs_hash) 
             (current_vars_in_fun) (elements_of_arrays) (adding_pref_flag)
             ~negative:negative pred (uniqueness_hash)
        in
        let new_pred_list = List.map (temp_fun) (pred_list) in
        And new_pred_list
      end
    | Or pred_list -> begin
        let temp_fun (pred : predicate) : predicate = 
            change_index_of_variables_in_condition_predicate (index_hash) 
             (int_hash) (var_to_index) (index_to_var) (structs_hash) 
             (current_vars_in_fun) (elements_of_arrays) (adding_pref_flag)
             ~negative:negative pred (uniqueness_hash)
        in
        let new_pred_list = List.map (temp_fun) (pred_list) in
        Or new_pred_list
      end
    | Not (Eq (Application ("unique_addr", [p]), Constant 0.)) -> (
        let new_expr1 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (p)
              (uniqueness_hash)
        in
        let new_expr2 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun)
              ((Application ("array", [Application ("memory", []); Constant 0.0])))
              (uniqueness_hash)
        in
        (* update int_hash *)
        let index1 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          new_expr1 in
        let index2 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          (new_expr2) in
        let index3 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          (Application ("addrof", [new_expr1])) in
        (try
            let indexs, a_index = Hashtbl.find int_hash index1 in
            Hashtbl.replace int_hash index1 (indexs, index2)
        with Not_found -> ());
        if not negative then (
            And [Not (Eq ((Application ("addrof", [new_expr2])), Constant 0.0));
                 (Eq (new_expr1, (Application ("addrof", [new_expr2]))));
                 Eq (Application ("pref", [new_expr1; Application ("NIL", []); Constant 0.]), new_expr2);
                 Not (Eq (Application ("unique_addr", [new_expr1]), Constant 0.))
                 ]
        )
        else Not (Eq (Application ("unique_addr", [new_expr1]), Constant 0.))
    )
    | Not (Eq (Application ("unique_block", [p; size]), Constant 0.)) -> (
        let new_expr1 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (p)
              (uniqueness_hash)
        in
        let new_expr2 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun)
              ((Application ("array", [Application ("memory", []); Constant 0.0])))
              (uniqueness_hash)
        in
        (* update int_hash *)
        let index1 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          new_expr1 in
        let index2 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          (new_expr2) in
        let index3 = find_in_var_to_index int_hash
                                          var_to_index
                                          index_to_var
                                          (Application ("addrof", [new_expr1])) in
        let new_size =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (size)
              (uniqueness_hash)
        in
        (try
            let indexs, a_index = Hashtbl.find int_hash index1 in
            Hashtbl.replace int_hash index1 (indexs, index2)
        with Not_found -> ());
        if not negative then (
            (* update uniqueness_hash *)
            Hashtbl.replace uniqueness_hash index1 new_size;
            And [Not (Eq ((Application ("addrof", [new_expr2])), Constant 0.0));
                 (Eq (new_expr1, (Application ("addrof", [new_expr2]))));
                 Eq (Application ("pref", [new_expr1; Application ("NIL", []); Constant 0.]), new_expr2);
                 Not (Eq (Application ("unique_block", [new_expr1; new_size]), Constant 0.))
                 ]
        )
        else Not (Eq (Application ("unique_block", [new_expr1; new_size]), Constant 0.))
    )
    | Not a_pred ->
        let a_new_pred : predicate =
            change_index_of_variables_in_condition_predicate (index_hash) 
              (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0)) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (adding_pref_flag) 
              ~negative:(not negative) (a_pred) (uniqueness_hash)
        in
        Not a_new_pred
    | Eq (expr1, expr2) -> begin
        let new_eq : predicate = 
            change_index_of_variables_in_eq_condition_predicate  
              (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (adding_pref_flag) (pred) (uniqueness_hash)
        in
        let updated_eq : predicate = 
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays) (new_eq)
              (uniqueness_hash)
        in
        (* return *)
        updated_eq
      end
    | Lt (expr1, expr2) ->
        let new_expr1 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0)) (structs_hash)
              (current_vars_in_fun) (expr1)
              (Hashtbl.create 0)
        and new_expr2 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0)) (structs_hash)
              (current_vars_in_fun) (expr2)
              (Hashtbl.create 0)
        in
        let updated_lt : predicate = 
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays)
              (Lt (new_expr1, new_expr2)) (uniqueness_hash)
        in
        (* Return *)
        updated_lt
    | Leq (expr1, expr2) ->
        let new_expr1 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0)) (structs_hash)
              (current_vars_in_fun) (expr1)
              (Hashtbl.create 0)
        and new_expr2 : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (Hashtbl.create(0)) (Hashtbl.create(0)) (Hashtbl.create(0)) (structs_hash)
              (current_vars_in_fun) (expr2)
              (Hashtbl.create 0)
        in
        let updated_leq : predicate = 
            update_elements_of_arrays_in_condition_predicate (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (elements_of_arrays)
              (Leq (new_expr1, new_expr2)) (uniqueness_hash)
        in
        (* Return *)
        updated_leq
    | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let change_index_of_variables_in_condition 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (structs_hash : (expression, expression list) Hashtbl.t) 
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (adding_pref_flag : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (pred : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate =
begin
    (*if check_pointer = false then begin
        change_index_of_variables_in_condition_predicate (index_hash) (int_hash) 
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (elements_of_arrays) (adding_pref_flag) (pred) 
    end else begin*)
        (* Preproces condition *)
        let (constraints, preprocessed_condition) = 
            preprocess_pred_for_change_index (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (pred) 
              (check_pointer) (null_pointer_exceptions) (uniqueness_hash)
        in 
        let normalized_preprocessed_cond = normalize preprocessed_condition in
        (* Change index of vars in the preprocessed_condition *)
        let new_preprocessed_condition : predicate = 
            change_index_of_variables_in_condition_predicate (index_hash) (int_hash) 
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
              (elements_of_arrays) (adding_pref_flag) (normalized_preprocessed_cond)
              (uniqueness_hash) 
        in
        (* Change index of vars in the pred *)
        let new_pred : predicate = 
              change_index_of_variables_in_condition_predicate(index_hash) (int_hash) 
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
              (elements_of_arrays) (adding_pref_flag) (pred)
              (uniqueness_hash)
        in
        (* Create new condtion & return *)
        if constraints != [] then begin
            (*And (new_pred :: new_preprocessed_condition :: constraints)*)
            Or [Not (And constraints); new_preprocessed_condition]
        end else begin
            new_pred
        end
    (*end*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Update information "left_expr points to right_expr" into int_hash       *)
(* Note: left_expr & right_expr have been in int_hash                      *)
let update_points_to_int_hash (int_hash : (int, (int list) * int) Hashtbl.t)
                              (var_to_index : (expression, int) Hashtbl.t)
                              (index_to_var : (int, expression) Hashtbl.t)
                              (left_var : expression) 
                              (right_var : expression) 
                              (uniqueness_hash : (int, expression) Hashtbl.t)
: unit =
begin
    try (
        (* Step1. find index of left_var: *)
        let index_of_left_var : int = Hashtbl.find (var_to_index) (left_var) in
        (* Step2. find index of right_var: *)
        let index_of_right_var : int = Hashtbl.find (var_to_index) (right_var) in
        (* Step3. Update information of left_var to int_hash: *)
        let (points_to_left_var, left_var_points_to) =
            try Hashtbl.find (int_hash) (index_of_left_var)
            with Not_found -> assert false
        in
        Hashtbl.replace (int_hash) (index_of_left_var) 
                        (points_to_left_var, index_of_right_var);
        (* Step4. Find information of right_var in int_hash: *)
        let (points_to_right_var, right_var_points_to) =
            try Hashtbl.find (int_hash) (index_of_right_var)
            with Not_found -> (*assert false*) ([], -1)
        in
        (* Step5. Check if points_to_right_var contains index_of_left_var or not: *)
        let is_mem : bool = List.mem (index_of_left_var) (points_to_right_var) in
        (* Step6. If not, add index_of_left_var to it: *)
        if is_mem = false then begin
            (* update uniqueness_hash *)
            try
                let size = Hashtbl.find uniqueness_hash index_of_right_var in
                Hashtbl.replace uniqueness_hash index_of_left_var size;
            with Not_found -> ();
            let new_points_to_right_var = index_of_left_var :: points_to_right_var in 
            Hashtbl.replace (int_hash) (index_of_right_var) 
                            (new_points_to_right_var, right_var_points_to);
        end;
    ) with Not_found -> ()
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* find all variables which are pointing to the var                        *)
let find_all_points_to_var (var : expression)
                           (int_hash : (int, (int list) * int) Hashtbl.t)
                           (var_to_index : (expression, int) Hashtbl.t)
                           (index_to_var : (int, expression) Hashtbl.t) 
: (expression list) =
begin
    (* Find index of var in int_hash *)
    let index_of_var : int = try Hashtbl.find (var_to_index) (var)
                             with Not_found -> -1
    in
    if index_of_var >= 0 then begin
        (* Find index of variables which are pointing to this var *)
        let (index_of_points_to_var, index_of_var_points_to) =
            try Hashtbl.find (int_hash) (index_of_var)
            with Not_found -> assert false
        in
        (* Find variables which are pointing to this var *)
        let list_of_points_to_var : expression list ref = ref [] in
        List.iter
            (fun an_index ->
                let a_point_to_var : expression = 
                    try Hashtbl.find (index_to_var) (an_index)
                    with Not_found -> assert false
                in
                (* Update list_of_points_to_var: *)
                list_of_points_to_var := !list_of_points_to_var @ [a_point_to_var]
            )
            index_of_points_to_var;
        (* Return *)
        !list_of_points_to_var;
    end else []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used to change index of srefp when sref changed index  *)
(* Example: when we changed index of a sref (x.f) -> we need to change     *)
(* index of srefp which points to the sref (x.f)                           *)
(* Predicate:                                                              *)
(*   {y = addrof (x)} => {<y->f, new_index_of_yf> = <x.f, new_index_of_xf} *)
(* Note: sfield has changed index                                          *)
let update_points_to_field_of_struct_in_int_hash (sfield : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
: predicate list =
begin
    (* Check sfield *)
    match sfield with
    | Application ("sref", [str; field; version]) -> begin
        (* Step1. Find index of str *)            
        let index_of_str = try Hashtbl.find (var_to_index) (str) 
                           with
                           | Not_found -> (*assert false*) -1
        in
        (* Step2. Check if have some points to str or not *)
        let (points_to_str_list, str_points_to) =
            try Hashtbl.find (int_hash) (index_of_str)
            with
            | Not_found -> (*assert false*)([], -1)
        in
        (* Step3. Create predicates *)
        if points_to_str_list = [] then []
        else begin
            (* Create list *)
            let list_of_preds = ref [] in
            (* For each pointer which point to str, create predicate: *)
            (* {y = addrof (x)} =>                                    *)
            (*   {<y->f, new_index_of_yf> = <x.f, new_index_of_xf}    *)
            List.iter
                (fun a_points_to_str -> 
                    (* Find pointer: *)
                    let pointer : expression =
                        try Hashtbl.find (index_to_var) (a_points_to_str)
                        with
                        | Not_found -> assert false
                    in
                    (* Create srefp: *)
                    let a_srefp = Application ("srefp", [pointer; field]) in
                    (* Change version cua srefp *)
                    let a_new_srefp : expression =
                        change_index_of_srefp (a_srefp) (true) (index_hash)
                          (structs_hash) (current_vars_in_fun)
                    in
                    (* create predicate {y->f = x.f} *)
                    let pred2 = Eq (a_new_srefp, sfield) in
                    if is_flatten = true then begin
                        list_of_preds := !list_of_preds @ [pred2]
                    end else begin
                        (* Create predicate {y = addrof (x)} => {y->f = x.f} *)
                        (* Note that we always get addr of first-new-var (var%0) *)
                        (* Create first-index_str *)
                        let based_str : expression = 
                            create_based_vars_in_expression (str) in
                        let first_index_str : expression = 
                            change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                              (Hashtbl.create (0)) (ref []) (based_str)
                              (Hashtbl.create 0)
                        in
                        let pred1 : predicate =
                            Eq (pointer, Application ("addrof", [first_index_str])) in
                        let implication = Or ((Not pred1) :: [pred2]) in
                        (* Add the implication to list to return *)    
                        list_of_preds := !list_of_preds @ [implication]
                    end  
                )
                points_to_str_list;
            (* Return *)
            !list_of_preds                 
        end
      end
    | anything_else -> []
end     
  
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used when a variable is changed value by an assignment *)
(* Note: old_var & new_var have changed index                              *)
let rec update_value_to_var_in_int_hash (old_var : expression)
          (new_var : expression)
          (index_hash : (expression, int) Hashtbl.t)
          (int_hash : (int, (int list) * int) Hashtbl.t)
          (var_to_index : (expression, int) Hashtbl.t)
          (index_to_var : (int, expression) Hashtbl.t) 
          (structs_hash : (expression, expression list) Hashtbl.t)
          (current_vars_in_fun : expression list ref)
          (is_flatten : bool)
: predicate list =
begin
    (* debug *)
    (*debug "\nupdate_value_to_var_in_int_hash";
    debug ("old_var: " ^ (string_of_expression old_var));
    debug ("new_var: " ^ (string_of_expression new_var));*) 
    (*let pred_of_updated_str = ref [] in*)    
    try (
        let index_of_old_var : int = Hashtbl.find (var_to_index) (old_var) in
        (* Find all variables which point to old_var *)
        let points_to_old_var : expression list =
            find_all_points_to_var (old_var) (int_hash) (var_to_index) (index_to_var)
        in
        (* Note that we always get addr of first-new-var (var%0) *)
        (* Create first-changed-index-expr *)
        let based_var : expression = 
            create_based_vars_in_expression (old_var) in
        let first_replaced_var : expression = 
            change_index_of_variables_in_expression (false) (Hashtbl.create(0))
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (based_var)
              (Hashtbl.create 0)
        in
        (* debug *)
        (* debug "all_points_to_var:";
        List.iter (fun a_p -> debug (string_of_expression a_p))
                  points_to_old_var;*)
        (*read_int();*)
      
        (* Update information for int_hash *)
        Hashtbl.replace (var_to_index) (new_var) (index_of_old_var);
        Hashtbl.replace (index_to_var) (index_of_old_var) (new_var);
    
        if points_to_old_var = [] then (*[]*) begin
            (* Create predicate: {addrof (old_var) = addrof (new_var)} *)
            (*let addrof_eq : predicate= Eq (Application ("addrof", [new_var]), 
                                           Application ("addrof", [old_var]))
            in
            [addrof_eq]*)
            []
        end else begin
		        (* Create predicate: {addrof (old_var) = addrof (new_var)} *)
    		    (*let addrof_eq : predicate= Eq (Application ("addrof", [new_var]), 
                                           Application ("addrof", [old_var]))
		        in*)
		        (* Create predicates: (a1 = addrof (i1) => i2 = pref (a1, 2) *)
            (* Or predicates {i2 = pref (a1, 2)} if is_flatten = true *)
		        let new_pred_list : predicate list ref = ref [] in
		        List.iter
		            (fun (a_points_to_old_var : expression) ->
		                let pred1 : predicate = 
                        Eq (a_points_to_old_var, 
                            Application ("addrof", [first_replaced_var (*old_var*)]))
		                in
                    let old_pref = Application ("pref", [a_points_to_old_var;
                                                         Application ("NIL", [])]) in
		                let new_pref : expression = 
		                    change_index_of_pref (old_pref) (true) (index_hash) 
                        (structs_hash) (current_vars_in_fun)
		                in
		                let pred2 : predicate = Eq (new_pref, new_var) in
                    if is_flatten = true then begin
                        new_pred_list := !new_pred_list @ [pred2]
                    end else begin
        		            let implication : predicate = Or [Not pred1; pred2] in
		                    new_pred_list := !new_pred_list @ [implication]
                    end
                )
		            points_to_old_var;
		    
		        (* return *)
            ((*addrof_eq ::*) !new_pred_list)(* @ !pred_of_updated_str*)
        end
    ) with Not_found -> []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create assignments when addrof_expr is struct                           *)
(* Example: struct s {int a; int b};                                       *)
(*          struct s x;                                                    *)
(*          struct s* y = addrof (x)                                       *)
(* ==> y->a = x.a /\ y->b = x.b                                            *)
(* Note: addrof_expr & left_expr have changed index                        *)
and addrof_expr_is_struct (addrof_expr : expression) (left_expr : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    (* Find fields of struct *)
    let fields_of_addrof_expr : expression list = 
        try Hashtbl.find (structs_hash) (addrof_expr)
        with
        | Not_found -> []
    in
    (* If we don't found any field of struct, return [] *)
    if fields_of_addrof_expr = [] then []
    (* Else, create new assignments *)
    else begin
        let preds_list_to_return : predicate list ref = ref [] in
        let based_addrof_expr : expression = 
            create_based_vars_in_expression (*false*) (addrof_expr) in
        let based_left_expr : expression =
            create_based_vars_in_expression (*false*) (left_expr) in
        List.iter
            (fun a_field -> 
                (* Create new variables: *)
                let a_based_sref : expression = 
                    Application ("sref", [based_addrof_expr; a_field]) in
                let new_sref : expression = 
                    change_index_of_variables_in_expression (false)
                      (index_hash) (int_hash) (var_to_index) (index_to_var) 
                      (structs_hash) (current_vars_in_fun) (a_based_sref)
                      (uniqueness_hash)
                in
                let a_based_srefp : expression = 
                    Application ("srefp", [based_left_expr; a_field]) in
                let new_srefp : expression = 
                    change_index_of_variables_in_expression (true)
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (structs_hash) (current_vars_in_fun) (a_based_srefp)
                      (uniqueness_hash)
                in
                (* Create new assignments: *)
                let new_assignment : predicate = Eq (new_srefp, new_sref) in
                preds_list_to_return := new_assignment :: !preds_list_to_return
            )
            fields_of_addrof_expr;
        (* return *)
        !preds_list_to_return
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used when assigns a struct to another struct (in       *)
(* function "assign_var_for_var"                                           *)
and right_var_is_struct (right_expr : expression) (left_expr : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    (* Find fields of right_expr if the right_expr has some *)
    let fields_of_right_expr : expression list = 
        try Hashtbl.find (structs_hash) (right_expr)
        with
        | Not_found -> []
    in
    (* If we don't find any field, return [] *)
    if fields_of_right_expr = [] then []
    (* Else, create new assignments *)
    else begin
        let preds_list_to_return = ref [] in
        let based_right_expr : expression = 
            create_based_vars_in_expression (*false*) (right_expr) in
        let based_left_expr : expression =
            create_based_vars_in_expression (*false*) (left_expr) in
        List.iter
            (fun a_field -> 
                (* Create right_sref *)
                let a_based_right_sref : expression = 
                    Application ("sref", [based_right_expr; a_field]) in
                let new_right_sref : expression = 
                    change_index_of_variables_in_expression (false)
                      (index_hash) (int_hash) (var_to_index) (index_to_var) 
                      (structs_hash) (current_vars_in_fun) (a_based_right_sref)
                      (uniqueness_hash)
                in
                (* Create left_sref *)
                let a_based_left_sref : expression = 
                    Application ("sref", [based_left_expr; a_field]) in
                let new_left_sref : expression = 
                    change_index_of_variables_in_expression (true)
                      (index_hash) (int_hash) (var_to_index) (index_to_var)
                      (structs_hash) (current_vars_in_fun) (a_based_left_sref)
                      (uniqueness_hash)
                in
                (* update left_sref to var_to_index *)
                let index_of_right =
                    find_in_var_to_index (int_hash) (var_to_index)
                                         (index_to_var) (new_right_sref)
                in
                Hashtbl.add (var_to_index) new_left_sref index_of_right;
                (* Create new assignment *)
                let new_assignment : predicate = 
                    Eq (new_left_sref, new_right_sref) in
                (* Update to list to return *)
                preds_list_to_return := new_assignment :: !preds_list_to_return
            )
            fields_of_right_expr;
        (* Return *)
        !preds_list_to_return
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates when assign addrof for var                            *)
and assign_addrof_for_var (addrof_expr : expression) 
      (left_expr : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* Create first-changed-index-expr *)
    let first_addrof_expr : expression = 
        change_index_of_variables_in_expression (false) (Hashtbl.create(0))
          (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
          (Hashtbl.create (0)) (ref []) (addrof_expr)
          (Hashtbl.create 0)
    in
	  (* Step1. change index of right_expr & update information: *)
	  let new_addrof_expr : expression =
	      change_index_of_variables_in_expression (false) (index_hash)
	        (int_hash) (var_to_index) (index_to_var) (structs_hash)
	        (current_vars_in_fun) (addrof_expr)
          (uniqueness_hash)
	  in
	  let _ = 
	      find_in_var_to_index (int_hash) (var_to_index) (index_to_var) 
          (new_addrof_expr)
	  in
	  (* Step2. change index of left_expr: *)
	  let old_version_of_left_expr = (*try Hashtbl.find (index_hash) (left_expr)
	                                 with
	                                 | Not_found -> - 1*)
        find_index_of_expr_in_index_hash (left_expr) (index_hash)
	  in
	  (* If left_expr hasn't appeared before *)
	  if old_version_of_left_expr = - 1 then begin
	      let new_left_expr : expression =
	          change_index_of_variables_in_expression (true) (index_hash)
	            (int_hash) (var_to_index) (index_to_var) (structs_hash)
	            (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
	      in
	      let _ = 
	          find_in_var_to_index (int_hash) (var_to_index) 
	            (index_to_var) (new_left_expr)
	      in
        (* Step3. Update information "left points to right" *)  
        update_points_to_int_hash (int_hash) (var_to_index) (index_to_var)
          (new_left_expr) (new_addrof_expr) (uniqueness_hash);
        (* Step4. Create predicates if left_expr is field of struct & *)
        (* the struct is pointed by some var                          *)
        let pred_of_updated_points_to_field : predicate list = 
            update_points_to_field_of_struct_in_int_hash (new_left_expr)
	            (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
	      in
        (* Step5. Create predicates when addrof_expr is struct *)
        let preds_of_addrof_struct : predicate list = 
            addrof_expr_is_struct (new_addrof_expr) (new_left_expr) 
              (index_hash) (int_hash) (var_to_index)  (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
              (uniqueness_hash)
        in
	      (* Step6. Create real assignment *)
	      let assign1 : predicate =
            (*Eq (new_left_expr, Application ("addrof", [new_addrof_expr])) in*)
	          Eq (new_left_expr, Application ("addrof", [first_addrof_expr])) in
        (* Get value of pref left *)
	      let new_pref : expression =
	          change_index_of_pref (Application ("pref", [new_left_expr; Application ("NIL", [])])) 
              (true) (index_hash) (structs_hash) (current_vars_in_fun)
	      in
	      let assign2 : predicate = Eq (new_pref, new_addrof_expr) in
        (* 20090703 Create constraint for addrof *)
        let addrof_constraint : predicate = 
            Not (Eq (Application ("addrof", [first_addrof_expr]), Constant 0.0)) in
            (*Not (Eq (Application ("addrof", [new_addrof_expr]), Constant 0.0)) in*)
	      (* Step7. return *)
        [addrof_constraint; assign1; assign2] @ pred_of_updated_points_to_field 
        @ preds_of_addrof_struct
	                
	  (* left_expr was in index_hash *)
	  end else begin
	      let old_left_expr : expression =
	          change_index_of_variables_in_expression (false) (index_hash)
	            (int_hash) (var_to_index) (index_to_var) (structs_hash)
	            (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
	      in
	      let new_left_expr : expression =
	          change_index_of_variables_in_expression (true) (index_hash)
	            (int_hash) (var_to_index) (index_to_var) (structs_hash)
	            (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
	      in
	      (* Step3. create predicates when left_expr changes value *)
	      let points_to_a_eq : predicate list =
	          update_value_to_var_in_int_hash (old_left_expr) (new_left_expr)
	            (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
	            (current_vars_in_fun) (is_flatten)
	      in
	      (* Step4. Update information "a points to b" *)
	      update_points_to_int_hash (int_hash) (var_to_index) (index_to_var)
	        (new_left_expr) (new_addrof_expr) (uniqueness_hash);
	      (* Step5. Create predicates if left_expr is field of struct & *)
        (* the struct is pointed by some var                          *)
	      let pred_of_updated_points_to_field : predicate list = 
	          update_points_to_field_of_struct_in_int_hash (new_left_expr)
	            (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
	            (current_vars_in_fun) (is_flatten)
	      in
        (* Step6. Create predicates when addrof_expr is struct *)
        let preds_of_addrof_struct : predicate list = 
            addrof_expr_is_struct (new_addrof_expr) (new_left_expr) 
              (index_hash) (int_hash) (var_to_index)  (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
              (uniqueness_hash)
        in
	      (* Step7. Create real assignment *)
	      let assign1 : predicate = 
	          Eq (new_left_expr, Application ("addrof", [first_addrof_expr])) in
            (*Eq (new_left_expr, Application ("addrof", [new_addrof_expr])) in*)
	      let new_pref : expression =
	          change_index_of_pref (Application ("pref", [new_left_expr; Application ("NIL", [])])) (true)
	            (index_hash) (structs_hash) (current_vars_in_fun)
	      in
	      let assign2 : predicate = Eq (new_pref, new_addrof_expr) in
        (* 20090703 Create constraint for addrof *)
        let addrof_constraint : predicate = 
            Not (Eq (Application ("addrof", [first_addrof_expr]), Constant 0.0)) in
            (*Not (Eq (Application ("addrof", [new_addrof_expr]), Constant 0.0)) in*)
	      (* Step8. return *)
	      ([addrof_constraint; assign1; assign2] @ points_to_a_eq @ 
        pred_of_updated_points_to_field @ preds_of_addrof_struct)
	  end;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used when assigns var to var                           *)
(* Right_var & left_var is alias when right_var is a pointer               *)
and update_alias_in_int_hash (left_var : expression) (right_var : expression)
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    try (
        (* Find index of variables: *)
        let index_of_right_var : int = Hashtbl.find (var_to_index) (right_var) in
        let index_of_left_var : int = Hashtbl.find (var_to_index) (left_var) in
        (* update uniqueness_hash *)
        (* Get information of right_var & left_var in int_hash *)
        let (_, right_var_points_to) = Hashtbl.find (int_hash) (index_of_right_var) in
        let (points_to_left_var, left_var_points_to) = 
            Hashtbl.find (int_hash) (index_of_left_var) in
        (try
            let size = Hashtbl.find uniqueness_hash index_of_right_var in
            List.iter (fun expr_index ->
                Hashtbl.replace uniqueness_hash index_of_left_var size;
            ) (index_of_left_var :: points_to_left_var)
        with Not_found -> ());
        if right_var_points_to <= - 1 then begin
            if left_var_points_to >= 0 then begin
                Hashtbl.replace (int_hash) (index_of_left_var) 
                                (points_to_left_var, -1);
                []
            end else []
        (* Update information "left_var points to where right_var points to" *)
        end else begin
            Hashtbl.replace (int_hash) (index_of_left_var) 
                            (points_to_left_var, right_var_points_to);
            (* Update information "where right_var points to is pointed by *)
            (* left_var, too"                                              *)
            let (points_to_var, var_points_to) =
                try Hashtbl.find (int_hash) (right_var_points_to)
                with Not_found -> assert false
            in
            if (List.mem (index_of_left_var) (points_to_var)) = false then begin
                (* update uniqueness_hash *)
                try
                    let size = Hashtbl.find uniqueness_hash right_var_points_to in
                    Hashtbl.replace uniqueness_hash index_of_left_var size;
                with Not_found -> ();
                Hashtbl.replace (int_hash) (right_var_points_to)
                                ((points_to_var @ [index_of_left_var]), var_points_to);
            end;
            (* Find pointed_expr: *)
            let pointed_expr : expression =
                try Hashtbl.find (index_to_var) (right_var_points_to)
                with Not_found -> assert false
            in
            (* Note that we always get addr of first-new-var (var%0) *)
            let based_pointed_expr : expression = 
                create_based_vars_in_expression (pointed_expr) in
            let first_index_pointed_expr : expression = 
                change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                  (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                  (Hashtbl.create (0)) (ref []) (based_pointed_expr)
                  (Hashtbl.create 0)
            in
            (* Create predicates: {pref (a, 1) = x} & {a = addrof (x)} *)
            let new_addrof : expression = Application ("addrof", [first_index_pointed_expr]) in
            (*let new_addrof : expression = Application ("addrof", [pointed_expr]) in*)
            let pred1 : predicate = Eq (right_var, new_addrof) in
            let new_pref : expression =
                change_index_of_pref (Application ("pref", [left_var; Application ("NIL", [])])) (true) 
                  (index_hash) (structs_hash) (current_vars_in_fun)
            in
            let pred2 : predicate = Eq (new_pref, pointed_expr) in
            let pred3 : predicate = Eq (left_var, new_addrof) in
            (* Create predicates when pointed_expr is struct *)
            let preds_of_addrof_struct : predicate list = 
                addrof_expr_is_struct (pointed_expr) (left_var) (index_hash)
                  (int_hash) (var_to_index)  (index_to_var) (structs_hash)
                  (current_vars_in_fun) (is_flatten)
                  (uniqueness_hash)
            in
            (*let assign_addrof_x_to_a : predicate = 
                 create_based_vars_in_predicate (Eq (left_var, new_addrof))
            in
            let preds_of_assign_addrof_x_to_a : predicate list = 
                change_index_of_variables_in_an_assignment (assign_addrof_x_to_a)
                  (index_hash) (int_hash) (var_to_index) (index_to_var) 
                  (structs_hash) (current_vars_in_fun)
            in
            let pred3 : predicate = And preds_of_assign_addrof_x_to_a in*)
            (* return *) 
            if is_flatten = true then 
                [pred2; pred3] @ preds_of_addrof_struct
            else
                [Or ((Not pred1) :: [And ([pred2; pred3] @ preds_of_addrof_struct)])] 
        end
    ) with Not_found -> []
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and assign_var_for_var (right_expr : expression) (left_expr : expression) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* Change index of right_expr *)
    let new_right_expr : expression =
        change_index_of_variables_in_expression (false) (index_hash)
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (right_expr)
          (uniqueness_hash)
    in
    let _ = 
	      find_in_var_to_index (int_hash) (var_to_index)
  	      (index_to_var)  (new_right_expr)
    in
    (* Find version of left_expr *)
    let old_version_of_left_expr : int =
        find_index_of_expr_in_index_hash (left_expr) (index_hash) 
        (*try Hashtbl.find (index_hash) (left_expr)
        with
        | Not_found -> - 1*)
    in
    (* If left_expr has never been in index_hash yet *)
    if old_version_of_left_expr = - 1 then begin
        let new_left_expr : expression =
            change_index_of_variables_in_expression (true) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
        in
        let _ = 
	          find_in_var_to_index (int_hash) (var_to_index) 
	            (index_to_var) (new_left_expr)
        in
        (* Create predicates if left_epxr is field of struct and the struct *)
        (* is pointed by some var                                           *)
        let pred_of_updated_points_to_field : predicate list = 
            update_points_to_field_of_struct_in_int_hash (new_left_expr)
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
        in
        (* Create predicates & update information to int_hash if a & b *)
        (* become alias                                                *)
        let pred_of_updated_alias : predicate list =
            update_alias_in_int_hash (new_left_expr) (new_right_expr) 
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
              (uniqueness_hash)
        in
        (* Create predicates if right_expr is a struct *)
        let preds_of_right_expr_is_struct : predicate list =
            right_var_is_struct (new_right_expr) (new_left_expr) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (is_flatten) (uniqueness_hash)
        in
        (* Create real assignment *)
        let assignment = Eq (new_left_expr, new_right_expr) in
        (* Return *)
        (assignment :: pred_of_updated_alias) @ pred_of_updated_points_to_field
        @ preds_of_right_expr_is_struct
    
    (* If left_expr has been in index_hash *)
    end else begin
        let old_left_expr : expression =
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
        in  
        let new_left_expr : expression =
            change_index_of_variables_in_expression (true) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
        in
        (* Create predicates if left_epxr is field of struct and the struct *)
        (* is pointed by some var                                           *)
        let pred_of_updated_points_to_field : predicate list = 
            update_points_to_field_of_struct_in_int_hash (new_left_expr)
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
        in
        (* Create predicates when left_expr is changed value *)
        let points_to_a_eq : predicate list =
            update_value_to_var_in_int_hash (old_left_expr) (new_left_expr) 
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
        in
        (* Create predicates & update information to int_hash if a & b *)
        (* become alias                                                *)
        let pred_of_updated_alias : predicate list =
            update_alias_in_int_hash (new_left_expr) (new_right_expr) 
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten)
              (uniqueness_hash)
        in
        (* Create predicates if right_expr is a struct *)
        let preds_of_right_expr_is_struct : predicate list =
            right_var_is_struct (new_right_expr) (new_left_expr) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (is_flatten) (uniqueness_hash)
        in
        (* Create real assignment *)
        let assignment = Eq (new_left_expr, new_right_expr) in
        (* Return *)
        [assignment] @ pred_of_updated_alias @ points_to_a_eq @ 
        pred_of_updated_points_to_field @ preds_of_right_expr_is_struct
    end;
  end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Check if complex_right_expr is point to an element of array or not      *)
and is_an_element_of_array (complex_right_expr : expression) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate * expression) option =
begin
    (* debug *)
    (*debug "is_an_element_of_array"; read_int();*)
    match complex_right_expr with
    | Sum (array_pointer :: index_list) -> begin
        (* Change index of array_pointer *)
        let new_array_pointer : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (array_pointer)
              (uniqueness_hash)
        in
        (* Check if array_pointer really points to an array or not *)
        let index_of_array_pointer : int = 
            try Hashtbl.find (var_to_index) (new_array_pointer)
            with
            | Not_found -> -1
        in
        if index_of_array_pointer = -1 then None
        else begin
            let (_, array_pointer_points_to) = 
                try Hashtbl.find (int_hash) (index_of_array_pointer)
                with
                | Not_found -> assert false
            in
            if array_pointer_points_to = -1 then None
            else begin
                (* Find name of var which array_pointer points to *)
                let array_expr = 
                    try Hashtbl.find (index_to_var) (array_pointer_points_to)
                    with
                    | Not_found -> assert false
                in
                (* Note that we always get addr of first-new-var (var%0) *)
                let based_array_expr : expression = 
                    create_based_vars_in_expression (array_expr) in
                let first_index_array : expression = 
                    change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                      (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                      (Hashtbl.create (0)) (ref []) (based_array_expr)
                      (Hashtbl.create 0)
                in
                (* Check if array_expr is real array or not *)
                (match array_expr with
                | Application ("array", [memory_expr; index_expr]) ->
                    let based_memory_expr : expression = 
                        create_based_vars_in_expression (*false*) (memory_expr)
                    in
                    let i = match index_list with
                            | [index] -> index
                            | _ -> Sum index_list
                    in
                    let array_expr_of_right : expression = 
                        Application ("array", [based_memory_expr; i])
                    in
                    let condition : predicate = 
                        Eq (new_array_pointer, 
                            Application ("addrof", [first_index_array(*rray_expr*)]))
                    in
                    (* Return *)
                    Some (condition, array_expr_of_right) 
                | anything_else -> None)
            end
          end         
        end
      | anything_else -> None
  end
    

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and assign_complex_expr_for_var (complex_right_expr : expression) 
      (left_expr : expression) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* debug *)
    (*debug "assign_complex_expr_for_var"; (*read_int();*)
    debug ("left_expr: " ^ (string_of_expression left_expr));
    debug ("right_expr: " ^ (string_of_expression complex_right_expr));*) 
    (*(if complex_right_expr = (Constant 2.0) then (let _ = read_int () in ()));*)
    (* Check if right_expr is mem of an array or not: *)
    let is_mem_of_array : (predicate * expression) option = 
        is_an_element_of_array (complex_right_expr) (index_hash) (int_hash) 
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (uniqueness_hash)
    in
    (* Change index of vars in right_expr *)
    let new_right_expr : expression =
        change_index_of_variables_in_expression (false)
          (index_hash) (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (complex_right_expr)
          (uniqueness_hash)
    in
    let _ = 
	      find_in_var_to_index (int_hash) (var_to_index) 
	        (index_to_var) (new_right_expr)
    in
      
    (match is_mem_of_array with
    | Some (points_to_array_pred, array_expr_of_right) -> begin
        (* Note that we always get addr of first-new-var (var%0) *)
        let based_array_expr : expression = 
            (*create_based_vars_in_expression*) (array_expr_of_right) in
        let new_assignment = 
            Eq (left_expr, Application ("addrof", [based_array_expr]))
        in
        let preds_of_new_assigment : predicate list = 
            change_index_of_variables_in_an_assignment 
              (index_hash) (int_hash) (var_to_index) (index_to_var) 
              (structs_hash) (current_vars_in_fun) (is_flatten) 
              (array_asgn_is_flatten) (true) (elements_of_arrays) 
              (check_pointer) (null_pointer_exceptions) (var_eq_hash) (new_assignment)
              (uniqueness_hash) 
        in
        (* Create predicate: pred_of_old_assignment /\                 *)
        (* points_to_array_pred => And pred_of_new_assigment           *)
        (* Note that if we change index for vars in left, flag = false *)
        let new_left_expr : expression = 
            change_index_of_variables_in_expression (false) 
              (index_hash) (int_hash) (var_to_index) (index_to_var)
              (structs_hash) (current_vars_in_fun) (left_expr)
              (uniqueness_hash)
        in
        let pred_of_old_assignment = Eq (new_left_expr, new_right_expr) in
        let implication = Or ((Not points_to_array_pred) :: [And preds_of_new_assigment]) in         
        (* return *)
        if is_flatten = true then 
              pred_of_old_assignment :: preds_of_new_assigment
        else pred_of_old_assignment :: [implication] 
      end
    | None -> begin
        (* Find verstion of left_expr *)
			  (*let old_version_of_left_expr : int = (try Hashtbl.find (index_hash) (left_expr)
			                                         with Not_found -> - 1)
			  in*)
        let old_version_of_left_expr : int = 
            find_index_of_expr_in_index_hash (left_expr) (index_hash) in
			  (* if left_expr has never been in index_hash yet: *)
			  if old_version_of_left_expr = - 1 then begin
			      let new_left_expr : expression =
			          change_index_of_variables_in_expression (true) 
                  (index_hash) (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (left_expr)
                  (uniqueness_hash)
			      in
			      let _ = 
			          find_in_var_to_index (int_hash) (var_to_index) 
			            (index_to_var) (new_left_expr)
			      in
            (*let _ = read_int () in*)
			      (* Create predicates if left_expr is field of struct and the 
            struct is pointed by some var *)
			      let pred_of_updated_points_to_field : predicate list = 
			          update_points_to_field_of_struct_in_int_hash (new_left_expr)
			            (index_hash) (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (is_flatten)
			      in
			      [Eq (new_left_expr, new_right_expr)] @ pred_of_updated_points_to_field
			      
			  end else begin
			      let old_left_expr : expression =
			          change_index_of_variables_in_expression (false) 
                  (index_hash) (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (left_expr)
                  (uniqueness_hash)
			      in
			      let new_left_expr : expression =
			          change_index_of_variables_in_expression (true) 
                  (index_hash) (int_hash) (var_to_index) (index_to_var)
                  (structs_hash) (current_vars_in_fun) (left_expr)
                  (uniqueness_hash)
			      in
			      let points_to_a_eq : predicate list =
			        update_value_to_var_in_int_hash (old_left_expr) (new_left_expr) 
			          (index_hash) (int_hash) (var_to_index) (index_to_var)
                (structs_hash) (current_vars_in_fun) (is_flatten)
			      in
			      let pred_of_updated_points_to_field : predicate list = 
			        update_points_to_field_of_struct_in_int_hash (new_left_expr)
			          (index_hash) (int_hash) (var_to_index) (index_to_var)
                (structs_hash) (current_vars_in_fun) (is_flatten)
			      in
            (* Set where left_expr points to to -1 *)
            (try (
			          let index_of_left_expr = Hashtbl.find (var_to_index) (new_left_expr) in
			          let (points_to_left, left_points_to) =
			              Hashtbl.find (int_hash) (index_of_left_expr)
    			      in
		    	      (*Hashtbl.remove (int_hash) (index_of_left_expr);*)
			          Hashtbl.replace (int_hash) (index_of_left_expr) (points_to_left, - 1);
            ) with Not_found -> ());
			      let assignment : predicate = Eq (new_left_expr, new_right_expr) in
			      (* return *)
			      (assignment :: points_to_a_eq) @ pred_of_updated_points_to_field
			    end
        end);
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and assign_memory_for_var (left_expr : expression) 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* change application "memory" *)
    let memory_expr = Application ("memory", []) in
    let new_memory_expr : expression = 
        change_index_of_variables_in_expression (false) (index_hash)
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (memory_expr)
          (uniqueness_hash)
    in
    let memory_var = Variable (string_of_expression (new_memory_expr)) in
    (* Create first-element of array *)
    let a0_of_memory = Application ("array", [memory_var; Constant 0.0]) in
    (* Create new assignment: *)
    let new_assignment = 
        Eq (left_expr, Application ("addrof", [a0_of_memory])) in
    (* Change index of vars in new-assignment, then return: *)
    change_index_of_variables_in_an_assignment (index_hash)
      (int_hash) (var_to_index) (index_to_var) (structs_hash) 
      (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten) (true) 
      (elements_of_arrays) (check_pointer) (null_pointer_exceptions) (var_eq_hash)
      (new_assignment) (uniqueness_hash)
end
  
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)    
(* Note: memory_expr has just changed index of its variables yet           *)
(*       index_expr has changed index                                      *)
and update_elements_of_arrays (memory_expr : expression)
      (index_expr : expression)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
: unit = 
begin
    (* Step1. Find all indices of the array_expr in hashtable *)
    let old_indices : expression list = 
        try Hashtbl.find (elements_of_arrays) (memory_expr)
        with
        | Not_found -> []
    in
    (* Step2. If the hashtbl hasn't any element of the array_expr, *)
    (* add array_expr to that hashtbl                              *)
    if old_indices = [] then begin
        Hashtbl.add (elements_of_arrays) (memory_expr) ([index_expr])
    (* Else, update information *)
    end else begin
        let is_mem : bool = is_mem_of_expr_list (index_expr) (old_indices) in
        if is_mem = false then 
            Hashtbl.replace (elements_of_arrays) (memory_expr) 
              (index_expr :: old_indices)
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and create_contraints_for_new_index_of_array (old_memory_expr : expression)
      (new_memory_expr : expression) 
      (old_indices : expression list)
      (new_index : expression)
: predicate list = 
begin
    let constraints : predicate list ref = ref [] in
    List.iter    
        (fun an_old_index -> 
            (* Create constraint for new_index *)
            let index_constraint : predicate = 
                Not (Eq (new_index, an_old_index)) in
            (* Create constraint for value of element of array *)
            let old_element : expression = 
                Application ("array", [old_memory_expr; an_old_index]) in
            let new_element : expression = 
                Application ("array", [new_memory_expr; an_old_index]) in
            let element_constraint : predicate= Eq (new_element, old_element) in
            (* Create final constraint: *)
            let final_constraint : predicate = 
                Or [Not index_constraint; element_constraint] in
            (* Add to list to return *)
            constraints := final_constraint :: !constraints;  
        )
        old_indices;
    (* Return *)
    !constraints
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and assign_for_array_element (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t) 
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool) 
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (array_app : expression)
      (right_expr : expression)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* debug *)
    (*debug "assign_for_array_element";
    debug ("lenght hash: " ^ string_of_int (Hashtbl.length elements_of_arrays));*)
    (*read_int();*)
    (* Step1. Check array_app *)
    match array_app with
    | Application ("array", [memory_expr; index_expr]) -> begin
        let old_memory_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (memory_expr)
              (uniqueness_hash)
        in
        let new_index_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (index_expr)
              (uniqueness_hash)
        in
        (* Find all indices of the memory_expr in elements_of_arrays *)
        let old_indices : expression list = 
            try Hashtbl.find (elements_of_arrays) (memory_expr)
            with
            | Not_found -> []
        in
        (* Generate predicates for real assignment: *)
        let old_eq : predicate = Eq (array_app, right_expr) in
        let preds_of_real_assignment : predicate list = 
            generate_preds_for_preprocessed_assignment (index_hash) (int_hash)
              (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
              (is_flatten) (array_asgn_is_flatten) (elements_of_arrays) (old_eq)
              (check_pointer) (null_pointer_exceptions) (Hashtbl.create 0)
              (uniqueness_hash)
        in
        let new_memory_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash)
              (current_vars_in_fun) (memory_expr)
              (uniqueness_hash)
        in
        (* Create constraints of new index and old_indices: *)
        let constraints : predicate list = 
            create_contraints_for_new_index_of_array (old_memory_expr)
              (new_memory_expr) (old_indices) (new_index_expr)
        in
        (* Create constraint for addrof *)
        let old_array_0 : expression = 
            Application ("array", [old_memory_expr; Constant 0.0]) in
        let new_array_0 : expression = 
            Application ("array", [new_memory_expr; Constant 0.0]) in
        let addrof_constraint : predicate = 
            Eq (Application ("addrof", [new_array_0]), 
                Application ("addrof", [old_array_0])) in
        (* Update information to elements_of_arrays *)
        update_elements_of_arrays (memory_expr) (new_index_expr)
          (elements_of_arrays);
        (* Return *)
        (* TEMPORARILY deal with flatten (to deal with function pointer) *)
        (if (is_flatten = true) & (array_asgn_is_flatten = true) then
            preds_of_real_assignment @ [addrof_constraint]
        else 
            preds_of_real_assignment @ constraints @ [addrof_constraint])
        (* 20090619. Because slicing path with criteria need to tract backward,
        therefore cannot flatten this case (do not know index this time) *)
        (*preds_of_real_assignment @ constraints @ [addrof_constraint]
        --> but analysis function pointer need flatten path!!!*)
      end
    | anything_else -> assert false
end

(** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The constraints (which concern with a pointer p) are only generated when having a
    assignment (p = &v) after that.
    Therefore, if a block before changing index is {p=q; pref(p)=1; pref(q)=0}
               then after change_index is {p%0=q%0; pref(p%0,0)=1; pref(q%0,0)=0}
    but we expect result is {p%0=q%0; pref(p%0,0)=1; pref(q%0,1)=0}
    This happen because we did not known where p and q point to, pointing
    to same memory by p and q can't be detained in table int_hash
    To fix this problem, we create table var_eq_hash to detain assignments (p=q)
    Before generating constraint from a assign *p = x:
        if don't find memory that p point to in int_hash,
           and exist assignment (p=q) in var_eq_hash
           and have never had *p in index_hash yet
           and have had *q in index_hash
        then update *p to index_hash with version = version of *q
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and is_mem_of_int_hash (int_hash : (int, (int list) * int) Hashtbl.t)
                       (var_to_index : (expression, int) Hashtbl.t)
                       (index_to_var : (int, expression) Hashtbl.t)
                       (expr : expression)
: bool = (
    try
        let expr_index = Hashtbl.find var_to_index expr in
        let _, memory_index = Hashtbl.find int_hash expr_index in
        let _ = Hashtbl.find index_to_var memory_index in
        true
    with Not_found -> false
)

and update_index_hash_before_gen_preds
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (preprocessed_assignment : predicate) 
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: unit = (
    (* if exist a expr = pref_expr and not exist pref_expr = &v
       then update version of *pref_expr = version of *expr
    *)
    (match preprocessed_assignment with
    | Eq (Application ("srefp", pref_expr :: _), _) -> ()
    | Eq (Application ("pref", pref_expr :: temp_expr :: l), _) -> (
        let left_expr = Application ("pref", pref_expr :: temp_expr :: l) in
        let new_current_vars_in_fun = current_vars_in_fun in
        let pref_expr_changed_index =
            change_index_of_variables_in_expression false 
                                                    (Hashtbl.copy index_hash)
                                                    (Hashtbl.copy int_hash)
                                                    (Hashtbl.copy var_to_index)
                                                    (Hashtbl.copy index_to_var) 
                                                    (Hashtbl.copy structs_hash)
                                                    new_current_vars_in_fun
                                                    pref_expr
                                                    (Hashtbl.copy uniqueness_hash)
        in
        let temp_expr_changed_index =
            change_index_of_variables_in_expression false 
                                                    (Hashtbl.copy index_hash)
                                                    (Hashtbl.copy int_hash)
                                                    (Hashtbl.copy var_to_index)
                                                    (Hashtbl.copy index_to_var) 
                                                    (Hashtbl.copy structs_hash)
                                                    new_current_vars_in_fun
                                                    temp_expr
                                                    (Hashtbl.copy uniqueness_hash)
        in
        if is_mem_of_int_hash int_hash
                              var_to_index
                              index_to_var
                              pref_expr_changed_index then ()
        else (
            try
                let old_version_of_left_expr =
                    find_index_of_expr_in_index_hash left_expr
                                                     index_hash
                in
                if old_version_of_left_expr = -1 then (
                    let expr_eq_pref_expr =
                        Hashtbl.find var_eq_hash pref_expr_changed_index
                    in
                    let old_index =
                        (Hashtbl.find index_hash
                                      (Application ("pref", [expr_eq_pref_expr; Application ("NIL", [])])))
                    in
                    if old_index <> -1 then
                        Hashtbl.add index_hash
                                    (Application ("pref", [pref_expr_changed_index;
                                                           temp_expr_changed_index]))
                                    old_index
                    else ())
                 else ()
            with Not_found -> ()
        )
      )
    | _ -> ());

)
(**~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    After generating predicates from assignment *p = x:
         find all expression q = p in var_eq_hash
         and update version of *q = version of *p to index_hash
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and update_index_hash_after_gen_preds
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (preds : predicate list) 
      (var_eq_hash : (expression, expression) Hashtbl.t)
: unit = (
    (* update index_hash *)
    List.iter (fun a_pred ->
        match a_pred with
        | Eq (Application ("pref", pref_pointer :: temp_expr :: _), _) -> (
            try
                if (is_mem_of_int_hash int_hash
                                       var_to_index
                                       index_to_var
                                       pref_pointer) = false then (
                    let new_version =
                        Hashtbl.find index_hash
                                     (Application ("pref", [pref_pointer;
                                                            temp_expr])) in
                    let exprs_eq_pref_pointer =
                        Hashtbl.find_all var_eq_hash pref_pointer in
                    List.iter (fun a_expr ->
                        Hashtbl.replace index_hash (Application ("pref", [a_expr; Application ("NIL", [])]))
                                        new_version;
                    ) exprs_eq_pref_pointer;)
                else ()
            with Not_found -> ()
          )
       | _ -> ()
    ) preds;
)

(** after generating predicates from assignment (p = q)
    update to var_eq_hash
 *)
and update_var_eq_hash
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (preds : predicate list) = (
    (* update var_eq_hash *)
    List.iter (fun a_pred ->
        match a_pred with
        | Eq (left_expr, right_expr) -> (
            match left_expr, right_expr with
            | Constant _, _
            | _, Constant _ -> ()
            | _ -> (
                let exprs = Hashtbl.find_all var_eq_hash right_expr in
                let exprs_replaced_left_expr =
                    List.filter (fun a_expr ->
                        (La_common.compare_expr_with_expr left_expr a_expr) = 0
                    ) exprs
                in
                if List.exists (fun a_expr ->
                       match a_expr with
                       | Constant _ -> true
                       | _ -> false
                   ) exprs then ()
                else (
                    let add_to_var_eq_hash var_eq_hash left_expr right_expr
                    : unit = (
                        let right_exprs = Hashtbl.find_all var_eq_hash left_expr in
                        if La_common.is_mem_of_expr_list right_expr right_exprs then
                            ()
                        else (
                            Hashtbl.add var_eq_hash left_expr right_expr;
                            Hashtbl.add var_eq_hash right_expr left_expr;)
                    ) in
                    add_to_var_eq_hash var_eq_hash left_expr right_expr;
                    List.iter (fun a_expr ->
                        add_to_var_eq_hash var_eq_hash left_expr a_expr
                    ) exprs;))
          )
        | _ -> ()
    ) preds;
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and generate_preds_for_preprocessed_assignment 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (preprocessed_assignment : predicate) 
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    update_index_hash_before_gen_preds index_hash
                                     int_hash
                                     var_to_index
                                     index_to_var 
                                     structs_hash
                                     current_vars_in_fun
                                     preprocessed_assignment 
                                     var_eq_hash
                                     uniqueness_hash;
    (* debug *)
    (*debug " In function generate_preds_for_preprocessed_assignment: ";
    debug ("preprocessed_assignment: " ^ (string_of_predicate preprocessed_assignment));*)
    let result =
    (match (*assignment*) preprocessed_assignment with
    | Eq (left_expr, Application ("memory", [])) -> (*assert false*)
        assign_memory_for_var (left_expr) (index_hash) (int_hash) 
          (var_to_index) (index_to_var) (structs_hash) 
          (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
          (elements_of_arrays) (check_pointer) (null_pointer_exceptions)
          (var_eq_hash) (uniqueness_hash)
    | Eq (left_expr, Application ("addrof", addrof_expr :: [])) ->
        assign_addrof_for_var (addrof_expr) (left_expr) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (structs_hash) 
          (current_vars_in_fun) (is_flatten) (uniqueness_hash)
    | Eq (left_expr, Application ("pref", (pref_expr) :: temp_expr :: [])) ->
        (*assign_pref_for_var (pref_expr) (left_expr) (index_hash) (int_hash)
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (is_flatten)*)
        assign_var_for_var (Application("pref", [pref_expr; temp_expr]))
          (left_expr) (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash)(current_vars_in_fun) (is_flatten) var_eq_hash
          (uniqueness_hash)
    | Eq (left_expr, Application ("srefp", [pstruct; field])) -> 
        (*assign_srefp_for_var (pstruct) (field) (left_expr) (index_hash)
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (is_flatten)*)
        assign_var_for_var (Application ("srefp", [pstruct; field]))
          (left_expr) (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash)(current_vars_in_fun) (is_flatten) var_eq_hash
          (uniqueness_hash)
    (*| Eq (Application ("pref", (pref_expr) :: []), right_expr) ->
        assign_expr_for_pref (right_expr) (pref_expr) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (structs_hash) 
          (current_vars_in_fun) (is_flatten)
    | Eq (Application ("srefp", [pstruct; field]), right_expr) ->
        assign_expr_for_srefp (right_expr) (pstruct) (field) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (structs_hash)
          (current_vars_in_fun) (is_flatten)*)
    | Eq (left_expr, Application ("sref", [str; field])) ->
        assign_var_for_var (Application ("sref", [str; field])) (left_expr) 
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (is_flatten) var_eq_hash
          (uniqueness_hash)
    | Eq (left_expr, Application ("array", [array; index])) ->
        assign_var_for_var (Application ("array", [array; index])) (left_expr)
          (index_hash) (int_hash) (var_to_index) (index_to_var)
          (structs_hash) (current_vars_in_fun) (is_flatten) var_eq_hash
          (uniqueness_hash)
    | Eq (left_expr, Variable b) ->
        assign_var_for_var (Variable b) (left_expr) (index_hash) (int_hash) 
          (var_to_index) (index_to_var) (structs_hash)(current_vars_in_fun)
          (is_flatten) var_eq_hash
          (uniqueness_hash)
    | Eq (left_expr, complex_right_expr) ->
        assign_complex_expr_for_var (complex_right_expr) (left_expr)
          (index_hash) (int_hash) (var_to_index) (index_to_var) 
          (structs_hash) (current_vars_in_fun) (is_flatten)
          (array_asgn_is_flatten) (elements_of_arrays) (check_pointer)
          (null_pointer_exceptions) (var_eq_hash) (uniqueness_hash)
    (*| Eq _ -> []*)
    (* 14.4.2009. don't need deal with "assume" here *)
    | any_pred -> begin 
        let new_pred : predicate = 
            change_index_of_variables_in_non_condition_predicate (index_hash)
              (int_hash) (var_to_index) (index_to_var) (structs_hash) 
              (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
              (elements_of_arrays) (check_pointer) (null_pointer_exceptions)
              (var_eq_hash) (any_pred) (uniqueness_hash)
        in
        (* return *)
        [new_pred]
      end) in

    (* update var_eq_hash *)
    update_var_eq_hash var_eq_hash result;
    (* update index_hash *)
    update_index_hash_after_gen_preds index_hash
                                      int_hash
                                      var_to_index
                                      index_to_var
                                      result var_eq_hash;
    result
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and change_index_of_variables_in_an_assignment
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (preprocessed_array : bool)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (assignment : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list =
begin
    (* debug *)
    (*debug ("``````````````````````````````````````````````````````");
    debug ("assignment: " ^ string_of_predicate assignment);*)
    (*debug ("lenght hash: " ^ string_of_int (Hashtbl.length elements_of_arrays));*)
    (* Step1. Preprocess pointers: *)
    let (conditions, preprocessed_assignment) = 
        preprocess_pred_for_change_index (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (assignment) 
          (check_pointer) (null_pointer_exceptions) (uniqueness_hash)
    in
    (* debug *)
    (*debug ("conditions: " ^ (string_of_predicate (And conditions)));
    debug ("preprocessed_assignment: " ^ (string_of_predicate preprocessed_assignment));
    debug ("null_pointer_exceptions: " ^ (string_of_int (List.length !null_pointer_exceptions)));*)
    (* Step2. Preprocess array in left of assignment *)
    let result : predicate list = 
        (if preprocessed_array = false then begin
            (match preprocessed_assignment with
            | Eq (Application ("array", [memory_expr; index_expr]), right_expr) -> 
                assign_for_array_element (index_hash) (int_hash) (var_to_index)
                  (index_to_var) (structs_hash) (current_vars_in_fun)
                  (is_flatten) (array_asgn_is_flatten) (elements_of_arrays)
                  (Application ("array", [memory_expr; index_expr])) (right_expr)
                  (check_pointer) (null_pointer_exceptions)
                  (uniqueness_hash)
            | (*anything_else*) _ -> begin 
                let temp_pred_list : predicate list = 
                    generate_preds_for_preprocessed_assignment (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (structs_hash)
                      (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
                      (elements_of_arrays) (preprocessed_assignment) (check_pointer)
                      (null_pointer_exceptions) (var_eq_hash) (uniqueness_hash)
                in
                if conditions = [] then temp_pred_list
                else if is_flatten = true then temp_pred_list 
                else [Or ((Not (And conditions)) :: [And temp_pred_list])]
              end);
        end else begin
            let temp_pred_list : predicate list = 
                generate_preds_for_preprocessed_assignment (index_hash)
                  (int_hash) (var_to_index) (index_to_var) (structs_hash)
                  (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
                  (elements_of_arrays) (preprocessed_assignment) (check_pointer)
                  (null_pointer_exceptions) (var_eq_hash) (uniqueness_hash)
            in
            if conditions = [] then temp_pred_list
            else if is_flatten = true then temp_pred_list 
            else [Or ((Not (And conditions)) :: [And temp_pred_list])]
        end)
    in
    (* debug:  *)
    (*debug ("result of func: change_index_of_variables_in_an_assignment");
    debug (string_of_predicate (And result));*)
    (*let _ = read_int () in*)
    (* Return *)
    result
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Input: an expression of assume                                          *)
(* Output: new expression of assume & constraints of array in assume       *)
and replace_pointer_by_array_element_in_pref 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pref_expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (expression * (predicate list)) =
begin
    match pref_expr with
    | Application ("pref", [Sum (pointer_expr :: adding_value_exprs); 
                            temp_expr]) -> (begin
        let old_pointer_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (pointer_expr)
              (uniqueness_hash)
        in
        (* Check if pointer_expr points to a memory or not: *)
        let index_of_pointer : int = 
            try Hashtbl.find (var_to_index) (old_pointer_expr)
            with
            | Not_found -> -1
        in
        if index_of_pointer >= 0 then begin
            let (points_to_pointer, pointer_points_to) = 
                try Hashtbl.find (int_hash) (index_of_pointer)
                with
                | Not_found -> assert false
            in
            if pointer_points_to >= 0 then begin
                (* Find first element of memory which the pointer points to *)
                let first_element : expression = 
                    try Hashtbl.find (index_to_var) (pointer_points_to)
                    with
                    | Not_found -> assert false
                in
                (* Find memory *)
                let memory_expr = (match first_element with
                                  | Application ("array", [m; fi]) -> m
                                  | anything_else -> assert false)
                in
                let based_memory_expr : expression = 
                    create_based_vars_in_expression (*false*) (memory_expr) in
                (* Create constraints: *)
                let addr_constraint : predicate = 
                    Eq (old_pointer_expr, Application ("addrof", [first_element]))
                in
                let i = match adding_value_exprs with
                        | [adding_value_expr] -> adding_value_expr
                        | _ -> Sum adding_value_exprs
                in          
                (* create new element of array *)
                let new_array_element : expression = 
                    Application ("array", [based_memory_expr; i])
                in          
                (* Return *)
                (new_array_element, [addr_constraint])
            end else begin
                (pref_expr, [])
            end  
        end else begin
            (pref_expr, [])
        end;
      end)
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Input: an expression of assume                                          *)
(* Output: new expression of assume & constraints of array in assume       *)
and replace_pointer_by_array_element_in_expression
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (expression * (predicate list)) = 
begin
    match expr with
    | Constant f -> (Constant f, [])
    | Variable s -> (Variable s, [])
    | Application ("pref", [Sum (pointer_expr :: adding_value_exprs);
                            temp_expr]) -> begin
        replace_pointer_by_array_element_in_pref (index_hash) 
          (int_hash) (var_to_index) (index_to_var) (expr) (uniqueness_hash)
      end
    | Application (s, expr_list) -> begin
        let new_expr_list : expression list ref = ref [] in
        let constraints : predicate list ref = ref [] in
        List.iter
            (fun an_expr -> 
                let (a_new_epxr, new_constraints) = 
                    replace_pointer_by_array_element_in_expression (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (an_expr)
                      (uniqueness_hash)
                in
                new_expr_list := !new_expr_list @ [a_new_epxr];
                constraints := !constraints @ new_constraints              
            )
            expr_list;
        (Application (s, !new_expr_list), !constraints)
      end
    | Sum expr_list -> begin
        let new_expr_list : expression list ref = ref [] in
        let constraints : predicate list ref = ref [] in
        List.iter
            (fun an_expr -> 
                let (a_new_epxr, new_constraints) = 
                    replace_pointer_by_array_element_in_expression (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (an_expr)
                      (uniqueness_hash)
                in
                new_expr_list := !new_expr_list @ [a_new_epxr];
                constraints := !constraints @ new_constraints              
            )
            expr_list;
        (Sum !new_expr_list, !constraints)
      end
    | Coeff (f, an_expr) -> begin
        let (a_new_expr, new_constraints) = 
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (an_expr)
              (uniqueness_hash)
        in
        (Coeff (f, a_new_expr), new_constraints)
      end 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Input: an assume                                                        *)
(* Output: new assume & constraints of array in the assume                 *)
and replace_pointer_by_array_element_in_predicate 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t)
      (pred : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: (predicate * (predicate list))= 
begin
    match pred with
    | True -> (True, [])
    | False -> (False, [])
    | And pred_list -> begin
        let new_pred_list : predicate list ref = ref [] in
        let constraints : predicate list ref = ref [] in
        List.iter
            (fun a_pred -> 
                let (a_new_pred, new_constraints) = 
                    replace_pointer_by_array_element_in_predicate (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (a_pred)
                      (uniqueness_hash)
                in
                new_pred_list := !new_pred_list @ [a_new_pred];
                constraints := !constraints @ new_constraints; 
            )
            pred_list;
        (* Return *)
        (And !new_pred_list, !constraints)
      end
    | Or pred_list -> begin
        let new_pred_list : predicate list ref = ref [] in
        let constraints : predicate list ref = ref [] in
        List.iter
            (fun a_pred -> 
                let (a_new_pred, new_constraints) = 
                    replace_pointer_by_array_element_in_predicate (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (a_pred)
                      (uniqueness_hash)
                in
                new_pred_list := !new_pred_list @ [a_new_pred];
                constraints := !constraints @ new_constraints; 
            )
            pred_list;
        (* Return *)
        (Or !new_pred_list, !constraints)
      end
    | Not p -> begin
        let (new_p, new_constraints) = 
            replace_pointer_by_array_element_in_predicate (index_hash)
              (int_hash) (var_to_index) (index_to_var) (p)
              (uniqueness_hash)
        in
        (Not new_p, new_constraints)
      end
    | Eq (expr1, expr2) -> begin
        let (new_expr1, constraints1) = 
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr1)
              (uniqueness_hash)
        in
        let (new_expr2, constraints2) =
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr2)
              (uniqueness_hash)
        in
        (Eq (new_expr1, new_expr2), constraints1 @ constraints2)
      end
    | Lt (expr1, expr2) -> begin
        let (new_expr1, constraints1) = 
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr1)
              (uniqueness_hash)
        in
        let (new_expr2, constraints2) =
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr2)
              (uniqueness_hash)
        in
        (Lt (new_expr1, new_expr2), constraints1 @ constraints2)
      end
    | Leq (expr1, expr2) -> begin
        let (new_expr1, constraints1) = 
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr1)
              (uniqueness_hash)
        in
        let (new_expr2, constraints2) =
            replace_pointer_by_array_element_in_expression (index_hash)
              (int_hash) (var_to_index) (index_to_var) (expr2)
              (uniqueness_hash)
        in
        (Leq (new_expr1, new_expr2), constraints1 @ constraints2)
      end
    | Atom i -> (Atom i, [])
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of var                                 *)
(* This function is used in reset_value_of_pref                            *)
(* Input: a var need to reset                                              *)
(* Output: constraints to reset the var                                    *)
and reset_value_of_variable (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (is_flatten : bool)
      (var : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    (* Find old_var *)
    let old_var : expression = 
        change_index_of_variables_in_expression (false) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) 
          (Hashtbl.create(0)) (ref []) (var)
          (uniqueness_hash)
    in
    let _ = 
        find_in_var_to_index (int_hash) (var_to_index) (index_to_var)  (old_var)
    in
    (* Create new_var *)
    let new_var : expression = 
        change_index_of_variables_in_expression (true) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) 
          (Hashtbl.create(0)) (ref []) (var)
          (uniqueness_hash)
    in
    let first_indexed_var : expression =
        change_index_of_variables_in_expression (true) (Hashtbl.create (0))
          (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
          (Hashtbl.create (0)) (ref []) (var)
          (Hashtbl.create 0)
    in
    (* Generate constraint for addrof var *)
    let addrof_constraint = 
        Eq (Application ("addrof", [new_var]), Application ("addrof", [old_var]))
    in
    (* Find all vars pointing to this var *)
    let points_to_var : expression list = 
        find_all_points_to_var (old_var) (int_hash) (var_to_index) (index_to_var)
    in
    let based_points_to_var : expression list = 
        List.map (create_based_vars_in_expression (*false*))
                 points_to_var
    in
    (* Create constraints *)
    let constraints : predicate list ref = ref [] in
    (try
        (List.iter2
            (fun (a_point_to_var) (a_based_point_to_var) -> 
                let a_pref : expression = 
                    Application ("pref", [a_based_point_to_var; Application ("NIL", [])]) in
                let a_new_pref : expression = 
                    change_index_of_variables_in_expression (true) (index_hash) 
                      (int_hash) (var_to_index) (index_to_var) 
                      (Hashtbl.create(0)) (ref []) (a_pref)
                      (uniqueness_hash)
                in      
                (* Create condition: *)
                let condition : predicate = 
                    Eq (a_point_to_var, Application ("addrof", [first_indexed_var(*old_var*)])) in
                (* Create reset predicate: *)
                let reset_predicate : predicate = Eq (a_new_pref, new_var) in
                (* Create implication *)
                let implication : predicate = 
                    Or [Not condition; reset_predicate] in
                if is_flatten = true then 
                    constraints := !constraints @ [reset_predicate]
                else constraints := !constraints @ [implication]
            )
            points_to_var
            based_points_to_var
        )
    with
    | Invalid_argument _ -> assert false); 
    
    (* Reset var_points_to *)
    let index_of_var : int = try Hashtbl.find (var_to_index) (old_var)
                             with
                             | Not_found -> assert false
    in
    let (points_to_var, var_points_to) = 
        try Hashtbl.find (int_hash) (index_of_var)
        with
        | Not_found -> assert false
    in
    Hashtbl.replace (int_hash) (index_of_var) (points_to_var, -1);
    (* Reset var_to_index *)
    Hashtbl.remove (var_to_index) (old_var);
    Hashtbl.add (var_to_index) (new_var) (index_of_var);
    (* Reset index_to_var *)
    Hashtbl.replace (index_to_var) (index_of_var) (new_var);
    (* Return *)
    addrof_constraint :: !constraints    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of pref                                *)
(* This function is used in change_index_of_variables_in_assume            *)
(* Input: a pref need to reset                                             *)
(* Output: constraints to reset the pref                                   *)
and reset_value_of_pref (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (is_flatten : bool)
      (pref_expr : expression)
      (check_pointer : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    match pref_expr with
    | Application ("pref", [pointer_expr; temp_expr]) -> begin 
        (* Create old_pointer_expr *)
        let old_pointer_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (pointer_expr)
              (uniqueness_hash)
        in
        (* Find index of the pointer: *)
        let index_of_pointer : int = 
            try Hashtbl.find (var_to_index) (old_pointer_expr)
            with
            | Not_found -> -1 
        in
        if index_of_pointer >= 0 then begin
            (* Check if pointer_expr points to a var or not *)
            let (points_to_pointer, pointer_points_to) = 
                try Hashtbl.find (int_hash) (index_of_pointer)
                with
                | Not_found -> assert false
            in
            if pointer_points_to >= 0 then begin
                (* Find var which pointer points to *)
                let pointed_var : expression = 
                    try Hashtbl.find (index_to_var) (pointer_points_to) 
                    with
                    | Not_found -> assert false
                in
                let based_var : expression = 
                    create_based_vars_in_expression (*false*) (pointed_var) in
                (* Reset value of variable *)
                let reset_pointed_var_constraints : predicate list = 
                    reset_value_of_variable (index_hash) (int_hash) 
                      (var_to_index) (index_to_var) (is_flatten) (based_var)
                      (uniqueness_hash)
                in 
                (* Create impication for before reset constraints: *)  
                (* Note that we always get addr of first-new-var (var%0) *)
                (* Create first-index_str *)
                let first_index_pointed_var : expression = 
                    change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                      (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                      (Hashtbl.create (0)) (ref []) (based_var)
                      (Hashtbl.create 0)
                in
                let condition : predicate = 
                    Eq (old_pointer_expr, 
                        Application ("addrof", [first_index_pointed_var(*pointed_var*)]))
                in
                let new_pointed_var : expression = 
                    change_index_of_variables_in_expression (false) (index_hash) 
                      (int_hash) (var_to_index) (index_to_var) 
                      (Hashtbl.create(0)) (ref []) (based_var)
                      (uniqueness_hash)
                in
                let new_pref_expr = 
                    change_index_of_variables_in_expression (false) (index_hash) 
                      (int_hash) (var_to_index) (index_to_var) 
                      (Hashtbl.create(0)) (ref []) (pref_expr)
                      (uniqueness_hash)
                in
                let reset_pred = Eq (new_pointed_var, new_pref_expr) in
                let implication : predicate = 
                    Or [Not condition; And (reset_pred :: reset_pointed_var_constraints)]
                in
                (* Return *)
                if is_flatten = true then 
                    reset_pred :: reset_pointed_var_constraints
                else [implication] 
            (* If pointer_points_to < 0 *)
            end else begin
                if check_pointer = true then begin
                    (* error ((string_of_expression pointer_expr) ^ 
                                   " points to nowhere");
                    assert false;*)
                    raise (La_exception.Not_point_to (("", -1), pointer_expr, old_pointer_expr, []));
                end else []                
            end
        end else []
      end
    | Application ("pref", expr_list) -> []
    | anything_else -> assert false    
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of sref                                *)
(* This function is used in change_index_of_variables_in_assume            *)
(* Input: a sref need to reset                                             *)
(* Output: constraints to reset the sref                                   *)
and reset_value_of_sref (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (is_flatten : bool)
      (sref_expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
   (* debug *)
    (*debug ("reset_value_of_sref: " ^ 
                  (string_of_expression sref_expr));*)
    match sref_expr with
    | Application ("sref", [str_expr; field_expr]) -> begin
        (* Step1. get old_str_expr *)
        let old_str_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (str_expr)
              (uniqueness_hash)
        in
        (* Create first-index_str *)
        let first_index_str_expr : expression = 
            change_index_of_variables_in_expression (false) (Hashtbl.create(0))
              (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
              (Hashtbl.create (0)) (ref []) (str_expr)
              (Hashtbl.create 0)
        in
        (* debug *)
        (*debug ("old_str_expr " ^ (string_of_expression old_str_expr));*)
        
        let _ = 
            find_in_var_to_index (int_hash) (var_to_index)
              (index_to_var)  (old_str_expr)
        in
        (* Get new sref *)
        (*let old_sref_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (sref_expr)
        in*)
        let new_sref_expr : expression = 
            change_index_of_variables_in_expression (true) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (sref_expr)
              (uniqueness_hash)
        in
        
        (* Step2. Find all vars pointing to this str_expr *)
        let points_to_str : expression list = 
            find_all_points_to_var (old_str_expr) (int_hash) (var_to_index)
              (index_to_var)
        in
        let based_points_to_str : expression list = 
            List.map (create_based_vars_in_expression (*false*))
                     (points_to_str)
        in
        let srefp_constraints : predicate list ref = ref [] in
        (* foreach point_to_str, create new srefp & Constraints for them *)
        (try
            (List.iter2
                (fun a_point_to_str a_based_point_to_str -> 
                    (* Create new srefp *)
                    let based_srefp : expression = 
                        Application ("srefp", [a_based_point_to_str; field_expr])
                    in
                    let new_srefp : expression = 
                        change_index_of_variables_in_expression (true) (index_hash) 
                          (int_hash) (var_to_index) (index_to_var) 
                          (Hashtbl.create(0)) (ref []) (based_srefp)
                          (uniqueness_hash)
                    in
                    (* Create condition *)
                    let condition : predicate = 
                        Eq (a_point_to_str, 
                            Application ("addrof", [first_index_str_expr(*old_str_expr*)]))
                    in
                    (* Create reset predicate *)
                    let reset_pred : predicate = Eq (new_srefp, new_sref_expr) in
                    (* Create implication & update *)
                    let implication = Or [Not condition; reset_pred] in
                    if is_flatten = true then 
                        srefp_constraints := !srefp_constraints @ [reset_pred]
                    else srefp_constraints := !srefp_constraints @ [implication]
                )
                points_to_str
                based_points_to_str
            )
        with
        | Invalid_argument _ -> assert false);
          
        (* Restore binding of sref_expr in index_hash *)
        let sref_key = Application ("sref", [old_str_expr; field_expr]) in
        let new_version : int = try Hashtbl.find (index_hash) (sref_key)
                                with
                                | Not_found -> assert false
        in
        let old_version = new_version - 1 in
        Hashtbl.remove (index_hash) (sref_key);
        Hashtbl.add (index_hash) (sref_key) (old_version);
        (* Generate value_constraints: *)
        let value_constraints : predicate list = 
            reset_value_of_variable (index_hash) (int_hash) (var_to_index)
              (index_to_var) (is_flatten) (sref_expr) (uniqueness_hash)
        in
        (* Return *)
        value_constraints @ !srefp_constraints
      end
    | anything_else -> assert false
end
    
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of srefp                               *)
(* This function is used in change_index_of_variables_in_assume            *)
(* Input: a srefp need to reset                                            *)
(* Output: constraints to reset the srefp                                  *)
and reset_value_of_srefp (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (is_flatten : bool)
      (srefp_expr : expression)
      (check_pointer : bool)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    (* debug *)
    (*debug ("reset_value_of_srefp: " ^ 
                   (string_of_expression srefp_expr));*)
    match srefp_expr with
    | Application ("srefp", [pointer_expr; field_expr]) -> begin 
        (* Create old_pointer_expr *)
        let old_pointer_expr : expression = 
            change_index_of_variables_in_expression (false) (index_hash) 
              (int_hash) (var_to_index) (index_to_var) 
              (Hashtbl.create(0)) (ref []) (pointer_expr)
              (uniqueness_hash)
        in
        (* Find index of the pointer: *)
        let index_of_pointer : int = 
            try Hashtbl.find (var_to_index) (old_pointer_expr)
            with
            | Not_found -> -1 
        in
        if index_of_pointer >= 0 then begin
            (* Check if pointer_expr points to a var or not *)
            let (points_to_pointer, pointer_points_to) = 
                try Hashtbl.find (int_hash) (index_of_pointer)
                with
                | Not_found -> assert false
            in
            if pointer_points_to >= 0 then begin
                (* Find var which pointer points to *)
                let pointed_var = 
                    try Hashtbl.find (index_to_var) (pointer_points_to) 
                    with
                    | Not_found -> assert false
                in
                let based_var : expression = 
                    create_based_vars_in_expression (*false*) (pointed_var) in
                (* Create sref *)
                let sref_expr = Application ("sref", [based_var; field_expr]) in
                (* Reset value of sref *)
                let reset_sref_constraints : predicate list = 
                    reset_value_of_sref (index_hash) (int_hash) (var_to_index)
                      (index_to_var) (is_flatten) (sref_expr) (uniqueness_hash) in
                (* Create implication for before reset constraints: *)
                (* Create first-index_str *)
                let first_index_pointed_var : expression = 
                    change_index_of_variables_in_expression (false) (Hashtbl.create(0))
                      (Hashtbl.create (0)) (Hashtbl.create (0)) (Hashtbl.create (0))
                      (Hashtbl.create (0)) (ref []) (pointed_var)
                      (Hashtbl.create 0)
                in
                let condition : predicate = 
                    Eq (old_pointer_expr, 
                        Application ("addrof", [first_index_pointed_var(*pointed_var*)]))
                in
                let new_sref_expr : expression = 
                    change_index_of_variables_in_expression (false) (index_hash) 
                      (int_hash) (var_to_index) (index_to_var) 
                      (Hashtbl.create(0)) (ref []) (sref_expr)
                      (uniqueness_hash)
                in
                let new_srefp_expr : expression = 
                    change_index_of_variables_in_expression (false) (index_hash) 
                      (int_hash) (var_to_index) (index_to_var) 
                      (Hashtbl.create(0)) (ref []) (srefp_expr)
                      (uniqueness_hash)
                in
                let reset_pred = Eq (new_sref_expr, new_srefp_expr) in
                let implication : predicate = 
                    Or [Not condition; And (reset_pred :: reset_sref_constraints)]
                in
                (* Return *)
                if is_flatten = true then reset_pred :: reset_sref_constraints
                else [implication] 
            (* If pointer_points_to < 0 *)
            end else begin
                if check_pointer = true then begin
                    (* error ((string_of_expression pointer_expr) ^ 
                                   " points to nowhere");
                    assert false;*)
                    raise (La_exception.Not_point_to (("", -1), pointer_expr, old_pointer_expr, []));
                end else []
            end
        end else []
      end
    | anything_else -> assert false 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of array                               *)
(* This function is used in change_index_of_variables_in_assume            *)
(* Input: a array need to reset                                            *)
(* Output: constraints to reset the array                                  *)
and reset_value_of_array (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (array_expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    let _ = change_index_of_variables_in_expression (true) (index_hash) 
          (int_hash) (var_to_index) (index_to_var) 
          (Hashtbl.create(0)) (ref []) (array_expr)
          (uniqueness_hash)
    in
    []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create predicates to reset value of var                                 *)
(* This function is used in change_index_of_variables_in_assume            *)
(* Input: a var need to reset                                              *)
(* Output: constraints to reset the var                                    *)
and reset_value_of_var_expr (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (expr : expression)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    (* debug *)
    (*debug ("reset_value_of_var_expr: " ^ 
                  (string_of_expression expr));*)
    
    (* Preprocess expr *)
    let (addrof_constraints, new_expr) = 
        preprocess_expr_for_change_index (index_hash) (int_hash) (var_to_index)
          (index_to_var) (expr) (check_pointer) (null_pointer_exceptions) (true)
          (uniqueness_hash) 
    in
    (* debug *)
    (*debug ("addrof_constraints: " ^ (string_of_predicate (And addrof_constraints)));
    debug ("new_expr: " ^ (string_of_expression new_expr));*)
    (* Create constraints for the reset *)
    let reset_constraints : predicate list = 
        (match new_expr with
        | Variable v -> (*[]*)
            reset_value_of_variable (index_hash) (int_hash) (var_to_index) 
              (index_to_var) (is_flatten) (new_expr) (uniqueness_hash)
        | Application ("pref", _) ->
            reset_value_of_pref (index_hash) (int_hash) (var_to_index) 
              (index_to_var) (is_flatten) (new_expr) (check_pointer)
              (uniqueness_hash)
        | Application ("sref", _) -> (*[]*)
            reset_value_of_sref (index_hash) (int_hash) (var_to_index) 
              (index_to_var) (is_flatten) (new_expr) (uniqueness_hash)
        | Application ("srefp", _) ->
            reset_value_of_srefp (index_hash) (int_hash) (var_to_index) 
              (index_to_var) (is_flatten) (new_expr) (check_pointer)
              uniqueness_hash
        | Application ("array", _) ->
            reset_value_of_array (index_hash) (int_hash) (var_to_index)
              (index_to_var) (elements_of_arrays) (is_flatten) (new_expr)
              (uniqueness_hash)
        | Application ("addrof", _) -> []
        | anything_else -> assert false)
    in
    (* Compose constraints & return *)
    if addrof_constraints = [] then reset_constraints
    else [Or [Not (And addrof_constraints); And reset_constraints]]
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and reset_value_of_vars (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (expr_list : expression list)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate list = 
begin
    let reset_constraints : predicate list ref = ref [] in
    (* Foreach expr in expr_list, reset it and update to reset_constraints *)
    List.iter
        (fun an_expr -> 
            let new_constraints : predicate list = 
                reset_value_of_var_expr (index_hash) (int_hash) (var_to_index)
                  (index_to_var) (elements_of_arrays) (is_flatten) 
                  (false) (null_pointer_exceptions) (an_expr) (uniqueness_hash)
            in
            reset_constraints := !reset_constraints @ new_constraints
        )
        expr_list;
    (* Return *)
    !reset_constraints
end 

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Foreach variable in reset-set, increase version of it, and reset        *)
(* points_to information in int_hash, then create constraint for pointers  *)
(* pointing to it                                                          *)
and change_index_of_variables_in_reset
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (pred : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate = 
begin
    (* debug *)
    (*debug ("change_index_of_variables_in_reset");
    debug (string_of_predicate pred);
    read_int();*)
    (* Step1. Find all reset-vars *)
    match pred with
    | Eq (Application ("reset", reset_vars), Application ("", [])) -> begin
        let replaced_reset_vars : expression list ref = ref [] in
        let replaced_constraints : predicate list ref = ref [] in
        (* Step1. Replace pref by array *)
        List.iter 
            (fun (a_reset_var : expression) -> 
                let (new_replaced_var, new_replaced_constraints) = 
                    replace_pointer_by_array_element_in_expression (index_hash)
                      (int_hash) (var_to_index) (index_to_var) (a_reset_var)
                      (uniqueness_hash)
                in
                replaced_reset_vars := !replaced_reset_vars @ [new_replaced_var];
                replaced_constraints := !replaced_constraints @ new_replaced_constraints;
            )
            reset_vars;
        (* Step2. Reset all vars replaced_reset_vars *)
        let reset_constraints : predicate list = 
            reset_value_of_vars (index_hash) (int_hash) (var_to_index)
              (index_to_var) (elements_of_arrays) (is_flatten) 
              (check_pointer) (null_pointer_exceptions) (!replaced_reset_vars)
              (uniqueness_hash) 
        in
        (* Step3. Normalize constraints *)
        let normalized_replaced_constraints : predicate = 
            match !replaced_constraints with
            | [] -> True
            | [an_element] -> an_element
            | list_of_constraints -> And !replaced_constraints
        in
        let normalized_reset_constraints : predicate = 
            match reset_constraints with
            | [] -> True
            | [an_element] -> an_element
            | list_of_constraints -> And reset_constraints
        in
        (* Return *)
        if is_flatten = true then normalized_reset_constraints
        else begin
            if !replaced_constraints = [] then normalized_reset_constraints
            else
                Or [Not normalized_replaced_constraints; normalized_reset_constraints]
        end
      end
    | anything_else -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and change_index_of_variables_in_assume 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (is_flatten : bool)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (pred : predicate)
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate = 
begin
    (* debug *)
    (*debug "change_index_of_variables_in_assume ";
    let _ = read_int() in*)
    
    (* Step1. Replace pref by array *)
    let (replaced_pred, replace_constraints) = 
        replace_pointer_by_array_element_in_predicate (index_hash)
          (int_hash) (var_to_index) (index_to_var) (pred) (uniqueness_hash)
    in
    (* Step2. Change index of variables in assume *)
    let new_assume : predicate = 
        change_index_of_variables_in_condition (index_hash) (int_hash)
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (elements_of_arrays) (false) (check_pointer) 
          (null_pointer_exceptions) (replaced_pred) (uniqueness_hash)
    in
    (*let final_constraints : predicate list = 
        replace_constraints @ reset_constraints in*)
    (* Return *)
    if replace_constraints = [] then begin
        new_assume
        (*if reset_constraints = [] then new_assume
        else And (new_assume :: reset_constraints)*)
    end else if is_flatten = true then begin 
        new_assume
    end else begin
        (match replace_constraints with
        | [] -> new_assume
        | [a_constraint] -> Or [Not a_constraint; new_assume]
        | list_of_constraints -> 
            Or [Not (And replace_constraints); new_assume])
        (*if reset_constraints = [] then 
            Or [Not (condition); new_assume]
        else
            Or [Not (condition); And (new_assume :: reset_constraints)]*)
    end
    (*if final_constraints = [] then new_assume
    else And ([new_assume] @ final_constraints)*)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and change_index_of_variables_in_non_condition_predicate 
      (index_hash : (expression, int) Hashtbl.t)
      (int_hash : (int, (int list) * int) Hashtbl.t)
      (var_to_index : (expression, int) Hashtbl.t)
      (index_to_var : (int, expression) Hashtbl.t) 
      (structs_hash : (expression, expression list) Hashtbl.t)
      (current_vars_in_fun : expression list ref)
      (is_flatten : bool)
      (array_asgn_is_flatten : bool)
      (elements_of_arrays : (expression, expression list) Hashtbl.t)
      (check_pointer : bool)
      (null_pointer_exceptions : (expression * expression * (expression list)) list ref)
      (var_eq_hash : (expression, expression) Hashtbl.t)
      (pred : predicate) 
      (uniqueness_hash : (int, expression) Hashtbl.t)
: predicate =
begin
    match pred with
    | True -> True
    | False -> False
    (* 14.4.2009. don't need deal with "assume" here *)
    (* 24.2.2009. Added the following to deal with "assume" *)
    | Not _ ->
        change_index_of_variables_in_assume (index_hash) (int_hash)
          (var_to_index) (index_to_var) (structs_hash) (current_vars_in_fun)
          (elements_of_arrays) (is_flatten) (check_pointer) (null_pointer_exceptions)
          (pred) (uniqueness_hash)
    | Or _ | Lt _ | Leq _ | Atom _ -> assert false
    | And pred_list (* list of assignments *) -> 
        begin
            let temp_fun (pred : predicate) : predicate list = 
                change_index_of_variables_in_an_assignment (index_hash) 
                  (int_hash) (var_to_index) (index_to_var) (structs_hash) 
                  (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
                  (false)(elements_of_arrays) (check_pointer) 
                  (null_pointer_exceptions) (var_eq_hash) (pred)
                  (uniqueness_hash)
            in 
            let new_pred_list_list : (predicate list) list = 
                List.map (temp_fun) (pred_list) in
            let new_pred_list : predicate list = 
                List.flatten (new_pred_list_list) in 
            And new_pred_list
        end
    | Eq (left, right) (* an assignment *) ->
        begin
            let pred_list_of_eq : predicate list =
                change_index_of_variables_in_an_assignment (index_hash) 
                  (int_hash) (var_to_index) (index_to_var) (structs_hash) 
                  (current_vars_in_fun) (is_flatten) (array_asgn_is_flatten)
                  (false) (elements_of_arrays) (check_pointer) 
                  (null_pointer_exceptions) (var_eq_hash) (Eq (left, right))
                  (uniqueness_hash)
            in
            And pred_list_of_eq
        end
end

let rec make_assignments (parameters: string list)
                         (arguments: expression list)
: predicate list =
begin
    match (parameters, arguments) with
    | ([], [])       -> [];
    | ([], an_argu :: argus)    -> assert false;
    | (a_para :: paras, _) -> (
        if (((String.length a_para) >= 3)
              && (String.sub a_para 0 3 = Cexpr.ellipsis_string)) then
            []
        else
        begin
            match arguments with
            | [] -> assert false;
            | an_argu :: argus -> (
                let an_assignment = Eq (Variable a_para, an_argu) in
                [an_assignment] @ (make_assignments paras argus)
              );
        end
          )
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let assign_arguments_for_parameters (cfa_of_called_function : cfa)
                                    (cfa_call : cfa_stmt)
                                    new_tail
                                    func_level
: predicate list =
begin
    (* Find arguments: *) 
    let arguments : expression list =
        (match cfa_call with
        | CFA_CALL (fc, _) -> (match fc with
                         | _ :: args -> args
                         | [] -> assert false)
        | _ -> assert false)
    in
    (* Find parameters: *)
    let string_parameters : string list =
        (match cfa_of_called_function with
        | CFA_FUNDEF f -> f.cfa_f_meta_data.cfa_params
        (* 20090704 maybe CFA_VARDEF *)
        | CFA_VARDEF v -> v.cfa_v_meta_data.cfa_params
        | _ -> assert false)
    in
    let string_parameters =
         List.map (fun str ->
             concat_tail_with_a_symbol str new_tail func_level
    ) string_parameters in
    (* return *)
    make_assignments string_parameters arguments;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let assign_value_from_return_to_var (fc_cfa_node : cfa_node)
                                    (index_of_function : int) 
                                    (cfa_fundef : cfa_fundef_t)
: predicate option =
begin
    (* check if fc_cfa_node is CFA_CALL or not: *)
    (match fc_cfa_node.cfa_node_stmt with
    | CFA_CALL _ -> ()
    | anything_else -> assert false);
    (* find name of left_var of assignment: *)
    let name_of_left_var : string = try List.hd (fc_cfa_node.cfa_outputs)
                                    with
                                    | Failure f -> ""
    in
    if (String.compare name_of_left_var  "") != 0 then begin
        (* create assignment *)
        let new_left_var = Variable (name_of_left_var) in
        let new_right_var = 
            (*Variable ("#ACC%0$" ^ string_of_int index_of_function) in*)
            Variable ("$result%1") in
        let assignment = Eq (new_left_var (*left_expr*), new_right_var) in
        (* return *)
        Some (assignment)
    end else None
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(*let is_noreturn (current_cfa : cfa_fundef_t) : bool =
begin
    (* debug debug ("is_noreturn"); read_int(); *)
    
    let flag = ref false in
    List.iter
        (fun a_cfa_attribute ->
            if a_cfa_attribute = CFA_NORETURN_ATTR then
                flag := true
        )
        current_cfa.cfa_f_meta_data.cfa_attributes;
      
    (* return *)
    !flag
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_next_nodes_and_edges_for_cfa_return_node (current_cnode : cfa_node) 
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    (* Step1. Create some vars: *)
    let new_call_stack = ref call_stack in
    let new_current_cfa = ref current_cfa in
    let outgoing_tuples = ref [] in
    
    (* Step2. Check if current_cnode is CFA_RETURN or not *)
    (match current_cnode.cfa_node_stmt with
    | CFA_RETURN -> ()
    | anything_else -> assert false);
      
    (* Step3. Check in call_stack: *)
    (*if (is_noreturn current_cfa) != true then begin*)
    (match call_stack with
    | [] -> (* CFA_RETURN in main-func *) outgoing_tuples := [] 
    | (index_of_cfa, cnode_in_stack, cfa_in_stack) :: tail_of_stack_list-> begin
        (* find outgoing_label: *)
        let outgoing_label = try List.hd (cnode_in_stack.cfa_outgoing_labels)
                             with
                             | Failure _ -> assert false
        in
        (* find outgoing_node in cfa: *)
        let outgoing_node =
            try Hashtbl.find (cfa_in_stack.cfa_node_hashtbl) (outgoing_label)
            with
            | Not_found -> assert false
        in
        (* update information: *)
        new_current_cfa := cfa_in_stack;
        new_call_stack := tail_of_stack_list;
        (* create assignment, which assigns var of return to var in funcall *)
        let assignment_option = 
            assign_value_from_return_to_var (cnode_in_stack) 
              (index_of_cfa) (cfa_in_stack)
        in
        (* update information: *)
        (match assignment_option with
        | Some a_assignment -> outgoing_tuples := [outgoing_node, a_assignment];
        | None -> outgoing_tuples := [outgoing_node, True]);
      end);
    (*end else begin
    (* debug debug "is_noreturn"; read_int(); *)
        new_call_stack := [];
    end*)
      
    (* return *)
    (!outgoing_tuples, !new_call_stack, !new_current_cfa);
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_next_nodes_and_edges_for_cfa_reset_node (current_cfa_node : cfa_node)
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    (* Step1. Create some vars to return: *)
    let new_call_stack = ref call_stack in
    let new_current_cfa = ref current_cfa in
    let outgoing_tuples = ref [] in
    (* Step2. Check if current_cfa_node is CFA_RESET or not: *)
    (match current_cfa_node.cfa_node_stmt with
    | CFA_RESET reset_vars -> begin
        (* CFA_RESET has only one outgoing: *)
        let outgoing_label = try List.hd (current_cfa_node.cfa_outgoing_labels)
                             with
                             | Failure _ -> assert false
        in
        let new_child : cfa_node = 
            try Hashtbl.find (current_cfa.cfa_node_hashtbl) (outgoing_label)
            with
            | Not_found -> assert false
        in
        (* Create temp_reset_predicate *)
        let temp_reset_predicate : predicate = 
            Eq (Application ("reset", reset_vars), Application ("", []))
        in
        (* Update into outgoing_tuples to return: *)
        outgoing_tuples := [new_child, temp_reset_predicate];
      end
    | anything_else -> assert false);
    (* Return *)
    (!outgoing_tuples, !new_call_stack, !new_current_cfa);  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_next_nodes_and_edges_for_cfa_assume_node (current_cfa_node : cfa_node)
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    (* debug *)
    (*debug ("get_next_nodes_and_edges_for_cfa_assume_node");
    debug (get_string_of_cfa_node current_cfa_node);*)

    (* Step1. Create some vars to return: *)
    let new_call_stack = ref call_stack in
    let new_current_cfa = ref current_cfa in
    let outgoing_tuples = ref [] in
    (* Step2. Check if current_cfa_node is CFA_ASSUME or not: *)
    (match current_cfa_node.cfa_node_stmt with
    | CFA_ASSUME cfa_ass -> begin
        (* CFA_ASSUME has only one outgoing: *)
        let outgoing_label = try List.hd (current_cfa_node.cfa_outgoing_labels)
                             with
                             | Failure _ -> assert false
        in
        let new_child : cfa_node = 
            try Hashtbl.find (current_cfa.cfa_node_hashtbl) (outgoing_label)
            with
            | Not_found -> assert false
        in
        (* Update into outgoing_tuples to return: *)
        outgoing_tuples := [new_child, cfa_ass];
      end
    | anything_else -> assert false);
    (* Return *)
    (!outgoing_tuples, !new_call_stack, !new_current_cfa);      
end
      
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Note: if we see "elsewhere", we must check in call_stack to know where  *)
(* to go (where calls current function)                                    *)
let get_next_nodes_and_edges_for_cfa_form_node (current_cfa_node : cfa_node) 
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    (* Step1. Create some vars to return: *)
    let new_call_stack = ref call_stack in
    let new_current_cfa = ref current_cfa in
    let outgoing_tuples = ref [] in
    
    (* Step2. Check if cfa_node is CFA_FORM or not: *)
    (match current_cfa_node.cfa_node_stmt with
    | CFA_ENTRY cfa_form | CFA_FORM cfa_form ->
        begin
            (* CFA_FORM has only one outgoing: *)
            let out_going_label =
                try List.hd (current_cfa_node.cfa_outgoing_labels)
                with
                | Failure _ -> 
                    (*let message = "not found outgoing_label" in*) 
                    assert false
                    (*raise (ConfigUtil.Message (Logging.Error, message))*)
            in
            (* Find next node: *)
            let new_child : cfa_node =
                try Hashtbl.find (current_cfa.cfa_node_hashtbl)(out_going_label)
                with
                | Not_found ->
                    (*let message = "not found outgoing_node" ^ out_going_label in*)
                    assert false 
                    (*raise (ConfigUtil.Message (Logging.Error, message));*)
            in
            (* Update into outgoing_tuples to return: *)
            outgoing_tuples := [new_child, cfa_form];
        end
    | _ -> assert false);
    (*raise (ConfigUtil.Message (Logging.Error, "isn't CFA_FORM")));*)
    
    (* Return *)
    (!outgoing_tuples, !new_call_stack, !new_current_cfa);
end
  
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* In condition, outgoings can not elsewhere. Therefore, if we see         *)
(* elsewhere, assert false.                                                *)
(* IMPORTANT: When we process a condition-node, what should we do? We have *)
(* 2 options:                                                              *)
(*     1. Goto "True-edge" first, "False-edge" second                      *)
(*     2. Goto "False-edge" firse, "True-edge" second                      *)
let get_next_nodes_and_edges_for_cfa_cond_node (current_cfa_node : cfa_node) 
                                               (current_cfa : cfa_fundef_t) 
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) =
begin 
    (* Create var which contains (child, edge1, edge2) to return: *)
    let tuples_to_return = ref [] in
    (* Check if current_cfa_node is CFA_COND or not: *)
    (match current_cfa_node.cfa_node_stmt with
    | CFA_COND (cfa_cond, s1, s2) -> begin
        (* =================== Temporarily do not use it ================== *)
        (* Find child2 & update to tuples_to_return: *)
        (*let new_child_2 : cfa_node =
            try Hashtbl.find (current_cfa.cfa_node_hashtbl) (s2)
            with
            | Not_found ->
                let message = "not found outgoing_node" ^ s2 in
                raise (ConfigUtil.Message (Logging.Error, message));
        in
        tuples_to_return := !tuples_to_return @ [new_child_2, Not cfa_cond];
        (* Find child1 & update to tuples_to_return: *)
        let new_child_1 =
          try Hashtbl.find (current_cfa.cfa_node_hashtbl) (s1)
          with
          | Not_found ->
              let message = "not found outgoing_node" ^ s1 in
              raise (ConfigUtil.Message (Logging.Error, message));
        in*)
        (* ================================================================ *)
        (* 20.2.2009. Xuan checked s1, s2 *)
        let true_label = ref "" in
        let false_label = ref "" in
        (match current_cfa_node.cfa_outgoing_labels with
        | [ol1; ol2] -> true_label := ol1; false_label := ol2;
        | anything_else -> assert false);
        (* Find child2 & update to tuples_to_return: *)
        let new_child_2 : cfa_node =
            try Hashtbl.find (current_cfa.cfa_node_hashtbl) (!false_label)
            with
            | Not_found ->
                (*let message = "not found outgoing_node" ^ !false_label in*)
                assert false
                (*raise (ConfigUtil.Message (Logging.Error, message));*)
        in
        tuples_to_return := (*!tuples_to_return @*)
            [new_child_2, Not cfa_cond];
        (* Find child1 & update to tuples_to_return: *)
        let new_child_1 =
          try Hashtbl.find (current_cfa.cfa_node_hashtbl) (!true_label)
          with
          | Not_found ->
              (*let message = "not found outgoing_node" ^ !true_label in*)
              assert false
              (*raise (ConfigUtil.Message (Logging.Error, message));*)
        in
        (*tuples_to_return := !tuples_to_return @ [new_child_1, cfa_cond];*)
        tuples_to_return := [new_child_1, cfa_cond]
                                @ !tuples_to_return;
      end
    | anything_else ->
        assert false
        (*raise (ConfigUtil.Message (Logging.Error, "isn't a CFA_COND"));*)
    );
    
    (* return *)
    !tuples_to_return
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_next_node_and_edge_for_nodef_func_call (exprs_of_fc : expression list) 
                                               (current_cnode : cfa_node) 
                                               (current_cfa_fundef : cfa_fundef_t) 
                                               (funcall_number : int) 
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) = 
begin
    (* debug *)
    (*debug ("called-function is not found: " ^ 
                  (string_of_expression (List.hd exprs_of_fc)));*)
    (*assert false;*)
    (* Step1. define var to return: *)
    let outgoing_tuples = ref [] in
    
    (* Step2. Find output & outgoing: *)
    let current_output : string = try List.hd (current_cnode.cfa_outputs)
                                  with | Failure _ -> ""
    in
    let out_going_label = try List.hd (current_cnode.cfa_outgoing_labels)
                          with | Failure _ -> assert false
    in
    let new_child : cfa_node = 
        try Hashtbl.find (current_cfa_fundef.cfa_node_hashtbl)(out_going_label)
        with Not_found -> assert false
    in
    (* Step3. Create pseudo assignment *)
    let expr_of_func = try List.hd exprs_of_fc
                       with Failure _ -> assert false
    in
    (*let name_of_func = 
      (string_of_expression expr_of_func) ^ "$" ^ (string_of_int funcall_number) 
    in*)
    (* 20090629. Create new pseudo assignment *)
    let sfn : string = string_of_int funcall_number in
    let new_assignment : predicate ref = ref True in
    let new_expr_of_func : expression = 
        (match expr_of_func with
        | Variable _ | Application ("sref", (Variable _) :: _) ->
            expr_of_func
        | anything_else -> begin
            let temp_var : expression = 
                Variable ("temp_var: " ^ string_of_expression expr_of_func) in 
            new_assignment := Eq (temp_var, expr_of_func);
            temp_var
          end)
    in
    (* Create new function_call *)
    let funcall_asgn : predicate = 
        Eq (Variable current_output, 
            Application ("$function_call$", 
                         (Variable sfn) :: new_expr_of_func :: (List.tl exprs_of_fc)))
    in
    (* NOTE: Create new assignment to denote name of function *)
    let denoted_asgn = 
        Eq (Variable "temp_out", create_based_vars_in_expression new_expr_of_func) 
    in
    (* Check new_assignment & update to eq_to_return *)
    let eq_to_return : predicate = 
        (match current_cnode.cfa_node_stmt with
        | CFA_FORM (And cfs) -> begin
            if !new_assignment = True then begin
                And (denoted_asgn :: cfs @ [funcall_asgn])
            end else begin
                And (denoted_asgn :: cfs @ [!new_assignment; funcall_asgn])
            end;
          end
        | CFA_CALL _ -> begin
            if !new_assignment = True then funcall_asgn
            else And [denoted_asgn; !new_assignment; funcall_asgn]
          end
        | _ -> assert false)
    in
    outgoing_tuples := [new_child, eq_to_return (*new_eq*)];
    (* Step4. return *)
    !outgoing_tuples
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in function "find_function_in_cfa_call" to check  *)
(* if calling function maybe points to somewhere                           *)
(*let rec function_pointer_in_current_rr_points_to (current_rr : predicate)
                                                 (possible_fp : expression)
: string option (* where function pointer points to *) = 
begin
    (* debug *)
    (*debug "function_pointer_in_current_rr_points_to";
    debug ("possible_fp: " ^  (string_of_expression possible_fp));
    debug ("current_rr: " ^ (string_of_predicate current_rr));*)
    match current_rr with
    | True | False -> None  
    | And pred_list | Or pred_list -> begin
        let fc_result : string option ref = ref None in
        (* Foreach apred in pred_list, check if it has an assignment assigned *)
        (* a function for possible_function pointer,  then return             *)
        List.iter
            (fun apred ->
                (* Check if fc_result != None or not *)
                if !fc_result = None then begin
                    fc_result := 
                        function_pointer_in_current_rr_points_to (apred) 
                          (possible_fp);    
                end
            )
            pred_list;
        (* Return *)
        !fc_result  
      end
    | Not apred -> None
        (*function_pointer_in_current_rr_points_to (apred) (possible_fp)*)
    | Eq (expr1, expr2) -> begin
        (* Check if expr1 is possible_fp, then expr2 is where it points to *)
        (if (compare_expr_with_expr expr1 possible_fp) = 0 then begin
            (match expr2 with
            | Variable v -> Some v
            (* 20090619 Xuan fix this bug *)
            | _ -> None)
        (* Check if expr2 is possible_fp, then expr1 is where it points to *)
        end else if (compare_expr_with_expr expr2 possible_fp) = 0 then begin
            (match expr1 with
            | Variable v -> Some v
            (* 20090619 Xuan fix this bug *)
            | _ -> None)
        end else None)
      end
    | Lt _ | Leq _ | Atom _ -> None
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in function "find_function_in_cfa_call" to check  *)
(* if calling function maybe points to somewhere                           *)
(* Note that possible_fp may be Varible, Sref, or Srefp,...  *)
let rec function_pointer_in_current_rr_points_to (current_rr : predicate)
                                                 (possible_fp : expression)
: string option (* where function pointer points to *) = 
begin
    (* debug *)
    (*debug ("function_pointer_in_current_rr_points_to: " );
    debug ("current_rr: " ^ (string_of_predicate current_rr));
    debug ("possible_fp: " ^ (string_of_expression possible_fp));*)

    match possible_fp with
    | Constant c -> assert false
    | Variable v -> begin
        find_value_of_var_in_pred (current_rr) (possible_fp)
      end;
    | Application _ -> begin
        (* First, create hash to save point-to informations *)
        let index_hash : (expression, int) Hashtbl.t = Hashtbl.create (0) in
        let int_hash : (int, (int list) * int) Hashtbl.t = Hashtbl.create (0) in
        let var_to_index : (expression, int) Hashtbl.t = Hashtbl.create (0) in
        let index_to_var : (int, expression) Hashtbl.t = Hashtbl.create (0) in
        let structs_hash : (expression, expression list) Hashtbl.t =
            Hashtbl.create (0) in
        let elements_of_arrays : (expression , expression list) Hashtbl.t = 
            Hashtbl.create (0) in
        let uniqueness_hash : (int, expression) Hashtbl.t = Hashtbl.create 0 in
        (* Second, change index of current_rr to update the hashtbls *)
        let _ = 
            change_index_of_variables_in_condition (index_hash) (int_hash)
              (var_to_index) (index_to_var) (structs_hash) (ref [])
              (elements_of_arrays) (true) (false) (ref []) (current_rr)
              (uniqueness_hash)
        in 
        (* Third, preprocess possible_fp *)
        let (conditions, pps_possible_fp) = 
            preprocess_expr_for_change_index (index_hash) (int_hash) 
              (var_to_index) (index_to_var) (possible_fp) (false)
              (ref []) (false) (uniqueness_hash)
        in
        find_value_of_var_in_pred (current_rr) (pps_possible_fp);
      end
    | Sum _ | Coeff _ -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let find_name_and_level_of_expr_func (expr_func : expression)
: string * int = 
begin
    (match expr_func with
    | Variable string_func -> (begin
        (* FIXME rindex or lindex??? *)
        let left_index : int = try String.rindex string_func '%'
                                with
                                | Not_found -> assert false
        in
        (* find name & level of the function *)
        let func_name : string = (try String.sub string_func 0 left_index
                                 with 
                                 | Invalid_argument _ -> assert false) 
        in
        let func_level : int = 1 in
        (func_name, func_level)
      end)
    | anything_else -> begin (*assert false*)
        (* 20090627. Maybe sref *)
        ("", -1)
      end
    );
end

let create_require_label (func_name : string) (funcall_number : int)
   (count : int) : string = (
    "##check_requires-" ^ func_name ^ "-" ^ (string_of_int funcall_number)
    ^ "-" ^ (string_of_int count)
)

let create_ensure_label (func_name : string) (funcall_number : int) (count : int)
: string = (
    "FP-" ^ func_name ^ "-" ^ (string_of_int funcall_number)
    ^ "Ensures-" ^ (string_of_int count)
)

let create_reset_label (func_name : string) (funcall_number : int) (count : int)
: string = (
    "FP-" ^ func_name ^ "-" ^ (string_of_int funcall_number)
    ^ "Reset" ^ "-" ^ (string_of_int count)
)

let create_assign_label (func_name : string) (funcall_number : int)
: string = (
    "FP-" ^ func_name ^ "-" ^ (string_of_int funcall_number)
    ^ "Assigns-parameters"
)

let replace_node_label (cfa_nodes : (string, cfa_node) Hashtbl.t)
    (funcall_number : int)
: (string, cfa_node) Hashtbl.t = (
    let replace_label (label : string) (funcall_number : int) : string = (
        if ((String.compare label "#Lreturn") = 0)
           or ((String.compare label "#Lentry") = 0)
           or ((String.compare label "#Lelsewhere") = 0) then
            label
        else
            label ^ "-" ^ (string_of_int funcall_number)
    ) in
    let replace_labels (labels : string list) (funcall_number : int)
    : string list = (
        List.map (fun label -> replace_label label funcall_number) labels
    ) in
    let new_cfa_nodes = Hashtbl.copy cfa_nodes in
    Hashtbl.iter (
        fun node_label cfa_node -> (
            let new_node_label = replace_label node_label funcall_number in
            let new_outgoing_labels =
                replace_labels cfa_node.cfa_outgoing_labels
                               funcall_number in
            let new_incoming_labels =
                replace_labels cfa_node.cfa_incoming_labels
                               funcall_number in
            let new_node =
              { cfa_node with cfa_label = new_node_label;
                              cfa_outgoing_labels = new_outgoing_labels;
                              cfa_incoming_labels = new_incoming_labels } in
            Hashtbl.remove new_cfa_nodes node_label;
            Hashtbl.add new_cfa_nodes new_node_label new_node;)
    ) cfa_nodes;
    new_cfa_nodes
)

(** translate expr_loc to memo *)
let expr_loc_to_memo (expr_loc : Lexing.position * Lexing.position)
: string * string * string * string * string = (
    let (start_pos, end_pos) = expr_loc in
    (start_pos.Lexing.pos_fname,
     string_of_int start_pos.Lexing.pos_lnum,
     string_of_int start_pos.Lexing.pos_cnum,
     string_of_int end_pos.Lexing.pos_lnum,
     string_of_int end_pos.Lexing.pos_cnum)
)

(** insert reset to after function call in cutlist *)
let rec inline_resets (current_cfa : cfa) (funcall_node_name : string)
                      (func_name : string) (funcall_number : int)
                      ?(count = 0) (agent_exprs : agent_expr_t list)
: cfa = (
    match agent_exprs with
    | [] -> current_cfa
    | agent_expr :: remain_agent_exprs -> (
        let reset = LogicUtil.Logic.Variable agent_expr.agent_temp_varid in
        let new_memo = [expr_loc_to_memo agent_expr.agent_expr_loc] in
        let new_reset_label : string =
            create_reset_label func_name funcall_number count
        in
        let incoming_label = funcall_node_name in
        let funcall_node = find_node_in_cfa current_cfa funcall_node_name in
        let outgoing_labels = funcall_node.cfa_outgoing_labels in
        assert ((List.length outgoing_labels) = 1);
        let outgoing_label = List.hd outgoing_labels in
        let assign_nodes = agent_expr.agent_assign_nodes in
        let assign_nodes = replace_node_label assign_nodes funcall_number in
        let new_current_cfa = inline_insert_cfa assign_nodes current_cfa
                                                funcall_node_name in

        inline_resets new_current_cfa funcall_node_name func_name funcall_number
                      ~count:(count+1) remain_agent_exprs)
)

(** insert postcondition into current_cfa at node has name is funcall_node_name *)
let rec inline_ensures (current_cfa : cfa) (funcall_node_name : string)
                       (func_name : string) (funcall_number : int)
                       ?(count = 0) (agent_exprs : agent_expr_t list)
: cfa = (
    match agent_exprs with
    | [] -> current_cfa
    | agent_expr :: remain_agent_exprs -> (
        let assign_nodes = agent_expr.agent_assign_nodes in
        let ensure = LogicUtil.Logic.Not (LogicUtil.Logic.Eq (
                          LogicUtil.Logic.Variable agent_expr.agent_temp_varid,
                          LogicUtil.Logic.Constant 0.0)) in
        let new_memo = [expr_loc_to_memo agent_expr.agent_expr_loc] in
        (* insert ensure node to cfa after function call node *)
        let new_ensure_label : string =
            create_ensure_label func_name funcall_number count
        in
        let incoming_label = funcall_node_name in
        let funcall_node = find_node_in_cfa current_cfa funcall_node_name in
        let outgoing_labels = funcall_node.cfa_outgoing_labels in
        assert ((List.length outgoing_labels) = 1);
        let outgoing_label = List.hd outgoing_labels in
        let new_current_cfa =
            insert_edge_into_cfa_fundef current_cfa incoming_label outgoing_label
                (CFA_ASSUME (ensure))
                new_ensure_label new_memo in
        (* insert assign nodes to after function call node *)
        let assign_nodes = replace_node_label assign_nodes funcall_number in
        let new_current_cfa = inline_insert_cfa assign_nodes new_current_cfa
                                                funcall_node_name in
        inline_ensures new_current_cfa funcall_node_name func_name funcall_number
                       ~count:(count+1) remain_agent_exprs;)
)

(** insert precondition of function has name is func_name into current_cfa
    at node has name is funcall_node_name *)
let rec inline_asserts (current_cfa : cfa) (funcall_node_name : string)
                       (func_name : string) (funcall_number : int)
                       ?(count = 0) (agent_exprs : agent_expr_t list)
: cfa = (
    match agent_exprs with
    | [] -> current_cfa
    | agent_expr :: remain_agent_exprs -> (
        let assign_nodes = agent_expr.agent_assign_nodes in
        let require = LogicUtil.Logic.Not (LogicUtil.Logic.Eq (
                          LogicUtil.Logic.Variable agent_expr.agent_temp_varid,
                          LogicUtil.Logic.Constant 0.0)) in
        (* insert reqquire *)
        let incoming_label = funcall_node_name in
        let funcall_node = find_node_in_cfa current_cfa funcall_node_name in
        let outgoing_labels = funcall_node.cfa_outgoing_labels in
        assert ((List.length outgoing_labels) = 1);
        let outgoing_label = List.hd outgoing_labels in
        (* Create new function call *)
        let start_pos, end_pos = agent_expr.agent_expr_loc in
        let range = Some (Sexpr.make_range start_pos end_pos) in
        let new_current_cfa =
            inline_assert_func_into_cfa_fundef current_cfa incoming_label
                outgoing_label funcall_number count
                (require) range in
        (* insert assign nodes *)
        let assign_nodes = replace_node_label assign_nodes funcall_number in
        let new_current_cfa = inline_insert_cfa assign_nodes new_current_cfa
                                                funcall_node_name in
        inline_asserts new_current_cfa funcall_node_name func_name
                       funcall_number ~count:(count+1) remain_agent_exprs
      )
)

(** replace old_exprs in cfa_stmt by new_exprs *)
let replace_exprs_in_cfa_stmt (old_exprs : LogicUtil.Logic.expression list)
                              (new_exprs : LogicUtil.Logic.expression list)
                              (cfa_stmt : cfa_stmt)
: cfa_stmt = (
    match cfa_stmt with
    | CFA_FORM pred ->
        CFA_FORM (replace_exprs_in_predicate old_exprs new_exprs pred)
    | CFA_COND (pred, str1, str2) ->
        CFA_COND (replace_exprs_in_predicate old_exprs new_exprs pred, str1, str2)
    | CFA_CALL (exprs, agent_fspec_op) ->
        CFA_CALL (replace_exprs_in_expressions old_exprs new_exprs exprs, agent_fspec_op)
    | CFA_ASSUME pred ->
        CFA_ASSUME (replace_exprs_in_predicate old_exprs new_exprs pred)
    | CFA_RESET exprs ->
        CFA_RESET (replace_exprs_in_expressions old_exprs new_exprs exprs)
    | CFA_ENTRY pred ->
        CFA_ENTRY (replace_exprs_in_predicate old_exprs new_exprs pred)
    | CFA_RETURN
    | CFA_ERROR
    | CFA_FAIL -> cfa_stmt
)

(** replace old_exprs in agent_expr_t by new_exprs *)
let replace_exprs_in_agent_expr (old_exprs : LogicUtil.Logic.expression list)
                                (new_exprs : LogicUtil.Logic.expression list)
                                (agent_expr : agent_expr_t)
: agent_expr_t = (
    match agent_expr with
      {
        agent_assign_nodes = assign_nodes;
        agent_temp_varid = var_name;
        agent_expr_loc = memo
      } -> (
        let new_assign_nodes = Hashtbl.create (Hashtbl.length assign_nodes) in
        Hashtbl.iter (fun str assign_node ->
            let cfa_stmt = assign_node.cfa_node_stmt in
            let new_cfa_stmt = replace_exprs_in_cfa_stmt old_exprs new_exprs
                                                         cfa_stmt in
            let new_assign_node =
                {assign_node with cfa_node_stmt = new_cfa_stmt} in
            Hashtbl.add new_assign_nodes str new_assign_node
        ) assign_nodes;
        let new_var_name = (
            match replace_exprs_in_expression old_exprs new_exprs
                       (LogicUtil.Logic.Variable var_name) with
            | LogicUtil.Logic.Variable new_var_name_ -> new_var_name_
            | _ -> assert false) in
        {
          agent_assign_nodes = new_assign_nodes;
          agent_temp_varid = new_var_name;
          agent_expr_loc = memo
        }
    )
)

let get_info_cutlist_of_cfa (cfa : cfa)
: (agent_fspec_t * old_term_t) option = (
    let meta_data = (match cfa with
                    | CFA_FUNDEF f -> f.cfa_f_meta_data
                    | CFA_VARDEF v -> v.cfa_v_meta_data
                    | _ -> assert false) in
    meta_data.cfa_specs
) 

let get_funcall_number (slabel : string)
: int = (
    try int_of_string (List.hd (List.rev (ExtString.String.nsplit slabel "*")))
    with _ -> assert false
)

let replace_return (func_call_node : cfa_node)
                   (info_cutlist : agent_expr_t list * agent_expr_t list *
                                   agent_expr_t list)
                   (new_tail : string)
                   (func_level : int)
: agent_expr_t list * agent_expr_t list * agent_expr_t list = (
    (* Find outgoing of the function *)
    let outgoing_expr : expression option =
        try Some (Variable (List.hd (func_call_node.cfa_outputs)))
        with Failure f -> None
    in
    (* get new names of parameters and output of function*)
    let replacing_vars, exprs_to_replace =
        (match outgoing_expr with
        | None ->
            ([], [])
        | Some out -> (
            ([out], [Variable (concat_tail_with_a_symbol
                                   (Ccommon.return_expr_name ^ "%3")
                                   new_tail
                                   func_level)]))) in
    let requires, assigns, ensures = info_cutlist in
    (List.map (replace_exprs_in_agent_expr exprs_to_replace replacing_vars)
         requires,
     List.map (replace_exprs_in_agent_expr exprs_to_replace replacing_vars)
         assigns,
     List.map (replace_exprs_in_agent_expr exprs_to_replace replacing_vars)
         ensures)
)

(** remove node in current_cfa has name move_node_name *)
let remove_cfa_node (current_cfa : cfa) (move_node_name : string) : cfa = (
    let move_node = find_node_in_cfa current_cfa move_node_name in
    let incomings = move_node.cfa_incoming_labels in
    let outgoings = move_node.cfa_outgoing_labels in
    let move_node_label = move_node.cfa_label in
    let fundef = (match current_cfa with
                 | CFA_FUNDEF f -> f
                 | _ -> assert false) in
    let new_cfa_nodes = Hashtbl.copy fundef.cfa_node_hashtbl in
    (* replace outgoings of incoming_nodes of move_node *)
    replace_nodes_outgoings new_cfa_nodes incomings move_node_label outgoings;
    (* replace incomings of outgoing_nodes of move_node *)
    replace_nodes_incomings new_cfa_nodes outgoings move_node_label incomings;
    (* remove move_node *)
    Hashtbl.remove new_cfa_nodes move_node_label;
    CFA_FUNDEF {fundef with cfa_node_hashtbl = new_cfa_nodes}
)

let rec find_null_addr_in_expr (expr : expression) : expression list = (
    match expr with
    | Constant _
    | Variable _
    | Sum _
    | Coeff _ ->
        []
    | Application ("null_addr%1", expr :: _) ->
        [expr]
    | Application (str, exprs) ->
        find_null_addr_in_exprs exprs
)

and find_null_addr_in_exprs (exprs : expression list) : expression list = (
    List.flatten (List.map find_null_addr_in_expr exprs)
)

let rec find_null_addr_in_pred (pred : predicate) : expression list = (
    match pred with
    | And preds
    | Or preds ->
        find_null_addr_in_preds preds
    | Not pred ->
        find_null_addr_in_pred pred
    | Eq (expr1, expr2)
    | Lt (expr1, expr2)
    | Leq (expr1, expr2) ->
        (find_null_addr_in_expr expr1) @ (find_null_addr_in_expr expr2)
    | True
    | False
    | Atom _ ->
        []
)

and find_null_addr_in_preds (preds : predicate list) : expression list = (
    List.flatten (List.map find_null_addr_in_pred preds)
)

let find_null_addr_in_cfa_stmt (stmt : cfa_stmt) : expression list = (
    match stmt with
    | CFA_FORM pred
    | CFA_COND (pred, _, _)
    | CFA_ASSUME pred
    | CFA_ENTRY pred ->
        find_null_addr_in_pred pred
    | CFA_CALL ((Variable ("null_addr%1")) :: expr1 :: _, _) ->
        [expr1]
    | CFA_CALL (exprs, _)
    | CFA_RESET (exprs) ->
        find_null_addr_in_exprs exprs
    | CFA_RETURN
    | CFA_ERROR
    | CFA_FAIL ->
        []
)

(** find null_addr in agent_expr *)
let find_null_addr_in_agent_expr (agent_expr : agent_expr_t)
: expression list = (
    let node_hash = agent_expr.agent_assign_nodes in
    Hashtbl.fold (fun _ node exprs ->
        exprs @ (find_null_addr_in_cfa_stmt node.cfa_node_stmt)
    ) node_hash [];
)

let rec replace_null_addr_in_expr (expr : expression)
                                  (parameters : expression list)
                                  (arguments : expression list)
: expression = (
    match expr with
    | Constant _
    | Variable _
    | Sum _
    | Coeff _ ->
        expr
    | Application ("null_addr%1", expr :: _) ->
        let replaced_expr =
            La_common.replace_exprs_in_expression parameters arguments expr in
        (Variable ("null_addr_" ^ (string_of_expression replaced_expr)))
    | Application (str, exprs) -> (
        Application (str,replace_null_addr_in_exprs exprs parameters arguments))
)

and replace_null_addr_in_exprs (exprs : expression list)
                               (parameters : expression list)
                               (arguments : expression list)
: expression list = (
    List.map (fun expr ->
        (replace_null_addr_in_expr expr parameters arguments)) exprs
)

let rec replace_null_addr_in_pred (pred : predicate)
                                  (parameters : expression list)
                                  (arguments : expression list)
: predicate = (
    match pred with
    | And preds ->
        And (replace_null_addr_in_preds preds parameters arguments)
    | Or preds ->
        Or (replace_null_addr_in_preds preds parameters arguments)
    | Not pred ->
        Not (replace_null_addr_in_pred pred parameters arguments)
    | Eq (expr1, expr2) ->
        Eq (replace_null_addr_in_expr expr1 parameters arguments,
            replace_null_addr_in_expr expr2 parameters arguments)
    | Lt (expr1, expr2) ->
        Lt (replace_null_addr_in_expr expr1 parameters arguments,
            replace_null_addr_in_expr expr2 parameters arguments)
    | Leq (expr1, expr2) ->
        Leq (replace_null_addr_in_expr expr1 parameters arguments,
             replace_null_addr_in_expr expr2 parameters arguments)
    | True
    | False
    | Atom _ ->
        pred
)

and replace_null_addr_in_preds (preds : predicate list)
                               (parameters : expression list)
                               (arguments : expression list)
: predicate list = (
    List.map (fun pred ->
        (replace_null_addr_in_pred pred parameters arguments)) preds
)

let replace_null_addr_in_cfa_stmt (stmt : cfa_stmt)
                                  (cfa_outputs : string list)
                                  (parameters : expression list)
                                  (arguments : expression list)
: cfa_stmt = (
    match stmt with
    | CFA_FORM pred ->
        CFA_FORM (replace_null_addr_in_pred pred parameters arguments)
    | CFA_COND (pred, label1, label2) ->
        CFA_COND (replace_null_addr_in_pred pred parameters arguments,
                  label1,
                  label2)
    | CFA_ASSUME pred ->
        CFA_ASSUME (replace_null_addr_in_pred pred parameters arguments)
    | CFA_ENTRY pred ->
        CFA_ENTRY (replace_null_addr_in_pred pred parameters arguments)
    | CFA_CALL (Variable "null_addr%1" :: expr1 :: _, spec) -> (
        let replaced_expr =
            La_common.replace_exprs_in_expression parameters arguments expr1 in
        CFA_FORM (Eq (Variable (List.hd cfa_outputs),
                      Variable ("null_addr_" ^ ((string_of_expression replaced_expr)))))
      )
    | CFA_CALL (exprs, spec) ->
        CFA_CALL (replace_null_addr_in_exprs exprs parameters arguments, spec)
    | CFA_RESET (exprs) ->
        CFA_RESET (replace_null_addr_in_exprs exprs parameters arguments)
    | CFA_RETURN
    | CFA_ERROR
    | CFA_FAIL ->
        stmt
)

(** replace null_addr in assumes *)
let replace_null_addr_in_agent_expr (agent_expr : agent_expr_t)
                                    (parameters : expression list)
                                    (arguments : expression list)
: agent_expr_t = (
    let node_hash = agent_expr.agent_assign_nodes in
    let new_node_hash = Hashtbl.create (Hashtbl.length node_hash) in
    Hashtbl.iter (fun node_name node ->
        let new_node_stmt =
            replace_null_addr_in_cfa_stmt node.cfa_node_stmt
                                          node.cfa_outputs
                                          parameters
                                          arguments
        in
        let new_node = {node with cfa_node_stmt = new_node_stmt} in
        Hashtbl.add new_node_hash node_name new_node
    ) node_hash;
    {
        agent_assign_nodes = new_node_hash;
        agent_temp_varid = agent_expr.agent_temp_varid;
        agent_expr_loc = agent_expr.agent_expr_loc
    }
)

(** Process null_addr(p) in assumes:
    First step, find all null_addr(p) in assumes and replace p that is parameter
        of called_function by argument in function_call
    Second step, null_addr(p) <==> reset(p) and (p = 0)
        ==> With every null_addr(p)
            insert reset(p)
            insert (null_addr_p = (p == 0))
    third step, replace null_addr(p) by null_addr_p
    This function process first step and third step
*)
let find_and_replace_null_addr (assumes : agent_expr_t list)
                               (funcall_node : cfa_node)
                               (called_cfa : cfa)
: expression list * agent_expr_t list = (
    (* find all null_addr in assumes *)
    let null_addr_exprs =
        remove_duplicated_exprs(
            List.flatten (List.map find_null_addr_in_agent_expr assumes)) in
    (* Find arguments: *)
    let arguments : expression list =
        (match funcall_node.cfa_node_stmt with
        | CFA_CALL (fc, _) -> (match fc with
                         | _ :: args -> args
                         | [] -> assert false)
        | _ -> assert false)
    in
    (* Find parameters: *)
    let string_parameters : string list =
        (match called_cfa with
        | CFA_FUNDEF f -> f.cfa_f_meta_data.cfa_params
        (* 20090704 maybe CFA_VARDEF *)
        | CFA_VARDEF v -> v.cfa_v_meta_data.cfa_params
        | _ -> assert false)
    in
    let parameters =
        List.map (fun string_param ->
            Variable (string_param);
        ) string_parameters in
    let replaced_exprs =
        La_common.replace_exprs_in_expressions parameters
                                               arguments
                                               null_addr_exprs
    in
    (* replace null_addr in assumes *)
    let assumes =
        List.map (fun assume ->
            replace_null_addr_in_agent_expr assume parameters arguments
        ) assumes in
    (replaced_exprs, assumes)
)

(** Process null_addr(p) in assumes:
    First step, find all null_addr(p) in assumes and replace p that is parameter
        of called_function by argument in function_call
    Second step, null_addr(p) <==> reset(p) and (p = 0)
        ==> With every null_addr(p)
            insert reset(p)
            insert (null_addr_p = (p == 0))
    third step, replace null_addr(p) by null_addr_p
    This function process second step
*)
let inline_process_null_addr (pointers : expression list)
                             (current_cfa : cfa)
                             (behind_node_name : string)
: cfa = (
    match pointers with
    | [] -> current_cfa
    | _ -> (
        let behind_node = find_node_in_cfa current_cfa behind_node_name in
        let new_cfa =
            List.fold_left (fun new_cfa_ expr ->
                (* insert null_add_p = (p == 0) *)
                let cond_node_label = Cexpr.make_new_label () in
                let if_node_label = Cexpr.make_new_label () in
                let else_node_label = Cexpr.make_new_label () in
                let return_node_label = Cexpr.make_new_label () in
                let return_stmt = CFA_FORM True in
                let new_cfa_ =
                    insert_stmt_below_to_cfa return_stmt
                                             return_node_label
                                             new_cfa_
                                             behind_node_name in
                let new_var = Variable ("null_addr_"
                                        ^ (string_of_expression expr)) in
                let cond_node_stmt =
                    CFA_COND (Eq (expr, Constant 0.0), if_node_label, else_node_label)
                in
                let new_cfa_ =
                    insert_stmt_below_to_cfa cond_node_stmt
                                             cond_node_label
                                             new_cfa_
                                             behind_node_name in
                let if_node_stmt = CFA_FORM (Eq (new_var, Constant 1.0)) in
                let new_cfa_ =
                    insert_edge_into_cfa_fundef new_cfa_ 
                                                cond_node_label
                                                return_node_label
                                                if_node_stmt
                                                if_node_label
                                                behind_node.cfa_memo
                in
                let else_node_stmt = CFA_FORM (Eq (new_var, Constant 0.0)) in
                let new_cfa_ =
                    insert_edge_into_cfa_fundef1 new_cfa_ 
                                                 cond_node_label
                                                 return_node_label
                                                 else_node_stmt
                                                 else_node_label
                                                 behind_node.cfa_memo
                in
                new_cfa_
            ) current_cfa pointers in
        (* insert reset(p) *)
        let new_reset : cfa_node =
            {behind_node with cfa_node_stmt =
                                   CFA_RESET (pointers);
                               cfa_label = Cexpr.make_new_label ()} in
        insert_cfa_node_below_to_cfa new_reset
                                     new_cfa
                                     behind_node_name)
)

let inter_cfas = ref []
(** Get info_cutlist of function has name is func_name.
    Insert this info into current_cfa at node has name is funcall_node_name *)
let inline_func_contract (current_cfa : cfa)
                         (afile : afile_t)
                         (funcall_node_name : string)
                         (called_cfa : cfa)
                         (func_name : string)
                         (funcall_number : int ref)
: cfa * cfa_node = (
    (* get info_cutlist *)
    let funcall_node = find_node_in_cfa current_cfa funcall_node_name in
    let agent_fspec_op =
        match funcall_node.cfa_node_stmt with
        | CFA_CALL (_, Some agent_fspec) ->
            Some agent_fspec
        | _ ->
            get_info_cutlist_of_cfa called_cfa
    in
    (* concat tail in agent_fspec_op *)
    let func_level =
        match funcall_node.cfa_node_stmt with
        | CFA_CALL (expr :: _, _) ->
            snd (find_name_and_level_of_expr_func expr)
        | _ -> assert false
    in
    let new_tail =
        "*" ^ (string_of_int (get_funcall_number funcall_node.cfa_label))
    in
    let agent_fspec_op_with_tail =
        concat_tail_in_agent_fspec_op agent_fspec_op new_tail func_level
    in
    (* divide agent_fspec into requires, assigns, assumes and old_terms
       and replace return_expr by cfa_output
    *)
    let (requires, assigns, assumes), old_terms =
        match agent_fspec_op_with_tail with
        | None -> (([], [], []), Hashtbl.create 0)
        | Some (agent_fspec, old_terms) ->
            (replace_return funcall_node
                            (divide_fspec agent_fspec)
                            new_tail
                            func_level,
             old_terms)
    in
    (* find and replace null_addr(p) in assumes *)
    let null_addr_pointers, assumes_replaced_null_addr_pointers =
        find_and_replace_null_addr assumes
                                   funcall_node
                                   called_cfa
    in
    (* insert info_cutlist to current cfa:
         insert assumes after funcall_node
         insert assigns after funcall_node but before assumes
         insert asserts after funcall_node but befor assigns *)
    let cfa_inserted_assumes = inline_ensures current_cfa funcall_node_name
                                              func_name !funcall_number
                                              assumes_replaced_null_addr_pointers
    in
    (* insert processing null_addr of assumes before assumes *)
    let cfa_inserted_processing_null_addr =
        inline_process_null_addr null_addr_pointers
                                 cfa_inserted_assumes
                                 funcall_node_name
    in
    let cfa_inserted_resets = inline_resets cfa_inserted_processing_null_addr
                                            funcall_node_name
                                            func_name
                                            !funcall_number assigns in
    let cfa_inserted_asserts = inline_asserts cfa_inserted_resets
                                              funcall_node_name func_name
                                              !funcall_number requires in
    funcall_number := !funcall_number + 1;

    let funcall_node = find_node_in_cfa cfa_inserted_asserts
                                        funcall_node_name in
    let assignments : predicate list = 
        assign_arguments_for_parameters called_cfa
                                        funcall_node.cfa_node_stmt
                                        new_tail
                                        func_level in
    let old_term_nodes = replace_node_label old_terms !funcall_number in
    let cfa_inserted_old_terms =
        if (Hashtbl.length old_term_nodes) = 0 then
            cfa_inserted_asserts
        else
            inline_insert_cfa old_term_nodes cfa_inserted_asserts
                                             funcall_node_name
    in
    let new_current_node : cfa_node =
        {funcall_node with cfa_node_stmt = CFA_FORM (And assignments)} in
    let cfa_replaced_funcall_node = (
         match cfa_inserted_old_terms with
         | CFA_FUNDEF f -> (
             let new_node_hash = Hashtbl.copy f.cfa_node_hashtbl in
             Hashtbl.replace new_node_hash funcall_node_name new_current_node;
             CFA_FUNDEF {f with cfa_node_hashtbl = new_node_hash}
             )
         | _ -> assert false) in
    inter_cfas := !inter_cfas @ [cfa_replaced_funcall_node];
    (cfa_replaced_funcall_node, new_current_node)
)

let is_cutentry cfa = (
    match cfa with
    | CFA_VARDEF v ->
        v.cfa_v_meta_data.cfa_is_cutentry
    | CFA_FUNDEF f ->
        f.cfa_f_meta_data.cfa_is_cutentry
    | CFA_OTHERDEF ->
        false)

let create_new_cfa_for_func_in_cutlist (func_name : string)
                                       (current_cfa : cfa)
                                       (current_node : cfa_node)
                                       (afile : afile_t)
                                       (funcall_number : int)
: bool * cfa * cfa_node = (
    try
        let called_cfa = find_cfa_in_afile afile
                                           ~caller:(Some current_cfa)
                                           func_name in
        let node_name = current_node.cfa_label in
        if is_cutentry called_cfa then (
            let new_cfa, new_current_node =
                inline_func_contract current_cfa afile node_name
                                     called_cfa func_name
                                     (ref funcall_number) in
            (true, new_cfa, new_current_node)
        )
        else (false, current_cfa, current_node)
    with Not_found -> (false, current_cfa, current_node)
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* NOTE: must normalize current_rr before calling function                 *)
(* function_pointer_in_current_rr_points_to to  deal with function pointer *)
let find_function_in_cfa_call
      (afile : afile_t)
      (current_rr : predicate)
      (current_cfa : cfa)
      (funcall_number : int) 
      (current_cnode : cfa_node)
: (cfa option) (* cfa of calling function *) 
  * cfa (* current cfa or new cfa which contains new reset & assume of 
calling function *)
  * cfa_node (* current_cnode or new cfa_node which get new_reset_label 
is outgoing_label *)
 =
begin      
    (* debug *)
    (*debug "find_function_in_cfa_call";*)
    (*debug ("current_rr: " ^ (string_of_predicate current_rr));
    let _ = read_int() in*)
    (*debug ("current_label: " ^ current_label);
    debug ("next_label: " ^ next_label);*)

    (* Before deal with function pointer, SHOULD normalize current_rr *)
    let normalized_current_rr : predicate = normalize current_rr in
    (* Get function call *)
    let func_c : expression list = match current_cnode.cfa_node_stmt with
                                   | CFA_CALL (f, _) -> f
                                   | _ -> assert false
    in
    (* Find expression of called-function *)
    let expr0 = try List.nth func_c 0 with Not_found -> assert false in
    (* debug *)
    (*debug ("string_of_expr0: " ^ (string_of_expression expr0));*)

    (* Check if calling function is in cut-list *)
    let (func_name, func_level) = find_name_and_level_of_expr_func (expr0) in
    let (is_in_cutlist, new_cfa, new_current_cnode) = 
        create_new_cfa_for_func_in_cutlist (func_name)
            (current_cfa) (current_cnode) (afile) (funcall_number)
    in
    (* If calling function is in cut-list, then return new_cfa *)
    if is_in_cutlist = true then (None, new_cfa, new_current_cnode)
    else begin
        (* Try to find calling function *)
        let called_cfa : cfa option =
            (try
                let cfa = find_cfa_in_afile afile
                                            ~caller:(Some current_cfa)
                                            func_name
                in
                Some cfa
            with Not_found -> None)
        in
        (* Check if we found calling function or not *)
        (match called_cfa with
        | Some CFA_FUNDEF _ -> (called_cfa, current_cfa, current_cnode)
        | Some CFA_OTHERDEF -> assert false 
        | Some CFA_VARDEF _
        | None -> begin
            (* Debug *)
            (*debug "Find possible Function Pointer";*)
            (* Find possible Function Pointer *)
            (*let name_of_possible_fp : string = 
                func_name ^ "%" ^ (string_of_int func_level) in*)
            (* 20090627. Maybe sref *)
            (*let based_expr0 = create_based_vars_in_expression (expr0) in*)
            (*debug ("based_expr0: " ^ (string_of_expression based_expr0));*)
            (* Check if possible_fp points to somewhere or not *)
            let possible_fp_points_to : string option = 
                function_pointer_in_current_rr_points_to (normalized_current_rr) 
                  (*Variable name_of_possible_fp*) (*based_expr0*) (expr0) 
            in
            (match possible_fp_points_to with
            | None -> (
                (None, current_cfa, current_cnode))
            | Some pointed_var -> begin 
                (* debug *)
                (*debug ("pointed_var : " ^ pointed_var);*)
                let (pointed_name, pointed_level) = 
                    find_name_and_level_of_expr_func (Variable pointed_var) in
                let (fp_is_in_cutlist, fp_cfa, fp_current_cnode) = 
                    create_new_cfa_for_func_in_cutlist (pointed_name) 
                      (current_cfa) (current_cnode) 
                      (afile) (funcall_number) 
                in
                if fp_is_in_cutlist = true then (None, fp_cfa, fp_current_cnode)
                else begin
                    (* debug *)
                    (*debug ("pointed_fun: " ^ pointed_name);*)
                    (* Try to find calling pointed function *)
                    let calling_pointed_cfa : cfa option = 
                        (try
                            Some (find_cfa_in_afile afile
                                                    ~caller:(Some current_cfa)
                                                    pointed_name)
                        with Not_found -> None)
                    in
                    (match calling_pointed_cfa with
                    | Some CFA_FUNDEF _ -> (calling_pointed_cfa, current_cfa, current_cnode)
                    | Some CFA_OTHERDEF -> assert false 
                    | Some CFA_VARDEF _
                    | None -> (None, current_cfa, current_cnode))
                end
              end) 
          end)
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let get_next_nodes_and_edges_for_cfa_call_node 
      (afile : afile_t)
      (current_cnode : cfa_node) 
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
      (funcall_number : int) 
      (current_rr : predicate)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    let func_level =
        match current_cnode.cfa_node_stmt with
        | CFA_CALL (expr :: _, _) ->
            snd (find_name_and_level_of_expr_func expr)
        | _ -> assert false
    in
    (* debug *)
    (*debug "get_next_nodes_and_edges_for_cfa_call_node ";
    debug ("------>current_rr: " ^ (string_of_predicate current_rr));*)
    (*read_int();*)
       
    (* Step1: create some vars: *)
    let new_call_stack = ref call_stack in
    let new_current_cfa = ref current_cfa in
    let outgoing_tuples = ref [] in
    (* Step2: check if current_cfa_node call a function or not: *)
    let expr_list, agent_fspec_op =
        (match current_cnode.cfa_node_stmt with
        | CFA_CALL (expr_list, agent_fspec_op) -> expr_list, agent_fspec_op
        | _ -> assert false)
    in
    (* debug: *)
    (*debug ("get_next_nodes_and_edges_for_cfa_call_node " ^ 
        string_of_expression (Sum expr_list));*)
    (* Step4: find called_function: *)
    let (called_cfa_option, concated_current_cfa, concated_current_cnode) = 
        find_function_in_cfa_call (afile) (current_rr)(CFA_FUNDEF current_cfa) 
          (funcall_number) (current_cnode)
    in
    let current_cfa = match concated_current_cfa with
                      | CFA_FUNDEF fundef -> fundef
                      | _ -> assert false in
    (* Step5: check if we find define of the function call or not: *)
    (match called_cfa_option with
    | Some called_cfa -> begin
        (* Save cfa_node & cfa to stack *)
        new_call_stack := 
            (funcall_number, current_cnode, current_cfa) :: call_stack;
        (* Find level_of_fc to create new_tail for vars called_cfa: *)
        let (name_of_fc, level_of_fc) =
            (match called_cfa with
            | CFA_FUNDEF fundef -> (fundef.cfa_f_meta_data.cfa_id, 1)
            | CFA_VARDEF vardef -> (
                error ("vardef - not fundef: " ^ vardef.cfa_v_meta_data.cfa_id);
                assert false)
            | _ -> assert false)
        in
        (* Create new_tail for vars called_cfa: *)
        let new_tail = "*" ^ (string_of_int funcall_number) in
        let new_called_cfa =
            concat_tail_with_var_name_in_cfa_function (called_cfa) 
              (new_tail) (level_of_fc) in
        (* Update new_current_cfa: *)
        new_current_cfa := (match new_called_cfa with
	                         | CFA_FUNDEF fundef -> fundef
	                         | _ -> assert false);
        (* Find entry_node of called_cfa to return *)
        let entry_node = 
            try Hashtbl.find (!new_current_cfa.cfa_node_hashtbl) 
                             (!new_current_cfa.cfa_entry_label)
            with
            | Not_found -> assert false
        in
        (* Create assignment, which assigns arguments for parameters: *)
        let assignments =
            assign_arguments_for_parameters (new_called_cfa) 
                                            (CFA_CALL (expr_list, agent_fspec_op)) 
                                            new_tail
                                            func_level
        in
        (* Save to outgoing_tuples to return: *)
        outgoing_tuples := [entry_node, 
                           And ([Eq (Variable "temp_out", Variable name_of_fc)] @ assignments)];
      end
    | None -> begin 
        new_current_cfa := (match concated_current_cfa with
                           | CFA_FUNDEF f -> f
                           | anything_else -> assert false);
        (* If the called-function not found -> only create an assignment: *)
        outgoing_tuples :=
            get_next_node_and_edge_for_nodef_func_call (expr_list) 
              (concated_current_cnode) (!new_current_cfa) (funcall_number)
      end);
    
    (* Step6: return: *)
    (!outgoing_tuples, !new_call_stack, !new_current_cfa)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~ get all children of one cfa_node ~~~~~~~~~~~~~~ *)
let get_next_nodes_and_edges_of_cfa_node
      (afile : afile_t)
      (current_cfa_node : cfa_node)
      (current_cfa : cfa_fundef_t)
      (call_stack : (int * cfa_node * cfa_fundef_t) list)
      (number_of_function_call : int) 
      (current_rr : predicate)
: (cfa_node * LogicUtil.Logic.predicate) list (* outgoing tuples *) 
  * (int * cfa_node * cfa_fundef_t) list (* new cfa_call_stack *)
  * cfa_fundef_t (* new cfa fundef *) =
begin
    (* debug *)
    (*debug ("get_next_nodes_and_edges_of_cfa_node");
    debug (get_string_of_cfa_node (current_cfa_node));*)

    match current_cfa_node.cfa_node_stmt with
    | CFA_ENTRY _ | CFA_FORM _ ->
        get_next_nodes_and_edges_for_cfa_form_node (current_cfa_node) 
          (current_cfa) (call_stack)
    | CFA_ASSUME _ -> 
        get_next_nodes_and_edges_for_cfa_assume_node (current_cfa_node)
          (current_cfa) (call_stack)
    | CFA_RESET _ -> 
        get_next_nodes_and_edges_for_cfa_reset_node (current_cfa_node)
          (current_cfa) (call_stack)
    | CFA_COND _ ->
        let outgoing_tuples =
            get_next_nodes_and_edges_for_cfa_cond_node (current_cfa_node) 
              (current_cfa) 
        in
        (outgoing_tuples, call_stack, current_cfa)
    | CFA_CALL _ ->
        get_next_nodes_and_edges_for_cfa_call_node (afile) (current_cfa_node) 
          (current_cfa) (call_stack) (number_of_function_call) (current_rr)
    | CFA_RETURN ->
        get_next_nodes_and_edges_for_cfa_return_node (current_cfa_node) 
          (current_cfa) (call_stack) 
    | CFA_ERROR | CFA_FAIL -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Create Cfa ofile from Translate.ofile_t.
   Mostly an 1-1 mapping. *)

let create_cfa_ofile (ofile : Translate.ofile_t)
		     : ofile_t = (begin
    match ofile with
    | {
        Translate.ofile_hash_name = (hashname : string) ;
        Translate.ofile_file_name = (filename : string) ;
        Translate.ofile_defs = (definitions : (string, Translate.definition_t) Hashtbl.t) ;
        Translate.ofile_global_predicates = (global_preds : Form.form_t list)
      } -> 
        (begin
	  let cfas = Hashtbl.create (Hashtbl.length definitions) in
	  (Hashtbl.iter
	      (fun name a_def ->
	           let new_name, a_cfa =
                       (transform_definition_t_to_cfa a_def "__ERROR__")
		   in Hashtbl.add cfas new_name a_cfa)
              definitions) ;
          (* Create & return new ofile *)
          let new_ofile =
              {
                ofile_hash_name = hashname ;
                ofile_file_name = filename ;
                ofile_cfas = cfas ;
                ofile_global_predicates = global_preds ;
                ofile_contract_of_field = Hashtbl.create 0;
              }
          in
          (Hashtbl.iter
              (fun _ cfa ->
                   match cfa with
                   | CFA_VARDEF v -> (
                       assert (v.cfa_v_meta_data.cfa_container = None);
                       v.cfa_v_meta_data.cfa_container <- (Some new_ofile))
                   | CFA_FUNDEF f -> (
                       assert (f.cfa_f_meta_data.cfa_container = None);
                       f.cfa_f_meta_data.cfa_container <- (Some new_ofile))
                   | CFA_OTHERDEF -> ())
              new_ofile.ofile_cfas);
          new_ofile
        end)
end)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let xmls_of_cfa_node (cn : cfa_node) : string = begin
    let xmls = ref "" in
    List.iter
        (fun (filename, startline, startposition, endline, endposition) -> 
            let xmlfile = "<filename>" ^ filename ^ "</filename>" in
            let xmlstartline = "<startline>" ^ startline ^ "</startline>" in
            let xmlstartposition = 
                (if (String.compare startposition "") = 0 then "" 
                else "<startposition>" ^ startposition ^ "</startposition>") 
            in
            let xmlendline = 
                (if (String.compare endline "") = 0 then "" 
                else "<endline>" ^ endline ^ "</endline>") 
            in
            let xmlendposition = 
                (if (String.compare endposition "") = 0 then "" 
                else "<endposition>" ^ endposition ^ "</endposition>") 
            in  
            xmls := !xmls ^ "<statement>" ^ xmlfile ^ xmlstartline ^ xmlstartposition ^
                            xmlendline ^ xmlendposition ^ "</statement>\n";
        )
        cn.cfa_memo;
    !xmls
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let xmls_of_unchecked_nodes (uns : cfa_node list) : string = begin
    let xmls = ref "<unchecked-statements>\n" in
    List.iter 
        (fun cn -> 
            xmls := !xmls ^ (xmls_of_cfa_node cn);
        )
        uns;
    xmls := !xmls ^ "</unchecked-statements>\n";
    !xmls
end

