(* parser/cexpr.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Expression/Statement/Type Node *)

(* NOTE: (1) "restrict" and "inline" are ignored.  (2?) Dropping
   braces in initializers does not work.  (3) Alignment calculation
   on bit-fields is incomplete and does not match with GCC. *)

let cgTRACE = ref false
let cgTRACE_dump_cast = false

open Num

module StringKeyedMap = Map.Make (
  struct
  type t = string
  let compare x y = (compare x y)
  end)

type stringmap_t = string StringKeyedMap.t

let is_in_cparser = ref false

let string_equal = Misc.string_equal

(* Some auxiliary logging functions *)
let warning (s : string) : unit =
    LogUtil.log LogUtil.PARSER LogUtil.WARNING s
   
let debug (s : string) : unit =
    LogUtil.log LogUtil.PARSER LogUtil.DEBUG s

let error (s : string) : unit =
    LogUtil.log LogUtil.PARSER LogUtil.ERROR s
    
let info (s : string) : unit =
    LogUtil.log LogUtil.PARSER LogUtil.INFO s

let emerg (s : string) : unit =
    LogUtil.log LogUtil.PARSER LogUtil.EMERG s

(* Disallows equality. *)

let (=) = (Misc.(=))
let (<>) = (Misc.(<>))

let num_ZERO = (num_of_int 0)
let num_ONE = (num_of_int 1)
let num_MINUS_ONE = (num_of_int (-1))

(* Pre-standard C treat conditional expressions as lvalue. *)

let traditional_conditional_expression_lvalue_p = false

(* ================================================================ *)

(* NOTES *)
(* There are two types of Notes :
      + line note
      + range note *)
type note_t =
    | LINE of bool * string * int * int
    | RANGE of string * int * int * int * int

(* VARIABLES *)

type varid_t = (string * int)
type vartag_t = (string * int * int)

let varid_equal ((id0, level0) : varid_t) ((id1, level1) : varid_t) : bool =
  ((string_equal id0 id1) && (level0 == level1))

(* STORAGE CLASS *)

type storage_class_t = NILSCLS | AUTO | EXTERN | REGISTER | STATIC | TYPEDEF

(* TYPE QUALIFIERS *)

type type_qualifier_t = NILQUAL | INLINE | CONSTANT | VOLATILE | RESTRICT
			| CONSTANT_VOLATILE | CONSTANT_RESTRICT
			| VOLATILE_RESTRICT | CONSTANT_VOLATILE_RESTRICT

(* EXPRESSION NODE *)

(* {typenode_t} represents a C type.  (1) Typenodes may possibly be
   wrapped by pseudo types, to indicate a defining context,
   {TYPEDEF_TYPE} for a typedef name, {STRCDEF_TYPE} for a structure
   or a union, {ENUMDEF_TYPE} for an enumeration, or {ATTRIBUTED_TYPE}
   for GCC attributes.  {ATTRIBUTED_TYPE} only appears at the toplevel
   in declarations or in structure components.  (2) {SPECIFR_TYPE} is
   used to hold a list of basic type specifiers (such as
   ["long";"int"]) temporarily in making <declaration_specifiers>.
   (3) {ELLIPSIS_TYPE} only appears in a parameter list.  (4)
   {PSEUDO_VA_LIST_TYPE} may be replaced to a type defined in
   {type_of_va_list}.  (5) {ARRAY_OF} has an expression; {None} for
   incomplete array, ({Some} {NIL}) for a bracket-star [*] (a variable
   length array in parameters), or {Some} for a value (constant or
   non-constant).  (6) {FUNCTION_OF} takes {None} for traditional
   parameters.  (7) {BITFIELD_TYPE} has an expression for width.  (8)
   {ENUMERATION_TYPE} and {STRUCTURE_TYPE} only has tag information
   and body is stored in a hashtable to avoid cyclic objects.  The
   hashtable is per file, but stored in types in order not to pass it
   around. *)

type typenode_t =
    NOTYPE
    (* Pseudo Types *)
    | SPECIFR_TYPE of string list
    | QUALIFIED_TYPE of typenode_t * type_qualifier_t
    | ATTRIBUTED_TYPE of typenode_t * gcc_attribute_t list
    | TYPEDEF_TYPE of typenode_t * string
    | STRCDEF_TYPE of typenode_t * unit
    | ENUMDEF_TYPE of typenode_t * (string * expr_t) list
    | ELLIPSIS_TYPE
    | PSEUDO_VA_LIST_TYPE
    (* Basic Types *)
    | VOID_TYPE
    | BOOLEAN
    | CHARACTER
    | SIGNED_CHARACTER
    | UNSIGNED_CHARACTER
    | SHORT_INTEGER
    | UNSIGNED_SHORT_INTEGER
    | INTEGER
    | UNSIGNED_INTEGER
    | LONG_INTEGER
    | UNSIGNED_LONG_INTEGER
    | LONG_LONG_INTEGER
    | UNSIGNED_LONG_LONG_INTEGER
    | SINGLE_FLOAT
    | DOUBLE_FLOAT
    | LONG_DOUBLE_FLOAT
    | SINGLE_COMPLEX
    | DOUBLE_COMPLEX
    | LONG_DOUBLE_COMPLEX
    | SINGLE_IMAGINARY
    | DOUBLE_IMAGINARY
    | LONG_DOUBLE_IMAGINARY
    (* Composit Types *)
    | POINTER_TO of typenode_t
    | ARRAY_OF of typenode_t * expr_t option
    | FUNCTION_OF of typenode_t * typenode_t list option
    | BITFIELD_TYPE of typenode_t * expr_t
    | ENUMERATION_TYPE of enumeration_tag_t
    | STRUCTURE_TYPE of structure_tag_t
    (* GCC Extension *)
    | GCC_VA_ARG

and structure_tag_t = {
  strc_struct_or_union_p : bool ; (* true=struct, false=union *)
  strc_tag : string ;
  strc_unique_id : string ;
  strc_typebody_table : (string, typebody_t) Hashtbl.t
}

and enumeration_tag_t = {
  enum_tag : string ;
  enum_unique_id : string ;
  enum_typebody_table : (string, typebody_t) Hashtbl.t
}

(* NOTE: {strc_unnamed_slots} is non-empty when some components are
   unnamed.  It maps a subcomponent in an unnamed component to a
   generated name given to an unnamed one. *)

and structure_body_t = {
  strc_utag : structure_tag_t ;
  mutable strc_components : (string * typenode_t) list ;
  (*GCC*) mutable strc_unnamed_slots : (string * string) list
}

and enumeration_body_t = {
  enum_utag : enumeration_tag_t ;
  mutable enum_components : (string * expr_t) list
}

and typebody_t =
    STRC_BODY of structure_body_t
    | ENUM_BODY of enumeration_body_t

and gcc_attribute_t = ATTR of (string * expr_t)

(* Operators of an SETF/OPER expression node. *)

and op_t =
    (* Simple Assignment (SETF only) *)
    SET
    (* Incr/Decr (SETF only) *)
    | INCR_PRE | DECR_PRE | INCR_POST | DECR_POST
    (* Binary Arithmetic (both SETF/OPER) *)
    | ADD | SUB | MUL | DIV | MOD | LSH | RSH | AND | IOR | XOR
    (* Unary Cast (OPER only) *)
    | CAST of bool (* false when implicitly added *)
    (* Unary Arithmetic (OPER only) *)
    | NEG | BITNOT
    (* Unary Logical (OPER only) *)
    | LOGNOT
    (* Binary Logical (OPER only) *)
    | EQ | NE | GE | GT | LE | LT
    (* Conditionals (OPER only) *)
    | COND | OROR | ANDAND
    (* Comma (OPER only) *)
    | COMMA
    (* Conditionals (ORER only) *)
    (*NONSTANDARD*)
    | IMP | IFF
    (* Pseudo Nodes (used for backward reference in Cg/Clex module). *)
    | PSEUDO_AMP | PSEUDO_STR | PSEUDO_PLS
    | PSEUDO_MNS | PSEUDO_TLD | PSEUDO_BAN

(* {expr_t} is an expression node.  (1) {SETF} expressions are
   replaced with {ASGN} statements in flattening.  (1) {VREF} has a
   quadruple (tn, id, scope-level, version-number).  Scope levels are
   0 for temporaries, 1 for globals, 2 for parameters, and 3 and more
   for locals.  Version number is for SSA and it is not used in the
   parser (always zero).  (2) {STRING} has a boolean indicating
   wide-string, a second string for a token, and a third string for a
   content (ie, backslashes are handled).  (3) {CLITERAL} represents
   compound literals, and is expanded to an ordinary form in
   {flatten_syntax} in "csyntax.ml".  {ASMSEQUENCE} is an expression
   list used in ASM statements.  (4) {PREF} is a pointer reference, a
   boolean indicating actual dereference (dereference may not be
   generated in case of multiple dimensional arrays).  {SREF} is a
   structure/union reference, a boolean indicating pointer
   dereference.  (5) {DSGN_AREF} only temporarily appears in
   (designation) initializers.  It shows an offset expression.  It is
   a range initializer, when the 2nd expression is not NIL.  (6) The
   2nd expression in {COND} can be NIL for a GCC extension of "?:".
   (7) {PROGN} represents statement expressions, and is expanded to
   ordinary forms in {flatten_syntax} in "csyntax.ml".  (8)
   {INIT_FILL} is a marker appearing only in initializers indicating
   to fill structures or arrays by default values. *)

and expr_t =
    NIL
    | PSEUDO_NONDET_VALUE of typenode_t * string
    (* Arithmetic/Logical/Conditional *)
    | OPER of typenode_t * op_t * expr_t * expr_t * expr_t
    (* Assignment/Increment/Decrement *)
    | SETF of typenode_t * op_t * expr_t * expr_t
    (* Primaries *)
    | CALL of typenode_t * expr_t * expr_t list * (cspec_t option)
    | VREF of typenode_t * string * int * int
    | PREF of typenode_t * bool * expr_t * expr_t
    | SREF of typenode_t * bool * expr_t * string
    | DSGN_AREF of typenode_t * expr_t * expr_t
    | ADDROF of typenode_t * expr_t
    | SIZEOF of typenode_t * typenode_t
    | PROGN of typenode_t * stmt_t list
    (* Value Node *)
    | CONST_I of (typenode_t * num * string)
    | CONST_F of (typenode_t * float * string)
    | STRING of bool * string * (int array)
    | CLITERAL of typenode_t * initializers_t
    | ASMSEQUENCE of expr_t list
    (* Internal Use *)
    | PHI of typenode_t * (string * expr_t) list
    | PSEUDO_MEMORY of typenode_t * typenode_t
    | INIT_FILL of typenode_t * expr_t
    (* Pragma Extension *)
    | PRAGMA of string
    | QUANTIFIER of string * (expr_t list) * expr_t
    (* Pseudo Node (not appear in parser) *)
    | PSEUDO_PROLOGUE
    | PSEUDO_EPILOGUE
    | PSEUDO_ACSL_RESULT of (typenode_t * expr_t)  (* actual return expr *)
    | PSEUDO_ACSL_OLD of (typenode_t * expr_t * expr_t)
                                          (* expr inside (), actual expr *)
    | PSEUDO_ACSL_UNIQUE of (typenode_t * expr_t)
    | PSEUDO_ACSL_UNIQUEBLOCK of (typenode_t * expr_t * expr_t)
    | PSEUDO_ACSL_VALID of (typenode_t * expr_t)
    | ALL_FIELD of (typenode_t * expr_t)

and full_expr_t = {
    expr_node : expr_t;
    expr_loc : Lexing.position * Lexing.position
}

and cspec_t = (full_expr_t, full_expr_t, typenode_t) Sexpr.spec_t

and initializers_t =
    INIT_EXPR of expr_t
    | INIT_LIST of initializers_t list
    | INIT_DSGN of (expr_t list) * initializers_t

and lable_type_t =
    NAME_LABEL of string
    | CASE_LABEL of expr_t * expr_t
    | DFLT_LABEL

(* {DOWHILE} has a boolean flag which tells while-loop (true) or
   do-loop (false).  {BEGIN} is for compound statements.  {ASGN} and
   {CALLASGN} are not introduced in the parser but in the flattening.
   {CALLASGN} takes a pair of a call expression and list of
   assignments.  {NOTE} with true are line number information, and
   with false are of syntaxing. *)

and stmt_t =
    DEFUN of declaration_t * stmt_t
    | DEFVAR of declaration_t
    | STMT of expr_t
    | ASGN of expr_t * expr_t
    | CALLASGN of expr_t * stmt_t list * expr_t list (* list of variable will be reseted after the function call*)
    | BEGIN of stmt_t list
    | IF of expr_t * stmt_t list * stmt_t list
    | SWITCH of expr_t * stmt_t list
    | DOWHILE of bool * expr_t * stmt_t list
    | FOR of stmt_t list * expr_t * expr_t * stmt_t list
    | BREAK
    | CONTINUE
    | LABEL of lable_type_t
    | GOTO of string
    | RETURN of expr_t option
    | ANNOTATION of string * expr_t
    | ASM of expr_t
    | NOTE of note_t
    (* FIXME: DzungDK
           - stupid solution, will be replaced with a better one later
           - (raw_spec, fn, ln, cnum_from_bol) *)
    | SPEC of Sexpr.lspec_t
    | ASSUME of expr_t
    | RESET of expr_t list

and ofile_t = {
    ofile_hash_name : string ;
    ofile_file_name : string ;
    ofile_defs : (string * stmt_t) list ref;
    (* save all of name information of function pointer field which is
       in a structure *)
    f_pointer_field_hashtbl : (string * string, string) Hashtbl.t;
    (* save all of func_contract information of function pointer field which is
       in a structure *)
    contract_of_field_hashtbl :
        ((string * string), (((full_expr_t, full_expr_t) Sexpr.function_contract_t)
                             * expr_t list)) Hashtbl.t;
}

and afile_t = {
    afile_libs : (string, ofile_t) Hashtbl.t ;
    afile_directory : (string, string) Hashtbl.t ;
    afile_start : string ;
    (* afile_global_predicates : (string, form_t list) Hashtbl.t *)
}

and block_memo_t = {
    blk_memo : string
}

(* Declaration record.  {decl_inits} is a list of inits when it is a
   variable, {decl_params} is a list of identifiers for a function,
   with ellipsis omitted which is represented in parameter types.
   MEMO: A name in {decl_id} can be empty for structure/union
   definitions. *)

and declaration_t = {
  decl_id : varid_t ;
  decl_tn : typenode_t ;
  decl_scls : storage_class_t ;
  decl_inits : expr_t list ;
  decl_params : string list ;
  decl_struct_defined_p : bool ;
  decl_specs : cspec_t option;
  decl_location : Sexpr.global_pos_t option;
}

and attribute_handler_t = declaration_t -> gcc_attribute_t -> declaration_t

(* Types used in grammar rules. *)

and specifier_t = (typenode_t * storage_class_t * type_qualifier_t
		     * gcc_attribute_t list)
and declarator_t = (typenode_t * varid_t * string list * gcc_attribute_t list)

let location_of_cur_fun : Sexpr.global_pos_t option ref = ref None

let nil_declaration : declaration_t = {
  decl_id = ("", 0) ;
  decl_tn = NOTYPE ;
  decl_scls = NILSCLS ;
  decl_inits = [] ;
  decl_params = [] ;
  decl_struct_defined_p = false ;
  decl_specs = None;
  decl_location = None;
}

let void_pointer_type = (POINTER_TO VOID_TYPE)
let char_pointer_type = (POINTER_TO CHARACTER)

(* ================================================================ *)
(* BinhLT *)
let note_of_curr_spec = ref (NOTE(LINE(false, "", 0, 0)))
(* ================================================================ *)

(* TARGET PROPERTY: 32BIT LONG CURRENTLY *)

let most_positive_int = (num_of_string "2147483647")
let most_positive_unsigned_int = (num_of_string "4294967295")
let most_positive_long_long = (num_of_string "9223372036854775807")
let most_positive_unsigned_long_long = (num_of_string "18446744073709551615")
let most_positive_long = most_positive_long_long
let most_positive_unsigned_long = most_positive_unsigned_long_long
let most_positive_short_float = 3.4028234663852885981170e+38
let maximum_characters_in_character_constant = 2

let unsigned_int_fit_to_long_int_p = true
let unsigned_short_fit_to_int_p = true
let pointer_difference_integral_type = INTEGER
let result_of_sizeof_type = UNSIGNED_LONG_INTEGER
let wide_character_type = INTEGER
let wide_string_type = (POINTER_TO SHORT_INTEGER)

let enumeration_compatible_type (tn : typenode_t) = INTEGER

let size_of_character = (num_of_int 1)
let size_of_short_integer = (num_of_int 2)
let size_of_integer = (num_of_int 4)
let size_of_long_integer = (num_of_int 4)
let size_of_long_long_integer = (num_of_int 8)
let size_of_single_float = (num_of_int 4)
let size_of_double_float = (num_of_int 8)
let size_of_long_double_float = (num_of_int 16)
let size_of_pointer = (num_of_int 4)

let alignment_of_character = (num_of_int 1)
let alignment_of_short_integer = (num_of_int 2)
let alignment_of_integer = (num_of_int 4)
let alignment_of_long_integer = (num_of_int 4)
let alignment_of_long_long_integer = (num_of_int 8)
let alignment_of_single_float = (num_of_int 4)
let alignment_of_double_float = (num_of_int 8)
let alignment_of_long_double_float = (num_of_int 16)
let alignment_of_pointer = (num_of_int 4)

let type_of_va_list = void_pointer_type

let rec struct_get_field_in_expr (expr : expr_t)
: expr_t list = (
    match expr with
    | NIL
    | PSEUDO_NONDET_VALUE _
    | VREF _
    | SIZEOF _
    | PROGN _
    | CONST_I _
    | CONST_F _
    | STRING _
    | CLITERAL _
    | PSEUDO_MEMORY _
    | PRAGMA _
    | PSEUDO_PROLOGUE
    | PSEUDO_EPILOGUE -> []
    | PREF (_, _, expr1, _)
    | SREF (_, _, expr1, _)
    | ADDROF (_, expr1)
    | INIT_FILL (_, expr1)
    | PSEUDO_ACSL_RESULT (_, expr1)
    | PSEUDO_ACSL_UNIQUE (_, expr1)
    | PSEUDO_ACSL_UNIQUEBLOCK (_, expr1, _)
    | PSEUDO_ACSL_VALID (_, expr1) ->
        struct_get_field_in_expr expr1;
    | ALL_FIELD (_, expr1) ->
        [expr1]
    | DSGN_AREF (_, expr1, expr2)
    | PSEUDO_ACSL_OLD (_, expr1, expr2)
    | SETF (_, _, expr1, expr2) ->
        (struct_get_field_in_expr expr1) @
        (struct_get_field_in_expr expr2)
    | OPER (_, _, expr1, expr2, expr3) ->
        (struct_get_field_in_expr expr1) @
        (struct_get_field_in_expr expr2) @
        (struct_get_field_in_expr expr3)
    | ASMSEQUENCE exprs ->
        List.flatten (
            List.map (fun a_expr ->
                struct_get_field_in_expr a_expr
            ) exprs)
    | PHI (_, str_exprs) ->
        List.flatten (
            List.map (fun (_, a_expr) ->
                struct_get_field_in_expr a_expr
            ) str_exprs)
    | QUANTIFIER (_, exprs, expr1)
    | CALL (_, expr1, exprs, _) -> (
        (struct_get_field_in_expr expr1) @
        (List.flatten (
            List.map (fun a_expr ->
                struct_get_field_in_expr a_expr
            ) exprs))
      )
)

let struct_get_field_in_full_expr (full_expr : full_expr_t)
: expr_t list = (
    struct_get_field_in_expr full_expr.expr_node;
)
(* ================================================================ *)

(* Strips off pseudo nodes ({STRCDEF_TYPE} and {TYPEDEF_TYPE}) from a
   type node.  Type nodes may possibly be wrapped to indicate a
   defining context (either by {STRCDEF_TYPE}, {ENUMDEF_TYPE}, or
   {TYPEDEF_TYPE}).  It is called in the middle of declaration
   construction, thus, pseudo type nodes may appear.  It removes type
   qualifiers, too, if {noqualifiersp} is true. *)

let rec strip_off_pseudo_type (tn : typenode_t) (noqualifiersp : bool) : typenode_t = (
  match tn with
    (ATTRIBUTED_TYPE (rt, a)) ->
      (strip_off_pseudo_type rt noqualifiersp)
  | (QUALIFIED_TYPE (rt, quals)) when noqualifiersp ->
      if (noqualifiersp) then
	(strip_off_pseudo_type rt noqualifiersp)
      else
	tn
  | (SPECIFR_TYPE _) -> (assert false)
  | (TYPEDEF_TYPE (rt, name)) ->
      (* It may be a nested renaming typedef. *)
      (strip_off_pseudo_type rt noqualifiersp)
  | (STRCDEF_TYPE (rt, _)) ->
      (strip_off_pseudo_type rt noqualifiersp)
  | (ENUMDEF_TYPE (rt, memo)) ->
      (strip_off_pseudo_type rt noqualifiersp)
  | _ ->
      tn)

(* Strips off pseudo and qualifier nodes as {strip_off_pseudo_type},
   but returns a pair of a typenode and a qualifier. *)

let rec strip_off_pseudo_type_return_qualifiers (tn : typenode_t) : (typenode_t * type_qualifier_t) = (
  match tn with
    (ATTRIBUTED_TYPE (rt, a)) ->
      (strip_off_pseudo_type_return_qualifiers rt)
  | (QUALIFIED_TYPE (rt, quals)) ->
      ((strip_off_pseudo_type rt true), quals)
  | (SPECIFR_TYPE _) -> (assert false)
  | (TYPEDEF_TYPE (rt, name)) ->
      (* It may be a nested renaming typedef. *)
      (strip_off_pseudo_type_return_qualifiers rt)
  | (STRCDEF_TYPE (rt, _)) ->
      (strip_off_pseudo_type_return_qualifiers rt)
  | (ENUMDEF_TYPE (rt, memo)) ->
      (strip_off_pseudo_type_return_qualifiers rt)
  | _ ->
      (tn, NILQUAL))

let strip_off_attributes (tn : typenode_t) : (typenode_t * gcc_attribute_t list) = (
  match tn with
    (ATTRIBUTED_TYPE (rt, a)) ->
      (tn, a)
  | _ ->
      (tn, []))

(* C source file name and line number for error messages.
   {current_trueline} holds a line number ignoring the line
   directive. *)
(* {current_cnum_from_bol} holds the number of character
   from the beginning of line. *)

let current_file = ref "(noname)"
let current_line = ref 0
let current_cnum_from_bol = ref 0
let current_trueline = ref 0
let spec_startpos_ln = ref 0
let spec_startpos_cnum_from_bol = ref 0

let get_current_line_info_string () : string = (
  (!current_file ^ ":" ^ (string_of_int !current_line)))

exception Parser_error of string

let syntaxerror (msg : string) : 'a = (
  if !is_in_cparser then
    (raise (Parser_error ((get_current_line_info_string ()) ^ ": " ^ msg)))
  else
    (raise (Parser_error msg)))

let internalerror (msg : string) : 'a = (
  (raise (Parser_error ("Internal error: " ^ msg))))

let generalerror (msg : string) : 'a = (
  (raise (Parser_error msg)))

let file_typebody_table : (string, typebody_t) Hashtbl.t option ref = ref None

(* LABEL VALUES.  {file_label_table} holds information of labels whose
   value is taken.  It consists of pairs of variables and label
   positions.  A variable is created to represent a unique label
   value.  A label position is a triple of names of a file, a
   function, and a label.  Intended use of this is to assign specific
   values to the created variables at linking. *)

let file_label_table : (string * (string * string * string)) list ref = ref []

let get_structure_body (tn : typenode_t) : structure_body_t = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE {
      strc_unique_id = uid ;
      strc_typebody_table = table
    }) ->
      (begin
	if (not (Hashtbl.mem table uid)) then
	  (internalerror "No structure in table") ;
	let bodyentry = (Hashtbl.find table uid) in
	match bodyentry with
	  STRC_BODY body -> body
	| ENUM_BODY body -> (assert false)
      end)
  | _ -> (assert false))

let put_structure_body (tn : typenode_t) (body : structure_body_t) : typenode_t = (
  match tn with
    (STRUCTURE_TYPE {
      strc_unique_id = uid ;
      strc_typebody_table = table
    }) ->
      (begin
	assert (not (Hashtbl.mem table uid)) ;
	(Hashtbl.replace table uid (STRC_BODY body)) ;
	tn
      end)
  | _ -> (assert false))

let get_enumeration_body (tn : typenode_t) : enumeration_body_t = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ENUMERATION_TYPE {
      enum_unique_id = uid ;
      enum_typebody_table = table
    }) ->
      (begin
	if (not (Hashtbl.mem table uid)) then
	  (internalerror "No enumeration in table") ;
	let bodyentry = (Hashtbl.find table uid) in
	match bodyentry with
	  STRC_BODY body -> (assert false)
	| ENUM_BODY body -> body
      end)
  | _ -> (assert false))

let put_enumeration_body (tn : typenode_t) (body : enumeration_body_t) : typenode_t = (
  match tn with
    (ENUMERATION_TYPE {
      enum_unique_id = uid ;
      enum_typebody_table = table
    }) ->
      (begin
	assert (not (Hashtbl.mem table uid)) ;
	(Hashtbl.replace table uid (ENUM_BODY body)) ;
	tn
      end)
  | _ -> (assert false))

(* PRINTER FOR STATEMENT/EXPRESSION (FOR TRACING) *)

let rec stmt_to_string_body (level : int) (s : stmt_t) : string = (
  let decl_params_to_string p : string = (
    if (p != []) then
      ("/(" ^ (String.concat "," p) ^ ")")
    else
      "") in
  match s with
    (BEGIN l) ->
      (String.concat ";" (List.map (stmt_to_string_body level) l))
  | SPEC _ -> "" | ASSUME _ -> "assume:" | RESET _ -> "reset:"
  | (DEFUN (d, s)) ->
      assert (d.decl_inits == []) ;
      assert ((d.decl_struct_defined_p) == false) ;
      ("(defun " ^ (varid_to_string d.decl_id)
	^ " : " ^ (type_to_string_body level d.decl_tn)
	^ "/" ^ (scls_to_string d.decl_scls)
	^ (decl_params_to_string d.decl_params)
	^ " {"
	^ (stmt_to_string_body level s)
	^ "})")
  | (DEFVAR d) ->
      ("(defvar " ^ (varid_to_string d.decl_id)
	^ " : " ^ (type_to_string_body level d.decl_tn)
	^ "/" ^ (scls_to_string d.decl_scls)
	^ (decl_params_to_string d.decl_params)
	^ (if (d.decl_struct_defined_p) then "/structdefined" else "")
	^ ((if (d.decl_inits != []) then " init: " else "")
	    ^ (exprlist_to_string level d.decl_inits))
	^ ")")
  | (STMT e) ->
      (expr_to_string_body level [] e)
  | (ASGN (e0, e1)) ->
      ("(" ^ (expr_to_string_body level [] e0) ^ " := "
	^ (expr_to_string_body level [] e1) ^ ")")
  | (CALLASGN (e0, l, _)) ->
      ("CALL(" ^ (expr_to_string_body level ["hidecast"] e0) ^ ";"
	^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	^ ")")
  | (IF (e, l0, l1)) ->
      ("if (" ^ (expr_to_string_body level [] e) ^ ") {"
	^ (String.concat ";" (List.map (stmt_to_string_body level) l0))
	^ "} else {"
	^ (String.concat ";" (List.map (stmt_to_string_body level) l1))
	^ "}")
  | (SWITCH (e, l)) ->
      ("switch (" ^ (expr_to_string_body level [] e) ^ ") {"
	^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	^ "}")
  | (DOWHILE (b, e, l)) ->
      if (b) then
	("while (" ^ (expr_to_string_body level [] e) ^ ") {"
	  ^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	  ^ "}")
      else
	("do {"
	  ^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	  ^ "} while (" ^ (expr_to_string_body level [] e) ^ ")")
  | (FOR (i, e1, e2, l)) ->
      ("for ("
	^ (String.concat "," (List.map (stmt_to_string_body level) i))
	^ ";" ^ (expr_to_string_body level [] e1)
	^ ";" ^ (expr_to_string_body level [] e2) ^ ") {"
	^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	^ "}")
  | BREAK -> "break"
  | (RETURN None) -> "return"
  | (RETURN (Some e)) -> ("return (" ^ (expr_to_string_body level [] e) ^ ")")
  | (GOTO s) -> ("goto " ^ s)
  | CONTINUE -> "continue"
  | (LABEL l) -> (lable_to_string l)
  | (ASM e) -> ("asm(" ^ (expr_to_string_body level [] e) ^ ")")
  | (NOTE (LINE(b, s, l, k))) ->
      if (b) then ("# " ^ (string_of_int l) ^ " \"" ^ s ^ "\"") else s
  | (NOTE (RANGE (fn, s_ln, s_cn, e_ln, e_cn))) ->
      ("# RANGE("
      ^ (string_of_int s_ln) ^ ", " ^ (string_of_int s_cn)
      ^ ") | ("
      ^ (string_of_int e_ln) ^ ", " ^ (string_of_int e_cn)
      ^ ") \"" ^ fn ^ "\"")
  | (ANNOTATION (s, e0)) ->
      (s ^ "(" ^ (expr_to_string_body level [] e0) ^ ")"))

and lable_to_string (l : lable_type_t) : string = (
  match l with
    (NAME_LABEL s) -> (s ^ ":")
  | (CASE_LABEL (e0, e1)) ->
      if (e1 == NIL) then
	("case (" ^ (expr_to_string e0) ^ "):")
      else
	("case (" ^ (expr_to_string e0) ^ "..." ^ (expr_to_string e1) ^ "):")
  | DFLT_LABEL -> "default:")

and stmtlist_to_string (level : int) (ss : stmt_t list) : string = (
  (String.concat ";" (List.map (stmt_to_string_body level) ss)))

and expr_to_string_body (level : int) (flags : string list) (e :expr_t) : string = (
  match e with
    NIL -> "NIL"
  | (PSEUDO_NONDET_VALUE (tn, s)) ->
      ("$nondet_" ^ s)
  | (OPER (tn, op, e0, e1, e2)) ->
      (nonassignment_operation_to_string level flags tn op e0 e1 e2)
  | (SETF (tn, op, e0, e1)) ->
      (assignment_operation_to_string level flags tn op e0 e1)
  | (VREF (tn, id, level, sequence)) ->
      (vartag_to_string (id, level, sequence))
  | (PREF (tn, deref, e0, e1)) ->
      if (deref) then
	("*(" ^ (expr_to_string_body level flags e0)
              ^ ", " ^ (expr_to_string_body level flags e1) ^ ")")
      else
	("(" ^ (expr_to_string_body level flags e0)
             ^ ", " ^ (expr_to_string_body level flags e1) ^ ")")
  | (SREF (tn, deref, e0, s)) ->
      if (deref) then
	("(" ^ (expr_to_string_body level flags e0) ^ "->" ^ s ^ ")")
      else
	("(" ^ (expr_to_string_body level flags e0) ^ "." ^ s ^ ")")
  | (DSGN_AREF (tn, e0, e1)) ->
      if (e1 == NIL) then
	("([" ^ (expr_to_string_body level flags e0) ^ "])")
      else
	("([" ^ (expr_to_string_body level flags e0)
	  ^ "..." ^ (expr_to_string_body level flags e1) ^ "])")
  | (CALL (tn, e0, ee, _)) ->
      ((expr_to_string_body level flags e0) ^ "("
	^ (exprlist_to_string level ee) ^ ")")
  | (ADDROF (tn, e0)) ->
      ("&(" ^ (expr_to_string_body level flags e0) ^ ")")
  | (SIZEOF (tn, tt)) ->
      ("sizeof(" ^ (type_to_string_body level tt) ^ ")")
  | (PROGN (tn, l)) ->
      ("(progn "
 	^ (String.concat ";" (List.map (stmt_to_string_body level) l))
	^ ")")
  | (CONST_I (tn, v, s)) ->
      if (not (string_equal s "")) then
	s
      else
	(string_of_num v)
  | (CONST_F (tn, v, s)) ->
      ("CONST(" ^ s ^ ")")
  | (STRING (b, s, _)) ->
      ("STRING(\"" ^ s ^ "\")")
  | (CLITERAL (tn, l)) ->
      ("{" ^ (type_to_string_body (level + 1) tn) ^ ","
	^ (init_to_string level l) ^ "}")
  | (ASMSEQUENCE l) ->
      ("(" ^ (exprlist_to_string level l) ^ ")")
  | (PHI (tn, phi)) ->
      ("#phi("
	^ (String.concat ";"
	      (List.map
		  (fun (label, e) ->
		    (label ^ ":" ^ (expr_to_string_body level flags e)))
		  phi))
	^ ")")
  | (PSEUDO_MEMORY (tn, tt)) ->
      ("MEMORY(" ^ (type_to_string_body level tt) ^ ")")
  | (INIT_FILL (tt, e)) ->
      ("INIT_FILL(" ^ (type_to_string_body level tt)
	^ " " ^ (expr_to_string_body level flags e) ^ ")")
  | (PRAGMA s) ->
      ("PRAGMA(" ^ s ^ ")")
  | (QUANTIFIER (s, v, e0)) ->
      ("(" ^ s ^ " "
	^ (String.concat " " (List.map (expr_to_string_body level flags) v))
	^ "," ^ (expr_to_string_body level flags e0) ^ ")")
  | PSEUDO_PROLOGUE ->
      "PROLOGUE()"
  | PSEUDO_EPILOGUE ->
      "EPILOGUE()"
  | PSEUDO_ACSL_RESULT (_, e0) ->
      ("$result(" ^ (expr_to_string_body level flags e0) ^ ")")
  | PSEUDO_ACSL_OLD (_, e0, _) ->
      ("$old(" ^ (expr_to_string_body level flags e0) ^ ")")
  | PSEUDO_ACSL_UNIQUE (_, e0) ->
      ("$unique(" ^ (expr_to_string_body level flags e0) ^ ")")
  | PSEUDO_ACSL_UNIQUEBLOCK (_, e0, e1) ->
      ("$unique(" ^ (expr_to_string_body level flags e0) ^ ", "
                    ^ (expr_to_string_body level flags e1) ^ ")")
  | PSEUDO_ACSL_VALID (_, e0) ->
      ("$valid(" ^ (expr_to_string_body level flags e0) ^ ")")
  | ALL_FIELD (_, e0) ->
      ("ALL_FIELD(" ^ (expr_to_string_body level flags e0) ^ ")")
)

and init_to_string (level : int) (ie : initializers_t) : string = (
  match ie with
    (INIT_EXPR e) ->
      ("(" ^ (expr_to_string_body level [] e) ^ ")")
  | (INIT_LIST l) ->
      ("{" ^ (String.concat ";" (List.map (init_to_string level) l)) ^ "}")
  | (INIT_DSGN (d, ie2)) ->
      ("(" ^ (String.concat "," (List.map (expr_to_string_body level []) d))
	^ "=" ^ (init_to_string level ie2) ^ ")"))

and operator_to_string (op : op_t) : string = (
  match op with
    SET -> "SET"
  | INCR_PRE -> "INCR_PRE"
  | DECR_PRE -> "DECR_PRE"
  | INCR_POST -> "INCR_POST"
  | DECR_POST -> "DECR_POST"
  | CAST _ -> "CAST"
  | NEG -> "NEG"
  | BITNOT -> "BITNOT"
  | ADD -> "ADD"
  | SUB -> "SUB"
  | MUL -> "MUL"
  | DIV -> "DIV"
  | MOD -> "MOD"
  | LSH -> "LSH"
  | RSH -> "RSH"
  | AND -> "AND"
  | IOR -> "IOR"
  | XOR -> "XOR"
  | LOGNOT -> "LOGNOT"
  | EQ -> "EQ"
  | NE -> "NE"
  | GE -> "GE"
  | GT -> "GT"
  | LE -> "LE"
  | LT -> "LT"
  | COND -> "COND"
  | OROR -> "OROR"
  | ANDAND -> "ANDAND"
  | COMMA -> "COMMA"
  | IMP -> "IMP"
  | IFF -> "IFF"
  | PSEUDO_AMP -> "PSEUDO_AMP"
  | PSEUDO_STR -> "PSEUDO_STR"
  | PSEUDO_PLS -> "PSEUDO_PLS"
  | PSEUDO_MNS -> "PSEUDO_MNS"
  | PSEUDO_TLD -> "PSEUDO_TLD"
  | PSEUDO_BAN -> "PSEUDO_BAN")

and nonassignment_operation_to_string (level : int) (flags : string list) (tn : typenode_t) (op : op_t) (e0 : expr_t) (e1 : expr_t) (e2 : expr_t) : string = (
  let s0 = (expr_to_string_body level [] e0) in
  let s1 = if (e1 != NIL) then (expr_to_string_body level [] e1) else "" in
  let s2 = if (e2 != NIL) then (expr_to_string_body level [] e2) else "" in
  match op with
  | (CAST explicitp) ->
      let cast = (type_to_string_body level tn) in
      if (List.mem "showcast" flags) then
	("((" ^ cast ^ ")" ^ s0 ^ ")")
      else if (List.mem "hidecast" flags) then
	("(" ^ s0 ^ ")")
      else if (cgTRACE_dump_cast) then
	("CAST" ^ "((" ^ cast ^ ")" ^ s0 ^ ")")
      else if (explicitp) then
	("((" ^ cast ^ ")" ^ s0 ^ ")")
      else
	("(" ^ s0 ^ ")")
  | SET -> (assert false)
  | INCR_PRE -> (assert false)
  | DECR_PRE -> (assert false)
  | INCR_POST -> (assert false)
  | DECR_POST -> (assert false)
  | NEG ->
      ("(- " ^ s0 ^ ")")
  | BITNOT ->
      ("(~ " ^ s0 ^ ")")
  | ADD ->
      ("(" ^ s0 ^ " + " ^ s1 ^ ")")
  | SUB ->
      ("(" ^ s0 ^ " - " ^ s1 ^ ")")
  | MUL ->
      ("(" ^ s0 ^ " * " ^ s1 ^ ")")
  | DIV ->
      ("(" ^ s0 ^ " / " ^ s1 ^ ")")
  | MOD ->
      ("(" ^ s0 ^ " % " ^ s1 ^ ")")
  | LSH ->
      ("(" ^ s0 ^ " << " ^ s1 ^ ")")
  | RSH ->
      ("(" ^ s0 ^ " >> " ^ s1 ^ ")")
  | AND ->
      ("(" ^ s0 ^ " & " ^ s1 ^ ")")
  | IOR ->
      ("(" ^ s0 ^ " | " ^ s1 ^ ")")
  | XOR ->
      ("(" ^ s0 ^ " ^ " ^ s1 ^ ")")
  | LOGNOT ->
      ("(! " ^ s0 ^ ")")
  | EQ ->
      ("(" ^ s0 ^ " == " ^ s1 ^ ")")
  | NE ->
      ("(" ^ s0 ^ " != " ^ s1 ^ ")")
  | GE ->
      ("(" ^ s0 ^ " >= " ^ s1 ^ ")")
  | GT ->
      ("(" ^ s0 ^ " > " ^ s1 ^ ")")
  | LE ->
      ("(" ^ s0 ^ " <= " ^ s1 ^ ")")
  | LT ->
      ("(" ^ s0 ^ " < " ^ s1 ^ ")")
  | COND ->
      ("(" ^ s0 ^ " ? " ^ s1 ^ " : " ^ s2 ^ ")")
  | OROR ->
      ("(" ^ s0 ^ " || " ^ s1 ^ ")")
  | ANDAND ->
      ("(" ^ s0 ^ " && " ^ s1 ^ ")")
  | COMMA ->
      ("(" ^ s0 ^ " , " ^ s1 ^ ")")
  | IMP ->
      ("(" ^ s0 ^ " ==> " ^ s1 ^ ")")
  | IFF ->
      ("(" ^ s0 ^ " <=> " ^ s1 ^ ")")
  | PSEUDO_AMP -> (assert false)
  | PSEUDO_STR -> (assert false)
  | PSEUDO_PLS -> (assert false)
  | PSEUDO_MNS -> (assert false)
  | PSEUDO_TLD -> (assert false)
  | PSEUDO_BAN -> (assert false))

and assignment_operation_to_string (level : int) (flags : string list) (tn : typenode_t) (op : op_t) (e0 : expr_t) (e1 : expr_t) : string = (
  let s0 = (expr_to_string_body level [] e0) in
  let s1 = if (e1 != NIL) then (expr_to_string_body level [] e1) else "" in
  match op with
    SET ->
      ("(" ^ s0 ^ " := " ^ s1 ^ ")")
  | INCR_PRE ->
      ("(++" ^ s0 ^ ")")
  | DECR_PRE ->
      ("(--" ^ s0 ^ ")")
  | INCR_POST ->
      ("(" ^ s0 ^ "++)")
  | DECR_POST ->
      ("(" ^ s0 ^ "--)")
  | (CAST _) -> (assert false)
  | NEG -> (assert false)
  | BITNOT -> (assert false)
  | ADD ->
      ("(" ^ s0 ^ " += " ^ s1 ^ ")")
  | SUB ->
      ("(" ^ s0 ^ " -= " ^ s1 ^ ")")
  | MUL ->
      ("(" ^ s0 ^ " *= " ^ s1 ^ ")")
  | DIV ->
      ("(" ^ s0 ^ " /= " ^ s1 ^ ")")
  | MOD ->
      ("(" ^ s0 ^ " %= " ^ s1 ^ ")")
  | LSH ->
      ("(" ^ s0 ^ " <<= " ^ s1 ^ ")")
  | RSH ->
      ("(" ^ s0 ^ " >>= " ^ s1 ^ ")")
  | AND ->
      ("(" ^ s0 ^ " &= " ^ s1 ^ ")")
  | IOR ->
      ("(" ^ s0 ^ " |= " ^ s1 ^ ")")
  | XOR ->
      ("(" ^ s0 ^ " ^= " ^ s1 ^ ")")
  | LOGNOT -> (assert false)
  | EQ -> (assert false)
  | NE -> (assert false)
  | GE -> (assert false)
  | GT -> (assert false)
  | LE -> (assert false)
  | LT -> (assert false)
  | COND -> (assert false)
  | OROR -> (assert false)
  | ANDAND -> (assert false)
  | COMMA -> (assert false)
  | IMP -> (assert false)
  | IFF -> (assert false)
  | PSEUDO_AMP -> (assert false)
  | PSEUDO_STR -> (assert false)
  | PSEUDO_PLS -> (assert false)
  | PSEUDO_MNS -> (assert false)
  | PSEUDO_TLD -> (assert false)
  | PSEUDO_BAN -> (assert false))

and exprlist_to_string (level : int) (ee : expr_t list) : string = (
  (String.concat "," (List.map (expr_to_string_body level []) ee)))

and scls_to_string (s : storage_class_t) : string = (
  match s with
    NILSCLS -> "NILSCLS"
  | AUTO -> "AUTO"
  | EXTERN -> "EXTERN"
  | REGISTER -> "REGISTER"
  | STATIC -> "STATIC"
  | TYPEDEF -> "TYPEDEF")

and qual_to_string (q : type_qualifier_t) : string = (
  match q with
    NILQUAL -> "NILQUAL"
  | CONSTANT -> "CONSTANT"
  | VOLATILE -> "VOLATILE"
  | RESTRICT -> "RESTRICT"
  | INLINE -> "INLINE"
  | CONSTANT_VOLATILE -> "CONSTANT_VOLATILE"
  | CONSTANT_RESTRICT -> "CONSTANT_RESTRICT"
  | VOLATILE_RESTRICT -> "VOLATILE_RESTRICT"
  | CONSTANT_VOLATILE_RESTRICT -> "CONSTANT_VOLATILE_RESTRICT")

and type_to_string_body (level : int) (tn : typenode_t) : string = (
  match tn with
    NOTYPE -> "NOTYPE"
  | (ATTRIBUTED_TYPE (t, a)) ->
      ("(" ^ (type_to_string_body level t)
	^ (attributes_to_string a) ^ ")")
  | (QUALIFIED_TYPE (t, q)) ->
      ("(" ^ (qual_to_string q) ^ ":"
	^ (type_to_string_body level t) ^ ")")
  | ELLIPSIS_TYPE -> "..."
  | (STRCDEF_TYPE (t, _)) ->
      ("STRCDEF(" ^ (type_to_string_body 0 t) ^ ")")
  | (ENUMDEF_TYPE (t, l)) ->
      ("ENUMDEF(" ^ (type_to_string_body 0 t) ^ ")")
  | (TYPEDEF_TYPE (t, s)) ->
      ("TYPEDEF(" ^ s ^ ")")
  | (SPECIFR_TYPE l) -> "SPECIFR"
  | VOID_TYPE -> "VOID"
  | BOOLEAN -> "BOOL"
  | CHARACTER -> "CHAR"
  | SIGNED_CHARACTER -> "SIGNED_CHAR"
  | UNSIGNED_CHARACTER -> "UNSIGNED_CHAR"
  | SHORT_INTEGER -> "SHORT_INT"
  | UNSIGNED_SHORT_INTEGER -> "UNSIGNED_SHORT_INT"
  | INTEGER -> "INT"
  | UNSIGNED_INTEGER -> "UNSIGNED_INT"
  | LONG_INTEGER -> "LONG_INT"
  | UNSIGNED_LONG_INTEGER -> "UNSIGNED_LONG_INT"
  | LONG_LONG_INTEGER -> "LONG_LONG_INT"
  | UNSIGNED_LONG_LONG_INTEGER -> "UNSIGNED_LONG_LONG_INT"
  | SINGLE_FLOAT -> "SINGLE_FLOAT"
  | DOUBLE_FLOAT -> "DOUBLE_FLOAT"
  | LONG_DOUBLE_FLOAT -> "LONG_DOUBLE_FLOAT"
  | SINGLE_COMPLEX -> "SINGLE_COMPLEX"
  | DOUBLE_COMPLEX -> "DOUBLE_COMPLEX"
  | LONG_DOUBLE_COMPLEX -> "LONG_DOUBLE_COMPLEX"
  | SINGLE_IMAGINARY -> "SINGLE_IMAGINARY"
  | DOUBLE_IMAGINARY -> "DOUBLE_IMAGINARY"
  | LONG_DOUBLE_IMAGINARY -> "LONG_DOUBLE_IMAGINARY"
  | (POINTER_TO t) ->
      ("PTR_TO(" ^ (type_to_string_body level t) ^ ")")
  | (ARRAY_OF (tn1, None)) ->
      (type_to_string_array_dimensions level tn1 "[]")
  | (ARRAY_OF (tn1, Some e)) ->
      (type_to_string_array_dimensions level tn1
	  ("[" ^ (expr_to_string_body (level + 1) [] e) ^ "]"))
  | (FUNCTION_OF (t, None)) ->
      ((type_to_string_body level t) ^ "()")
  | (FUNCTION_OF (t, Some tt)) ->
      ((type_to_string_body level t)
	^ "(" ^ (typelist_to_string level tt) ^ ")")
  | (BITFIELD_TYPE (t, e)) ->
      ("BITFIELD(" ^ (expr_to_string_body level [] e) ^ ")")
  | (STRUCTURE_TYPE {
      strc_struct_or_union_p = structp ;
      strc_tag = tag ;
      strc_unique_id = uid
    }) ->
      let key = (if (structp) then "STRUCT" else "UNION") in
      if (level == 0) then
	(type_to_string_structure level tn)
      else
	(key ^ ":" ^ tag ^ "(" ^ uid ^ ")" ^ "{...}")
  | (ENUMERATION_TYPE {
      enum_tag = tag ;
      enum_unique_id = uid
    }) ->
      (type_to_string_enumeration level tn)
  | PSEUDO_VA_LIST_TYPE -> "VA_LIST_TYPE"
  | GCC_VA_ARG -> "GCC_VA_ARG")

and type_to_string_structure (level : int) (tn : typenode_t) : string = (
  let body : structure_body_t = (get_structure_body tn) in
  match body with
    {
      strc_utag = {
	strc_struct_or_union_p = structp ;
	strc_tag = tag ;
	strc_unique_id = uid
      } ;
      strc_components = components ;
      strc_unnamed_slots = unnamedslots
    } ->
      let key = (if (structp) then "struct" else "union") in
      (key ^ ":" ^ tag ^ "(" ^ uid ^ ")"
	^ ("{"
	    ^ (String.concat ";"
		  (List.map
		      (fun (n, t) -> n ^ ":"
			^ (type_to_string_body (level + 1) t))
		      components))
	    ^ "}")))

and type_to_string_enumeration (level : int) (tn : typenode_t) : string = (
  let body : enumeration_body_t = (get_enumeration_body tn) in
  match body with
    {
      enum_utag = {
	enum_tag = tag ;
	enum_unique_id = uid
      } ;
      enum_components = components
    } ->
      ("enum" ^ ":" ^ tag ^ "{"
	^ (String.concat ";"
	      (List.map
		  (fun (id, e) ->
		    (id ^ "=" ^ (expr_to_string_body (level + 1) [] e)
		      ^ ": int"))
		  components))
	^ "}"))

and type_to_string_array_dimensions (level : int) (tn : typenode_t) (dims : string) : string = (
  match tn with
    (ARRAY_OF (tn1, None)) ->
      (type_to_string_array_dimensions level tn1 (dims ^ "[]"))
  | (ARRAY_OF (tn1, Some e)) ->
      (type_to_string_array_dimensions level tn1
	  (dims ^ "[" ^ (expr_to_string_body (level + 1) [] e) ^ "]"))
  | _ ->
      ((type_to_string_body level tn) ^ dims))

and typelist_to_string (level : int) (tt : typenode_t list) : string = (
  (String.concat "," (List.map (type_to_string_body level) tt)))

and stmt_to_string (s : stmt_t) : string = (stmt_to_string_body 0 s)

and expr_to_string (e : expr_t) : string = (expr_to_string_body 0 [] e)

and type_to_string (tn : typenode_t) : string = (type_to_string_body 0 tn)

and decl_to_string (d : declaration_t) : string = (decl_to_string_body 1 d)

and decl_to_simple_string (d : declaration_t) : string = (
  let decl_params_to_string p = (
    if (p != []) then
      ("/(" ^ (String.concat "," p) ^ ")")
    else
      "") in
  ("(" ^ (varid_to_string d.decl_id)
    ^ " : " ^ (type_to_string_body 1 d.decl_tn)
    ^ "/" ^ (scls_to_string d.decl_scls)
    ^ (decl_params_to_string d.decl_params) ^ ")"))

and decl_to_string_body (level : int) (d : declaration_t) : string = (
  let decl_params_to_string p = (
    if (p != []) then
      ("/(" ^ (String.concat "," p) ^ ")")
    else
      "") in
  ("(" ^ (varid_to_string d.decl_id)
    ^ " : " ^ (type_to_string_body level d.decl_tn)
    ^ "/" ^ (scls_to_string d.decl_scls)
    ^ (decl_params_to_string d.decl_params)
    ^ " = " ^ (exprlist_to_string level d.decl_inits)
    ^ (if (d.decl_struct_defined_p) then "/structdefined" else "") ^ ")"))

and attributes_to_string (l : gcc_attribute_t list) : string = (
  if (l == []) then
    ""
  else
    ("["
      ^ (String.concat ";"
	    (List.map
		(fun (a : gcc_attribute_t) ->
		  match a with
		    ATTR (k, e) ->
		      ("ATTRIBUTE:" ^ k ^ "(" ^ (expr_to_string e) ^ ")"))
		l))
      ^ "]"))

and attribute_names (l : gcc_attribute_t list) (separator : string) : string = (
  (String.concat separator
      (List.map
	  (fun (a : gcc_attribute_t) -> match a with ATTR (k, e) -> k)
	  l)))

and varid_to_string (v : varid_t) : string = (
  let ((id : string), (level : int)) = v in
  if (string_equal id "") then
    ("<none>%" ^ (string_of_int level))
  else
    (id ^ "%" ^ (string_of_int level)))

and vartag_to_string (v : vartag_t) : string = (
  let ((id : string), (level : int), (sequence: int)) = v in
  (* 16.4.2009. Xuan temporarily remove *)
  (*if (sequence == 0) then
    (id ^ "%" ^ (string_of_int level))
  else*)
    (id ^ "%" ^ (string_of_int level) ^ "%" ^ (string_of_int sequence)))

let rec dump_stmts (l : stmt_t list) : unit = (
  (List.iter (dump_stmt 1) l) ;
  (*(List.iter
      (fun s2 ->
	(print_endline (stmt_to_string_body 1 s2)))
    l)*))

and dump_stmt (level : int) (s : stmt_t) : unit = (
  let decl_params_to_string p : string = (
    if (p != []) then
      ("/(" ^ (String.concat "," p) ^ ")")
    else
      "") in
  match s with
    (BEGIN l) ->
      (print_endline "{") ;
      (List.iter (dump_stmt level) l) ;
      (print_endline "}")
  | SPEC _ | ASSUME _ | RESET _ -> ()
  | (DEFUN (d, s)) ->
      (*assert (d.decl_inits == []) ;*)
      assert ((d.decl_struct_defined_p) == false) ;
      (print_endline
	  ("(defun " ^ (varid_to_string d.decl_id)
	    ^ " : " ^ (type_to_string_body level d.decl_tn)
	    ^ "/" ^ (scls_to_string d.decl_scls)
	    ^ (decl_params_to_string d.decl_params) ^ ")")) ;
      (dump_stmt level s)
  | (DEFVAR d) ->
      let fakelevel = (if (string_equal (fst d.decl_id) "") then 0 else level) in
      (print_endline
	  ("(defvar " ^ (varid_to_string d.decl_id)
	    ^ " : " ^ (type_to_string_body fakelevel d.decl_tn)
	    ^ "/" ^ (scls_to_string d.decl_scls)
	    ^ (decl_params_to_string d.decl_params)
	    ^ (if (d.decl_struct_defined_p) then "/structdefined" else "")
	    ^ ((if (d.decl_inits != []) then " init: " else "")
		^ (exprlist_to_string level d.decl_inits))
	    ^ ")"))
  | (STMT e) ->
      (print_endline
	  (expr_to_string_body level [] e))
  | (ASGN (e0, e1)) ->
      (print_endline (stmt_to_string_body level s))
  | (CALLASGN (e0, l, _)) ->
      (print_endline (stmt_to_string_body level s))
  | (IF (e, [(GOTO label0)], [(GOTO label1)])) ->
      (print_endline
	  ("if (" ^ (expr_to_string_body level [] e) ^ ") {"
	    ^ (stmt_to_string_body level (GOTO label0))
	    ^ "} else {"
	    ^ (stmt_to_string_body level (GOTO label1)) ^ "}"))
  | (IF (e, l0, l1)) ->
      (print_endline
	  ("if (" ^ (expr_to_string_body level [] e) ^ ") {")) ;
      (List.iter (dump_stmt level) l0) ;
      (print_endline
	  ("} else {")) ;
      (List.iter (dump_stmt level) l1) ;
      (print_endline
	  ("}"))
  | (SWITCH (e, l)) ->
      (print_endline
	  ("switch (" ^ (expr_to_string_body level [] e) ^ ") {")) ;
      (List.iter (dump_stmt level) l) ;
      (print_endline
	  ("}"))
  | (DOWHILE (true, e, l)) ->
      (print_endline
	  ("while (" ^ (expr_to_string_body level [] e) ^ ") {")) ;
      (List.iter (dump_stmt level) l) ;
      (print_endline
	  ("}"))
  | (DOWHILE (false, e, l)) ->
      (print_endline
	  ("do {")) ;
      (List.iter (dump_stmt level) l) ;
      (print_endline
	  ("} while (" ^ (expr_to_string_body level [] e) ^ ")"))
  | (FOR (i, e1, e2, l)) ->
      (print_endline
	  ("for ("
	    ^ (String.concat "," (List.map (stmt_to_string_body level) i))
	    ^ ";" ^ (expr_to_string_body level [] e1)
	    ^ ";" ^ (expr_to_string_body level [] e2) ^ ") {")) ;
      (List.iter (dump_stmt level) l) ;
      (print_endline
	  ("}"))
  | BREAK ->
      (print_endline "break")
  | (RETURN None) ->
      (print_endline "return")
  | (RETURN (Some e)) ->
      (print_endline ("return (" ^ (expr_to_string_body level [] e) ^ ")"))
  | (GOTO s) ->
      (print_endline("goto " ^ s))
  | CONTINUE ->
      (print_endline "continue")
  | (LABEL l) ->
      (print_endline (lable_to_string l))
  | (ASM e) ->
      (print_endline ("asm(" ^ (expr_to_string_body level [] e) ^ ")"))
  | (NOTE (LINE (true, s, l, k))) ->
      (print_endline ("# " ^ (string_of_int l) ^ " \"" ^ s ^ "\""))
  | (NOTE (LINE (false, s, l, k))) ->
      (print_endline s)
  | (NOTE (RANGE _)) -> ()		(*FIXME: does nothing for the moment *)
  | (ANNOTATION (s, e0)) ->
      (print_endline
	  (s ^ "(" ^ (expr_to_string_body level [] e0) ^ ")")))

let full_expr_to_string (fexpr : full_expr_t) : string =
    let {
          expr_node = expr;
          expr_loc = _
        } = fexpr
    in expr_to_string expr

let cspecs_to_string = Sexpr.string_of_specs full_expr_to_string
                                             full_expr_to_string
                                             type_to_string

(* ================================================================ *)

(* {process_declaration} holds a function (stmt_t -> unit) which is
   called for each translation unit.  This body function is dummy. *)

let process_declaration = ref
  (fun (decl : stmt_t) ->
    (Printf.fprintf stderr "toplevel.\n") ;
    (Printf.fprintf stderr "%s\n" (stmt_to_string decl)))

(* Values of scope level for specific meanings. *)

let temporary_scope_level : int = 0
let global_scope_level : int = 1
let parameter_scope_level : int = 2
let implicit_function_scope_level : int = -1

(* Warn on non-C89 features. *)

let pedantic_warn_p = ref true
let traditional_p = ref false

(* Current function under defining. *)

let current_function = ref (nil_declaration : declaration_t)

(* Flag to suppress checks in {check_undeclared_reference} while
   scanning GCC attributes. *)

let in_attribute_scanning_p = ref false

let warn_in_parse (message : string) : unit = (
  let function_name = "(noname)" in
  (warning
      ((get_current_line_info_string ())
	^ ": warning"
	^ " in function '" ^ function_name ^ "': "
	^ message ^ ".")))

let warn_pedantic_in_parse : string -> unit = (
  let once : (string, unit) Hashtbl.t = (Hashtbl.create 10) in
  (fun (message : string) ->
    if (!pedantic_warn_p) then
      (begin
	if (not (Hashtbl.mem once message)) then
	  (begin
	    (Hashtbl.add once message ()) ;
	    (warn_in_parse message)
	  end)
      end)))

let ellipsis_string : string = "..."

let ellipsis_parameter_p (s : string) : bool = (
  (string_equal s ellipsis_string))

let ellipsis_type_p (tn : typenode_t) : bool = (
  (tn == ELLIPSIS_TYPE))

let epilogue_statement_p (s : stmt_t list) : bool = (
  match s with
    [(STMT PSEUDO_EPILOGUE)] -> true
  | _ -> false)

(* Place holder of a parameter declaration for "...". *)

let pseudo_ellipsis : declaration_t = {
  decl_id = (ellipsis_string, 0) ;
  decl_tn = NOTYPE ;
  decl_scls = NILSCLS ;
  decl_inits = [] ;
  decl_params = [] ;
  decl_struct_defined_p = false ;
  decl_specs = None;
  decl_location = None;
}

let empty_declarator : declarator_t = (NOTYPE, ("", 0), [], [])

(* Constant nodes. *)

let constant_zero = (CONST_I (INTEGER, num_ZERO, "0"))
let constant_one = (CONST_I (INTEGER, num_ONE, "1"))

let constant_integer (n : num) : expr_t = (
  (CONST_I (INTEGER, n, (string_of_num n))))

(* Type of {size_t} typedefed. *)

let defined_size_t_type = ref None

(* Returns the type of {size_t} which is typedefed. *)

let size_t_type () : typenode_t = (
  match !defined_size_t_type with
    None ->
      (warn_in_parse "Undefined size_t yet (use unsigned long)") ;
      UNSIGNED_LONG_INTEGER
  | (Some tn) ->
      tn)

let gensym_prefix = "#"

(* Makes a new variable/label. *)

let gensym : (string -> string) = (
  let counter = ref 10000 in
  (fun (prefix : string) ->
    (incr counter) ;
    assert (not (string_equal prefix "")) ;
    (gensym_prefix ^ prefix ^ (string_of_int !counter))))

let check_generated_symbol (s : string) : bool = (
  (Misc.string_starts_with gensym_prefix s))

let make_new_label () : string = (
  (gensym "L"))

let temporary_label_p (label : string) : bool = (
  (check_generated_symbol label))

let make_new_temporary_variable_for (e :expr_t) : expr_t = (
  match e with
    (VREF (tn, oldid, level, _)) ->
      let newid = (gensym "V") in
      (VREF (tn, newid, temporary_scope_level, 0))
  | _ -> (assert false))

let make_new_temporary_variable (tn : typenode_t) : expr_t = (
  (VREF (tn, (gensym "V"), temporary_scope_level, 0)))

let make_new_variable (tn : typenode_t) : expr_t = (
  (VREF (tn, (gensym "V"), parameter_scope_level, 0)))

let temporary_variable_p (v : expr_t) : bool = (
  match v with
    (VREF (tn, id, level, _)) when (level == temporary_scope_level) ->
      assert (check_generated_symbol id) ;
      true
  | (VREF (tn, id, level, _)) ->
      false
  | _ -> (assert false))

(* Checks if a variable is global.  It checks the scope, since
   extern-declared globals/locals are given with the global scope. *)

let global_variable_p (v : expr_t) : bool = (
  match v with
    (VREF (tn, id, level, _)) when (level == global_scope_level) ->
      assert (not (check_generated_symbol id)) ;
      true
  | (VREF (tn, id, level, _)) ->
      false
  | _ -> (assert false))

let varid_temporary_p ((id, level, _) : vartag_t) : bool = (
  (level == temporary_scope_level))

let varid_global_p ((id, level, _) : vartag_t) : bool = (
  (level == global_scope_level))

(* Checks if two variables are identical. *)

let identical_variable_p (v0 : expr_t) (v1 : expr_t) : bool = (
  match (v0, v1) with
    ((VREF (_, id0, level0, sequence0)), (VREF (_, id1, level1, sequence1))) ->
      ((string_equal id0 id1) && level0 == level1 && sequence0 == sequence1)
  | _ -> (assert false))

(* Checks if a variable reference is a function (not including a
   pointer-to a function).  It only appears as a global reference. *)

let function_reference_p (v : expr_t) : bool = (
  match v with
    (VREF (tn, id, level, _)) ->
      (begin
	match tn with
	  (FUNCTION_OF _) ->
	    (*assert (level == 1) ;*)
	    true
	| _ ->
	    false
      end)
  | _ -> (assert false))

let epilogue_statement_p (s : stmt_t list) : bool = (
  match s with
    [(STMT PSEUDO_EPILOGUE)] ->
      true
  | _ ->
      false)

(* GCC attribute handling functions. *)

let attribute_handlers : (string, attribute_handler_t) Hashtbl.t =
  (Hashtbl.create 40)

let attribute_null_handler (decl : declaration_t) (_ : gcc_attribute_t) : declaration_t = (
  decl)

(* ================================================================ *)

let make_attributed_type (tn : typenode_t) (attrs : gcc_attribute_t list) : typenode_t = (
  if (attrs == []) then
    tn
  else
    (ATTRIBUTED_TYPE (tn, attrs)))

let make_pointer_type (tn : typenode_t) : typenode_t = (
  (POINTER_TO tn))

let make_function_type (tn : typenode_t) (arguments : typenode_t list) : typenode_t = (
  (FUNCTION_OF (tn, Some arguments)))

let make_function_type_traditional (tn : typenode_t) : typenode_t = (
  (FUNCTION_OF (tn, None)))

(* Makes a bitfield.  Type {tn} is NOTYPE at call, since specifier is
   not resolved yet. *)

let make_bitfield_type (tn : typenode_t) (width : expr_t) : typenode_t = (
  (BITFIELD_TYPE (tn, width)))

let make_typedef_type name (tn : typenode_t) : typenode_t = (
  (TYPEDEF_TYPE (tn, name)))

let make_structuredef_specifier ((tn, scls, quals, attrs) : specifier_t) (body : (string * typenode_t) list) : specifier_t = (
  ((STRCDEF_TYPE (tn, ())), scls, quals, attrs))

let make_enumeratordef_specifier ((tn, scls, quals, attrs) : specifier_t) (body : (string * expr_t) list) : specifier_t = (
  ((ENUMDEF_TYPE (tn, body)), scls, quals, attrs))

(* Makes a structure with empty definition (that is incomplete). *)

let make_structure_type (structp : bool) (tag : string) : typenode_t = (
  let uid = (gensym "T") in
  let utag : structure_tag_t = {
    strc_struct_or_union_p = structp ;
    strc_tag = tag ;
    strc_unique_id = uid ;
    strc_typebody_table = Hashtbl.create 10
                          (*(Misc.some_of !file_typebody_table)*)
  } in

  let utag_none : structure_tag_t = {
    strc_struct_or_union_p = structp ;
    strc_tag = tag ;
    strc_unique_id = uid ;
    strc_typebody_table = Hashtbl.create 10
                          (*(Misc.some_of !file_typebody_table)*)
  } in
  let body : structure_body_t = {
    strc_utag = utag_none ;
    strc_components = [] ;
    strc_unnamed_slots = []
  } in
  (put_structure_body (STRUCTURE_TYPE utag) body))

let make_enumeration_type (tag : string) : typenode_t = (
  let uid = (gensym "T") in
  let utag : enumeration_tag_t = {
    enum_tag = tag ;
    enum_unique_id = uid ;
    enum_typebody_table = (Misc.some_of !file_typebody_table)
  } in
  let body : enumeration_body_t = {
    enum_utag = utag ;
    enum_components = []
  } in
  (put_enumeration_body (ENUMERATION_TYPE utag) body))

(* ================================================================ *)

(* QUALIFIER MANIPULATION *)

let qualifiers_as_bits (quals : type_qualifier_t) : int = (
  match quals with
    NILQUAL -> 0
  | INLINE -> 0 (*ignored*)
  | CONSTANT -> 4
  | VOLATILE -> 2
  | RESTRICT -> 1
  | CONSTANT_VOLATILE -> 6
  | CONSTANT_RESTRICT -> 5
  | VOLATILE_RESTRICT -> 3
  | CONSTANT_VOLATILE_RESTRICT -> 7)

let qualifiers_for_bits (bits : int) : type_qualifier_t = (
  match bits with
    0 -> NILQUAL
  | 4 -> CONSTANT
  | 2 -> VOLATILE
  | 1 -> RESTRICT
  | 6 -> CONSTANT_VOLATILE
  | 5 -> CONSTANT_RESTRICT
  | 3 -> VOLATILE_RESTRICT
  | 7 -> CONSTANT_VOLATILE_RESTRICT
  | _ -> (assert false))

let qualifiers_equal (quals0 : type_qualifier_t) (quals1 : type_qualifier_t) : bool = (
  (quals0 == quals1))

(* Returns a combination of the qualifiers.  IT IGNORES {INLINE}.
   Duplication signals an error when {combinep} is true. *)

let merge_qualifiers (combinep : bool) (q0 : type_qualifier_t) (q1 : type_qualifier_t) : type_qualifier_t = (
  let b0 = (qualifiers_as_bits q0) in
  let b1 = (qualifiers_as_bits q1) in
  if (combinep && ((b0 land b1) != 0)) then
    (syntaxerror "Bad type qualifier combination")
  else
    (qualifiers_for_bits (b0 lor b1)))

(* Merges given qualifiers to a given type.  If the given qualifiers
   are not empty, the returned type is of (QUALIFIED_TYPE
   <subordinate_type> <qualifiers>). *)

let rec add_type_qualifiers (tn : typenode_t) (qualifiers : type_qualifier_t) : typenode_t = (
  match tn with
    NOTYPE -> (assert false)
  | (QUALIFIED_TYPE (tx, quals)) ->
      (make_qualified_type tx (merge_qualifiers true quals qualifiers))
  | (ATTRIBUTED_TYPE (tn0, attrs)) ->
      let tx = (add_type_qualifiers tn0 qualifiers) in
      (ATTRIBUTED_TYPE (tx, attrs))
  | (TYPEDEF_TYPE (tn0, name)) ->
      (* Move up qualifiers from the inferior node. *)
      let tx = (add_type_qualifiers tn0 qualifiers) in
      (TYPEDEF_TYPE (tx, name))
  | (STRCDEF_TYPE (tn0, anything)) ->
      let tx = (add_type_qualifiers tn0 qualifiers) in
      (STRCDEF_TYPE (tx, anything))
  | (ENUMDEF_TYPE (tn0, anything)) ->
      let tx = (add_type_qualifiers tn0 qualifiers) in
      (ENUMDEF_TYPE (tx, anything))
  | _ ->
      (make_qualified_type tn qualifiers))

and make_qualified_type (tn : typenode_t) (quals : type_qualifier_t) : typenode_t = (
  (* Check if type is not qualified. *)
  match tn with
    (QUALIFIED_TYPE _) -> (assert false)
  | _ ->
      (begin
	if (qualifiers_equal quals NILQUAL) then
	  tn
	else
	  (QUALIFIED_TYPE (tn, quals))
      end))

let constant_qualifier_p (quals : type_qualifier_t) : bool = (
  match quals with
      NILQUAL -> false
    | INLINE -> false
    | CONSTANT -> true
    | VOLATILE -> false
    | RESTRICT -> false
    | CONSTANT_VOLATILE -> true
    | CONSTANT_RESTRICT -> true
    | VOLATILE_RESTRICT -> false
    | CONSTANT_VOLATILE_RESTRICT -> true)

let volatile_qualifier_p (quals : type_qualifier_t) : bool = (
  match quals with
      NILQUAL -> false
    | INLINE -> false
    | CONSTANT -> false
    | VOLATILE -> true
    | RESTRICT -> false
    | CONSTANT_VOLATILE -> true
    | CONSTANT_RESTRICT -> false
    | VOLATILE_RESTRICT -> true
    | CONSTANT_VOLATILE_RESTRICT -> true)

(* Checks if quals0 is a subset of quals1. *)

let qualifier_subset_p (quals0 : type_qualifier_t) (quals1 : type_qualifier_t) : bool = (
  let b0 = (qualifiers_as_bits quals0) in
  let b1 = (qualifiers_as_bits quals1) in
  ((b0 lor b1) == b1))

(* ================================================================ *)

(* Basic Types (C Object Types) *)

(* {basic_type_mapping} is a mapping from a list of types to an
   internal representations of basic type, like
   [SHORT_INTEGER;INTEGER] to a typenode SHORT_INTEGER. *)

let basic_type_mapping = (Hashtbl.create 40)

let init_basic_type_mapping () : unit = (
  (List.iter
      (fun (k, v) ->
	(Hashtbl.add basic_type_mapping (List.sort compare k) v))
      [
	(["void"], VOID_TYPE) ;
	(["_Bool"], BOOLEAN) ;
	(["short";"int"], SHORT_INTEGER) ;
	(["short"], SHORT_INTEGER) ;
	(["signed";"short"], SHORT_INTEGER) ;
	(["signed";"short";"int"], SHORT_INTEGER) ;
	(["int"], INTEGER) ;
	(["signed"], INTEGER) ;
	(["signed";"int"], INTEGER) ;
	(["long";"int"], LONG_INTEGER) ;
	(["long"], LONG_INTEGER) ;
	(["signed";"long"], LONG_INTEGER) ;
	(["signed";"long";"int"], LONG_INTEGER) ;
	(["long";"long";"int"], LONG_LONG_INTEGER) ;
	(["long";"long"], LONG_LONG_INTEGER) ;
	(["signed";"long";"long"], LONG_LONG_INTEGER) ;
	(["signed";"long";"long";"int"], LONG_LONG_INTEGER) ;
	(["unsigned";"short"], UNSIGNED_SHORT_INTEGER) ;
	(["unsigned";"short";"int"], UNSIGNED_SHORT_INTEGER) ;
	(["unsigned";"int"], UNSIGNED_INTEGER) ;
	(["unsigned"], UNSIGNED_INTEGER) ;
	(["unsigned";"long";"int"], UNSIGNED_LONG_INTEGER) ;
	(["unsigned";"long"], UNSIGNED_LONG_INTEGER) ;
	(["unsigned";"long";"long";"int"], UNSIGNED_LONG_LONG_INTEGER) ;
	(["unsigned";"long";"long"], UNSIGNED_LONG_LONG_INTEGER) ;
	(["char"], CHARACTER) ;
	(["signed";"char"], SIGNED_CHARACTER) ;
	(["unsigned";"char"], UNSIGNED_CHARACTER) ;
	(["float"], SINGLE_FLOAT) ;
	(["double"], DOUBLE_FLOAT) ;
	(["long";"double"], LONG_DOUBLE_FLOAT) ;
	(["float";"_Complex"], SINGLE_COMPLEX) ;
	(["double";"_Complex"], DOUBLE_COMPLEX) ;
	(["long";"double";"_Complex"], LONG_DOUBLE_COMPLEX) ;
	(["float";"_Imaginary"], SINGLE_IMAGINARY) ;
	(["double";"_Imaginary"], DOUBLE_IMAGINARY) ;
	(["long";"double";"_Imaginary"], LONG_DOUBLE_IMAGINARY) ;
	(["__builtin_va_list"], PSEUDO_VA_LIST_TYPE) ;
	(["__builtin_va_arg"], GCC_VA_ARG)
      ]))

(* Returns a typenode for a list of names of a basic type, like
   ["long";"int"].  It returns NOTYPE, when it is not a basic one. *)

let match_basic_type (typedesc : typenode_t) : typenode_t = (
  match typedesc with
    (SPECIFR_TYPE specifiers) ->
      (begin
	let spec = (List.sort compare specifiers) in
	if (Hashtbl.mem basic_type_mapping spec) then
	  (Hashtbl.find basic_type_mapping spec)
	else
	  NOTYPE
      end)
  | _ ->
      NOTYPE)

(*let expr_node expr = `(nth 0 ,expr)*)

let type_of_expr (e : expr_t) : typenode_t = (
  match e with
    NIL -> VOID_TYPE
  | (PSEUDO_NONDET_VALUE (tn, _)) -> tn
  | (OPER (tn, _, _, _, _)) -> tn
  | (SETF (tn, _, _, _)) -> tn
  | (VREF (tn, _, _, _)) -> tn
  | (PREF (tn, _, _, _)) -> tn
  | (SREF (tn, _, _, _)) -> tn
  | (DSGN_AREF _) -> (assert false)
  | (CALL (tn, _, _, _)) -> tn
  | (ADDROF (tn, _)) -> tn
  | (SIZEOF (tn, _)) -> tn
  | (PROGN (tn, _)) -> tn
  | (CONST_I (tn, _, _)) -> tn
  | (CONST_F (tn, _, _)) -> tn
  | (STRING (wide, _, _)) ->
      (if (not wide) then char_pointer_type else wide_string_type)
  | (CLITERAL (tn, _)) -> tn
  | (ASMSEQUENCE _) -> VOID_TYPE
  | (PHI (tn, _)) -> tn
  | (PSEUDO_MEMORY (tn, tt)) -> tn
  | (INIT_FILL _) -> VOID_TYPE
  | (PRAGMA _) -> VOID_TYPE
  | (QUANTIFIER _) -> VOID_TYPE
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
  | PSEUDO_ACSL_OLD (tn, _, _) -> tn
  | PSEUDO_ACSL_RESULT (tn, _) -> tn
  | ALL_FIELD (tn, _) -> tn
  | PSEUDO_ACSL_UNIQUE (tn, _) -> tn
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, _, _) -> tn
  | PSEUDO_ACSL_VALID (tn, _) -> tn)

(* ================================================================ *)

(* Mapping of builtin functions. *)

let gcc_builtin_identifiers : (string, typenode_t) Hashtbl.t = (Hashtbl.create 40)

let pseudo_builtin_identifiers : (string, typenode_t) Hashtbl.t = (Hashtbl.create 40)

let init_gcc_builtin_identifiers () : unit = (
  (List.iter
      (fun (s, (t0, t1)) ->
	let funtn =
	  if (t0 == NOTYPE) then
	    NOTYPE
	  else
	    (POINTER_TO (FUNCTION_OF (t0, t1))) in
	(Hashtbl.add gcc_builtin_identifiers s funtn))
      [("__builtin_abort", (NOTYPE, None)) ;
       ("__builtin_alloca", (NOTYPE, None)) ;
       ("__builtin_alpha_cmpbge", (NOTYPE, None)) ;
       ("__builtin_alpha_extbl", (NOTYPE, None)) ;
       ("__builtin_alpha_extwl", (NOTYPE, None)) ;
       ("__builtin_alpha_insbl", (NOTYPE, None)) ;
       ("__builtin_alpha_inslh", (NOTYPE, None)) ;
       ("__builtin_alpha_insql", (NOTYPE, None)) ;
       ("__builtin_alpha_inswl", (NOTYPE, None)) ;
       ("__builtin_bfin_norm_fr1x32", (NOTYPE, None)) ;
       ("__builtin_bfin_ssync", (NOTYPE, None)) ;
       ("__builtin_bswap_16", (NOTYPE, None)) ;
       ("__builtin_bswap_32", (NOTYPE, None)) ;
       ("__builtin_choose_expr", (NOTYPE, None)) ;
       ("__builtin_clzl", (NOTYPE, None)) ;
       ("__builtin_cmdline", (NOTYPE, None)) ;
       ("__builtin_constant_p", (INTEGER, None)) ;
       ("__builtin_ctzl", (NOTYPE, None)) ;
       ("__builtin_expect", (INTEGER, (Some [INTEGER ; INTEGER]))) ;
       ("__builtin_extract_return_addr", (NOTYPE, None)) ;
       ("__builtin_ffs", (NOTYPE, None)) ;
       ("__builtin_memcpy", ((POINTER_TO VOID_TYPE),
			    (Some [(POINTER_TO VOID_TYPE) ;
				   (POINTER_TO VOID_TYPE) ;
				   result_of_sizeof_type]))) ;
       ("__builtin_memset", ((POINTER_TO VOID_TYPE),
			    (Some [(POINTER_TO VOID_TYPE) ;
				   INTEGER ;
				   result_of_sizeof_type]))) ;
       ("__builtin_mfdr", (NOTYPE, None)) ;
       ("__builtin_mfsr", (NOTYPE, None)) ;
       ("__builtin_mtdr", (NOTYPE, None)) ;
       ("__builtin_popcountl", (NOTYPE, None)) ;
       ("__builtin_prefetch", (VOID_TYPE, (Some [(POINTER_TO VOID_TYPE)]))) ;
       ("__builtin_read16", (NOTYPE, None)) ;
       ("__builtin_read32", (NOTYPE, None)) ;
       ("__builtin_read8", (NOTYPE, None)) ;
       ("__builtin_return_address", ((POINTER_TO VOID_TYPE), Some [INTEGER])) ;
       ("__builtin_trap", (NOTYPE, None)) ;
       ("__builtin_types_compatible_p", (INTEGER, None)) ;
       ("__builtin_warning", (NOTYPE, None)) ;
       ("__builtin_write16", (NOTYPE, None)) ;
       ("__builtin_write32", (NOTYPE, None)) ;
       ("__builtin_write8", (NOTYPE, None))]))

let init_pseudo_builtin_identifiers () : unit = (
  let function_type (tn : typenode_t) (a : typenode_t list) : typenode_t = (
    (FUNCTION_OF (tn, (Some a)))) in
  (List.iter
      (fun (s, tn) ->
	(Hashtbl.add pseudo_builtin_identifiers s tn))
      [("$mem", (function_type INTEGER [INTEGER ; INTEGER]))]))

let init_builtin_identifiers () : unit = (
  init_gcc_builtin_identifiers() ;
  init_pseudo_builtin_identifiers())

(* ================================================================ *)

(* TYPE DISCRIMINATORS *)

let proper_type_p (tn : typenode_t) : bool = (
  (not (tn == NOTYPE)))

let basic_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    VOID_TYPE
  | BOOLEAN | CHARACTER | SIGNED_CHARACTER | UNSIGNED_CHARACTER
  | SHORT_INTEGER | UNSIGNED_SHORT_INTEGER
  | INTEGER | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      true
  | _ ->
      false)

let void_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  (nqtn == VOID_TYPE))

let pointer_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (POINTER_TO _) ->
      true
  | _ ->
      false)

(* Checks a type.  Integral is either integer (including character) or
   enumeration.  See [H&S5] 5.1 (p.124). *)

let integral_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    BOOLEAN | CHARACTER | SIGNED_CHARACTER | UNSIGNED_CHARACTER
  | SHORT_INTEGER | UNSIGNED_SHORT_INTEGER | INTEGER
  | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER ->
      true
  | (BITFIELD_TYPE _) ->
      true
  | (ENUMERATION_TYPE _) ->
      true
  | _ ->
      false)

let floating_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      true
  | _ ->
      false)

let complex_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      true
  | _ ->
      false)

(* Checks a type.  Arithmetic is a union of integral and
   floating-point. *)

let arithmetic_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  ((integral_type_p nqtn) || (floating_type_p nqtn)))

(* Checks a type.  Scaler is a union of pointer type and arithmetic
   type. *)

let scalar_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  ((pointer_type_p nqtn) || (arithmetic_type_p nqtn)))

let array_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ARRAY_OF _) ->
      true
  | _ ->
      false)

let incomplete_array_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ARRAY_OF (_, None)) ->
      true
  | _ ->
      false)

let function_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (FUNCTION_OF _) ->
      true
  | _ ->
      false)

let void_pointer_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (POINTER_TO ref) ->
      (void_type_p ref)
  | _ ->
      false)

let function_pointer_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (POINTER_TO ptr) ->
      (begin
	match (strip_off_pseudo_type ptr true) with
	  (FUNCTION_OF _) ->
	    true
	| _ ->
	    false
      end)
  | _ ->
      false)

(* Checks a type.  Object pointer is a non-void, non-function pointer.
   This allows incomplete types. *)

let object_pointer_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  ((pointer_type_p nqtn)
    && (not (function_pointer_type_p nqtn))
    && (not (void_pointer_type_p nqtn))))

let string_type_p (widep : bool) (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ARRAY_OF (reftn, dim)) ->
      (begin
	let nqreftn = (strip_off_pseudo_type reftn true) in
	let ctn = if (not widep) then CHARACTER else wide_character_type in
	match (nqreftn, ctn) with
	  ((CHARACTER|SIGNED_CHARACTER|UNSIGNED_CHARACTER),
	  (CHARACTER|SIGNED_CHARACTER|UNSIGNED_CHARACTER)) ->
	    true
	| ((SHORT_INTEGER|UNSIGNED_SHORT_INTEGER),
	  (SHORT_INTEGER|UNSIGNED_SHORT_INTEGER))->
	    true
	| ((INTEGER|UNSIGNED_INTEGER),
	  (INTEGER|UNSIGNED_INTEGER)) ->
	    true
	| ((LONG_INTEGER|UNSIGNED_LONG_INTEGER),
	  (LONG_INTEGER|UNSIGNED_LONG_INTEGER)) ->
	    true
	| ((LONG_LONG_INTEGER|UNSIGNED_LONG_LONG_INTEGER),
	  (LONG_LONG_INTEGER|UNSIGNED_LONG_LONG_INTEGER)) ->
	    true
	| _ ->
	    false
      end)
  | _ ->
      false)

let structure_or_union_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE _) ->
      true
  | _ ->
      false)

let structure_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE { strc_struct_or_union_p = true }) ->
      true
  | _ ->
      false)

let union_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE { strc_struct_or_union_p = false }) ->
      true
  | _ ->
      false)

let enumeration_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ENUMERATION_TYPE _) ->
      true
  | _ ->
      false)

let constant_qualified_type_p (tn : typenode_t) : bool = (
  let (_, quals) = (strip_off_pseudo_type_return_qualifiers tn) in
  (constant_qualifier_p quals))

let external_declaration_p (decl : declaration_t) : bool = (
  match decl with
    {
      decl_id = (id, level) ;
      decl_tn = tn ;
      decl_scls = scls ;
      decl_inits = _ ;
      decl_params = _ ;
      decl_struct_defined_p = _;
      decl_specs = _;
      decl_location = _;
    } ->
      (scls == EXTERN))

let storage_class_external_p (scls : storage_class_t) : bool = (
  (scls == EXTERN))

let float_type_of_complex (tn : typenode_t) : typenode_t = (
  match tn with
    LONG_DOUBLE_COMPLEX ->
      LONG_DOUBLE_FLOAT
  | DOUBLE_COMPLEX ->
      DOUBLE_FLOAT
  | SINGLE_COMPLEX ->
      SINGLE_FLOAT
  | LONG_DOUBLE_IMAGINARY ->
      LONG_DOUBLE_FLOAT
  | DOUBLE_IMAGINARY ->
      DOUBLE_FLOAT
  | SINGLE_IMAGINARY ->
      SINGLE_FLOAT
  | _ -> (assert false))

let complex_type_of_float (tn : typenode_t) : typenode_t = (
  match tn with
    LONG_DOUBLE_FLOAT ->
      LONG_DOUBLE_COMPLEX
  | DOUBLE_FLOAT ->
      DOUBLE_COMPLEX
  | SINGLE_FLOAT ->
      SINGLE_COMPLEX
  | _ -> (assert false))

(*GOMI*)(*
let rec variable_length_array_type_p (tn : typenode_t) = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ARRAY_OF (reftn, None)) ->
      if (not (array_type_p refty)) then
	false
      else
	((not (incomplete_array_type_p reftn))
	  && (variable_length_array_type_p refrn))
  | (ARRAY_OF (reftn, Some NIL)) ->
      if (not (array_type_p refty)) then
	true
      else
	(not (incomplete_array_type_p reftn))
  | (ARRAY_OF (reftn, Some e)) ->
      if (not (array_type_p refty)) then
	(not (constant_expr_p e))
      else
	(((not (constant_expr_p e))
	 && (not (incomplete_array_type_p reftn)))
	|| ((constant_expr_p e)
	     && (not (incomplete_array_type_p reftn))
	     && (variable_length_array_type_p refrn)))
  | _ ->
      false)
*)

let deref_pointer_type (tn : typenode_t) : typenode_t = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (POINTER_TO ref) ->
      ref
  | _ ->
      (internalerror "Dereferencing a non-pointer type"))

(* Finds the last value generating expression in a list (for statement
   expressions).  It returns (STMT NIL) when nothing.  *)

let rec last_effective_expression (l : stmt_t list) : stmt_t = (
  let nothing = (STMT NIL) in
  (last_effective_expression_loop l nothing))

and last_effective_expression_loop (l : stmt_t list) (es : stmt_t) : stmt_t = (
  match l with
    [] ->
      es
  | (s :: tl) ->
      (begin
	match s with
	  (DEFUN _) | (SPEC _) | (ASSUME _) | (RESET _)-> (assert false)
	| (DEFVAR _) ->
	    (last_effective_expression_loop tl es)
	| (STMT _) ->
	    (last_effective_expression_loop tl s)
	| (ASGN _) -> (assert false)
	| (CALLASGN _) -> (assert false)
	| (IF _) ->
	    (last_effective_expression_loop tl es)
	| (BEGIN l) ->
	    let x = (last_effective_expression_loop l es) in
	    (last_effective_expression_loop (x :: tl) x)
	| (SWITCH _) ->
	    (last_effective_expression_loop tl es)
	| (DOWHILE _) ->
	    (last_effective_expression_loop tl es)
	| (FOR _) ->
	    (last_effective_expression_loop tl es)
	| BREAK ->
	    (last_effective_expression_loop tl es)
	| (RETURN _) ->
	    (last_effective_expression_loop tl es)
	| (GOTO _) ->
	    (last_effective_expression_loop tl es)
	| CONTINUE ->
	    (last_effective_expression_loop tl es)
	| (LABEL _) ->
	    (last_effective_expression_loop tl es)
	| (ASM _) ->
	    (last_effective_expression_loop tl es)
	| (NOTE _) ->
	    (last_effective_expression_loop tl es)
	| (ANNOTATION _) ->
	    (last_effective_expression_loop tl es)
      end))

(* Checks if an expression is a modifiable lvalue.  It returns a pair
   of boolean and its type.  In the case of a direct component
   selection for structure or union, it checks if the structure is an
   lvalue and the selected component is not an array.  A structure is
   normally an lvalue except a return value from a function call.  See
   [H&S5] 7.1 (p.204). *)

let rec check_lvalue_p (e : expr_t) (modifiablep : bool) : (bool * typenode_t) = (
  match e with
    NIL ->
      (false, NOTYPE)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      (false, NOTYPE)
  | (OPER (tn, COND, e0, e1, e2)) ->
      if (not (traditional_conditional_expression_lvalue_p)) then
	(false, NOTYPE)
      else
	(begin
	  if (e1 == NIL) then
	    (*GCC*)
	    (false, NOTYPE)
	  else
	    (begin
	      let (b1, tn1) = (check_lvalue_p e1 false) in
	      let (b2, tn2) = (check_lvalue_p e2 false) in
	      if (not (b1 && b2)) then
		(false, NOTYPE)
	      else if (modifiablep && (constant_qualified_type_p tn)) then
		(false, NOTYPE)
	      else
		(true, tn1)
	    end)
	end)
  | (OPER (tn, COMMA, e0, e1, e2)) ->
      assert (e2 == NIL) ;
      (false, NOTYPE)
  | (OPER _) ->
      (false, NOTYPE)
  | (SETF _) ->
      (false, NOTYPE)
  | (VREF (tn, _, _, _))
  | (PREF (tn, _, _, _)) ->
      if (modifiablep && (constant_qualified_type_p tn)) then
	(false, NOTYPE)
      else
	(true, tn)
  | (SREF (tn, deref, e0, id)) ->
      if (modifiablep && (constant_qualified_type_p tn)) then
	(false, NOTYPE)
      else if (deref) then
	(true, tn)
      else
	(begin
	  let (lvalp, _) = (check_lvalue_p e0 modifiablep) in
	  if (not lvalp) then
	    (false, NOTYPE)
	  else
	    (true, tn)
	end)
  | (DSGN_AREF (tn, _, _)) ->
      (false, NOTYPE)
  | (CALL _) ->
      (false, NOTYPE)
  | (ADDROF _) | (SIZEOF _) ->
      (false, NOTYPE)
  | (PROGN (NOTYPE, l)) (*GCC*) ->
      (syntaxerror "No value in statement expression")
  | (PROGN (tn, l)) (*GCC*) ->
      (begin
	let x = (last_effective_expression l) in
	match x with
	  (STMT e) ->
	    assert (e != NIL) ;
	    (check_lvalue_p e modifiablep)
	| _ -> (assert false)
      end)
  | (CONST_I _) | (CONST_F _) ->
      (false, NOTYPE)
  | (STRING (widep, s, a)) ->
      if (not widep) then
	(true, char_pointer_type)
      else
	(true, wide_string_type)
  | (CLITERAL (tn, l)) ->
      (false, NOTYPE)
  | (ASMSEQUENCE _) ->
      (false, NOTYPE)
  | (PHI _) -> (assert false)
  | (PSEUDO_MEMORY _) -> (assert false)
  | (INIT_FILL _) -> (assert false)
  | (PRAGMA _) | (QUANTIFIER _) ->
      (false, NOTYPE)
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE | ALL_FIELD _
  | PSEUDO_ACSL_RESULT _ | PSEUDO_ACSL_OLD _ | PSEUDO_ACSL_UNIQUE _
  | PSEUDO_ACSL_VALID _ | PSEUDO_ACSL_UNIQUEBLOCK _ ->
      (assert false)
      (*| _ -> (internalerror "Strange expression node")*))

let enumerator_p (tn : typenode_t) (id : string) : bool = (
  if (enumeration_type_p tn) then
    let {
      enum_utag = _ ;
      enum_components = components
    } = (get_enumeration_body tn) in
    (List.mem_assoc id components)
  else
    false)

(* Checks constant-value-ness of an expression.  It checks operands of
   operators, because constant qualifiers are not properly propagated.
   NOTE: Constant qualified variables are not considered as constants.
   It considers only the RHS of assignments, and also considers only
   the last expression in a statement-expression. *)

let rec constant_expression_p (e : expr_t) : bool = (
  match e with
    NIL ->
      (assert false)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      false
  | (OPER (tn, (CAST _), e1, _, _)) ->
      (constant_expression_p e1)
  | (OPER (tn, (NEG|BITNOT), e0, _, _)) ->
      (constant_expression_p e0)
  | (OPER (tn, (LOGNOT), e0, _, _)) ->
      (constant_expression_p e0)
  | (OPER (tn, (ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR), e0, e1, _)) ->
      ((constant_expression_p e0) && (constant_expression_p e1))
  | (OPER (tn, (EQ|NE|GE|GT|LE|LT), e0, e1, _)) ->
      ((constant_expression_p e0) && (constant_expression_p e1))
  | (OPER (tn, COND, e0, e1, e2)) ->
      ((constant_expression_p e0)
	&& ((e1 == NIL) || (constant_expression_p e1))
	&& (constant_expression_p e2))
  | (OPER (tn, (OROR|ANDAND), e0, e1, _)) ->
      ((constant_expression_p e0) && (constant_expression_p e1))
  | (OPER (tn, COMMA, e0, e1, _)) ->
      (constant_expression_p e1)
  | (OPER _) -> (assert false)
  | (SETF (tn, SET, e0, e1)) ->
      (constant_expression_p e1)
  | (SETF (tn, (INCR_PRE|DECR_PRE|INCR_POST|DECR_POST), e0, e1)) ->
      false
  | (SETF (tn, (ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR), e0, e1)) ->
      false
  | (SETF _) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      (enumerator_p tn id)
  | (PREF (tn, _, e0, _)) ->
      false
  | (SREF (tn, _, e0, id)) ->
      (constant_expression_p e0)
  | (DSGN_AREF _) -> (assert false)
  | (CALL _) ->
      false
  | (ADDROF (tn, e0)) ->
      false
  | (SIZEOF (tn, tn1)) ->
      (constant_sized_type_p tn1)
  | (PROGN (NOTYPE, l)) ->
      (syntaxerror "No value in statement expression")
  | (PROGN (tn, l)) ->
      (begin
	let x = (last_effective_expression l) in
	match x with
	  (STMT e) ->
	    assert (e != NIL) ;
	    (constant_expression_p e)
	| _ -> (assert false)
      end)
  | (CONST_I _) | (CONST_F _) | (STRING _) ->
      true
  | (CLITERAL (tn, l)) ->
      false
  | (ASMSEQUENCE l) ->
      (List.for_all constant_expression_p l)
  | (PHI _) -> (assert false)
  | (PSEUDO_MEMORY _) -> (assert false)
  | (INIT_FILL _) -> (assert false)
  | (PRAGMA _) | (QUANTIFIER _) ->
      false
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE | ALL_FIELD _
  | PSEUDO_ACSL_RESULT _ | PSEUDO_ACSL_OLD _ | PSEUDO_ACSL_UNIQUE _
  | PSEUDO_ACSL_VALID _ | PSEUDO_ACSL_UNIQUEBLOCK _ ->
      (assert false))

and constant_sized_type_p (tn : typenode_t) : bool = (
  match tn with
    NOTYPE -> (assert false)
  | (ATTRIBUTED_TYPE (tn1, _)) ->
      (constant_sized_type_p tn1)
  | (QUALIFIED_TYPE (tn1, _)) ->
      (constant_sized_type_p tn1)
  | (SPECIFR_TYPE _) -> (assert false)
  | (STRCDEF_TYPE (tn1, _))
  | (ENUMDEF_TYPE (tn1, _))
  | (TYPEDEF_TYPE (tn1, _)) ->
      (constant_sized_type_p tn1)
  | ELLIPSIS_TYPE -> (assert false)
  | VOID_TYPE
  | BOOLEAN | CHARACTER | SIGNED_CHARACTER | UNSIGNED_CHARACTER
  | SHORT_INTEGER | UNSIGNED_SHORT_INTEGER
  | INTEGER | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      true
  | (POINTER_TO tn1) ->
      true
  | (ARRAY_OF (tn1, None)) ->
      (* Incomplete array size is one for a pointer. *)
      true
  | (ARRAY_OF (tn1, Some e1)) ->
      ((constant_sized_type_p tn1) && (constant_expression_p e1))
  | (FUNCTION_OF (tn1, _)) ->
      false
  | (BITFIELD_TYPE (tn1, e)) ->
      true
  | (STRUCTURE_TYPE _) ->
      (begin
	let body = (get_structure_body tn) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	(List.for_all (fun (s, tn1) -> (constant_sized_type_p tn1)) components)
      end)
  | (ENUMERATION_TYPE _) ->
      true
  | PSEUDO_VA_LIST_TYPE ->
      (constant_sized_type_p type_of_va_list)
  | GCC_VA_ARG -> (assert false))

(* ================================================================ *)

(* EXPRESSION NODE ACCESSORS *)

(* Strips off cast operations. *)

let rec ignore_cast (e : expr_t) : expr_t = (
  match e with
    (OPER (tn, (CAST _), e1, _, _)) ->
      (ignore_cast e1)
  | _ ->
      e)

(* Checks if an expression is a variable reference after stripping off
   the casts. *)

let variable_reference_p (e : expr_t) : bool = (
  let e1 = (ignore_cast e) in
  match e1 with
    (VREF (_, _, _, _)) ->
      true
  | _ ->
      false)

(* Checks if the OPER is constant null. *)

(* NOTE: NOT CERTAIN THIS IS ENOUGH. *)

let rec constant_zero_p (e : expr_t) : bool = (
  match e with
    (CONST_I (tn, v, s)) ->
      ((integral_type_p tn) && (eq_num num_ZERO v))
  | (CONST_F (tn, v, s)) ->
      false
  | (OPER (tn, (CAST _), e1, _, _)) ->
      (* Ignore castings. *)
      (constant_zero_p e1)
  | (PREF (tn, _, e1, _)) ->
      (* Special case of *&0.  Ignore castings between star and ampersand. *)
      (begin
	let e2 = (ignore_cast e) in
	match e2 with
	  (ADDROF (tn, e1)) ->
	    (constant_zero_p e1)
	| _ ->
	    false
      end)
  | _ ->
      false)

let constant_string_p (widep : bool) (e :expr_t) : bool = (
  match e with
    (STRING (b, s, _)) ->
      (widep == b)
  | _ ->
      false)

(* NOTUSED: Checks if the constant expr is null, ie, 0 or its cast to
   a pointer. *)

let null_pointer_p (e : expr_t) : bool = (
  match e with
    (CONST_I (tn, v, s)) when (integral_type_p tn) ->
      (eq_num num_ZERO v)
  | (CONST_I (tn, v, s)) ->
      false
  | (OPER (tn, (CAST _), e1, _, _)) when (pointer_type_p tn) ->
      (constant_zero_p e1)
  | _ ->
      false)

let expr_arg0 (expr : expr_t) : expr_t = (
  match expr with
    (OPER (tn, op, e0, e1, e2)) ->
      e0
  | (SETF (tn, op, e0, e1)) ->
      e0
  | _ ->
      (internalerror "Bad accessor to an expression"))

let expr_arg1 (expr : expr_t) : expr_t = (
  match expr with
    (OPER (tn, op, e0, e1, e2)) ->
      e1
  | (SETF (tn, op, e0, e1)) ->
      e1
  | _ ->
      (internalerror "Bad accessor to an expression"))

let expr_args (expr : expr_t) (expr : expr_t) : expr_t list = (
  match expr with
    (CALL (tn, f, al, _)) ->
      al
  | _ ->
      (internalerror "Bad accessor to an expression"))

(* Returns a return type of a function. *)

let rec type_of_return_value (tn : typenode_t) : typenode_t = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (POINTER_TO ptrtn) ->
      (type_of_return_value ptrtn)
  | (FUNCTION_OF (rettn, params)) ->
      rettn
  | _ -> (assert false))

(* ================================================================ *)

(* FOLDING CONSTANTS *)

let constant_integer_p (e : expr_t) : bool = (
  match e with
    (CONST_I _) ->
      true
  | _ ->
      false)

let constant_integer_value (noerr : bool) (e : expr_t) : num = (
  match e with
    (CONST_I (_, v, _)) ->
      v
  | _ ->
      if (noerr) then
	(num_of_int 0)
      else
	(assert false))

(* Folds a constant of an expression.  It returns a reduced
   expression, which is {CONST_I} when it is a constant value.  NOTE:
   IT NEVER RETURNS CONST_I IF AN EXPRESSION HAS EFFECT. *)

let rec fold_constant_integer (e : expr_t) : expr_t = (
  match e with
    NIL -> (e)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      e
  | (OPER (tn, (CAST b), e0, _, _)) ->
      let x0 = (fold_constant_integer e0) in
      if (constant_integer_p x0) then
	x0
      else
	(OPER (tn, (CAST b), x0, NIL, NIL))
  | (OPER (tn, ((NEG|BITNOT) as op), e0, _, _)) ->
      let x0 = (fold_constant_integer e0) in
      if (constant_integer_p x0) then
	(begin
	  let v0 = (constant_integer_value false x0) in
	  match op with
	    NEG ->
	      let vv = (minus_num v0) in
	      (CONST_I (tn, vv, ""))
	  | BITNOT ->
	      (warn_in_parse "Constant folding with bitnot-op") ;
	      let x = (Misc.pseudo_bitnot_num v0) in
	      (CONST_I (tn, x, ""))
	  | _ -> (assert false)
	end)
      else
	(OPER (tn, op, x0, NIL, NIL))
  | (OPER (tn, LOGNOT, e0, _, _)) ->
      let x0 = (fold_constant_integer e0) in
      if (constant_integer_p x0) then
	(begin
	  let v0 = (constant_integer_value false x0) in
	  let vv =
	    if (eq_num v0 num_ZERO) then
	      num_ONE
	    else
	      num_ZERO in
	  (CONST_I (tn, vv, ""))
	end)
      else
	(OPER (tn, LOGNOT, x0, NIL, NIL))
  | (OPER (tn, ((ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR) as op), e0, e1, _)) ->
      (begin
	let x0 = (fold_constant_integer e0) in
	let x1 = (fold_constant_integer e1) in
	if ((constant_integer_p x0) && (constant_integer_p x1)) then
	  (begin
	    let v0 = (constant_integer_value false x0) in
	    let v1 = (constant_integer_value false x1) in
	    match op with
	      ADD ->
		let x = (add_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | SUB ->
		let x = (sub_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | MUL ->
		let x = (mult_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | DIV ->
		let x = (quo_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | MOD ->
		let x = (mod_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | LSH ->
		let x = (mult_num v0 (power_num (num_of_int 2) v1)) in
		(CONST_I (tn, x, ""))
	    | RSH ->
		let x = (quo_num v0 (power_num (num_of_int 2) v1)) in
		(CONST_I (tn, x, ""))
	    | AND ->
		(warn_in_parse "Constant folding with and-op") ;
		let x = (Misc.pseudo_and_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | IOR ->
		(warn_in_parse "Constant folding with ior-op") ;
		let x = (Misc.pseudo_ior_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | XOR ->
		(warn_in_parse "Constant folding with xor-op") ;
		let x = (Misc.pseudo_xor_num v0 v1) in
		(CONST_I (tn, x, ""))
	    | _ -> (assert false)
	  end)
	else
	  (OPER (tn, op, x0, x1, NIL))
      end)
  | (OPER (tn, ((EQ|NE|GE|GT|LE|LT) as op), e0, e1, _)) ->
      (begin
	let x0 = (fold_constant_integer e0) in
	let x1 = (fold_constant_integer e1) in
	if ((constant_integer_p x0) && (constant_integer_p x1)) then
	  (begin
	    let v0 = (constant_integer_value false x0) in
	    let v1 = (constant_integer_value false x1) in
	    match op with
	      EQ ->
		let x = (num_of_int (if (eq_num v0 v1) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | NE ->
		let x = (num_of_int (if (not (eq_num v0 v1)) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | GE ->
		let x = (num_of_int (if (ge_num v0 v1) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | GT ->
		let x= (num_of_int (if (gt_num v0 v1) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | LE ->
		let x = (num_of_int (if (le_num v0 v1) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | LT ->
		let x = (num_of_int (if (lt_num v0 v1) then 1 else 0)) in
		(CONST_I (tn, x, ""))
	    | _ -> (assert false)
	  end)
	else
	  (OPER (tn, op, x0, x1, NIL))
      end)
  | (OPER (tn, COND, e0, NIL, e2)) ->
      let x0 = (fold_constant_integer e0) in
      let x2 = (fold_constant_integer e2) in
      if (constant_integer_p x0) then
	let v0 = (constant_integer_value false x0) in
	if (not (eq_num v0 num_ZERO)) then x0 else x2
      else
	(OPER (tn, COND, x0, NIL, x2))
  | (OPER (tn, COND, e0, e1, e2)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      let x2 = (fold_constant_integer e2) in
      if (constant_integer_p x0) then
	let v0 = (constant_integer_value false x0) in
	if (not (eq_num v0 num_ZERO)) then x1 else x2
      else
	(OPER (tn, COND, x0, x1, x2))
  | (OPER (tn, OROR, e0, e1, _)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      if (constant_integer_p x0) then
	(begin
	  let v0 = (constant_integer_value false x0) in
	  if (not (eq_num v0 num_ZERO)) then
	    let x = num_ONE in
	    (CONST_I (tn, x, ""))
	  else
	    x1
	end)
      else
	(OPER (tn, OROR, x0, x1, NIL))
  | (OPER (tn, ANDAND, e0, e1, _)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      if (constant_integer_p x0) then
	(begin
	  let v0 = (constant_integer_value false x0) in
	  if (eq_num v0 num_ZERO) then
	    let x = num_ZERO in
	    (CONST_I (tn, x, ""))
	  else
	    x1
	end)
      else
	(OPER (tn, ANDAND, x0, x1, NIL))
  | (OPER (tn, COMMA, e0, e1, _)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      if (constant_integer_p x0) then
	(fold_constant_integer e1)
      else
	(OPER (tn, COMMA, x0, x1, NIL))
  | (OPER (_, SET, _, _, _)) -> (assert false)
  | (OPER (_, (INCR_PRE|DECR_PRE), _, _, _)) -> (assert false)
  | (OPER (_, (INCR_POST|DECR_POST), _, _, _)) -> (assert false)
  | (OPER (_, (IFF|IMP), _, _, _)) -> (assert false)
  | (OPER (_, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), _, _, _)) -> (assert false)
  | (OPER (_, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), _, _, _)) -> (assert false)
  | (SETF (tn, SET, e0, e1)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      (SETF (tn, SET, x0, x1))
  | (SETF (tn, ((INCR_PRE|DECR_PRE|INCR_POST|DECR_POST) as op), e0, _)) ->
      let x0 = (fold_constant_integer e0) in
      (SETF (tn, op, x0, NIL))
  | (SETF (tn, ((ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR) as op), e0, e1)) ->
      let x0 = (fold_constant_integer e0) in
      let x1 = (fold_constant_integer e1) in
      (SETF (tn, op, x0, x1))
  | (SETF (_, (CAST b), _, _)) -> (assert false)
  | (SETF (_, (NEG|BITNOT|LOGNOT), _, _)) -> (assert false)
  | (SETF (_, (EQ|NE|GE|GT|LE|LT), _, _)) -> (assert false)
  | (SETF (_, (COND|OROR|ANDAND|COMMA|IMP|IFF), _, _)) -> (assert false)
  | (SETF (_, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), _, _)) -> (assert false)
  | (SETF (_, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), _, _)) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      (begin
	if (enumerator_p tn id) then
	  (begin
	    let v = (get_enumerator_value tn id) in
	    (CONST_I (tn, v, id))
	  end)
	else if (constant_qualified_type_p tn) then
	  (begin
	    (warn_in_parse "Constant variable not constant folded") ;
	    e
	  end)
	else
	  e
      end)
  | (PREF (tn, deref, e0, e1)) ->
      let x0 = (fold_constant_integer e0)
      and x1 = (fold_constant_integer e1) in
      (PREF (tn, deref, x0, x1))
  | (SREF (tn, deref, e0, s)) ->
      let x0 = (fold_constant_integer e0) in
      (SREF (tn, deref, x0, s))
  | (DSGN_AREF _) -> (assert false)
  | (CALL _) ->
      (*AHO*)
      e
  | (ADDROF _) ->
      (*AHO*)
      e
  | (SIZEOF (tn, tn1)) ->
      let (s, a) = (storage_layout true false tn1) in
      s
  | (PROGN (tn, l)) ->
      (*AHO*)
      e
  | (CONST_I (tn1, v, _)) ->
      if (not (integral_type_p tn1)) then
	(syntaxerror "Bad constant expression (non-integer)")
      else
	e
  | (CONST_F _) ->
      (syntaxerror "Bad constant expression (float)")
  | (STRING _) ->
      e
  | (CLITERAL (tn, l)) ->
      (*AHO*)
      e
  | ASMSEQUENCE _ -> (assert false)
  | PHI _ -> (assert false)
  | PSEUDO_MEMORY _ -> (assert false)
  | INIT_FILL _ -> (assert false)
  | PRAGMA _ -> (assert false)
  | QUANTIFIER _ -> (assert false)
  | PSEUDO_PROLOGUE -> (assert false)
  | PSEUDO_EPILOGUE -> (assert false)
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_OLD _ | PSEUDO_ACSL_RESULT _ | PSEUDO_ACSL_UNIQUE _
  | PSEUDO_ACSL_VALID _ | PSEUDO_ACSL_UNIQUEBLOCK _ ->
      e)

(* Returns size and alignment as a pair.  {slotp} is true when it is
   called for a structure member, where incomplete arrays are allowed
   as flexible arrays (size is zero). *)

and storage_layout (sizeofp : bool) (slotp : bool) (tn : typenode_t) : (expr_t * expr_t) = (
  match tn with
    NOTYPE -> (assert false)
  | (ATTRIBUTED_TYPE (tn1, _)) ->
      (storage_layout sizeofp slotp tn1)
  | (QUALIFIED_TYPE (tn1, _)) ->
      (storage_layout sizeofp slotp tn1)
  | (TYPEDEF_TYPE (tn1, _))
  | (STRCDEF_TYPE (tn1, _))
  | (ENUMDEF_TYPE (tn1, _)) ->
      (storage_layout sizeofp slotp tn1)
  | (SPECIFR_TYPE l) -> (assert false)
  | ELLIPSIS_TYPE -> (assert false)
  | VOID_TYPE ->
      (syntaxerror "Bad constant expression (void type)")
  | BOOLEAN ->
      ((constant_integer size_of_character),
      (constant_integer alignment_of_character))
  | CHARACTER | SIGNED_CHARACTER | UNSIGNED_CHARACTER ->
      ((constant_integer size_of_character),
      (constant_integer alignment_of_character))
  | SHORT_INTEGER | UNSIGNED_SHORT_INTEGER ->
      ((constant_integer size_of_short_integer),
      (constant_integer alignment_of_short_integer))
  | INTEGER | UNSIGNED_INTEGER ->
      ((constant_integer size_of_integer),
      (constant_integer alignment_of_integer))
  | LONG_INTEGER | UNSIGNED_LONG_INTEGER ->
      ((constant_integer size_of_long_integer),
      (constant_integer alignment_of_long_integer))
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER ->
      ((constant_integer size_of_long_long_integer),
      (constant_integer alignment_of_long_long_integer))
  | SINGLE_FLOAT ->
      ((constant_integer size_of_single_float),
      (constant_integer alignment_of_single_float))
  | DOUBLE_FLOAT ->
      ((constant_integer size_of_double_float),
      (constant_integer alignment_of_double_float))
  | LONG_DOUBLE_FLOAT ->
      ((constant_integer size_of_long_double_float),
      (constant_integer alignment_of_long_double_float))
  | SINGLE_COMPLEX ->
      ((constant_integer (mult_num (num_of_int 2) size_of_single_float)),
      (constant_integer alignment_of_single_float))
  | DOUBLE_COMPLEX ->
      ((constant_integer (mult_num (num_of_int 2) size_of_double_float)),
      (constant_integer alignment_of_double_float))
  | LONG_DOUBLE_COMPLEX ->
      ((constant_integer (mult_num (num_of_int 2) size_of_long_double_float)),
      (constant_integer alignment_of_long_double_float))
  | SINGLE_IMAGINARY ->
      ((constant_integer size_of_single_float),
      (constant_integer alignment_of_single_float))
  | DOUBLE_IMAGINARY ->
      ((constant_integer size_of_double_float),
      (constant_integer alignment_of_double_float))
  | LONG_DOUBLE_IMAGINARY ->
      ((constant_integer size_of_long_double_float),
      (constant_integer alignment_of_long_double_float))
  | (POINTER_TO _) ->
      ((constant_integer size_of_pointer),
      (constant_integer alignment_of_pointer))
  | (ARRAY_OF (tn1, None)) ->
      (begin
	if (sizeofp && (not slotp)) then
	  (syntaxerror ("Taking sizeof an incomplete array: "
			 ^ (type_to_string_body 0 tn)))
	else if (sizeofp) then
	  (* Flexible arrays have size zero. *)
	  let (s, a) = (storage_layout sizeofp slotp tn1) in
	  ((constant_integer (num_of_int 0)), a)
	else
	  let (s, a) = (storage_layout sizeofp slotp tn1) in
	  ((constant_integer (num_of_int (-1))), a)
      end)
  | (ARRAY_OF (tn1, Some e)) ->
      (begin
	let (s, a) = (storage_layout sizeofp slotp tn1) in
	let v1 = (fold_constant_integer e) in
	if (not (constant_integer_p v1)) then
	  (warn_in_parse "Taking sizeof a non-constant sized array") ;
	let s0 = (OPER (INTEGER, MUL, s, v1, NIL)) in
	let s1 = (fold_constant_integer s0) in
	(s1, a)
      end)
  | (FUNCTION_OF (_, None)) ->
      (syntaxerror "Taking sizeof a function")
  | (FUNCTION_OF (_, Some tt)) ->
      (syntaxerror "Taking sizeof a function")
  | (BITFIELD_TYPE _) ->
      (* Bitfield be handled in handling structures. *)
      (assert false)
  | (STRUCTURE_TYPE { strc_struct_or_union_p = structp }) ->
      (begin
	let body = (get_structure_body tn) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	if (structp) then
	  (storage_layout_struct sizeofp components ""
	      num_ZERO num_ONE num_ZERO num_ZERO num_ONE)
	else
	  (storage_layout_union sizeofp components num_ZERO num_ONE)
      end)
  | (ENUMERATION_TYPE _) ->
      ((constant_integer size_of_integer),
      (constant_integer alignment_of_integer))
  | PSEUDO_VA_LIST_TYPE ->
      (storage_layout sizeofp slotp type_of_va_list)
  | GCC_VA_ARG -> (syntaxerror "Bad constant expression (bad type)"))

(* Calculates the offset to the {selector} or the size of a structure
   when {selector} is "".  Note that it counts {bitoffset} in bits.
   Alignment for bit-field is slack compared to GCC. *)

and storage_layout_struct (sizeofp : bool) (l : (string * typenode_t) list) (selector : string) (offset : num) (maxalign : num) (bitoffset : num) (bfsize : num) (bfalign : num) : (expr_t * expr_t) = (
  match l with
    [] ->
      (begin
	assert (string_equal selector "") ;
	if (not (eq_num bitoffset num_ZERO)) then
	  (storage_layout_bitfield_drop sizeofp [] "" offset maxalign
	      bitoffset bfsize bfalign)
	else
	  let o = (Misc.ceiling_to offset maxalign) in
	  ((constant_integer o), (constant_integer maxalign))
      end)
  | ((s, _) :: _) when (string_equal selector s) ->
      (begin
	if (not (eq_num bitoffset num_ZERO)) then
	  (storage_layout_bitfield_drop sizeofp [] "" offset maxalign
	      bitoffset bfsize bfalign)
	else
	  let o = (Misc.ceiling_to offset maxalign) in
	  ((constant_integer o), (constant_integer maxalign))
      end)
  | ((s, tn) :: tl) ->
      (begin
	let nqtn = (strip_off_pseudo_type tn true) in
	match nqtn with
	  (BITFIELD_TYPE (tx, e)) ->
	    (begin
	      let (s0, a0) = (storage_layout sizeofp true tx) in
	      if (not (constant_integer_p s0)) then
		(syntaxerror "Bad type for bitfield") ;
	      let s = (constant_integer_value false s0) in
	      let a = (constant_integer_value false a0) in
	      let bfsize1 = (max_num bfsize s) in
	      let bfalign1 = (max_num bfalign a) in
	      let v = (fold_constant_integer e) in
	      if (constant_integer_p v) then
		(begin
		  let bits = (constant_integer_value false v) in
		  if (gt_num bits (mult_num (num_of_int 8) s)) then
		    (syntaxerror "Bad bitfield (too large)") ;
		  let bitoffset1 = (add_num bitoffset bits) in
		  if (lt_num bitoffset1 (mult_num (num_of_int 8) bfsize1)) then
		    (storage_layout_struct sizeofp tl selector
			offset maxalign bitoffset1 bfsize1 bfalign1)
		  else
		    (storage_layout_bitfield_drop sizeofp tl selector offset maxalign
			bitoffset1 bfsize1 bfalign1)
		end)
	      else
		(syntaxerror "Bad bitfield size (non-constant)")
	    end)
	| _ ->
	    (begin
	      if (not (eq_num bitoffset num_ZERO)) then
		(storage_layout_bitfield_drop sizeofp l selector offset maxalign
		    bitoffset bfsize bfalign)
	      else
		let (s0, a0) = (storage_layout sizeofp true tn) in
		if (not (constant_integer_p s0)) then
		  (syntaxerror "Bad type for structure (variable size)") ;
		let s = (constant_integer_value false s0) in
		let a = (constant_integer_value false a0) in
		let o = (Misc.ceiling_to offset a) in
		(storage_layout_struct sizeofp tl selector
		    (add_num o s) (max_num maxalign a)
		    num_ZERO num_ZERO num_ONE)
	    end)
      end))

(* Calculates the size of collected bitfields. *)

and storage_layout_bitfield_drop (sizeofp : bool) (l : (string * typenode_t) list) (selector : string) (offset : num) (maxalign : num) (bfoffset : num) (bfsize : num) (bfalign :num) : (expr_t * expr_t) = (
  let o = (Misc.ceiling_to offset bfalign) in
  (storage_layout_struct sizeofp l selector
      (add_num o bfsize) (max_num maxalign bfalign)
      num_ZERO num_ZERO num_ONE))

(* Calculates the size of a union. *)

and storage_layout_union (sizeofp : bool) (l : (string * typenode_t) list) (maxsize : num) (maxalign : num) : (expr_t * expr_t) = (
  match l with
    [] ->
      ((constant_integer (Misc.ceiling_to maxsize maxalign)),
      (constant_integer maxalign))
  | ((_, tn) :: tl) ->
      let (s0, a0) = (storage_layout sizeofp false tn) in
      if (not (constant_integer_p s0)) then
	(syntaxerror "Bad type for union (variable size)") ;
      let s = (constant_integer_value false s0) in
      let a = (constant_integer_value false a0) in
      (storage_layout_union sizeofp tl (max_num maxsize s) (max_num maxalign a)))

and get_enumerator_value (tn : typenode_t) (id : string) : num = (
  let {
    enum_utag = _ ;
    enum_components = components
  } = (get_enumeration_body tn) in
  let e = (List.assoc id components) in
  (begin
    match e with
      (CONST_I (tn, v, s)) ->
	v
    | _ -> (assert false)
  end))

let size_of_type (tn : typenode_t) : expr_t = (
  let (s, a) = (storage_layout true false tn) in
  s)

let rec offset_of_expression (e : expr_t) : num = (
  match e with
    (SREF (tn, deref, e1, s)) ->
      (begin
	let o0 = (offset_of_expression e1) in
	let tx = (strip_off_pseudo_type (type_of_expr e1) true) in
      	match tx with
	  (STRUCTURE_TYPE { strc_struct_or_union_p = structp }) ->
	    (begin
	      let body = (get_structure_body tx) in
	      let {
		strc_utag = _ ;
		strc_components = components ;
		strc_unnamed_slots = unnamedslots
	      } = body in
	      if (structp) then
		(begin
		  let (s0, _) =
		    (storage_layout_struct true components s
			num_ZERO num_ONE num_ZERO num_ZERO num_ONE) in
		  if (not (constant_integer_p s0)) then
		    (syntaxerror "Bad type for offsetof (variable size)") ;
		  let o1 = (constant_integer_value false s0) in
		  (add_num o0 o1)
		end)
	      else
		o0
	    end)
	  | _ -> (assert false)
      end)
  | (OPER (tn, (CAST false), NIL, NIL, NIL)) ->
      num_ZERO
  | _ -> (assert false))

(* ================================================================ *)

(* Returns a type of a unary conversion.  It is for (1) an operand of
   unary operators: ('!', '-', '+', '~', and '*'), (2) each operand of
   binary operators: ('<<' and '>>'), (3) simple assignment on
   arithmetic types, and (4) simple assignment with array name on the
   right hand side.  See [H&S5] 6.3.3 (p.197). *)

let rec usual_unary_conversion (srctype : typenode_t) : typenode_t = (
  let tn = (strip_off_pseudo_type srctype true) in
  match tn with
    NOTYPE ->
      (internalerror "Pseudo appears (NOTYPE)")
  | ATTRIBUTED_TYPE _ | QUALIFIED_TYPE _ -> (assert false)
  | SPECIFR_TYPE _ | TYPEDEF_TYPE _ -> (assert false)
  | STRCDEF_TYPE _ | ENUMDEF_TYPE _ -> (assert false)
  | ELLIPSIS_TYPE ->
      (internalerror "Ellipsis appears")
  | VOID_TYPE ->
      (syntaxerror "Usual unary conversion for void")
  | BOOLEAN ->
     INTEGER
  | CHARACTER | SIGNED_CHARACTER | SHORT_INTEGER ->
      INTEGER
  | UNSIGNED_CHARACTER | UNSIGNED_SHORT_INTEGER ->
      (if (unsigned_short_fit_to_int_p) then INTEGER else UNSIGNED_INTEGER)
  | INTEGER | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER ->
      tn
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT ->
      tn
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX ->
      tn
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      tn
  | (POINTER_TO _) ->
      tn
  | (ARRAY_OF (arr, _)) ->
      (make_pointer_type arr)
  | (FUNCTION_OF _) ->
      (make_pointer_type tn)
  | (BITFIELD_TYPE (reftn, _)) ->
      (usual_unary_conversion reftn)
  | (STRUCTURE_TYPE _) ->
      tn
  | (ENUMERATION_TYPE _) ->
      (enumeration_compatible_type tn)
  | PSEUDO_VA_LIST_TYPE ->
      (usual_unary_conversion type_of_va_list)
  | GCC_VA_ARG -> GCC_VA_ARG)

(* Returns a type after an argument passing conversion.  This applies
   arguments specified by a traditional form, and is mostly similar to
   the usual unary conversion except for float conversion from single
   to double. *)

let usual_argument_conversion (tn : typenode_t) : typenode_t = (
  match tn with
    SINGLE_FLOAT ->
      DOUBLE_FLOAT
  | _ ->
      (usual_unary_conversion tn))

let parameter_type_conversion (srctn : typenode_t) : typenode_t = (
  let tn = (strip_off_pseudo_type srctn true) in
  match tn with
    NOTYPE -> (internalerror "Pseudo appears (NOTYPE)")
  | ATTRIBUTED_TYPE _ | QUALIFIED_TYPE _ -> (assert false)
  | SPECIFR_TYPE _ | TYPEDEF_TYPE _ -> (assert false)
  | STRCDEF_TYPE _ | ENUMDEF_TYPE _ -> (assert false)
  | ELLIPSIS_TYPE -> (internalerror "Ellipsis appears")
  | VOID_TYPE
  | BOOLEAN
  | CHARACTER | SIGNED_CHARACTER | SHORT_INTEGER
  | UNSIGNED_CHARACTER | UNSIGNED_SHORT_INTEGER
  | INTEGER | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      tn
  | (POINTER_TO _) ->
      tn
  | (ARRAY_OF (arr, _)) ->
      (make_pointer_type arr)
  | (FUNCTION_OF _) ->
      (make_pointer_type tn)
  | (BITFIELD_TYPE (reftn, _)) ->
      tn
  | (STRUCTURE_TYPE _) ->
      tn
  | (ENUMERATION_TYPE _) ->
      tn
  | PSEUDO_VA_LIST_TYPE ->
      tn
  | GCC_VA_ARG ->
      tn)

(* Returns the common type of given types.  The usual binary
   conversion is for arguments of binary operators, and is applied to
   basic type.  See [H&S5] 6.3.4 (p.198). *)

let rec usual_binary_conversion (tn1 : typenode_t) (tn2 : typenode_t) : typenode_t = (
  if (not ((basic_type_p tn1) && (basic_type_p tn2))) then
    (* Both should be basic type. *)
    NOTYPE
  else if ((complex_type_p tn1) || (complex_type_p tn2)) then
    (begin
      let tx1 = (float_type_of_complex tn1) in
      let tx2 = (float_type_of_complex tn2) in
      (complex_type_of_float (usual_binary_conversion tx1 tx2))
    end)
  else
    (begin
      match (tn1, tn2) with
	(LONG_DOUBLE_FLOAT, _) | (_, LONG_DOUBLE_FLOAT) ->
	  LONG_DOUBLE_FLOAT
      | (DOUBLE_FLOAT, _) | (_, DOUBLE_FLOAT) ->
	  DOUBLE_FLOAT
      | (SINGLE_FLOAT, _) | (_, SINGLE_FLOAT) ->
	  SINGLE_FLOAT
      | (UNSIGNED_LONG_LONG_INTEGER, _) | (_, UNSIGNED_LONG_LONG_INTEGER) ->
	  UNSIGNED_LONG_LONG_INTEGER
      | (LONG_LONG_INTEGER, _) | (_, LONG_LONG_INTEGER) ->
	  LONG_LONG_INTEGER
      | (UNSIGNED_LONG_INTEGER, _) | (_, UNSIGNED_LONG_INTEGER) ->
	  UNSIGNED_LONG_INTEGER
      | (LONG_INTEGER, UNSIGNED_INTEGER) | (UNSIGNED_INTEGER, LONG_INTEGER) ->
	  if (unsigned_int_fit_to_long_int_p) then
	    LONG_INTEGER
	  else
	    UNSIGNED_LONG_INTEGER
      | (LONG_INTEGER, _) | (_, LONG_INTEGER) ->
	  LONG_INTEGER
      | (UNSIGNED_INTEGER, _) | (_, UNSIGNED_INTEGER) ->
	  UNSIGNED_INTEGER
      | (INTEGER, INTEGER) ->
	  INTEGER
      | _ ->
	  (internalerror "Strange type combination")
    end))

(* ================================================================ *)

(* TYPE COMPATIBILITY CHECK *)

let constant_equal_nofold_p (e0 : expr_t) (e1 : expr_t) : bool = (
  match (e0, e1) with
    ((CONST_I (t0, v0, _)), (CONST_I (t1, v1, _))) ->
      ((integral_type_p t0) && (integral_type_p t1) && (eq_num v0 v1))
  | ((CONST_I _), (CONST_F _)) | ((CONST_F _), (CONST_I _)) ->
      false
  | _ ->
      (internalerror "Needs constant folding"))

let rec compatible_type_p (linking : bool) (tn0 : typenode_t) (tn1 : typenode_t) : bool = (
  let (compat, _) = (check_compatible_type
			(if (not linking) then None else (Some StringKeyedMap.empty))
			tn0 tn1) in
  (proper_type_p compat))

(* Checks compatibility ignoring qualifiers on a pointed-to type. *)

and pointer_type_comparable_p (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : bool = (
  assert ((pointer_type_p tn0) && (pointer_type_p tn1)) ;
  let tx0 = (strip_off_pseudo_type tn0 true) in
  let tx1 = (strip_off_pseudo_type tn1 true) in
  match (tx0, tx1) with
    ((POINTER_TO ty0), (POINTER_TO ty1)) ->
      let tz0 = (strip_off_pseudo_type ty0 true) in
      let tz1 = (strip_off_pseudo_type ty1 true) in
      (compatible_type_p (linking != None) tz0 tz1)
  | _ -> (assert false))

(* Checks if two given types are compatible, and returns the composite
   type if they are compatible or returns {NOTYPE}.  Normally,
   structure/union/enumeration are not compatible across compilation
   units, and for which behavior {linking} be {None}.  For testing
   compatibility across compilation units, {linking} be specified as
   {Some []}.  A list in {linking} records checked structures to avoid
   looping by recurrence.  GCC attributes are ignored.  See [H&S5]
   5.11 (p.172). *)

and check_compatible_type (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : (typenode_t * stringmap_t option) = (
  let tx0 = (strip_off_pseudo_type tn0 false) in
  let tx1 = (strip_off_pseudo_type tn1 false) in
  match (tx0, tx1) with
    ((QUALIFIED_TYPE (ty0, quals0)), (QUALIFIED_TYPE (ty1, quals1))) ->
      let (compat, linking1) = (check_compatible_type_noquals linking ty0 ty1) in
      if ((qualifiers_equal quals0 quals1) && (proper_type_p compat)) then
	((make_qualified_type compat quals0), linking1)
      else
	(NOTYPE, linking1)
  | ((QUALIFIED_TYPE _), _) | (_, (QUALIFIED_TYPE _)) ->
      (NOTYPE, linking)
  | _ ->
      (check_compatible_type_noquals linking tx0 tx1))

and check_compatible_type_noquals (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : (typenode_t * stringmap_t option) = (
  let tx0 = (strip_off_pseudo_type tn0 false) in
  let tx1 = (strip_off_pseudo_type tn1 false) in
  if (tx0 == tx1) then
    (tx0, linking)
  else if ((basic_type_p tx0) && (basic_type_p tx1)) then
    ((check_compatible_basic_type tx0 tx1), linking)
  else
    (begin
      match (tx0, tx1) with
	((QUALIFIED_TYPE _), _) | (_, (QUALIFIED_TYPE _)) -> (assert false)
      | ((ENUMERATION_TYPE _), (ENUMERATION_TYPE _)) ->
	  (* Parser keeps identity of enumeration types. *)
	  if (tx0 == tx1) then
	    (tx0, linking)
	  else if (linking != None) then
	    (check_compatible_type_at_linking linking tx0 tx1)
	  else
	    (NOTYPE, linking)
      | ((ENUMERATION_TYPE _), _) | (_, (ENUMERATION_TYPE _)) ->
	  if (basic_type_p tx0) then
	    (begin
	      let ty1 = (enumeration_compatible_type tx1) in
	      let compat = (check_compatible_basic_type tx0 ty1) in
	      if (proper_type_p compat) then
		(tx1, linking)
	      else
		(NOTYPE, linking)
	    end)
	  else if (basic_type_p tx1) then
	    (begin
	      let ty0 = (enumeration_compatible_type tx0) in
	      let compat = (check_compatible_basic_type ty0 tx1) in
	      if (proper_type_p compat) then
		(tx0, linking)
	      else
		(NOTYPE, linking)
	    end)
	  else
	    (NOTYPE, linking)
      | ((STRUCTURE_TYPE strc0), (STRUCTURE_TYPE strc1)) ->
	  (* Parser keeps identity of structure types. *)
	  if (tx0 == tx1) then
	    (tx0, linking)
	  else if (linking != None) then
	    (check_compatible_type_at_linking linking tx0 tx1)
	  else
	    (NOTYPE, linking)
      | ((STRUCTURE_TYPE _), _) | (_, (STRUCTURE_TYPE _)) ->
	  (NOTYPE, linking)
      | ((BITFIELD_TYPE (ty0, width0)), (BITFIELD_TYPE (ty1, width1))) ->
	  if (not (constant_equal_nofold_p width0 width1)) then
	    (NOTYPE, linking)
	  else
	    (check_compatible_type linking ty0 ty1)
      | ((BITFIELD_TYPE _), _) | (_, (BITFIELD_TYPE _)) ->
	  (* TODO: UNCLEAR BETWEEN INT AND BITFIELD. *)
	  (NOTYPE, linking)
      | ((POINTER_TO ty0), (POINTER_TO ty1)) ->
	  let (compat, linking1) = (check_compatible_type linking ty0 ty1) in
	  if (proper_type_p compat) then
	    ((POINTER_TO compat), linking1)
	  else
	    (NOTYPE, linking1)
      | ((POINTER_TO _), _) | (_, (POINTER_TO _)) ->
	  (NOTYPE, linking)
      | ((ARRAY_OF (ty0, (Some dim0))), (ARRAY_OF (ty1, (Some dim1)))) ->
	  let (compat, linking1) = (check_compatible_type linking ty0 ty1) in
	  if (not (proper_type_p compat)) then
	    (NOTYPE, linking1)
	  else if ((constant_expression_p dim0)
		    && (constant_expression_p dim1)) then
	    (begin
	      (* Constants of array dimension are folded. *)
	      if (constant_equal_nofold_p dim0 dim1) then
		((ARRAY_OF (compat, Some dim0)), linking1)
	      else
		(NOTYPE, linking1)
	    end)
	  else
	    (* Compatible when one or both are variable length. *)
	    ((ARRAY_OF (compat, None)), linking1)
      | ((ARRAY_OF (ty0, _)), (ARRAY_OF (ty1, _))) ->
	  (* Either dimension is None. *)
	  let (compat, linking1) = (check_compatible_type linking ty0 ty1) in
	  if (proper_type_p compat) then
	    ((ARRAY_OF (compat, None)), linking1)
	  else
	    (NOTYPE, linking1)
      | ((ARRAY_OF _), _) | (_, (ARRAY_OF _)) ->
	  (NOTYPE, linking)
      | ((FUNCTION_OF (t0, params0)), (FUNCTION_OF (t1, params1))) ->
	  (check_compatible_function_type linking (t0, params0) (t1, params1))
      | ((FUNCTION_OF _), _) | (_, (FUNCTION_OF _)) ->
	  (NOTYPE, linking)
      | ((ATTRIBUTED_TYPE _), _) | (_, (ATTRIBUTED_TYPE _)) -> (assert false)
      | ((SPECIFR_TYPE _), _) | (_, (SPECIFR_TYPE _)) -> (assert false)
      | ((TYPEDEF_TYPE _), _) | (_, (TYPEDEF_TYPE _)) -> (assert false)
      | ((STRCDEF_TYPE _), _) | (_, (STRCDEF_TYPE _)) -> (assert false)
      | ((ENUMDEF_TYPE _), _) | (_, (ENUMDEF_TYPE _)) -> (assert false)
      | (PSEUDO_VA_LIST_TYPE, PSEUDO_VA_LIST_TYPE) ->
	  (type_of_va_list, linking)
      | (PSEUDO_VA_LIST_TYPE, _) ->
	  (check_compatible_type linking type_of_va_list tx1)
      | (_, PSEUDO_VA_LIST_TYPE) ->
	  (check_compatible_type linking tx0 type_of_va_list)
      | (GCC_VA_ARG, _) | (_, GCC_VA_ARG) -> (assert false)
      | _ -> (assert false)
    end))

(* {same_type_p} is similar to {compatible_type_p}.  The differences
   are in the treatment of incomplete array types, enumeration vs
   integer types, bitfield vs integer types, and traditional vs
   prototyped function types. *)

and same_type_p (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : bool = (
  let tx0 = (strip_off_pseudo_type tn0 false) in
  let tx1 = (strip_off_pseudo_type tn1 false) in
  match (tx0, tx1) with
    ((QUALIFIED_TYPE (ty0, quals0)), (QUALIFIED_TYPE (ty1, quals1))) ->
      ((qualifiers_equal quals0 quals1) && (same_type_noquals_p linking ty0 ty1))
  | ((QUALIFIED_TYPE _), _) | (_, (QUALIFIED_TYPE _)) ->
      false
  | _ ->
      (same_type_noquals_p linking tx0 tx1))

and same_type_noquals_p (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : bool = (
  let tx0 = (strip_off_pseudo_type tn0 false) in
  let tx1 = (strip_off_pseudo_type tn1 false) in
  if (tx0 == tx1) then
    true
  else if ((basic_type_p tx0) && (basic_type_p tx1)) then
    (proper_type_p (check_compatible_basic_type tx0 tx1))
  else
    (begin
      match (tx0, tx1) with
	((QUALIFIED_TYPE _), _) | (_, (QUALIFIED_TYPE _)) -> (assert false)
      | ((ENUMERATION_TYPE _), (ENUMERATION_TYPE _)) ->
	  (* Parser keeps identity of enumeration types. *)
	  (tx0 == tx1)
      | ((BITFIELD_TYPE (ty0, _)), (BITFIELD_TYPE (ty1, _))) ->
	  (* Parser keeps identity of bitfield types. *)
	  (ty0 == ty1)
      | ((STRUCTURE_TYPE _), (STRUCTURE_TYPE _)) ->
	  (* Parser keeps identity of structure types. *)
	  (tx0 == tx1)
      | ((POINTER_TO ty0), (POINTER_TO ty1)) ->
	  (same_type_p linking ty0 ty1)
      | ((ARRAY_OF (ty0, None)), (ARRAY_OF (ty1, None))) ->
	  (same_type_p linking ty0 ty1)
      | ((ARRAY_OF (ty0, (Some dim0))), (ARRAY_OF (ty1, (Some dim1)))) ->
	  if ((constant_expression_p dim0)
	       && (constant_expression_p dim1)) then
	    ((same_type_p linking ty0 ty1) && (constant_equal_nofold_p dim0 dim1))
	  else
	    false
      | ((FUNCTION_OF (ty0, params0)), (FUNCTION_OF (ty1, params1))) ->
	  (same_function_type_p linking (ty0, params0) (ty1, params1))
      | ((ATTRIBUTED_TYPE _), _) | (_, (ATTRIBUTED_TYPE _)) -> (assert false)
      | ((SPECIFR_TYPE _), _) | (_, (SPECIFR_TYPE _)) -> (assert false)
      | ((TYPEDEF_TYPE _), _) | (_, (TYPEDEF_TYPE _)) -> (assert false)
      | ((STRCDEF_TYPE _), _) | (_, (STRCDEF_TYPE _)) -> (assert false)
      | ((ENUMDEF_TYPE _), _) | (_, (ENUMDEF_TYPE _)) -> (assert false)
      | (PSEUDO_VA_LIST_TYPE, PSEUDO_VA_LIST_TYPE) ->
	  true
      | (PSEUDO_VA_LIST_TYPE, _) ->
	  (same_type_p linking type_of_va_list tx1)
      | (_, PSEUDO_VA_LIST_TYPE) ->
	  (same_type_p linking tx0 type_of_va_list)
      | (GCC_VA_ARG, _) | (_, GCC_VA_ARG) -> (assert false)
      | _ ->
	  false
    end))

and check_compatible_basic_type (tn0 : typenode_t) (tn1 : typenode_t) : typenode_t = (
  let tx0 = (strip_off_pseudo_type tn0 false) in
  let tx1 = (strip_off_pseudo_type tn1 false) in
  let (_, quals0) = (strip_off_pseudo_type_return_qualifiers tx0) in
  let (_, quals1) = (strip_off_pseudo_type_return_qualifiers tx1) in
  assert (qualifiers_equal quals0 NILQUAL) ;
  assert (qualifiers_equal quals1 NILQUAL) ;
  if (tn0 == tn1) then
    tn0
  else
    NOTYPE)

(* Matches lists of types of parameters, and returns a list of types.
   It signals an error when lists mismatch. *)

and check_compatible_type_on_parameters (linking : stringmap_t option) (l0 : typenode_t list) (l1 : typenode_t list) (compatlist : typenode_t list) : typenode_t list = (
  match (l0, l1) with
    ([], []) ->
      compatlist
  | ([], _) | (_, []) ->
      (syntaxerror "Parameters incompatible")
  | ([ELLIPSIS_TYPE], [ELLIPSIS_TYPE]) ->
      (compatlist @ [ELLIPSIS_TYPE])
  | ([ELLIPSIS_TYPE], _) | (_, [ELLIPSIS_TYPE]) ->
      (syntaxerror "Parameters incompatible")
  | ((hd0 :: tl0), (hd1 :: tl1)) ->
      let (compat, _) = (check_compatible_type linking hd0 hd1) in
      if (proper_type_p compat) then
	(check_compatible_type_on_parameters linking tl0 tl1 (compatlist @ [compat]))
      else
	(syntaxerror "Parameters incompatible"))

(* Checks a list of types of parameters appropriate for an traditional
   parameters, and returns a list of types.  It returns signals an
   error when lists mismatch. *)

and check_compatible_type_on_parameters_traditional (linking : stringmap_t option) (tl : typenode_t list) (compatlist : typenode_t list) : typenode_t list = (
  match tl with
    [] ->
      compatlist
  | [ELLIPSIS_TYPE] ->
      (syntaxerror "Parameters incompatible")
  | (hd :: tl) ->
      let (compat, linking1) = (check_compatible_type linking (usual_argument_conversion hd) hd) in
      if (proper_type_p compat) then
	(check_compatible_type_on_parameters_traditional linking1 tl
	    (compatlist @ [compat]))
      else
	(syntaxerror "Parameters incompatible"))

and same_function_type_p (linking : stringmap_t option) ((tn0, params0) : (typenode_t * (typenode_t list) option)) ((tn1, params1) : (typenode_t * (typenode_t list) option)) : bool = (
  if (same_type_p linking tn0 tn1) then
    match (params0, params1) with
      (None, None) ->
	(* Both are in traditional. *)
	true
    | ((Some []), (Some [])) ->
	(* Both are with no arguments (in prototype). *)
	true
    | ((Some tl0), (Some tl1)) ->
	(* Both are in prototype. *)
	let l = (check_compatible_type_on_parameters linking tl0 tl1 []) in
	(List.length l > 0)
    | _ ->
	false
  else
    false)

and check_compatible_function_type (linking : stringmap_t option) ((tn0, params0) : (typenode_t * (typenode_t list) option)) ((tn1, params1) : (typenode_t * (typenode_t list) option)) : (typenode_t * stringmap_t option) = (
  let (returntype, linking1) = (check_compatible_type linking tn0 tn1) in
  if (not (proper_type_p returntype)) then
    (NOTYPE, linking1)
  else
    (begin
      match (params0, params1) with
	(None, None) ->
	  (* Both are traditional. *)
	  ((make_function_type_traditional returntype), linking1)
      | ((Some tl0), (Some tl1)) ->
	  (* Both are prototype. *)
	  let compats = (check_compatible_type_on_parameters linking1 tl0 tl1 []) in
	  if (List.memq NOTYPE compats) then
	    (* NEVER; Check singnals an error. *)
	    (NOTYPE, linking1)
	  else
	    ((make_function_type returntype compats), linking1)
      | ((Some tl), None) | (None, (Some tl)) ->
	  (* One is traditional, the other is prototype. *)
	  let compats = (check_compatible_type_on_parameters_traditional linking1
			    tl []) in
	  if (List.memq NOTYPE compats) then
	    (* NEVER; Check singnals an error. *)
	    (NOTYPE, linking1)
	  else
	    ((make_function_type returntype compats), linking1)
    end))

(* Checks type compatibility at linking, thus, checks equality of
   members.  It returns the first argument when types are compatible.
   It does not handle empty structures (GCC extention) properly. *)

and check_compatible_type_at_linking (linking : stringmap_t option) (tn0 : typenode_t) (tn1 : typenode_t) : (typenode_t * stringmap_t option) = (
  match (tn0, tn1) with
    ((ENUMERATION_TYPE _), (ENUMERATION_TYPE _)) ->
      (begin
	let body0 = (get_enumeration_body tn0) in
	let {
	  enum_utag = { enum_tag = tag0 ; enum_unique_id = uid0 } ;
	  enum_components = components0
	} = body0 in
	let body1 = (get_enumeration_body tn1) in
	let {
	  enum_utag = { enum_tag = tag1 ; enum_unique_id = uid1 } ;
	  enum_components = components1
	} = body1 in
	assert (linking != None) ;
	if (StringKeyedMap.mem uid0 (Misc.some_of linking)) then
	  (begin
	    if (string_equal (StringKeyedMap.find uid0 (Misc.some_of linking))
		   uid1) then
	      (tn0, linking)
	    else
	      (NOTYPE, linking)
	  end)
	else if (not (string_equal tag0 tag1)) then
	  (NOTYPE, linking)
	else 
	  (begin
	    let (compat, linking1) =
	      (check_compatible_type_enumeration
		  (Some (StringKeyedMap.add uid0 uid1 (Misc.some_of linking)))
		  components0 components1) in
	    if (not compat) then
	      (NOTYPE, linking1)
	    else
	      (tn0, linking1)
	  end)
      end)
  | ((STRUCTURE_TYPE strc0), (STRUCTURE_TYPE strc1)) ->
      (begin
	let body0 = (get_structure_body tn0) in
	let {
	  strc_utag = { strc_struct_or_union_p = structp0 ;
			strc_tag = tag0 ; strc_unique_id = uid0 } ;
	  strc_components = components0 ;
	  strc_unnamed_slots = _
	} = body0 in
	let body1 = (get_structure_body tn1) in
	let {
	  strc_utag = { strc_struct_or_union_p = structp1 ;
			strc_tag = tag1 ; strc_unique_id = uid1 } ;
	  strc_components = components1 ;
	  strc_unnamed_slots = _
	} = body1 in
	assert (linking != None) ;
	if (StringKeyedMap.mem uid0 (Misc.some_of linking)) then
	  (begin
	    if (string_equal (StringKeyedMap.find uid0 (Misc.some_of linking))
		   uid1) then
	      (tn0, linking)
	    else 
	      (NOTYPE, linking)
	  end)
	else if ((structp0 != structp1) || (not (string_equal tag0 tag1))) then
	  (NOTYPE, linking)
        else if (((List.length components0) == 0)
              || ((List.length components1) == 0)) then
          (tn0, linking)
	else 
	  (begin
	    let (compat, linking1) =
	      (check_compatible_type_structure
		  (Some (StringKeyedMap.add uid0 uid1 (Misc.some_of linking)))
		  components0 components1) in
	    if (not compat) then
	      (NOTYPE, linking1)
	    else
	      (tn0, linking1)
	  end)
      end)
  | _ -> (assert false))

and check_compatible_type_enumeration (linking : stringmap_t option) (components0 : (string * expr_t) list) (components1 : (string * expr_t) list) : (bool * stringmap_t option) = (
  match (components0, components1) with
    ([], []) ->
      (true, linking)
  | ([], _) | (_, []) ->
      (false, linking)
  | (((slot0, e0) :: tl0), ((slot1, e1) :: tl1)) ->
      (* Constants of enumerator are folded. *)
      if (not (string_equal slot0 slot1)) then
	(false, linking)
      else if (not (constant_equal_nofold_p e0 e1)) then
	(false, linking)
      else
	(check_compatible_type_enumeration linking tl0 tl1))

and check_compatible_type_structure (linking : stringmap_t option) (components0 : (string * typenode_t) list) (components1 : (string * typenode_t) list) : (bool * stringmap_t option) = (
  match (components0, components1) with
    ([], []) ->
      (true, linking)
  | ([], _) | (_, []) ->
      (false, linking)
  | (((slot0, tn0) :: tl0), ((slot1, tn1) :: tl1)) ->
      if (not (string_equal slot0 slot1)) then
	(false, linking)
      else 
	(begin
	  let (compat, linking1) = (check_compatible_type linking tn0 tn1) in
	  if (not (proper_type_p compat)) then
	    (false, linking1)
	  else
	    (check_compatible_type_structure linking1 tl0 tl1)
	end))

let rec compare_expr (expr1 : expr_t) (expr2 : expr_t) : bool = (
    match expr1, expr2 with
    | PSEUDO_NONDET_VALUE (tn1, id1),
      PSEUDO_NONDET_VALUE (tn2, id2) ->
        ((String.compare id1 id2) == 0)
        && (same_type_p None tn1 tn2)
    (* Arithmetic/Logical/Conditional *)
    | OPER (tn1, op1, expr11, expr12, expr13),
      OPER (tn2, op2, expr21, expr22, expr23) ->
        (op1 == op2)
        && (compare_expr expr11 expr21)
        && (compare_expr expr12 expr22)
        && (compare_expr expr13 expr23)
        && (same_type_p None tn1 tn2)
    (* Assignment/Increment/Decrement *)
    | SETF (tn1, op1, expr11, expr12),
      SETF (tn2, op2, expr21, expr22) ->
        (op1 == op2)
        && (compare_expr expr11 expr21)
        && (compare_expr expr12 expr22)
        && (same_type_p None tn1 tn2)
    | CALL (tn1, expr11, exprs1, cspec_op1),
      CALL (tn2, expr21, exprs2, cspec_op2) ->
        (compare_expr expr11 expr21)
        && (compare_exprs exprs1 exprs2)
        && (cspec_op1 == cspec_op2)
        && (same_type_p None tn1 tn2)
    | VREF (tn1, id1, i11, i12),
      VREF (tn2, id2, i21, i22) ->
        ((String.compare id1 id2) == 0)
        && (i11 == i21)
        && (i12 == i22)
        && (same_type_p None tn1 tn2)
    | PREF (tn1, b1, expr11, expr12),
      PREF (tn2, b2, expr21, expr22) ->
        (b1 == b2)
        && (compare_expr expr11 expr21)
        && (compare_expr expr12 expr22)
        && (same_type_p None tn1 tn2)
    | SREF (tn1, b1, expr11, id1),
      SREF (tn2, b2, expr21, id2) ->
        (b1 == b2)
        && (compare_expr expr11 expr21)
        && ((String.compare id1 id2) == 0)
        && (same_type_p None tn1 tn2)
    | PROGN (tn1, stmts1),
      PROGN (tn2, stmts2) ->
        (stmts1 == stmts2)
        && (same_type_p None tn1 tn2)
    | CONST_I (tn1, num1, id1),
      CONST_I (tn2, num2, id2) ->
        ((Num.compare_num num1 num2) == 0)
        && ((String.compare id1 id2) == 0)
        && (same_type_p None tn1 tn2)
    | CONST_F (tn1, f1, id1),
      CONST_F (tn2, f2, id2) ->
        (f1 == f2)
        && ((String.compare id1 id2) == 0)
        && (same_type_p None tn1 tn2)
    | STRING (b1, id1, int_array1),
      STRING (b2, id2, int_array2) ->
        (b1 == b2)
        && ((String.compare id1 id2) == 0)
        && (int_array1 == int_array2)
    | CLITERAL (tn1, ini1),
      CLITERAL (tn2, ini2) ->
        (ini1 == ini2)
        && (same_type_p None tn1 tn2)
    | ASMSEQUENCE (exprs1),
      ASMSEQUENCE (exprs2) ->
        (compare_exprs exprs1 exprs2)
    | PHI (tn1, (str_exprs1)),
      PHI (tn2, (str_exprs2)) -> (
        (List.fold_left2 (fun b (str1, expr11) (str2, expr21) ->
                b && ((String.compare str1 str2) == 0)
                && (compare_expr expr11 expr21)
           ) true str_exprs1 str_exprs2)
        && (same_type_p None tn1 tn2)
      )
    | SIZEOF (tn11, tn12),
      SIZEOF (tn21, tn22)
    | PSEUDO_MEMORY (tn11, tn12),
      PSEUDO_MEMORY (tn21, tn22) ->
        (same_type_p None tn11 tn21)
        && (same_type_p None tn12 tn22)
    | PRAGMA (id1),
      PRAGMA (id2) ->
        (String.compare id1 id2) == 0
    | QUANTIFIER (id1, exprs1, expr11),
      QUANTIFIER (id2, exprs2, expr21) ->
        ((String.compare id1 id2) == 0)
        && (compare_exprs exprs1 exprs2)
        && (compare_expr expr11 expr21)
    | NIL, NIL
    | PSEUDO_PROLOGUE, PSEUDO_PROLOGUE
    | PSEUDO_EPILOGUE, PSEUDO_PROLOGUE -> true
    | PSEUDO_ACSL_OLD (tn1, expr11, expr12),
      PSEUDO_ACSL_OLD (tn2, expr21, expr22)
    | PSEUDO_ACSL_UNIQUEBLOCK (tn1, expr11, expr12),
      PSEUDO_ACSL_UNIQUEBLOCK (tn2, expr21, expr22)
    | DSGN_AREF (tn1, expr11, expr12),
      DSGN_AREF (tn2, expr21, expr22) ->
        (compare_expr expr11 expr21)
        && (compare_expr expr12 expr22)
        && (same_type_p None tn1 tn2)
    | PSEUDO_ACSL_RESULT (tn1, expr11),
      PSEUDO_ACSL_RESULT (tn2, expr21)
    | PSEUDO_ACSL_UNIQUE (tn1, expr11),
      PSEUDO_ACSL_UNIQUE (tn2, expr21)
    | PSEUDO_ACSL_VALID (tn1, expr11),
      PSEUDO_ACSL_VALID (tn2, expr21)
    | ALL_FIELD (tn1, expr11),
      ALL_FIELD (tn2, expr21)
    | ADDROF (tn1, expr11),
      ADDROF (tn2, expr21)
    | INIT_FILL (tn1, expr11),
      INIT_FILL (tn2, expr21) ->
        (compare_expr expr11 expr21)
        && (same_type_p None tn1 tn2)
    | _, _ -> false
)

and compare_exprs (exprs1 : expr_t list) (exprs2 : expr_t list)
: bool = (
    match exprs1, exprs2 with
    | [], [] -> true
    | expr1 :: remain_exprs1, expr2 :: remain_exprs2 -> (
        if (compare_expr expr1 expr2) == false then false
        else compare_exprs remain_exprs1 remain_exprs2
      )
    | _, _ -> false
)

let is_mem_of_exprs (ex : expr_t) (ex_lst : expr_t list) : bool =
begin
    List.exists
        (fun a_ex ->
            compare_expr ex a_ex
        ) ex_lst
end

(* Checks if the first argument is a completion of the second type.
   It ignores properness of the incomplete type, ie, int[][10][] is
   granted as proper. *)

let rec check_completion_array_type_p (tn0 : typenode_t) (tn1 : typenode_t) : bool = (
  let t0 = (strip_off_pseudo_type tn0 true) in
  let t1 = (strip_off_pseudo_type tn1 true) in
  match (t0, t1) with
    ((ARRAY_OF (tx0, _)), (ARRAY_OF (tx1, None))) ->
      (check_completion_array_type_p tx0 tx1)
  | ((ARRAY_OF (tx0, Some d0)), (ARRAY_OF (tx1, Some d1))) ->
      if (constant_equal_nofold_p d0 d1) then
	(check_completion_array_type_p tx0 tx1)
      else
	false
  | ((ARRAY_OF _), (ARRAY_OF _)) | ((ARRAY_OF _), _) | (_, (ARRAY_OF _)) ->
      false
  | _ ->
      (compatible_type_p false t0 t1))

(* Checks if an array is of dimension size zero. *)

(* NOTE: ASSUME CONSTANT FOLDING. *)

let gcc_flexible_array_type_p (tn : typenode_t) : bool = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (ARRAY_OF (_, (Some e))) ->
      (constant_equal_nofold_p e constant_zero)
  | _ ->
      false)

(* Checks if the conversion is permitted.  Note that arithmetic type
   is an union of integral and floating types, and pointer types are
   disjoint for object, void, and function types.  See [H&S5] 6.3.1
   (p.194). *)

let rec check_casting_conversion_p (dsttn : typenode_t) (srctn : typenode_t) : bool = (
  let nqdst = (strip_off_pseudo_type dsttn true) in
  let nqsrc = (usual_array_conversion (strip_off_pseudo_type srctn true)) in
  if (integral_type_p nqdst) then
    ((arithmetic_type_p nqsrc) || (pointer_type_p nqsrc))
  else if (floating_type_p nqdst) then
    (arithmetic_type_p nqsrc)
  else if (void_pointer_type_p nqdst) then
    (begin
      if (function_pointer_type_p nqsrc) then
	(warn_pedantic_in_parse "Casting function pointer to void pointer") ;
      ((integral_type_p nqsrc)
	|| (void_pointer_type_p nqsrc)
	|| (object_pointer_type_p nqsrc)
	|| (function_pointer_type_p nqsrc))
    end)
  else if (object_pointer_type_p nqdst) then
    (begin
      if (function_pointer_type_p nqsrc) then
	(warn_pedantic_in_parse "Casting function pointer to object pointer") ;
      ((integral_type_p nqsrc)
	|| (void_pointer_type_p nqsrc)
	|| (object_pointer_type_p nqsrc)
	|| (function_pointer_type_p nqsrc))
    end)
  else if (function_pointer_type_p nqdst) then
    (begin
      if ((object_pointer_type_p nqsrc) || (void_pointer_type_p nqsrc)) then
	(warn_pedantic_in_parse "Casting object pointer to function pointer") ;
      ((integral_type_p nqsrc)
	|| (void_pointer_type_p nqsrc)
	|| (function_pointer_type_p nqsrc)
	|| (object_pointer_type_p nqsrc)
        || (function_type_p nqsrc))
    end)
  else if (structure_type_p nqdst) then
    false
  else if (array_type_p nqdst) then
    false
  else if (function_type_p nqdst) then
    false
  else if (void_type_p nqdst) then
    true
  else
    false)

and usual_array_conversion (tn : typenode_t) : typenode_t = (
  let tx = (strip_off_pseudo_type tn true) in
  match tx with
  | (ARRAY_OF (arr, _)) ->
      (make_pointer_type arr)
  | _ -> tn)

(* Checks of a simple assignment conversion (a subset of casting
   conversion).  The source data needs to be checked if it is a
   constant zero.  NOTE: This does not check the constant
   qualification of an lvalue.  See [H&S5] 6.3.2 (p.195).  MEMO: If
   both are object pointers, check if the pointed objects are
   compatible, where qualifiers of the pointed objects have the
   property that destination one includes source one. *)

let check_assignment_conversion_p (tndst : typenode_t) (tnsrc : typenode_t) (srcdata : expr_t) : bool = (
  let nqdst = (strip_off_pseudo_type tndst true) in
  let nqsrc = (usual_unary_conversion tnsrc) in
  if (integral_type_p nqdst) then
    (begin
      if (nqdst == BOOLEAN) then
	((arithmetic_type_p nqsrc) || (pointer_type_p nqsrc))
      else
	(arithmetic_type_p nqsrc)
    end)
  else if (floating_type_p nqdst) then
    (arithmetic_type_p nqsrc)
  else if (void_pointer_type_p nqdst) then
    (begin
      if (function_pointer_type_p nqsrc) then
	(warn_pedantic_in_parse "Assigning function pointer to void pointer") ;
      ((constant_zero_p srcdata)
	|| (void_pointer_type_p nqsrc)
	|| (object_pointer_type_p nqsrc)
      	|| (function_pointer_type_p nqsrc))
    end)
  else if (object_pointer_type_p nqdst) then
    (begin
      if (object_pointer_type_p nqsrc) then
	(begin
	  match (nqdst, nqsrc) with
	    ((POINTER_TO d), (POINTER_TO s)) ->
	      let (dstref, dstquals) =
		(strip_off_pseudo_type_return_qualifiers d) in
	      let (srcref, srcquals) =
		(strip_off_pseudo_type_return_qualifiers s) in
	      ((qualifier_subset_p srcquals dstquals)
		&& (compatible_type_p false dstref srcref))
	  | _ -> (assert false)
	end)
      else
	((constant_zero_p srcdata)
	  || (void_pointer_type_p nqsrc))
    end)
  else if (function_pointer_type_p nqdst) then
    ((constant_zero_p srcdata)
      || (compatible_type_p false nqdst nqsrc))
  else if (structure_or_union_type_p nqdst) then
    (compatible_type_p false nqdst nqsrc)
  else if (array_type_p nqdst) then
    (syntaxerror "Assignment to array")
  else if (function_type_p nqdst) then
    (syntaxerror "Assignment to function")
  else
    (internalerror "Strange type"))

(* ================================================================ *)

(* Structure/Enumeration Type Manipulators *)

let rec get_structure_components_in_declaration (declaration : declaration_t) : (string * typenode_t) = (
  match declaration with
    {
      decl_id = (id, level) ;
      decl_tn = tn ;
      decl_scls = scls ;
      decl_inits = inits ;
      decl_params = params ;
      decl_struct_defined_p = structdefinedp;
      decl_specs = _;
      decl_location = _;
    } ->
      assert ((scls == NILSCLS) && (inits == []) (* && (params == [])*)) ;
      assert (structdefinedp == false) ;
      (id, tn))

and list_structure_components (tn : typenode_t) : string list = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE _) ->
      (begin
	let body = (get_structure_body nqtn) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	((List.map fst components) @ (List.map fst unnamedslots))
      end)
  | _ -> (assert false))

(* Sets the definition to the given structure or union type.  It
   checks the type is not defined already (ie, the type is
   incomplete).  NOTE: It should check slot name conflict, but not. *)

let rec define_structure_type ((structtype, scls, quals, attrs) : specifier_t) (definition : (string * typenode_t) list) : unit = (
  let definition =
      if union_type_p structtype then
          definition @ [("permit", INTEGER)]
      else definition
  in
  if (not (structure_incomplete_definition_p structtype)) then
    (syntaxerror "Redefining structure")
  else
    (begin
      match structtype with
	(STRUCTURE_TYPE _) ->
	  (begin
	    let (components, unnamedslots, _) =
	      (List.fold_left
		  (fun (components0, unnamedslots0, i) (slot, tn) ->
		    if ((string_equal slot "")
			 && (structure_or_union_type_p tn)) then
		      (begin
			(*GCC*) (* Handle unnamed slots. *)
			let xslot = (gensym_prefix ^ "S"
				      ^ (string_of_int (100 + i))) in
			let d = (xslot, tn) in
			((components0 @ [d]), (unnamedslots0 @ [d]), (i + 1))
		      end)
		    else
		      let d = (slot, tn) in
		      ((components0 @ [d]), unnamedslots0, (i + 1)))
		  ([], [], 0) definition) in
	    let body = (get_structure_body structtype) in
	    let subslots =
	      (List.flatten
		  (List.map
		      (fun (xslot, tn) ->
			(List.map
			    (fun subslot -> (subslot, xslot))
			    (list_structure_components tn)))
		      unnamedslots)) in
	    (body.strc_components <- components) ;
	    (body.strc_unnamed_slots <- subslots)
	  end)
      | _ ->
	  (internalerror "Non-structure type")
    end))

and structure_incomplete_definition_p (structtype : typenode_t) : bool = (
  (begin
    match structtype with
      (STRUCTURE_TYPE _) ->
	let body = (get_structure_body structtype) in
	(body.strc_components == [])
    | _ ->
	(assert false)
  end))

(* Gives a definition to an enumeration. *)

let rec define_enumeration_type ((enumtype, scls, quals, attrs) : specifier_t) (definition : (string * expr_t) list) : unit = (
  if (not (enumeration_incomplete_definition_p enumtype)) then
    (syntaxerror "Redefining enumeration")
  else
    (begin
      let components = (enumerate_enumeration_list enumtype
			   definition num_ZERO []) in
      match enumtype with
	(ENUMERATION_TYPE _) ->
	  let body = (get_enumeration_body enumtype) in
	  body.enum_components <- components
      | _ ->
	  (internalerror "Non-enumeration type")
    end))

and enumeration_incomplete_definition_p (tn : typenode_t) : bool = (
  (begin
    match tn with
      (ENUMERATION_TYPE _) ->
	let body = (get_enumeration_body tn) in
	(body.enum_components == [])
    | _ ->
	(assert false)
  end))

(* Enumerates enumeration list, and calculates enumerator values.
   Note that it temporarily assigns the partial enumeration definition
   for constant evaluation. *)

and enumerate_enumeration_list (tn : typenode_t) (l : (string * expr_t) list) (counter : num) (r : (string * expr_t) list) : (string * expr_t) list = (
  match l with
    [] ->
      r
  | ((id, e) :: tl) ->
      if (e == NIL) then
	(begin
	  let rr = (r @ [(id, (constant_integer counter))]) in
	  (enumerate_enumeration_list tn tl (succ_num counter) rr)
	end)
      else
	(begin
	  match tn with
	    (ENUMERATION_TYPE _) ->
	      let body = (get_enumeration_body tn) in
	      body.enum_components <- r ;
	      let x = (fold_constant_integer e) in
	      if (not (constant_integer_p x)) then
		(syntaxerror "Non-constant enumerator value") ;
	      let v = (constant_integer_value false x) in
	      let rr = (r @ [(id, x)]) in
	      (enumerate_enumeration_list tn tl (succ_num v) rr)
	  | _ ->
	      (internalerror "Non-enumeration type")
	end))

(* ================================================================ *)

(* Makes a cast node. *)

let make_cast (tn : typenode_t) (e : expr_t) : expr_t = (
  match e with
    (PSEUDO_NONDET_VALUE (_, s)) ->
      (* Not generate a cast node. *)
      (PSEUDO_NONDET_VALUE (tn, s))
  | _ ->
      if (check_casting_conversion_p tn (type_of_expr e)) then
	(OPER (tn, (CAST true), e, NIL, NIL))
      else
	(syntaxerror "Casting not permitted"))

(* Makes an implicit cast node if necessary.  It does not check
   whether the cast is permitted.  MEMO: It allows unchecked cast such
   as from array to pointer which appears as part of the usual unary
   conversion, and it does not appear as casting operation. *)

let make_implicit_cast (tn : typenode_t) (expr : expr_t) : expr_t = (
  if (same_type_p None tn (type_of_expr expr)) then
    expr
  else
    (OPER (tn, (CAST false), expr, NIL, NIL)))

(* Makes a unary conversion.  See comments attached to
   {usual_unary_conversion} for more details. *)

let make_usual_unary_conversion (e : expr_t) : expr_t = (
  let t1 = (usual_unary_conversion (type_of_expr e)) in
  (make_implicit_cast t1 e))

let make_usual_argument_conversion (e : expr_t) : expr_t = (
  let t1 = (usual_argument_conversion (type_of_expr e)) in
  (make_implicit_cast t1 e))

(* Makes a component selection expression for structure reference or
   union reference.  It fills the unnamed slot with a generated slot
   name. *)

let rec make_sref_dot (e : expr_t) (identifier : string) : expr_t = (
  let tn = (type_of_expr e) in
  (make_structure_reference false tn e identifier))

and make_sref_arw (e : expr_t) (identifier : string) : expr_t = (
  let ptrtn = (strip_off_pseudo_type (type_of_expr e) true) in
  match ptrtn with
    (POINTER_TO reftn) ->
      (begin
	(make_structure_reference true reftn e identifier)
      end)
  | _ ->
      (syntaxerror "Referencing to a non-pointer"))

and make_structure_reference (deref : bool) (tn0 : typenode_t) (e : expr_t) (identifier : string) : expr_t = (
  let (tn, quals) = (strip_off_pseudo_type_return_qualifiers tn0) in
  match tn with
    (STRUCTURE_TYPE _) ->
      (begin
	let unnamedslot = (check_structure_slot_unnamed tn identifier) in
	match unnamedslot with
	  (Some missingname) ->
	    (warn_pedantic_in_parse "Unnamed structure/union member") ;
	    let membertn = (get_structure_slot tn missingname) in
	    let qualtn =
	      if (deref) then
		membertn
	      else
		(add_type_qualifiers membertn quals) in
	    let subref = (SREF (qualtn, deref, e, missingname)) in
	    (make_structure_reference false membertn subref identifier)
	| None ->
	    let membertn = (get_structure_slot tn identifier) in
	    let qualtn =
	      if (deref) then
		membertn
	      else
		(add_type_qualifiers membertn quals) in
	    (SREF (qualtn, deref, e, identifier))
      end)
  | _ ->
      (syntaxerror "Referencing to a non-struct/union"))

(* Checks if a slot is unnamed, and returns the generated slot name if
   unnamed. *)

and check_structure_slot_unnamed (tn : typenode_t) (identifier : string) : string option = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE _) ->
      (begin
	let body = (get_structure_body nqtn) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	if (List.mem_assoc identifier unnamedslots) then
	  (Some (List.assoc identifier unnamedslots))
	else
	  None
      end)
  | _ -> (assert false))

(* Retrieves a type of a slot from a structure/union. *)

and get_structure_slot (tn : typenode_t) (identifier : string) : typenode_t = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    (STRUCTURE_TYPE _) ->
      (begin
	let body = (get_structure_body nqtn) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	if (List.mem_assoc identifier components) then
	  (List.assoc identifier components)
	else
	  (syntaxerror ("No member for structure or union: " ^ identifier))
      end)
  | _ -> (assert false))

(* Makes arguments list with casts. *)

let rec make_arguments (pars : typenode_t list) (args : expr_t list) (position : int) : expr_t list = (
  match (pars, args) with
    ([], []) ->
      []
  | ([ELLIPSIS_TYPE], _) ->
      (* Make a list of the rest of arguments. *)
      (List.map make_usual_argument_conversion args)
  | ((_ :: _), []) | ([], (_ :: _)) ->
      (syntaxerror "Number of arguments mismatch")
  | ((p :: ptl), (a :: atl)) ->
      if (check_assignment_conversion_p p (type_of_expr a) a) then
	((make_implicit_cast (type_of_expr a) a)
  :: (make_arguments ptl atl (succ position)))

      else
	(syntaxerror ("Arguments incompatible at position "
		       ^ (string_of_int (succ position)))))

let make_sizeof_operation (tn : typenode_t) : expr_t = (
  (SIZEOF ((make_qualified_type result_of_sizeof_type CONSTANT),
	  (strip_off_pseudo_type tn true))))

(* Makes a component selection expression for {offsetof} argument,
   whose type is {NOTYPE}. *)

let rec make_offsetof_operation (spec : specifier_t) (e : expr_t) : expr_t = (
  let (tn, scls, quals, attrs) = spec in
  let x = (make_offsetof_expression tn e) in
  let n = (offset_of_expression x) in
  let tx = (size_t_type ()) in
  (CONST_I (tx, n, (string_of_num n))))

and make_offsetof_expression (tn : typenode_t) (e : expr_t) : expr_t = (
  match e with
    (SREF (NOTYPE, deref, NIL, s)) ->
      let base = (OPER (tn, (CAST false), NIL, NIL, NIL)) in
      if (deref) then
	(make_sref_arw base s)
      else
	(make_sref_dot base s)
  | (SREF (NOTYPE, deref, e1, s)) ->
      if (deref) then
	(make_sref_arw (make_offsetof_expression tn e1) s)
      else
	(make_sref_dot (make_offsetof_expression tn e1) s)
  | _ -> (assert false))

(* Makes a component selection expression for {offsetof} argument,
   which is soon decoded in {make_offsetof_operation}. *)

let make_sref_pseudo (deref : bool) (expr : expr_t) (identifier : string) : expr_t = (
  (SREF (NOTYPE, deref, expr, identifier)))

(* Note alignment is in a small range. *)

let make_alignof_operation (tn : typenode_t) : expr_t = (
  let (s, a) = (storage_layout false false tn) in
  a)

(* Builds a unary operator node.  The argument expression is of
   <cast-expression>.  Operators and its operands, see [H&S5] 7.5
   (p.219): '-', '+': arithmetic; '!': scalar; '~': integral; '*':
   pointer reference; '&': lvalue/function; 'PRE++', 'PRE--': scalar
   lvalue; 'SIZEOF': type-name. *)

let make_unary_operation (operator : op_t) (castexpr : expr_t) : expr_t = (
  match operator with
    PSEUDO_PLS | PSEUDO_MNS ->
      let expr = (make_usual_unary_conversion castexpr) in
      let tn = (type_of_expr expr) in
      if (not (arithmetic_type_p tn)) then
	(syntaxerror "Non-arithmetic type operand")
      else if (operator == PSEUDO_MNS) then
	(OPER (tn, NEG, expr, NIL, NIL))
      else
	expr
  | PSEUDO_BAN ->
      let expr = (make_usual_unary_conversion castexpr) in
      let tn = (type_of_expr expr) in
      if (not (scalar_type_p tn)) then
	(syntaxerror "Non-scalar type operand")
      else
	(OPER (INTEGER, LOGNOT, expr, NIL, NIL))
  | PSEUDO_TLD ->
      let expr = (make_usual_unary_conversion castexpr) in
      let tn = (type_of_expr expr) in
      if (not (scalar_type_p tn)) then
	(syntaxerror "Non-integral type operand")
      else
	(OPER (tn, BITNOT, expr, NIL, NIL))
  | PSEUDO_AMP ->
      (begin
	let tn = (snd (check_lvalue_p castexpr false)) in
	let e1 = (ignore_cast castexpr) in
	match e1 with
	  (PREF (tn, _, e, _)) ->
	    (make_usual_unary_conversion e)
	| _ ->
	    assert (proper_type_p (type_of_expr castexpr)) ;
	    (ADDROF ((make_pointer_type tn), castexpr))
      end)
  | PSEUDO_STR ->
      let expr = (make_usual_unary_conversion castexpr) in
      let tn = (type_of_expr expr) in
      if (not (pointer_type_p tn)) then
	(syntaxerror "Referencing non-pointer type")
      else
	(begin
	  let reftn = (deref_pointer_type tn) in
	  (* No actual dereference for multi-dimensional arrays. *)
	  if (array_type_p reftn) then
	    (PREF (reftn, false, expr, expr))
	  else
	    (PREF (reftn, true, expr, expr))
	end)
  | INCR_PRE | DECR_PRE | INCR_POST | DECR_POST ->
      let tn = (snd (check_lvalue_p castexpr false)) in
      if (not (scalar_type_p tn)) then
	(syntaxerror "Non-scalar type operand")
      else
	(SETF (tn, operator, castexpr, NIL))
  | _ ->
      (internalerror "Strange unary operator"))

(* Binary Operator Node Constructor *)

(* Makes a binary operation expression.  The arguments are of
   <cast_expression>.  For operators, see [H&S5] 7.6 (p.227). *)

let make_binary_operation (operator : op_t) (expr0 : expr_t) (expr1 : expr_t) : expr_t = (
  let tn0 = (type_of_expr expr0) in
  let tn1 = (type_of_expr expr1) in
  match operator with
    MUL | DIV ->
      (begin
	if ((arithmetic_type_p tn0)
	     && (arithmetic_type_p tn1)) then
	  let tx0 = (usual_unary_conversion tn0) in
	  let tx1 = (usual_unary_conversion tn1) in
	  let tc = (usual_binary_conversion tx0 tx1) in
	  (OPER (tc, operator, (make_implicit_cast tc expr0),
		(make_implicit_cast tc expr1), NIL))
	else
	  (syntaxerror "Non-arithmetic type operand")
      end)
  | MOD | AND | XOR | IOR ->
      if ((integral_type_p tn0) && (integral_type_p tn1)) then
	let tx0 = (usual_unary_conversion tn0) in
	let tx1 = (usual_unary_conversion tn1) in
	let tc = (usual_binary_conversion tx0 tx1) in
	(OPER (tc, operator, (make_implicit_cast tc expr0),
	      (make_implicit_cast tc expr1), NIL))
      else
	(syntaxerror "Non-integral type operand")
  | LSH | RSH ->
      if ((integral_type_p tn0) && (integral_type_p tn1)) then
	let tx0 = (usual_unary_conversion tn0) in
	let tx1 = (usual_unary_conversion tn1) in
	(OPER (tx0, operator, (make_implicit_cast tx0 expr0),
	      (make_implicit_cast tx1 expr1), NIL))
      else
	(syntaxerror "Non-integral type operand")
  | ADD ->
      let tx0 = (usual_unary_conversion tn0) in
      let tx1 = (usual_unary_conversion tn1) in
      if ((arithmetic_type_p tx0) && (arithmetic_type_p tx1)) then
	let tc = (usual_binary_conversion tx0 tx1) in
	let e0 = (make_implicit_cast tc expr0) in
	let e1 = (make_implicit_cast tc expr1) in
	(OPER (tc, operator, e0, e1, NIL))
      else if ((pointer_type_p tx0) && (integral_type_p tx1)) then
	let e0 = (make_usual_unary_conversion expr0) in
	let e1 = (make_usual_unary_conversion expr1) in
	(OPER (tx0, operator, e0, e1, NIL))
      else if ((integral_type_p tx0) && (pointer_type_p tx1)) then
	let e0 = (make_usual_unary_conversion expr0) in
	let e1 = (make_usual_unary_conversion expr1) in
	(OPER (tx1, operator, e0, e1, NIL))
      else
	(syntaxerror "Non-applicable type combination (in addition)")
  | SUB ->
      let tx0 = (usual_unary_conversion tn0) in
      let tx1 = (usual_unary_conversion tn1) in
      if ((arithmetic_type_p tx0) && (arithmetic_type_p tx1)) then
	let tc = (usual_binary_conversion tx0 tx1) in
	let e0 = (make_implicit_cast tc expr0) in
	let e1 = (make_implicit_cast tc expr1) in
	(OPER (tc, operator, e0, e1, NIL))
      else if ((pointer_type_p tx0) && (integral_type_p tx1)) then
	let e0 = (make_usual_unary_conversion expr0) in
	let e1 = (make_usual_unary_conversion expr1) in
	(OPER (tx0, operator, e0, e1, NIL))
      else if ((pointer_type_p tx0) && (pointer_type_p tx1)
		&& (compatible_type_p false tx0 tx1)) then
	let e0 = (make_usual_unary_conversion expr0) in
	let e1 = (make_usual_unary_conversion expr1) in
	(OPER (pointer_difference_integral_type, operator, e0, e1, NIL))
      else
	(syntaxerror "Non-applicable type combination (in subtraction)")
  | LT | LE | GT | GE | EQ | NE ->
      let tx0 = (usual_unary_conversion tn0) in
      let tx1 = (usual_unary_conversion tn1) in
      if ((arithmetic_type_p tx0) && (arithmetic_type_p tx1)) then
	let tc = (usual_binary_conversion tx0 tx1) in
	let e0 = (make_implicit_cast tc expr0) in
	let e1 = (make_implicit_cast tc expr1) in
	(OPER (INTEGER, operator, e0, e1, NIL))
      else if ((pointer_type_p tx0) && (pointer_type_p tx1)
		&& (pointer_type_comparable_p None tx0 tx1)) then
	(OPER (INTEGER, operator, expr0, expr1, NIL))
      else if ((pointer_type_p tx0) && (constant_zero_p expr1)) then
	let e0 = (make_usual_unary_conversion expr0) in
	let e1 = (make_implicit_cast tx0 expr1) in
	(OPER (INTEGER, operator, e0, e1, NIL))
      else if ((constant_zero_p expr0) && (pointer_type_p tx1)) then
	let e0 = (make_implicit_cast tx1 expr0) in
	let e1 = (make_usual_unary_conversion expr1) in
	(OPER (INTEGER, operator, e0, e1, NIL))
      else
	(syntaxerror "Non-applicable type combination (in relational)")
  | _ ->
      (internalerror "Strange binary operator"))

(* Removes casts on "$nondet".  It is called at reducing assignment
   expression to expression in the grammar rule. *)

let rec make_pseudo_value_simpler (e : expr_t) : expr_t = (
  match e with
    NIL ->
      e
  | (PSEUDO_NONDET_VALUE (tn, s)) ->
      e
  | (OPER (tn, (CAST _), e1, _, _)) ->
      (* Note call below with {e} (not {e1}). *)
      (remove_cast_on_pseudo_value tn e)
  | (OPER (tn, op, e0, e1, e2)) ->
      (OPER (tn, op,
	    (make_pseudo_value_simpler e0),
	    (make_pseudo_value_simpler e1),
	    (make_pseudo_value_simpler e2)))
  | (SETF (tn, op, e0, e1)) ->
      (SETF (tn, op,
	    (make_pseudo_value_simpler e0),
	    (make_pseudo_value_simpler e1)))
  | (VREF (tn, id, level, sequence)) ->
      e
  | (PREF (tn, deref, e0, e1)) ->
      (PREF (tn, deref, (make_pseudo_value_simpler e0),
                        (make_pseudo_value_simpler e1)))
  | (SREF (tn, deref, e0, s)) ->
      (SREF (tn, deref, (make_pseudo_value_simpler e0), s))
  | (DSGN_AREF _) -> (assert false)
  | (CALL (tn, e0, ee, _)) ->
      e
  | (ADDROF (tn, e0)) ->
      e
  | (SIZEOF (tn, tt)) ->
      e
  | (PROGN (tn, l)) ->
      e
  | (CONST_I (tn, v, s)) ->
      e
  | (CONST_F (tn, v, s)) ->
      e
  | (STRING (b, s, _)) ->
      e
  | (CLITERAL (tn, l)) ->
      e
  | (ASMSEQUENCE l) ->
      e
  | (PHI (tn, phi)) ->
      e
  | (PSEUDO_MEMORY (tn, tt)) ->
      e
  | (INIT_FILL _) -> (assert false)
  | (PRAGMA s) ->
      e
  | (QUANTIFIER (s, v, e0)) ->
      e
  | PSEUDO_PROLOGUE ->
      e
  | PSEUDO_EPILOGUE ->
      e
  | PSEUDO_ACSL_RESULT (tn, e0) ->
      PSEUDO_ACSL_RESULT (tn, make_pseudo_value_simpler e0)
  | PSEUDO_ACSL_OLD (tn, e0, e1) ->
      PSEUDO_ACSL_OLD(tn, make_pseudo_value_simpler e0,
                          make_pseudo_value_simpler e1)
  | ALL_FIELD (tn, e0) ->
      ALL_FIELD (tn, make_pseudo_value_simpler e0)
  | PSEUDO_ACSL_UNIQUE (tn, e0) ->
      PSEUDO_ACSL_UNIQUE (tn, make_pseudo_value_simpler e0)
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) ->
      PSEUDO_ACSL_UNIQUEBLOCK (tn, make_pseudo_value_simpler e0,
                                   make_pseudo_value_simpler e1)
  | PSEUDO_ACSL_VALID (tn, e0) ->
      PSEUDO_ACSL_VALID (tn, make_pseudo_value_simpler e0))

and remove_cast_on_pseudo_value (tndst : typenode_t) (e : expr_t) : expr_t = (
  match e with
    (PSEUDO_NONDET_VALUE (tn, s)) ->
      if (tndst == tn) then
	e
      else
	(begin
	  assert (check_casting_conversion_p tndst tn) ;
	  (PSEUDO_NONDET_VALUE (tndst, s))
	end)
  | (OPER (tn, (CAST explicitp), e1, _, _)) ->
      (begin
	let x1 = (remove_cast_on_pseudo_value tndst e1) in
	match x1 with
	  (PSEUDO_NONDET_VALUE (_, _)) ->
	    x1
	| _ ->
	    (OPER (tn, (CAST explicitp), x1, NIL, NIL))
      end)
  | _ ->
      (make_pseudo_value_simpler e))

(* Makes assignment operator expression.  See [H&S5] 7.9 (p.246). *)

let rec make_assignment (op : op_t) (e0 : expr_t) (e1 : expr_t) : expr_t = (
  (make_pseudo_value_simpler (make_assignment_body op e0 e1)))

and make_assignment_body (op : op_t) (expr0 : expr_t) (expr1 : expr_t) : expr_t = (
  let tn0 = (type_of_expr expr0) in
  let tn1 = (type_of_expr expr1) in
  let (nqtn0, _) = (strip_off_pseudo_type_return_qualifiers tn0) in
  let (nqtn1, _) = (strip_off_pseudo_type_return_qualifiers tn1) in
  if (not (fst (check_lvalue_p expr0 false))) then
    (syntaxerror "Assigning to non-lvalue") ;
  (begin
    match op with
      SET ->
	(begin
	  if (check_assignment_conversion_p nqtn0 nqtn1 expr1) then
	    (SETF (tn0, op, expr0, (make_implicit_cast nqtn0 expr1)))
	  else
	    (syntaxerror "Non-allowable assignment")
	end)
    | ADD | SUB ->
	if ((arithmetic_type_p tn0) && (arithmetic_type_p tn1)) then
	  (SETF (tn0, op, expr0,
		(make_usual_unary_conversion expr1)))
	else if ((pointer_type_p tn0) && (integral_type_p tn1)) then
	  (SETF (tn0, op, expr0,
		(make_usual_unary_conversion expr1)))
	else
	  (syntaxerror "Non-allowable assignment")
    | MUL | DIV ->
	if ((arithmetic_type_p tn0) && (arithmetic_type_p tn1)) then
	  (SETF (tn0, op, expr0,
		(make_usual_unary_conversion expr1)))
	else
	  (syntaxerror "Non-allowable assignment")
    | MOD | LSH | RSH | AND | XOR | IOR ->
	if ((integral_type_p tn0) && (integral_type_p tn1)) then
	  (SETF (tn0, op, expr0,
		(make_usual_unary_conversion expr1)))
	else
	  (syntaxerror "Non-allowable assignment")
    | _ ->
	(internalerror "Bad operator")
  end))

(* Makes conditional operator expression.  The conditional expressions
   include question-conditional, logical and/or expressions.  See
   [H&S5] 7.8 (p.244).  The 2nd argument of COND can be NIL for "?:"
   in GCC extension. *)

let rec make_conditional_expr (operator : op_t) (expr0 : expr_t) (expr1 : expr_t) (expr2 : expr_t) : expr_t = (
  let tx0 = (type_of_expr expr0) in
  let tx1 = (type_of_expr expr1) in
  let tx2 = (type_of_expr expr2) in
  match operator with
    COND ->
      (begin
	let tn =
	  if (expr1 == NIL) then
	    tx2
	  else
	    (check_type_of_conditional tx1 tx2 expr1 expr2) in
	if (proper_type_p tn) then
	  (OPER (tn, COND, expr0, expr1, expr2))
	else
	  (syntaxerror "Illegal type combination for ?:")
      end)
   | OROR ->
       let (tn, _) = (check_compatible_type None tx0 tx1) in
       if (tn == NOTYPE) then
	 (OPER (INTEGER, OROR, expr0, expr1, NIL))
       else
	 (OPER (tn, OROR, expr0, expr1, NIL))
   | ANDAND ->
       let (tn, _) = (check_compatible_type None tx0 tx1) in
       if (tn == NOTYPE) then
	 (OPER (INTEGER, ANDAND, expr0, expr1, NIL))
       else
	 (OPER (tn, ANDAND, expr0, expr1, NIL))
   (*NONSTANDARD*)
   | IMP ->
       (OPER (INTEGER, IMP, expr0, expr1, NIL))
   | IFF ->
       (OPER (INTEGER, IFF, expr0, expr1, NIL))
   | _ ->
       (assert false))

(* Makes a result type for conditional.  See [H&S5] 7.8 (p.244). *)

and check_type_of_conditional (tn0 : typenode_t) (tn1 : typenode_t) (e0 : expr_t) (e1 : expr_t) : typenode_t = (
  if ((void_type_p tn0) || (void_type_p tn1)) then
    VOID_TYPE
  else
    (begin
      let tx0 = (usual_unary_conversion tn0) in
      let tx1 = (usual_unary_conversion tn1) in
      if ((arithmetic_type_p tx0) || (arithmetic_type_p tx1)) then
	(begin
	  if ((arithmetic_type_p tx0) && (arithmetic_type_p tx1)) then
	    (usual_binary_conversion tx0 tx1)
	  else
	    NOTYPE
	end)
      else if (((structure_type_p tx0) || (structure_type_p tx1))
		|| ((union_type_p tx0) || (union_type_p tx1))) then
	(begin
	  let (compat, _) = (check_compatible_type None tx0 tx1) in
	  if (proper_type_p compat) then
	    compat
	  else
	    NOTYPE
	end)
      else if ((void_type_p tx0) || (void_type_p tx1)) then
	VOID_TYPE
      else if ((pointer_type_p tx0) || (pointer_type_p tx1)) then
	(begin
	  let tr0 = (deref_pointer_type tx0) in
	  let tr1 = (deref_pointer_type tx1) in
	  let (tz0, quals0) = (strip_off_pseudo_type_return_qualifiers tr0) in
	  let (tz1, quals1) = (strip_off_pseudo_type_return_qualifiers tr1) in
	  let (compat, _) = (check_compatible_type None tz0 tz1) in
	  let newquals = (merge_qualifiers false quals0 quals1) in
	  if (proper_type_p compat) then
	    (make_pointer_type (make_qualified_type compat newquals))
	  else if ((void_pointer_type_p tx0) || (void_pointer_type_p tx1)) then
	    void_pointer_type
	  else if (null_pointer_p e0) then
	    (make_pointer_type (make_qualified_type tz1 newquals))
	  else if (null_pointer_p e1) then
	    (make_pointer_type (make_qualified_type tz0 newquals))
	  else
	    NOTYPE
	end)
      else
	NOTYPE
    end))

let make_return_expression (expr : expr_t option) : stmt_t = (
  let f = !current_function in
  assert (not (f == nil_declaration)) ;
  let tn = (type_of_return_value (f.decl_tn)) in
  match expr with
    None ->
      if (tn == VOID_TYPE) then
	(RETURN None)
      else
	(syntaxerror ("No return value for non-void function"))
  | (Some e1) ->
      if (tn == VOID_TYPE) then
	(syntaxerror ("Return value for void function"))
      else if (check_assignment_conversion_p tn (type_of_expr e1) e1) then
	let e2 = (make_pseudo_value_simpler (make_implicit_cast tn e1)) in
	(RETURN (Some e2))
      else
	(syntaxerror "Non-allowable assignment"))

(* Makes an array subscript expression.  It first combines expressions
   by addition to generate a proper cast.  MEMO: Pointer dereference
   when a referent is a pointer returns just an identity. *)

let make_aref (expr0 : expr_t) (expr1 : expr_t) : expr_t = (
  let ex = (make_binary_operation ADD expr0 expr1) in
  let e0 = (expr_arg0 ex) in
  let e1 = (expr_arg1 ex) in
  let t0 = (type_of_expr e0) in
  let t1 = (type_of_expr e1) in
  let ptr, base_expr, index_expr =
    if ((pointer_type_p t0) && (integral_type_p t1)) then
      e0, expr0, expr1
    else if ((integral_type_p t0) && (pointer_type_p t1)) then
      e1, expr1, expr0
    else
      (syntaxerror ("Array reference to a non-pointer")) in
  let tn = (type_of_expr ptr) in
  let ex = (make_binary_operation ADD base_expr index_expr) in
  let (nqtn, quals) = (strip_off_pseudo_type_return_qualifiers tn) in
  let reftn = (deref_pointer_type nqtn) in
  let sizeof_ex = make_sizeof_operation reftn in
  let actual_index_ex = (make_binary_operation MUL sizeof_ex index_expr) in
  let ex1 = (make_binary_operation ADD base_expr actual_index_ex) in
  (* No actual dereference for multi-dimensional arrays. *)
  if (array_type_p reftn) then
    (PREF (reftn, false, ex, ex1))
  else
    (PREF (reftn, true, ex, ex1)))

(* ================================================================ *)

(* **** SCOPE STACK **** *)

(* This section of code implements some scope stack operations and
   type definitions. *)

(* Scope Stack Operations *)

(* BINDING.  Pairs of identifiers and bindings are maintained in the
   scope stack.  The binding has the contents of: ((id, level) type
   sclass).  ID is an identifier or struct/union tag, LEVEL is a scope
   level, SCLASS is a storage class, and TYPE is a type.  For tags,
   storage class is always nil.  One appears in storage_class only one
   of TYPEDEF EXTERN STATIC AUTO REGISTER . *)

type binding_t = {
  binding_name : (string * int) ;
  binding_tn : typenode_t ;
  binding_scls : storage_class_t
}

let nil_binding = {
  binding_name = ("", 0) ;
  binding_tn = INTEGER ;
  binding_scls = AUTO
}

let binding_eql (x : binding_t) (y : binding_t) : bool = (x == y)
let binding_exists_p (b : binding_t) : bool = (not (b == nil_binding))

let binding_to_string (b : binding_t) : string = (
  ("binding(" ^ (fst b.binding_name) ^ "%"
    ^ (string_of_int (snd b.binding_name))
    ^ ":" ^ (type_to_string b.binding_tn) ^ ")"))

(* {name_scope_stack}/{tag_scope_stack} are lists of hashtables with
   the bottom one is for external definitions.  {scope_block_stack} is
   a list of block numbers distinguishing the scope of a variable in a
   function.  {scope_block_counter} is incremented on seeing a block,
   and its value is pushed on to {scope_block_stack}.  Value 1 is for
   external definitions. *)

let name_scope_stack = ref ([] : (string, binding_t) Hashtbl.t list)
let tag_scope_stack = ref ([] : (string, binding_t) Hashtbl.t list)
let scope_block_stack = ref ([] : int list)
let scope_block_counter = ref 1
let pragma_scope_stack = ref ([] : (string list) Vector.t list)

(* {scope_stack_history} is for debugging: for each compound
   statement, a pair of the two scope_stacks are pushed onto it. *)

let scope_stack_history =
  (ref ([] : (((string, binding_t) Hashtbl.t list) list) list))

(*let program = ref ([] : stmt_t list)*)

(* Checks if a given type is in the current scope level.  It is used
   to check the structure/union definition is in the scope. *)

let check_type_in_current_scope_p (tn : typenode_t) (scope_stack : (string, binding_t) Hashtbl.t list) : bool = (
  (Misc.hashtbl_mem_if
      (fun k b -> (b.binding_tn == tn))
      (List.hd scope_stack)))

(* Checks if a given identifier is in the current scope level.  It
   returns a binding or {nil_binding}. *)

let check_id_in_current_scope (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) : binding_t = (
  if (Hashtbl.mem (List.hd scope_stack) id) then
    (Hashtbl.find (List.hd scope_stack) id)
  else
    nil_binding)

(* Searches an {id} in {scope_stack} and returns its binding.  True
   {in_current_scope} limits search only in the current scope. *)

let rec get_binding (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) (in_current_scope_p : bool) : binding_t = (
  if (in_current_scope_p) then
    (begin
      if (Hashtbl.mem (List.hd scope_stack) id) then
	(Hashtbl.find (List.hd scope_stack) id)
      else
	nil_binding
    end)
  else
    (begin
      match scope_stack with
	[] ->
	  nil_binding
      | (hd :: tl) ->
	  let b : binding_t = (get_binding id scope_stack true) in
	  if (binding_exists_p b) then
	    b
	  else
	    (get_binding id tl false)
    end))

let add_binding_shadowing ((id, level) : varid_t) (scls : storage_class_t) (tn : typenode_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  (* Record size_t definition. *)
  if (scls == TYPEDEF
       && (varid_equal (id, level) ("size_t", global_scope_level))) then
    (begin
      if (!defined_size_t_type != None) then
	(warn_in_parse "Defining size_t type twice") ;
      defined_size_t_type := (Some tn)
    end) ;
  let binding = {
    binding_name = (id, level) ;
    binding_tn = tn ;
    binding_scls = scls
  }
  in
  (Misc.hashtbl_unique_add (List.hd scope_stack) id binding))

let add_binding_replacing ((id, level) : varid_t) (scls : storage_class_t) (tn : typenode_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  let binding = {
    binding_name = (id, level) ;
    binding_tn = tn ;
    binding_scls = scls
  }
  in
  (Hashtbl.replace (List.hd scope_stack) id binding))

(* Adds a new binding to the scope-stack.  Also, it checks the
   existence in the current scope, because it is an error.  This
   returns the newly created binding.  REBINDING ERROR SHOULD BE
   DETECTED IN UPPER LEVELS. *)

let add_binding (id, level) (scls : storage_class_t) (tn : typenode_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  let b = (check_id_in_current_scope id scope_stack) in
  assert (not (binding_exists_p b)) ;
  (add_binding_shadowing (id, level) scls tn scope_stack))

(* Defines a given identifier and adds a binding in the current scope.
   NOTE: An extern-declared local identifiers are given the global
   scope level. *)

let define_identifier_in_scope (decl : declaration_t) : declaration_t = (
  let scope_stack = !name_scope_stack in
  match decl with
    {
      decl_id = (id, level) ;
      decl_tn = tn ;
      decl_scls = scls ;
      decl_inits = _ ;
      decl_params = _ ;
      decl_struct_defined_p = _;
      decl_specs = _;
      decl_location = _;
    } ->
      assert (not (string_equal id "")) ;
      assert (level == (List.hd !scope_block_stack)) ;
      let b = (check_id_in_current_scope id scope_stack) in
      if (not (binding_exists_p b)) then
	(begin
	  if (external_declaration_p decl) then
	    (begin
	      (* Fix external identfier as global level. *)
	      let fixedlevel = global_scope_level in
	      (add_binding_shadowing (id, fixedlevel) scls tn scope_stack) ;
	      { decl with decl_id = (id, fixedlevel) }
	    end)
	  else
	    (begin
	      (add_binding_shadowing (id, level) scls tn scope_stack) ;
	      decl
	    end)
	end)
      else if ((function_type_p tn)
		&& (compatible_type_p false tn b.binding_tn)) then
	(begin
	  (* Replace previous definition. *)
	  (add_binding_replacing (id, level) scls tn scope_stack) ;
	  decl
	end)
      else if ((array_type_p tn)
		&& (incomplete_array_type_p b.binding_tn)
		&& (check_completion_array_type_p tn b.binding_tn)) then
	(begin
	  (* Replace previous definition. *)
	  (add_binding_replacing (id, level) scls tn scope_stack) ;
	  decl
	end)
      else if ((external_declaration_p decl)
		&& (compatible_type_p false tn b.binding_tn)) then
	(begin
	  (* Multiple extern variables are allowed. *)
	  decl
	end)
      else if ((storage_class_external_p b.binding_scls)
		&& (compatible_type_p false tn b.binding_tn)) then
	(begin
	  (* Multiple extern variables are allowed. *)
	  decl
	end)
      else if ((compatible_type_p false tn b.binding_tn)) then (begin
	  (add_binding_replacing (id, level) scls tn scope_stack) ;
           decl
      end)
      else
	(syntaxerror ("Rebinding identifier: "
		       ^ id ^ "%" ^ (string_of_int level))))

(* Defines an ID as enumerator in the current scope. *)

let define_enumerator (enumtn : typenode_t) (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  match enumtn with
    (ENUMERATION_TYPE _) ->
      let level = (List.hd !scope_block_stack) in
      (add_binding (id, level) NILSCLS enumtn scope_stack) ;
  | _ -> (assert false))

(* Defines an identifier as a local label in the current scope.
   Identifiers with AUTO and NOTYPE are regarded as labels. *)

let define_local_label (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  assert (not (string_equal id "")) ;
  let level = (List.hd !scope_block_stack) in
  (add_binding_shadowing (id, level) AUTO NOTYPE scope_stack))

let binding_is_label_p (b : binding_t) : bool = (
  ((binding_exists_p b) && b.binding_scls == AUTO && b.binding_tn == NOTYPE))

let make_goto_stmt (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) : stmt_t = (
  let b = (get_binding id scope_stack false) in
  if (binding_is_label_p b) then
    let (id, level) = (b.binding_name) in
    (GOTO (id ^ "%" ^ (string_of_int level)))
  else
    (GOTO id))

let make_named_label (id : string) (attrs : gcc_attribute_t list) (scope_stack : (string, binding_t) Hashtbl.t list) : stmt_t = (
  if (attrs != []) then
    (warn_in_parse ("GCC attribute to label ignored")) ;
  let b = (get_binding id scope_stack false) in
  if (binding_is_label_p b) then
    let (id, level) = (b.binding_name) in
    (LABEL (NAME_LABEL (id ^ "%" ^ (string_of_int level))))
  else
    (LABEL (NAME_LABEL id)))

(* Takes a label value (a GCC extension).  It returns an address-of
   expression with a reference of {NOTYPE} (other address-of
   expressions have proper types).  Normal labels have the parameter
   scope level, and local labels have levels at their declarations.
   An address-of expression will later be replaced by a variable
   created to represent the label value, where the variable name is a
   concatenation of (file-name "%" hash-value "%" function-name "%"
   scope-level "%" label-name "%" 1 "%" 1).  The last two ones are of
   the scope level and the SSA version.  See {make_value_for_label} in
   "csyntax.ml". *)

let make_label_value (id : string) (scope_stack : (string, binding_t) Hashtbl.t list) : expr_t = (
  let b = (get_binding id scope_stack false) in
  let level =
    if (binding_is_label_p b) then
      (snd (b.binding_name))
    else
      parameter_scope_level in
  (ADDROF (void_pointer_type, (VREF (NOTYPE, id, level, 0)))))

(* ================================================================ *)

(* Checks if function declaration is proper. *)

let check_function_declaration (declaration : declaration_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  let (id, level) = (declaration.decl_id) in
  let tn = (declaration.decl_tn) in
  assert (level == 1) ;
  current_function := declaration ;
  let externalscope = [(Misc.list_last scope_stack)] in
  let b = (get_binding id externalscope false) in
  (* It should be declared already in {declarator} rule. *)
  assert (binding_exists_p b) ;
  let (compat, _) = (check_compatible_type None (b.binding_tn) tn) in
  if (not (proper_type_p compat)) then
    (* It has incompatible type. *)
    (syntaxerror "Already declared in incompatible type"))

(* Separates types and identifiers in a <parameter_type_list>
   declaration.  It returns a tuple of two lists of identifiers and
   types.  If no identifiers are specified, return [] as
   identifiers. *)

let rec split_parameter_types_and_ids (paramdecls : declaration_t list) : (string list * typenode_t list) = (
  let (ids, types) = (split_parameter_types_and_ids_loop paramdecls [] []) in
  if (types == []) then
    (* No parameters. *)
    ([], [])
  else if ((List.length types) == 1 && ((List.hd types) == VOID_TYPE)) then
    (* No parameters. *)
    ([], [])
  else
    (ids, types))

and split_parameter_types_and_ids_loop (decls : declaration_t list) (ids : string list) (types : typenode_t list) : (string list * typenode_t list) = (
  match decls with
    [] ->
      (ids, types)
  | (decl :: tl) ->
      if (decl == pseudo_ellipsis) then
	((ids @ [ellipsis_string]), (types @ [ELLIPSIS_TYPE]))
      else
	(begin
	  match decl with
	    {
	      decl_id = (id, level) ;
	      decl_tn = tn ;
	      decl_scls = _ ;
	      decl_inits = _ ;
	      decl_params = _ ;
	      decl_struct_defined_p = _
	    } ->
	      let tx = (parameter_type_conversion tn) in
	      (split_parameter_types_and_ids_loop tl
		  (ids @ [id]) (types @ [tx]))
	end))

(* Searches a {tag} in the scope stack and returns a bound
   structure/union type if it is found, or make and returns a new one.
   If a given {tag} is "", this just makes and returns a new one.  If
   the key {definingp} is true, search is performed only in the
   current scope. *)

let find_or_make_structure_type (structp : bool) (tag : string) (scope_stack : (string, binding_t) Hashtbl.t list) (definingp : bool) : typenode_t = (
  if (string_equal tag "") then
    (make_structure_type structp tag)
  else
    (begin
      let binding = (get_binding tag scope_stack definingp) in
      if (binding_exists_p binding) then
	(begin
	  let tn = (binding.binding_tn) in
	  match tn with
	    (STRUCTURE_TYPE _) ->
	      tn
	  | _ ->
	      (syntaxerror "Struct/union/enum tag abuse")
	end)
      else
	(begin
	  let tn = (make_structure_type structp tag) in
	  let level = (List.hd !scope_block_stack) in
	  (add_binding (tag, level) NILSCLS tn scope_stack) ;
	  tn
	end)
    end))

(* Searches {tag}, similar to the one for the structure type. *)

let find_or_make_enumeration_type (tag : string) (scope_stack : (string, binding_t) Hashtbl.t list) (definingp : bool) : typenode_t = (
  if (string_equal tag "") then
    (make_enumeration_type tag)
  else
    (begin
      let binding = (get_binding tag scope_stack definingp) in
      if (binding_exists_p binding) then
	(begin
	  let enumtype = binding.binding_tn in
	  match enumtype with
	    (ENUMERATION_TYPE _) ->
	      enumtype
	  | _ ->
	      (syntaxerror "struct/union/enum tag abuse")
	end)
      else
	(begin
	  let tn = (make_enumeration_type tag) in
	  let level = (List.hd !scope_block_stack) in
	  (add_binding (tag, level) NILSCLS tn scope_stack) ;
	  tn
	end)
    end))

(* ================================================================ *)

(* **** PROTOTYPE PARAMETER DECLARATION **** *)

let check_traditional_parameters (fundecl : declaration_t) (paramdecls : stmt_t list): unit = (
  (warn_in_parse ("CHECK_TRADITIONAL_PARAMETERS:"
		   ^ " Nothing is checked currently.\n")))

(* ================================================================ *)

(* Pseudo variable names. *)

let pseudo_result_variable = "$result"
let pseudo_varargs = "$varargs"

(* Initializes pseudo variables for annotation. *)

let add_pseudo_variables_in_scope (rettn : typenode_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  (add_binding_shadowing ("$t", 2) STATIC INTEGER scope_stack) ;
  (add_binding_shadowing ("$f", 2) STATIC INTEGER scope_stack) ;
  (add_binding_shadowing (pseudo_result_variable, 2) AUTO rettn scope_stack) ;
  (add_binding_shadowing ("$length", 2) STATIC INTEGER scope_stack))

(* Appends bindings of prototype parameters of a function
   declaration. *)

let add_binding_of_parameters (declaration : declaration_t) (scope_stack : (string, binding_t) Hashtbl.t list) : unit = (
  match declaration with
    {
      decl_id = (id, level) ;
      decl_tn = tn ;
      decl_scls = _ ;
      decl_inits = _ ;
      decl_params = params ;
      decl_struct_defined_p = _
    } ->
      (begin
	match (strip_off_pseudo_type tn true) with
	  (FUNCTION_OF (rettn, None)) ->
	    (* Traditional parameters, OK. *)
	    (add_pseudo_variables_in_scope rettn scope_stack) ;
	    ()
	| (FUNCTION_OF (rettn, (Some protos))) ->
	    (begin
	      assert ((List.length protos) == (List.length params)) ;
	      (add_pseudo_variables_in_scope rettn scope_stack) ;
	      (* Remove ellipsis in protos. *)
	      let (trueparams, trueprotos) =
		if (List.mem ELLIPSIS_TYPE protos) then
		  (begin
		    assert (ellipsis_parameter_p (Misc.list_last params)) ;
		    assert (ellipsis_type_p (Misc.list_last protos)) ;
		    ((Misc.list_droptail params), (Misc.list_droptail protos))
		  end)
		else
		  (params, protos) in
	      (List.iter2
		  (fun id tn ->
		    (add_binding (id, parameter_scope_level)
			AUTO tn scope_stack))
		  trueparams
		  trueprotos)
	    end)
	| _ ->
	    (internalerror ("Not a function type: " ^ id))
      end))

(* ================================================================ *)

(* Substitutes the base-type in {modifier} with {base}, where the
   base-type in modifiers is {NOTYPE} during declarator making.
   Modifiers are, eg, {POINTER_TO}, {ARRAY_OF}, and {FUNCTION_OF}.
   NOTE: The combination is reversed in two pointers in {**X}.  That
   is, the first {*(1)} is a base, and the 2nd {*(2)} is a modifier
   (pointer-to(2) a pointer-to(1) a some type). *)

let rec substitute_basetype_of (modifier : typenode_t) (base : typenode_t) : typenode_t = (
  match modifier with
    NOTYPE ->
      base
  | (ATTRIBUTED_TYPE (tn, a)) ->
      (ATTRIBUTED_TYPE ((substitute_basetype_of tn base), a))
  | (QUALIFIED_TYPE (tn, quals)) ->
      (QUALIFIED_TYPE ((substitute_basetype_of tn base), quals))
  | (TYPEDEF_TYPE (tn, name)) ->
      (TYPEDEF_TYPE ((substitute_basetype_of tn base), name))
  | (STRCDEF_TYPE (tn, body)) ->
      (STRCDEF_TYPE ((substitute_basetype_of tn base), body))
  | (ENUMDEF_TYPE (tn, body)) ->
      (ENUMDEF_TYPE ((substitute_basetype_of tn base), body))
  | (POINTER_TO tn) ->
      (POINTER_TO (substitute_basetype_of tn base))
  | (ARRAY_OF (tn, e)) ->
      (ARRAY_OF ((substitute_basetype_of tn base), e))
  | (FUNCTION_OF (tn, params)) ->
      (FUNCTION_OF ((substitute_basetype_of tn base), params))
  | (BITFIELD_TYPE (tn, e)) ->
      (BITFIELD_TYPE ((substitute_basetype_of tn base), e))
  | (STRUCTURE_TYPE _) ->
      (begin
	let body = (get_structure_body modifier) in
	let {
	  strc_utag = _ ;
	  strc_components = components ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	let l = (List.map
		    (fun (s, tn) -> (s, (substitute_basetype_of tn base)))
		    components) in
	(body.strc_components <- l) ;
	modifier
      end)
  | (ENUMERATION_TYPE _) ->
      modifier
  | _ ->
      (internalerror "No base type for pointer"))

(* ================================================================ *)

(* Makes a structure tag defined on the current scope, if it is not
   already.  A structure/union tag definition (without a body) without
   declarators shall shadow the tag in the outer scope.  Note that it
   is not known until to see an empty declarator.  It returns a new
   one when a new one is defined.  seeing an empty declarator. *)

(* NOTE/TODO: It creates an empty structure/union/enumeration.  Reason
   unclear. *)

let make_structure_tag_on_current_scope (tn : typenode_t) : typenode_t = (
  let truetn = (strip_off_pseudo_type tn true) in
  if (check_type_in_current_scope_p truetn !tag_scope_stack) then
    (* Already in the current scope. *)
    tn
  else
    (* Move structure definition. *)
    match truetn with
      (STRUCTURE_TYPE {
	strc_struct_or_union_p = structp ;
	strc_tag = tag
      }) ->
	(begin
	  let body = (get_structure_body truetn) in
	  let {
	    strc_utag = _ ;
	    strc_components = components ;
	    strc_unnamed_slots = unnamedslots
	  } = body in
	  if (not (string_equal tag "")) then
	  (begin
	    assert (components == []) ;
	    let newtn = (make_structure_type structp tag) in
	    let level = (List.hd !scope_block_stack) in
	    (add_binding (tag, level) NILSCLS newtn !tag_scope_stack) ;
	    newtn
	  end)
	else
	  tn
	end)
    | (ENUMERATION_TYPE { enum_tag = tag }) ->
	(begin
	  let body = (get_enumeration_body truetn) in
	  let {
	    enum_utag = _ ;
	    enum_components = components
	  } = body in
	  if (not (string_equal tag "")) then
	    (begin
	      assert (components == []) ;
	      let newtn = (make_enumeration_type tag) in
	      let level = (List.hd !scope_block_stack) in
	      (add_binding (tag, level) NILSCLS newtn !tag_scope_stack) ;
	      newtn
	    end)
	  else
	    tn
	end)
    | _ ->
	(internalerror "Not a structure type"))

(* ================================================================ *)

(* Makes a simple (ie, without storage-class and qualifiers) type
   specifier of a given type. *)

let make_specifier (tn : typenode_t) (scls : storage_class_t) (quals : type_qualifier_t) (attrs : gcc_attribute_t list) : specifier_t = (
  (tn, scls, quals, attrs))

let make_simple_specifier (tn : typenode_t) : specifier_t = (
  (tn, NILSCLS, NILQUAL, []))

(* Resolves and removes a pseudo {SPECIFR_TYPE} making a basic type in
   {specifier_t}.  Note that the given type specifier is either a
   pseudo {SPECIFR_TYPE} or a type-node of structure, union, or
   enumeration. *)

let make_proper_specifier ((typedesc, sclass, qualifiers, attributes) : specifier_t) : specifier_t = (
  let tn = (match_basic_type typedesc) in
  if (proper_type_p tn) then
    (* A basic type. *)
    (tn, sclass, qualifiers, attributes)
  else
    match typedesc with
      (SPECIFR_TYPE [typedefname]) ->
	(* Typedef name.  It is bound because it appears. *)
	let binding = (get_binding typedefname !name_scope_stack false) in
	assert (binding_exists_p binding) ;
	let tx = (make_typedef_type typedefname binding.binding_tn) in
	(tx, sclass, qualifiers, attributes)
    | (SPECIFR_TYPE []) | (SPECIFR_TYPE _) ->
	(internalerror "Empty/many basic type specifiers")
    | (STRCDEF_TYPE (tn, _)) ->
	(typedesc, sclass, qualifiers, attributes)
    | (STRUCTURE_TYPE _) ->
	(typedesc, sclass, qualifiers, attributes)
    | (ENUMDEF_TYPE (tn, body)) ->
	(typedesc, sclass, qualifiers, attributes)
    | (ENUMERATION_TYPE _) ->
	(typedesc, sclass, qualifiers, attributes)
    | _ ->
	(typedesc, sclass, qualifiers, attributes))

(* Merges (<type-specifier>,<storage-class>,<type_qualifiers>). *)

let merge_specifiers ((ts0, sc0, tq0, at0) : specifier_t) ((ts1, sc1, tq1, at1) : specifier_t) : specifier_t = (
  let ts =
    match (ts0, ts1) with
      (NOTYPE, _) -> ts1
    | (_, NOTYPE) -> ts0
    | ((SPECIFR_TYPE l0), (SPECIFR_TYPE l1)) ->
	(SPECIFR_TYPE (l0 @ l1))
    | _ ->
	(internalerror "Conflicting merge of type specifiers")
  in
  let sc =
    if (sc0 != NILSCLS && sc1 != NILSCLS) then
      (syntaxerror "Multiple storage class specified")
    else if (sc0 != NILSCLS) then
      sc0
    else
      sc1
  in
  (ts, sc, (merge_qualifiers true tq0 tq1), (at0 @ at1)))

let append_attributes_to_specifier ((tn, scls, quals, attrs) : specifier_t) (newattrs : gcc_attribute_t list) : specifier_t = (
  (tn, scls, quals, (attrs @ newattrs)))

(* Makes a simple (ie, without storage-class and qualifiers) type
   specifier of a given type. *)

let make_simple_declarator (varid : varid_t) : declarator_t = (
  (NOTYPE, varid, [], []))

let make_declarator (tn : typenode_t) (varid : varid_t) (params : string list) (attrs : gcc_attribute_t list) : declarator_t = (
  (tn, varid, params, attrs))

let modify_declarator ((tn, varid, params, attrs) : declarator_t) (newtn : typenode_t) : declarator_t = (
  (newtn, varid, params, attrs))

let append_attributes_to_declarator ((tn, varid, params, attrs) : declarator_t) (newattrs : gcc_attribute_t list) : declarator_t = (
  (tn, varid, params, (attrs @ newattrs)))

let append_to_sequnce (seq : expr_t) (l : expr_t list) : expr_t = (
  match seq with
    (ASMSEQUENCE s) ->
      (ASMSEQUENCE (s @ l))
  | _ -> (assert false))

(* ================================================================ *)

(* INITIALIZERS *)

(* Initializer handler.  It takes remaining initializers and partial
   results of initializers. *)

type initializer_follower_t = initializers_t list -> (expr_t list * num) -> (expr_t list * num)

(* Handles the end of initializers (it is a follower). *)

let initializer_empty_follower (remains : initializers_t list) ((exprs, maxindex) : (expr_t list * num)) : (expr_t list * num) = (
  if (remains != []) then
    (syntaxerror "Initializer too long")
  else
    (exprs, maxindex))

let constant_array_size (e : expr_t option) : num = (
  match e with
    None ->
      (*infinity*) most_positive_unsigned_long_long
  | Some x ->
      let x0 = (fold_constant_integer x) in
      if (constant_integer_p x0) then
	(constant_integer_value false x0)
      else
	most_positive_unsigned_long_long)

(* Makes an assignment for an expression {v} with an initializer
   {inits}.  It returns a pair of an initializer expressions and a
   maximum index (which is valid only for incomplete arrays).
   Returned initializers are empty or consist of {INIT_FILL} and zero
   or more {SETF} expressions, whose the LHS with {VREF}, {SREF}, and
   {PREF} (for arrays). *)

let rec make_initializers (v : expr_t) (tn : typenode_t) (inits : initializers_t) : (expr_t list * num) = (
  (make_initializers_rest true v tn
      initializer_empty_follower [inits] ([], num_MINUS_ONE)))

(* Makes an assignment for an expression {v} with the head of
   initializers {propers}.  Generated initializers are appended to
   {exprs}.  {toplevel} is true at the outermost initializers, and
   false in embedded initializers.  It calls {follower} for handling
   the remaining initializers. *)

and make_initializers_rest (toplevel : bool) (v : expr_t) (tn : typenode_t) (follower : initializer_follower_t) (propers : initializers_t list) ((exprs, maxindex) : (expr_t list * num)) : (expr_t list * num) = (
  let nqtn = (strip_off_pseudo_type tn true) in
  if (gcc_flexible_array_type_p nqtn) then
    (syntaxerror "Initializer to zero-sized array unsupported")
  else if (basic_type_p nqtn) then
    (begin
      match propers with
	[] -> (assert false)
      | (inits :: tl) ->
	  let x = (make_basic_initializer v nqtn inits) in
	  (follower tl ((exprs @ x), maxindex))
    end)
  else
    match nqtn with
      NOTYPE -> (assert false)
    | QUALIFIED_TYPE _ | ELLIPSIS_TYPE | SPECIFR_TYPE _ -> (assert false)
    | TYPEDEF_TYPE _ | STRCDEF_TYPE _ | ENUMDEF_TYPE _ -> (assert false)
    | (POINTER_TO tn1) ->
	(begin
	  match propers with
	    [] -> (assert false)
	  | ((INIT_EXPR e) :: tl) ->
	      let x = [(make_assignment SET v e)] in
	      (follower tl ((exprs @ x), maxindex))
	  | ((INIT_LIST []) :: tl) (*GCC*) ->
	      (syntaxerror "Bad initializer to pointer (list)")
	  | ((INIT_LIST _) :: tl) ->
	      (syntaxerror "Bad initializer to pointer (list)")
	  | ((INIT_DSGN _) :: tl) ->
	      (syntaxerror "Bad initializer to pointer (designator)")
	end)
    | (ARRAY_OF (tn1, optionsize)) ->
	(begin
	  if ((not toplevel) && optionsize == None) then
	    (syntaxerror "Incomplete non-first dimension") ;
	  match propers with
	    [] -> (assert false)
	  | ((INIT_EXPR e) :: tl) ->
	      (begin
		if ((string_type_p false nqtn)
		     && (constant_string_p false e)) then
		  (begin
		    let (x, s) = (make_string_initializer v nqtn e) in
		    let newmaxindex =
		      if (toplevel) then (max_num maxindex s) else maxindex in
		    (follower tl ((exprs @ x), newmaxindex))
		  end)
		else if ((string_type_p true nqtn)
			  && (constant_string_p true e)) then
		  (begin
		    let (x, s) = (make_string_initializer v nqtn e) in
		    let newmaxindex =
		      if (toplevel) then (max_num maxindex s) else maxindex in
		    (follower tl ((exprs @ x), newmaxindex))
		  end)
		else if (toplevel) then
		  (syntaxerror "Bad initializer to array")
		else
		  (begin
		    (* Eliding-braces case. *)
		    (warn_pedantic_in_parse "Eliding braces in initializers") ;
		    let size = (constant_array_size optionsize) in
		    (make_initializers_array toplevel false
			v tn1 size num_ZERO
			[] follower propers (exprs, maxindex))
		  end)
	      end)
	  | ((INIT_LIST []) :: tl) (*GCC*) ->
	      let z = if (toplevel) then [(INIT_FILL (nqtn, v))] else [] in
	      (follower tl ((exprs @ z), maxindex))
	  | ((INIT_LIST ee) :: tl) ->
	      (begin
		let z = if (toplevel) then [(INIT_FILL (nqtn, v))] else [] in
		let size = (constant_array_size optionsize) in
		(make_initializers_array toplevel true
		    v tn1 size num_ZERO
		    tl follower ee ((exprs @ z), maxindex))
	      end)
	  | ((INIT_DSGN _) :: tl) ->
	      (syntaxerror "Bad initializer to array (designator)")
	end)
    | (FUNCTION_OF (tn1, tl)) ->
	(begin
	  match propers with
	    [] -> (assert false)
	  | (inits :: tl) ->
	      let x = (make_basic_initializer v nqtn inits) in
	      (follower tl ((exprs @ x), maxindex))
	end)
    | (BITFIELD_TYPE (tn1, ex)) ->
	(begin
	  match propers with
	    [] -> (assert false)
	  | (inits :: tl) ->
	      let x = (make_basic_initializer v tn1 inits) in
	      (follower tl ((exprs @ x), maxindex))
	end)
    | (STRUCTURE_TYPE _) ->
	(begin
	  let body = (get_structure_body nqtn) in
	  let {
	    strc_utag = _ ;
	    strc_components = wholecomponents ;
	    strc_unnamed_slots = unnamedslots
	  } = body in
	  match propers with
	    [] -> (assert false)
	  | ((INIT_EXPR e) :: tl) ->
	      (begin
		if (check_assignment_conversion_p nqtn (type_of_expr e) e) then
		  (begin
		    (* Structure copy. *)
		    let x = [(make_assignment SET v e)] in
		    (follower tl ((exprs @ x), maxindex))
		  end)
		else
		  (begin
		    (* Eliding-braces case. *)
		    (warn_pedantic_in_parse "Eliding braces in initializers") ;
		    (make_initializers_struct false
			v wholecomponents wholecomponents
			[] follower propers (exprs, maxindex))
		  end)
	      end)
	  | ((INIT_LIST []) :: tl) (*GCC*) ->
	      let z = if (toplevel) then [(INIT_FILL (nqtn, v))] else [] in
	      (follower tl ((exprs @ z), maxindex))
	  | ((INIT_LIST ee) :: tl) ->
	      let z = if (toplevel) then [(INIT_FILL (nqtn, v))] else [] in
	      (make_initializers_struct true v wholecomponents wholecomponents
		  tl follower ee ((exprs @ z), maxindex))
	  | ((INIT_DSGN _) :: tl) ->
	      (syntaxerror "Initializer to structure (designator)")
	end)
    | (ENUMERATION_TYPE _) ->
	(begin
	  match propers with
	    [] -> (assert false)
	  | (inits :: tl) ->
	      let x = (make_basic_initializer v nqtn inits) in
	      (follower tl ((exprs @ x), maxindex))
	end)
    | _ -> (assert false))

(* Makes initializers for an array {v} of an element type {elemtn} with
   elements in {propers}.  {propers} and {remains} are both
   initializers where {remains} is for followers, and concatenating
   them makes a true list.  On eliding-braces case, {remains} is empty
   and {propers} serves also for followers. *)

and make_initializers_array (toplevel : bool) (braced : bool) (v : expr_t) (elemtn : typenode_t) (size : num) (index : num) (remains : initializers_t list) (follower : initializer_follower_t) (propers : initializers_t list) ((exprs, maxindex) : (expr_t list * num)) : (expr_t list * num) = (
  assert (braced || remains == []);
  if (eq_num size index) then
    (begin
      if (not braced) then
	(follower propers (exprs, maxindex))
      else if (propers != []) then
	(syntaxerror "Initializer too long")
      else
	(follower remains (exprs, maxindex))
    end)
  else
    (begin
      match propers with
	[] ->
	  (begin
	    if (not braced) then
	      (follower [] (exprs, maxindex))
	    else
	      (follower remains (exprs, maxindex))
	  end)
      | ((INIT_EXPR _) :: tl)
      | ((INIT_LIST _) :: tl) ->
	  (begin
	    let newmaxindex =
	      (if (toplevel) then (max_num maxindex index) else maxindex) in
	    let newfollower = (make_initializers_array toplevel braced
				  v elemtn size (succ_num index)
				  remains follower) in
	    let ref = (make_aref v (constant_integer index)) in
	    (make_initializers_rest false ref elemtn newfollower
		propers (exprs, newmaxindex))
	  end)
      | ((INIT_DSGN (((DSGN_AREF (_, c0, NIL)) :: el), e)) :: tl) ->
	  (begin
	    let inits2 =
	      (if (el == []) then e else (INIT_LIST [(INIT_DSGN (el, e))])) in
	    let k0 = (fold_constant_integer c0) in
	    if (not (constant_integer_p k0)) then
	      (syntaxerror "Initializer with non-constant designator")
	    else
	      (begin
		let j = (constant_integer_value false k0) in
		let newmaxindex =
		  (if (toplevel) then (max_num maxindex j) else maxindex) in
		let newfollower = (make_initializers_array toplevel braced
				      v elemtn size (succ_num j)
				      remains follower) in
		let ref = (make_aref v c0) in
		(make_initializers_rest false ref elemtn newfollower
		    (inits2 :: tl) (exprs, newmaxindex))
	      end)
	  end)
      | ((INIT_DSGN (((DSGN_AREF (tn, c0, c1)) :: el), e)) :: tl) (*GCC*) ->
	  (begin
	    (* Range designator [lb...ub].  Note ub is inclusive. *)
	    let k0 = (fold_constant_integer c0) in
	    let k1 = (fold_constant_integer c1) in
	    if (not ((constant_integer_p k0) && (constant_integer_p k1))) then
	      (syntaxerror "Initializer with non-constant designator")
	    else
	      (begin
		let j0 = (constant_integer_value false k0) in
		let ub = (constant_integer_value false k1) in
		if (not (le_num j0 ub)) then
		  (syntaxerror "Initializer with bad range designator")
		else
		  (begin
		    let j1 = (constant_integer j0) in
		    let j2 = (constant_integer (succ_num j0)) in
		    let propers2 =
		      if (eq_num j0 ub) then
			((INIT_DSGN (((DSGN_AREF (tn, j1, NIL)) :: el), e))
			  :: tl)
		      else
			((INIT_DSGN (((DSGN_AREF (tn, j1, NIL)) :: el), e))
			  :: (INIT_DSGN (((DSGN_AREF (tn, j2, c1)) :: el), e))
			  :: tl) in
		    (make_initializers_array toplevel braced
			v elemtn size index remains follower
			propers2 (exprs, maxindex))
		  end)
	      end)
	  end)
      | ((INIT_DSGN _) :: _) ->
	  (syntaxerror "Bad designator to array")
    end))

(* Makes initializers for {v} for {components} with elements in
   {propers}.  Note that designators restart scanning components from
   the beginning. *)

and make_initializers_struct (braced : bool) (v : expr_t) (wholecomponents : (string * typenode_t) list) (components : (string * typenode_t) list) (remains : initializers_t list) (follower : initializer_follower_t) (propers : initializers_t list) ((exprs, maxindex) : (expr_t list * num)) : (expr_t list * num) = (
  let deref = false in
  match (components, propers) with
    (_, []) ->
      (follower remains (exprs, maxindex))
  | (((s, tn) :: components1), ((INIT_EXPR e) :: inits1)) ->
      let newfollower = (make_initializers_struct braced v
			    wholecomponents components1 remains follower) in
      let ref = (SREF (tn, deref, v, s)) in
      (make_initializers_rest false ref tn newfollower
	  propers (exprs, maxindex))
  | (((s, tn) :: components1), ((INIT_LIST []) :: inits1)) (*GCC*) ->
      (make_initializers_struct braced v wholecomponents components1
	  remains follower inits1 (exprs, maxindex))
  | (((s, tn) :: components1), ((INIT_LIST ee) :: inits1)) ->
      let newfollower = (make_initializers_struct braced v
			    wholecomponents components1 remains follower) in
      let ref = (SREF (tn, deref, v, s)) in
      (make_initializers_rest false ref tn newfollower
	  propers (exprs, maxindex))
  | (_, ((INIT_DSGN _) :: _)) ->
      (make_initializers_struct_designator braced v
	  wholecomponents wholecomponents
	  remains follower propers (exprs, maxindex))
  | ([], _) ->
      (begin
	if (not braced) then
	  (begin
	    assert (remains == []) ;
	    (follower propers (exprs, maxindex))
	  end)
	else
	  (syntaxerror "Initializer to structure (too many)")
      end))

(* Makes initializers for {v} for {components} with elements in
   {propers}. *)

and make_initializers_struct_designator (braced : bool) (v : expr_t) (wholecomponents : (string * typenode_t) list) (components : (string * typenode_t) list) (remains : initializers_t list) (follower : initializer_follower_t) (propers : initializers_t list) ((exprs, maxindex) : (expr_t list * num)) : (expr_t list * num) = (
  let deref = false in
  match propers with
    ((INIT_DSGN (((SREF (_, _, _, slotname)) :: tld), e)) :: inits1) ->
      (begin
	match components with
	  [] ->
	    (syntaxerror "Initializer to structure (no matching designator)")
	| ((s, tn) :: components1) when (string_equal s slotname) ->
	    (begin
	      if (tld == []) then
		(begin
		  let newfollower = (make_initializers_struct braced v
					wholecomponents components1
					remains follower) in
		  let ref = (SREF (tn, deref, v, s)) in
		  (make_initializers_rest false ref tn newfollower
		      (e :: inits1) (exprs, maxindex))
		end)
	      else
		(begin
		  let newfollower = (make_initializers_struct braced v
					wholecomponents components1
					remains follower) in
		  let ref = (SREF (tn, deref, v, s)) in
		  (make_initializers_rest false ref tn newfollower
		      ((INIT_LIST [(INIT_DSGN (tld, e))]) :: inits1)
		      (exprs, maxindex))
		end)
	    end)
	| (_ :: components1) ->
	    (begin
	      (make_initializers_struct_designator braced v
		  wholecomponents components1
		  remains follower propers (exprs, maxindex))
	    end)
      end)
  | _ -> (assert false))

(* Makes an initializer.  It accepts enumeration type. *)

and make_basic_initializer (v : expr_t) (tn : typenode_t) (inits : initializers_t) : expr_t list = (
  let nqtn = (strip_off_pseudo_type tn true) in
  match nqtn with
    VOID_TYPE ->
      (syntaxerror "Initializer to void type")
  | BOOLEAN
  | CHARACTER
  | SIGNED_CHARACTER
  | UNSIGNED_CHARACTER
  | SHORT_INTEGER
  | UNSIGNED_SHORT_INTEGER
  | INTEGER
  | UNSIGNED_INTEGER
  | LONG_INTEGER
  | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER
  | UNSIGNED_LONG_LONG_INTEGER
  | SINGLE_FLOAT
  | DOUBLE_FLOAT
  | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX
  | DOUBLE_COMPLEX
  | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY
  | DOUBLE_IMAGINARY
  | LONG_DOUBLE_IMAGINARY
  | (ENUMERATION_TYPE _) ->
      (begin
	match inits with
	  (INIT_EXPR e) ->
	    [(make_assignment SET v e)]
	| (INIT_LIST []) (*GCC*) ->
	    (syntaxerror "Bad initializer")
	| (INIT_LIST l) ->
	    (syntaxerror "Bad initializer")
	| (INIT_DSGN (dsgn, il)) ->
	    (syntaxerror "Bad initializer")
      end)
  | _ -> (assert false))

and make_string_initializer (v : expr_t) (tn : typenode_t) (e : expr_t) : (expr_t list * num) = (
  match (tn, e) with
    ((ARRAY_OF (tn1, optionsize)), (STRING (b, s, content))) ->
      (begin
	let len = (num_of_int (Array.length content)) in
	let size = (constant_array_size optionsize) in
	let truesize = (int_of_num (min_num len size)) in
	let truecontent = (Array.sub content 0 truesize) in
	let ex = (Array.mapi
		   (fun i c ->
		     let code = (constant_integer (num_of_int c)) in
		     let index = (constant_integer (num_of_int i)) in
		     (make_assignment SET (make_aref v index) code))
		   truecontent) in
	let x = (Array.to_list ex) in
      	(x, (num_of_int (truesize - 1)))
      end)
  | _ -> (assert false))

(* ================================================================ *)

(* VARIABLE LENGTH ARRAY HANDLING (NON-CONSTANT DIMENSIONS) *)

(* Substitutes an expression (including a variable) appearing in array
   dimensions with a temporary variable to copy a value.  It returns a
   (maybe) new type and a list of assignments of substitutions.  It
   substitutes only program variables, assuming introduced temporaries
   are constants.  It recurses on a structure definition only when it
   is in a defining context.  NOTE: Types with variable length array
   cannot be used in casts.  So, substitutions can be performed in
   declarations (appearing in sizeof is safe).  NOTE: {subst} is
   reversed because array dimensions are reversed as they appear.. *)

let rec substitute_variables_in_array_type (tn : typenode_t) : (typenode_t * (expr_t list)) = (
  let subst : (expr_t * expr_t) Vector.t = (Vector.create 0 (NIL, NIL)) in
  let tnx = (substitute_variables_in_type subst tn) in
  let inits = (List.rev
		  (Vector.map
		      (fun (v, e) -> (make_assignment SET v e))
		      subst)) in
  (tnx, inits))

and substitute_variables_in_type (subst : (expr_t * expr_t) Vector.t) (tn : typenode_t) : typenode_t = (
  match tn with
    NOTYPE ->
      (* No-type specifier. *)
      tn
  | (ATTRIBUTED_TYPE (tn0, a)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(ATTRIBUTED_TYPE (tn1, a))
  | (QUALIFIED_TYPE (tn0, q)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(QUALIFIED_TYPE (tn1, q))
  | ELLIPSIS_TYPE ->
      tn
  | (STRCDEF_TYPE (tn0, body)) ->
      (begin
	let body = (get_structure_body tn0) in
	let {
	  strc_utag = _ ;
	  strc_components = components1 ;
	  strc_unnamed_slots = unnamedslots
	} = body in
	let components2 =
	  (List.map
	      (fun (s, tn1) ->
		(s, (substitute_variables_in_type subst tn1)))
	      components1) in
	let identicalp =
	  (List.for_all2
	      (fun (s1, t1) (s2, t2) ->
		((string_equal s1 s2) && (t1 == t2)))
	      components1 components2) in
	if (not (identicalp)) then
	  (body.strc_components <- components2) ;
	tn
      end)
  | (ENUMDEF_TYPE (tn0, l)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(ENUMDEF_TYPE (tn1, l))
  | (TYPEDEF_TYPE (tn0, l)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(TYPEDEF_TYPE (tn1, l))
  | SPECIFR_TYPE _ -> (assert false)
  | VOID_TYPE
  | BOOLEAN | CHARACTER | SIGNED_CHARACTER | UNSIGNED_CHARACTER
  | SHORT_INTEGER | UNSIGNED_SHORT_INTEGER
  | INTEGER | UNSIGNED_INTEGER | LONG_INTEGER | UNSIGNED_LONG_INTEGER
  | LONG_LONG_INTEGER | UNSIGNED_LONG_LONG_INTEGER
  | SINGLE_FLOAT | DOUBLE_FLOAT | LONG_DOUBLE_FLOAT
  | SINGLE_COMPLEX | DOUBLE_COMPLEX | LONG_DOUBLE_COMPLEX
  | SINGLE_IMAGINARY | DOUBLE_IMAGINARY | LONG_DOUBLE_IMAGINARY ->
      tn
  | (POINTER_TO tn0) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(POINTER_TO tn1)
  | (ARRAY_OF (tn0, None)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(ARRAY_OF (tn1, None))
  | (ARRAY_OF (tn0, Some e0)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      let e2 = (substitute_variables_in_type_expr subst e0) in
      if ((tn0 == tn1) && (e0 == e2)) then
	tn
      else
	(ARRAY_OF (tn1, Some e2))
  | (FUNCTION_OF (tn0, None)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      if (tn0 == tn1) then
	tn
      else
	(FUNCTION_OF (tn1, None))
  | (FUNCTION_OF (tn0, Some tt0)) ->
      let tn1 = (substitute_variables_in_type subst tn0) in
      let tt2 = (List.map (substitute_variables_in_type subst) tt0) in
      if ((tn0 == tn1) && (List.for_all2 (==) tt0 tt2)) then
	tn
      else
	(FUNCTION_OF (tn1, Some tt2))
  | (BITFIELD_TYPE _) ->
      tn
  | (STRUCTURE_TYPE _) ->
      tn
  | (ENUMERATION_TYPE _) ->
      tn
  | PSEUDO_VA_LIST_TYPE ->
      type_of_va_list
  | GCC_VA_ARG ->
      tn)

(* Substitutes variables appearing in an expression (not temporary,
   but program variables) with new temporary variables. *)

and substitute_variables_in_type_expr (subst : (expr_t * expr_t) Vector.t) (e : expr_t) : expr_t = (
  match e with
    NIL ->
      e
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      e
  | (OPER (tn, op, e0, e1, e2)) ->
      let x0 = (substitute_variables_in_type_expr subst e0) in
      let x1 = (substitute_variables_in_type_expr subst e1) in
      let x2 = (substitute_variables_in_type_expr subst e2) in
      if ((e0 == x0) && (e1 == x1) && (e2 == x2)) then
	e
      else
	(OPER (tn, op, x0, x1, x2))
  | (SETF (tn, op, e0, e1)) ->
      let x0 = (substitute_variables_in_type_expr subst e0) in
      let x1 = (substitute_variables_in_type_expr subst e1) in
      if ((e0 == x0) && (e1 == x1)) then
	e
      else
	(SETF (tn, op, x0, x1))
  | (VREF (tn, id, level, _)) ->
      if (temporary_variable_p e) then
	e
      else if (constant_expression_p e) then
	e
      else
	(begin
	  let substitution =
	    (Vector.find_if
		(fun (tmpvar, oldvar) ->
		  match oldvar with
		    (VREF (tn, id1, level1, _)) ->
		      ((string_equal id id1) && level == level1)
		  | _ -> false)
		subst) in
	  if (substitution != None) then
	    let (tmpvar, oldvar) = (Misc.some_of substitution) in
	    tmpvar
	  else
	    (begin
	      let tmpvar = (make_new_temporary_variable_for e) in
	      (Vector.add subst (tmpvar, e)) ;
	      tmpvar
	    end)
	end)
  | (PREF (tn, _, _, _)) ->
      let tmpvar = (make_new_temporary_variable tn) in
      (Vector.add subst (tmpvar, e)) ;
      tmpvar
  | (SREF (tn, _, _, _)) ->
      let tmpvar = (make_new_temporary_variable tn) in
      (Vector.add subst (tmpvar, e)) ;
      tmpvar
  | (DSGN_AREF _) -> (assert false)
  | (CALL (tn, _, _, _)) ->
      let tmpvar = (make_new_temporary_variable tn) in
      (Vector.add subst (tmpvar, e)) ;
      tmpvar
  | (ADDROF (tn, _)) ->
      let tmpvar = (make_new_temporary_variable tn) in
      (Vector.add subst (tmpvar, e)) ;
      tmpvar
  | (SIZEOF (tn, tn1)) ->
      let sizeexpr = (size_of_type tn1) in
      (substitute_variables_in_type_expr subst sizeexpr)
  | (PROGN (tn, l)) ->
      let x = (List.map (substitute_variables_in_type_stmt subst) l) in
      (PROGN (tn, x))
  | (CONST_I _) | (CONST_F _) | (STRING _) ->
      e
  | (CLITERAL (tn, l)) ->
      (syntaxerror "Bad compound literal in type")
  | (ASMSEQUENCE _) -> (assert false)
  | (PHI _) -> (assert false)
  | (PSEUDO_MEMORY _) -> (assert false)
  | (INIT_FILL _) -> (assert false)
  | (PRAGMA _) | (QUANTIFIER _) -> (assert false)
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      PSEUDO_ACSL_RESULT (tn, x0))
  | PSEUDO_ACSL_OLD (tn, e0, e1) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      let x1 = substitute_variables_in_type_expr subst e1 in
      PSEUDO_ACSL_OLD (tn, x0, x1))
  | ALL_FIELD (tn, e0) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      ALL_FIELD (tn, x0)
    )
  | PSEUDO_ACSL_UNIQUE (tn, e0) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      PSEUDO_ACSL_UNIQUE (tn, x0))
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      let x1 = substitute_variables_in_type_expr subst e1 in
      PSEUDO_ACSL_UNIQUEBLOCK (tn, x0, x1))
  | PSEUDO_ACSL_VALID (tn, e0) -> (
      let x0 = substitute_variables_in_type_expr subst e0 in
      PSEUDO_ACSL_VALID (tn, x0))
)

and substitute_variables_in_type_stmt (subst : (expr_t * expr_t) Vector.t) (s : stmt_t) : stmt_t = (
  match s with
    (DEFUN _) | (SPEC _) | (ASSUME _) | (RESET _)-> (assert false)
  | (DEFVAR d) ->
      (begin
	match d with
	  {
	    decl_id = id ;
	    decl_tn = _ ;
	    decl_scls = _ ;
	    decl_inits = inits ;
	    decl_params = _ ;
	    decl_struct_defined_p = _
	  } ->
	    if (inits != []) then
	      let v = (List.map (substitute_variables_in_type_expr subst)
			  inits) in
	      (DEFVAR { d with decl_inits = v })
	    else
	      s
      end)
  | (STMT e) ->
      let x = (substitute_variables_in_type_expr subst e) in
      (STMT x)
  | (ASGN (e0, e1)) -> (assert false)
  | (CALLASGN _) -> (assert false)
  | (IF (e, u0, u1)) ->
      let x = (substitute_variables_in_type_expr subst e) in
      let v0 = (List.map (substitute_variables_in_type_stmt subst) u0) in
      let v1 = (List.map (substitute_variables_in_type_stmt subst) u1) in
      (IF (x, v0, v1))
  | (BEGIN u) ->
      let v = (List.map (substitute_variables_in_type_stmt subst) u) in
      (BEGIN v)
  | (SWITCH (e, u)) ->
      let x = (substitute_variables_in_type_expr subst e) in
      let v = (List.map (substitute_variables_in_type_stmt subst) u) in
      (SWITCH (x, v))
  | (DOWHILE (b, e, u)) ->
      let x = (substitute_variables_in_type_expr subst e) in
      let v = (List.map (substitute_variables_in_type_stmt subst) u) in
      (DOWHILE (b, x, v))
  | (FOR (u0, e1, e2, u3)) ->
      let v0 = (List.map (substitute_variables_in_type_stmt subst) u0) in
      let x1 = (substitute_variables_in_type_expr subst e1) in
      let x2 = (substitute_variables_in_type_expr subst e2) in
      let v3 = (List.map (substitute_variables_in_type_stmt subst) u3) in
      (FOR (v0, x1, x2, v3))
  | BREAK ->
      s
  | (RETURN None) ->
      s
  | (RETURN (Some e)) ->
      let x = (substitute_variables_in_type_expr subst e) in
      (RETURN (Some x))
  | (GOTO _) ->
      s
  | CONTINUE ->
      s
  | (LABEL _) ->
      s
  | (ASM _) ->
      s
  | (NOTE _) ->
      s
  | (ANNOTATION (t, e)) ->
      let x = (substitute_variables_in_type_expr subst e) in
      (ANNOTATION (t, x)))

(* ================================================================ *)

(* Makes a declaration of {declaration_t} by combining specifiers and
   declarators.  It substitutes variables in types with
   temporaries. *)

let rec combine_declaration (explicittypep : bool) (specifiers : specifier_t) (declarator : declarator_t) : declaration_t = (
  let (basetn0, scls, quals, attrs0) = specifiers in
  let (decltn, varid, params, attrs1) = declarator in
  (begin
    assert (explicittypep || basetn0 == NOTYPE) ;
    if (!cgTRACE) then
      (debug
	  (";; combine_declaration: " ^ (varid_to_string varid)
	    ^ "(" ^ (get_current_line_info_string ()) ^ ").")) ;
    let basetn = if (explicittypep) then basetn0 else INTEGER in
    let qualtn = (add_type_qualifiers basetn quals) in
    let tn = (substitute_basetype_of decltn qualtn) in
    let abstractdeclarationp = (string_equal (fst varid) "") in
    if (function_type_p decltn) then
      (begin
	(* It is a function, if parameters are given. *)
	let attrtn = (make_attributed_type tn (attrs0 @ attrs1)) in
	let decl = {
	  decl_id = varid ;
	  decl_tn = attrtn ;
	  decl_scls = scls ;
	  decl_inits = [] ;
	  decl_params = params ;
	  decl_struct_defined_p = false ;
	  decl_specs = None;
	  decl_location = !location_of_cur_fun;
	} in
	(apply_attributes decl (attrs0 @ attrs1))
      end)
    else if (scls == TYPEDEF) then
      (begin
	(* It is a typedef. *)
	let (tx, typeinits) =
	  if (not abstractdeclarationp) then
	    (substitute_variables_in_array_type tn)
	  else
	    (tn, []) in
	let attrtn = (make_attributed_type tx (attrs0 @ attrs1)) in
	let decl = {
	  decl_id = varid ;
	  decl_tn = attrtn ;
	  decl_scls = scls ;
	  decl_inits = typeinits ;
	  decl_params = params ;
	  decl_struct_defined_p = false ;
	  decl_specs = None;
	  decl_location = !location_of_cur_fun;
	} in
	(apply_attributes decl (attrs0 @ attrs1))
      end)
    else
      (begin
	(* It is a variable (without initializer). *)
	let (tx, typeinits) =
	  if (not abstractdeclarationp) then
	    (substitute_variables_in_array_type tn)
	  else
	    (tn, []) in
	let attrtn = (make_attributed_type tx (attrs0 @ attrs1)) in
	let decl = {
	  decl_id = varid ;
	  decl_tn = attrtn ;
	  decl_scls = scls ;
	  decl_inits = typeinits ;
	  decl_params = params ;
	  decl_struct_defined_p = false ;
	  decl_specs = None;
	  decl_location = !location_of_cur_fun;
	} in
	(apply_attributes decl (attrs0 @ attrs1))
      end)
  end))

(* Applies handlers of attributes to declarations. *)

and apply_attributes (decl : declaration_t) (attrs : gcc_attribute_t list) : declaration_t = (
  (List.fold_left
      (fun (decl : declaration_t) (attr : gcc_attribute_t) ->
	match attr with
	  ATTR ((name : string), (expr : expr_t)) ->
	    if (not (Hashtbl.mem attribute_handlers name)) then
	      (begin
		(warn_pedantic_in_parse
		    ("Attribute ignored: '" ^ name ^ "'")) ;
		decl
	      end)
	    else
	      (begin
		let h = (Hashtbl.find attribute_handlers name) in
		(h decl attr)
	      end))
      decl attrs))

(* Merge initializers to the declaration.  Note that it makes
   initializers twice for an incomplete array, once to find the size
   and once to make initializers. *)

let rec merge_inits_to_declaration (declaration : declaration_t) (inits : initializers_t) : declaration_t = (
  match declaration with
    {
      decl_id = (id, level) ;
      decl_tn = tn ;
      decl_scls = scls ;
      decl_inits = typeinits ;
      decl_params = params ;
      decl_struct_defined_p = structdefinedp;
      decl_specs = specs_op;
      decl_location = location;
    } ->
      (begin
	assert ((*params == [] && *)structdefinedp == false) ;
        assert (specs_op == None) ;
	let (initexpr, newattrtn, typefixed) =
	  if (not (incomplete_array_type_p tn)) then
	    (begin
	      let v = (VREF (tn, id, level, 0)) in
	      let (exprs, _) = (make_initializers v tn inits) in
	      (exprs, tn, false)
	    end)
	  else
	    (begin
	      let (noattrtn, attrs) = (strip_off_attributes tn) in
	      let dummy = (VREF (tn, id, level, 0)) in
	      let (_, maxindex) = (make_initializers dummy tn inits) in
	      let fixtn = (make_attributed_type
			      (adjust_incomplete_array_size noattrtn
				  (succ_num maxindex))
			      attrs) in
	      let v = (VREF (fixtn, id, level, 0)) in
	      let (exprs, _) = (make_initializers v fixtn inits) in
	      (exprs, fixtn, true)
	    end) in
	let decl = {
	  decl_id = (id, level) ;
	  decl_tn = newattrtn ;
	  decl_scls = scls ;
	  decl_inits = (typeinits @ initexpr) ;
	  decl_params = params ;
	  decl_struct_defined_p = false ;
	  decl_specs = None;
	  decl_location = location;
	} in
	if (typefixed) then
	  (define_identifier_in_scope decl)
	else
	  decl
      end))

and adjust_incomplete_array_size (tn : typenode_t) (size : num) : typenode_t = (
  let (nqtn, quals) = (strip_off_pseudo_type_return_qualifiers tn) in
  match nqtn with
    (ARRAY_OF (elemtn, None)) ->
      let c = (constant_integer size) in
      (make_qualified_type (ARRAY_OF (elemtn, (Some c))) quals)
  | _ -> (assert false))

(* Makes a compound literal form.  It is treated specially for
   (char[]){"abc"} because the brace needs be stripped off. *)

let make_compound_literal (tn : typenode_t) (i : initializers_t) : expr_t = (
  match i with
    (INIT_EXPR _) | (INIT_DSGN _) -> (assert false)
  | (INIT_LIST []) (*GCC*) ->
      (syntaxerror "Compound literal empty")
  | (INIT_LIST l) ->
      (begin
	let tn0 = (strip_off_pseudo_type tn true) in
	match l with
	  [(INIT_EXPR e)] ->
	    if ((string_type_p false tn0)
		 && (constant_string_p false e)) then
	      (CLITERAL (tn, (INIT_EXPR e)))
	    else if ((string_type_p true tn0)
		      && (constant_string_p true e)) then
	      (CLITERAL (tn, (INIT_EXPR e)))
	    else
	      (CLITERAL (tn, i))

	| _ ->
	    (CLITERAL (tn, i))
      end))

(* Makes a statement expression (with {PROGN}) from a {BEGIN} block.
   (GCC) *)

let rec make_statement_expression (s : stmt_t) : expr_t = (
  match s with
    (BEGIN l) ->
      (begin
	let x = (last_effective_expression l) in
	match x with
	  (STMT NIL) ->
	    (* No value generating expression. *)
	    (PROGN (NOTYPE, l))
	| (STMT e) ->
	    let tn = (type_of_expr e) in
	    (PROGN (tn, l))
	| _ ->
	    (assert false)
      end)
  | _ -> (assert false))

(* ================================================================ *)

let make_array_type (tn : typenode_t) (dimension : expr_t option) : typenode_t = (
  match dimension with
    None ->
      (ARRAY_OF (tn, dimension))
  | (Some NIL) ->
      (ARRAY_OF (tn, dimension))
  | (Some e) ->
      let v = (fold_constant_integer e) in
      if (not (constant_integer_p v)) then
	(begin
	  (warn_in_parse "Variable array dimension") ;
	  (ARRAY_OF (tn, dimension))
	end)
      else
	(begin
	  (ARRAY_OF (tn, (Some v)))
	end))

(* Makes an identifier reference node.  It does not check if the given
   identifier is declared or not, since there is a possibility of
   implicit function declaration.  The return value is of (VREF
   (type,id,scope-level,sequence)). (VREF (NOTYPE,id,-1,0)) is for
   undeclared case. *)

let make_variable_reference (id : string) : expr_t = (
  let b = (get_binding id !name_scope_stack false) in
  if (string_equal id "$nondet") then
    (PSEUDO_NONDET_VALUE (INTEGER, "integer"))
  else if (string_equal id "$nondet_positive") then
    (PSEUDO_NONDET_VALUE (INTEGER, "positive"))
  else if (string_equal id "$nondet_zero_one") then
    (PSEUDO_NONDET_VALUE (INTEGER, "zero_one"))
  else if ((binding_exists_p b) && (b.binding_scls == TYPEDEF)) then
    (syntaxerror ("Referencing to typedef name as variable: " ^ id))
  else if (binding_exists_p b) then
    (VREF ((b.binding_tn), (fst b.binding_name), (snd b.binding_name), 0))
  else if (Hashtbl.mem gcc_builtin_identifiers id) then
    (begin
      (* Builtin functions. *)
      let tn = (Hashtbl.find gcc_builtin_identifiers id) in
      if (tn == NOTYPE) then
	(syntaxerror ("Builtin '" ^ id ^ "' undefined"))
      else
	(VREF (tn, id, global_scope_level, 0))
    end)
  else if (Hashtbl.mem pseudo_builtin_identifiers id) then
    (begin
      (* Builtin variables/functions. *)
      let tn = (Hashtbl.find pseudo_builtin_identifiers id) in
      (VREF (tn, id, global_scope_level, 0))
    end)
  else
    (begin
      (* Implicit functions. *)
      (VREF (NOTYPE, id, implicit_function_scope_level, 0))
    end))

(* Checks if the argument is an undeclared identifier reference.  It
   returns an expression or errs.  When {declareherep} is true, and
   the identifier is a name of an undeclared function, it is declared
   here as function type.  IT IS NOT ADDED IN THE SCOPE AND FORCED A
   SCOPE-LEVEL TO GLOBAL. *)

let check_undeclared_reference (e : expr_t) (declareherep : bool) : expr_t = (
  if (!in_attribute_scanning_p) then
    (* No check, while reading GCC attributes. *)
    e
  else
    (begin
      match e with
	(VREF (tn, id, level, _)) ->
	  if (tn == NOTYPE) then
	    (begin
	      (* Null type means no binding. *)
	      if (not declareherep) then
		(syntaxerror ("Referencing undefined identifier: " ^ id))
	      else
		(begin
                  (warn_in_parse ("Referencing undefined identifier: " ^ id)) ;
		  let x = (FUNCTION_OF (INTEGER, None)) in
		  (VREF (x, id, global_scope_level, 0))
		end)
	    end)
	  else
	    e
      | _ ->
	  e
    end))

let make_linenote () : stmt_t = (
  (NOTE (LINE (true, !current_file, !current_line, !current_trueline))))

let make_quantified_expression (quantifier : string) (vs : stmt_t list) (expr : expr_t) = (
  let vars =
    (List.map
	(fun (v) ->
	  match v with
	    (DEFVAR (decl : declaration_t)) ->
	      (VREF (NOTYPE, (fst decl.decl_id), (snd decl.decl_id), 0))
	  | _ ->
	      (assert false))
	vs) in
  (QUANTIFIER (quantifier, vars, expr)))

let make_annotation (operator : string) (expr : expr_t) : stmt_t = (
  (ANNOTATION (operator, expr)))

let rec make_function_call (expr : expr_t) (args : expr_t list) : expr_t = (
  (* Check if it is a builtin function call. *)
  match expr with
    (VREF (tn, id, level, sequence)) ->
      if (Hashtbl.mem gcc_builtin_identifiers id) then
	(make_builtin_function_call id expr args)
      else
	(make_function_call_expr expr args)
  | _ ->
      (make_function_call_expr expr args))

and make_function_call_expr (expr : expr_t) (args : expr_t list) : expr_t = (
  let func = (make_usual_unary_conversion expr) in
  let fptn = (type_of_expr func) in
  if (not (function_pointer_type_p fptn)) then
    (syntaxerror "Funcall to non-function")
  else
    (begin
      (* Check if the arguments are proper for this funcall, and
	 convert them if necessary. *)
      let fntn = (deref_pointer_type fptn) in
      match fntn with
	(FUNCTION_OF (rettn, None)) ->
	  (* Traditional parameter case. *)
	  let eargs = (List.map make_usual_argument_conversion args) in
	  (CALL (rettn, func, eargs, None))
      | (FUNCTION_OF (rettn, Some params)) ->
	  (* Prototype case. *)
	  let eargs = (make_arguments params args 0) in
	  (CALL (rettn, func, eargs, None))
      | _ ->
	  (syntaxerror "Non-arithmetic type operand")
    end))

and make_builtin_function_call (id : string) (expr : expr_t) (args : expr_t list) : expr_t = (
  let tn = (Hashtbl.find gcc_builtin_identifiers id) in
  if (tn == NOTYPE) then
    (syntaxerror ("Builtin '" ^ id ^ "' undefined"))
  else
    (begin
      match id with
	"__builtin_constant_p" ->
	  (begin
	    (*AHO*) (* Check if folding constants is needed. *)
	    if (constant_expression_p (List.hd args)) then
	      (CONST_I (INTEGER, num_ONE, "1"))
	    else
	      (CONST_I (INTEGER, num_ZERO, "0"))
	  end)
      | "__builtin_return_address" ->
	  (warn_pedantic_in_parse ("Builtin '" ^ id ^ "' as funcall")) ;
	  (make_function_call_expr expr args)
      | "__builtin_trap" ->
	  (warn_pedantic_in_parse ("Builtin '" ^ id ^ "' as funcall")) ;
	  (make_function_call_expr expr args)
      | "__builtin_expect" ->
	  (begin
	    (warn_pedantic_in_parse ("Builtin '" ^ id ^ "' ignored")) ;
	    match args with
	      [(OPER (_, LOGNOT, (OPER (_, LOGNOT, e0, _, _)), _, _)) ; e1] ->
		e0
	    | [e0 ; e1] ->
		e0
	    | _ ->
		(syntaxerror "Bad arguments for '__builtin_expect'")
	  end)
      | "__builtin_prefetch" ->
	  (warn_pedantic_in_parse ("Builtin '" ^ id ^ "' ignored")) ;
	  (CONST_I (INTEGER, num_ZERO, "0"))
      | _ ->
	  (warn_pedantic_in_parse ("Builtin '" ^ id ^ "' as funcall")) ;
	  (make_function_call_expr expr args)
    end))

let builtin_types_compatible_p (u0 : specifier_t) (u1 : specifier_t) : expr_t = (
  let (tn0, _, _, _) = u0 in
  let (tn1, _, _, _) = u1 in
  if (compatible_type_p false tn0 tn1) then
    (CONST_I (INTEGER, num_ONE, "1"))
  else
    (CONST_I (INTEGER, num_ZERO, "0")))

(* ================================================================ *)

(* Utilities Not Used in the Parser (ie, inessential) *)

(* Checks if an expr list may have an effect.  NOTE: {PROGN} is
   assumed to have an effect (ie, it is imprecise).  *)

let rec check_has_effect_or_call_p (l : expr_t list) : bool = (
  match l with
    [] ->
      false
  | (e :: tl) ->
      (begin
	match e with
	  NIL ->
	    (check_has_effect_or_call_p tl)
	| (PSEUDO_NONDET_VALUE (tn, _)) ->
	    (check_has_effect_or_call_p tl)
	| (OPER (tn, op, e0, e1, e2)) ->
	    (check_has_effect_or_call_p ([e0 ; e1 ; e2] @ tl))
	| (SETF (tn, op, e0, e1)) ->
	    true
	| (VREF (tn, s, l, _)) ->
	    (check_has_effect_or_call_p tl)
	| (PREF (tn, _, e0, _)) ->
	    (check_has_effect_or_call_p ([e0] @ tl))
	| (SREF (tn, _, e0, s)) ->
	    (check_has_effect_or_call_p ([e0] @ tl))
	| (DSGN_AREF _) ->  (assert false)
	| (CALL (tn, e0, ee, _)) ->
	    true
	| (ADDROF (tn, e0)) ->
	    (check_has_effect_or_call_p ([e0] @ tl))
	| (SIZEOF (tn, tt)) ->
	    (check_has_effect_or_call_p tl)
	| (PROGN (tn, l)) ->
	    true
	| (CONST_I _) | (CONST_F _) ->
	    (check_has_effect_or_call_p tl)
	| (STRING _) ->
	    (check_has_effect_or_call_p tl)
	| (CLITERAL (tn, l1)) ->
	    if (check_has_effect_or_call_in_initializer_p [l1]) then
	      true
	    else
	      (check_has_effect_or_call_p tl)
	| (ASMSEQUENCE l) ->
	    (check_has_effect_or_call_p (l @ tl))
	| (PHI _) -> (assert false)
	| (PSEUDO_MEMORY _) -> (assert false)
	| (INIT_FILL _) -> (assert false)
	| (PRAGMA s) ->
	    (check_has_effect_or_call_p tl)
	| (QUANTIFIER (s, vs, e0)) ->
	    (check_has_effect_or_call_p ([e0] @ tl))
	| PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
	| PSEUDO_ACSL_RESULT (tn, e0) ->
	    (check_has_effect_or_call_p (e0 :: tl))
	| ALL_FIELD (_, e0) ->
	    (check_has_effect_or_call_p (e0 :: tl))
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1)
	| PSEUDO_ACSL_OLD (tn, e0, e1) ->
	    (check_has_effect_or_call_p (e0 :: e1 :: tl))
        | PSEUDO_ACSL_UNIQUE (tn, e0) ->
	    (check_has_effect_or_call_p (e0 :: tl))
        | PSEUDO_ACSL_VALID (tn, e0) ->
	    (check_has_effect_or_call_p (e0 :: tl))
      end))

and check_has_effect_or_call_in_initializer_p (l : initializers_t list) : bool = (
  match l with
    [] -> false
  | (ie :: tl) ->
      (begin
	match ie with
	  (INIT_EXPR e0) ->
	    if (check_has_effect_or_call_p [e0]) then
	      false
	    else
	      (check_has_effect_or_call_in_initializer_p tl)
	| (INIT_LIST []) (*GCC*) ->
	    (check_has_effect_or_call_in_initializer_p tl)
	| (INIT_LIST il) ->
	    (check_has_effect_or_call_in_initializer_p (il @ tl))
	| (INIT_DSGN (dsgn, ie2)) ->
	    (check_has_effect_or_call_in_initializer_p ([ie2] @ tl))
      end))

(* Checks if an expression list have a statement expression (that is
   {PROGN}). *)

let rec check_has_stmt_expr_p (l : expr_t list) : bool = (
  match l with
    [] ->
      false
  | (e :: tl) ->
      (begin
	match e with
	  NIL ->
	    (check_has_stmt_expr_p tl)
	| (PSEUDO_NONDET_VALUE (tn, _)) ->
	    (check_has_stmt_expr_p tl)
	| (OPER (tn, op, e0, e1, e2)) ->
	    (check_has_stmt_expr_p ([e0 ; e1 ; e2] @ tl))
	| (SETF (tn, op, e0, e1)) ->
	    (check_has_stmt_expr_p ([e0 ; e1] @ tl))
	| (VREF (tn, s, l, _)) ->
	    (check_has_stmt_expr_p tl)
	| (PREF (tn, _, e0, _)) ->
	    (check_has_stmt_expr_p ([e0] @ tl))
	| (SREF (tn, _, e0, s)) ->
	    (check_has_stmt_expr_p ([e0] @ tl))
	| (DSGN_AREF _) ->  (assert false)
	| (CALL (tn, e0, ee, _)) ->
	    (check_has_stmt_expr_p ([e0] @ ee @ tl))
	| (ADDROF (tn, e0)) ->
	    (check_has_stmt_expr_p ([e0] @ tl))
	| (SIZEOF (tn, tt)) ->
	    (check_has_stmt_expr_p tl)
	| (PROGN (tn, l)) ->
	    true
	| (CONST_I _) | (CONST_F _) ->
	    (check_has_stmt_expr_p tl)
	| (STRING _) ->
	    (check_has_stmt_expr_p tl)
	| (CLITERAL (tn, l1)) ->
	    if (check_has_stmt_expr_in_initializer_p [l1]) then
	      true
	    else
	      (check_has_stmt_expr_p tl)
	| (ASMSEQUENCE l) ->
	    (check_has_stmt_expr_p (l @ tl))
	| (PHI _) -> (assert false)
	| (PSEUDO_MEMORY _) -> (assert false)
	| (INIT_FILL _) -> (assert false)
	| (PRAGMA s) ->
	    (check_has_stmt_expr_p tl)
	| (QUANTIFIER (s, vs, e0)) ->
	    (check_has_stmt_expr_p ([e0] @ tl))
	| PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
	| PSEUDO_ACSL_RESULT (tn, e0) ->
	    (check_has_stmt_expr_p (e0 :: tl))
	| ALL_FIELD (_, e0) ->
	    (check_has_stmt_expr_p (e0 :: tl))
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1)
	| PSEUDO_ACSL_OLD (tn, e0, e1) ->
	    (check_has_stmt_expr_p (e0 :: e1 :: tl))
        | PSEUDO_ACSL_UNIQUE (tn, e0) ->
	    (check_has_stmt_expr_p (e0 :: tl))
        | PSEUDO_ACSL_VALID (tn, e0) ->
	    (check_has_stmt_expr_p (e0 :: tl))
      end))

and check_has_stmt_expr_in_initializer_p (l : initializers_t list) : bool = (
  match l with
    [] -> false
  | (ie :: tl) ->
      (begin
	match ie with
	  (INIT_EXPR e0) ->
	    if (check_has_stmt_expr_p [e0]) then
	      false
	    else
	      (check_has_stmt_expr_in_initializer_p tl)
	| (INIT_LIST []) (*GCC*) ->
	    (check_has_stmt_expr_in_initializer_p tl)
	| (INIT_LIST il) ->
	    (check_has_stmt_expr_in_initializer_p (il @ tl))
	| (INIT_DSGN (dsgn, ie2)) ->
	    (check_has_stmt_expr_in_initializer_p ([ie2] @ tl))
      end))

(* Checks if an expression is a logical/relational operation. *)

let rec logical_value_p (e : expr_t) : bool = (
  match e with
    NIL -> (assert false)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      false
  | (OPER (tn, COND, e0, e1, e2)) ->
      ((logical_value_p e1) && (logical_value_p e2))
  | (OPER (tn, (OROR|ANDAND), e0, e1, NIL)) ->
      ((logical_value_p e0) && (logical_value_p e1))
  | (OPER (tn, (CAST b), e0, NIL, NIL)) ->
      (logical_value_p e0)
  | (OPER (tn, COMMA, e0, e1, NIL)) ->
      (logical_value_p e1)
  | (OPER (tn, LOGNOT, e0, NIL, NIL)) ->
      true
  | (OPER (tn, (NEG|BITNOT), e0, NIL, NIL)) ->
      false
  | (OPER (tn, (ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR), e0, e1, NIL)) ->
      false
  | (OPER (tn, (EQ|NE|GE|GT|LE|LT), e0, e1, NIL)) ->
      true
  | (OPER _) -> (assert false)
  | (SETF (tn, SET, e0, e1)) ->
      false
  | (SETF (tn, (INCR_PRE|DECR_PRE|INCR_POST|DECR_POST), e0, NIL)) ->
      false
  | (SETF (tn, (ADD|SUB|MUL|DIV|MOD|LSH|RSH|AND|IOR|XOR), e0, e1)) ->
      false
  | (SETF _) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      false
  | (PREF (tn, _, e0, _)) ->
      false
  | (DSGN_AREF _) ->  (assert false)
  | (SREF (tn, _, e0, s)) ->
      false
  | (CALL (tn, e0, ee, _)) ->
      false
  | (ADDROF (tn, e0)) ->
      false
  | (SIZEOF (tn, tt)) ->
      false
  | (PROGN (tn, l)) -> (assert false)
  | (CONST_I (tn, v, s)) ->
      false
  | (CONST_F (tn, v, s)) ->
      false
  | (STRING (b, s, _)) ->
      false
  | (CLITERAL (tn, l)) ->
      false
  | (ASMSEQUENCE ee) ->
      false
  | (PHI _) -> (assert false)
  | (PRAGMA s) ->
      false
  | (QUANTIFIER (s, vs, e0)) ->
      false
  | (PSEUDO_MEMORY _) -> (assert false)
  | (INIT_FILL _) -> (assert false)
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) ->
      logical_value_p e0
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1)
  | PSEUDO_ACSL_OLD (tn, e0, e1) ->
      (logical_value_p e0) && (logical_value_p e1)
  | ALL_FIELD (_, e0) ->
      logical_value_p e0
  | PSEUDO_ACSL_UNIQUE (tn, e0) ->
      logical_value_p e0
  | PSEUDO_ACSL_VALID (tn, e0) ->
      logical_value_p e0)

let rec substitute_variables_in_expr (varmap : ((string * int * int), (string * int * int)) Hashtbl.t) (e : expr_t) : expr_t = (
  match e with
    NIL -> (assert false)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      e
  | (OPER (tn, op, e0, e1, e2)) ->
      (begin
	match op with
	  SET -> (assert false)
	| INCR_PRE | DECR_PRE -> (assert false)
	| INCR_POST | DECR_POST -> (assert false)
	| ADD | SUB | MUL | DIV | MOD
	| LSH | RSH | AND | IOR | XOR ->
	    assert (e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    let x1 = (substitute_variables_in_expr varmap e1) in
	    (OPER (tn, op, x0, x1, NIL))
	| (CAST _) ->
	    assert (e1 == NIL && e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    (OPER (tn, op, x0, NIL, NIL))
	| NEG | BITNOT ->
	    assert (e1 == NIL && e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    (OPER (tn, op, x0, NIL, NIL))
	| LOGNOT ->
	    assert (e1 == NIL && e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    (OPER (tn, op, x0, NIL, NIL))
	| EQ | NE | GE | GT | LE | LT ->
	    assert (e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    let x1 = (substitute_variables_in_expr varmap e1) in
	    (OPER (tn, op, x0, x1, NIL))
	| COND ->
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    let x1 = (substitute_variables_in_expr varmap e1) in
	    let x2 = (substitute_variables_in_expr varmap e2) in
	    (OPER (tn, op, x0, x1, x2))
	| OROR | ANDAND
	| COMMA
	| IMP | IFF ->
	    assert (e2 == NIL) ;
	    let x0 = (substitute_variables_in_expr varmap e0) in
	    let x1 = (substitute_variables_in_expr varmap e1) in
	    (OPER (tn, op, x0, x1, NIL))
	| PSEUDO_AMP | PSEUDO_STR -> (assert false)
	| PSEUDO_PLS | PSEUDO_MNS -> (assert false)
	| PSEUDO_TLD | PSEUDO_BAN -> (assert false)
      end)
  | (SETF _) -> (assert false)
  | (CALL (tn, e0, ee, _)) -> (assert false)
  | (VREF (tn, id, level, sequence)) ->
      let (i, l, s) = (Hashtbl.find varmap (id, level, sequence)) in
      (VREF (tn, i, l, s))
  | (PREF (tn, deref, e0, e1)) ->
      let x0 = (substitute_variables_in_expr varmap e0)
      and x1 = (substitute_variables_in_expr varmap e1)
      in
      (PREF (tn, deref, x0, x1))
  | (SREF (tn, deref, e0, s)) ->
      let x0 = (substitute_variables_in_expr varmap e0) in
      (SREF (tn, deref, x0, s))
  | (ADDROF (tn, e0)) ->
      let x0 = (substitute_variables_in_expr varmap e0) in
      (ADDROF (tn, x0))
  | (SIZEOF (tn, t0)) ->
      e
  | (PROGN (tn, l)) -> (assert false)
  | (CONST_I _) ->
      e
  | (CONST_F _) ->
      e
  | (STRING _) ->
      e
  | (CLITERAL (tn, l)) -> (assert false)
  | (DSGN_AREF _) -> (assert false)
  | (ASMSEQUENCE _) -> (assert false)
  | (PHI _) -> (assert false)
  | (PSEUDO_MEMORY (tn, t0)) ->
      e
  | (INIT_FILL _) -> (assert false)
  | (PRAGMA s) -> (assert false)
  | (QUANTIFIER (s, vs, e0)) -> (assert false)
  | (PSEUDO_PROLOGUE) ->
      e
  | (PSEUDO_EPILOGUE) ->
      e
  | PSEUDO_ACSL_RESULT (tn, e0) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      PSEUDO_ACSL_RESULT (tn, x0)
    )
  | PSEUDO_ACSL_OLD (tn, e0, e1) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      let x1 = substitute_variables_in_expr varmap e1 in
      PSEUDO_ACSL_OLD (tn, x0, x1)
    )
  | ALL_FIELD (tn, e0) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      ALL_FIELD (tn, x0)
    )
  | PSEUDO_ACSL_UNIQUE (tn, e0) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      PSEUDO_ACSL_UNIQUE (tn, x0)
    )
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      let x1 = substitute_variables_in_expr varmap e1 in
      PSEUDO_ACSL_UNIQUEBLOCK (tn, x0, x1)
    )
  | PSEUDO_ACSL_VALID (tn, e0) -> (
      let x0 = substitute_variables_in_expr varmap e0 in
      PSEUDO_ACSL_VALID (tn, x0)
    ))

(* given an expression that appears in a post-condition,
   the following are to determine whether the expression has 
   the full effect in the caller side, i.e., could be transformed
   into an assertion or an assumption in the caller side *)

let rec full_effect_in_caller (expr : expr_t) : bool = (
    match expr with
    | NIL -> true
    | PSEUDO_NONDET_VALUE _ -> false
    | OPER (_, _, expr1, expr2, expr3) ->
        ((full_effect_in_caller expr1)
         && (full_effect_in_caller expr2)
         && (full_effect_in_caller expr3))
    | SETF (_, _, expr1, expr2) ->
        ((full_effect_in_caller expr1)
         & (full_effect_in_caller expr2))
    | CALL (_, func, params, _) ->
        List.for_all full_effect_in_caller params
    | VREF (_, _, level, _) ->
           (level == 1 || level == 2)
    | PREF _ -> true
    | SREF (_, _, expr, _) ->
        (full_effect_in_caller expr)
    | DSGN_AREF _ ->
        (assert false)
    | ADDROF (_, expr) ->
        (full_effect_in_caller expr)
    | ALL_FIELD (_, expr1) ->
        (full_effect_in_caller expr1)
    | SIZEOF _ -> true
    | PROGN _ ->
        (assert false)
    | CONST_I _
    | CONST_F _
    | STRING _ -> true
    | CLITERAL _
    | ASMSEQUENCE _ ->
        (assert false)
    | PHI _
    | PSEUDO_MEMORY _
    | INIT_FILL _ ->
        (assert false)
    | PRAGMA _
    | QUANTIFIER _ ->
        (assert false)
    | PSEUDO_PROLOGUE
    | PSEUDO_EPILOGUE ->
        (assert false)
    | PSEUDO_ACSL_RESULT _
    | PSEUDO_ACSL_OLD _
    | PSEUDO_ACSL_UNIQUE _
    | PSEUDO_ACSL_UNIQUEBLOCK _
    | PSEUDO_ACSL_VALID _ -> true)

let full_effect_in_caller (full_expr : full_expr_t) : bool =
    full_effect_in_caller full_expr.expr_node

let make_stmts_from_assume (ex : expr_t) : stmt_t list =
begin
    let tmp_var = VREF (INTEGER, (gensym "V"), 3, 0) in
    let name_of_var = 
        match tmp_var with
        | VREF (_, str, _, _) -> str
        | _ -> assert false
    in
    let defvar = 
        {
            decl_id = (name_of_var, 3);
            decl_tn = INTEGER;
            decl_scls = NILSCLS;
            decl_inits = 
                [SETF (INTEGER, SET, tmp_var, ex)];
            decl_params = [];
            decl_struct_defined_p = false;
            decl_specs = None;
            decl_location = None;                
        } in
    let st_1 = DEFVAR (defvar) in
    let const_1 = CONST_I (INTEGER, Num.Int(1), "1") in
    let new_oper = make_binary_operation EQ tmp_var const_1 in
    let st_2 = ASSUME (new_oper) in
    [st_1; st_2]      
end
	
let rec preprocess_assume_in_stmts (init_stmts : stmt_t list)
: stmt_t list =
begin
    List.flatten (
        List.map
            (fun a_st ->
                preprocess_assume a_st 
            )
            init_stmts
   )
end

and preprocess_assume (init_stmt : stmt_t) : stmt_t list =
begin
    match init_stmt with
    | DEFUN (decl, a_stmt) -> (
        let new_stmts = preprocess_assume a_stmt in
        match new_stmts with
        | [st] -> [DEFUN (decl, st)]
        | _ -> assert false
    )
    | DEFVAR _
    | STMT _
    | ASGN _
    | BREAK
    | CONTINUE
    | LABEL _
    | GOTO _
    | RETURN _
    | ASM _
    | NOTE _
    | ANNOTATION _
    | SPEC _
    | RESET _ -> [init_stmt]
    | CALLASGN (ex_1, stmts, exs_2) -> (
        let new_stmts = preprocess_assume_in_stmts stmts in    
        [CALLASGN (ex_1, new_stmts, exs_2)]
    ) 
    | BEGIN stmts -> (
        let new_stmts = preprocess_assume_in_stmts stmts in
        [BEGIN (new_stmts)]
    )
    | SWITCH (ex, stmts) -> (
        let new_stmts = preprocess_assume_in_stmts stmts in
        [SWITCH (ex, new_stmts)]
    )
    | DOWHILE (bo, ex, stmts) -> (
        let new_stmts = preprocess_assume_in_stmts stmts in
        [DOWHILE (bo, ex, new_stmts)]
    ) 
    | IF (ex, stmts_1, stmts_2) -> (
        let new_stmts_1 = preprocess_assume_in_stmts stmts_1 in
        let new_stmts_2 = preprocess_assume_in_stmts stmts_2 in
        [IF (ex, new_stmts_1, new_stmts_2)]
    )
    | FOR (stmts_1 , ex_1, ex_2, stmts_2) -> (
        let new_stmts_1 = preprocess_assume_in_stmts stmts_1 in
        let new_stmts_2 = preprocess_assume_in_stmts stmts_2 in
        [FOR (new_stmts_1, ex_1, ex_2, new_stmts_2)]
    )
    | ASSUME ex ->
        make_stmts_from_assume ex
end

let create_stmt_ofile (decls : stmt_t list)
                      (file_name : string)
                      (hash_name : string)
                      (f_pointer_field_hashtbl : (string * string, string) Hashtbl.t)
                      (contract_of_field_hashtbl : (string * string,
                                 (full_expr_t, full_expr_t) Sexpr.function_contract_t
                                  * expr_t list) Hashtbl.t)
: ofile_t =
begin
    let defs = 
        List.fold_left
            (fun curr_defs a_decl ->
                match a_decl with
                | DEFUN (decl, st) -> (
                    let (name_defun, _) = decl.decl_id in
                    (curr_defs) @ [(name_defun, a_decl)]
                )
                | DEFVAR (decl) -> (
                    let (name_defvar, _) = decl.decl_id in
                    (curr_defs) @ [(name_defvar, a_decl)]
                )
                | _ -> curr_defs
            ) [] decls in
    let ofile =
        {
            ofile_hash_name = hash_name;
            ofile_file_name = file_name;
            ofile_defs = ref defs;
            f_pointer_field_hashtbl = Hashtbl.copy f_pointer_field_hashtbl;
            contract_of_field_hashtbl = Hashtbl.copy contract_of_field_hashtbl;
        } in
    (ofile)
end

let is_result_expr (curr_expr : expr_t) : bool =
begin
    match curr_expr with
    | PSEUDO_ACSL_RESULT _ -> true
    | _ -> false
end

let rec get_result_expr (curr_expr : expr_t) : expr_t option =
begin
    if (is_result_expr curr_expr) then
        Some curr_expr
    else (
        match curr_expr with
        | CALL (_, func_var, args, _) ->
            get_result_exprs ([func_var] @ args)
        | NIL
        | VREF _
        | SIZEOF _
        | CONST_I _
        | CONST_F _
        | STRING _
        | PSEUDO_MEMORY _
        | PRAGMA _
        | PSEUDO_PROLOGUE
        | PSEUDO_EPILOGUE
        | PSEUDO_NONDET_VALUE _ -> None
        | OPER (_, _, expr1, expr2, expr3) ->
            get_result_exprs [expr1; expr2; expr3]
        | SETF (_, _, expr1, expr2)
        | DSGN_AREF (_, expr1, expr2)
        | PSEUDO_ACSL_UNIQUEBLOCK (_, expr1, expr2)
        | PSEUDO_ACSL_OLD (_, expr1, expr2) ->
            get_result_exprs [expr1; expr2]
        | PREF (_, _, expr, _)
        | SREF (_, _, expr, _)
        | ADDROF (_, expr)
        | INIT_FILL (_, expr)
        | PSEUDO_ACSL_RESULT (_, expr)
        | PSEUDO_ACSL_UNIQUE (_, expr) 
        | PSEUDO_ACSL_VALID (_, expr)
        | ALL_FIELD (_, expr) ->
            get_result_expr expr
        | PROGN (ty, sts) -> (* DODO process sts (Cexpr.stmt_t list) later *)
            None
        | CLITERAL (ty, init) ->
            get_result_init init
        | ASMSEQUENCE exprs ->
            get_result_exprs exprs
        | PHI (ty, phi_lst) -> (
            let (_, exprs) = List.split phi_lst in
            get_result_exprs exprs
        ) 
        | QUANTIFIER (str, exprs, expr) ->
            get_result_exprs (exprs @ [expr])
    )
end

and get_result_exprs (curr_exprs : expr_t list) : expr_t option =
begin
    match curr_exprs with
    | [] -> None
    | head :: tails -> (
        let ex_op = get_result_expr head in
        match ex_op with
        | None -> get_result_exprs tails
        | _ -> ex_op
    )  
end

and get_result_init (init : initializers_t) : expr_t option =
begin
    match init with
    | INIT_EXPR expr ->
        get_result_expr expr
    | INIT_LIST inits ->
        get_result_inits inits
    | INIT_DSGN (exprs, init) -> (
        let ex_op = get_result_exprs exprs in
        match ex_op with
        | None -> get_result_init init
        | _ -> ex_op   
    )
end

and get_result_inits (curr_inits : initializers_t list)
: expr_t option =
begin
    match curr_inits with
    | [] -> None
    | head :: tails -> (
        let ex_op = get_result_init head in
        match ex_op with
        | None -> get_result_inits tails
        | _ -> ex_op
    )  
end

let is_local (decls : stmt_t list) : bool =
begin
    List.exists
        (fun a_stmt ->
            let header =
                match a_stmt with
                | DEFUN (decl, st) -> decl
                | DEFVAR (decl) -> decl                       
                | _ -> assert false in 
            match header.decl_scls with
            | STATIC -> true
            | _ -> false
        ) decls
end
